{******************************************}
{                                          }
{          FastReport VCL/FMX/LCL          }
{              Core Library                }
{                                          }
{         Copyright (c) 1998-2022          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

{$IFNDEF FMX}
unit frCoreClasses;
{$ENDIF}

{$I frVer.inc}

interface

uses
{$IFNDEF FMX}
  SysUtils, Types, Classes, Controls,
  frUnicodeUtils, frCore, frTypes, frUtils;
{$ELSE}
  System.SysUtils, System.Types, System.Classes,
  FMX.Types,
  FMX.frUnicodeUtils, FMX.frCore, FMX.frTypes, FMX.frUtils;
{$ENDIF}

type
  EfrException = class(Exception);

  TfrPointerList = array of Pointer;

  { TfrMap }

  TfrMap = class
  protected type
    THashProc = function(const AKey: Pointer): Integer;
    TAllocProc = function(const AValue: Pointer): Pointer;
    TDisposeProc = procedure(const AValue: Pointer);
    TSameKeyProc = function(const AKey1, AKey2: Pointer): Boolean;
    PItem = ^TItem;
    TItem = record
      Key: Pointer;
      Hash: Integer;
      Value: Pointer;
      Next: PItem;
    end;
    TItems = array of PItem;
  strict private const
    FDefaultItemsLength = 691;
    function GetItem(const AKey: Pointer): Pointer;
    procedure SetItem(const AKey, AValue: Pointer);
  strict private
    FAllocKeyProc: TAllocProc;
    FAllocValueProc: TAllocProc;
    FCount: Cardinal;
    FDisposeKeyProc: TDisposeProc;
    FDisposeValueProc: TDisposeProc;
    FItems: TItems;
    FHashProc: THashProc;
    FKeySize: Integer;
    FSameKeyProc: TSameKeyProc;
    FValueSize: Integer;
  protected
    function AllocKey(const AKey: Pointer): Pointer; virtual;
    function AllocValue(const AValue: Pointer): Pointer; virtual;
    procedure DoAdd(const AKey, AValue: Pointer; ACheckDuplicates: Boolean); virtual;
    procedure DisposeKey(AKey: Pointer); virtual;
    procedure DisposeItem(AItem: PItem); virtual;
    procedure DisposeValue(AValue: Pointer); virtual;
    function Hash(const AKey: Pointer): Integer;
    function NewItem(const AKey, AValue: Pointer; AHash: Cardinal): PItem;
    function SameKey(const AKey1, AKey2: Pointer): Boolean;

    property AllocKeyProc: TAllocProc read FAllocKeyProc write FAllocKeyProc;
    property AllocValueProc: TAllocProc read FAllocValueProc write FAllocValueProc;
    property DisposeKeyProc: TDisposeProc read FDisposeKeyProc write FDisposeKeyProc;
    property DisposeValueProc: TDisposeProc read FDisposeValueProc write FDisposeValueProc;
    property KeySize: Integer read FKeySize;
    property SameKeyProc: TSameKeyProc read FSameKeyProc write FSameKeyProc;
    property ValueSize: Integer read FValueSize;
  public
    constructor Create(AKeySize, AValueSize: Integer; AHashProc: THashProc);
    destructor Destroy; override;
    procedure Add(const AKey, AValue: Pointer);
    procedure AddOrSetValue(const AKey, AValue: Pointer);
    procedure Clear;
    function ContainsKey(const AKey: Pointer): Boolean;
    function Keys: TfrPointerList;
    function Remove(const AKey: Pointer): Boolean;
    function TryGetValue(const AKey: Pointer; out AValue: Pointer): Boolean;

    property Items[const AKey: Pointer]: Pointer read GetItem write SetItem; default;
    property Count: Cardinal read FCount;
  end;

  { TfrObjectMap }

  TfrObjectMap = class(TfrMap)
  strict private
    FOwnObjects: Boolean;
  protected
    function AllocValue(const AValue: Pointer): Pointer; override;
    procedure DisposeValue(AValue: Pointer); override;
  public
    constructor Create(AOwnObjects: Boolean; AKeySize: Integer; AHashProc: TfrMap.THashProc); reintroduce;
  end;

  { TfrCustomDictionary }

  TfrCustomDictionary = class
  strict private
    FMap: TfrMap;
    function GetCount: Integer;
  protected
    function CreateMap: TfrMap; virtual; abstract;

    property Map: TfrMap read FMap;
  public
    constructor Create;
    destructor Destroy; override;

    procedure Clear;

    property Count: Integer read GetCount;
  end;

  { TfrNamedDictionary }

  TfrNamedDictionary = class abstract(TfrCustomDictionary)
  private
    FKeyCaseSensetive: Boolean;
  protected
    function CreateMap: TfrMap; override;
    function GetValueSize: Integer; virtual; abstract;
  public
    constructor Create(AKeyCaseSensetive: Boolean = True); reintroduce;

    function ContainsKey(const AKey: string): Boolean;
    function Keys: TStringDynArray;
    function Remove(const AKey: string): Boolean;
  end;

  { TfrNamedObjectDictionary }

  TfrNamedObjectDictionary = class(TfrNamedDictionary)
  strict private
    FOwnObjects: Boolean;
    function GetItem(const AKey: string): TObject;
    procedure SetItem(const AKey: string; const Value: TObject);
  protected
    function CreateMap: TfrMap; override;
    function GetValueSize: Integer; override;
  public
    constructor Create(AOwnObjects: Boolean = True; AKeyCaseSensetive: Boolean = True); reintroduce;

    procedure Add(const AKey: string; const AValue: TObject);
    procedure AddOrSetValue(const AKey: string; const AValue: TObject);
    function TryGetValue(const AKey: string; out AValue: TObject): Boolean;

    property Items[const AKey: string]: TObject read GetItem write SetItem; default;
  end;

  { TfrStringDictionary }

  TfrStringDictionary = class(TfrNamedDictionary)
  strict private
    function GetItem(const AKey: string): string;
    procedure SetItem(const AKey, AValue: string);
  protected
    function CreateMap: TfrMap; override;
    function GetValueSize: Integer; override;
  public
    procedure Add(const AKey, AValue: string);
    procedure AddOrSetValue(const AKey, AValue: string);
    function TryGetValue(const AKey: string; out AValue: string): Boolean;

    property Items[const AKey: string]: string read GetItem write SetItem; default;
  end;

  { TfrInterfacedObject }

  TfrInterfacedObject = class(TObject, IUnknown)
  protected
  {$REGION 'IUnknown'}
    {$IFDEF FPC}
      function QueryInterface({$IFDEF FPC_HAS_CONSTREF}constref{$ELSE}const{$ENDIF} iid : tguid;out obj) : longint;{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};
    {$ELSE}
      function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    {$ENDIF}
    function _AddRef: Integer; {$IFDEF FPC}{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};{$ELSE} stdcall;{$ENDIF}
    function _Release: Integer; {$IFDEF FPC}{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};{$ELSE} stdcall;{$ENDIF}
  {$ENDREGION}
  end;

  { TfrList }

  TfrList = class
  strict private
    FCapacity: Integer;
    FCount: Integer;
    FList: PfrPointerArray;

    procedure InternalDelete(AIndex: Integer);
  protected
    function GetItem(AIndex: Integer): Pointer;
    procedure Grow;
    procedure Notify(const AItem: Pointer; AAction: TListNotification); virtual;
    procedure SetCapacity(AValue: Integer);
    procedure SetCount(AValue: Integer);
    procedure SetItem(AIndex: Integer; AItem: Pointer);
  public
  {$IF NOT DEFINED(FPC) AND NOT DEFINED(DELPHI17)}
    type
      TDirection = (FromBeginning, FromEnd);
  {$IFEND}

    destructor Destroy; override;
    function Add(AItem: Pointer): Integer;
    procedure Assign(ASource: TList); overload;
    procedure Assign(ASource: TfrList; AOperator: TListAssignOp = laCopy); overload;
    procedure Clear; virtual;
    procedure Delete(AIndex: Integer);
    class procedure Error(const Msg: String; Data: Integer); overload; virtual;
    class procedure Error(Msg: PResStringRec; Data: Integer); overload;
    procedure Exchange(AIndex1, AIndex2: Integer);
    function Expand: TfrList;
    function Extract(AItem: Pointer): Pointer; inline;
    function ExtractItem(AItem: Pointer; ADirection: TDirection): Pointer;
    function First: Pointer; inline;
    function IndexOf(AItem: Pointer; ADirection: TDirection = TDirection.FromBeginning): Integer;
    procedure Insert(AIndex: Integer; AItem: Pointer);
    function Last: Pointer;
    procedure Move(ACurIndex, ANewIndex: Integer);
    procedure Pack;
    function Remove(AItem: Pointer): Integer; inline;
    function RemoveItem(AItem: Pointer; ADirection: TDirection): Integer;
    procedure Reverse;
    procedure Sort(ACompare: TfrSortPointerCompare); overload;
    procedure Sort(ACompare: TfrListSortPointerCompare); overload;
  {$IFNDEF FPC}
    procedure SortList(const ACompare: TfrListSortCompareFunc);
  {$ENDIF}

    property Capacity: Integer read FCapacity write SetCapacity;
    property Count: Integer read FCount write SetCount;
    property Items[Index: Integer]: Pointer read GetItem write SetItem; default;
    property List: PfrPointerArray read FList;
  end;

  /// <summary>
  /// This interface used to pass standard list as a read only reference.
  /// It does not allow modification of a list
  /// </summary>
  IfrReadonlyList = interface ['{B677D5AF-26A7-4CAE-9D5B-9EEE102BA78C}']
    function Get(Index: TfrListInteger): Pointer;
    function GetCount: TfrListInteger;
    function IndexOf(const Item: Pointer): TfrListInteger;
    property Count: TfrListInteger read GetCount;
    property Items[Index: TfrListInteger]: Pointer read Get; default;
  end;

  /// <summary>
  ///   This class represents standard list with IfrReadonlyList interface implementation
  /// </summary>
  TfrReadOnlyList = class(TList, IfrReadonlyList)
  protected
  {$REGION 'IUnknown'}
    {$IFDEF FPC}
      function QueryInterface({$IFDEF FPC_HAS_CONSTREF}constref{$ELSE}const{$ENDIF} iid : tguid;out obj) : longint;{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};
    {$ELSE}
      function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    {$ENDIF}
    function _AddRef: Integer; {$IFDEF FPC}{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};{$ELSE} stdcall;{$ENDIF}
    function _Release: Integer; {$IFDEF FPC}{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};{$ELSE} stdcall;{$ENDIF}
  {$ENDREGION}
  {$REGION 'IfrReadonlyList'}
    function Get(Index: TfrListInteger): Pointer;
    function GetCount: TfrListInteger;
    function IndexOf(const Item: Pointer): TfrListInteger;
  {$ENDREGION}
  end;

  { TfrCustomComponent }

  TfrCustomComponent = class(TComponent);
  TfrCollection = class;
  { TfrCollectionItem }

  TfrCollectionItem = class(TCollectionItem)
  protected
    FUniqueIndex: Integer;
    FIsInherited: Boolean;
    function GetInheritedName: String; virtual;
    procedure SetInheritedName(const Value: String); virtual;
  public
    constructor Create(ACollection: TCollection); override;
    procedure CreateUniqueIName(ACollection: TCollection);
    procedure WriteProperties(Writer: TObject; Ancestor: TfrCollection; Owner: TComponent); virtual;
    procedure ReadProperties(PropName: String; Reader: TObject; Ancestor: TfrCollection); virtual;
    function IsUniqueNameStored: Boolean; virtual;
    property IsInherited: Boolean read FIsInherited write FIsInherited;
    property InheritedName: String read GetInheritedName write SetInheritedName;
    property UniqueIndex: Integer read FUniqueIndex write FUniqueIndex;
  end;

  { TfrCollection }

  TfrCollection = class(TCollection)
  public
    procedure SerializeProperties(Writer: TObject; Ancestor: TfrCollection; Owner: TComponent); virtual;
    procedure DeserializeProperties(PropName: String; Reader: TObject; Ancestor: TfrCollection); virtual;
    function FindByName(Name: String): TfrCollectionItem; virtual;
  end;

  { TfrOwnObjList }

  TfrOwnObjList = class(TList)
  private
    FOwnsObjects: Boolean;
  protected
    function GetItem(Index: TfrListInteger): TObject;
  public
    constructor Create(AOwnsObjects: Boolean = True);
    procedure Clear; override;
    procedure AddNonZero(Item: Pointer);
    procedure Assign(SourceList: TList);

    property Items[Index: TfrListInteger]: TObject read GetItem; default;
  end;

  { TfrRegistredUnitRepository }

  TfrRegistredUnitRepository = class
  strict private
    FDictionary: TfrNamedObjectDictionary;
  public
    constructor Create;
    destructor Destroy; override;

    procedure RegisterUnit(const AClass: TClass; const AUnit: string);
    procedure UnregisterUnit(const AClass: TClass; const AUnit: string);
    function GetUnits(const AClass: TClass): TStrings;
  end;

procedure frRegisterFrameworkComponentClass(AComponentClass: TPersistentClass); overload;
procedure frRegisterFrameworkComponentClass(AComponentClasses: array of TPersistentClass); overload;
function frRegistredUnitRepository: TfrRegistredUnitRepository;

implementation
end.
