
{******************************************}
{                                          }
{          FastReport VCL/FMX/LCL          }
{               XML document               }
{                                          }
{         Copyright (c) 1998-2022          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

{$IFNDEF FMX}
unit frXML;
{$ENDIF}

interface

{$I frVer.inc}

uses
{$IFNDEF FPC}
  {$IFDEF MSWINDOWS}
    Windows,
  {$ENDIF}
{$ENDIF}
{$IFNDEF FMX}
  {$IFNDEF FPC} Variants, {$ELSE} LazFileUtils, LCLProc, {$ENDIF}
  Types, Math, SysUtils, Classes, frCore;
{$ELSE}
  System.Classes, System.SysUtils, System.Types, System.IOUtils, System.Math, System.Variants,
  FMX.frCore;
{$ENDIF}

type
  TfrInvalidXMLException = class(Exception);

  TfrXMLItem = class(TObject)
  private
    FData: Pointer;              { optional item data }
{$IFDEF FR_XML_LOW_MEM_USAGE}
    FHiOffset: Byte;             { hi-part of the offset }
    FLoOffset: Integer;          { lo-part of the offset }
{$ELSE}
    FOffset: Int64;
{$ENDIF}
    FLoaded: Boolean;            { item is loaded, no need to call LoadItem }
    FModified: Boolean;          { item is modified (used by preview designer) }
    FName: string;               { item name }
    FParent: TfrXMLItem;        { item parent }
    FText: string;               { item attributes }
    FUnloadable: Boolean;
    FValue: string;              { item value <item>Value</item> }
    function GetCount: Integer;
    function GetItems(Index: Integer): TfrXMLItem;
    function GetOffset: Int64;
    procedure SetOffset(const Value: Int64);
    function GetProp(APropertyName: string): string;
    procedure SetProp(APropertyName: string; const AValue: string);
    function GetPropAsFloat(APropertyName: string): Double;
    procedure SetPropAsFloat(APropertyName: string; const Value: Double);
    function GetPropAsBoolean(APropertyName: string): Boolean;
    procedure SetPropAsBoolean(APropertyName: string; const Value: Boolean);
    function GetPropAsInteger(APropertyName: string): Integer;
    procedure SetPropAsInteger(APropertyName: string; const Value: Integer);
    function GetPropAsDateTime(APropertyName: string): TDateTime;
    function GetPropAsInt64(APropertyName: string): Int64;
    procedure SetPropAsDateTime(APropertyName: string; const Value: TDateTime);
    procedure SetPropAsInt64(APropertyName: string; const Value: Int64);
  protected
    FItems: TList;               { subitems }
  public
    constructor Create;
    destructor Destroy; override;
    procedure AddItem(Item: TfrXMLItem); virtual;
    procedure Clear;
    procedure InsertItem(Index: Integer; Item: TfrXMLItem);

    function Add: TfrXMLItem; overload;
    function Add(Name: string): TfrXMLItem; overload;
    procedure DeleteItem(const AIndex: Integer); overload;
    procedure DeleteItem(const AIndex: String); overload;
    function Find(const Name: string): Integer;
    function FindItem(const Name: string): TfrXMLItem;
    function IndexOf(Item: TfrXMLItem): Integer;
    function PropExists(const Index: string): Boolean;
    function Root: TfrXMLItem;
    procedure DeleteProp(const Index: string);
    procedure SetData(const AName, AText, AValue: string); virtual;

    property Count: Integer read GetCount;
    property Data: Pointer read FData write FData;
    property Items[Index: Integer]: TfrXMLItem read GetItems; default;
    property Loaded: Boolean read FLoaded;
    property Modified: Boolean read FModified write FModified;
    property Name: string read FName write FName;
{ offset is the position of the item in the tempstream. This parameter is needed
  for dynamically loading large files. Items that can be loaded on-demand must
  have Unloadable = True (in run-time) or have 'ld="0"' parameter (in the file) }
    property Offset: Int64 read GetOffset write SetOffset;
    property Parent: TfrXMLItem read FParent;
    property Prop[APropertyName: string]: string read GetProp write SetProp; // for backward capatibility
    property PropAsBoolean[APropertyName: string]: Boolean read GetPropAsBoolean write SetPropAsBoolean;
    property PropAsDateTime[APropertyName: string]: TDateTime read GetPropAsDateTime write SetPropAsDateTime;
    property PropAsFloat[APropertyName: string]: Double read GetPropAsFloat write SetPropAsFloat;
    property PropAsInt[APropertyName: string]: Integer read GetPropAsInteger write SetPropAsInteger;
    property PropAsInt64[APropertyName: string]: Int64 read GetPropAsInt64 write SetPropAsInt64;
    property PropAsText[APropertyName: string]: string read GetProp write SetProp;
//    property PropAsFloat[Index: string]: string read GetPropAsFloat write SetPropAsFloat;
    property Text: string read FText write FText;
    property Unloadable: Boolean read FUnloadable write FUnloadable;
    property Value: string read FValue write FValue;
  end;

  TfrXMLDocument = class(TObject)
  private
    FAutoIndent: Boolean;        { use indents when writing document to a file }
    FTempDir: string;            { folder for temporary files }
    FTempFile: string;           { tempfile name }
    FTempFileCreated: Boolean;   { tempfile has been created - need to delete it }
    procedure CreateTempFile;
  protected
    FRoot: TfrXMLItem;          { root item }
    FTempStream: TStream;        { temp stream associated with tempfile }
    FOldVersion: Boolean;
    procedure DeleteTempFile;
    function GetAsText: String; virtual;
    procedure SetAsText(const AValue: String); virtual;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    procedure LoadItem(Item: TfrXMLItem);
    procedure UnloadItem(Item: TfrXMLItem);
    procedure SaveToStream(Stream: TStream);
    procedure LoadFromStream(Stream: TStream; AllowPartialLoading: Boolean = False);
    procedure SaveToFile(const FileName: string);
    procedure LoadFromFile(const FileName: string);

    property AutoIndent: Boolean read FAutoIndent write FAutoIndent;
    property Root: TfrXMLItem read FRoot;
    property TempDir: string read FTempDir write FTempDir;
    property OldVersion: Boolean read FOldVersion;
    property AsText: String read GetAsText write SetAsText;
  end;

{ TfrXMLReader and TfrXMLWriter are doing actual read/write to the XML file.
  Read/write process is buffered. }

  TfrXMLReader = class(TObject)
  private
    FBuffer: PAnsiChar;
    FBufPos: Integer;
    FBufEnd: Integer;
    FPosition: Int64;
    FSize: Int64;
    FStream: TStream;
    FOldFormat: Boolean;
    procedure SetPosition(const Value: Int64);
    procedure ReadBuffer;
  protected
    procedure ReadItem(var {$IFDEF DELPHI12}NameS{$ELSE}Name{$ENDIF}, Text, Value: string); overload;
    procedure ReadItem(var Name, Text: string); overload;
    function EndOfStream: Boolean;
    function ReadFromBuffer: Integer;
  public
    constructor Create(Stream: TStream);
    destructor Destroy; override;
    procedure RaiseException;
    procedure ReadHeader;
    procedure ReadRootItem(AItem: TfrXMLItem; AReadChildren: Boolean = True);
    property Position: Int64 read FPosition write SetPosition;
    property Size: Int64 read FSize;
    property OldFormat: Boolean read FOldFormat;
  end;

  TfrXMLWriter = class(TObject)
  private
    FAutoIndent: Boolean;
    FBuffer: AnsiString;
    FStream: TStream;
    FTempStream: TStream;
    procedure FlushBuffer;
    procedure WriteLn(const s: AnsiString);
    procedure WriteItem(Item: TfrXMLItem; Level: Integer = 0);
  public
    constructor Create(Stream: TStream);
    procedure WriteHeader;
    procedure WriteRootItem(RootItem: TfrXMLItem);
    property TempStream: TStream read FTempStream write FTempStream;
    property AutoIndent: Boolean read FAutoIndent write FAutoIndent;
  end;

  { TfrValuedXMLReader }

  TfrValuedXMLReader = class(TfrXMLReader)
  protected
    function IsLastSlash(const InSt: String): Boolean;
    function IsFirstSlash(const InSt: String): Boolean;
    function IsFirstNumberSign(const InSt: String): Boolean;
    procedure ReadValuedItem(out {$IFDEF Delphi12}NameS, ValueS{$ELSE}Name, Value{$ENDIF}, Text: String); virtual;
    procedure ProcessSecondLeftBrocket; virtual; abstract;
    function CreateItem: TfrXMLItem; virtual;
    procedure AddChar(var st: AnsiString; var stPos: Integer; ch: AnsiChar);
  public
    function IsReadValuedXMLRootItem(Item: TfrXMLItem): Boolean;
    function IsReadValuedXMLItem(Item: TfrXMLItem): Boolean;
  end;

  { TfrValuedXMLDocument }

  TfrValuedXMLDocument = class(TfrXMLDocument)
  protected
    FValuedXMLStream: TStream;
    FValuedXMLStreamReader: TfrValuedXMLReader;

    procedure ReadXMLHeader; virtual; abstract;
    function CreateReader: TfrValuedXMLReader; virtual; abstract;
  public
    procedure InitValuedXMLFile(const FileName: String);
    procedure DoneValuedXMLFile;
    function IsReadItem(Item: TfrXMLItem): Boolean;
  end;

{ StrToXML changes '<', '>', '"', cr, lf symbols to its ascii codes }
function frxStrToXML(const s: string): string;

{ ValueToXML convert a value to the valid XML string }
function frxValueToXML(const Value: Variant): string;

{ XMLToStr is opposite to StrToXML function }
function frxXMLToStr(const s: string): string;

function frXMLStrToWStr(const AValue: string; AOldVersion: Boolean): WideString;
//function frUTF8Decode(const AValue: ): WideString;

implementation
end.

