
{******************************************}
{                                          }
{             FastReport VCL               }
{         RaveReport -> FastReport         }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit ConverterRR2FR;

interface

uses

  { Delphi }

  Classes,
  SysUtils,
  Graphics,
  Printers,
  Forms,
  FileCtrl,
  Dialogs,
  StdCtrls,
  Controls,
  ComCtrls,
  frxBaseForm,
  { FastReport }

  frCore,
  frBaseGraphicsTypes,
  frxClass,
  frxBarCode,
  frxBarcod,
  frxADOComponents,
  frxBDEComponents,
  frxDBXComponents,
  frxIBXComponents,
  frxCustomDB,

  { RaveReport }

  RpDefine,
  RpRave,
  RpBase,
  RpSystem,
  RpBars,
  RvClass,
  RvProj,
  RvCsRpt,
  RvDefine,
  RvCsStd,
  RvCsBars,
  RvCsDraw,
  RvCsData,
  RvDatabase,
  RvDriverDataView;

type

  TRaveLoaderForm = class(TfrxBaseForm)

    lbTop: TLabel;
    pb: TProgressBar;
    bt: TButton;
    lbBottom: TLabel;

    procedure btClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);

  private

    FRunning: Boolean;
    FPage: string;
    FReport: string;



    procedure Stop;
    procedure UpdateLabel;
    procedure SetReport(const s: string);
    procedure SetPage(const s: string);
    procedure SetItemsCount(n: Integer);
    procedure SetComponentName(const s: string);
    function IsRunning: Boolean;

  public

    property Report: string write SetReport;
    property Page: string write SetPage;
    property Component: string write SetComponentName;
    property ItemsCount: Integer write SetItemsCount;

    { It's true if a user hasn't pressed the 'Cancel' button. }

    property Running: Boolean read IsRunning;

    { It's called when an item has been processed. }

    procedure NextItem;

    procedure Reset;

  end;

  { This function accepts two components:

      rc    Some Rave component that need be converted to an FR component.
      fc    Some FR component that's currently the parent. For example,
            fc can be a TfrxPage and then the function must create an FR component
            that'll be a child of fc and will be an FR analogue of rc.

    The function creates an FR analogue of rc Rave's component,
    adds this component to children of fc and returns the created component. }

  TRaveAddComponent = function(rc: TRaveComponent; fc: TfrxComponent):
    TfrxComponent of object;

  { This class contains Load function that loads a Rave report into FR.
    This Load function is needed for FR importing subsystem. }

  TRaveLoader = class(TObject)

    FOnUnknownComponent: TRaveAddComponent;
    FForm: TRaveLoaderForm;
    FLoadQueries: Boolean;
    FIsFirstBand: Boolean;
    FBandPos: real;

    { Converts the given Rave component to an FR component and
      does the same thing for all its children. }

    procedure EnumComponents(rc: TRaveComponent; fc: TfrxComponent);

    function HAlign(pj: TPrintJustify): TfrxHAlign;
    function FrameStyle(ps: TPenStyle): TfrxFrameStyle;

    { Rave designer allows to set the color of the brush fill pattern,
      but FR designer doesn't allow that. All brush fills will be black. }

    function BrushStyle(fs: TRaveFillStyle): TBrushStyle;

    procedure AssignBounds(rc: TRaveControl; fc: TfrxComponent);
    procedure AssignStyle(rc: TRaveSurface; fc: TfrxView);

    { Routine AddXXX(rc, fc) creates an FR equivalent of the given Rave rc
      component and makes the created component a child of the given FR fc
      component. If rc is a container, then the routine returns the newly
      created FR component. Otherwise, the routine returns fc. }

    function AddPage(rc: TRavePage; fc: TfrxComponent): TfrxComponent;
    function AddMemo(rc: TRaveMemo; fc: TfrxComponent): TfrxComponent;
    function AddText(rc: TRaveText; fc: TfrxComponent): TfrxComponent;
    function AddCustomText(rc: TRaveCustomText; fc: TfrxComponent): TfrxComponent;
    function AddRect(rc: TRaveRectangle; fc: TfrxComponent): TfrxComponent;
    function AddEllipse(rc: TRaveEllipse; fc: TfrxComponent): TfrxComponent;
    function AddBarCode(rc: TRaveBaseBarCode; fc: TfrxComponent): TfrxComponent;
    function AddImage(rc: TRaveGraphicImage; fc: TfrxComponent): TfrxComponent;
    function AddDataBand(rc: TRaveDataBand; fc: TfrxComponent): TfrxComponent;
    function AddBand(rc: TRaveBand; fc: TfrxComponent): TfrxComponent;
    function AddRegion(rc: TRaveRegion; fc: TfrxComponent): TfrxComponent;
    function AddDatabase(rc: TRaveDatabase; fc: TfrxComponent): TfrxComponent;
    function AddDataview(rc: TRaveDriverDataView; fc: TfrxComponent): TfrxComponent;
    function AddStub(rc: TRaveControl; fc: TfrxComponent): TfrxComponent;
    function AddUnknown(rc: TRaveComponent; fc: TfrxComponent): TfrxComponent;

    { There's a bug in Rave v5.0 - it doesn't save the style of lines to .rav files. }

    function AddLine(rc: TRaveLine; fc: TfrxComponent): TfrxComponent;

    { Returns the Tag of the component if it's not nil,
      or returns the default value if it's nil. }

    function GetTag(c: TComponent; default: Integer = 0): Integer;

    { Every string from a Rave report goes through this function (String Gate).
      This function simply returns its result. }

    function SG(const s: string): string; inline;

  public

    constructor Create;
    destructor Destroy; override;

    function Load(Report: TfrxReport; Stream: TStream): Boolean;

    { While analysing a Rave report, the import may find a component that
      is unknown to it. In this case it can call a handler for this situation. }

    property OnUnknownComponent: TRaveAddComponent
      read FOnUnknownComponent write FOnUnknownComponent;

    { Specifies whether to load SQL queries or not.
      By default this option is disabled, since some Rave builds corrupts
      memory while processing queries. }

    property LoadQueries: Boolean read FLoadQueries write FLoadQueries; {default False}

  end;

{ Rave classes hierarchy:

  TComponent - belongs to VCL
    TRaveComponent
      TRaveDataField
        TRaveStringField
        TRaveDateTimeField
        TRaveFloatField
        TRaveFormattedField
          TRaveIntegerField

      TRaveContainerComponent
        TRaveProjectItem
          TRavePage
          TRaveReport
          TRaveDataObject
            TRaveDatabase
            TRaveBaseDataView
              TRaveDriverDataView

          TRaveProjectManager

      TRaveControl
        TRaveBaseImage
          TRaveGraphicImage
            TRaveBitmap
            TRaveMetafile

        TRaveContainerControl
          TRaveBand
            TRaveIterateBand
              TRaveDataBand

          TRaveRegion
          TRaveSection
            TRaveDataMirrorSection

        TRaveBaseBarCode
        TRaveCustomText
          TRaveText
            TRaveDataText
            TRaveCalcText

          TRaveMemo
            TRaveDataMemo

        TRaveGraphicBase
          TRaveLine
            TRaveHLine
            TRaveVLine

          TRaveSurface
            TRaveRectangle
              TRaveSquare

            TRaveEllipse
              TRaveCircle
}

{ This routine adds to FR the ability to open rav files.
  After the returned TRaveLoader is no longer needed,
  it must be deleted by its Free method. }

function RegisterLoader: TRaveLoader;

implementation
end.
