
{******************************************}
{                                          }
{             FastReport VCL               }
{            BIFF8 Writing API             }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

{ This module provides API for writing BIFF
  streams. This format is documented in
  MSDN [MS-XLS] section.

  The only way to store BIFF2-BIFF4 documents is
  to write data to a simple stream file. Since BIFF5
  the documents can be stored as compound document files:
  see frxCBFF.pas }

unit frxBIFF;

interface

{$I frx.inc}

uses
  Windows,
  Graphics,
  SysUtils,
  Classes,
  Contnrs,
  frCore,
  frTypes,
  frxUtils,
  frxStorage,
  frxEscher,
  frxCrypto
{$IFDEF DELPHI16}
, System.UITypes
{$ENDIF}
;

type
  TBiffPaperSize = Word;
  TBiffFormatIndex = LongInt;
  TBiffStreamKind = LongInt;    // See bk*** values
  TBiffRecId = Word;            // BiffId*** values

const

  { BIFF8 allows the maximium record
  length of 8228 bytes including the record
  header (4 bytes) }

  BiffMaxRecLen       = $2020;

  { The maximum number of blocks in a MERGEDCELLS record.
    Required by [MS-XLS] 2.4.168 }

  BiffMaxMrgCellsNum  = 1026;

  //
  // Rows in BIFF8 are grouped into blocks.
  //

  BiffRowBlockSz      = 32;

  //
  // BIFF8 palette size.
  // This size includes the first 8 built-in colors.
  //

  BiffPaletteSize     = $40;

  { RRD table contains unique identifiers
    for all sheets in a workbook. If the
    count of sheets is greater that BiffRrdMaxCount
    then the RRD table is not emitted. }

  BiffRrdMaxCount     = $1010;

  //
  // BIFF8 allows cells with coordinates within
  // the following bounds:
  //
  //  row = 0 .. BiffMaxRow
  //  col = 0 .. BiffMaxCol
  //

  BiffMaxRow          = $fffe;
  BiffMaxCol          = $fe;

  //
  // Number formats are stored in a table.
  // The first 164 items are reserved and
  // define special widely used formats.
  // The first user format can begin from 164.
  //

  BiffUserFormat      = 164;

  //
  // BOF records have a field which specifies
  // a type of a stream which they start.
  // TBiffStreamKind lists all possible stream types.
  //

  bkWBGlobals         = 5;
  bkVBModule          = 6;
  bkSheet             = $10;
  bkChart             = $20;
  bkMacro             = $40;
  bkWorkspace         = $100;

  //
  // Font options
  //

  foBold              = $1;
  foItalic            = $2;
  foUnderline         = $4;
  foStruckOut         = $8;
  foOutline           = $10;
  foShadow            = $20;
  foCondense          = $40;
  foExtended          = $80;

  //
  // Font underline values
  //

  fuNone              = 0;
  fuSingle            = 1;
  fuDouble            = 2;
  fuSingleAcc         = $21;
  fuDoubleAcc         = $22;

  //
  // Normally, font weight value lies
  // in range 100..1000
  //

  fwNormal            = 400;
  fwBold              = 700;

  //
  // XF cell protection
  //

  xftpCellLocked      = $1;
  xftpHidden          = $2;
  xftpStyle           = $4;

  //
  // XF used attributes flags
  //

  BiffXfuaNumber      = $01;
  BiffXfuaFont        = $02;
  BiffXfuaText        = $04;
  BiffXfuaBorders     = $08;
  BiffXfuaBg          = $10;
  BiffXfuaCellProt    = $20;
  BiffXfuaAll         = $3f;

  //
  // Pattern style
  //

  psNone              = 0;
  psSolid             = 1;
  psChess             = 2;
  psHorThick          = 5;
  psVerThick          = 6;
  psChessThick        = 9;
  psHor               = $B;
  psVer               = $C;
  psDiagBack          = $D;
  psDiag              = $E;
  psCross             = $F;
  psCrossDiag         = $10;

  //
  // Window options
  //

  woFormula           = $1;
  woGrid              = $2;
  woHeaders           = $4;
  woFreezePanes       = $8;
  woShowZeros         = $10;
  woAutoGridCol       = $20;
  woColRTL            = $40;
  woOutline           = $80;
  woNoSplits          = $100;
  woSelected          = $200;
  woActive            = $400;
  woPageBreak         = $800;

  //
  // Sheet type
  //

  skWorksheet         = 0;
  skChart             = 2;
  skVB                = 6;

  //
  // BIFF keeps all format strings in a single list.
  // Several entries in this list are reserved.
  // A few indexes to these reserved formats are listed here.
  //

  BiffFmtGeneral      = 0;
  BiffFmtFixedPoint   = 2;      // 0.00
  BiffFmtThSep        = 4;      // #,##0.00
  BiffFmtCurrency     = 5;      // "$"#,##0_);("$"#,##0)
  BiffFmtDateTime     = 22;//32;   // M/D/YY h:mm

  //
  // Flags for options of WINDOW2 object
  //

  BiffWoFormulas      = 1;      // show formulas; clear = show results of formulas
  BiffWoGridLines     = 2;      // show grid lines
  BiffWoHeaders       = 4;      // show sheet headers
  BiffWoFrozen        = 8;      // panes are frozen
  BiffWoZeros         = $10;    // show zeros; clear = show zeros as empty cells
  BiffWoAutoGridColor = $20;    // automatic grid line color; clear = manual
  BiffWoColumnsRTL    = $40;    // columns from right to left; clear = left to right
  BiffWoOutline       = $80;    // show outline symbols
  BiffWoNoSplits      = $100;   // remove splits if pane freeze is removed
  BiffWoSelected      = $200;   // sheet selected
  BiffWoActive        = $400;   // sheet active
  BiffWoPageBreak     = $800;   // show in page break preview; clear = normal view

  { List of named paper sizes. }

  BiffPsUnknown       = 0;
  BiffPsA4            = 9;
  BiffPsReservedMin   = 118;
  BiffPsReservedMax   = 255;
  BiffPsCustomMin     = 256;

  { Flags for BOF record as specified
    in [MS-XLS] section 2.4.21. }

  BiffBoffWin         = 1;
  BiffBoffRisc        = 2;
  BiffBoffBeta        = 4;
  BiffBoffWinAny      = 8;
  BiffBoffMacAny      = $10;
  BiffBoffBetaAny     = $20;
  BiffBoffRiscAny     = $100;
  BiffBoffFontLim     = $2000;

  { Flags for WSBOOL record.
    See [MS-XLS] section 2.4.351 }

  BiffWsbShowBreaks   = $0001;  // show page breaks that added automatically
  BiffWsbDialog       = $0010;  // sheet is dialog
  BiffWsbOutline      = $0020;
  BiffWsbRowSums      = $0040;
  BiffWsbColSums      = $0080;
  BiffWsbFitPage      = $0100;  // fit the sheet to a print page
  BiffWsbHSync        = $1000;
  BiffWsbVSync        = $2000;
  BiffWsbAltExpr      = $4000;
  BiffWsbAltFormulas  = $8000;

type

  { BIFF stream.

    BIFF stream is a sequence of small
    data blocks named BIFF records. Each record
    has a header (2 byte Id and 2 byte Size field)
    and an optional data block. Normally, a BIFF stream
    starts with a BOF record and ends with a EOF record. }

  TBiffStream = class;

  TBiffRecord = class
  private

    FOwner: TBiffStream;
    FOffset: Cardinal;

    function  GetRecId: TBiffRecId;
    function  GetSize: Cardinal;
    procedure SetSize(n: Cardinal);

  public

    constructor Create(Owner: TBiffStream; Offset: Cardinal);

    { These methods append data to the record }

    procedure Write(const Buffer; Count: Cardinal);
    procedure WriteConst(Value, Size: Cardinal);
    procedure WriteZeros(Size: Cardinal);
    procedure WriteStream(Stream: TStream);

    { Allows to write data at arbitrary place inside the record }

    procedure WriteBytes(Offset, Data, Count: Cardinal);

    { Saves and loads contents of the record }

    procedure SaveToStream(Stream: TStream);
    procedure LoadFromStream(Stream: TStream);

    { Each BIFF record has an Id value that
      defines a type of the contents }

    property Id: TBiffRecId read GetRecId;

    { Size of the record's contents. This size doesn't include
      the 4-byte record header. }

    property Size: LongWord read GetSize write SetSize;

    { Each BIFF record is stored within a BIFF stream.
      This value contains an offset in bytes from the
      beginning of the BIFF stream to this BIFF record. }

    property Offset: LongWord read FOffset;

  end;

  TBiffStream = class
  private

    FStream: TStream;
    FRecords: TListCache;   // List of TBiffRecord
    FLastRec: TBiffRecord;  // Last added record

    function  ReadBytes(Offset, Count: Cardinal): Cardinal;
    procedure WriteBytes(Offset, Data, Count: Cardinal);

    procedure Append(Data: Cardinal; Count: Cardinal);
    procedure AppendData(const Data; Count: Cardinal);
    procedure AppendRecord(Rec: TBiffRecord; const Data; DataSize: Cardinal);

    function GetRecord(Index: Integer): TBiffRecord;
    function GetRecCount: Integer;
    function GetSize: Cardinal;

  public

    constructor Create(Cached: Boolean = False);
    destructor Destroy; override;

    function Add(Id: TBiffRecId): TBiffRecord;
    function AddBOF(k: TBiffStreamKind): TBiffRecord;
    function AddEOF: TBiffRecord;

    procedure SaveToStream(Stream: TStream);

    property Records[Index: LongInt]: TBiffRecord read GetRecord; default;
    property Count: Integer read GetRecCount;
    property Size: Cardinal read GetSize;

  end;

  { BIFF object.

    BIFF object is anything that has
    representation in a BIFF stream.
    BIFF objects are: workbook, worksheet,
    text cell, unicode string and so on. }

  TBiffObject = class

    { Every BIFF object must be capable to
      serialize itself in a BIFF stream. }

    procedure Flush(Stream: TBiffStream); virtual; abstract;

    function GetHashCode: LongInt; reintroduce; virtual;
    function Equals(s: TBiffObject): Boolean; reintroduce; virtual;

    { This method calls Flush method for each BIFF object
      in the list. }

    class procedure FlushList(list: TObjList; Stream: TBiffStream);

  end;

  //
  // BIFF8 unicode string
  //

  TBiffUCS = class(TBiffObject)
  private

    FData:      WideString;
    FRuns:      TMemoryStream;
    FHash:      LongInt;
    FCompress:  Boolean;

    procedure SetData(const Value: WideString);

    { UCS16 string is compressible if it contains only
      symbols in range 0..255, i.e. the high byte of each symbol
      is zero. }

    function IsCompressible: Boolean;

    procedure Init;

  public

    Len16:      Boolean;        // set, if the length of this string must occupy two bytes
    Tag:        LongInt;        // not used by this class
    SstIndex:   Integer;        // not used by this class

    constructor Create; overload;
    constructor Create(const S: WideString; UCS16: Boolean); overload;
    destructor Destroy; override;

    procedure AddFormat(Position, Font: LongInt);
    procedure Flush(Stream: TBiffStream); override;

    { Compares itself with a TBiffUCS using
      a created hash. }

    function Equals(s: TBiffObject): Boolean; override;

    { Returns a hash code for this string.
      The created hash code is cached. }

    function GetHashCode: LongInt; override;

    property Data: WideString read FData write SetData;
    property Compress: Boolean read FCompress write FCompress;

  end;

  //
  // STYLE
  //

  TBiffBuiltinStyleId = (
    bsiNormal,
    bsiRowLevel,
    bsiColLevel,
    bsiCOmma,
    bsiCurrency,
    bsiPercent,
    bsiComma0,
    bsiCurrency0,
    bsiHyperlink,
    bsiFHyperlink);

  TBiffStyle = class(TBiffObject)
  public

    XF:       LongInt;
    StyleId:  LongInt;
    Level:    LongInt;
    Name:     WideString;

    constructor Create;
    procedure Flush(Stream: TBiffStream); override;

  end;

  //
  // BLANK
  //

  TBiffCell = class(TBiffObject)
  public

    Row:    LongInt;
    Col:    LongInt;
    XF:     LongInt;

    constructor Create;
    procedure Flush(Stream: TBiffStream); override;

  end;

  //
  // Link Table (EXTRENALBOOK, EXTERNALSHEET, etc.)
  //

  TBiffLinkTable = class(TBiffObject)
  private

    { i-th REF is associated with a sheet named FIntSheetNames[i],
      which has index FIntSheetIndex[i] in the workbook }

    FIntSheetNames: array of string;
    FIntSheetIndex: array of Integer;
    FSheetsCount: Integer;

    function FindSheet(const Name: string): Integer;

    function GetRefsCount: Integer;
    procedure SetRefsCount(n: Integer);
  public

    { The Link Table contains a mapping from internal sheet names to
      indices to REF records. Formulas need this mapping: in expression
      like SheetABC!G8 the sheet name SheetABC is stored indirectly,
      via an associated with it an index to its REF record.  }

    function GetInternalSheetRef(const SheetName: string): Integer;

    { Links a sheet name with a corresponding sheet index. Sheet indices are known
      by TBiffWorkbook, so this method is called by the workbook. }

    procedure SetSheetIndex(const SheetName: string; Index: Integer);

    { Writes itself as a series of record. A link table is repersented with several
      records, unlike the most of other objects, which are represented by one record. }

    procedure Flush(Stream: TBiffStream); override;

    property RefsCount: Integer read GetRefsCount write SetRefsCount;
    property SheetsCount: Integer read FSheetsCount write FSheetsCount;
  end;

  //
  // FORMULA
  //

  TBiffFormulaCell = class(TBiffCell)
  private
    FInst: TStream;
  public
    constructor Create(Instructions: TStream);
    destructor Destroy; override;

    procedure Flush(Stream: TBiffStream); override;
  end;

  { Formulas lexer }

  TBiffFormulaLexemKind =
  (
    flkVoid,      // indicates that a lexem does not exist
    flkSpace,     // several characters that are considered to be whitespaces
    flkName,      // a sequence of letters and numbers: ABS, LOG, but not LOG10, A11
    flkInt,       // an unsigned integer
    flkString,    // a string
    flkOp,        // an operator, like < > <= >= <> & * + -
    flkSymbol     // some special symbol of unrecognized meaning
  );

  TBiffFormulaLexem = record
    Kind: TBiffFormulaLexemKind;
    Text: string;
  end;

  TBiffFormulaLexer = class
  private
    FText: string;
    FPos: Integer;
    FLexems: array of TBiffFormulaLexem;

    function IsAlpha(c: Char): Boolean;
    function IsDigit(c: Char): Boolean;

    function Read(Len: Integer): string;
    function GetChar: Char;
    function NextChar: Char;
    procedure SkipChar;
    function SubStr(Pos, Len: Integer): string;

    procedure Add(const Lex: TBiffFormulaLexem); overload;
    procedure Add(LexKind: TBiffFormulaLexemKind; const LexText: string); overload;

    function AddSpace: Boolean;
    function AddNumber: Boolean;
    function AddSymbol: Boolean;
    function AddName: Boolean;
    function AddString(Quote: Char): Boolean;
    function AddOp: Boolean;

    function GetLexemsCount: Integer;
    function GetLexem(i: Integer): TBiffFormulaLexem;

    procedure Analyse(const Formula: string);
  public
    property Formula: string read FText write Analyse;
    property Count: Integer read GetLexemsCount;
    property Lexems[i: Integer]: TBiffFormulaLexem read GetLexem; default;
  end;

  { Formulas RPN stack }

  TBiffFormulaTokenKind =
  (
    ftkVoid,
    ftkArg,
    ftkOp         // pops a number of args and pushes one result back; Flags = the number of args
  );

  TBiffFormulaOperatorKind =
  (
    fokVoid,
    fokPush,      // pushes to stack one argument

    fokDiv100,    // %
    fokNeg,       // -

    fokAdd,       // +
    fokSub,       // -
    fokDiv,       // /
    fokMul,       // *
    fokPow,       // ^

    fokArea,      // : creates a cell area (A2:G6)
    fokColon,     // : creates a cell area like fokArea, but more general (A2:INDIRECT("G6"))
    fokIsect,     // intersects two ranges, e.g. A1:G5 A2:H7; denoted by one space character
    fokExt,       // ! is used to refer to an external cell; e.g. "Sheet 10"!G6
    fokJoin,      // & joins two strings; e.g. "123" & "abc"

    fokL,         // <
    fokG,         // >
    fokE,         // =
    fokNE,        // <>
    fokLE,        // <=
    fokGE,        // >=

    fokNull,      // null value; it's used for representing a missing argument: SUM(A12,,7)
    fokNumber,    // a floating point number
    fokBool,      // a boolean value; Flags = 1 means "true", Flags = 0 means "false"
    fokString,    // a string
    fokCell,      // a cell reference; Flags: bit 0 = row is relative; bit 1 = column is relative
    fokFunc,      // a function call
    fokId         // the identity operator; pops one arg and pushes it back unmodified
  );

  TBiffFormulaToken = record
    Kind:   TBiffFormulaTokenKind;
    Op:     TBiffFormulaOperatorKind;
    Text:   string;
    Flags:  Integer;
  end;

  TBiffFormulaTokenArray = array of TBiffFormulaToken;

  TBiffFormulaRPNStack = class
  private
    FNumArgs: Integer;                // number of arguments in FCode
    FCode: TBiffFormulaTokenArray;    // resulting program
    FStack: TBiffFormulaTokenArray;   // stack of operators awaiting to being unrolled
    FFrame: TBiffFormulaTokenArray;   // stack of function calls

    procedure Error(const Msg: string);
    procedure Ensure(b: Boolean; const Msg: string = 'RPN stack failed');

    procedure Push(var Res: TBiffFormulaTokenArray; const t: TBiffFormulaToken);
    function  Pop(var a: TBiffFormulaTokenArray): TBiffFormulaToken;
    function  Top(const a: TBiffFormulaTokenArray): TBiffFormulaToken;

    procedure PushOp(Op: TBiffFormulaToken); overload; // pushes an arith op recognizing the precedence order
    procedure PopOp; // moves one operator from FStack to FCode

    procedure Unroll; // unrolls one frame
    function  Joinable(Op: TBiffFormulaOperatorKind): Boolean; // whether Op can be added on top of FStack

    function  OpPriority(Kind: TBiffFormulaOperatorKind): Integer;

    function  GetCount: Integer;
    function  GetInstruction(i: Integer): TBiffFormulaToken;
    function  GetFrameArgs: Integer; // number of available arguments in the current frame
    function  GetFrameOps: Integer; // number of operators in the current frame
  public
    procedure PushArg(Op: TBiffFormulaOperatorKind; const Text: string; Flags: Integer = 0);
    procedure PushOp(Op: TBiffFormulaOperatorKind; const Text: string; NumArgs: Integer = 0); overload;

    procedure PushFrame(const Func: string = ''); // pushes a function call
    procedure PopFrame; // it's called whenever the closing parenthesis occurs

    property Count: Integer read GetCount;
    property Instructions[i: Integer]: TBiffFormulaToken read GetInstruction; default;
  end;

  EBiffFormulaRPNStackError = class(Exception);

  { Formulas parser

    BNF of accepted formulas:

    formula   ::= ppterm (binop ppterm)*
    ppterm    :=  [prefop] term [postop]
    term      ::= cell | area | string | number | func | name | extcell | extarea
    func      ::= [name [int]] "(" [formula] ("," formula)* ")"
    number    ::= int [ "." int ]
    cell      ::= ["$"] name ["$"] int
    extcell   ::= sheet "!" cell
    extarea   ::= sheet "!" area
    area      ::= cell ":" cell
    sheet     ::= string | name
    binop     ::= "+" | "-" | "*" | "/" | "^" | etc.
    prefop    ::= "-" | "+"
    postop    ::= "%"

    (string, int, name - indivisible lexems; they are emitted by the lexer)

    After a formula analysed, the parser builds a sequence of tokens (instructions)
    available through its Tokens property. Each token represents an instruction or an argument
    for an instruction. This a token can have one of two kinds:

      ftkOp     A token of this kind represents an instruction that must be executed
                by a virtual processor. The Flags field defines the number of arguments
                taken.

      ftkArg    A token of this kind represents an argument, which can be a number, a string,
                a cell, the null arguments and so on. This token is not an instruction and
                its only purpose is to be an argument for a token of kind ftkOp.

    Here is an example of tokens produced from the formula "SUM(A3:B4 G5) + SheetABC!$G$8^2":

      cell(A3)        This is an argument token. It represents a cell.
      cell(B4)
      area:2          This is an operator token. It takes two previous tokens, makes an area (A3:B4) and
                      pushes the area to the stack.
      cell(G5)
      push:1          This is an operator token. It takes on previous token (G5) and pushes it to the stack.
      isect:2         This operator pops two values from the stack, intersects them and pushes the result back.
      call(SUM):1     This operator pops one value from the stack, calls SUM and pushes the result to the stack.
      str(SheetABC)   This is an argument token. It represents a string.
      cell(G8):3      Here "3" is the Flags field. It has bits 0 and 1 set. This means that
                      the row and the column are absolute, not relative.
      extcell:2       This operator takes 2 previous tokens ("SheetABC", "$G$8"), makes an external cell
                      reference from them (SheetABC!$G$8) and pushes it onto the stack.
      num(2)          This is an argument.
      pow:2           This operator takes two values from the stack (SheetABC!$G$8 and 2), does exponentiation
                      and pushes the result (SheetABC!$G$8^2) onto the stack.
      add:2           Like as pow:2, but does addition. }

  TBiffFormulaParser = class
  private
    FLexems: array of TBiffFormulaLexem;
    FPos: Integer; // current lexem
    FSavedPos: array of Integer;
    FRPN: TBiffFormulaRPNStack;
    FCode: TBiffFormulaTokenArray;

    function  CreateArgToken(Kind: TBiffFormulaOperatorKind;
      const Text: string = ''; Flags: Integer = 0): TBiffFormulaToken;

    { emits two tokens: arg and push }

    procedure Push(Arg: TBiffFormulaOperatorKind; const Text: string = ''; Flags: Integer = 0);

    { errors reporing }

    procedure Error(const ErrorMsg: string);
    procedure Ensure(b: Boolean; const ErrorMsg: string = 'Invalid formula');

    { lexems traversing }

    function  Lexem(i: Integer = 0): TBiffFormulaLexem; // returns (FPos + i) lexem or flkVoid
    procedure SkipLexem;
    function  SkipLexemIf(const Text: string): Boolean; overload;
    function  SkipLexemIf(Kind: TBiffFormulaLexemKind; out Text: string): Boolean; overload;
    function  IsOp(const Lex: TBiffFormulaLexem; const Text: string): Boolean;
    function  IsArgSep(const Lex: TBiffFormulaLexem): Boolean;

    { parser's state saving/restoring }

    procedure Save;     // saves the parser's state
    procedure Load;     // loads the parser's state
    procedure Discard;  // discards the last saved state

    { operators propeties }

    function  OpKind(const s: string): TBiffFormulaOperatorKind;

    { parsing methods }

    procedure Parse(const s: string);
    procedure BuildLexems(const s: string);
    procedure CleanLexems;
    procedure CopyCodeFromRPNStack;

    function  ReadSym(const s: string): Boolean;
    function  ReadString(out s: string): Boolean;
    function  ReadNumber(out s: string): Boolean;
    function  ReadName(out s: string): Boolean;
    function  ReadOp(out s: string): Boolean;
    function  ReadSheet(out s: string): Boolean;
    function  ReadCell(out t: TBiffFormulaToken): Boolean;

    function  ParseFormula: Boolean;
    function  ParsePPTerm: Boolean;
    function  ParsePrefOp: Boolean;
    function  ParsePostOp: Boolean;
    function  ParseTerm: Boolean;
    function  ParseString: Boolean;
    function  ParseNumber: Boolean;
    function  ParseCell: Boolean;
    function  ParseArea: Boolean;
    function  ParseBinOp: Boolean;
    function  ParseFuncCall: Boolean;
    function  ParseNameConst: Boolean;
    function  ParseExtCell: Boolean;
    function  ParseExtArea: Boolean;

    { property getters/setters }

    function  GetToken(i: Integer): TBiffFormulaToken;
    function  GetTokensCount: Integer;
  public
    property Formula: string write Parse;
    property Tokens[i: Integer]: TBiffFormulaToken read GetToken; default;
    property Count: Integer read GetTokensCount;
  end;

  EBiffFormulaParserError = class(Exception);

  { Formulas code emitter }

  TBiffFormulaCellRef = record
    Row: Integer;     // zero based row index
    Col: Integer;     // zero based column index
    AbsRow: Boolean;  // row index is absolute; in $G7 the index is absolute; in G7 it's relative
    AbsCol: Boolean;  // column index is absolute; in G$7 the index is absolute; in G7 it's relative
  end;

  TBiffFormulaRetType = (frtVoid, frtRef, frtVal, frtArray);

  TBiffFormulaCodeEmitter = class
  private
    FInst: TStream;
    FRetTypeMode: TBiffFormulaRetType;

    procedure Error(const ErrorMsg: string);
    procedure Ensure(b: Boolean; const ErrorMsg: string = 'Cannot emit opcode');
    procedure EnsureCellRange(const Cell: TBiffFormulaCellRef);

    function RelFlags(RelRow, RelCol: Boolean): Byte;

    procedure WriteOpCode(Op: Byte);
    procedure Write(Inst: Cardinal; Len: Cardinal = 1);
    procedure WriteCellRef(const Cell: TBiffFormulaCellRef);
    procedure WriteAreaRef(const Cell1, Cell2: TBiffFormulaCellRef);
  public

    { commands that push a value on the stack }

    procedure Push(Value: Integer); overload;
    procedure Push(Value: Double); overload;
    procedure Push(b: Boolean); overload;
    procedure Push(const s: string); overload;
    procedure PushNull;
    procedure PushCell(const Cell: TBiffFormulaCellRef);
    procedure PushArea(const Cell1, Cell2: TBiffFormulaCellRef);
    procedure PushExtCell(SheetRefId: Integer; const Cell: TBiffFormulaCellRef);
    procedure PushExtArea(SheetRefId: Integer; const Cell1, Cell2: TBiffFormulaCellRef);

    { function calls }

    procedure Call(Func: Cardinal); overload;
    procedure Call(Func, NumArgs: Cardinal); overload;
    procedure CallId; // identity function; emits parethesis for display purposes

    { unary operators }

    procedure Neg;      // unary -
    procedure Div100;   // %

    { binary arithmetic operators }

    procedure Add;      // +
    procedure Sub;      // -
    procedure Mul;      // *
    procedure Divide;   // /
    procedure Pow;      // ^

    { binary comparsion operators }

    procedure CmpL;     // <
    procedure CmpG;     // >
    procedure CmpLE;    // <=
    procedure CmpGE;    // >=
    procedure CmpE;     // =
    procedure CmpNE;    // <>

    { cell operators }

    procedure Intersect; // intersects two areas
    procedure Range;     // :

    { string operators }

    procedure Join;     // &

    property Output: TStream read FInst write FInst;
    property RetMode: TBiffFormulaRetType read FRetTypeMode write FRetTypeMode;
  end;

  EBiffFormulaCodeEmitterError = class(Exception);

  { Functions list.

    todo -cOptimisation: This clumsy class represents a map which stores items
    sorted by key. It can be represented by AVL, red-black or b-tree,
    but because of lack of generics in old Delphi versions,
    there's no easy way adapt existing classes in frxStorage for
    string keys. In far future this class should be replaced with something like
    TAVLTree<string, TBiffFormulaFunc>. }

  TBiffFormulaFunc = record
    Name:     string;
    Id:       Integer;
    MinArgs:  Integer;
    MaxArgs:  Integer;
    Volatile: Boolean;
    RetType:  Char;
    ArgTypes: string;
  end;

  TBiffFormulaFuncArray = array of TBiffFormulaFunc;

  TBiffFormulaFuncList = class
  private
    class procedure Init;
    class procedure Add(const f: TBiffFormulaFunc); overload;
    class function Find(Name: string): Integer;
    class function GetCount: Integer;
    class function GetFunc(i: Integer): TBiffFormulaFunc;
    class procedure SetFunc(i: Integer; const f: TBiffFormulaFunc);
    class procedure SetCount(n: Integer);
  public

    { [MS-XLS] Section 2.5.198.17

      Name      The function name. Letters case doesn't matter - this method upcases all letters.
      Id        The function id. It can be found in the documentation.
      MinArgs   The minimum number of arguments the function takes.
      MaxArgs   The maximum number of arguments the function takes.
      RetType   The result type. See note below.
      ArgTypes  Arguments types. The actual length of this array can be less than MaxArgs. In this
                case all missing entries are assumed to be equal to the last entry; e.g.
                SUM takes up to 30 args and all of them are references, so ArgTypes can be set to
                "r" or "rr" or "rrr" - all these strings are equally assumed to be equal to the string
                with 30 r's - "rrr....rrr".

      Result type of arguments types take one of three values: 'R' (reference), 'V' (value) and 'A' (array).
      See docs for meaning of these types. }

    class procedure Add(Id: Integer; Name: string; MinArgs, MaxArgs: Integer;
      RetType: Char; ArgTypes: string; Volatile: Boolean = False); overload;

    class function Exists(const Name: string): Boolean;
    class function Get(const Name: string): TBiffFormulaFunc;
    class function GetArgType(const Name: string; i: Integer): Char;
  end;

  { Formulas compiler }

  TBiffFormulaCompiler = class
  private
    FParser: TBiffFormulaParser;
    FPos: Integer; // current token
    FEmitter: TBiffFormulaCodeEmitter;
    FCode: TStream;
    FLinkTable: TBiffLinkTable;
    FRetTypes: array of TBiffFormulaRetType;

    function  GetLinkTable: TBiffLinkTable;

    function  Token(i: Integer = 0): TBiffFormulaToken;
    procedure SkipToken;
    procedure SelectToken(i: Integer);

    procedure Error(const Fmt: string; const Args: array of const);
    procedure Ensure(b: Boolean; const ErrorMsg: string = 'Cannot compile formula'); overload;
    procedure Ensure(b: Boolean; const Fmt: string; const Args: array of const); overload;

    function  IsCell(const t: TBiffFormulaToken): Boolean;
    function  IsStr(const t: TBiffFormulaToken): Boolean;

    function  GetCellPos(const t: TBiffFormulaToken): TBiffFormulaCellRef;

    { Excel doesn't perform types coercion. This method finds for each token
      what type it should return. }

    procedure CalcRetTypes;

    procedure Compile(const s: string);
    procedure CompileToken;

    procedure EmitNum(Num: Double);
    procedure EmitOp(Kind: TBiffFormulaOperatorKind; NumArgs: Integer);
    procedure EmitFunc(Name: string; NumArgs: Integer);
    procedure EmitIdFunc(NumArgs: Integer);
    procedure EmitArea(const Cell1, Cell2: TBiffFormulaToken);
    procedure EmitExtCell(const Sheet, Cell: TBiffFormulaToken);
    procedure EmitExtArea(const Sheet, Cell1, Cell2: TBiffFormulaToken);
    procedure EmitPush(const t: TBiffFormulaToken);
  public
    constructor Create;
    destructor Destroy; override;

    procedure SaveToStream(Stream: TStream);

    property Formula: string write Compile;

    { Link table is needed to compile external cell references: SheetABC!G8.
      If the formula doesn't contain such expressions, then LinkTable can be left nil. }

    property LinkTable: TBiffLinkTable read GetLinkTable write FLinkTable;
  end;

  EBiffFormulaCompilerError = class(Exception);

  //
  // LABELSST
  //

  TBiffTextCell = class(TBiffCell)
  public

    SST:    LongInt;

    constructor Create(SST: LongInt);
    procedure Flush(Stream: TBiffStream); override;

  end;

  //
  // NUMBER
  //

  TBiffNumberCell = class(TBiffCell)
  public

    Value:  Double; // IEEE 754 floating-point value (64-bit double precision)

    constructor Create(Value: Double);
    procedure Flush(Stream: TBiffStream); override;

  end;

  //
  // Palette.
  //
  // BIFF document can contain a PALETTE record
  // which specifies an array of used colors.
  // To refer to a color in this array, a color index
  // is used. There're several special built-in
  // colors. Their indexes are defined below.
  //

  TBiffColorIndex = (
    ciBlack,
    ciWhite,
    ciRed,
    ciGreen,
    ciBlue,
    ciYellow,
    ciMagenta,
    ciCyan
 );

  //
  // Font.
  //

  TBiffFontOptions = LongInt;   // See fo*** values

  TBiffFontFamily = (
    ffNone,
    ffRoman,
    ffSwiss,
    ffModern,
    ffScript,
    ffDecorative);

  TBiffFontEscapement = (
    feNone,
    feSuperScript,
    feSubScript);

  TBiffFontUnderline = LongInt; // See fu*** values
  TBiffFontWeight    = LongInt; // See fw*** values

  TBiffFontData = record

    Height:     LongInt;        // Font height in twips = 1/20 of a point
    Options:    Word;           // See TBiffFontOptions
    Color:      Word;           // Index to a color
    Weight:     Word;
    Esc:        TBiffFontEscapement;
    Underline:  TBiffFontUnderline;
    Family:     TBiffFontFamily;
    Charset:    Byte;           // For byte strings

  end;

  TBiffFont = class(TBiffObject)
  public

    Data: TBiffFontData;
    Name: WideString;
    Hash: Integer;
    FontIndex: Integer;         // not used by this class

    constructor Create;
    procedure Flush(Stream: TBiffStream); override;
    function Equals(Font: TBiffObject): Boolean; override;
    function GetHashCode: Integer; override;

    //
    // Returns width in points of a specified string as though
    // it's drawed with this font. On failure, returns 0.
    //
    // This function is complicated and its results
    // should be cached to increase performance.
    //

    function StrWidth(const Str: WideString): LongInt;

  end;

  //
  // XF.
  //
  // Extended cell formatting.
  //

  TBiffXFTypeProt = LongInt;  // See xftp** values
  TBiffXFUsedAttrib = LongInt;// See BiffXfua*** values
  TBiffPatternStyle = LongInt;// See ps*** values

  TBiffXFHAlign = (
    xfhaGeneral,
    xfhaLeft,
    xfhaCentered,
    xfhaRight,
    xfhaFilled,
    xfhaJustified,
    xfhaCAS,                    // Centred across selection
    xfhaDistributed);

  TBiffXFVAlign = (
    xfvaTop,
    xfvaCentered,
    xfvaBottom,
    xfvaJustified,
    xfvaDistributed);

  TBiffXFTextDir = (
    xftdAuto,
    xftdLTR,
    xftdRTL);

  TBiffXFOrientation = (
    xfoNone,
    xfoTop,                     // Letters are stacked top-to-bottom, but not rotated
    xfo90CCW,                   // 90 degrees counterclockwise
    xfo90CW);                  // 90 degrees clockwise

  TBiffLineStyle = (
    lsNone,
    lsThin,
    lsMedium,
    lsDashed,
    lsDotted,
    lsThick,
    lsDouble,
    lsHair,
    lsMediumDashed,
    lsThinDashDotted,
    lsMediumDashDotted,
    lsThinDashDotDotted,
    lsMediumDashDotDotted,
    lsSlantedMediumDashDotted
 );

  TBiffLine = record

    Style:    TBiffLineStyle;
    Color:    Byte;             // Index to a color

  end;

  TBiffXfData = record

    Font:       LongInt;        // Index to FONT record
    Format:     LongInt;        // Index to FORMAT record
    Prot:       Byte;           // See TBiffXFTypeProt
    Parent:     LongInt;        // Index to parent XF. Always -1 in style XFs.
    HAlign:     TBiffXFHAlign;
    WordWrap:   Boolean;
    VAlign:     TBiffXFVAlign;
    Justify:    Boolean;        // Justify last line in justified or distibuted text
    Rotation:   Byte;

    //
    // Indent level is measured in units,
    // each of them equals to length of three
    // blank characters.
    //

    Indent:     Byte;

    Shrink:     Boolean;        // Shrink content to fit into cell
    Direction:  TBiffXFTextDir;
    UsedAttrs:  Byte;           // See TBiffXFUsedAttrib

    L, T, R, B: TBiffLine;      // Left, top, right and bottom lines
    D:          TBiffLine;      // Diagonal line
    LTRB:       Boolean;        // Diagonal line from left-top to right-bottom
    LBRT:       Boolean;        // Diagonal line from left-bottom to right-top

    Patt:       Byte;           // See TBiffPatternStyle and docs
    PattColor:  Word;           // Index to a color
    PattBgColor:Word;           // Index to a color

  end;

  TBiffXF = class(TBiffObject)
  public

    XfIndex:    Integer;        // Index to XF table. Not used by this class.
    Data:       TBiffXfData;
    Hash:       Integer;

    constructor Create;
    procedure Flush(Stream: TBiffStream); override;
    function Equals(XF: TBiffObject): Boolean; override;
    function GetHashCode: Integer; override;

  end;

  //
  // COLINFO
  // Defines formatting for several consequent columns.
  //

  TBiffColInfo = class(TBiffObject)
  public

    First:      LongInt;        // Index to the first column
    Last:       LongInt;        // Index to the last column
    XF:         LongInt;        // Index to an XF record
    Hidden:     Boolean;
    Collapsed:  Boolean;
    Outline:    Byte;           // Must be in 0..7 range
    Width:      LongInt;        // In 1/256 of the zero character of the 0-th font

    constructor Create(Column, XF, Width: LongInt);
    procedure Flush(Stream: TBiffStream); override;

  end;

  //
  // WINDOW1
  //

  TBiffWindowOptions = LongInt; // See wo*** values

  TBiffWindow = class
  public

    HPos, VPos:     LongInt;    // In twips = 1/20 of a point
    Width, Height:  LongInt;    // In twips = 1/20 of a point
    ActiveSheet:    LongInt;
    FirstTab:       LongInt;
    SelSheet:       LongInt;
    TabWidth:       LongInt;    // Worksheet tab bar. 1/1000 of window width

    Visible:        Boolean;    // Window
    Open:           Boolean;    // Open/minimized
    HSBVisible:     Boolean;    // Hor. scroll bar
    VSBVisible:     Boolean;    // Vert. scroll bar
    TabVisible:     Boolean;    // Worksheet tab

    constructor Create;
    procedure Flush(Stream: TBiffStream);

  end;

  //
  // ROW
  //

  TBiffRow = class(TBiffObject)
  public

    Row:      LongInt;        // Index to this row
    Cells:    TObjList;       // List of TBiffCell
    FirstCol: LongInt;        // Index to column of the first cell
    LastCol:  LongInt;        // Index to column of the last cell
    Height:   LongInt;        // Height in twips
    Outline:  Byte;           // Valid outline levels are 0..7
    XF:       LongInt;        // Index to an XF record
    Hidden:   Boolean;

    { These two fields are used by TBiffSheet, not
      by this class.

        - FirstCell   - stream position of the first cell of this row
        - Offset      - stream position of this row }

    FirstCell: LongWord;
    Offset:   LongWord;

    constructor Create;
    destructor Destroy; override;
    procedure Flush(Stream: TBiffStream); override;

  end;

  //
  // PAGESETUP
  //

  TBiffPageOrientation = (
    bpoLandscape,
    bpoPortrait
  );

  TBiffPageSetup = class(TBiffObject)
  public

    Size:       TBiffPaperSize;
    Orient:     TBiffPageOrientation;
    Copies:     LongInt;
    Colored:    Boolean;      // Print in color or black and white
    Xdpi:       Word;
    Ydpi:       Word;
    PrintInRows:Boolean;
    Draft:      Boolean;      // Draft print quality
    CellNotes:  Boolean;      // Print cell notes
    SheetNotes: Boolean;      // Print notes at the end of the sheet

    constructor Create;
    procedure Flush(Stream: TBiffStream); override;

  end;

  { FILEPASS

    TBiffRC4 allows to encrypt excel workbook data using the RC4 cipher.
    This class is serialized into the FILEPASS record. }

  TBiffRC4 = class(TBiffObject)
  private

    { Specific data to be written to the output
      FILEPASS record }

    FSalt:      array[1..16] of Byte;   // Random data
    FVerifier:  array[1..16] of Byte;   // Encrypted verifier (that is random data)
    FHash:      array[1..16] of Byte;   // Encrypted MD5 of the verifier

    { Partial hash required for creating
      the encryption key. }

    FTruncHash: array[1..5] of Byte;

    { RC4 cipher }

    FCipher:    TCryptoRC4;

    { Positioon of the currently encrypted byte }

    FTail:      LongInt;
    FBlockId:   LongInt;

    procedure EncryptEx(Data: Pointer; Size: LongInt);
    procedure MD5(out Hash; const Data; Size: Integer);

  public

    { RC4 cipher performs a self test. If the test has failed,
      an exception is raised. }

    constructor Create;
    destructor Destroy; override;

    { Creates a partial hash as specified in [MS-OFFCRYPTO] section 2.3.6.2. }

    procedure Prepare(const Password; PassLen: LongInt);

    { Creates the FILEPASS record in a BIFF stream.
      This routine must be called after the Prepare routine. }

    procedure Flush(Stream: TBiffStream); override;

    { Initializes the RC4 cipher with a specified key }

    procedure Init(Key: Pointer; Len: LongInt);

    { Initializes the RC4 cipher as specified in [MS-OFFCRYPTO] section 2.3.6.2.
      This routine must be called after the Prepare routine, because
      Initialize uses the partial hash that's generated by Prepare }

    procedure Initialize(BlockId: LongWord);

    { Encrypts a data block "in place".
      The RC4 cipher generates a crypto stream (based on the password),
      that is combined with the original data with XOR operation.
      If Data is nil, the crypto stream is generated but is written
      nowhere. }

    procedure Encrypt(Data: Pointer; Size: LongInt); cdecl;

    { This routine encrypts a BIFF stream  as specified in
      [MS-XLS] section 2.2.10. }

    procedure EncryptStream(Stream: TBiffStream);
  end;

  TBiffBounds = record

    FR:       LongInt;        // First row
    LR:       LongInt;        // Last row
    FC:       LongInt;        // First column
    LC:       LongInt;        // Last column

  end;

  //
  // WINDOW2
  //

  TBiffWindow2 = class(TBiffObject)
  public

    Options:  Word;           // See BiffWo*** values
    FirstRow: Word;           // First visible row
    FirstCol: Word;           // First visible column
    GridCol:  Word;           // Grid line color
    MFPBP:    Word;           // Cached magnification factor in page break preview
    CMFNV:    Word;           // Cached magnification factor in normal view

    constructor Create;
    procedure Flush(Stream: TBiffStream); override;

  end;

  { BIFF sheet.

    Any XLS BIFF document consists of a few sheets.
    A sheet is represented by class TBiffSheet.
    An instance of this class can be accessed by a single
    thread at a time. }

  TBiffSheetKind = LongInt;   // See sk*** values

  TBiffSheetVisibility = (
    svVisible,
    svHidden,
    svVeryHidden);            // Such a sheet can only be shown/hidden by VB macro

  TBiffMargin = record

    Left:   Double;           // Inches
    Top:    Double;
    Right:  Double;
    Bottom: Double;

  end;

  TRectangle = class
  private
    FRect: TRect;
    function GetBottom: Integer;
    function GetBottomRight: TPoint;
    function GetLeft: Integer;
    function GetRight: Integer;
    function GetTop: Integer;
    function GetTopLeft: TPoint;
    procedure SetBottom(const Value: Integer);
    procedure SetBottomRight(const Value: TPoint);
    procedure SetLeft(const Value: Integer);
    procedure SetRight(const Value: Integer);
    procedure SetTop(const Value: Integer);
    procedure SetTopLeft(const Value: TPoint);
  public
    property Bottom: Integer read GetBottom write SetBottom;
    property BottomRight: TPoint read GetBottomRight write SetBottomRight;
    property Left: Integer read GetLeft write SetLeft;
    property Right: Integer read GetRight write SetRight;
    property Top: Integer read GetTop write SetTop;
    property TopLeft: TPoint read GetTopLeft write SetTopLeft;
  end;

  TBiffWorkbook = class;

  TBiffSheet = class(TBiffObject)
  private

    FRows:      TObjList;         // List of TBiffRow. Sorted by the row index.
    FCols:      TObjList;         // List of TBiffColInfo
    FMrgCells:  TObjectList;      // rectangles of merged cells
    FDGroup:    TEscherGroup;     // Drawing group. This class doesn't delete DGroup.
    FOwner:     TBiffWorkbook;
    FKind:      TBiffSheetKind;
    FVis:       TBiffSheetVisibility;
    FPB:        TList;            // List of LongInt (indexes to row where a page break occurs)
    FStrCount:  LongInt;          // number of text cells

    FLastReadRow: Integer;

    procedure SetColWidth(i: LongInt; w: LongInt);
    procedure SetRowHeight(i: LongInt; h: LongInt);
    function GetColWidth(i: LongInt): LongInt;
    function GetRowHeight(i: LongInt): LongInt;
    function GetRow(Index: LongInt): TBiffRow; // optimised for sequential access

  public

    Name:       WideString;
    Margin:     TBiffMargin;
    PageSetup:  TBiffPageSetup;
    View:       TBiffWindow2;
    Bounds:     TBiffBounds;      // Bounding rectangle of all cells

    { Index to the corresponding SHEET
      record which is placed in the Workbook
      globals stream. This index is formed by
      TBiffSheet.Flush method and used by
      TBiffWorkbook.Flush method to alter the
      SHEET record after all sheets are written
      to a BIFF stream. }

    RecIndex: Integer;

    constructor Create(Owner: TBiffWorkbook);
    destructor Destroy; override;

    //
    // Adda a new cell to the list of cells.
    //

    procedure AddCell(Cell: TBiffCell);

    { Merges a rectangle area of cells }

    procedure MergeCells(Rect: TRect);

    //
    // Adds a colinfo block to the list of colinfos
    //

    procedure AddColInfo(Info: TBiffColInfo);

    { Adds a new drawing to the sheet }

    function AddDrawing: TEscherShape;

    //
    // This function is to be called at the very end
    // to create a BIFF8 sheet substream.
    //

    procedure Flush(Stream: TBiffStream); override;

    { Adds a page break }

    procedure AddPageBreak(Row: LongInt);

    { Returns the index to the last row in this sheet.
      If there're no rows, -1 is returned. }

    function LastRowIndex: LongInt;

    { Column width is measured in 1/256 of width of the '0'
      character written by the 0-th font specified in
      the appropriate workbook. Row height is measured in twips.

        1 point = 20 twips

      When trying to read a non-existing column or a row, zero is
      returned. When trying to change size of an inaccesible column
      or a row (negative indexes, etc.), nothing is changed. }

    property ColWidth[i: LongInt]: LongInt read GetColWidth write SetColWidth;
    property RowHeight[i: LongInt]: LongInt read GetRowHeight write SetRowHeight;

    property TextCellsCount: Integer read FStrCount;
    property Visibility: TBiffSheetVisibility read FVis;
    property Kind: TBiffSheetKind read FKind;

  end;

  { BIFF workbook.

    The workbook is an Excel document containing some information
    and several sheets. The workbook is represented by class TBiffWorkbook.
    An instance of this class can be accessed by multiple threads at a moment.
    This is achieved via usage of critical section objects. }

  TBiffWorkbook = class(TBiffObject)
  private

    { Shared Strings Table }

    SST:      TObjList;         // List of TBiffUCS
    SSTHash:  TListHashTable;

    Win:      TBiffWindow;
    Sheets:   TObjList;         // List of TBiffSheet
    Styles:   TObjList;         // List of TBiffStyle
    FFormats: TObjList;         // List of TBiffUCS
    Escher:   TEscherStorage;   // Global Escher storage
    FLinkTbl: TBiffLinkTable;   // Contains external references; created on demand

    { Shared Fonts Table

      First four FONT records are reserved
      and have special meaning }

    Fonts:    TObjList;         // List of TBiffFont
    FontHash: TListHashTable;

    { Shared XFs Table

      Any BIFF document must contain
      at least 16 XF records used specially. }

    XFs:      TObjList;         // List of TBiffXF
    XFHash:   TListHashTable;

    { Shared Colors Table

      When anywhere a color is needed, this color
      is placed to PALETTE record and index to
      the color in the palette is used. Colors with
      indexes 0..7 are built-in, colors with indexes
      8..$3F are user defined.

      PALETTE record may be omitted and then the
      default Excel palette is used.

      To add a color to the palette
      use TboffWorkbook.AddColor method. }

    Palette:    array[0..BiffPaletteSize-1] of LongWord;
    PalUsed:    LongWord;

    { Optionally, the workbook can be encrypted }

    FCipher:    TBiffRC4;

    { The following critical section objects are used for
      mutual access to different parts of a workbook. }

    FCsStrings:   TRTLCriticalSection;
    FCsFonts:     TRTLCriticalSection;
    FCsXFs:       TRTLCriticalSection;
    FCsSheets:    TRTLCriticalSection;
    FCsStyles:    TRTLCriticalSection;
    FCsColors:    TRTLCriticalSection;
    FCsPictures:  TRTLCriticalSection;
    FCsFormats:   TRTLCriticalSection;

    { TBiffWorkbook can be accesses by a few threads.
      In this case, threads must synchronize access to
      the shared workbook object. Here is a typical example
      of use these functions:

         thread wants to add a string to a workbook
         thread calls TBiffWorkbook.LockSst
         thread adds a string via TBiffWorkbook.AddString
         thread calls TBiffWorkbook.UnlockSst

      TBiffWorkbook performs these locks automatically, i.e.
      client threads may not know about these. }

    procedure LockSst;
    procedure UnlockSst;

    procedure LockFonts;
    procedure UnlockFonts;

    procedure LockXfs;
    procedure UnlockXfs;

    procedure LockSheets;
    procedure UnlockSheets;

    procedure LockStyles;
    procedure UnlockStyles;

    procedure LockColors;
    procedure UnlockColors;

    procedure LockPictures;
    procedure UnlockPictures;

    procedure LockFormats;
    procedure UnlockFormats;

    { Internal  methods }

    function GetLinkTable: TBiffLinkTable;
    function GetSheetsCount: LongInt;
    function GetSheet(Index: LongInt): TBiffSheet;
    function GetFontsCount: LongInt;
    function GetFont(Index: LongInt): TBiffFont;
    procedure SetFormat(i: TBiffFormatIndex; const s: WideString);
    function AddBlip(Blip: TEscherPicture): LongInt;
    function AddColorInternal(C: LongWord): LongInt;

  public

    constructor Create;
    destructor Destroy; override;

    { Workbook can be protected with a password.
      The password can contain up to 255 unicode characters. }

    procedure SetPassword(const s: WideString);

    { BIFF8 contains all used strings in the SST table,
      all cells with strings refers to the SST table.
      If a cell needs to use a string, this function must be called.

      Returns an SST index.
      Multithread safe. }

    function AddString(S: TBiffUCS): LongWord; overload;
    function AddString(S: WideString): LongWord; overload;

    { In BIFF8 fonts have specific indexing: the 4-th FONT record
      is omitted. Use this function to add a new font and receive
      its valid index.

      Returns a font index.
      Multithread safe. }

    function AddFont(F: TBiffFont): LongWord;

    { These routines add a specified kind of object to the workbook.
      They are multithread safe. }

    function AddXF(X: TBiffXF): LongWord;
    function AddSheet(S: TBiffSheet): LongWord;
    function AddStyle(S: TBiffStyle): LongInt;
    function AddFormat(const FormatStr: WideString;
      FormatIndex: TBiffFormatIndex = -1): TBiffFormatIndex;

    { Adds a color to the palette and returns an index to the
      added color. If the color already exists, the index to
      the existing color is returned. If the palette is full
      and the color cannot be added, an index to the nearest color
      is returned.

      Byte structure of the color: RR GG BB AA
      Multithread safe. }

    function AddColor(C: LongWord): LongInt;

    { Adds a picture to the workbook.
      AddBitmap is applicable for BMP, PNG, JPEG, TIFF, etc.
      Multithread safe. }

    function AddBitmap(Kind: TEscherBlipKind; Contents: TStream): LongInt;
    function AddMetafile(Metafile: Graphics.TMetafile): LongInt;

    { Returns a count of strings occured in the workbook.
      Identical strings are counted as much as they occur.

      Multithread safe. }

    function StringsCount: LongInt;

    { Writes the workbook to data stream in BIFF8 format. }

    procedure Flush(Stream: TBiffStream); override;

    { The folowing properties are multithread safe. }

    property SheetsCount: LongInt read GetSheetsCount;
    property Sheet[Index: LongInt]: TBiffSheet read GetSheet;

    property FontsCount: LongInt read GetFontsCount;
    property Font[Index: LongInt]: TBiffFont read GetFont;

    property Format[i: TBiffFormatIndex]: WideString write SetFormat;

    property LinkTable: TBiffLinkTable read GetLinkTable;

  end;


{ MS Excel treats only #10 character as
  a line break, other characters (e.g. #13)
  it displays as squares (unknown symbols).

  This function deletes #13 chracaters.
  Returns a new length of the string.

  You must not use this function for an empty strings. }

procedure ValidateLineBreaks(var s: WideString);

{ Finds the nearest color within an array of colors. }

function NearestColor(Colors: Pointer; Count, Color: LongWord): LongWord;

implementation
end.
