
{******************************************}
{                                          }
{             FastReport VCL               }
{   Converter from TfrxIEMatrix to BIFF    }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxBiffConverter;

{$I frx.inc}

interface

uses
  Graphics,
  Windows,
  Classes,
{$IFDEF DELPHI16}
 System.UITypes,
{$ENDIF}
  frxClass,
  frxExportMatrix,
  frxBIFF,

  frxProgress,
  frGraphicUtils;

{ Printers.TPrinterOrientation
  This type is not defined in DELPHI4, therefore
  I have to define these constants manually. }

const
  frxPoPortrait = 0;  // poPortrait
  frxPoLandscape = 1; // poLandscape

type

  TfrxPrintOrient = LongInt; // Printers.TPrinterOrientation

  { Fields of this structure are
    copies of corresponding fields
    from TfrxReportPage. }

  TfrxPageInfo = record

    PaperWidth:   Extended;
    PaperHeight:  Extended;
    LeftMargin:   Extended;
    RightMargin:  Extended;
    TopMargin:    Extended;
    BottomMargin: Extended;
    Orientation:  TfrxPrintOrient;
    PageCount:    LongInt;
    PaperSize:    LongInt;
    Name:         string;

  end;

  { - Matrix  - the exported data matrix
    - Sheet   - the resulting Excel sheet
    - Workbook- a workbook that will contain the sheet
    - Page    - the current page
    - PageId  - index to the page
    - Size    - count of columns and rows in the mapped rectangle area
    - Source  - the top-left cell in TfrxIEMatrix
    - Dest    - the top-left cell in the Excel sheet
    - Images  - list of indexes to TfrxIEMObject with pictures
    - Surr    - if set, empty cells will be added for padding
    - Pictures - if set, pictures are exported
    - PageBreaks - if set, pages breaks are added to Excel sheet
    - FitPages - adjust page dimensions to fit to a print page
    - TSW     - with of a tab offset
    - ZCW     - width of zero character
    - GridLines - if set, the sheet will contain grid lines }

  TfrxBiffPageOptions = record
    Matrix:   TfrxIEMatrix;
    Sheet:    TBiffSheet;
    Page:     TfrxPageInfo;
    PageId:   LongInt;
    Source:   TPoint;
    Dest:     TPoint;
    Size:     TPoint;
    Images:   TList;
    Surr:     Boolean;
    Pictures: Boolean;
    PageBreaks: Boolean;
    WorkBook: TBiffWorkbook;
    FitPages: Boolean;
    TSW:      Integer;
    ZCW:      Integer;
    GridLines: Boolean;
    RHScale:  Extended;
    Formulas: Boolean;
  end;

  { Any object in the matrix has a style.
    This style must be converted to XF record
    that can be written to a BIFF document.
    The conversion is slow, so it need be cached. }

  TfrxBiffStyles = class
  private

    FXFi: array of Integer;   // XF indexes
    FFonti: array of Integer; // Font indexes
    FFonts: array of TFont;   // TFont objects
    FWorkbook: TBiffWorkbook;
    FTSW: Integer;            // the width of the tab character

    { Creates an XF record and returns an index to this record.
      Excel defines XF records of two types: cell XFs and style XFs.
      This function creates a cell XF record. }

    function CreateStyle(s: TfrxIEMStyle; BgPattern: Boolean;
      RTL: Boolean): LongInt;

    { Returns an index to an entry in FXFi that corresponds
      to the given arguments. }

    function GetEntryIndex(StyleIndex: Integer; Background, RTL: Boolean): Integer;

    { Returns True if the style index is valid }

    function IsValidStyleIndex(StyleIndex: Integer): Boolean;

  public

    { StylesCount - the maximum number of styles that will be added via AddStyle
      Workbook    - the workbook that will keep all added styles and fonts
      TSW         - the width of the tab character }

    constructor Create(StylesCount: Integer; Workbook: TBiffWorkbook; TSW: Integer);

    { Creates an XF record from the given style object and
      two boolean constants. The created XF record is associated
      with the given style index. The routine returns an index
      to the created XF record. }

    function AddStyle(StyleIndex: Integer; Style: TfrxIEMStyle;
      Background, RTL: Boolean): Integer;

    { Returns an index to a previously created XF record }

    function GetStyle(StyleIndex: Integer; Background, RTL: Boolean): Integer;

    { Creates a FONT record, adds it to a workbook and
      returns an index to the added FONT record. }

    function CreateFont(f: TFont; ss: TSubStyle = ssNormal): LongInt;

    { Creates a FONT recor from the given font object.
      The result is an index to the created FONT record. }

    function AddFont(Font: TFont): Integer;

    { Returns an index to a FONT record previously
      created via AddFont. If AddFont hasn't been
      called for this font object, then -1 is returned. }

    function GetFont(Font: TFont): Integer;
  end;

  TfrxBiffConverter = class
  private

    po: TfrxBiffPageOptions;
    FShowprogress: Boolean;
    FProgressBar: TfrxProgress;

    procedure Convert(Sheet: TBiffSheet; BiffMaxRow_: Longint = BiffMaxRow);

    procedure InitProgressBar(Steps: Integer; Text: string);
    procedure StepProgressBarIf(Condition: Boolean);
    procedure FreeProgressBar;

    procedure BreakIfTerminated;

    //
    // todo: any cell text seems to end with a line break
    // even no line breaks are written in the designer.
    // This will suppress the ending system symbols.
    //

    procedure SuppressEndingTrash(var s: WideString);

    function CreateFormulaCell(Obj: TfrxIEMObject): TBiffCell;
    function CreateHTMLCell(obj: TfrxIEMObject; Styles: TfrxBiffStyles): TBiffTextCell;
    function CreateTextCell(obj: TfrxIEMObject): TBiffCell;
    function CreateNumberCell(obj: TfrxIEMObject): TBiffCell;
    function CreateDateCell(obj: TfrxIEMObject): TBiffCell;
    function CreateCell(r, c: LongInt; obj: TfrxIEMObject; Styles: TfrxBiffStyles): TBiffCell;

    function IsFormula(Obj: TfrxIEMObject): Boolean;

    procedure SetColWidths(s: TBiffSheet);
    procedure SetRowHeights(s: TBiffSheet);
    procedure SetMargin(var m: TBiffMargin); // page margin in inches
    procedure SetPageSetup(ps: TBiffPageSetup);
    procedure AddImage(Sheet: TBiffSheet; ObjId: LongInt);
    procedure MergeCells(Sheet: TBiffSheet; ObjId: Integer);

    { Ratio of a length unit of an Excel column width to a pixel.
      Example:

      w := 300; // 300 pixels
      cw := w * GetColWidthFactor; // excel column's width is 300 pixels }

    function GetColWidthFactor: Double;

    { Ratio of a length unit of an Excel row height to a pixel.
      Example:

      h := 200; // 200 pixels
      rh := h * GetRowHeightFactor; // excel row's height is 200 pixels }

    function GetRowHeightFactor: Double;

    { Returns the width of a column in units
      defined in [MS-XLS] }

    function GetColWidth(Col: LongInt): Double;

    { Returns the height of a row in units
      defines in [MS-XLS] }

    function GetRowHeight(Row: LongInt): Double;

  public

    function GetSheet(BiffMaxRow_: Longint = BiffMaxRow): TBiffSheet;

    property Options: TfrxBiffPageOptions write po;
    property ShowProgress: Boolean read FShowProgress write FShowProgress;

  end;

  ETerminated = class(TObject);       // exporting terminated
  EInvalidFRFormat = class(TObject);  // invalid FR number format

function frxConvertMatrixToBiffSheet(po: TfrxBiffPageOptions;
  ShowProgress: Boolean;
  BiffMaxRow_: Longint = BiffMaxRow): TBiffSheet;

implementation
end.
