
{******************************************}
{                                          }
{             FastReport VCL               }
{             CBFF Writing API             }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

{ This module provides API for writing CBFF
  documents. The format is documented in
  MSDN [MS-CFB] section. }

unit frxCBFF;

interface

{$I frx.inc}

uses
  Classes, frxStorage, Windows, SysUtils;

type

  //
  // SAT writer.
  // Writes data streams with corresponding SAT table.
  //

  TCbffStream = class
  public

    Stream: TBlockStream;
    SAT:    TBlockStream;

    constructor Create; overload;
    constructor Create(SecShift: LongWord); overload;
    destructor Destroy; override;

    //
    // Writes a data block, writes a sequence of secIds and
    // returns the first secId.
    //

    function WriteStream(Data: Pointer; Size: LongWord): LongInt;

  end;

  //
  // CBFF directory.
  // Actually, a directory is a named data stream.
  //

  TCbffDirKind = (
    dkEmpty,
    dkStorage,
    dkStream,
    dkLock,
    dkProp,
    dkRoot);

  TCbffNodeColor = (
    ncRed,
    ncBlack);

  TCbffDir = class
  public

    Name:     WideString;       // Up to 31 characters
    Stream:   TBlockStream;     // Data stream
    Childs:   TObjList;         // List of TCbffDir
    Right:    TCbffDir;         // Right sibling
    Left:     TCbffDir;         // Left sibling
    Parent:   TCbffDir;         // Parent of this directory
    IsBlack:  Boolean;          // Node color in red-black tree
    Kind:     TCbffDirKind;

    SecID:    LongInt;          // SecID of the stream. For internal use.
    LeftID:   LongInt;          // DirID if the left sibling. For internal use.
    RightID:  LongInt;          // DirID if the right sibling. For internal use.
    ChildID:  LongInt;          // DirID if a child. For internal use.
    SelfID:   LongInt;          // DirID of itself. For internal use.

    constructor Create;
    destructor Destroy; override;
    procedure Flush(Stream: TStream);

    { Adds a new named child and returns that child. }

    function Add(const Name: WideString): TCbffDir;

  end;

  //
  // CBFF document.
  // Represents a CBFF document.
  //

  TCbffDocument = class
  public

    Root:           TCbffDir;     // Root directory entry
    SecShift:       LongInt;      // Sector size is 2**SecShift
    ShortSecShift:  LongInt;      // Short sector size is 2**ShortSecShift
    MinStreamSize:  LongWord;     // Minimal size of a standard stream

    constructor Create;
    destructor Destroy; override;

    //
    // CBFF document serializing
    //

    procedure Flush(Stream: TStream);

  end;

  //
  // CBFF file header
  // 512 bytes
  //

  TCbffHeader = packed record
    Sign:     array[1..8] of Byte;  // Signature: D0 CF 11 E0 A1 B1 1A E1
    Id:       array[1..8] of Word;
    Rev:      Word;                 // Revision: 003E
    Ver:      Word;                 // Version: 0003
    Order:    Word;                 // Byte order
    Sec:      Word;                 // Sector size is 2**Sec bytes
    SSec:     Word;                 // Short sector size is 2**SSec bytes
    NA1:      array[1..5] of Word;
    nSAT:     LongWord;             // Count of sectors used for SAT
    Dir:      LongInt;              // Directory stream
    NA2:      LongWord;
    MinSize:  LongWord;             // Minimal size of a standard stream
    SSAT:     LongInt;              // SSAT stream
    nSSAT:    LongWord;             // Count of sectors used for SSAT
    MSAT:     LongInt;              // MSAT stream
    nMSAT:    LongWord;             // Count of sectors used for MSAT
    sMSAT:    array[1..109] of LongInt; // First 109 MSAT SecID values
  end;

  PCbffHeader = ^TCbffHeader;

  //
  // Directory Entry
  // 128 bytes
  //

  TCbffDirEntry = packed record
    Name:     array[1..32] of WCHAR;// Unicode zero-ended name
    NameSize: Word;                 // Count of bytes occupied by Name
    Kind:     Byte;                 // Directory type, see TCbffDirKind
    Color:    Byte;                 // Node color, see TCbffNodeColor
    Left:     LongInt;              // Left sibling node
    Right:    LongInt;              // Right sibling node
    Child:    LongInt;              // One of child nodes
    Id:       array[1..8] of Word;
    Flags:    LongWord;
    Creation: array[1..2] of DWORD; // Creation time
    LastMod:  array[1..2] of DWORD; // Last modification time
    Stream:   LongInt;              // First SecID of the data stream
    Size:     LongWord;             // Size of the stream
    NA1:      LongWord;
  end;

  PCbffDirEntry = ^TCbffDirEntry;

const

  //
  // Special SecID values.
  //

  siFree  = -1;                     // SecID is not used
  siEOC   = -2;                     // This SecID terminates a chain of SecIDs
  siSAT   = -3;                     // Sector is used by SAT
  siMSAT  = -4;                     // Sector is used by MSAT

implementation
end.
