{******************************************}
{                                          }
{             FastReport VCL               }
{    Multithreaded Document Generating     }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxDraftPool;

interface

{$I frx.inc}

uses
  frxStorage, Windows, Classes, frCore;

type
  TDpThread = class;

  { Processing routine.
    This routine accepts a draft sheet and generates an actual
    document page. This routine is called within a thread context. }

  TDpProcessRoutine = procedure (Draft: TObject) of object;

  { Draft pool.
    This class performs parallel processing of document pages. }

  TDpDraftPool = class
  public

    { This list contains:

         Matrix objects that need be converted to document sheets
         Already created document sheets }

    Sheets: TObjList;

    { Count of currently added but not processed
      draft sheets. When threads will convert all
      draft sheets to document sheets, this value will
      become zero, and the event will become signalled. }

    DraftSheets: Integer;
    EvNoDraftSheets: TfrHandle;

    { Lock for Sheets }

    CsSheets: TRTLCriticalSection;

    { Lock for DraftSheets }

    CsDraftSheets: TRTLCriticalSection;

    { This is a list of threads that generate document sheets. }

    Threads: TObjList {of TfrxExportThread};

    { Draft process routine }

    DraftProcess: TDpProcessRoutine;

    { Access to Sheets must be performed via these routines }

    procedure LockSheets;
    procedure UnlockSheets;

    { Acces to DraftSheets must be performed via these routines }

    procedure LockDraftSheets;
    procedure UnlockDraftSheets;

    { This routine adds a new draft matrix object to the list. }

    procedure AddDraft(Draft: TObject);

    { When a draft sheet has been converted to a document sheet,
      this routine must be called to inform the draft pool. }

    procedure DraftProcessed;

    { This routine waits for all threads that are generating
      document sheets from given drafts }

    procedure WaitForThreads;

    constructor Create(ThreadsCount, SheetsCount: LongInt;
      DraftProcRoutine: TDpProcessRoutine);

    destructor Destroy; override;

  end;

  { Export thread.
    The export thread performs two actions:

       Waits for a draft matrix object
       Converts the found matrix object to a document sheet

    The export thread will never stop. It must be terminated
    explicitly. }

  TDpThread = class(TThread)
  private

    FDraftSheets: TList {of TfrxDraftSheet};
    FEvNewDraft: TfrHandle;
    FPool: TDpDraftPool;
    FCsDraftSheets: TRTLCriticalSection;
    FActiveDrafts: Integer;

    procedure LockDrafts;
    procedure UnlockDrafts;

  public

    ProcessedDrafts: Integer;
    MaxQueueLength: Integer;

    constructor Create(Pool: TDpDraftPool);
    destructor Destroy; override;

    procedure ForceStart;
    procedure PushDraft(Draft: TObject);
    function PopDraft: TObject;
    procedure Execute; override;

    property ActiveDrafts: Integer read FActiveDrafts write FActiveDrafts;

  end;

implementation
end.
