
{******************************************}
{                                          }
{             FastReport VCL               }
{        Escher Stream Writing API         }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

{ This module provides API for writing Escher
  streams. This format is documented in MSDN
  [MS-ODRAW] section. }

unit frxEscher;

interface

{$I frx.inc}

uses
  Classes, Windows,
  frTypes;

const

  //
  // Record types
  //

  EscherRkBse     = $f007;
  EscherRkBlip    = $f018;
  EscherRkDggc    = $f000;
  EscherRkDgg     = $f006;
  EscherRkBstore  = $f001;
  EscherRkSpCont  = $f004;
  EscherRkSp      = $f00a;
  EscherRkOpts    = $f00b;
  EscherRkAnchor  = $f00e;
  EscherRkDgCont  = $f002;
  EscherRkDg      = $f008;
  EscherRkSpgrCont= $f003;
  EscherRkSpgr    = $f009;
  EscherRkSMC     = $f11e;
  EscherRkCData   = $f011;
  EscherRkCAnchor = $f010;

  //
  // BLIP kinds
  //

  EscherBkEMF     = 2;
  EscherBkWMF     = 3;
  EscherBkJPEG    = 5;
  EscherBkPNG     = 6;
  EscherBkDIB     = 7;
  EscherBkTIFF    = 8;

  //
  // BLIP signatures
  // Actually these values are stored
  // in 12-bit fields.
  //

  EscherBsUnknown = $000;
  EscherBsWMF     = $216;
  EscherBsEMF     = $3d4;
  EscherBsPNG     = $6e0;
  EscherBsJPEG    = $46a;
  EscherBsDIB     = $7a8;
  EscherBsTIFF    = $6e4;

  //
  // BLIP usage
  //

  EscherBuDefault = 0;
  EscherBuTexture = 1;

  //
  // Metafile compression
  //

  EscherMcDeflate = 0;
  EscherMcNone    = $fe;

  //
  // Shape flags.
  // They are specified in Sp record.
  //

  EscherSfGroup   = $001;
  EscherSfChild   = $002;
  EscherSfRoot    = $004;
  EscherSfDeleted = $008;
  EscherSfOle     = $010;
  EscherSfMaster  = $020;
  EscherSfFlipHor = $040;
  EscherSfFlipVer = $080;
  EscherSfConn    = $100;
  EscherSfAnchor  = $200;
  EscherSfBg      = $400;
  EscherSfShape   = $800;

  //
  // Shape types.
  //
  // These values are written in 
  // the instance field of the Sp 
  // record.
  //
  // The complete list of 204
  // values can be found in the
  // documentation.
  //

  EscherStNone          = 0;
  EscherStPictureFrame  = 75;

  //
  // Each shape in an Escher stream has an
  // unique id, named "shape id". Shapes are
  // grouped into "drawing groups". Each drawing
  // group can contain up to a limited number
  // of shapes. This is not a restriction of
  // the Escher format, but this limit allow
  // to simplify generating shape id values.
  //

  EscherGroupLimit      = 1024;

type

  { Escher stream consists of consequent records.
    Each record has a fixed header and a variable
    length body. The body of an Escher record can
    be another Escher stream.

    When TEscherStream serializes its data,
    it looks over all its TEscherRec items and
    serializes them consequently.

    When TEscherRec serializes its data,
    it does that in three steps:

      - it writes the fixed header of 8 bytes

      - it serializes the own data, due to it is
        derived from TMemoryStream

      - it serializes all subrecords, due to TEscherRec
        is derived from TEscherStream }

  TEscherRec = class;

  TEscherStream = class(TMemoryStream)
  private

    Records:  TList;              // List of TEscherRec

  public

    constructor Create;
    destructor Destroy; override;

    procedure Flush(Stream: TStream); virtual;
    procedure WriteVal(Value: LongWord; Count: LongWord);

    function Add: TEscherRec; overload;
    function Add(Version: Byte; Instance: Word; Kind: Word): TEscherRec; overload;
    function AddCont(Instance: Word; Kind: Word): TEscherRec; overload;

  end;

  TEscherRec = class(TEscherStream)
  public

    Version:  Byte;
    Instance: Word;
    Kind:     Word;

    procedure Flush(Stream: TStream); override;
    function GetESize: LongWord;

  end;

  { There are two general kinds of drawings in Escher stream:

      - Metafiles (they are: EMF, WMF and PICT)
      - Bitmaps (they are: PNG, JPEG and DIB)

    The only difference is that metafiles start with one header
    and bitmaps start with another header. The actual data of images
    is not interesting for the Escher stream }

  TEscherBlipKind = Byte;         // See EscherBk values
  TEscherBlipSign = Word;         // See EscherBs values. Actually it's 12-bit length.

  TEscherPicture = class(TMemoryStream)
  public

    Kind: TEscherBlipKind;

    function GetESize: LongWord; virtual; abstract;
    procedure Flush(Stream: TStream); virtual; abstract;

  end;

  TEscherBitmap = class(TEscherPicture)
  public

    function GetESize: LongWord; override;
    procedure Flush(Stream: TStream); override;

  end;

  TEscherMetafile = class(TEscherPicture)
  public

    Bounds:   TRect;              // Boundary for metafile commands
    MFSize:   TPoint;             // Size of the metafile in EMUs (360000 EMU = 1 cm)
    Compr:    Boolean;            // Entire metafile can be LZ compressed

    function GetESize: LongWord; override;
    procedure Flush(Stream: TStream); override;

  end;

  TEscherBlipHash = array[1..16] of Byte;

  TEscherBlip = class
  public

    Hash:       TEscherBlipHash;
    RefCount:   LongWord;
    Pict:       TEscherPicture;
    Index:      LongWord;         // Index to this BLIP in the array of images

    constructor Create(Pict: TEscherPicture);
    destructor Destroy; override;
    procedure Flush(Stream: TEscherStream);

  end;

  { Properties of drawings are stored in a separate
    record. The contents of this record is an array
    of fixed size entries followed by an array of
    variable length elements such as Unicode strings }

  TEscherProp = class
  public

    Id:         Word;
    Blip:       Boolean;
    Complex:    Boolean;
    Value:      LongInt;

    constructor Create(Id: Word);

  end;

  TEscherPropList = class
  private

    FS: TList;                    // List of TEscherProp

  public

    constructor Create;
    destructor Destroy; override;

    function Add(Id: Word): TEscherProp;
    procedure Flush(Stream: TEscherStream);

  end;

  { Shape location.
    Location of a shape in an excel sheet can be
    specified relative to data cells. The left side
    of the sahpe is specified with two parameters:

      - Left - index to an excel data cell where
        the left side of the shape is located

      - LeftOffset - a distance between the left side
        of the data cell and the left side of the shape
        (this distance is measured in excel-specific
        units of length)

    The same rule is applicable to the other three
    sides of the shape. }

  TEscherShapePos = class
  public

    Left:   LongInt;
    Top:    LongInt;
    Right:  LongInt;
    Bottom: LongInt;

    LeftOffset:   LongInt;
    TopOffset:    LongInt;
    RightOffset:  LongInt;
    BottomOffset: LongInt;

    procedure Flush(Stream: TEscherStream);

  end;

  { Drawing group is a list of shapes.
    Each shape has:

      - shape id that is unique within a document
      - reference to an image object, such as a metafile
      - some options, such as a placemenet }

  TEscherShape = class
  public

    Id:       LongWord;           // Shape Id that's unique within a document
    Flags:    LongWord;           // Combination of EsherSf flags
    Image:    LongWord;           // Reference to an image object.
    Pos:      TEscherShapePos;

    constructor Create;
    destructor Destroy; override;
    procedure Flush(Stream: TEscherStream);

  end;

  TEscherStorage = class;

  TEscherGroup = class
  private

    SId:    LongWord;             // Used for generating unique Shape Ids.
    Shapes: TList;                // List of TEscherShape

    function GetShape(Index: LongInt): TEscherShape;
    function GetSId: LongWord;
    function GetRId: LongWord;

  public

    Id:     LongWord;             // Drawing Group Id that is unique within a document

    constructor Create;
    destructor Destroy; override;

    procedure Flush(Stream: TEscherStream);
    function Count: LongWord;
    function Add: TEscherShape;
    function GetMaxSId: LongWord;

    property Items[Index: LongInt]: TEscherShape read GetShape; default;

  end;

  { Escher Storage is a global per a document Escher stream
    that contains images. The storage consists of:

      - array of drawing groups (TEscherGroup)
      - array of images (TEscherBlip)
      - array of properties (TEscherPropList) }

  TEscherStorage = class
  private

    Groups:   TList;              // List of TEscherGroup
    Images:   TList;              // List of TEscherBlip

    function GetGroup(Index: LongInt): TEscherGroup;
    function GetImage(Index: LongInt): TEscherBlip;

  public

    Props: TEscherPropList;

    constructor Create;
    destructor Destroy; override;

    function AddGroup: TEscherGroup;
    function AddImage(Pict: TEscherPicture): TEscherBlip;

    function Empty: Boolean;

    procedure Flush(Stream: TEscherStream);

    property Image[Index: LongInt]: TEscherBlip read GetImage;
    property Group[Index: LongInt]: TEscherGroup read GetGroup;

  end;

{ Returns a BLIP signature corresponding to a BLIP type }

function EscherGetBlipSign(Kind: TEscherBlipKind): TEscherBlipSign;

{ Computes a 16-bytes hash value for a drawing }

procedure EscherGetBlipHash(out Hash: TEscherBlipHash; Blip: TMemoryStream);

implementation
end.
