
{******************************************}
{                                          }
{             FastReport VCL               }
{            HTML <div> Export             }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxExportHTMLDiv;

{ General advice for using this export.

   Avoid using vertical alignment in memos:
  it forces the export to create more
  complicated HTML. Leave the alignment vaTop
  whenever it's possible.

   Use @-type anchors in TfrxView.URL instead of
  #-type, because #-type is much slower to export. }

interface

{$I frx.inc}

uses
  {$IFNDEF FPC}
  Windows,
  {$ENDIF}
  SysUtils,
  Variants,
  Classes,
  {$IFDEF FPC}
    LCLType, LCLIntf, LCLProc,
  {$ENDIF}
  {$IFNDEF FPC}
  ShellAPI,
  {$ENDIF}
  Graphics,
  frxClass,
  frBaseGraphicsTypes,
  frxStorage, // for TCachedStream
  frxGradient,
{$IFDEF DELPHI16}
  System.UITypes,
{$ENDIF}
  frxExportHelpers, frxExportBaseDialog, frVectorCanvas, frCSSStyle;

type
  { Measures actual size of a TfrxView.
    Example:

    Gauge := TfrxBoundsGauge.Create;
    Gauge.Obj := MemoView;

    Gauge.Bounds.Left; // the leftmost coordinate including the left border
    Gauge.Borders.Left; // the left border width }

  TfrxBoundsGauge = class
  private
    FObj: TfrxView;
    FBoundsSet: Boolean;
    FBounds: TRect;
    FBorders: TRect;

    FX, FY, FX1, FY1, FDX, FDY: Integer;
    FFrameWidth: Integer;

    procedure SetObj(Obj: TfrxView);
    procedure AddBounds(r: TRect);
    function GetInnerWidth: Integer;
    function GetInnerHeight: Integer;
  protected
    procedure BeginDraw;
    procedure DrawBackground;
    procedure DrawFrame;
    procedure DrawLine(x1, y1, x2, y2, w: Integer; Side: TfrxFrameType);
  public
    property Obj: TfrxView read FObj write SetObj;
    property Bounds: TRect read FBounds;
    property Borders: TRect read FBorders;
    property InnerWidth: Integer read GetInnerWidth;
    property InnerHeight: Integer read GetInnerHeight;
  end;

  { Represents a HTML tag }

  TfrxHTMLItem = class
  private
    FName: string;
    FKeys: TStrings;
    FValues: TStrings;
    FValue: string;
    FRawValue: AnsiString;
    FChildren: TObjList;
    FLeft, FTop, FWidth, FHeight: Extended;
    FLeftSet, FTopSet, FWidthSet, FHeightSet: Boolean;
    FStyle: TfrxCSSStyle;
    FClass: string;
    FRotation: Integer;
    FAllowNegativeLeft: Boolean;
    FIsTransformMatrix: Boolean;
    FTM: array of Extended; // TransformMatrix

    procedure SetProp(Index: string; const Value: string);
    function GetProp(Index: string): string;
    function GetStyle: TfrxCSSStyle;

    procedure SetLeft(a: Extended);
    procedure SetTop(a: Extended);
    procedure SetWidth(a: Extended);
    procedure SetHeight(a: Extended);
  public

    constructor Create(const Name: string);
    destructor Destroy; override;

    function This: TfrxHTMLItem;

    procedure GaudeFrame(Obj: TfrxView);
    procedure Gaude(Obj: TfrxView);
    procedure WidenBy(Size: Extended);
    procedure DoPositive;
    procedure Expand(Size: Extended);

    class function EscapeAttribute(const s: string): string;

    procedure Save(Stream: TStream; Formatted: Boolean);

    function Add(const Tag: string): TfrxHTMLItem; overload;
    function Add(Item: TfrxHTMLItem): TfrxHTMLItem; overload;
    function AddRotated(const Tag: string; ARotation: Integer): TfrxHTMLItem;
    function AddTransformed(const Tag: string; ATransformMatrix: array of Extended): TfrxHTMLItem;
    procedure AddCSSClass(const s: string);

    property Prop[Index: string]: string read GetProp write SetProp; default;
    property Value: string write FValue;
    property RawValue: AnsiString write FRawValue;
    property Name: string write FName;

    { CSS style is created on demand }
    property Style: TfrxCSSStyle read GetStyle;
    { These lengths are measured in FR units (pixels) }
    property Left: Extended read FLeft write SetLeft;
    property Top: Extended read FTop write SetTop;
    property Width: Extended read FWidth write SetWidth;
    property Height: Extended read FHeight write SetHeight;

    property AllowNegativeLeft: Boolean read FAllowNegativeLeft write FAllowNegativeLeft;
  end;

  { Queue of automatically serialised HTML tags }

  TfrxHTMLItemQueue = class
  private
    FQueue: array of TfrxHTMLItem;
    FUsed: Integer;
    FStream: TStream;
    FFormatted: Boolean;
  protected
    procedure Flush;
  public
    constructor Create(Stream: TStream; Formatted: Boolean);
    destructor Destroy; override;

    procedure Push(Item: TfrxHTMLItem);
    procedure SetQueueLength(n: Integer);
  end;

  { HTML export filter }

{$IFDEF DELPHI16}
  [ComponentPlatformsAttribute(pidWin32 or pidWin64)]
{$ENDIF}

  TfrxHTMLDivExport = class(TExportHTMLDivSVGParent)
  private
    FTitle: string;
    FHTML5: Boolean;
    FAllPictures: Boolean;
    FPageStyle: TfrxCSSStyle;
    FExportAnchors: Boolean;
    FPictureTag: Integer;

    FQueue: TfrxHTMLItemQueue; // it represents the current page

    FUseTemplates: Boolean;
    FPrint: Boolean;
    FGetNavTemplate: TfrxHTMLExportGetNavTemplate;
    FGetMainTemplate: TfrxHTMLExportGetMainTemplate;
    FGetToolbarTemplate: TfrxHTMLExportGetToolbarTemplate;

    FPreviewOutline: TfrxCustomOutline;
    FOutlineTree: TfrxHTMLOutlineNode;
    FOutlineTreeNode :TfrxHTMLOutlineNode;

    FOnProcessHyperLink: TOnProcessHyperLink;
    FParentDetailURL: String;

    function GetPageStyle: TfrxCSSStyle;
    function GetAnchor(var Page: string; const Name: string): Boolean;
    function GetHRef(const URL: string): string;
    procedure PutImg(Obj: TfrxView; Pic: TGraphic; WriteSize: Boolean);
    procedure EndHTML;

    { Handlers for specific kinds of TfrxView objects.
      They return "True" if they succeed to export an object, or "False"
      if they want to pass the object further along the handlers chain. }
    function ExportTaggedView(Obj: TfrxView): Boolean;
    function ExportAllPictures(Obj: TfrxView): Boolean;
    function ExportMemo(Obj: TfrxView): Boolean;
    function ExportPicture(Obj: TfrxView): Boolean;
    function ExportShape(Obj: TfrxView): Boolean;
    function ExportLine(Obj: TfrxView): Boolean;
    function ExportGradient(Obj: TfrxView): Boolean;
    {$IFNDEF FPC}
    function ExportViaEMF(Obj: TfrxView): Boolean;
    {$ENDIF}
    function ExportAsPicture(Obj: TfrxView): Boolean;
  protected
    { Creates a new HTML tag and returns it. The tag don't need to be
      deleted or serialised to a stream: this is done automatically. }
    function AddTag(const Name: string): TfrxHTMLItem;
    { Creates an empty <div> tag with assigned styles }
    function CreateDiv(Obj: TfrxView; Widen: Integer = 0): TfrxHTMLItem;
    function CreateFrameDiv(Obj: TfrxView): TfrxHTMLItem;
    function CreateFillDiv(Obj: TfrxView): TfrxHTMLItem;
    procedure FillGraduienProps(Style: TfrxCSSStyle;
      BeginColor, EndColor: TColor; GradientStyle: TfrxGradientStyle);
    function FilterHTML(const Text: string): string;
    function EscapeText(const s: string): string;

    function DoHyperLink(HL: TfrxHyperlink): string;
    procedure DoExportAsPicture(Obj: TfrxView; Transparent: Boolean; IsVectorSource: Boolean; IsAlphaSource: Boolean;
      TransparentColor: TColor = clNone);

    function NavPageNumber(PageNumber: Integer): string;
    procedure StartAnchors;
    procedure StartNavigator;
    procedure CreateCSS; override;
    procedure StartHTML;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    class function GetDescription: String; override;
    class function ExportDialogClass: TfrxBaseExportDialogClass; override;
    { function ShowModal: TModalResult; override; }

    function Start: Boolean; override;
    procedure StartPage(Page: TfrxReportPage; Index: Integer); override;
    procedure FinishPage(Page: TfrxReportPage; Index: Integer); override;
    procedure Finish; override;

    { Report pages are represented as separate <div> tags. All these tags
      are associated with the same CSS style, so its possible to adjust
      appearance of pages by modifying this style. Example:

      PageStyle['border'] := '1mm solid orange';
      PageStyle['margin'] := '5mm';
      PageStyle.CSS3Style['box-shadow'] := '3mm 3mm 3mm gray';
      PageStyle.cSS3Style['border-radius'] := '3mm'; }
    property PageStyle: TfrxCSSStyle read GetPageStyle;

    property OnProcessHyperLink: TOnProcessHyperLink read FOnProcessHyperLink write FOnProcessHyperLink;
    property ParentDetailURL: string read FParentDetailURL write FParentDetailURL;

    property Print: Boolean read FPrint write FPrint;
    property UseTemplates: Boolean read FUseTemplates write FUseTemplates;
    property OnGetMainTemplate: TfrxHTMLExportGetMainTemplate read FGetMainTemplate
      write FGetMainTemplate;
    property OnGetToolbarTemplate: TfrxHTMLExportGetToolbarTemplate read FGetToolbarTemplate
      write FGetToolbarTemplate;
    property OnGetNavTemplate: TfrxHTMLExportGetNavTemplate read FGetNavTemplate
      write FGetNavTemplate;
  published
    property Title: string read FTitle write FTitle;

    { Allows using HTML5 features }
    property HTML5: Boolean read FHTML5 write FHTML5;

    { Exports all report components as pictures }
    property AllPictures: Boolean read FAllPictures write FAllPictures;

    { Creates anchors based on TfrxView.URL property. This option is useful only if
      the URLs begin with the "#" sign. }
    property ExportAnchors: Boolean read FExportAnchors write FExportAnchors;

    { If not equals zero, forces the export to save all report components
      with this value of the Tag property export as pictures. The format of pictures
      is defined by PictureFormat. }
    property PictureTag: Integer read FPictureTag write FPictureTag;
  end;

  { HTML export that uses modern HTML5 features }
{$IFDEF DELPHI16}
  [ComponentPlatformsAttribute(pidWin32 or pidWin64)]
{$ENDIF}

  TfrxHTML5DivExport = class(TfrxHTMLDivExport)
  public
    constructor Create(AOwner: TComponent); override;
    class function GetDescription: string; override;
  end;

  { HTML export that is compatible with old browsers, like IE6 }
{$IFDEF DELPHI16}
  [ComponentPlatformsAttribute(pidWin32 or pidWin64)]
{$ENDIF}

  TfrxHTML4DivExport = class(TfrxHTMLDivExport)
  public
    constructor Create(AOwner: TComponent); override;
    class function GetDescription: string; override;
  end;

implementation
end.
