
{******************************************}
{                                          }
{             FastReport VCL               }
{       Generalized Markup Language        }
{           Reading/Writing API            }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

{ This module contains API for reading and
  writing documents that have syntax defined
  in the ISO 8879 standard. }

unit frxGML;

{ The implementation has the following simple
  idea. The source document is represented as
  a tree of nodes. The document written as a
  string is assumed to be the root node. Any
  node can be of two kinds:

    - a plain text node
    - a node that contains a list of subnodes

  Any node has the following main fields:

    - Reference to a part of the source
      document that represents the header of
      the node. This reference is a pair of
      values: a zero based index to the first
      character of the header and a count of
      characters in the header. Example of
      headers:

        - b, i, table (HTML)
        - ul, ansi, f (RTF)

    - Reference to the body of the node.

    - List of subnodes ordered in the same
      way as they are ordered in the document.

  Note, that the sum of the header and the body
  not equals to the whole node. An example is:

    <td class="c">text with an <a>anchor</a></td>

  In this example the header is "td", the body
  is "text with an <a>anchor</a>". }

interface

{$I frx.inc}

uses
  Classes, Types;

const

  { Values for TGmlEaResult
    Obselete }

  GmlEaFailure    = 0;  // the node failed to expand
  GmlEaSuccess    = 1;  // the node and all subnodes expanded
  GmlEaIncomplete = 2;  // the node expanded but not all subnodes expanded

  { Values for TGmlElValue }

  GmlElNone       = 0;  // leave the node unexpanded
  GmlElOne        = 1;  // expand the first level of subnodes
  GmlElAll        = -1; // build the complete tree of nodes

  { RTF special characters.
    The choosen values for these
    constants don't have special
    meaning and can be replaced with
    any other values that not equal
    to known codes of characters. }

  GmlRtfScEscape: AnsiChar = #$81;  // backslash "\"
  GmlRtfScOpen:   AnsiChar = #$82;  // opening brace "{"
  GmlRtfScClose:  AnsiChar = #$83;  // closing brace "}"

  { Maximum count of characters in an Rtf
    control word, e.g. in \lang the
    count is 4. }

  GmlRtfMaxControl = 32;

  { Maximum count of characters in an Rtf
    control word's argument, i.e. in \lang123
    the count is 3. }

  GmlRtfMaxControlArg = 10;

type

  TGmlEaResult = LongInt;
  TGmlElValue = LongInt;

  { List of TObject instances }

  TGmlList = class(TList)
  public

    procedure Clear; override;

  end;

  { String reference.
    This auxiliary structure represents a string
    in the source document. The string can be represented
    in two ways:

       Indirect. In this case the string is defined
        by two numbers: an index to its first character
        within a source document, and the length of
        the string.

       Direct. In this case the string is represented
        by AnsiString object.

    If TGmlStr.Data field is not empty, then the direct
    way is used. Otherwise the indirect way is used. }

  TGmlStr = record

    Pos:        LongInt;    // zero based
    Len:        LongInt;    // can be zero
    Data:       AnsiString; // can be empty

  end;

  { Abstract node. }

  TGmlNode = class
  private

    procedure SetHeader(const s: AnsiString);
    procedure SetBody(const s: AnsiString);

  protected

    FHeader:    TGmlStr;  // can has zero length
    FBody:      TGmlStr;  // can has zero length

    function GetHeader: AnsiString; virtual; abstract;
    function GetBody: AnsiString; virtual; abstract;

  public

    FParent:    TGmlNode; // can be nil
    FSubNodes:  TList;    // can be nil or empty

    destructor Destroy; override;

    { Default destructor doesn't delete subnodes. }

    procedure DestroyTree;

    { Returns True if a subnode exists.
      Returns False if there're subnodes. }

    function Empty: Boolean;

    { Walks over all subnodes and updates
      their FParent field. Then calls the Update
      routine of theirs. }

    procedure UpdateParent;

    { Removes itself from the list of subnodes of its
      parent. Optionally, deletes itself from memory
      and all its subnodes. }

    procedure Remove(DestroyItself: Boolean = True);

    { This node is contained in a list of
      subnodes of a parent of this node.
      This routine returns the index to this node
      in the list. If this node doesn't have
      a parent node, -1 is returned. }

    function SelfSubIndex: LongInt;

    { Copies a range of subnodes.
      Arguments:

        - First - index to the first subnodes from
          which the search should be started

        - Last - index to the last subnode, if it
          equals to -1 all subnodes from First till
          the end will be copied }

    function Select(First: LongInt; Last: LongInt = -1): TList;

    function NodesCount: LongInt;

    property Header: AnsiString read GetHeader write SetHeader;
    property Body: AnsiString read GetBody write SetBody;

  end;

  { RTF nodes.

    The following classes represents different
    kinds of nodes in an RTF documents. }

  TGmlRtf = class;

  TGmlRtfNode = class(TGmlNode)
  protected

    function GetHeader: AnsiString; override;
    function GetBody: AnsiString; override;

  public

    FDoc: TGmlRtf;

    { Serializes this node and all subnodes
      to an output text stream. }

    procedure Serialize(Stream: TStream); virtual;

    { Serializes a range of subnodes.
      The Last argument can be -1, in this
      case all subnodes from First will be
      serialized. }

    procedure SerializeSubNodes(Stream: TStream; First, Last: LongInt);

    { Returns a specified node.
      If the specified node doesn't exist,
      the result is nil. }

    function Node(Index: LongInt): TGmlRtfNode;

    { Finds a subnode with a specified header.
      The routine looks for all subnodes, not only
      immediate subnodes. }

    function Find(const Hdr: AnsiString): TGmlRtfNode;
    function FindAll(const Hdr: AnsiString;
      MaxCount: LongInt = 0): TList {of TGmlRtfNode};

  end;

  TGmlRtfGroup = class(TGmlRtfNode)
  public
    procedure Serialize(Stream: TStream); override;
  end;

  TGmlRtfText = class(TGmlRtfNode)
  public
    procedure Serialize(Stream: TStream); override;
  end;

  TGmlRtfControl = class(TGmlRtfNode)
  private

    FArg: Boolean;
    FArgValue: LongInt;

    procedure SetValue(x: LongInt);
    function GetValue: LongInt;

  public

    procedure Serialize(Stream: TStream); override;

    { RTF control word can be without a value.
      If you try to read value of such a control,
      an exception will be raised. }

    property Value: LongInt read GetValue write SetValue;

  end;

  TGmlRtfNumber = class(TGmlRtfNode)
  public
    FValue: LongInt;
    procedure Serialize(Stream: TStream); override;
  end;

  { RTF Control Symbol

    A control symbol consists of a backslash followed
    by a single, nonalphabetic character. For example,
    \~ represents a nonbreaking space.
    Control symbols take no delimiters. }

  TGmlRtfSymbol = class(TGmlRtfNode)
  public
    Symbol: AnsiChar;
    procedure Serialize(Stream: TStream); override;
  end;

  { RTF font.

    RTF document keeps descriptions of fonts
    in a list of group nodes. Each group node
    has a list of subnodes which defines
    properties of a font. }

  TGmlRtfFont = class(TGmlRtfNode)
  private

    function GetIndex: LongInt;
    function GetCharset: LongInt;
    function GetName: AnsiString;

  public

    { Access to any of these properties
      will perform a search in the list
      of subnodes. }

    property Index: LongInt read GetIndex {default -1};
    property Charset: LongInt read GetCharset {default 0};
    property Name: AnsiString read GetName {default ''};

  end;

  { RTF color.

    Color in an RTF document is represented by
    three consequent nodes \red100\green200\blue50
    where a value of a node is the saturation
    of the corresponding color component.  }

  TGmlRtfColor = class
  public
    R, G, B: TGmlRtfControl;
    function Serialize: AnsiString;
  end;

  { RTF document }

  TGmlRtf = class
  private

    { The following is used everywhere.  }

    FText:  AnsiString;     // the whole document
    FRoot:  TGmlRtfNode;    // the root Rtf node

    { The following data is used by the Rtf parser. }

    FStack: TList;          // list of saved pointers to the document
    FPos:   LongInt;        // current pointer to the source document
    FLast:  LongInt;        // last available position for parsing

    FFontList: TList {of TGmlRtfFont};
    FColorList: TGmlList {of TGmlRtfColor};

    { Skips "invisible" symbols and returns the first
      visible character. The current position points
      to the returned character. }

    function Current(SkipInvisibles: Boolean = False): AnsiChar;

    { Returns the first visible character.
      The current position points after the
      returned character. }

    function Get: AnsiChar;

    { Returns the first visible character.
      This function notes the escaping symbol '\'
      and converts special symbols to their
      GmlRtfSc values. }

    function Prepare: AnsiChar;

    { Saves parser's state }

    procedure Push;

    { Takes the last saved state of the parser.
      If Discard = False, the last saved state
      is loaded to the parser. If Discard = True,
      the last saved state is not loaded.
      Anyway, the last saved state is removed from
      the stack of saved states. }

    procedure Pop(Discard: Boolean = False);

    { Simple operations on characters. }

    function IsVisible(c: AnsiChar): Boolean; cdecl;
    function Escape(c: AnsiChar): AnsiChar;
    function IsSpecChar(c: AnsiChar): Boolean; cdecl;
    function IsAlpha(c: AnsiChar): Boolean; cdecl;
    function IsDigit(c: AnsiChar): Boolean; cdecl;
    function HexDigit(c: AnsiChar): Byte; cdecl;

    { Parsing routines. }

    function SkipAlpha: Boolean;        // 'A'..'Z' | 'a'..'z'
    function SkipDigit: Boolean;        // '0'..'9'
    function SkipControl: TGmlRtfNode;  // '\' <word>[<integer>][<space>]
    function SkipGroup: TGmlRtfNode;    // '{' {<node>} '}'
    function SkipNumber: TGmlRtfNode;   // "\'" <hexdigit><hexdigit>
    function SkipText: TGmlRtfNode;     // {^<specsym>}
    function SkipControlSymbol: TGmlRtfNode; // '\' <nonalpha>
    function SkipList: TList;           // {<node>}

    { Parses a specified selection of the document and
      returns a list of Rtf nodes that represent the
      selection. }

    function Parse(First, Last: LongInt): TList;

    { Looks for \fonttbl node and updates FFontList.
      If \fonttbl doesn't exist, FFontList is set to nil. }

    procedure LoadFonts;
    procedure LoadColors;

  public

    constructor Create(const Text: AnsiString);
    destructor Destroy; override;

    { Returns the count of characters in the document. }

    function Length: LongInt;

    { This routine copies a substring from the
      source document. Arguments:

        - Pos - a zero based index to the first
          character in the string in the document

        - Len - a count of characters in the string

        - Dest - a resulting buffer that is capable
          to store Len characters

        - Returned value - the number of characters
          that is copied to resulting buffer

      Implementation of this routine must not raise
      an exception. }

    function Copy(Pos: LongInt; Len: LongInt; Dest: PAnsiChar): LongInt; overload;

    { Copies a string from the document.
      If the specified location of the string is out
      of bounds, an empty string is returned. }

    function Copy(Pos: LongInt; Len: LongInt): AnsiString; overload;
    function Copy(Str: TGmlStr): AnsiString; overload;

    { Returns a character at a specified position.
      The position is zero based.
      This routine must raise an exception if the
      specified position is out of bounds. }

    function Char(Pos: LongInt): AnsiChar;

    { Writes this document in the Rtf format. }

    procedure Serialize(Stream: TStream);

    { Returns the top level node.
      Normally, this node contains a subnode TGmlRtfGroup,
      that contains a list of subnodes with document
      contents. }

    property Root: TGmlRtfNode read FRoot;

    { Returns count of fonts in the font table.
      If the table doesn't exist, zero is returned. }

    function FontsCount: LongInt;

    { Returns a font by its index.
      Note, that this index is not the same as
      the font index. The font index is an attribute
      of a font. If the font doesn't exist,
      nil is returned. }

    function Font(i: LongInt): TGmlRtfFont;

    { Returns count of colors in the color table.
      If the table doesn't exist, zero is returned. }

    function ColorsCount: LongInt;

    { Returns a color by its index.
      If the color doesn't exist,
      nil is returned. }

    function Color(i: LongInt): TGmlRtfColor;

  end;

implementation
end.
