
{******************************************}
{                                          }
{             FastReport VCL               }
{            OLEPS Writing API             }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

{ This module provides API for writing
  OLEPS streams. The format is documented
  in MSDN [MS-OLEPS] section. }

unit frxOLEPS;

interface

{$I frx.inc}

uses
  Classes, Windows;

type

  TOlepsFmtId = array[1..4] of LongWord;
  TOlepsPropId = LongWord;
  TOlepsPropType = Word;
  TOlepsAccess = LongWord;

const

  { OleSi values represents a list of
    property id values for properties
    that are stored within a <05>SummaryInformation
    OLEPS stream. }

  OlepsSiCodepage   = $01;
  OlepsSiAuthor     = $04;
  OlepsSiLastAuthor = $08;
  OlepsSiAppName    = $12;
  OlepsSiAccess     = $13;
  OlepsSiComment    = $06;
  OlepsSiKeywords   = $05;
  OlepsSiSubject    = $03;
  OlepsSiTitle      = $02;
  OlepsSiRevision   = $09;
  OlepsSiCreation   = $0c;
  OlepsSiLastSave   = $0d;

  { OlepsAf values defines flags for access
    rights that are stored in the OlepsSiAccess
    property }

  OlepsAfAll        = 0;  // All access
  OlepsAfPassword   = 1;  // Password protected
  OlepsAfReadOnlyR  = 2;  // Read-only recommended
  OlepsAfReadOnlyF  = 4;  // Read-only enforced
  OlepsAfNoAnnot    = 8;  // Locked for annotations

  { OleDsi values represents a list of
    property id values for properties
    that are stored within a
    <05>DocumentSummaryInformation OLEPS stream }

  OlepsDsiCategory  = 2;
  OlepsDsiCompany   = $f;
  OlepsDsiManager   = $e;

  { Well known FMTID values.
    The abbreviations are:

      Si  = <05>SummaryInformation
      Dsi = <05>DocumentSummaryInformation }

  OlepsFmtIdSi: TOlepsFmtId = ($f29f85e0, $10684ff9, $000891ab, $d9b3272b);
  OlepsFmtIdDsi: TOlepsFmtId = ($d5cdd502, $101b2e9c, $00089793, $aef92c2b);

  { Property types }

  OlepsPtWord       = 2;
  OlepsPtInt        = 3;
  OlepsPtString     = $1e;
  OlepsPtUnicode    = $1f;
  OlepsPtTime       = $40;

type

  { TOlepsProp represents a single
    property in a property set. Such a
    property can define an author name
    or a document title.

    For the list of values for TOlepsPropId
    see OlepsSi and OlepsDsi values.

    For the list of values for TOlepsPropType
    see OlepsPt values. }

  TOlepsProp = class(TMemoryStream)
  public

    PropId: TOlepsPropId;
    PropType: TOlepsPropType;
    Offset: LongInt;

    procedure Flush(Stream: TStream);
    procedure WriteVal(Value: LongInt; BytesCount: LongInt);
    procedure WriteUCS(const Str: WideString);
    procedure WriteAnsi(const Str: AnsiString);

  end;

  { TOlepsPropSet represents a property set.
    Each property set is associated with a
    well known FMTID values. For an example,
    <05>SummaryInformation stream uses the
    [F29F85E0-4FF9-1068-AB91-08002B27B3D9] FMTID.
    For the list of values for TOlepsFmtId
    see OlepsFmtId values. }

  TOlepsPropSet = class
  public

    FmtId: TOlepsFmtId;
    Props: TList; // List of TOlepsProp
    Offset: LongInt;

    constructor Create;
    destructor Destroy; override;
    procedure Flush(Stream: TStream);
    function Add(PropId: TOlepsPropId; PropType: TOlepsPropType): TOlepsProp;

    { Adds a new non-empty string property.
      If the passed string is empty, no property
      will be added. The return value can be nil. }

    function AddUCS(PropId: TOlepsPropId; const Str: WideString): TOlepsProp;
    function AddAnsi(PropId: TOlepsPropId; const Str: AnsiString): TOlepsProp;
    function AddUnicode(PropId: TOlepsPropId; const Str: WideString): TOlepsProp;

    { Adds a property with type OlepsPtTime }

    function AddTime(PropId: TOlepsPropId; const t: FILETIME): TOlepsProp;

  end;

  { TOlepsStream represents the contents of
    a OLEPS stream. It contains of a few property sets.
    Normally, there one or two property sets
    in a stream.  }

  TOlepsStream = class
  public

    PropSets: TList; // List of TOlepsPropSet

    constructor Create;
    destructor Destroy; override;
    procedure Flush(Stream: TStream);
    function Add(const FmtId: TOlepsFmtId): TOlepsPropSet;

  end;

implementation
end.
