
{******************************************}
{                                          }
{             FastReport VCL               }
{            OpenType Fonts API            }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

{ http://www.microsoft.com/typography/otspec/ }

unit frxOTF;

{$IFDEF DEBUG}

{ Enables verification of all fields in font files,
  even those that are not needed for parsing. }

//{$DEFINE EXTENSIVE_CHECKS}

{$ENDIF} // {$IFDEF DEBUG}

{ Some fields in font files contain checksums of
  large file parts. Correct checksums are not required
  by most font consumer programs, but calculating them
  is a very slow process. }

//{$DEFINE EMIT_CHECKSUMS}

interface

{ Classes in this module fall into two groups: viewers and builders.

  A viewer represents a certain structure inside a font file and provides the
  read-only access to its contents. Viewers require minimum amount of memory
  and don't actually load contents of the font.

  A builder constructs a structure inside a font. It provides only one
  interface method - SaveToStream. It doesn't keep all the contents of the
  structure in memory, but rather it keeps minimum information about how to
  construct the stucture. Often a builder is connected with a corresponding viewer,
  from which it extracts needed information.

  Thus viewers and builders can completely change structure of a font file
  and write the changed font to another file without the need to load any considerable
  pieces of the source font. }

uses
  Classes,
  SysUtils,
  frxStorage;

type
  TGlyphIndex = Word;
  WordArray = array of Word;
  TPanose = array[1..10] of Byte;
  TFontTableTag = array[1..4] of AnsiChar;

  { Information about a glyph }

  TGlyphMetrics = record
    AdvWidth:   Word;     // advance width
    AdvHeight:  Word;     // advance height
    LeftSB:     SmallInt; // left side bearing
    TopSB:      SmallInt; // top side bearing
  end;

  { Bounding box for all glyphs }

  TFontBox = packed record
    XMin: SmallInt;
    YMin: SmallInt;
    XMax: SmallInt;
    YMax: SmallInt;
  end;

  { The header for a TTF or a OTF file }

  TFontFileHeader = packed record
    Version:    Cardinal;
    NumTables:  Word;
    Range:      Word;
    Selector:   Word;
    Shift:      Word;
  end;

  { The header for a named table }

  TTableRecordEntry = packed record
    Tag:        TFontTableTag;
    Checksum:   Cardinal;
    Offset:     Cardinal;
    Length:     Cardinal;
  end;

  { "head" table contents  }

  THeadTableHeader = packed record
    Version:    Cardinal;
    Revision:   Cardinal;
    Checksum:   Cardinal;
    Signature:  Cardinal; // must be 0x5F0F3CF5
    Flags:      Word;
    UnitsPerEm: Word;
    Created:    Int64;
    Modified:   Int64;
    FontBox:    TFontBox;
    Style:      Word;
    Smallest:   Word;
    DirHint:    SmallInt;
    LongOffset: SmallInt; // either 1 or 0
    Format:     SmallInt;
  end;

  { "maxp" table header }

  TMaxpTableHeader = packed record
    Version:    Cardinal;
    NumGlyphs:  Word;
    MaxPoints:  Word;
    MaxContours:Word;
    MaxCompPts: Word;
    MaxCompCtrs:Word;
    MaxZones:   Word;
    MaxTwPoints:Word;
    MaxStorage: Word;
    MaxFunc:    Word;
    MaxInst:    Word;
    MaxStack:   Word;
    MaxInstSize:Word;
    MaxComp:    Word;
    MaxDepth:   Word;
  end;

  { Each glyphs in the "glyf" table is begun with this header }

  TGlyphHeader = packed record
    NumContours:SmallInt;
    XMin:       SmallInt;
    YMin:       SmallInt;
    XMax:       SmallInt;
    YMax:       SmallInt;
  end;

  { Composite glyphs consists of a list of entries
    which are started with this header. }

  TGlyphEntryHeader = packed record
    Flags:      Word;
    Glyph:      Word;
  end;

  { "cmap" table header }

  TCharmapTableHeader = packed record
    Version:    Word;
    NumTables:  Word;
  end;

  { a "cmap" subtable record entry }

  TCharmapSubtableRecordEntry = packed record
    Platform:   Word;
    Encoding:   Word;
    Offset:     cardinal;
  end;

  { Header of the charmap subtable of the format 0 }

  TAsciiCharmapTableHeader = packed record
    Format:   Word;
    Length:   Word;
    Lang:     Word;
  end;

  { Header for the charmap subtable of the format 4 }

  TSegCharmapTableHeader = packed record
    Format:     Word;
    Length:     Word;
    Lang:       Word;
    SegCount2:  Word; // 2 * SegCount
    Range:      Word; // 2 * 2 ^ floor(log2(SegCount))
    Selector:   Word; // floor(log2(SegCount))
    Shift:      Word; // 2 * SegCount - Range
  end;

  { TTC file header }

  TFontCollectionFileHeader = packed record
    Tag:      array[1..4] of AnsiChar;
    Version:  Cardinal;
    NumFonts: Cardinal;
  end;

  { Contents of "hhea" and "vhea" tables }

  TMetricsTableHeader = packed record
    Version:    Cardinal;
    Ascent:     SmallInt;
    Descent:    SmallInt;
    LineGap:    SmallInt;
    MaxWidth:   Word;
    MinSB1:     SmallInt;
    MinSB2:     SmallInt;
    MaxExtent:  SmallInt;
    SlopeRise:  SmallInt;
    SlopeRun:   SmallInt;
    Offset:     SmallInt;
    Reserved:   array[1..4] of SmallInt;
    Metric:     SmallInt;
    NumMetrics: Word;
  end;

  { Contents of the "OS/2" table.
    Version 1. }

  TWinMetrics1TableHeader = packed record
    Version:    Word;
    AvgWidth:   SmallInt;
    Weight:     Word;
    Width:      Word;
    Rights:     Word;
    SubXSize:   SmallInt;
    SubYSize:   SmallInt;
    SubXOffset: SmallInt;
    SubYOffset: SmallInt;
    SupXSize:   SmallInt;
    SupYSize:   SmallInt;
    SupXOffset: SmallInt;
    SupYOffset: SmallInt;
    StrikeSize: SmallInt;
    StrikePos:  SmallInt;
    Family:     SmallInt;
    Panose:     TPanose;
    UniRange:   array[1..4] of Cardinal;
    Vendor:     array[1..4] of AnsiChar;
    Selection:  Word;
    FirstChar:  Word;
    LastChar:   Word;
    Ascender:   Word;
    Descender:  Word;
    LineGap:    Word;
    WinAscent:  Word;
    WinDescent: Word;
    PageRange:  array[1..2] of Cardinal;
    {
    Height:     SmallInt;
    CapHeight:  SmallInt;
    DefChar:    Word;
    BreakChar:  Word;
    MaxContext: Word;
    }
  end;

  { "name" table header }

  TStringTableHeader = packed record
    Format:     Word;
    Count:      Word;
    Offset:     Word;
  end;

  TStringRecord = packed record
    Platform:   Word;
    Encoding:   Word;
    Language:   Word;
    Id:         Word; // copyright notice, font family name, etc.
    Size:       Word; // in bytes
    Offset:     Word;
  end;

  TStringTableEntry = record
    Platform: Word;
    Encoding: Word;
    Language: Word;
    Id:       Word;
  end;

  { Conforms the Visitor pattern }

  TVisitor = class;

  { Represents a piece of a file with the font }

  TView = class
  private
    FStream: TStream;
    FOwnStream: Boolean;

    function GetSize: Cardinal;
  protected
    procedure Init; virtual;
  public
    constructor Create(Stream: TStream; OwnStream: Boolean);
    destructor Destroy; override;

    procedure Accept(v: TVisitor); virtual;

    property Stream: TStream read FStream;
    property Size: Cardinal read GetSize;
  end;

  { Represents a named table in a OpenType font file.
    Examples of this table: cmap, head, fpgm. }

  TFontTableView = class(TView);

  { "OS/2" table }

  TWinMetricsTableView = class(TFontTableView)
  private
    function GetVersion: Word;
  protected
    procedure Init; override;
  public
    procedure Accept(v: TVisitor); override;
    property Version: Word read GetVersion;
  end;

  TWinMetrics1TableView = class(TWinMetricsTableView)
    function GetHeader: TWinMetrics1TableHeader;
  protected
    procedure Init; override;
  public
    procedure Accept(v: TVisitor); override;
    property Header: TWinMetrics1TableHeader read GetHeader;
  end;

  { Represents a subtable in the "cmap" table }

  TCharmapSubtableView = class(TView)
  private
    function GetFormat: Word;
  protected
    function GetLanguage: Word; virtual; abstract;
    function GetGlyphIndex(Char: Word): TGlyphIndex; virtual; abstract;
  public
    class function CreateSubtable(Stream: TStream; OwnStream: Boolean): TCharmapSubtableView;

    procedure Accept(v: TVisitor); override;

    { Some subtables have the Length field at one offset,
      others have it at another offset. }

    class function GetSubtableLength(Stream: TStream): Cardinal;

    property Format: Word read GetFormat;
    property Language: Word read GetLanguage;
    property Glyph[Char: Word]: TGlyphIndex read GetGlyphIndex; default;
  end;

  { Represents a charmap subtable that maps all charcodes to the zero glyph.
    This class is used when it's needed to represent a subtable of an
    unknown format. }

  TEmptyCharmapTableView = class(TCharmapSubtableView)
  protected
    function GetLanguage: Word; override;
    function GetGlyphIndex(Char: Word): TGlyphIndex; override;
  end;

  { Represents a charmap subtable of the format 0 }

  TAsciiCharmapTableView = class(TCharmapSubtableView)
  private
    function GetHeader: TAsciiCharmapTableHeader;
    procedure CheckHeader(const h: TAsciiCharmapTableHeader);
  protected
    procedure Init; override;
    function GetLanguage: Word; override;
    function GetGlyphIndex(Char: Word): TGlyphIndex; override;
  public
    procedure Accept(v: TVisitor); override;
  end;

  { Represents a charmap subtable of the format 4 }

  TSegCharmapTableView = class(TCharmapSubtableView)
  private
    FSegNum: Integer;
    FGlyphNum: Integer;
    FFirst: WordArray;
    FLast: WordArray;
    FDelta: WordArray;
    FOffset: WordArray;

    function GetHeader: TSegCharmapTableHeader;
    procedure CheckHeader(const h: TSegCharmapTableHeader);

    function ReadWord(Offset: Cardinal): Word;
    function GetWord(var w: WordArray; i: Integer; p: Cardinal): Word;

    function GetFirst(i: Integer): Word;
    function GetLast(i: Integer): Word;
    function GetOffset(i: Integer): Word;
    function GetGlyph(i: Integer): Word;
    function GetDelta(i: Integer): Word;

    { Finds optimal set of segments for a set of used chars which
      produces the smallest charmap.

      Chars[i]  = charcode of the i-th char; this array must be ascending
      Glyphs[i] = glyph index the corresponds to the i-th char
      Seg[i]    = length of the i-th segment

      Result    = number of bytes needed for all these segments }

    class function FindOptimalMapping(const Chars, Glyphs: WordArray; out Seg: WordArray): Word;
  protected
    procedure Init; override;
    function GetLanguage: Word; override;
    function GetGlyphIndex(Char: Word): TGlyphIndex; override;
  public
    procedure Accept(v: TVisitor); override;

    property SegCount: Integer read FSegNum;
    property First[i: Integer]: Word read GetFirst;
    property Last[i: Integer]: Word read GetLast;
    property Delta[i: Integer]: Word read GetDelta;
    property Offset[i: Integer]: Word read GetOffset;
    property Glyph[i: Integer]: Word read GetGlyph;
  end;

  { "cmap" table }

  TCharmapTableView = class(TFontTableView)
  private
    FSubtables: array of TCharmapSubtableView;

    function GetSubtablesCount: Integer;
    function GetSubtable(i: Integer): TCharmapSubtableView;
    function GetPlatform(i: Integer): Word;
    function GetEncoding(i: Integer): Word;
    function GetHeader: TCharmapTableHeader;
    procedure LoadSubtable(i: Integer);
  protected
    procedure Init; override;
  public
    destructor Destroy; override;

    procedure Accept(v: TVisitor); override;
    function GetGlyphIndex(Char: Word): Cardinal;
    function GetSubtableRecordEntry(i: Integer): TCharmapSubtableRecordEntry;

    property SubtablesCount: Integer read GetSubtablesCount;
    property Subtable[i: Integer]: TCharmapSubtableView read GetSubtable;
    property Platform[i: Integer]: Word read GetPlatform;
    property Encoding[i: Integer]: Word read GetEncoding;
  end;

  { "head" table }

  THeadTableView = class(TFontTableView)
  private
    function GetHeader: THeadTableHeader;
  protected
    procedure Init; override;
  public
    procedure Accept(v: TVisitor); override;
    property Header: THeadTableHeader read GetHeader;
  end;

  { "hhea" and "vhea" tables }

  TMetricsInfoTableView = class(TFontTableView)
  private
    function GetHeader: TMetricsTableHeader;
  protected
    procedure Init; override;
  public
    property Header: TMetricsTableHeader read GetHeader;
  end;

  { "hmtx" and "vmtx" tables }

  TMetricsTableView = class(TFontTableView)
  private
    FNumMetrics: Cardinal;
    FNumGlyphs: Cardinal;

    function GetDim(i: Integer): Word;
    function GetSB(i: Integer): SmallInt;
  public
    // NumMetrics comes from hhea or vhea; NumGlyphs comes from maxp.
    constructor Create(Stream: TStream; Own: Boolean; NumMetrics, NumGlyphs: Cardinal);

    procedure Accept(v: TVisitor); override;

    property Count: Cardinal read FNumGlyphs;
    property Dim[i: Integer]: Word read GetDim;   // Width for hmtx; Height for vmtx
    property SB[i: Integer]: SmallInt read GetSB; // LSB for hmtx; TSB for vmtx
  end;

  { "name" table }

  TStringTableView = class(TFontTableView)
  private
    FCount: Word;
    FStorage: Word; // offset to the strings storage

    function GetHeader: TStringTableHeader;
    function GetEntry(i: Integer): TStringTableEntry;
    function getString(i: Integer): AnsiString;
    function GetStringRecord(i: Integer): TStringRecord;
    procedure CheckHeader(const h: TStringTableHeader);
    procedure CheckStringRecord(const h: TStringRecord);
  protected
    procedure Init; override;
  public
    procedure Accept(v: TVisitor); override;

    property Count: Word read FCount;
    property Entry[i: Integer]: TStringTableEntry read GetEntry;
    property Strings[i: Integer]: AnsiString read GetString; default;
  end;

  { "maxp" table.
    See http://www.microsoft.com/typography/otspec/maxp.htm for details.
    There're two versions of this table, but both have the same two first
    fields: Version (4 bytes) and NumGlyphs (2 bytes). }

  TMaxpTableView = class(TFontTableView)
  private
    function GetNumGlyphs: Word;
    function GetHeader: TMaxpTableHeader;
  protected
    procedure Init; override;
  public
    property Header: TMaxpTableHeader read GetHeader;
    property NumGlyphs: Word read GetNumGlyphs;
  end;

  { "loca" table.
    See for http://www.microsoft.com/typography/otspec/loca.htm details. }

  TGlyphPosTableView = class(TFontTableView)
  private
    FLongOffsets: Boolean;
    FNumOffsets: Cardinal;

    function GetGlyphOffset(i: Integer): Cardinal;
    function GetGlyphLength(i: Integer): Cardinal;
  public
    constructor Create(Stream: TStream; OwnStream, LongOffsets: Boolean;
      NumOffsets: Cardinal);

    procedure Accept(v: TVisitor); override;

    property LongOffsets: Boolean read FLongOffsets;
    property NumOffsets: Cardinal read FNumOffsets;

    property GlyphOffset[i: Integer]: Cardinal read GetGlyphOffset;
    property GlyphLength[i: Integer]: Cardinal read GetGlyphLength;
  end;

  { "glyf" table }

  TGlyphTableView = class(TFontTableView);

  { Represents a glyph.
    Glyphs can be wither simple or composite.
    Composite glyphs are constructed from other glyphs,
    and refer to these other glyphs by indices. }

  TGlyphView = class(TView)
  protected
    function GetWidth: Integer; virtual; abstract;
    function GetHeight: Integer; virtual; abstract;
    function IsComposite: Boolean; virtual; abstract;
  public
    class function CreateGlyph(Stream: TStream; Own: Boolean): TGlyphView;

    property Width: Integer read GetWidth;
    property Height: Integer read GetHeight;
    property Composite: Boolean read IsComposite;
  end;

  { Represents a glyph whose size is zero }

  TNullGlyphView = class(TGlyphView)
  protected
    function GetWidth: Integer; override;
    function GetHeight: Integer; override;
    function IsComposite: Boolean; override;
  end;

  { Introducing this intermediate class is made because
    fonts can contain glyphs of zero size. Such glyphs are
    invalid due to the specification (any glyph begins with
    a 10 byte header), but all font-consumer programs happily
    accept such glyphs and many standard font files contain
    such glyphs (e.g. the space glyph). So, this situation must be
    handled in some way. This class represents a glyph of non-zero
    size. }

  TBaseGlyphView = class(TGlyphView)
  private
    function GetHeader: TGlyphHeader;
    procedure CheckHeader(const h: TGlyphHeader);
  protected
    function IsComposite: Boolean; override;
    function GetWidth: Integer; override;
    function GetHeight: Integer; override;
    procedure Init; override;
  public
    property Header: TGlyphHeader read GetHeader;
  end;

  { Simple glyph }

  TSimpleGlyphView = class(TBaseGlyphView);

  { Composite glyph }

  TCompositeGlyphView = class(TBaseGlyphView)
  private
    FEntryOffset: array of Cardinal;

    function GetCount: Cardinal;
    function GetGlyph(i: Cardinal): TGlyphIndex;
    function GetEntryOffset(i: Integer): Cardinal;
  protected
    procedure Init; override;
  public
    property Count: Cardinal read GetCount;
    property Component[i: Cardinal]: TGlyphIndex read GetGlyph;
    property EntryOffset[i: Integer]: Cardinal read GetEntryOffset;
  end;

  { Represents a TTF or a OTF file.
    Consists of several named tables. }

  TFontView = class(TView)
  private
    FFontFile: TStream;
    FTables: array of TFontTableView;
    FGlyphsCount: Cardinal;
    FGlyphs: TObjList;
    FCurCharmap: TCharmapSubtableView; // currently used mapping; can be nil

    FCharmapTable: TCharmapTableView;
    FGlyphPosTable: TGlyphPosTableView;
    FHMetricstable: TMetricsTableView;
    FGlyphTable: TGlyphTableView;

    function GetString(NameId: Integer): AnsiString;
    function GetFontName: AnsiString;
    function GetFontStyleName: AnsiString;

    function CreateTable(Name: AnsiString; Stream: TStream; Own: Boolean): TFontTableView;
    procedure LoadTable(i: Integer);
    function FindTable(Name: AnsiString): Integer;

    function GetGlyph(g: TGlyphIndex): TGlyphView;
    function GetGlyphMetrics(g: TGlyphIndex): TGlyphMetrics;

    function GetTablesCount: Integer;
    function GetGlyphsCount: Cardinal;

    function GetCharmapTable: TCharmapTableView;
    function GetGlyphPosTable: TGlyphPosTableView;
    function GetHMetricsTable: TMetricsTableView;
    function GetHMetricsInfoTable: TMetricsInfoTableView;
    function GetVMetricsTable: TMetricsTableView;
    function GetVMetricsInfoTable: TMetricsInfoTableView;
    function GetGlyphTable: TGlyphTableView;
    function GetHeadTable: THeadTableView;
    function GetMaxpTable: TMaxpTableView;
    function GetStringsTable: TStringTableView;
    function GetWinMetricsTable: TWinMetricsTableView;

    function GetHeader: TFontFileHeader;
    function GetHeadTableHeader: THeadTableHeader;
    function GetHorzHeader: TMetricsTableHeader;
    function GetVertHeader: TMetricsTableHeader;

    procedure CheckHeader(const h: TFontFileHeader);
    procedure CheckTableHeader(const h: TTableRecordEntry);
  protected
    procedure Init; override;
  public

    { "Stream" contains data of this font. "FontFile" contains the file either
      with this font or with a fonts collection (.ttc files). In the case when
      the file contains this single font, FontFile equals Stream. }

    constructor Create(FontFile, Stream: TStream; OwnStream: Boolean);
    destructor Destroy; override;

    procedure Accept(v: TVisitor); override;

    { Selects a mapping from charcodes to glyph indices.
      If no mapping is selected, then GetGlyphIndex(c) will try
      to find a mapping that is capable to map the charcode c. }

    function SelectMapping(Platform, Encoding: Integer): Boolean;

    function GetGlyphIndex(Char: Word): Cardinal;
    function TableExists(Name: AnsiString): Boolean;

    function GetTable(Index: Integer): TFontTableView; overload;
    function GetTableHeader(i: Integer): TTableRecordEntry;
    function GetTable(Name: AnsiString): TFontTableView; overload;

    property TablesCount: Integer read GetTablesCount;

    property GlyphsCount: Cardinal read GetGlyphsCount;
    property Glyph[g: TGlyphIndex]: TGlyphView read GetGlyph;
    property GlyphMetrics[g: TGlyphIndex]: TGlyphMetrics read GetGlyphMetrics;

    property CharmapTable: TCharmapTableView read GetCharmapTable;
    property GlyphPosTable: TGlyphPosTableView read GetGlyphPosTable;
    property HMetricsTable: TMetricsTableView read GetHMetricsTable;
    property HMetricsInfoTable: TMetricsInfoTableView read GetHMetricsInfoTable;
    property VMetricsTable: TMetricsTableView read GetVMetricsTable;
    property VMetricsInfoTable: TMetricsInfoTableView read GetVMetricsInfoTable;
    property GlyphTable: TGlyphTableView read GetGlyphTable;
    property HeadTable: THeadTableView read GetHeadTable;
    property StringsTable: TStringTableView read GetStringsTable;
    property MaxpTable: TMaxpTableView read GetMaxpTable;
    property WinMetricsTable: TWinMetricsTableView read GetWinMetricsTable;

    property FamilyName: AnsiString read GetFontName;
    property SubfamilyName: AnsiString read GetFontStyleName;
  end;

  { TTC file - a collection of TrueType fonts }

  TFontCollectionView = class(TView)
  private
    FFonts: array of TFontView; // loaded on demand

    function GetHeader: TFontCollectionFileHeader;
    function GetFontsCount: Cardinal;
    procedure CheckHeader(const h: TFontCollectionFileHeader);
    function GetFont(i: Integer): TFontView;
    procedure LoadFont(i: Integer);
  protected
    procedure Init; override;
  public
    destructor Destroy; override;

    procedure Accept(v: TVisitor); override;

    property FontsCount: Cardinal read GetFontsCount;
    property Font[i: Integer]: TFontView read GetFont; default;
  end;

  { Exceptions }

  EOpenTypeException = class(Exception);

  EInvalidOpenTypeFont          = class(EOpenTypeException);
  ETableDoesNotExist            = class(EOpenTypeException);
  EUnknownCharmapSubtableFormat = class(EOpenTypeException);
  EGlyphDoesNotExist            = class(EOpenTypeException);
  ENotEnoughDataForGlyph        = class(EOpenTypeException);
  EGlyphIndexOutOfBounds        = class(EOpenTypeException);
  EFormatNotSupported           = class(EOpenTypeException);

  { Visitor interface }

  TVisitor = class
  public
    procedure Visit(Obj: TObject);                overload; virtual;
    procedure Visit(Obj: TFontView);              overload; virtual;
    procedure Visit(Obj: TCharmapTableView);      overload; virtual;
    procedure Visit(Obj: TSegCharmapTableView);   overload; virtual;
    procedure Visit(Obj: TGlyphPosTableView);     overload; virtual;
    procedure Visit(Obj: TFontCollectionView);    overload; virtual;
    procedure Visit(Obj: TStringTableView);       overload; virtual;
    procedure Visit(Obj: TMetricsTableView);      overload; virtual;
    procedure Visit(Obj: TAsciiCharmapTableView); overload; virtual;
    procedure Visit(Obj: THeadTableView);         overload; virtual;
    procedure Visit(Obj: TCompositeGlyphView);    overload; virtual;
    procedure Visit(Obj: TSimpleGlyphView);       overload; virtual;
    procedure Visit(Obj: TNullGlyphView);         overload; virtual;
    procedure Visit(Obj: TWinMetrics1TableView);  overload; virtual;
  end;

  { Builder interface }

  TBuilder = class
  public
    procedure SaveToStream(Stream: TStream); virtual; abstract;
  end;

  { This is a helper builer. It writes the holded stream when
    someone calls its SaveToStream. This class allows to turn any
    data stream into a builder. }

  TStreamBuilder = class(TBuilder)
  private
    FStream: TStream;
    FOwn: Boolean;
  public
    constructor Create(Stream: TStream; Own: Boolean = False);
    constructor CreateWithStream;
    destructor Destroy; override;

    procedure SaveToStream(Stream: TStream); override;

    property Stream: TStream read FStream;
  end;

  { Builds a font table }

  TFontTableBuilder = class(TBuilder);

  { Builds the "cmap" table }

  TCharmapTableBuilder = class(TFontTableBuilder)
  private
    FPlatforms: array of Word;
    FEncodings: array of Word;
    FSubtables: array of TBuilder;

    procedure WriteHeader(Stream: TStream);
    procedure WriteSubHeader(Stream: TStream; i: Integer; Ptr: Cardinal);
  public
    destructor Destroy; override;

    procedure AddSubtable(Subtable: TBuilder; Platform, Encoding: Word);
    procedure SaveToStream(Stream: TStream); override;
  end;

  { Builds a charmap subtable }

  TCharmapSubtableBuilder = class(TBuilder);

  { Builds a charmap subtable of the format 4 }

  TSegCharmapTableBuilder = class(TCharmapSubtableBuilder)
  private
    FLang: Word;
    FFirst: array of Word;
    FLast: array of Word;
    FDelta: array of Word;

    procedure WriteHeader(Stream: TStream);
    procedure WriteZero(Stream: TStream; n: Integer);
    procedure WriteWords(Stream: TStream; const w: array of Word);
  public
    constructor Create(Lang: Word);

    procedure AddSegment(First, Last: Word; Delta: Word);
    procedure SaveToStream(Stream: TStream); override;
  end;

  { Builds a charmap subtable of the format 0 }

  TAsciiCharmapTableBuilder = class(TCharmapSubtableBuilder)
  private
    FGlyphs: array[0..255] of Byte;
    FLang: Word;

    procedure SetGlyphIndex(Char, Glyph: Word);
    procedure WriteHeader(Stream: TStream);
  public
    constructor Create(Lang: Word);
    procedure SaveToStream(Stream: TStream); override;
    property GlyphIndex[Char: Word]: Word write SetGlyphIndex;
  end;

  { Builds "hmtx" and "vmtx" tables }

  TMetricsTableBuilder = class(TFontTableBuilder)
  private
    FDim: array of Word;
    FSB: array of Word;
  public
    procedure AddMetric(Dim: Word; SB: SmallInt);
    procedure SaveToStream(Stream: TStream); override;
  end;

  { Builds the "name" table }

  TStringTableBuilder = class(TFontTableBuilder)
  private
    FItems: array of TStringTableEntry;
    FStrings: array of AnsiString;
    FStrRef: array of Integer;

    procedure WriteHeader(Stream: TStream; FSO: Cardinal);
    procedure WriteRecord(Stream: TStream; i: Integer; SO: Cardinal);
    procedure WriteString(Stream: TStream; const s: AnsiString);

    function FindString(const s: AnsiString): Integer;
  public
    procedure AddString(const s: AnsiString; Platform, Encoding, Language, Id: Word);
    procedure SaveToStream(Stream: TStream); override;
  end;

  { Builds a glyph }

  TGlyphBuilder = class(TBuilder);

  { Builds a simple glyph }

  TSimpleGlyphBuilder = class(TGlyphBuilder);

  { Builds a composite glyph }

  TCompositeGlyphBuilder = class(TGlyphBuilder)
  private
    FGlyph: TCompositeGlyphView;
    FComponents: array of TGlyphIndex;

    procedure LoadComponents;
    procedure SetComponent(i: Cardinal; Value: TGlyphIndex);
  public
    constructor Create(Glyph: TCompositeGlyphView);
    procedure SaveToStream(Stream: TStream); override;

    property Component[i: Cardinal]: TGlyphIndex write SetComponent;
  end;

  { Builds the "loca" table }

  TGlyphPosTableBuilder = class(TFontTableBuilder)
  private
    FOffset: array of Cardinal;

    procedure AddOffset(Offset: Cardinal);
    function GetLongOffsets: Boolean;
    function GetMaxOffset: Cardinal;
  public
    constructor Create;

    function AddGlyph(Len: Cardinal): Cardinal;
    procedure SaveToStream(Stream: TStream); override;

    property LongOffsets: Boolean read GetLongOffsets;
  end;

  { Builds the "glyf" table }

  TGlyphTableBuilder = class(TFontTableBuilder)
  private
    FGlyphs: array of TBuilder;
    FLengths: array of Cardinal;
  public
    destructor Destroy; override;

    procedure AddGlyph(Glyph: TBuilder; DesiredLength: Cardinal);
    procedure SaveToStream(Stream: TStream); override;
  end;

  { Builds a font (a .ttf file) }

  TFontBuilder = class(TBuilder)
  private
    FFontFileBase: Cardinal;
    FFont: TFontView;
    FTags: array of TFontTableTag;
    FTables: array of TBuilder;
    FCompChecksums: Boolean;

    FCharmap: TOrderedMap;  // maps charcodes to glyph indices
    FGUsed: TOrderedMap;    // set of original indices of all used glyphs
    FGRemap: TOrderedMap;   // maps original glyph indices to new ones
    FGlyphs: WordArray;     // list of original indices for all used glyphs
    FChars: WordArray;      // list of all used charcodes

    function RemapGlyph(g: TGlyphIndex): TGlyphIndex; // maps original index to index in repacked font
    function MapCharToGlyph(c: Word): TGlyphIndex;    // maps charcode to glyph index in original font

    function CreateMetricsInfoTable(View: TMetricsInfoTableView; NumMetrics: Word): TBuilder;
    function CreateSegCharmapTable(Lang: Word): TSegCharmapTableBuilder;
    function CreateAsciiCharmapTable(Lang: Word): TAsciiCharmapTableBuilder;
    function CreateCharmapTable: TCharmapTableBuilder;
    function CreateMetricsTable(View: TMetricsTableView): TMetricsTableBuilder;
    function CreateMaxpTable(NumGlyphs: Word): TBuilder;
    function CreateHeadTable(LongOffsets: Boolean): TBuilder;
    function CreateStringTable: TStringTableBuilder;

    procedure RepackGlyphs(out Glyf, Loca: TFontTableBuilder; out LongOffsets: Boolean);

    procedure AddDescendantGlyphs(g: TGlyphIndex);
    procedure BuildGlyphsRemapping; // fills in FGRemap, FGlyphs
    procedure BuildCharsList; // fills in FChars

    procedure WriteHeader(Stream: TStream);
    procedure WriteTableRecord(Stream: TStream; Index, Offset, Len, Checksum: Cardinal);

    procedure CopyTableFromSrcFont(Name: TFontTableTag);
    procedure AddTable(Tag: TFontTableTag; Table: TBuilder);
    procedure SaveTablesToStream(Stream: TStream);
  public
    constructor Create(Font: TFontView);
    destructor Destroy; override;

    procedure AddCharset(Charset: TBitArray);
    procedure AddChar(c: Word);
    procedure SaveToStream(Stream: TStream); override;

    function GetUsedGlyphs: TBitArray;

    property ComputeChecksums: Boolean read FCompChecksums write FCompChecksums;
    property FontFileBase: Cardinal read FFontFileBase write FFontFileBase;
  end;

  { Builds a fonts collection (a .ttc file) }

  TFontCollectionBuilder = class(TBuilder)
  private
    FFonts: array of TFontBuilder;
    procedure WriteHeader(Stream: TStream);
  public
    destructor Destroy; override;
    procedure AddFont(Font: TFontBuilder);
    procedure SaveToStream(Stream: TStream); override;
  end;

{ Creates either a font (TFontView) or a font collection (TFontCollectionView).
  If the stream contains a malformed data - raises an exception. }

function OpenFont(Stream: TStream): TView;

{ Extracts from a font a specified set of characters.

  If Font is TFontView, then a builder of a single font (.ttf file) will returned.

  If Font is TFontCollectionView then a builder of a fonts collection
  (.ttc file) will be returned. The returned collection will have
  the same number of fonts as in Font and each font in the result
  will contain the specified set of characters from the corresponding font. }

function PackFont(Font: TView; const UsedChars: TBitArray): TBuilder;

implementation
end.

