{******************************************}
{                                          }
{             FastReport VCL               }
{            PDFView support               }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxPDFium;

{$I frx.inc}

interface

uses
{$IFNDEF Linux}
  Windows,
{$ELSE}
  LCLType, LCLIntf, LCLProc,
{$ENDIF}
  SysUtils, Classes, Graphics, Contnrs, SyncObjs, frCore;

type
  TPdfDocument = class;

  TPdfDocumentSaveOption = (
    dsoIncremental    = 1,
    dsoNoIncremental  = 2,
    dsoRemoveSecurity = 3
  );

  TPdfDrawOption = (
    pdoAnnotations,
    pdoLCDOptimized,
    pdoNoNativeText,
    pdoGrayScale,
    pdoNoCatch,
    pdoLimitedImageCacheSize,
    pdoForceHalftone,
    pdoPrinting,
    pdoUseMetafile,
    pdoPrintTextWithGDI,
    pdoFastDraw
  );
  TPdfDrawOptions = set of TPdfDrawOption;

  PFPDF_FILEWRITE = ^FPDF_FILEWRITE;
  FPDF_FILEWRITE = record
    version: Integer;
    WriteBlock: function(pThis: PFPDF_FILEWRITE; pData: Pointer; size: LongWord): Integer; cdecl;
  end;
  PFPdfFileWrite = ^TFPdfFileWrite;
  TFPdfFileWrite = FPDF_FILEWRITE;

  PFPDFFileWriteEx = ^TFPDFFileWriteEx;
  TFPDFFileWriteEx = record
    Inner: TFPDFFileWrite;
    Stream: TStream;
  end;

  FPDF_BOOL  = Integer;
  FPDF_DWORD = LongWord;
  FPDF_BYTESTRING = PAnsiChar;
  __FPDF_PTRREC = record end;
  __PFPDF_PTRREC = ^__FPDF_PTRREC;
  FPDF_DOCUMENT          = type __PFPDF_PTRREC;
  FPDF_FORMHANDLE        = type __PFPDF_PTRREC;
  FPDF_PAGE              = type __PFPDF_PTRREC;
  FPDF_BITMAP            = type __PFPDF_PTRREC;

  TPdfPage = class
  private
    FDocument: TPdfDocument;
    FPage: FPDF_PAGE;
    FWidth: Single;
    FHeight: Single;
    function GetDrawFlags(const Options: TPdfDrawOptions): Integer;
    procedure AfterOpen;
  public
    constructor Create(ADocument: TPdfDocument; APage: FPDF_PAGE);
    procedure Draw(ACanvas: TCanvas; X, Y, Width, Height, Rotation: Integer; const Options: TPdfDrawOptions; ADirectDraw: Boolean);
    function IsLoaded: Boolean;
    procedure Open;
    procedure Close;
    property Width: Single read FWidth;
    property Height: Single read FHeight;
  end;

{$IFNDEF FPC}
  TfrEMFDrawState = class
  private
    FNeedPathScaleDown: Boolean;
    FPenStyle: DWORD;
    FBrushStyle: UINT;
    FColor: COLORREF;
    FHatch: DWORD;
  public
    property NeedPathScaleDown: Boolean read FNeedPathScaleDown write FNeedPathScaleDown;
    property PenStyle: DWORD read FPenStyle write FPenStyle;
    property BrushStyle: UINT read FBrushStyle write FBrushStyle;
    property Color: COLORREF read FColor write FColor;
    property Hatch: DWORD read FHatch write FHatch;
  end;
{$ENDIF}

  TPdfDocument = class
  private
    FDocument: FPDF_DOCUMENT;
    FPages: TObjectList;
    FBuffer: PByte;
    FPageIndex: Integer;
    FErrorText: String;
    StrongError: Boolean;
{$IFNDEF FPC}
    FDrawState: TfrEMFDrawState;
{$ENDIF}
    procedure InitPDFium;
    procedure InternLoadFromMem(Buffer: PByte; Size: frInteger; const Password: String);

    function GetActive: Boolean;
    function GetPageCount: Integer;
    function IsPageValid: Boolean;
    function GetCurrentPage: TPdfPage;
    function GetPage(Index: Integer): TPdfPage;
    procedure SetPageIndex(Value: Integer);
    procedure SetErrorText(v: String);
    procedure PageContentChanged(Closing: Boolean);
    function IsPageLoaded(PageIndex: Integer): Boolean;
    procedure CheckActive;
  public

    constructor Create;
    destructor Destroy; override;
    procedure Close;
    procedure Clear;
    procedure LoadFromStream(AStream: TStream; const Password: String = '');
    procedure Paint(Canvas: Tcanvas; Rect: TRect; Rotation: Integer; DrawOptions: TPdfDrawOptions);
    function ReloadPage(APage: TPdfPage): FPDF_PAGE;

    function NewDocument: Boolean;
    function ImportPages(Source: TPdfDocument; const Range: string = ''; Index: Integer = -1): Boolean;
    procedure SaveToStream(Stream: TStream; Option: TPdfDocumentSaveOption = dsoRemoveSecurity; FileVersion: Integer = -1);

    property Active: Boolean read GetActive;
    property PageIndex: Integer read FPageIndex write SetPageIndex;
    property PageCount: Integer read GetPageCount;
    property CurrentPage: TPdfPage read GetCurrentPage;
    property Pages[Index: Integer]: TPdfPage read GetPage;
    property ErrorText: String read FErrorText write SetErrorText;
end;

procedure RaiseLastPdfError;

const
{$IFDEF Linux}
  {$IFDEF CPU64}
  pdfium_dll = 'frx_pdfium_64.so';
  {$ELSE}
  pdfium_dll = 'frx_pdfium.so';
  {$ENDIF}
{$ELSE}
  {$IFDEF WIN64}
    pdfium_dll = 'frx_pdfium_64.dll';
  {$ELSE}
    pdfium_dll = 'frx_pdfium.dll';
  {$ENDIF}
{$ENDIF}
  EMF_PPI = 600;

type
  TPDF_InitLibrary = procedure(); {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDF_CloseDocument = procedure(document: FPDF_DOCUMENT); {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDF_LoadMemDocument64 = function(data_buf: Pointer; size: frInteger; password: FPDF_BYTESTRING): FPDF_DOCUMENT; {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDF_LoadPage = function(document: FPDF_DOCUMENT; page_index: Integer): FPDF_PAGE; {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDF_GetPageCount = function(document: FPDF_DOCUMENT): Integer; {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDF_RenderPage = procedure(DC: HDC; page: FPDF_PAGE; start_x, start_y, size_x, size_y: Integer;
    rotate: Integer; flags: Integer); {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDFBitmap_Create = function(width, height: Integer; alpha: Integer): FPDF_BITMAP; {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDFBitmap_FillRect = procedure(bitmap: FPDF_BITMAP; left, top, width, height: Integer; color: FPDF_DWORD); {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDF_RenderPageBitmap = procedure(bitmap: FPDF_BITMAP; page: FPDF_PAGE; start_x, start_y, size_x, size_y: Integer; rotate: Integer; flags: Integer); {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDFBitmap_GetBuffer = function(bitmap: FPDF_BITMAP): Pointer; {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDFBitmap_Destroy = procedure(bitmap: FPDF_BITMAP); {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDF_GetPageWidthF = function(page: FPDF_PAGE): Single; {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDF_GetPageHeightF = function(page: FPDF_PAGE): Single; {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDFPage_HasTransparency = function(page: FPDF_PAGE): FPDF_BOOL; {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDFPage_GetRotation = function(page: FPDF_PAGE): Integer; {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDF_ClosePage = procedure(page: FPDF_PAGE); {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDF_ImportPages = function(dest_doc, src_doc: FPDF_DOCUMENT; pagerange: FPDF_BYTESTRING; index: Integer): FPDF_BOOL; {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDF_CreateNewDocument = function: FPDF_DOCUMENT; {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDF_SaveAsCopy = function(document: FPDF_DOCUMENT; pFileWrite: PFPDF_FILEWRITE; flags: FPDF_DWORD): FPDF_BOOL; {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDF_SaveWithVersion = function(document: FPDF_DOCUMENT; pFileWrite: PFPDF_FILEWRITE;
    flags: FPDF_DWORD; fileVersion: Integer): FPDF_BOOL; {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDF_GetLastError = function(): LongWord; {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};
  TPDF_SetPrintTextWithGDI = procedure(use_gdi: LongBool); {$IFDEF DLLEXPORT}stdcall{$ELSE}cdecl{$ENDIF};

var
  PdfiumModule: HMODULE;
  PDFiumInitCritSect: TCriticalSection;
//
  FPDF_InitLibrary: TPDF_InitLibrary;
  FPDF_CloseDocument: TPDF_CloseDocument;
  FPDF_LoadMemDocument64: TPDF_LoadMemDocument64;
  FPDF_LoadPage: TPDF_LoadPage;
  FPDF_GetPageCount: TPDF_GetPageCount;
  FPDF_RenderPage: TPDF_RenderPage;
  FPDFBitmap_Create: TPDFBitmap_Create;
  FPDFBitmap_FillRect: TPDFBitmap_FillRect;
  FPDF_RenderPageBitmap: TPDF_RenderPageBitmap;
  FPDFBitmap_GetBuffer: TPDFBitmap_GetBuffer;
  FPDFBitmap_Destroy: TPDFBitmap_Destroy;
  FPDF_GetPageWidthF: TPDF_GetPageWidthF;
  FPDF_GetPageHeightF: TPDF_GetPageHeightF;
  FPDFPage_HasTransparency: TPDFPage_HasTransparency;
  FPDFPage_GetRotation: TPDFPage_GetRotation;
  FPDF_ClosePage: TPDF_ClosePage;
  FPDF_ImportPages: TPDF_ImportPages;
  FPDF_CreateNewDocument: TPDF_CreateNewDocument;
  FPDF_SaveAsCopy: TPDF_SaveAsCopy;
  FPDF_SaveWithVersion: TPDF_SaveWithVersion;
  FPDF_GetLastError: TPDF_GetLastError;
  FPDF_SetPrintTextWithGDI: TPDF_SetPrintTextWithGDI;

  PDFiumDLLPath: String;

  procedure Lock;
  procedure UnLock;

  procedure frxPDF_InitLibrary();
  procedure frxPDF_CloseDocument(document: FPDF_DOCUMENT);
  function frxPDF_LoadMemDocument64(data_buf: Pointer; size: frInteger; password: FPDF_BYTESTRING): FPDF_DOCUMENT;
  function frxPDF_LoadPage(document: FPDF_DOCUMENT; page_index: Integer): FPDF_PAGE;
  function frxPDF_GetPageCount(document: FPDF_DOCUMENT): Integer;

  procedure frxPDF_RenderPage(DC: HDC; page: FPDF_PAGE; start_x, start_y, size_x, size_y: Integer;
    rotate: Integer; flags: Integer);
  function frxPDFBitmap_Create(width, height: Integer; alpha: Integer): FPDF_BITMAP;
  procedure frxPDFBitmap_FillRect(bitmap: FPDF_BITMAP; left, top, width, height: Integer; color: FPDF_DWORD);
  procedure frxPDF_RenderPageBitmap(bitmap: FPDF_BITMAP; page: FPDF_PAGE; start_x, start_y, size_x, size_y: Integer; rotate: Integer; flags: Integer);
  function frxPDFBitmap_GetBuffer(bitmap: FPDF_BITMAP): Pointer;
  procedure frxPDFBitmap_Destroy(bitmap: FPDF_BITMAP);

  function frxPDF_GetPageWidthF(page: FPDF_PAGE): Single;
  function frxPDF_GetPageHeightF(page: FPDF_PAGE): Single;
  function frxPDFPage_HasTransparency(page: FPDF_PAGE): FPDF_BOOL;
  function frxPDFPage_GetRotation(page: FPDF_PAGE): Integer;
  procedure frxPDF_ClosePage(page: FPDF_PAGE);
  function frxPDF_ImportPages(dest_doc, src_doc: FPDF_DOCUMENT; pagerange: FPDF_BYTESTRING; index: Integer): FPDF_BOOL;
  function frxPDF_CreateNewDocument: FPDF_DOCUMENT;
  function frxPDF_SaveAsCopy(document: FPDF_DOCUMENT; pFileWrite: PFPDF_FILEWRITE; flags: FPDF_DWORD): FPDF_BOOL;
  function frxPDF_SaveWithVersion(document: FPDF_DOCUMENT; pFileWrite: PFPDF_FILEWRITE;
    flags: FPDF_DWORD; fileVersion: Integer): FPDF_BOOL;
  function frxPDF_GetLastError(): LongWord;
  procedure frxPDF_SetPrintTextWithGDI(use_gdi: LongBool);

  procedure frxRenderPage(ACanvas: TCanvas; page: FPDF_PAGE; start_x, start_y, size_x, size_y: Integer;
    rotate: Integer; flags: Integer; AFastDraw: Boolean; ADirectDraw: Boolean);

const
  FPDF_ERR_SUCCESS   = 0;
  FPDF_ERR_UNKNOWN   = 1;
  FPDF_ERR_FILE      = 2;
  FPDF_ERR_FORMAT    = 3;
  FPDF_ERR_PASSWORD  = 4;
  FPDF_ERR_SECURITY  = 5;
  FPDF_ERR_PAGE      = 6;
  RsPdfErrorSuccess  = 'No error.';
  RsPdfErrorUnknown  = 'Unknown error.';
  RsPdfErrorFile     = 'File not found or can''t be opened.';
  RsPdfErrorFormat   = 'File is not a PDF document or is corrupted.';
  RsPdfErrorPassword = 'Password required or invalid password.';
  RsPdfErrorSecurity = 'Security schema is not supports.';
  RsPdfErrorPage     = 'Page does not exist or data error.';

  RsPdfInvalidPage   = 'Invalid page';
  RsPdfDLLNotFound   = pdfium_dll + ' was not found.';
  RsPdfCheckActive   = 'attempt to render before loading the document.';
  RsPdfEmptyDocument = 'Empty document.';

implementation
end.
