unit frxASN1;

{$I frx.inc}

interface

uses
{$IFNDEF Linux}
  Windows,
{$ENDIF}
  SysUtils, Classes, Math,
  frCoreClasses, frxHash, frxHelpers;

type
  TOIDName = (
    OID_undef,
    OID_rsaEncryption,                      // 1.2.840.113549.1.1.1
    OID_pkcs7_data,                         // 1.2.840.113549.1.7.1
    OID_pkcs7_signedData,                   // 1.2.840.113549.1.7.2
    OID_pkcs7_envelopedData,                // 1.2.840.113549.1.7.3
    OID_pkcs7_signedAndEnvelopedData,       // 1.2.840.113549.1.7.4
    OID_pkcs7_digestedData,                 // 1.2.840.113549.1.7.5
    OID_pkcs7_encryptedData,                // 1.2.840.113549.1.7.6
    OID_pkcs9_emailAddress,                 // 1.2.840.113549.1.9.1
    OID_pkcs9_contentType,                  // 1.2.840.113549.1.9.3
    OID_pkcs9_messageDigest,                // 1.2.840.113549.1.9.4
    OID_friendlyName,                       // 1.2.840.113549.1.9.20
    OID_localKeyID,                         // 1.2.840.113549.1.9.21
    OID_x509Certificate,                    // 1.2.840.113549.1.9.22.1
    OID_pbeWithSHA1And128BitRC4,            // 1.2.840.113549.1.12.1.1
    OID_pbeWithSHA1And40BitRC4,             // 1.2.840.113549.1.12.1.2
    OID_pbeWithSHA1And3_KeyTripleDES_CBC,   // 1.2.840.113549.1.12.1.3
    OID_pbeWithSHA1And2_KeyTripleDES_CBC,   // 1.2.840.113549.1.12.1.4
    OID_pbeWithSHA1And128BitRC2_CBC,        // 1.2.840.113549.1.12.1.5
    OID_pbeWithSHA1And40BitRC2_CBC,         // 1.2.840.113549.1.12.1.6
    OID_keyBag,                             // 1.2.840.113549.1.12.10.1.1
    OID_pkcs_8ShroudedKeyBag,               // 1.2.840.113549.1.12.10.1.2
    OID_certBag,                            // 1.2.840.113549.1.12.10.1.3
    OID_crlBag,                             // 1.2.840.113549.1.12.10.1.4
    OID_secretBag,                          // 1.2.840.113549.1.12.10.1.5
    OID_safeContentsBag,                    // 1.2.840.113549.1.12.10.1.6
    OID_md5,                                // 1.2.840.113549.2.5
    OID_sha1,                               // 1.3.14.3.2.26 - US Secure Hash Algorithm 1 (SHA1)
    OID_commonName,                         // 2.5.4.3
    OID_surname,                            // 2.5.4.4
    OID_serialNumber,                       // 2.5.4.5
    OID_countryName,                        // 2.5.4.6
    OID_localityName,                       // 2.5.4.7
    OID_stateOrProvinceName,                // 2.5.4.8
    OID_streetAddress,                      // 2.5.4.9
    OID_organizationName,                   // 2.5.4.10
    OID_organizationUnitName,               // 2.5.4.11
    OID_title,                              // 2.5.4.12
    OID_postalCode,                         // 2.5.4.17
    OID_givenName,                          // 2.5.4.42
    OID_initials,                           // 2.5.4.43
    OID_generationQualifier,                // 2.5.4.44
    OID_dnQualifier,                        // 2.5.4.46
    OID_pseudonym,                          // 2.5.4.65
    OID_sha256                              // 2.16.840.1.101.3.4.2.1
  );

  TfrxASN1Base = class
  private
    FTag: Cardinal;
    FClass: Byte;
  protected
    function CalcSize(Value, Threshold, Shift: Cardinal): Cardinal;
    function GetTagSize: Cardinal;
    function GetLenSize: Cardinal; virtual;
    function GetDataSize: Cardinal; virtual;
    function GetTotalSize: Cardinal;
    function Constructed: Byte; virtual;
    function HeaderToAnsi: AnsiString;
    function LengthToAnsi: AnsiString; virtual;
    function DataToAnsi: AnsiString; virtual;
  public
    constructor Create(ATag: Cardinal; AClass: Byte);
    function ToAnsi: AnsiString;
    function CreateCopy: TfrxASN1Base; virtual; abstract;
    function IsEqual(ASN1Object: TfrxASN1Base): Boolean; virtual;
    function IsTag(ATag: Cardinal): Boolean;
    function IsNot(ATag: Cardinal): Boolean;
    function IsClass(AClass: Byte): Boolean;
    function DataToWide: WideString; virtual;

    property Tag: Cardinal read FTag;
    property ASN1Class: Byte read FClass;
  end;

  TfrxASN1Null = class(TfrxASN1Base)
  protected
  public
    constructor Create;
    function CreateCopy: TfrxASN1Base; override;
  end;

  TfrxASN1Data = class(TfrxASN1Base)
  private
    procedure SetData(const Value: AnsiString); virtual;
  protected
    FData: AnsiString;
    function GetDataSize: Cardinal; override;
    function DataToAnsi: AnsiString; override;
  public
    constructor Create(ATag: Cardinal; AClass: Byte; AValue: AnsiString); overload;
    constructor Create(AValue: AnsiString); overload;
    function CreateCopy: TfrxASN1Base; override;
    function IsEqual(ASN1Object: TfrxASN1Base): Boolean; override;
    function DataToWide: WideString; override;

    property Data: AnsiString read FData write SetData;
  end;

  TfrxASN1ImmutableData = class(TfrxASN1Data)
  private
    procedure SetData(const Value: AnsiString); override;
  end;

  TfrxASN1Boolean = class(TfrxASN1ImmutableData)
  private
    FValue: Boolean;
  protected
  public
    constructor Create(AValue: Boolean); overload;
    constructor Create(AData: AnsiString); overload;
    function CreateCopy: TfrxASN1Base; override;
    property Value: Boolean read FValue;
  end;

  TfrxASN1Integer = class(TfrxASN1ImmutableData)
  private
    FValue: Int64;
    FIsOver64Bit: Boolean;
  public
    constructor Create(AValue: Int64); overload;
    constructor Create(AData: AnsiString); overload;
    function CreateCopy: TfrxASN1Base; override;
    property Value: Int64 read FValue;
    property IsOver64Bit: Boolean read FIsOver64Bit;
  end;

  TfrxASN1ObjectID = class(TfrxASN1ImmutableData)
  private
    FOIDName: TOIDName;
  public
    constructor Create(AData: AnsiString); overload;
    constructor Create(AOIDName: TOIDName); overload;
    function CreateCopy: TfrxASN1Base; override;
    property OIDName: TOIDName read FOIDName;
  end;

const
  ASN1_TAG_EOC = 0; // End-of-Content
  ASN1_TAG_BOOLEAN = 1;
  ASN1_TAG_INTEGER = 2;
  ASN1_TAG_OCTET_STRING = 4;
  ASN1_TAG_NULL = 5;
  ASN1_TAG_OBJECT_ID = 6;

  ASN1_TAG_UTF8STRING = 12;
  ASN1_TAG_NUMERICSTRING = 18;
  ASN1_TAG_PRINTABLESTRING = 19;
  ASN1_TAG_TG1STRING = 20;
  ASN1_TAG_VIDEOTEXSTRING = 21;
  ASN1_TAG_IA5STRING = 22;
  ASN1_TAG_GRAPHICSTRING = 25;
  ASN1_TAG_VISIBLESTRING = 26;
  ASN1_TAG_GENERALSTRING = 27;
  ASN1_TAG_UNIVERSALSTRING = 28;
  ASN1_TAG_BMPSTRING = 29;

  ASN1_TAG_SEQUENCE = 16;
  ASN1_TAG_SET = 17;

  ASN1_CLASS_UNIVERSAL = 0;
  ASN1_CLASS_CONTEXT = 2;

type
  TfrxASN1Container = class(TfrxASN1Base)
  private
    FList: TfrOwnObjList;
    FUnknowLength: Boolean;
    function GetItems(Index: Integer): TfrxASN1Base;
    function GetCount: Integer;
  protected
    function Constructed: Byte; override;
    function GetLenSize: Cardinal; override;
    function GetDataSize: Cardinal; override;
    function LengthToAnsi: AnsiString; override;
    function DataToAnsi: AnsiString; override;
  public
    constructor Create(ATag: Cardinal; AClass: Byte);
    constructor CreateSEQUENCE(AClass: Byte = ASN1_CLASS_UNIVERSAL);
    constructor CreateSET;
    destructor Destroy; override;
    function Add(AObject: TfrxASN1Base): Integer;
    function CreateCopy: TfrxASN1Base; override;
    function IsEqual(ASN1Object: TfrxASN1Base): Boolean; override;
    function IsEmpty: Boolean;
    property Items[Index: Integer]: TfrxASN1Base read GetItems; default;
    property Count: Integer read GetCount;
  end;

  TfrxASN1Document = class(TObject)
  private
    FList: TfrOwnObjList;
    procedure LoadFromBuffer(Buffer: Pointer; Size: Cardinal);
    function GetCount: Integer;
    function GetItems(Index: Integer): TfrxASN1Base;
  public
    constructor Create;
    destructor Destroy; override;
    procedure LoadFromStream(AStream: TStream);
    function Add(AObject: TfrxASN1Base): Integer;
    procedure Clear;
    property Count: Integer read GetCount;
    property Items[Index: Integer]: TfrxASN1Base read GetItems; default;
  end;

function ReadASN1Object(Data: AnsiString; Offset: Integer = 0): TfrxASN1Base;

procedure RaiseIf(Flag: Boolean; Text: string; DebugLog: TLogList = nil);
function OIDtoHashClass(OID: TOIDName): TfrxHashClass;

implementation
end.
