
{******************************************}
{                                          }
{             FastReport VCL               }
{            Certificate List              }
{                                          }
{         Copyright (c) 1998-2023          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxCertificateList;

interface

{$I frx.inc}

uses
{$IFNDEF NONWINFPC}
  Windows,
{$ENDIF}
  Contnrs, SysUtils, Classes, Types,
  frxWinCrypt, frxHelpers;

type
  TSignatureSatus = (ssOK,
    ssMessageFileNotSpecified,
    ssMessageFileNotFound,
    ssStoreHandleWasNotGot,
    ssCertificateNotFound,
    ssReadMessageDataFailed,
    ssCryptSignMessageFailed,
    ssSaveSignatureFailed,
    ssWrongLen,
    ssCertGetCertificateChainFailed,
    ssCertificateFileNotSpecified,
    ssCertificateFileNotFound,
    ssCryptoAPICADES_X_LONG_TYPE_1,
    ssWrongTSAURL,
    ssCryptMsgOpenToDecodeFailed,
    ssCryptMsgUpdateFailed,
    ssCryptMsgGetParamFailed,
    ssCryptRetrieveTimeStampFailed,
    ssCryptEncodeObjectFailed,
    ssCryptMsgControlFailed,
    ssCadesSignMessageFailed,
    ssCadesFreeBlobFailed,
    ssCadesEnhanceMessageFailed,
    ssCryptoProSDKNotFound,
    ssCadesSignMessageNotFound,
    ssCadesEnhanceMessageNotFound,
    ssCadesFreeBlobNotFound,
    ssWrongPassword,
    ssInvalidASN1File,
    ssCantParseCertificate,
    ssPrepareSignError,
    ssSignDigestError,
    ssUnknownEncryptionAlgorithm,
    ssUnknownSertificateStructure,
    ssInvalidEncryptedData,
    ssInvalidPrivateKey,
    ssUnsupportedAlgorithm,
    ssInvalidCertificateName,
    ssFastReportEngineRequireCertInPFX,
    ssFastReportEngineRequireCADES_BES,
    ssCantCreateTemporaryStore,
    ssAddCertificateContextToStoreFailed,
    ssPFXExportCertStoreExFailed,
    ssCertificateNotAssigned,
    ssNotUnique,
    ssPFXFileNotSpecified,
    ssPFXFileNotFound
    );

type
  TfrxCertificateStoreKind = (csWindowsMy, csPFXFile); // csToken

type
  TfrxCertificate = class
  private
    FContext: PCCERT_CONTEXT;
    FStoreKind: TfrxCertificateStoreKind;
    FPFXFile: TFileName;
    FPassword: AnsiString;

    function CertName(const NameBlob: TCertNameBlob; StrType: DWORD): string;
    function CertNameField(const NameBlob: TCertNameBlob; Field: string): string;
  public
    constructor Create(AContext: PCCERT_CONTEXT);
    constructor CreatePFX(APFXFile: TFileName; APassword: AnsiString);
    destructor Destroy; override;

    function ToStream(out Stream: TStream; out Password: AnsiString): TSignatureSatus;
    function Copy: TfrxCertificate;
    function This: TfrxCertificate;

    function IsExpired: Boolean;
    function IsNamesSuiable(SubjectContain, IssuerConain: string; IgnoreCase: Boolean): Boolean;
    function Subject(StrType: DWORD = CERT_SIMPLE_NAME_STR): string;
    function Issuer(StrType: DWORD = CERT_SIMPLE_NAME_STR): string;
    function SubjectCommonName: string;
    function IssuerCommonName: string;
    function NotBefore: TDateTime;
    function NotAfter: TDateTime;
    function HashOID: PAnsiChar;
    function IsGOST: Boolean;
    function HashOIDName: string;

    property Context: PCCERT_CONTEXT read FContext;
    property StoreKind: TfrxCertificateStoreKind read FStoreKind write FStoreKind;
    property PFXFile: TFileName read FPFXFile write FPFXFile;
    property Password: AnsiString read FPassword write FPassword;
  end;

  function frxCertificateDescription(Certificate: TfrxCertificate): string;

type
  TFillType = (ftAll, ftFirst, ftUnique);

  TfrxCertificateData = record
    StoreKind: TfrxCertificateStoreKind;
    PFXFile: TFileName;
    Password: AnsiString;
    IgnoreCase: Boolean;
    SubjectContain: string;
    IssuerContain: string;
    AllowExpired: Boolean;
    OnlyGOST: Boolean;
    FillType: TFillType;
  end;

function frxCertificateData(StoreKind: TfrxCertificateStoreKind; PFXFile: TFileName;
  Password: AnsiString; IgnoreCase: Boolean; SubjectContain: string;
  IssuerContain: string; AllowExpired: Boolean; OnlyGOST: Boolean;
  FillType: TFillType): TfrxCertificateData;
function frxCertificateDataStore(SubjectContain: string; IssuerContain: string): TfrxCertificateData;
function frxCertificateDataPFX(PFXFile: TFileName; Password: AnsiString): TfrxCertificateData;

type
  TfrxCertificateList = class(Contnrs.TObjectList)
  private
    FStatus: TSignatureSatus;
    FDebugLog: TLogList;
    FStoreHandle: HCERTSTORE;
    FCertificateData: TfrxCertificateData;

    function IsOpenCertificateStore: Boolean;
    function OpenPFXStore: HCERTSTORE;
    procedure ReadFromStore;
    function GetCertificate(Index: Integer): TfrxCertificate;
  public
    constructor Create(DebugLog: TLogList);

    function IsError(ErrorCondition: Boolean; ErrorStatus: TSignatureSatus;
      IsUseLastError: Boolean = False): Boolean;
    procedure FillFrom(ACertificateData: TfrxCertificateData);
    procedure GetStrings(Strings: TStrings);
    function ExtractByIndex(Index: Integer): TfrxCertificate;

    property Items[Index: Integer]: TfrxCertificate read GetCertificate; default;
    property Status: TSignatureSatus read FStatus;
  end;

function frxSignatureErrorDescription(SignatureSatus: TSignatureSatus): string;
function frxGetCertificate(CertificateData: TfrxCertificateData; var Certificate: TfrxCertificate): TSignatureSatus;

function IsReadFile2BDA(const FileName: TFileName; out BDA: TByteDynArray): Boolean;

const
  SignatureEncoding = X509_ASN_ENCODING or PKCS_7_ASN_ENCODING;

implementation
end.
