
{******************************************}
{                                          }
{             FastReport VCL               }
{         TeeChart Add-In Object           }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxChart;

interface

{$I frx.inc}
{$I frTeeChartVer.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Menus, Controls,
  frCoreClasses, frxClass, frThreading,
{$IFDEF DELPHI16}
  VCLTee.TeeProcs, VCLTee.TeEngine, VCLTee.Chart, VCLTee.Series, VCLTee.TeCanvas
{$ELSE}
  TeeProcs, TeEngine, Chart, Series, TeCanvas
{$ENDIF}
{$IFDEF DELPHI16}
 {$IFDEF TeeChartPro}, VCLTEE.TeeEdit{$IFNDEF TeeChart4}, VCLTEE.TeeEditCha{$ENDIF} {$ENDIF}
{$ELSE}
 {$IFDEF TeeChartPro}, TeeEdit{$IFNDEF TeeChart4}, TeeEditCha{$ENDIF} {$ENDIF}
{$ENDIF}
{$IFDEF DELPHI6}
, Variants
{$ENDIF};


type
{$IFDEF DELPHI16}
/// <summary>
///   The TfrxChartBoxObject lets you use the Chart component in your report.
///   TfrxChartObject is an empty component. It is used to add the frxChart.pas
///   file to the "uses" list. The main component is TfrxChartView.
/// </summary>
[ComponentPlatformsAttribute(pidWin32 or pidWin64)]
{$ENDIF}
  TfrxChartObject = class(TComponent);  // fake component
  TChartClass = class of TCustomChart;

  /// <summary>
  ///   Data type of the series: DB data, band data or fixed data.
  /// </summary>
  TfrxSeriesDataType = (dtDBData, dtBandData, dtFixedData);
  /// <summary>
  ///   Sort type of the series data.
  /// </summary>
  TfrxSeriesSortOrder = (soNone, soAscending, soDescending);
  /// <summary>
  ///   The X-axis data type: text, numeric, date.
  /// </summary>
  TfrxSeriesXType = (xtText, xtNumber, xtDate);
  TSeriesClass = class of TChartSeries;
  /// <summary>
  ///   Series types.
  /// </summary>
  TfrxChartSeries = (csLine, csArea, csPoint, csBar, csHorizBar,
    csPie, csGantt, csFastLine, csArrow, csBubble, csChartShape, csHorizArea,
    csHorizLine, csPolar, csRadar, csPolarBar, csGauge, csSmith, csPyramid,
    csDonut, csBezier, csCandle, csVolume, csPointFigure, csHistogram,
    csHorizHistogram, csErrorBar, csError, csHighLow, csFunnel, csBox,
    csHorizBox, csSurface, csContour, csWaterFall, csColorGrid, csVector3D,
    csTower, csTriSurface, csPoint3D, csBubble3D, csMyPoint, csBarJoin, csBar3D,
    //new
    csPolarContourSeries, csEquiVolumeSeries, csCircularGauge, csClockGauge,
    csNumericGauge, csBigCandle, csDeltaPoint, TImageBar, TImagePoint,
    csWindRose, csErrorPoint3D, csErrorPoint, csEqualizer, csPolarGrid,
    csOrg, csKagi, csRenko, csTagCloud, csHighLowLine, csVolumePipe, csTreeMap,
    csKnobGauge, csRose, csTernary
    );


  /// <summary>
  ///   The TfrxSeriesItem class represents one series.
  /// </summary>
  TfrxSeriesItem = class(TfrCollectionItem)
  private
    FDataBand: TfrxDataBand;
    FDataSet: TfrxDataSet;
    FDataSetName: String;
    FDataType: TfrxSeriesDataType;
    FSortOrder: TfrxSeriesSortOrder;
    FTopN: Integer;
    FTopNCaption: String;
    FSource1: String;
    FSource2: String;
    FSource3: String;
    FSource4: String;
    FSource5: String;
    FSource6: String;
    FSource7: String;
    FSource8: String;
    FSource9: String;
    FSource10: String;
    FXType: TfrxSeriesXType;
    FValues1: String;
    FValues2: String;
    FValues3: String;
    FValues4: String;
    FValues5: String;
    FValues6: String;
    FValues7: String;
    FValues8: String;
    FValues9: String;
    FValues10: String;
    Fsl1: TStringList;
    Fsl2: TStringList;
    Fsl3: TStringList;
    Fsl4: TStringList;
    Fsl5: TStringList;
    Fsl6: TStringList;
    Fsl7: TStringList;
    Fsl8: TStringList;
    Fsl9: TStringList;
    Fsl10: TStringList;
    FValueIndex1: Integer;
    FValueIndex2: Integer;
    FValueIndex3: Integer;
    FValueIndex4: Integer;
    FValueIndex5: Integer;
    FValueIndex6: Integer;
    FValueIndex7: Integer;
    FValueIndex8: Integer;
    FValueIndex9: Integer;
    FValueIndex10: Integer;
    procedure FillSeries(Series: TChartSeries);
    procedure SetDataSet(const Value: TfrxDataSet);
    procedure SetDataSetName(const Value: String);
    function GetDataSetName: String;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    function IsUniqueNameStored: Boolean; override;
  published
    /// <summary>
    ///   Type of series data.
    /// </summary>
    property DataType: TfrxSeriesDataType read FDataType write FDataType;
    /// <summary>
    ///   Band which the series is connected to. Used if DataType = dtBandData.
    /// </summary>
    property DataBand: TfrxDataBand read FDataBand write FDataBand;
    /// <summary>
    ///   Dataset which is series connected to. This property is used if
    ///   property DataType = dtDBData.
    /// </summary>
    property DataSet: TfrxDataSet read FDataSet write SetDataSet;
    property DataSetName: String read GetDataSetName write SetDataSetName;
    /// <summary>
    ///   Sort order of series data.
    /// </summary>
    property SortOrder: TfrxSeriesSortOrder read FSortOrder write FSortOrder;
    /// <summary>
    ///   Show TopN values. Sum of other values shows as a separate value with
    ///   caption set in the TopNCaption property.
    /// </summary>
    property TopN: Integer read FTopN write FTopN;
    /// <summary>
    ///   Show TopN values. Sum of other values shows as a separate value with
    ///   caption set in the TopNCaption property.
    /// </summary>
    property TopNCaption: String read FTopNCaption write FTopNCaption;
    /// <summary>
    ///   The X-axis data type: text, numeric, date.
    /// </summary>
    property XType: TfrxSeriesXType read FXType write FXType;

    { source expressions }
    /// <summary>
    ///   Datasource1 for axis values. This property may contain an expression
    ///   or list of fixed data (in case DataType property is dtFixedData).
    /// </summary>
    property Source1: String read FSource1 write FSource1;
    /// <summary>
    ///   Datasource2 for axis values. This property may contain an expression
    ///   or list of fixed data (in case DataType property is dtFixedData).
    /// </summary>
    property Source2: String read FSource2 write FSource2;
    /// <summary>
    ///   Datasource3 for axis values. This property may contain an expression
    ///   or list of fixed data (in case DataType property is dtFixedData).
    /// </summary>
    property Source3: String read FSource3 write FSource3;
    /// <summary>
    ///   Datasource4 for axis values. This property may contain an expression
    ///   or list of fixed data (in case DataType property is dtFixedData).
    /// </summary>
    property Source4: String read FSource4 write FSource4;
    /// <summary>
    ///   Datasource5 for axis values. This property may contain an expression
    ///   or list of fixed data (in case DataType property is dtFixedData).
    /// </summary>
    property Source5: String read FSource5 write FSource5;
    /// <summary>
    ///   Datasource6 for axis values. This property may contain an expression
    ///   or list of fixed data (in case DataType property is dtFixedData).
    /// </summary>
    property Source6: String read FSource6 write FSource6;
    /// <summary>
    ///   Datasource7 for axis values. This property may contain an expression
    ///   or list of fixed data (in case DataType property is dtFixedData).
    /// </summary>
    property Source7: String read FSource7 write FSource7;
    /// <summary>
    ///   Datasource8 for axis values. This property may contain an expression
    ///   or list of fixed data (in case DataType property is dtFixedData).
    /// </summary>
    property Source8: String read FSource8 write FSource8;
    /// <summary>
    ///   Datasource9 for axis values. This property may contain an expression
    ///   or list of fixed data (in case DataType property is dtFixedData).
    /// </summary>
    property Source9: String read FSource9 write FSource9;
    /// <summary>
    ///   Datasource10 for axis values. This property may contain an expression
    ///   or list of fixed data (in case DataType property is dtFixedData).
    /// </summary>
    property Source10: String read FSource10 write FSource10;

    { ready values. For internal use only. }
    property Values1: String read FValues1 write FValues1;
    property Values2: String read FValues2 write FValues2;
    property Values3: String read FValues3 write FValues3;
    property Values4: String read FValues4 write FValues4;
    property Values5: String read FValues5 write FValues5;
    property Values6: String read FValues6 write FValues6;
    property Values7: String read FValues7 write FValues7;
    property Values8: String read FValues8 write FValues8;
    property Values9: String read FValues9 write FValues9;
    property Values10: String read FValues10 write FValues10;

    { backward compatibility }
    property XSource: String read FSource1 write FSource1;
    property YSource: String read FSource2 write FSource2;
    property XValues: String read FValues1 write FValues1;
    property YValues: String read FValues2 write FValues2;

    { deprecated V5 }
    property InheritedName: String read GetInheritedName write SetInheritedName stored False;
  end;

  /// <summary>
  ///   The TfrxSeriesData class represents collection of series.
  /// </summary>
  TfrxSeriesData = class(TfrCollection)
  private
    FReport: TfrxReport;
    function GetSeries(Index: Integer): TfrxSeriesItem;
  public
    constructor Create(Report: TfrxReport);
    /// <summary>
    ///   Adds a new item to the collection.
    /// </summary>
    function Add: TfrxSeriesItem;
    /// <summary>
    ///   List of series.
    /// </summary>
    property Items[Index: Integer]: TfrxSeriesItem read GetSeries; default;
  end;


  /// <summary>
  ///   The TfrxChartView component represents a chart. The component uses
  ///   TeeChart library to render charts. It may show several series on one
  ///   chart. The series settings are stored in the SeriesData collection.
  /// </summary>
  TfrxChartView = class(TfrxView)
  private
    FChart: TCustomChart;
    FSeriesData: TfrxSeriesData;
    FIgnoreNulls: Boolean;
    FLockFillSeries: Boolean;
    FSeriesFilled: Boolean;
    FSavedColor: TColor;
    FHlIndex: Integer;
    FClickedVal1: String;
    FClickedVal2: String;
    FClickedVal3: String;
    FClickedVal4: String;
    FClickedVal5: String;
    FClickedVal6: String;
    FClickedVal7: String;
    FClickedVal8: String;
    FClickedVal9: String;
    FClickedVal10: String;
    FHighlightColor: TColor;
    FPenWidth: Integer;
{$IFDEF DELPHI12}
    FPenStyle: TPenStyle;
{$ENDIF}
    FSeriesIndex: Integer;
    FMouseDown: Boolean;
    FMouseOffsetX: Extended;
    FMouseOffsetY: Extended;
    FMacroLoaded: Boolean;
    FEMF: TMetafile;
    //FChart: TCustomChart;
    procedure ReadData(Stream: TStream);
    procedure ReadData1(Reader: TReader);
    procedure ReadData2(Reader: TReader);
    procedure WriteData(Stream: TStream);
    procedure WriteData1(Writer: TWriter);
    procedure WriteData2(Writer: TWriter);
    function CreateMetafile: TMetafile;
    function ResetHighlightSelection: Boolean;
    procedure DrawOnMetaFile;
    procedure ResetClickedValues;
  protected
    procedure DefineProperties(Filer: TFiler); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure CreateChart; virtual;
    class function GetChartClass: TChartClass; virtual;
    function GetNeedSelectSeries: Boolean; virtual;
    function CheckMoveSelector(X, Y: Extended): Boolean;
    procedure ReaderError(Reader: TReader; const Message: string;
      var Handled: Boolean);
    procedure FillChart; virtual;

    property NeedSelectSeries: Boolean read GetNeedSelectSeries;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DoMouseMove(X, Y: Integer; Shift: TShiftState;
      var EventParams: TfrxInteractiveEventsParams); override;
    procedure DoMouseEnter(aPreviousObject: TfrxComponent;
      var EventParams: TfrxInteractiveEventsParams); override;
    procedure DoMouseLeave(aNextObject: TfrxComponent;
      var EventParams: TfrxInteractiveEventsParams); override;
    procedure DoMouseUp(X, Y: Integer; Button: TMouseButton;
      Shift: TShiftState;
      var EventParams: TfrxInteractiveEventsParams); override;
    function DoMouseDown(X, Y: Integer; Button: TMouseButton;
      Shift: TShiftState; var EventParams: TfrxInteractiveEventsParams)
      : Boolean; override;
    function DoMouseWheel(Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint;
      var EventParams: TfrxInteractiveEventsParams): Boolean; override;
    class function GetDescription: String; override;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;
    procedure AfterPrint; override;
    procedure GetData; override;
    procedure BeforeStartReport; override;
    procedure OnNotify(Sender: TObject); override;
    procedure ClearSeries;
    procedure AddSeries(Series: TfrxChartSeries);
    procedure UpdateSeriesData;

    procedure SaveContentToDictionary(aReport: TfrxReport; PostProcessor: TfrxPostProcessor); override;
    procedure ProcessDictionary(aItem: TfrxMacrosItem; aReport: TfrxReport; PostProcessor: TfrxPostProcessor); override;
    function LoadContentFromDictionary(aReport: TfrxReport; aItem: TfrxMacrosItem): Boolean; override;

    /// <summary>
    ///   Reference to internal TChart object.
    /// </summary>
    property Chart: TCustomChart read FChart;
    /// <summary>
    ///   Settings of the series.
    /// </summary>
    property SeriesData: TfrxSeriesData read FSeriesData;
    property HighlightIndex: Integer read FHLIndex;
    property ClickedVal1: String read FClickedVal1;
    property ClickedVal2: String read FClickedVal2;
    property ClickedVal3: String read FClickedVal3;
    property ClickedVal4: String read FClickedVal4;
    property ClickedVal5: String read FClickedVal5;
    property ClickedVal6: String read FClickedVal6;
    property ClickedVal7: String read FClickedVal7;
    property ClickedVal8: String read FClickedVal8;
    property ClickedVal9: String read FClickedVal9;
    property ClickedVal10: String read FClickedVal10;
  published
    property IgnoreNulls: Boolean read FIgnoreNulls write FIgnoreNulls default False;
    property HighlightColor: TColor read FHighlightColor write FHighlightColor default clHotLight;
    property FillType;
    property Fill;
    property Cursor;
    property Frame;
    property Editable default [ferAllowInPreview];
    property TagStr;
    property URL;
    property Processing;
  end;


implementation
end.
