
{******************************************}
{                                          }
{             FastReport VCL               }
{      Custom TDataSet-based classes       }
{        for enduser DB components         }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxCustomDB;

interface

{$I frx.inc}

uses
{$IFNDEF FPC}
  Windows,
{$ENDIF}
  Types, Classes, SysUtils, DB, frxClass, frxDBSet, DBCtrls, Variants,
  frCoreClasses, frXML;

type
  TfrxCustomQuery = class;

  { TfrQueryAssistant }

  TfrQueryAssistant = class
  strict private
    FOwner: TfrxCustomQuery;
  public
    constructor Create(AOwner: TfrxCustomQuery); virtual;

    function IsActive: Boolean; virtual;
    function ShowDialog(var ASQLText, ASQLSchema: string): Boolean; virtual;

    property Owner: TfrxCustomQuery read FOwner;
  end;
  TfrQueryAssistantClass = class of TfrQueryAssistant;

  /// <summary>
  ///   The TfrxCustomDataset class is the base class for DB engine components
  ///   such as TfrxBDETable, based on TDataSet.
  /// </summary>
  TfrxCustomDataset = class(TfrxDBDataSet)
  private
    FDBConnected: Boolean;
    FDataSource: TDataSource;
    FMaster: TfrxDBDataSet;
    FMasterFields: String;
    procedure SetActive(Value: Boolean);
    procedure SetFilter(const Value: String);
    procedure SetFiltered(Value: Boolean);
    function GetActive: Boolean;
    function GetFields: TFields;
    function GetFilter: String;
    function GetFiltered: Boolean;
    procedure InternalSetMaster(const Value: TfrxDBDataSet);
    procedure InternalSetMasterFields(const Value: String);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetParent(AParent: TfrxComponent); override;
    procedure SetUserName(const Value: String); override;
    procedure SetMaster(const Value: TDataSource); virtual;
    procedure SetMasterFields(const Value: String); virtual;
    property DataSet;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure OnPaste; override;
    /// <summary>
    ///   Determines whether the dataset is connected to a Database component.
    /// </summary>
    property DBConnected: Boolean read FDBConnected write FDBConnected;
    /// <summary>
    ///   Reference to internal TDataSet.Fields.
    /// </summary>
    property Fields: TFields read GetFields;
    /// <summary>
    ///   List of linked fields (used for master-detail relationship). List has
    ///   a form: DetailField=MasterField;...
    /// </summary>
    property MasterFields: String read FMasterFields write InternalSetMasterFields;
    /// <summary>
    ///   Determines whether the dataset is active.
    /// </summary>
    property Active: Boolean read GetActive write SetActive default False;
  published
    /// <summary>
    ///   The expression used for filtering records in a dataset. It is
    ///   equivalent to TDataSet.Filter.
    /// </summary>
    property Filter: String read GetFilter write SetFilter;
    /// <summary>
    ///   Determines whether to filter dataset records. It is equivalent to
    ///   TDataSet.Filtered.
    /// </summary>
    property Filtered: Boolean read GetFiltered write SetFiltered default False;
    /// <summary>
    ///   Reference to master dataset (used for master-detail relationships).
    ///   You also should set up the MasterFields property.
    /// </summary>
    property Master: TfrxDBDataset read FMaster write InternalSetMaster;
    property Left;
    property Top;
  end;

  /// <summary>
  ///   The TfrxCustomTable is a base class for table components of a DB
  ///   engine.
  /// </summary>
  TfrxCustomTable = class(TfrxCustomDataset)
  protected
    function GetIndexFieldNames: String; virtual;
    function GetIndexName: String; virtual;
    function GetTableName: String; virtual;
    procedure SetIndexFieldNames(const Value: String); virtual;
    procedure SetIndexName(const Value: String); virtual;
    procedure SetTableName(const Value: String); virtual;
    property DataSet;
  published
    property MasterFields;
    /// <summary>
    ///   The table name.
    /// </summary>
    property TableName: String read GetTableName write SetTableName;
    /// <summary>
    ///   The index name, used for sorting data records.
    /// </summary>
    property IndexName: String read GetIndexName write SetIndexName;
    /// <summary>
    ///   Use IndexFieldNames as an alternative method of specifying the index
    ///   to use for a table. In IndexFieldNames, specify the name of each
    ///   column to use as an index for a table. Ordering of column names is
    ///   significant. Separate names with semicolon.
    /// </summary>
    property IndexFieldNames: String read GetIndexFieldNames write SetIndexFieldNames;
  end;

  /// <summary>
  ///   The TfrxParamItem represents one query parameter.
  /// </summary>
  TfrxParamItem = class(TfrCollectionItem)
  private
    FDataType: TFieldType;
    FExpression: String;
    FName: String;
    FValue: Variant;
  protected
    function GetInheritedName: String; override;
  public
    procedure Assign(Source: TPersistent); override;
    /// <summary>
    ///   Value of the parameter. This property is used if Expression property
    ///   is empty.
    /// </summary>
    property Value: Variant read FValue write FValue;
  published
    /// <summary>
    ///   Name of parameter. Don't change this property directly, because the
    ///   list of parameters rebuilds automatically when changing SQL query.
    /// </summary>
    property Name: String read FName write FName;
    /// <summary>
    ///   Type of the parameter.
    /// </summary>
    property DataType: TFieldType read FDataType write FDataType;
    /// <summary>
    ///   Value of the parameter - expression that will be calculated when
    ///   opening a query.
    /// </summary>
    property Expression: String read FExpression write FExpression;
  end;

  /// <summary>
  ///   The TfrxParams represents a list of query parameters.
  /// </summary>
  TfrxParams = class(TfrCollection)
  private
    FIgnoreDuplicates: Boolean;
    function GetParam(Index: Integer): TfrxParamItem;
  public
    constructor Create;
    /// <summary>
    ///   Adds a parameter. Don't use this method directly, because the list of
    ///   parameters rebuilds automatically when changing SQL query.
    /// </summary>
    function Add: TfrxParamItem;
    /// <summary>
    ///   Finds a parameter with specified name.
    /// </summary>
    function Find(const Name: String): TfrxParamItem;
    /// <summary>
    ///   Returns index of a parameter with specified name, or -1 if such
    ///   parameter not exists.
    /// </summary>
    function IndexOf(const Name: String): Integer;
    procedure UpdateParams(const SQL: String);
    /// <summary>
    ///   List of parameters.
    /// </summary>
    property Items[Index: Integer]: TfrxParamItem read GetParam; default;
    /// <summary>
    ///   If a property set to True, then parameters with similar names are
    ///   ignored.
    /// </summary>
    property IgnoreDuplicates: Boolean read FIgnoreDuplicates write FIgnoreDuplicates;
  end;

  /// <summary>
  ///   The TfrxCustomQuery class is the base class for query components of the
  ///   DB engine.
  /// </summary>
  TfrxCustomQuery = class(TfrxCustomDataset)
  private
    FAssistant: TfrQueryAssistant;
    FParams: TfrxParams;
    FSaveOnBeforeOpen: TDataSetNotifyEvent;
    FSaveOnBeforeRefresh: TDataSetNotifyEvent;
    FSaveOnChange: TNotifyEvent;
    FSQLSchema: String;

    procedure DataSetBeforeOpenHandler(DataSet: TDataSet);
    procedure DataSetBeforeRefreshHandler(DataSet: TDataSet);
    procedure SQLChangeHandler(Sender: TObject);

    procedure ReadData(Reader: TReader);
    procedure SetParams(Value: TfrxParams);
    procedure WriteData(Writer: TWriter);
    function GetIgnoreDupParams: Boolean;
    procedure SetIgnoreDupParams(const Value: Boolean);
  protected
    function CreateQueryAssistant: TfrQueryAssistant; virtual;
    procedure DefineProperties(Filer: TFiler); override;
    procedure DoDataSetBeforeRefreshHandler(DataSet: TDataSet); virtual;
    procedure DoSQLChange; virtual;
    procedure SetSQL(Value: TStrings); virtual;
    function GetSQL: TStrings; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure UpdateParams; virtual;
    procedure WriteNestedProperties(Item: TfrXmlItem; AAncestor: TPersistent = nil); override;
    function ReadNestedProperty(Item: TfrXmlItem): Boolean; override;
    /// <summary>
    ///   Finds a parameter with specified name.
    /// </summary>
    function ParamByName(const Value: String): TfrxParamItem;
    property Assistant: TfrQueryAssistant read FAssistant;
  published
    /// <summary>
    ///   If a property set to True, then parameters with similar names are
    ///   ignored.
    /// </summary>
    property IgnoreDupParams: Boolean read GetIgnoreDupParams write SetIgnoreDupParams;
    /// <summary>
    ///   List of query parameters. List of parameters rebuilds automatically
    ///   when changing SQL query.
    /// </summary>
    property Params: TfrxParams read FParams write SetParams;
    /// <summary>
    ///   Text of SQL query.
    /// </summary>
    property SQL: TStrings read GetSQL write SetSQL;
    /// <summary>
    ///   The SQL schema used in the FastQueryBuilder.
    /// </summary>
    property SQLSchema: String read FSQLSchema write FSQLSchema;
  end;
  TfrxCustomQueryClass = class of TfrxCustomQuery;

  /// <summary>
  ///   The TfrxDBLookupComboBox control represents TDBLookupComboBox control
  ///   adapter.
  /// </summary>
  TfrxDBLookupComboBox = class(TfrxDialogControl)
  private
    FDataSet: TfrxDBDataSet;
    FDataSetName: String;
    FDataSource: TDataSource;
    FDBLookupComboBox: TDBLookupComboBox;
    FAutoOpenDataSet: Boolean;

    function GetDataSetName: String;
    function GetKeyField: String;
    function GetKeyValue: Variant;
    function GetListField: String;
    function GetText: String;
    procedure SetDataSet(const Value: TfrxDBDataSet);
    procedure SetDataSetName(const Value: String);
    procedure SetKeyField(Value: String);
    procedure SetKeyValue(const Value: Variant);
    procedure SetListField(Value: String);
    procedure UpdateDataSet;
    procedure OnOpenDS(Sender: TObject);
    function GetDropDownWidth: Integer;
    procedure SetDropDownWidth(const Value: Integer);
    function GetDropDownRows: Integer;
    procedure SetDropDownRows(const Value: Integer);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    class function GetDescription: String; override;
    procedure BeforeStartReport; override;
    /// <summary>
    ///   Reference to internal TDBLookupComboBox object.
    /// </summary>
    property DBLookupComboBox: TDBLookupComboBox read FDBLookupComboBox;
    /// <summary>
    ///   Value of key field that was selected in the control.
    /// </summary>
    property KeyValue: Variant read GetKeyValue write SetKeyValue;
    /// <summary>
    ///   Value of list field that was selected in the control.
    /// </summary>
    property Text: String read GetText;
  published
    /// <summary>
    ///   If a property is set to True , then during execution FastReport will
    ///   try to open the attached data source.
    /// </summary>
    property AutoOpenDataSet: Boolean read FAutoOpenDataSet write FAutoOpenDataSet default False;
    /// <summary>
    ///   Dataset which the control is connected to.
    /// </summary>
    property DataSet: TfrxDBDataset read FDataSet write SetDataSet;
    /// <summary>
    ///   Name of dataset, which the control is connected to. This property
    ///   duplicates the DataSet property.
    /// </summary>
    property DataSetName: String read GetDataSetName write SetDataSetName;
    /// <summary>
    ///   Name of DB field that will be shown in the control.
    /// </summary>
    property ListField: String read GetListField write SetListField;
    /// <summary>
    ///   Name of DB field that will represent the key field.
    /// </summary>
    property KeyField: String read GetKeyField write SetKeyField;
    property DropDownWidth: Integer read GetDropDownWidth write SetDropDownWidth;
    property DropDownRows: Integer read GetDropDownRows write SetDropDownRows;
    property OnClick;
    property OnDblClick;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

  { TfrQueryAssistantClassRepository }

  TfrQueryAssistantClassRepository = class
  strict private
    FItems: TStringList;
    procedure Clear;
  public
    constructor Create;
    destructor Destroy; override;

    procedure RegisterClass(AQueryClass: TfrxCustomQueryClass; AAssistantClass: TfrQueryAssistantClass);
    procedure UnregisterClass(AQueryClass: TfrxCustomQueryClass; AAssistantClass: TfrQueryAssistantClass);

    function GetAssistantClass(AQueryClass: TfrxCustomQueryClass): TfrQueryAssistantClass;
  end;

procedure frxParamsToTParams(Query: TfrxCustomQuery; Params: TParams);
function frQueryAssistantClassRepository: TfrQueryAssistantClassRepository;
implementation
end.
