
{******************************************}
{                                          }
{             FastReport VCL               }
{                Designer                  }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxDesgn;

interface

{$I frx.inc}

uses
{$IFNDEF FPC}
  Windows, Messages,
{$ENDIF}
  SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Types, ComCtrls, ImgList, Menus, Buttons, StdCtrls, ToolWin, ExtCtrls, ActnList, frxBaseForm,
{$IFDEF FPC}
  LResources, LCLType, LMessages, LCLIntf, LCLProc, LazHelper, LazarusPackageIntf,
{$ELSE}
  CommCtrl,
{$ENDIF}
  frCore,
  frxClass, frxDock, frxCtrls, frxDesgnCtrls, frxDesgnWorkspace,
  frxInsp, frxDialogForm, frxDataTree, frxReportTree, frxSimpleMemo,
  fs_iinterpreter, Printers,
{$IFNDEF FPC}
  frxWatchForm, frxBreakPointsForm,
{$ELSE}
  frxLazBWForm,
{$ENDIF}
  frPictureCache, frxSearchDForm, Variants, frxDPIAwareBaseControls,
{$IFDEF DELPHI17}
  System.Actions,
{$ENDIF}
{$IFDEF DELPHI9}
  GraphUtil, Tabs,
{$ENDIF}
{$IFDEF DELPHI16}
  System.UITypes,
{$ENDIF}
  frBaseForm, frTypes, frTreeView;

type
  /// <summary>
  ///   Units used in the designer.
  /// </summary>
  TfrxDesignerUnits = (
    /// <summary>
    ///   entimeters.
    /// </summary>
    duCM,
    /// <summary>
    ///   Inches.
    /// </summary>
    duInches,
    /// <summary>
    ///   Pixels.
    /// </summary>
    duPixels,
    /// <summary>
    ///   Chars(dot-matrix report).
    /// </summary>
    duChars);
  TfrxDesignerUpdateEvent = (fduLoadReport, fduSaveReport);
  TfrxAfterUpdateReportEvent = procedure(Sender: TObject; EventType: TfrxDesignerUpdateEvent; Report: TfrxReport) of object;
  /// <summary>
  ///   The event occurs when loading a report. With the help of this event,
  ///   you can manage report's loading from DB. Event handler must load a
  ///   report and return True if operation was successful.
  /// </summary>
  TfrxLoadReportEvent = function(Report: TfrxReport): Boolean of object;
  TfrxLoadRecentReportEvent = function(Report: TfrxReport; FileName: String): Boolean of object;
  /// <summary>
  ///   The event occurs when saving a report. With the help of this event, you
  ///   can manage process of report's saving in DB. Event handler must save a
  ///   report and return True if operation was successful. SaveAs parameter is
  ///   True in case the user selects "Save As..." menu item.
  /// </summary>
  TfrxSaveReportEvent = function(Report: TfrxReport; SaveAs: Boolean): Boolean of object;
  /// <summary>
  ///   The event occurs when user goes File|New... or Report|Options... menu
  ///   items to choose the report template. Event handler must return a list
  ///   of available templates in the List parameter.
  /// </summary>
  TfrxGetTemplateListEvent = procedure(List: TStrings) of object;
  /// <summary>
  ///   A set of flags, which inhibit specific operations in the designer.
  /// </summary>
  TfrxDesignerRestrictions = set of TfrxDesignerRestriction;
  TSampleFormat = class;

  TScriptStepMode = (ssmNone, ssmStepTraceInto, ssmStepOver, ssmStepReturn);

{$IFDEF DELPHI16}
/// <summary>
///   The TfrxDesigner component is a report designer. When using this
///   component, you have the ability to design reports in your project at
///   runtime. The component itself contains only a few designer settings; the
///   designer itself is connected by adding the "frxDesgn" unit into the
///   "Uses" list (this occurs automatically if you place the component on the
///   form).
/// </summary>
[ComponentPlatformsAttribute(pidWin32 or pidWin64)]
{$ENDIF}
  TfrxDesigner = class(TComponent)
  private
    FCloseQuery: Boolean;
    FDefaultScriptLanguage: String;
    FDefaultFont: TFont;
    FDefaultLeftMargin: Extended;
    FDefaultBottomMargin: Extended;
    FDefaultRightMargin: Extended;
    FDefaultTopMargin: Extended;
    FDefaultPaperSize: Integer;
    FDefaultOrientation: TPrinterOrientation;
{$IFDEF DELPHI10}
    FGradient: Boolean;
    FGradientEnd: TColor;
    FGradientStart: TColor;
{$ENDIF}
    FOpenDir: String;
    FSaveDir: String;
    FTemplatesExt: String;
    FTemplateDir: String;
    FStandalone: Boolean;
    FRestrictions: TfrxDesignerRestrictions;
    FRTLLanguage: Boolean;
    FMemoParentFont: Boolean;
    FOnAfterUpdateReport: TfrxAfterUpdateReportEvent;
    FOnLoadReport: TfrxLoadReportEvent;
    FOnLoadRecentReport: TfrxLoadRecentReportEvent;
    FOnSaveReport: TfrxSaveReportEvent;
    FOnShow: TNotifyEvent;
    FOnInsertObject: TNotifyEvent;
    FOnGetTemplateList: TfrxGetTemplateListEvent;
    FOnShowStartupScreen: TNotifyEvent;

    procedure SetDefaultFont(const Value: TFont);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    /// <summary>
    ///   Defines, if it is necessary to ask about report saving on designer
    ///   closing.
    /// </summary>
    property CloseQuery: Boolean read FCloseQuery write FCloseQuery default True;
    /// <summary>
    ///   The default script language.
    /// </summary>
    property DefaultScriptLanguage: String read FDefaultScriptLanguage write FDefaultScriptLanguage;
    /// <summary>
    ///   The default font for a new "Text" object.
    /// </summary>
    property DefaultFont: TFont read FDefaultFont write SetDefaultFont;
    /// <summary>
    ///   The default left margin for a new report page.
    /// </summary>
    property DefaultLeftMargin: Extended read FDefaultLeftMargin write FDefaultLeftMargin;
    /// <summary>
    ///   The default right margin for a new report page.
    /// </summary>
    property DefaultRightMargin: Extended read FDefaultRightMargin write FDefaultRightMargin;
    /// <summary>
    ///   The default top margin for a new report page.
    /// </summary>
    property DefaultTopMargin: Extended read FDefaultTopMargin write FDefaultTopMargin;
    /// <summary>
    ///   The default bottom margin for a new report page.
    /// </summary>
    property DefaultBottomMargin: Extended read FDefaultBottomMargin write FDefaultBottomMargin;
    /// <summary>
    ///   The default paper size for a new report page.
    /// </summary>
    property DefaultPaperSize: Integer read FDefaultPaperSize write FDefaultPaperSize;
    /// <summary>
    ///   The default orientation for a new report page.
    /// </summary>
    property DefaultOrientation: TPrinterOrientation read FDefaultOrientation write FDefaultOrientation;
{$IFDEF DELPHI10}
    /// <summary>
    ///   The gradient fill for toolbars.
    /// </summary>
    property Gradient: Boolean read FGradient write FGradient default False;
    /// <summary>
    ///   The gradient fill for toolbars, end color.
    /// </summary>
    property GradientEnd: TColor read FGradientEnd write FGradientEnd;
    /// <summary>
    ///   The gradient fill for toolbars, start color.
    /// </summary>
    property GradientStart: TColor read FGradientStart write FGradientStart;
{$ENDIF}
    /// <summary>
    ///   The name of a folder, from which the file is opened by default.
    /// </summary>
    property OpenDir: String read FOpenDir write FOpenDir;
    /// <summary>
    ///   The name of a folder, where the file is saved to by default.
    /// </summary>
    property SaveDir: String read FSaveDir write FSaveDir;
    /// <summary>
    ///   Extension for report files (during changes you need to change a line
    ///   with dialogue filter in resources as well).
    /// </summary>
    property TemplatesExt: String read FTemplatesExt write FTemplatesExt;
    /// <summary>
    ///   The directory name that contains report templates.
    /// </summary>
    property TemplateDir: String read FTemplateDir write FTemplateDir;
    /// <summary>
    ///   Deprecated. Not used.
    /// </summary>
    property Standalone: Boolean read FStandalone write FStandalone default False;
    /// <summary>
    ///   A set of flags, which forbids certain operations in the designer.
    /// </summary>
    property Restrictions: TfrxDesignerRestrictions read FRestrictions write FRestrictions;
    /// <summary>
    ///   Determines whether to set the RTLReading property for a new "Text"
    ///   object.
    /// </summary>
    property RTLLanguage: Boolean read FRTLLanguage write FRTLLanguage;
    property MemoParentFont: Boolean read FMemoParentFont write FMemoParentFont;
    /// <summary>
    ///   The event occurs when loading a report. With the help of this event,
    ///   you can manage report's loading from DB.
    /// </summary>
    property OnLoadReport: TfrxLoadReportEvent read FOnLoadReport write FOnLoadReport;
    property OnAfterUpdateReport: TfrxAfterUpdateReportEvent read FOnAfterUpdateReport write FOnAfterUpdateReport;
    property OnLoadRecentReport: TfrxLoadRecentReportEvent read FOnLoadRecentReport write FOnLoadRecentReport;
    /// <summary>
    ///   The event occurs when saving a report. With the help of this event,
    ///   you can manage process of report's saving in DB.
    /// </summary>
    property OnSaveReport: TfrxSaveReportEvent read FOnSaveReport write FOnSaveReport;
    /// <summary>
    ///   The event occurs when the designer starts.
    /// </summary>
    property OnShow: TNotifyEvent read FOnShow write FOnShow;
    /// <summary>
    ///   The event occurs when inserting a new object.
    /// </summary>
    property OnInsertObject: TNotifyEvent read FOnInsertObject write FOnInsertObject;
    /// <summary>
    ///   The event occurs before displaying a designer window.
    /// </summary>
    property OnShowStartupScreen: TNotifyEvent read FOnShowStartupScreen write FOnShowStartupScreen;
    /// <summary>
    ///   The event occurs when user selects File|New... or Report|Options...
    ///   menu items to choose the report template. Event handler must return a
    ///   list of available templates in the List parameter.
    /// </summary>
    property OnGetTemplateList: TfrxGetTemplateListEvent read FOnGetTemplateList write FOnGetTemplateList;
  end;

  { TfrxDesignerForm }

  TfrxDesignerForm = class(TfrxCustomDesigner)
    Bevel1: TBevel;
    StatusBar: TStatusBar;
    DockBottom: {$IFDEF FPC}TfrxControlBar{$ELSE}TControlBar{$ENDIF};
    DockTop: {$IFDEF FPC}TfrxControlBar{$ELSE}TControlBar{$ENDIF};
    TextTB: TfrxDockToolBar;
    PanelTB1: TfrxTBPanel;
    FontSizeCB: TfrxComboBox;
    FontNameCB: TfrxFontComboBox;
    BoldB: TToolButton;
    ItalicB: TToolButton;
    UnderlineB: TToolButton;
    SepTB8: TToolButton;
    FontColorB: TToolButton;
    HighlightB: TToolButton;
    SepTB9: TToolButton;
    TextAlignLeftB: TToolButton;
    TextAlignCenterB: TToolButton;
    TextAlignRightB: TToolButton;
    TextAlignBlockB: TToolButton;
    SepTB10: TToolButton;
    TextAlignTopB: TToolButton;
    TextAlignMiddleB: TToolButton;
    TextAlignBottomB: TToolButton;
    FrameTB: TfrxDockToolBar;
    FrameTopB: TToolButton;
    FrameBottomB: TToolButton;
    FrameLeftB: TToolButton;
    FrameRightB: TToolButton;
    SepTB11: TToolButton;
    FrameAllB: TToolButton;
    FrameNoB: TToolButton;
    SepTB12: TToolButton;
    FillColorB: TToolButton;
    FrameColorB: TToolButton;
    FrameStyleB: TToolButton;
    PanelTB2: TfrxTBPanel;
    FrameWidthCB: TfrxComboBox;
    StandardTB: TfrxDockToolBar;
    NewB: TToolButton;
    OpenB: TToolButton;
    SaveB: TToolButton;
    PreviewB: TToolButton;
    CutB: TToolButton;
    CopyB: TToolButton;
    PasteB: TToolButton;
    SepTB2: TToolButton;
    UndoB: TToolButton;
    RedoB: TToolButton;
    SepTB3: TToolButton;
    SepTB4: TToolButton;
    NewPageB: TToolButton;
    NewDialogB: TToolButton;
    DeletePageB: TToolButton;
    PageSettingsB: TToolButton;
    ShowGridB: TToolButton;
    AlignToGridB: TToolButton;
    ExtraToolsTB: TToolBar;
    PagePopup: TPopupMenu;
    CutMI1: TMenuItem;
    CopyMI1: TMenuItem;
    PasteMI1: TMenuItem;
    DeleteMI1: TMenuItem;
    SelectAllMI1: TMenuItem;
    SepMI8: TMenuItem;
    EditMI1: TMenuItem;
    MainMenu: TMainMenu;
    FileMenu: TMenuItem;
    EditMenu: TMenuItem;
    ViewMenu: TMenuItem;
    ToolbarsMI: TMenuItem;
    StandardMI: TMenuItem;
    TextMI: TMenuItem;
    FrameMI: TMenuItem;
    AlignmentMI: TMenuItem;
    ToolsMI: TMenuItem;
    InspectorMI: TMenuItem;
    DataTreeMI: TMenuItem;
    OptionsMI: TMenuItem;
    HelpMenu: TMenuItem;
    HelpContentsMI: TMenuItem;
    SepMI7: TMenuItem;
    AboutMI: TMenuItem;
    OpenDialog: TOpenDialog;
    TabPopup: TPopupMenu;
    NewPageMI1: TMenuItem;
    NewDialogMI1: TMenuItem;
    DeletePageMI1: TMenuItem;
    PageSettingsMI1: TMenuItem;
    ActionList: TActionList;
    ExitCmd: TAction;
    CutCmd: TAction;
    CopyCmd: TAction;
    PasteCmd: TAction;
    UndoCmd: TAction;
    RedoCmd: TAction;
    DeleteCmd: TAction;
    SelectAllCmd: TAction;
    EditCmd: TAction;
    BringToFrontCmd: TAction;
    SendToBackCmd: TAction;
    DeletePageCmd: TAction;
    NewItemCmd: TAction;
    NewPageCmd: TAction;
    NewDialogCmd: TAction;
    NewReportCmd: TAction;
    OpenCmd: TAction;
    SaveCmd: TAction;
    SaveAsCmd: TAction;
    VariablesCmd: TAction;
    PageSettingsCmd: TAction;
    PreviewCmd: TAction;
    NewMI: TMenuItem;
    NewReportMI: TMenuItem;
    NewPageMI: TMenuItem;
    NewDialogMI: TMenuItem;
    SepMI1: TMenuItem;
    OpenMI: TMenuItem;
    SaveMI: TMenuItem;
    SaveAsMI: TMenuItem;
    VariablesMI: TMenuItem;
    SepMI3: TMenuItem;
    PreviewMI: TMenuItem;
    SepMI4: TMenuItem;
    ExitMI: TMenuItem;
    UndoMI: TMenuItem;
    RedoMI: TMenuItem;
    SepMI5: TMenuItem;
    CutMI: TMenuItem;
    CopyMI: TMenuItem;
    PasteMI: TMenuItem;
    DeleteMI: TMenuItem;
    DeletePageMI: TMenuItem;
    SelectAllMI: TMenuItem;
    SepMI6: TMenuItem;
    BringtoFrontMI: TMenuItem;
    SendtoBackMI: TMenuItem;
    EditMI: TMenuItem;
    PanelTB3: TfrxTBPanel;
    ScaleCB: TfrxComboBox;
    ObjectsTB1: TToolBar;
    BandsPopup: TPopupMenu;
    ReportTitleMI: TMenuItem;
    ReportSummaryMI: TMenuItem;
    PageHeaderMI: TMenuItem;
    PageFooterMI: TMenuItem;
    HeaderMI: TMenuItem;
    FooterMI: TMenuItem;
    MasterDataMI: TMenuItem;
    DetailDataMI: TMenuItem;
    SubdetailDataMI: TMenuItem;
    GroupHeaderMI: TMenuItem;
    GroupFooterMI: TMenuItem;
    ColumnHeaderMI: TMenuItem;
    ColumnFooterMI: TMenuItem;
    ChildMI: TMenuItem;
    LeftDockSite1: TfrxDockSite;
    SepTB13: TToolButton;
    PageSettingsMI: TMenuItem;
    Timer: TTimer;
    ReportSettingsMI: TMenuItem;
    Data4levelMI: TMenuItem;
    Data5levelMI: TMenuItem;
    Data6levelMI: TMenuItem;
    SepMI10: TMenuItem;
    SepMI9: TMenuItem;
    ShowGuidesMI: TMenuItem;
    AutoGuidesMI: TMenuItem;
    ShowRulersMI: TMenuItem;
    DeleteHGuidesMI: TMenuItem;
    DeleteVGuidesMI: TMenuItem;
    SepMI11: TMenuItem;
    N1: TMenuItem;
    BringtoFrontMI1: TMenuItem;
    SendtoBackMI1: TMenuItem;
    SepMI12: TMenuItem;
    RotateB: TToolButton;
    RotationPopup: TPopupMenu;
    R0MI: TMenuItem;
    R45MI: TMenuItem;
    R90MI: TMenuItem;
    R180MI: TMenuItem;
    R270MI: TMenuItem;
    SetToGridB: TToolButton;
    FrameEditB: TToolButton;
    ReportMenu: TMenuItem;
    ReportDataMI: TMenuItem;
    OpenScriptDialog: TOpenDialog;
    SaveScriptDialog: TSaveDialog;
    ReportTreeMI: TMenuItem;
    ObjectsPopup: TPopupMenu;
    AlignTB: TfrxDockToolBar;
    AlignLeftsB: TToolButton;
    AlignHorzCentersB: TToolButton;
    AlignRightsB: TToolButton;
    AlignTopsB: TToolButton;
    AlignVertCentersB: TToolButton;
    AlignBottomsB: TToolButton;
    SpaceHorzB: TToolButton;
    SpaceVertB: TToolButton;
    CenterHorzB: TToolButton;
    CenterVertB: TToolButton;
    SameWidthB: TToolButton;
    SameHeightB: TToolButton;
    SepTB15: TToolButton;
    SepTB16: TToolButton;
    SepTB18: TToolButton;
    SepTB17: TToolButton;
    OverlayMI: TMenuItem;
    ReportStylesMI: TMenuItem;
    TabOrderMI: TMenuItem;
    N2: TMenuItem;
    FindMI: TMenuItem;
    FindNextMI: TMenuItem;
    ReplaceMI: TMenuItem;
    DMPPopup: TPopupMenu;
    BoldMI: TMenuItem;
    ItalicMI: TMenuItem;
    UnderlineMI: TMenuItem;
    SuperScriptMI: TMenuItem;
    SubScriptMI: TMenuItem;
    CondensedMI: TMenuItem;
    WideMI: TMenuItem;
    N12cpiMI: TMenuItem;
    N15cpiMI: TMenuItem;
    FontB: TToolButton;
    VerticalbandsMI: TMenuItem;
    HeaderMI1: TMenuItem;
    FooterMI1: TMenuItem;
    MasterDataMI1: TMenuItem;
    DetailDataMI1: TMenuItem;
    SubdetailDataMI1: TMenuItem;
    GroupHeaderMI1: TMenuItem;
    GroupFooterMI1: TMenuItem;
    ChildMI1: TMenuItem;
    N3: TMenuItem;
    GroupB: TToolButton;
    UngroupB: TToolButton;
    SepTB20: TToolButton;
    IntersectionsSelectionB: TToolButton;
    SepTB21: TToolButton;
    GroupCmd: TAction;
    UngroupCmd: TAction;
    GroupMI: TMenuItem;
    UngroupMI: TMenuItem;
    ConnectionsMI: TMenuItem;
    BackPanel: TPanel;
    ScrollBoxPanel: TPanel;
    ScrollBox: TfrxScrollBox;
    LeftRuler: TfrxRuler;
    TopRuler: TfrxRuler;
    CodePanel: TPanel;
    CodeTB: TToolBar;
    frTBPanel1: TfrxTBPanel;
    LangL: TLabel;
    LangCB: TfrxComboBox;
    OpenScriptB: TToolButton;
    SaveScriptB: TToolButton;
    SepTB19: TToolButton;
    RunScriptB: TToolButton;
    RunToCursorB: TToolButton;
    StepScriptB: TToolButton;
    StopScriptB: TToolButton;
    EvaluateB: TToolButton;
    BreakPointB: TToolButton;
    CodeDockSite: TfrxDockSite;
    LeftDockSite2: TfrxDockSite;
    RightDockSite: TfrxDockSite;
    TabPanel: TPanel;
    Panel1: TPanel;
    AddChildMI: TMenuItem;
    FindCmd: TAction;
    ReplaceCmd: TAction;
    FindNextCmd: TAction;
    ReportDataCmd: TAction;
    ReportStylesCmd: TAction;
    ReportOptionsCmd: TAction;
    ShowRulersCmd: TAction;
    ShowGuidesCmd: TAction;
    AutoGuidesCmd: TAction;
    DeleteHGuidesCmd: TAction;
    DeleteVGuidesCmd: TAction;
    OptionsCmd: TAction;
    HelpContentsCmd: TAction;
    AboutCmd: TAction;
    StandardTBCmd: TAction;
    TextTBCmd: TAction;
    FrameTBCmd: TAction;
    AlignTBCmd: TAction;
    ExtraTBCmd: TAction;
    InspectorTBCmd: TAction;
    DataTreeTBCmd: TAction;
    ReportTreeTBCmd: TAction;
    ToolbarsCmd: TAction;
{$IFNDEF FPC}
    WatchTBCmd: TAction;
    BreakPointsTBCmd: TAction;
{$ENDIF}
    FontColorPopupMenu: TPopupMenu;
    FillColorPopupMenu: TPopupMenu;
    FrameColorPopupMenu: TPopupMenu;
    ToolButton1: TToolButton;
    FillEditB: TToolButton;
    RevertInheritedMI: TMenuItem;
    RevertInheritedChildMI: TMenuItem;
    SelectAllOfTypeMI: TMenuItem;
    SelectAllOfTypeCmd: TAction;
    SelectAllOfTypeOnPageMI: TMenuItem;
    EdConfigMI: TMenuItem;
    EdConfigCmd: TAction;
    CopyContentCmd: TAction;
    CopyContent1: TMenuItem;
    CopyContent2: TMenuItem;
    FindB: TToolButton;
    StepOverScriptB: TToolButton;
    StepReturnScriptB: TToolButton;
    StylesDockTB: TfrxDockToolBar;
    StylesPan: TfrxTBPanel;
    StyleCB: TfrxComboBox;
    StyleSheetPan: TfrxTBPanel;
    StyleSheetCB: TfrxComboBox;
    StyleBookTB: TToolButton;
    StyleSheetAddTB: TToolButton;
    StyleSheetDeleteTB: TToolButton;
    StyleModeTB: TToolButton;
    StyleAddTB: TToolButton;
    StyleDeleteTB: TToolButton;
    SepStyle2: TToolButton;
    SepStyle3: TToolButton;
    SepStyle1: TToolButton;
    StyleToolBarMI: TMenuItem;
    StyleToolBarCmd: TAction;
    StyleModePopup: TPopupMenu;
    StyleBookDesignMI: TMenuItem;
    StyleBookPreviewMI: TMenuItem;
    StyleAddNewCmd: TAction;
    StyleDeleteCmd: TAction;
    StyleSheetAddCmd: TAction;
    StyleSheetDeleteCmd: TAction;
    StyleEditModeCmd: TAction;
{$IFNDEF FPC}
    WatchMI: TMenuItem;
    BreakPoints: TMenuItem;
{$ENDIF}
    {$IFDEF FPC}
    TimerFDock: TTimer;
    procedure FirstSetDock(Sender: TObject);
    {$ENDIF}
    procedure ExitCmdExecute(Sender: TObject);
    procedure ObjectsButtonClick(Sender: TObject);
    procedure StatusBarDrawPanel(StatusBar: TStatusBar;
      Panel: TStatusPanel; const ARect: TRect);
    procedure ScrollBoxMouseWheelUp(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
    procedure ScrollBoxMouseWheelDown(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
    procedure ScrollBoxResize(Sender: TObject);
    procedure ScaleCBClick(Sender: TObject);
    procedure ShowGridBClick(Sender: TObject);
    procedure AlignToGridBClick(Sender: TObject);
    procedure StatusBarDblClick(Sender: TObject);
    procedure StatusBarMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure InsertBandClick(Sender: TObject);
    procedure BandsPopupPopup(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure NewReportCmdExecute(Sender: TObject);
    procedure ToolButtonClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FontColorBClick(Sender: TObject);
    procedure FrameStyleBClick(Sender: TObject);
    procedure TabChange(Sender: TObject);
    procedure UndoCmdExecute(Sender: TObject);
    procedure RedoCmdExecute(Sender: TObject);
    procedure CutCmdExecute(Sender: TObject);
    procedure CopyCmdExecute(Sender: TObject);
    procedure PasteCmdExecute(Sender: TObject);
    procedure TimerTimer(Sender: TObject);
    procedure DeletePageCmdExecute(Sender: TObject);
    procedure NewDialogCmdExecute(Sender: TObject);
    procedure NewPageCmdExecute(Sender: TObject);
    procedure SaveCmdExecute(Sender: TObject);
    procedure SaveAsCmdExecute(Sender: TObject);
    procedure OpenCmdExecute(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure DeleteCmdExecute(Sender: TObject);
    procedure SelectAllCmdExecute(Sender: TObject);
    procedure EditCmdExecute(Sender: TObject);
    procedure TabChanging(Sender: TObject; var AllowChange: Boolean);
    procedure PageSettingsCmdExecute(Sender: TObject);
    procedure TopRulerDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure AlignLeftsBClick(Sender: TObject);
    procedure AlignRightsBClick(Sender: TObject);
    procedure AlignTopsBClick(Sender: TObject);
    procedure AlignBottomsBClick(Sender: TObject);
    procedure AlignHorzCentersBClick(Sender: TObject);
    procedure AlignVertCentersBClick(Sender: TObject);
    procedure CenterHorzBClick(Sender: TObject);
    procedure CenterVertBClick(Sender: TObject);
    procedure SpaceHorzBClick(Sender: TObject);
    procedure SpaceVertBClick(Sender: TObject);
    procedure SelectToolBClick(Sender: TObject);
    procedure PagePopupPopup(Sender: TObject);
    procedure BringToFrontCmdExecute(Sender: TObject);
    procedure SendToBackCmdExecute(Sender: TObject);
    procedure LangCBClick(Sender: TObject);
    procedure OpenScriptBClick(Sender: TObject);
    procedure SaveScriptBClick(Sender: TObject);
    procedure CodeWindowDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure CodeWindowDragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure VariablesCmdExecute(Sender: TObject);
    procedure ObjectBandBClick(Sender: TObject);
    procedure PreviewCmdExecute(Sender: TObject);
    procedure HighlightBClick(Sender: TObject);
    procedure TabMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure TabMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure TabMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure TabDragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure TabDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure SameWidthBClick(Sender: TObject);
    procedure SameHeightBClick(Sender: TObject);
    procedure NewItemCmdExecute(Sender: TObject);
    procedure TabOrderMIClick(Sender: TObject);
    procedure RunScriptBClick(Sender: TObject);
    procedure StopScriptBClick(Sender: TObject);
    procedure EvaluateBClick(Sender: TObject);
    procedure GroupCmdExecute(Sender: TObject);
    procedure UngroupCmdExecute(Sender: TObject);
    procedure ConnectionsMIClick(Sender: TObject);
    procedure LangSelectClick(Sender: TObject);
    procedure BreakPointBClick(Sender: TObject);
    procedure RunToCursorBClick(Sender: TObject);
    procedure CodeDockSiteDockOver(Sender: TObject;
      Source: TDragDockObject; X, Y: Integer; State: TDragState;
      var Accept: Boolean);
    procedure TabSetChange(Sender: TObject; NewTab: Integer;
      var AllowChange: Boolean);
    procedure FormShow(Sender: TObject);
    procedure AddChildMIClick(Sender: TObject);
    procedure FindCmdExecute(Sender: TObject);
    procedure ReplaceCmdExecute(Sender: TObject);
    procedure FindNextCmdExecute(Sender: TObject);
    procedure ReportDataCmdExecute(Sender: TObject);
    procedure ReportStylesCmdExecute(Sender: TObject);
    procedure ReportOptionsCmdExecute(Sender: TObject);
    procedure ShowRulersCmdExecute(Sender: TObject);
    procedure ShowGuidesCmdExecute(Sender: TObject);
    procedure AutoGuidesCmdExecute(Sender: TObject);
    procedure DeleteHGuidesCmdExecute(Sender: TObject);
    procedure DeleteVGuidesCmdExecute(Sender: TObject);
    procedure OptionsCmdExecute(Sender: TObject);
    procedure HelpContentsCmdExecute(Sender: TObject);
    procedure AboutCmdExecute(Sender: TObject);
    procedure StandardTBCmdExecute(Sender: TObject);
    procedure TextTBCmdExecute(Sender: TObject);
    procedure FrameTBCmdExecute(Sender: TObject);
    procedure AlignTBCmdExecute(Sender: TObject);
    procedure ExtraTBCmdExecute(Sender: TObject);
    procedure InspectorTBCmdExecute(Sender: TObject);
    procedure DataTreeTBCmdExecute(Sender: TObject);
    procedure ReportTreeTBCmdExecute(Sender: TObject);
    {$IFNDEF FPC}
    procedure WatchTBCmdExecute(Sender: TObject);
    procedure BreakPointsTBCmdExecute(Sender: TObject);
    {$ENDIF}
    procedure ToolbarsCmdExecute(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure FontColorPopupMenuPopup(Sender: TObject);
    procedure FillColorBClick(Sender: TObject);
    procedure FillColorPopupMenuPopup(Sender: TObject);
    procedure FrameColorBClick(Sender: TObject);
    procedure FrameColorPopupMenuPopup(Sender: TObject);
    procedure RevertInheritedMIClick(Sender: TObject);
    procedure SelectAllOfTypeCmdExecute(Sender: TObject);
    procedure EdConfigCmdExecute(Sender: TObject);
    procedure CopyContentCmdExecute(Sender: TObject);
    procedure FindBClick(Sender: TObject);
    procedure IntersectionsSelectionBClick(Sender: TObject);
    procedure StyleToolBarCmdExecute(Sender: TObject);
    procedure StyleBookDesignMIClick(Sender: TObject);
    procedure StyleEditModeCmdExecute(Sender: TObject);
    procedure StyleAddNewCmdExecute(Sender: TObject);
    procedure StyleDeleteCmdExecute(Sender: TObject);
    procedure StyleSheetCBChange(Sender: TObject);
    procedure StyleSheetAddCmdExecute(Sender: TObject);
    procedure StyleSheetDeleteCmdExecute(Sender: TObject);
  private
    { Private declarations }
    ObjectSelectB: TToolButton;
    HandToolB: TToolButton;
    ZoomToolB: TToolButton;
    TextToolB: TToolButton;
    FormatToolB: TToolButton;
    ObjectBandB: TToolButton;

    FClipboard: TfrxClipboard;
    FCodeWindow: TfrxSyntaxMemo;
    FColor: TColor;
    FFontColor: TColor;
    FFillColor: TColor;
    FFrameColor: TColor;
    FCoord1: String;
    FCoord2: String;
    FCoord3: String;
    FDialogForm: TfrxDialogForm;
    FEditAfterInsert: Boolean;
    FDataTree: TfrxDataTreeForm;
    FDropFields: Boolean;
    FGridAlign: Boolean;
    FGridSize1: Extended;
    FGridSize2: Extended;
    FGridSize3: Extended;
    FGridSize4: Extended;
    FInspector: TfrxObjectInspector;
    FLineStyle: TfrxFrameStyle;
    FLocalizedOI: Boolean;
    FLockSelectionChanged: Boolean;
    FModifiedBy: TObject;
    FMouseDown: Boolean;
    FOldDesignerComp: TfrxDesigner;
    FOldUnits: TfrxDesignerUnits;
    FPagePositions: TStrings;
    FPictureCache: TfrxPictureCache;
    FRecentFiles: TStringList;
    FRecentMenuIndex: Integer;
    FReportTree: TfrxReportTreeForm;
    FSampleFormat: TSampleFormat;
    FScale: Extended;
    FScriptFirstTime: Boolean;
    FScriptRunning: Boolean;
    FScriptStep: TScriptStepMode;
    FPrevScript: TfsScript;
    FScriptStopped: Boolean;
    FShowGrid: Boolean;
    FShowGuides: Boolean;
    FAutoGuides: Boolean;
    FAutoGuidesH: Boolean;
    FAutoGuidesV: Boolean;
    FShowRulers: Boolean;
    FShowStartup: Boolean;
    FStickToGuides: Boolean;
    FGuidesAsAnchor: Boolean;
  {$IFDEF UseTabset}
    FTabs: TTabSet;
  {$ELSE}
    FTabs: TTabControl;
  {$ENDIF}
    FToolsColor: TColor;
    FUndoBuffer: TfrxUndoBuffer;
    FUnits: TfrxDesignerUnits;
    FUnitsDblClicked: Boolean;
    FUpdatingControls: Boolean;
  {$IFNDEF FPC}
    FWatchList: TfrxWatchForm;
    FBreakPoints: TfrxBreakPointsForm;
  {$ELSE}
    FBWForm: TfrxBWForm;
  {$ENDIF}
    FWorkspace: TfrxDesignerWorkspace;
    FWorkspaceColor: TColor;
    FIntersections: Boolean;
    FIntersectionsColor: TColor;
    FStatusBarOldWindowProc: TWndMethod;
    FTemplatePath: String;
    FTemplateExt: String;
    FFilterList: TStringList;
    FSearchForm: TfrxSearchDesignForm;
    FFindSplitter: TSplitter;
    FSearchFormVisible: Boolean;
    FGuidesBtn: TSpeedButton;
    FEditStyles: Boolean;
    procedure AttachDialogFormEvents(Attach: Boolean);
    procedure CreateColorSelector(Sender: TToolButton);
    procedure CreateExtraToolbar;
    procedure CreateToolWindows;
    procedure CreateObjectsToolbar;
    procedure CreateWorkspace;
    function ChangeStyleSheetName(var ANewText: string; const AOldText: string; const AItemIndex: Integer): Boolean;
    procedure DialogFormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure DialogFormKeyPress(Sender: TObject; var Key: Char);
    procedure DialogFormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure DialogFormModify(Sender: TObject);
    procedure Done;
    procedure DoTopmosts(Enable: Boolean);
    procedure Init;
    procedure InitControls;
    procedure NormalizeTopmosts;
    procedure OnCodeChanged(Sender: TObject);
    procedure OnCodeCompletion(const Name: String; List: TfrxCompletionList);
    procedure OnColorChanged(Sender: TObject);
    procedure OnComponentMenuClick(Sender: TObject);
    procedure OnChangePosition(Sender: TObject);
    procedure OnDataTreeDblClick(Sender: TObject);
    procedure OnDisableDock(Sender: TObject; var DragObject: TDragDockObject);
    procedure OnEditObject(Sender: TObject);
    procedure OnEnableDock(Sender, Target: TObject; X, Y: Integer);
    procedure OnExtraToolClick(Sender: TObject);
    procedure OnInsertObject(Sender: TObject);
    procedure OnModify(Sender: TObject);
    procedure OnNotifyPosition(ARect: TfrxRect);
    function SearchParentStatement(Current: TfsScript): Boolean;
    procedure OnRunLine(Sender: TfsScript; const UnitName, SourcePos: String);
    procedure OnSelectionChanged(Sender: TObject);
    procedure OnStyleChanged(Sender: TObject);
    procedure OpenRecentFile(Sender: TObject);
    procedure OnSaveFilterExecute(Sender: TObject);
    procedure OnEnterWorkSpace(Sender: TObject);
    procedure ReadButtonImages;
    procedure RestorePagePosition;
    procedure RestoreTopmosts;
    procedure SavePagePosition;
    procedure SetScale(Value: Extended);
    procedure SetGridAlign(const Value: Boolean);
    procedure SetShowGrid(const Value: Boolean);
    procedure SetShowRulers(const Value: Boolean);
    procedure SetToolsColor(const Value: TColor);
    procedure SetUnits(const Value: TfrxDesignerUnits);
    procedure SetWorkspaceColor(const Value: TColor);
    procedure SetIntersections(const Value: Boolean);
    procedure SetIntersectionsColor(const Value: TColor);
    procedure SwitchToolbar;
    procedure UpdateCaption;
    procedure UpdateControls;
    procedure UpdatePageDimensions;
    procedure UpdateRecentFiles(NewFile: String);
    procedure UpdateStyles;
    procedure UpdateSyntaxType;
    procedure UpdateWatches;
    function AskSave: Word;
    function GetPageIndex: Integer;
    function GetReportName: String;
    function GetStickAccuracy: Extended;
    procedure SetShowGuides(const Value: Boolean);
    procedure SetAutoGuides(const Value: Boolean);
    procedure SetAutoGuidesH(const Value: Boolean);
    procedure SetAutoGuidesV(const Value: Boolean);
    procedure Localize;
    procedure Localize_Guides;
    procedure CreateLangMenu;
    procedure CMStartup(var Message: TMessage); message WM_USER + 1;
    procedure WMSysCommand(var Message: TWMSysCommand); message WM_SYSCOMMAND;
    procedure WMEnable(var Message: TMessage); message WM_ENABLE;
    {$IFNDEF FPC}
    procedure WMActivateApp(var Message: TWMActivateApp); message WM_ACTIVATEAPP;
    {$ENDIF}
    procedure StatusBarWndProc(var Message: TMessage);
    procedure CMDialogKey(var Message: TCMDialogKey); message CM_DIALOGKEY;
    procedure SetGridSize1(const Value: Extended);
    procedure SetGridSize2(const Value: Extended);
    procedure SetGridSize3(const Value: Extended);
    procedure SetGridSize4(const Value: Extended);
    procedure SetStickToGuides(const Value: Boolean);
    procedure SetGuidesAsAnchor(const Value: Boolean);
    procedure SetStickAccuracy(const Value: Extended);
	  procedure SetChange();
	  procedure UnSetChange();
    procedure DoAfterUpdateReport(EventType: TfrxDesignerUpdateEvent);
    function GetCurrentForm: TWinControl;
    function TopRulerDialogForm: Integer;
    function LeftRulerDialogForm: Integer;
    function GetTabName(AIndex: Integer): string;
    function GetSearchFormVisible: Boolean;
    procedure SetSearchFormVisible(Value: Boolean);
    procedure OnUpdateSerch(Sender: TObject; FindNext: Boolean);
    procedure GuideChanged(Sender: TObject; Action: TfrxRulerPointerAction; Position: Extended);
    procedure OnClickGuidesMode(Sender: TObject);
    procedure UpdateGuidesMode;
    procedure UpdateStyleMode;
{$IFNDEF FPC}
    procedure OnCodePopup(Sender: TObject);
    procedure AddAutoWatch(Sender: TObject);
{$ENDIF}
  protected
    procedure Activate; override;
    procedure Deactivate; override;
    procedure UpdateWorkSpace(Sender: TObject);
    procedure ScaleControls;
    procedure UpdateVirtualGuids(Sender: TObject; Position: Extended);
    procedure SetModified(const Value: Boolean); override;
    procedure SetPage(const Value: TfrxPage); override;
    function GetCode: TStrings; override;
    procedure ScrollBoxMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure AfterPPIChange; override;
    procedure BeforePPIChange; override;
    procedure AfterPPIMessage; override;
    function GetAvailablePreferences: TfrxPreferencesTypes; override;
    function GetDefaultPreferences: TObject; override;
    procedure FormShowBeforeLoad; override;
    procedure LoadFormPreferences(PreferencesStorage: TObject; DefPreferencesStorage: TObject); override;
    procedure SaveFormPreferences(PreferencesStorage: TObject; DefPreferencesStorage: TObject); override;
    procedure ResetFormPreferences(PreferencesStorage: TObject); override;
    function ShortCutRedirect(Action: TfrxShortcutAction): Boolean;
    function IsEditControlActive: Boolean;
  public
    destructor Destroy; override;
    { Public declarations }
    procedure DeleteObject(vComp: TfrxComponent); override;
    procedure AddPictureToCache(PicView: TfrxPictureView); override;
    function CheckOp(Op: TfrxDesignerRestriction): Boolean; override;
    procedure ClearLastActiveObject; override;
    function InsertExpression(const Expr: String): String; override;
    function IsChangedExpression(const InExpr: String; out OutExpr: String): Boolean; override;
    function IsDataTree(aObject: TObject): Boolean; override;
    function GetDataSelectedAsExp: String; override;
    procedure HostControls(Host: TWinControl); override;
    function SetFocusedControl(Control: TWinControl): Boolean; override;

    procedure InternalCopy; override;
    procedure InternalPaste; override;
    function InternalIsPasteAvailable: Boolean; override;
    procedure ReportDataEdit; override;
    procedure ReportEditVariable; override;

    procedure LoadFile(FileName: String; UseOnLoadEvent: Boolean);
    procedure LoadFilter(aFilter: TfrxCustomIOTransport; FileName: String; UseOnLoadEvent: Boolean);
    procedure Lock; override;
    procedure ReloadPages(Index: Integer); override;
    procedure ReloadReport; override;
    procedure ReloadObjects(ResetSelection: Boolean = True); override;
    procedure RestoreState(RestoreDefault: Boolean = False;
      RestoreMainForm: Boolean = False);
    function SaveFile(SaveAs: Boolean; UseOnSaveEvent: Boolean): Boolean;
    function IOTransport(aFilter: TfrxCustomIOTransport; SaveAs: Boolean; UseOnSaveEvent: Boolean): Boolean;
    procedure SetReportDefaults;
    procedure SwitchToCodeWindow;
    procedure SwitchToPage(vPage: TfrxPage);
    procedure UpdateDataTree; override;
    procedure UpdatePage; override;
    procedure UpdateFormPPI(ANewPPI: Integer); override;
    procedure UpdateResouces; override;
    function GetDefaultObjectSize: TfrxPoint;
    function mmToUnits(mm: Extended; X: Boolean = True): Extended;
    function UnitsTomm(mm: Extended; X: Boolean = True): Extended;
    procedure GetTemplateList(List: TStrings);
    procedure UpdateInspector; override;
    procedure DisableInspectorUpdate; override;
    procedure EnableInspectorUpdate; override;
    procedure SetSelection(AComponent: TfrxComponent; const PropName: String; SelStart, SelEnd: TPoint); override;
    property CodeWindow: TfrxSyntaxMemo read FCodeWindow;
    property DataTree: TfrxDataTreeForm read FDataTree;
    property DropFields: Boolean read FDropFields write FDropFields;
    property EditAfterInsert: Boolean read FEditAfterInsert write FEditAfterInsert;
    property FillColor: TColor read FFillColor write FFillColor;
    property GridAlign: Boolean read FGridAlign write SetGridAlign;
    property GridSize1: Extended read FGridSize1 write SetGridSize1;
    property GridSize2: Extended read FGridSize2 write SetGridSize2;
    property GridSize3: Extended read FGridSize3 write SetGridSize3;
    property GridSize4: Extended read FGridSize4 write SetGridSize4;
    property Inspector: TfrxObjectInspector read FInspector;
    property PictureCache: TfrxPictureCache read FPictureCache;
    property RecentFiles: TStringList read FRecentFiles;
    property ReportTree: TfrxReportTreeForm read FReportTree;
    property SampleFormat: TSampleFormat read FSampleFormat;
    property Scale: Extended read FScale write SetScale;
    property ShowGrid: Boolean read FShowGrid write SetShowGrid;
    property ShowGuides: Boolean read FShowGuides write SetShowGuides;
    property AutoGuides: Boolean read FAutoGuides write SetAutoGuides;
    property AutoGuidesH: Boolean read FAutoGuidesH write SetAutoGuidesH;
    property AutoGuidesV: Boolean read FAutoGuidesV write SetAutoGuidesV;
    property ShowRulers: Boolean read FShowRulers write SetShowRulers;
    property ShowStartup: Boolean read FShowStartup write FShowStartup;
    property ToolsColor: TColor read FToolsColor write SetToolsColor;
    property Units: TfrxDesignerUnits read FUnits write SetUnits;
    property Workspace: TfrxDesignerWorkspace read FWorkspace;
    property WorkspaceColor: TColor read FWorkspaceColor write SetWorkspaceColor;
    property Intersections: Boolean read FIntersections write SetIntersections;
    property IntersectionsColor: TColor read FIntersectionsColor write SetIntersectionsColor;
    property TemplatePath: String read FTemplatePath write FTemplatePath;
    property StickToGuides: Boolean read FStickToGuides write SetStickToGuides;
    property GuidesAsAnchor: Boolean read FGuidesAsAnchor write SetGuidesAsAnchor;
    property StickAccuracy: Extended read GetStickAccuracy write SetStickAccuracy;
    {$IFDEF FPC}
    property LockSelectionChanged: Boolean read FLockSelectionChanged write FLockSelectionChanged;
    {$ENDIF}
    property SearchFormVisible: Boolean read GetSearchFormVisible write SetSearchFormVisible;
    property SearchForm: TfrxSearchDesignForm read FSearchForm;
  end;

  TSampleFormat = class(TObject)
  private
    FMemo: TfrxCustomMemoView;
    procedure Clear;
  public
    constructor Create;
    destructor Destroy; override;
    procedure ApplySample(Memo: TfrxCustomMemoView);
    procedure SetAsSample(Memo: TfrxCustomMemoView);
    property Memo: TfrxCustomMemoView read FMemo;
  end;

  TfrxCustomSavePlugin = class(TObject)
  public
    FileFilter: String;
    procedure Save(Report: TfrxReport; const FileName: String); virtual; abstract;
  end;

{$IFDEF FPC}
//procedure Register;
{$ENDIF}

var
  frxDesignerComp: TfrxDesigner;
///  frxSavePlugin: TfrxCustomSavePlugin;

implementation
end.
