
{******************************************}
{                                          }
{             FastReport VCL               }
{             Design interface             }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxDsgnIntf;

interface

{$I frx.inc}

uses
  SysUtils, Types,
{$IFNDEF FPC}
  Windows, Messages,
{$ENDIF}
  Classes, Graphics, Controls, StdCtrls, Menus, TypInfo, Variants,
{$IFDEF FPC}
  LCLType, LazHelper, LCLProc,
{$ENDIF}  
  frxClass, frCore;
  

type
  /// <summary>
  ///   Property editor attributes.
  /// </summary>
  TfrxPropertyAttribute = (
    /// <summary>
    ///   The property represents the dropdown list of values. (This function
    ///   is exemplified in the "Color" property). If this attribute is
    ///   presented, the "GetValues" method should be overridden. <br />
    /// </summary>
    paValueList,
    /// <summary>
    ///   Sorts the list's elements. It is used together with paValueList.
    /// </summary>
    paSortList,
    /// <summary>
    ///   The property has an editor. If this attribute is presented, the ...
    ///   button is displayed in the right part of the editing line. The Edit
    ///   method is called on by clicking on it.
    /// </summary>
    paDialog,
    /// <summary>
    ///   Allow editing of the given property in some objects selected at the
    ///   same time. Some properties (such as "Name", etc) do not have this
    ///   attribute.
    /// </summary>
    paMultiSelect,
    /// <summary>
    ///   The property is an object of the TPersistent type and has its own
    ///   properties, which are also should be displayed. (This function is
    ///   exemplified in the "Font" property)
    /// </summary>
    paSubProperties,
    /// <summary>
    ///   It is impossible to modify a value in the editor line. Some
    ///   properties, being the "Class" or "Set" types, possess this attribute.
    /// </summary>
    paReadOnly,
    /// <summary>
    ///   Drawing of the property's value is performed via the "OnDrawItem"
    ///   method. If the "paValueList" attribute is defined, then drawing of
    ///   the drop-down list is performed via the OnDrawLBItem method.
    /// </summary>
    paOwnerDraw);
  /// <summary>
  ///   Property editor attributes.
  /// </summary>
  TfrxPropertyAttributes = set of TfrxPropertyAttribute;

  /// <summary>
  ///   The TfrxPropertyEditor class is the base class for all property
  ///   editors.
  /// </summary>
  TfrxPropertyEditor = class(TObject)
  private
    FDesigner: TfrxCustomDesigner;
    FPropList: TList;
    FItemHeight: Integer;
    FValues: TStrings;
    function GetPropInfo: PPropInfo;
    function GetComponent: TPersistent;
    function GetfrComponent: TfrxComponent;
  protected
    FCompList: TList;
    procedure GetStrProc(const s: String);
    /// <summary>
    ///   Method get a property value of type Extended.
    /// </summary>
    function GetFloatValue: Extended;
    /// <summary>
    ///   Method get a property value of type Integer.
    /// </summary>
    function GetOrdValue: frInteger;
    /// <summary>
    ///   Method get a property value of type Integer.
    /// </summary>
    function GetInt64Value: Int64;
    /// <summary>
    ///   Method get a property value of type String.
    /// </summary>
    function GetStrValue: String;
    /// <summary>
    ///   Method get a property value of type Variant.
    /// </summary>
    function GetVarValue: Variant;
    /// <summary>
    ///   Method sets a property value of type Extended.
    /// </summary>
    procedure SetFloatValue(Value: Extended);
    /// <summary>
    ///   Method sets a property value of type Integer.
    /// </summary>
    procedure SetOrdValue(Value: Int64);
    /// <summary>
    ///   Method sets a property value of type String.
    /// </summary>
    procedure SetStrValue(const Value: String);
    /// <summary>
    ///   Method sets a property value of type Variant.
    /// </summary>
    procedure SetVarValue(Value: Variant);
  public
    /// <summary>
    ///   The constructor.
    /// </summary>
    constructor Create(Designer: TfrxCustomDesigner); virtual;
    destructor Destroy; override;
    /// <summary>
    ///   Method is called in two cases: either by selecting a property, by
    ///   double-clicking its value, or (if a property has the paDialog
    ///   attribute) by clicking the ... button. The method should return
    ///   "True," if the property's value was modified.
    /// </summary>
    function Edit: Boolean; virtual;
    /// <summary>
    ///   The "GetAttributes" method is to return a set of the property's
    ///   attributes. You should override this method in most cases.
    /// </summary>
    function GetAttributes: TfrxPropertyAttributes; virtual;
    /// <summary>
    ///   Method returns the name of the property.
    /// </summary>
    function GetName: String; virtual;
    /// <summary>
    ///   The "GetExtraLBSize" method is called in case you defined the
    ///   "paValueList" attribute. The method returns the number of pixels, by
    ///   which the width of the drop-down list should be adjusted in order to
    ///   find room for the displayed picture. By default, this method returns
    ///   the value corresponding to the cell's height for property's
    ///   enveloping. If you need to deduce a picture, width of which is larger
    ///   than its height, the given method should be overridden.
    /// </summary>
    function GetExtraLBSize: Integer; virtual;
    /// <summary>
    ///   The "GetValue" method should return the property's value as a string
    ///   (it will be displayed in the object inspector). If you inherit from
    ///   the TfrxPropertyEditor basic class, it is necessary to override the
    ///   method.
    /// </summary>
    function GetValue: String; virtual;
    /// <summary>
    ///   The "GetValues" method should be overridden in case you defined the
    ///   "paValueList" attribute. This method should fill the "Values"
    ///   property with values.
    /// </summary>
    procedure GetValues; virtual;
    /// <summary>
    ///   The "SetValue" method is to set the property's value transferred as a
    ///   string. If you inherit from the TfrxPropertyEditor basic class, it is
    ///   necessary to override the method.
    /// </summary>
    procedure SetValue(const Value: String); virtual;
    /// <summary>
    ///   The "OnDrawLBItem" method is called when drawing a string in the
    ///   drop-down list, if you defined the paValueList attribute. In fact,
    ///   this method is the TListBox.OnDrawItem event's handler and has the
    ///   same set of parameters.
    /// </summary>
    procedure OnDrawLBItem(Control: TWinControl; Index: Integer;
      ARect: TRect; State: TOwnerDrawState); virtual;
    /// <summary>
    ///   The "OnDrawItem" method is called when drawing the property's value
    ///   in the object inspector (when the property is not selected; otherwise
    ///   its value is simply displayed in the editing line). For example, the
    ///   Color property's editor draws a rectangle, filled with the color
    ///   according to the value, at the left of the property's value.
    /// </summary>
    procedure OnDrawItem(Canvas: TCanvas; ARect: TRect); virtual;
    /// <summary>
    ///   Reference to the parent component (not to the property itself!), to
    ///   which the given property belongs.
    /// </summary>
    property Component: TPersistent read GetComponent;
    /// <summary>
    ///   Reference to the parent component (not to the property itself!), to
    ///   which the given property belongs. Cast to the TfrxComponent type (for
    ///   convenience of use in some cases).
    /// </summary>
    property frComponent: TfrxComponent read GetfrComponent;
    /// <summary>
    ///   Reference to the report's designer.
    /// </summary>
    property Designer: TfrxCustomDesigner read FDesigner;
    /// <summary>
    ///   Height of the item, in which the property is displayed. It can be
    ///   useful in the OnDrawXXX methods.
    /// </summary>
    property ItemHeight: Integer read FItemHeight write FItemHeight;
    /// <summary>
    ///   Link to the PPropInfo structure, which contains information about the
    ///   edited property.
    /// </summary>
    property PropInfo: PPropInfo read GetPropInfo;
    /// <summary>
    ///   Property's value displayed as a string.
    /// </summary>
    property Value: String read GetValue write SetValue;
    /// <summary>
    ///   The list of values. This property is to be filled in the "GetValue"
    ///   method, if the "paValueList" attribute is defined.
    /// </summary>
    property Values: TStrings read FValues;
  end;

  TfrxPropertyEditorClass = class of TfrxPropertyEditor;

  /// <summary>
  ///   The TfrxComponentEditor class is the base class for all component
  ///   editors.
  /// </summary>
  TfrxComponentEditor = class(TObject)
  private
    FComponent: TfrxComponent;
    FDesigner: TfrxCustomDesigner;
    FMenu: TMenu;
  protected
    /// <summary>
    ///   Adds a menu item to the context menu of the edited component.
    /// </summary>
    function AddItem(const Caption: String; Tag: Integer;
      Checked: Boolean = False; Own: TMenuItem = nil): TMenuItem;
  public
    /// <summary>
    ///   The constructor.
    /// </summary>
    constructor Create(Component: TfrxComponent; Designer: TfrxCustomDesigner;
      Menu: TMenu); virtual;
    /// <summary>
    ///   Method should perform necessary actions (for example, display a
    ///   dialogue box) and returns "True," if the component's state was
    ///   modified.
    /// </summary>
    function Edit: Boolean; virtual;
    /// <summary>
    ///   The "HasEditor" method should return "True" if your component has an
    ///   editor. If it either returns "False" or you do not override this
    ///   method, the editor will not be called.
    /// </summary>
    function HasEditor: Boolean; virtual;
    /// <summary>
    ///   You should override this method if you need to add items in the
    ///   component's context menu. Items should be added by AddItem method.
    /// </summary>
    procedure GetMenuItems; virtual;
    /// <summary>
    ///   Method is called, when you select one of your items in the
    ///   component's menu; response to the selected menu item should be
    ///   described here.
    /// </summary>
    function Execute(Tag: Integer; Checked: Boolean): Boolean; virtual;
    /// <summary>
    ///   Link to the edited component.
    /// </summary>
    property Component: TfrxComponent read FComponent write FComponent;
    /// <summary>
    ///   Link to the designer.
    /// </summary>
    property Designer: TfrxCustomDesigner read FDesigner;
  end;

  TfrxComponentEditorClass = class of TfrxComponentEditor;

  /// <summary>
  ///   The standard editor of properties of type Integer.
  /// </summary>
  TfrxIntegerProperty = class(TfrxPropertyEditor)
  public
    function GetValue: String; override;
    procedure SetValue(const Value: String); override;
  end;

  TfrxInt64Property = class(TfrxPropertyEditor)
  public
    function GetValue: String; override;
    procedure SetValue(const Value: String); override;
  end;

  /// <summary>
  ///   The standard editor of properties of type Extended.
  /// </summary>
  TfrxFloatProperty = class(TfrxPropertyEditor)
  public
    function GetValue: String; override;
    procedure SetValue(const Value: String); override;
  end;

  /// <summary>
  ///   The standard editor of properties of type Char.
  /// </summary>
  TfrxCharProperty = class(TfrxPropertyEditor)
  public
    function GetValue: String; override;
    procedure SetValue(const Value: String); override;
  end;

  /// <summary>
  ///   The standard editor of properties of type String.
  /// </summary>
  TfrxStringProperty = class(TfrxPropertyEditor)
  public
    function GetValue: String; override;
    procedure SetValue(const Value: String); override;
  end;

  /// <summary>
  ///   The standard editor of properties of type Enum.
  /// </summary>
  TfrxEnumProperty = class(TfrxPropertyEditor)
  public
    function GetAttributes: TfrxPropertyAttributes; override;
    function GetValue: String; override;
    procedure GetValues; override;
    procedure SetValue(const Value: String); override;
  end;

  /// <summary>
  ///   The standard editor of properties of type Set.
  /// </summary>
  TfrxSetProperty = class(TfrxPropertyEditor)
  public
    function GetAttributes: TfrxPropertyAttributes; override;
    function GetValue: String; override;
  end;

  /// <summary>
  ///   The standard editor of properties of type Set element.
  /// </summary>
  TfrxSetElementProperty = class(TfrxPropertyEditor)
  private
    FElement: Integer;
  public
    function GetAttributes: TfrxPropertyAttributes; override;
    function GetName: String; override;
    function GetValue: String; override;
    procedure GetValues; override;
    procedure SetValue(const Value: String); override;
   end;

  /// <summary>
  ///   The standard editor of properties of type TObject.
  /// </summary>
  TfrxClassProperty = class(TfrxPropertyEditor)
  public
    function GetAttributes: TfrxPropertyAttributes; override;
    function GetValue: String; override;
  end;

  /// <summary>
  ///   The standard editor of properties of type TComponent.
  /// </summary>
  TfrxComponentProperty = class(TfrxPropertyEditor)
  public
    function GetAttributes: TfrxPropertyAttributes; override;
    function GetValue: String; override;
    procedure GetValues; override;
    procedure SetValue(const Value: String); override;
  end;

  /// <summary>
  ///   The standard editor of properties of type TComponentName.
  /// </summary>
  TfrxNameProperty = class(TfrxStringProperty)
  public
    function GetAttributes: TfrxPropertyAttributes; override;
  end;

  /// <summary>
  ///   The standard editor of properties of type TClass.
  /// </summary>
  TfrxClassNameProperty = class(TfrxStringProperty)
  public
    function GetAttributes: TfrxPropertyAttributes; override;
    procedure GetValues; override;
    procedure SetValue(const Value: String); override;
  end;

  /// <summary>
  ///   The standard editor of properties of type TColor.
  /// </summary>
  TfrxColorProperty = class(TfrxIntegerProperty)
  public
    function GetAttributes: TfrxPropertyAttributes; override;
    function Edit: Boolean; override;
    function GetValue: String; override;
    procedure GetValues; override;
    procedure SetValue(const Value: String); override;
    procedure OnDrawLBItem(Control: TWinControl; Index: Integer;
      ARect: TRect; State: TOwnerDrawState); override;
    procedure OnDrawItem(Canvas: TCanvas; ARect: TRect); override;
  end;

  /// <summary>
  ///   The standard editor of properties of type TFont.
  /// </summary>
  TfrxFontProperty = class(TfrxClassProperty)
  public
    function Edit: Boolean; override;
    function GetAttributes: TfrxPropertyAttributes; override;
  end;

  /// <summary>
  ///   The standard editor of properties of type TFont.Name.
  /// </summary>
  TfrxFontNameProperty = class(TfrxStringProperty)
  public
    function GetAttributes: TfrxPropertyAttributes; override;
    procedure GetValues; override;
  end;

  /// <summary>
  ///   The standard editor of properties of type TFont.Charset.
  /// </summary>
  TfrxFontCharsetProperty = class(TfrxIntegerProperty)
  public
    function GetAttributes: TfrxPropertyAttributes; override;
    function GetValue: String; override;
    procedure GetValues; override;
    procedure SetValue(const Value: String); override;
  end;

  /// <summary>
  ///   The standard editor of properties of type TModalResult.
  /// </summary>
  TfrxModalResultProperty = class(TfrxIntegerProperty)
  public
    function GetAttributes: TfrxPropertyAttributes; override;
    function GetValue: String; override;
    procedure GetValues; override;
    procedure SetValue(const Value: String); override;
  end;

  /// <summary>
  ///   The standard editor of properties of type TShortCut.
  /// </summary>
  TfrxShortCutProperty = class(TfrxPropertyEditor)
  public
    function GetAttributes: TfrxPropertyAttributes; override;
    function GetValue: String; override;
    procedure GetValues; override;
    procedure SetValue(const Value: String); override;
  end;

  /// <summary>
  ///   The standard editor of properties of type TCursor.
  /// </summary>
  TfrxCursorProperty = class(TfrxIntegerProperty)
  public
    function GetAttributes: TfrxPropertyAttributes; override;
    function GetValue: String; override;
    procedure GetValues; override;
    procedure SetValue(const Value: String); override;
  end;

  TfrxDateTimeProperty = class(TfrxPropertyEditor)
  public
    function GetValue: String; override;
    procedure SetValue(const Value: String); override;
  end;

  TfrxDateProperty = class(TfrxPropertyEditor)
  public
    function GetValue: String; override;
    procedure SetValue(const Value: String); override;
  end;

  TfrxTimeProperty = class(TfrxPropertyEditor)
  public
    function GetValue: String; override;
    procedure SetValue(const Value: String); override;
  end;

{ Internal classes used by Object Inspector }

  TfrxPropertyList = class;

  TfrxPropertyItem = class(TCollectionItem)
  private
    FEditor: TfrxPropertyEditor;
    FExpanded: Boolean;
    FSubProperty: TfrxPropertyList;
  public
    destructor Destroy; override;
    property Editor: TfrxPropertyEditor read FEditor;
    property Expanded: Boolean read FExpanded write FExpanded;
    property SubProperty: TfrxPropertyList read FSubProperty;
  end;

  TfrxPropertyList = class(TCollection)
  private
    FComponent: TPersistent;
    FDesigner: TfrxCustomDesigner;
    FParent: TfrxPropertyList;
    FFilter: String;
    procedure AddProperties(PropertyList: TfrxPropertyList);
    procedure FillProperties(AClass: TPersistent);
    procedure FillCommonProperties(PropertyList: TfrxPropertyList);
    procedure SetComponent(Value: TPersistent);
    function GetPropertyItem(Index: Integer): TfrxPropertyItem;
    procedure SetFilter(const Value: String);
  public
    constructor Create(Designer: TfrxCustomDesigner);
    function Add: TfrxPropertyItem;
    function FindSame(const AToFind: TfrxPropertyItem): TfrxPropertyItem;
    property Component: TPersistent read FComponent write SetComponent;
    property Items[Index: Integer]: TfrxPropertyItem read GetPropertyItem; default;
    property Parent: TfrxPropertyList read FParent;
    property Filter: String read FFilter write SetFilter;
  end;


{ registered items }

  TfrxObjectCategory = (ctData, ctReport, ctDialog, ctDMP, ctNone);
  TfrxObjectCategories = set of TfrxObjectCategory;

  TfrxObjectItem = class(TCollectionItem)
  public
    ClassRef: TfrxComponentClass;
    ButtonBmp: TBitmap;
    ButtonImageIndex: Integer;
    ButtonHint: String;
    CategoryName: String;
    Flags: Word;
    Category: TfrxObjectCategories;
  end;

  TfrxComponentEditorItem = class(TCollectionItem)
  public
    ComponentClass: TfrxComponentClass;
    ComponentEditor: TfrxComponentEditorClass;
  end;

  TfrxPropertyEditorItem = class(TCollectionItem)
  public
    PropertyType: PTypeInfo;
    ComponentClass: TClass;
    PropertyName: String;
    EditorClass: TfrxPropertyEditorClass;
  end;

  TfrxExportFilterItem = class(TCollectionItem)
  public
    Filter: TfrxCustomExportFilter;
  end;

  TfrxWizardItem = class(TCollectionItem)
  public
    ClassRef: TfrxWizardClass;
    ButtonBmp: TBitmap;
    WizardButtonBmp: TBitmap;
    ButtonImageIndex: Integer;
    ToolBarButtonImageIndex: Integer;
    IsToolbarWizard: Boolean;
  end;

  /// <summary>
  ///   The TfrxObjectCollection class represents list of objects registered in
  ///   the FastReport. frxObjects global object contains an instance of this
  ///   class.
  /// </summary>
  TfrxObjectCollection = class(TCollection)
  private
    function GetObjectItem(Index: Integer): TfrxObjectItem;
  public
    constructor Create;
    /// <summary>
    ///   Registers a category. <br />
    /// </summary>
    /// <param name="CategoryName">
    ///   Name of the category.This name is used for identification of the
    ///   category only and does not appear anywhere else. It should also be
    ///   used when calling the "frxObjects.RegisterObject" procedure, if you
    ///   want to place the registered object inside the given category.
    /// </param>
    /// <param name="ButtonBmp">
    ///   Category bitmap of size 16x16.
    /// </param>
    /// <param name="ButtonHint">
    ///   Category hint. It is displayed as a hint, if the button with a
    ///   category is selected by the cursor.
    /// </param>
    /// <param name="ImageIndex">
    ///   Index of the bitmap from standard set of bitmaps if the ButtonBmp
    ///   parameter is nil.
    /// </param>
    procedure RegisterCategory(const CategoryName: String; ButtonBmp: TBitmap;
      const ButtonHint: String; ImageIndex: Integer = -1);
    /// <summary>
    ///   Registers an add-in object. During registration, you should specify
    ///   control class' name, its picture, and the name of a category, to
    ///   which it should be placed. If the name of a category is not
    ///   specified, the control is placed in the basic component palette. The
    ///   ButtonBmp size should be 16x16 pixels.
    /// </summary>
    /// <param name="ClassRef">
    ///   Component class reference.
    /// </param>
    /// <param name="ButtonBmp">
    ///   Component image.
    /// </param>
    /// <param name="CategoryName">
    ///   Category name.
    /// </param>
    procedure RegisterObject(ClassRef: TfrxComponentClass; ButtonBmp: TBitmap;
      const CategoryName: String = '');
    /// <summary>
    ///   Method is used for registration of internal FastReport objects.
    /// </summary>
    procedure RegisterObject1(ClassRef: TfrxComponentClass; ButtonBmp: TBitmap;
      const ButtonHint: String = ''; const CategoryName: String = '';
      Flags: Integer = 0; ImageIndex: Integer = -1;
      Category: TfrxObjectCategories = []);
    /// <summary>
    ///   Deletes the object from registered object's list.
    /// </summary>
    procedure Unregister(ClassRef: TfrxComponentClass);
    function FindByClassName(const Name: String): TfrxComponentClass;
    /// <summary>
    ///   list of registered components.
    /// </summary>
    property Items[Index: Integer]: TfrxObjectItem read GetObjectItem; default;
  end;

  /// <summary>
  ///   The TfrxComponentEditorCollection class represents a list of component
  ///   editors registered in the FastReport. frxComponentEditors global object
  ///   contains an instance of this class.
  /// </summary>
  TfrxComponentEditorCollection = class(TCollection)
  private
    function GetComponentEditorItem(Index: Integer): TfrxComponentEditorItem;
  public
    constructor Create;
    /// <summary>
    ///   Registers a component editor. The first parameter is the class' name,
    ///   for which the editor is to be created. The second parameter is the
    ///   name of the editor's class.
    /// </summary>
    /// <param name="ComponentClass">
    ///   Component class reference.
    /// </param>
    /// <param name="ComponentEditor">
    ///   Editor class reference.
    /// </param>
    procedure Register(ComponentClass: TfrxComponentClass;
      ComponentEditor: TfrxComponentEditorClass);
    /// <summary>
    ///   Remove editor class from registered list.
    /// </summary>
    procedure UnRegister(ComponentEditor: TfrxComponentEditorClass);
    /// <summary>
    ///   Creates new instance of component editor.
    /// </summary>
    function GetComponentEditor(Component: TfrxComponent;
      Designer: TfrxCustomDesigner; Menu: TMenu): TfrxComponentEditor;
    /// <summary>
    ///   List of registered component editors.
    /// </summary>
    property Items[Index: Integer]: TfrxComponentEditorItem
      read GetComponentEditorItem; default;
  end;

  /// <summary>
  ///   The TfrxPropertyEditorCollection class represents a list of property
  ///   editors registered in the FastReport. frxPropertyEditors global object
  ///   contains an instance of this class.
  /// </summary>
  TfrxPropertyEditorCollection = class(TCollection)
  private
    FEventEditorItem: Integer;
    function GetPropertyEditorItem(Index: Integer): TfrxPropertyEditorItem;
  public
    constructor Create;
    /// <summary>
    ///   Registers a property editor.
    /// </summary>
    /// <param name="PropertyType">
    ///   Information about the property's type, transferred via the "TypeInfo"
    ///   system function, for example TypeInfo(String).
    /// </param>
    /// <param name="ComponentClass">
    ///   Name of the component, the property of which you want to edit (may be
    ///   nil).
    /// </param>
    /// <param name="PropertyName">
    ///   Name of the property you want to edit (may be a blank string).
    /// </param>
    /// <param name="EditorClass">
    ///   Name of the property's editor.
    /// </param>
    procedure Register(PropertyType: PTypeInfo; ComponentClass: TClass;
      const PropertyName: String; EditorClass: TfrxPropertyEditorClass);
    procedure RegisterEventEditor(EditorClass: TfrxPropertyEditorClass);
    procedure UnRegister(EditorClass: TfrxPropertyEditorClass);
    function GetPropertyEditor(PropertyType: PTypeInfo; Component: TPersistent;
      PropertyName: String): Integer;
    property Items[Index: Integer]: TfrxPropertyEditorItem
      read GetPropertyEditorItem; default;
  end;

  /// <summary>
  ///   The TfrxExportFilterCollection class represents a list of export
  ///   filters registered in the FastReport. frxExportFilters global object
  ///   contains an instance of this class.
  /// </summary>
  TfrxExportFilterCollection = class(TCollection)
  private
    function GetExportFilterItem(Index: Integer): TfrxExportFilterItem;
  public
    constructor Create;
    /// <summary>
    ///   Registers an export filter.
    /// </summary>
    procedure Register(Filter: TfrxCustomExportFilter);
    /// <summary>
    ///   Deletes an export filter from the list of registered items.
    /// </summary>
    procedure Unregister(Filter: TfrxCustomExportFilter);
    property Items[Index: Integer]: TfrxExportFilterItem
      read GetExportFilterItem; default;
  end;

  TfrxSaveFilterItem = class(TCollectionItem)
  public
    SaveFilter: TfrxCustomIOTransport;
  end;

  TfrxSaveFiltersCollection = class(TCollection)
  private
    function GetSaveFilterItem(Index: Integer): TfrxSaveFilterItem;
  public
    constructor Create;
    procedure Register(Filter: TfrxCustomIOTransport);
    procedure Unregister(Filter: TfrxCustomIOTransport);
    property Items[Index: Integer]: TfrxSaveFilterItem
      read GetSaveFilterItem; default;
  end;

  /// <summary>
  ///   The TfrxWizardCollection class represents a list of wizards registered
  ///   in the FastReport. frxWizards global object contains an instance of
  ///   this class.
  /// </summary>
  TfrxWizardCollection = class(TCollection)
  private
    function GetWizardItem(Index: Integer): TfrxWizardItem;
  public
    constructor Create;
    /// <summary>
    ///   Registers a wizard. At registration, one enters the name of the
    ///   wizard's class, its picture, and specifies if the wizard is placed in
    ///   the "Wizards" toolbar. If the wizard should be placed in the toolbar,
    ///   the ButtonBmp size must be either 16x16 pixels, or 32x32 pixels
    ///   otherwise.
    /// </summary>
    procedure Register(ClassRef: TfrxWizardClass; ButtonBmp: TBitmap;
      IsToolbarWizard: Boolean = False; Bmp32x32: TBitmap = nil);
    procedure Register1(ClassRef: TfrxWizardClass; ImageIndex: Integer);
    /// <summary>
    ///   Deletes a wizard from the list of registered items.
    /// </summary>
    procedure Unregister(ClassRef: TfrxWizardClass);
    property Items[Index: Integer]: TfrxWizardItem read GetWizardItem; default;
  end;


{ internal methods }

function frxCreatePropertyList(ComponentList: TList; Designer: TfrxCustomDesigner; aFilter: String = ''): TfrxPropertyList;
procedure frxHideProperties(ComponentClass: TClass; const Properties: String);
function frxIsHiddenProperty(ComponentClass: TClass; const PropName: String): Boolean;


/// <summary>
///   List of add-in objects registered in the FastReport.
/// </summary>
function frxObjects: TfrxObjectCollection;
/// <summary>
///   List of component editors registered in the FastReport.
/// </summary>
function frxComponentEditors: TfrxComponentEditorCollection;
/// <summary>
///   List of property editors registered in the FastReport.
/// </summary>
function frxPropertyEditors: TfrxPropertyEditorCollection;
/// <summary>
///   List of export filters registered in the FastReport.
/// </summary>
function frxExportFilters: TfrxExportFilterCollection;
/// <summary>
///   List of wizards registered in the FastReport.
/// </summary>
function frxWizards: TfrxWizardCollection;


implementation
end.
