
{******************************************}
{                                          }
{             FastReport VCL               }
{           ERSI Shape Map File            }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxERSIShapeFileFormat;

interface

{$I frx.inc}

uses
  Classes, Contnrs,
{$IFNDEF FPC}
  Windows,
{$ELSE}
  LCLType, LCLIntf, LCLProc,
{$ENDIF}
  Graphics, Types, frxERSIShapeDBFImport, frxClass,
  frxMapHelpers, frxAnaliticGeometry;

const
  { ERSI Shape Types }
  ERSI_Null = 0;
  ERSI_Point = 1;
  ERSI_PolyLine = 3;
  ERSI_Polygon = 5;
  ERSI_MultiPoint = 8;
  ERSI_PointZ = 11;
  ERSI_PolyLineZ = 13;
  ERSI_PolygonZ = 15;
  ERSI_MultiPointZ = 18;
  ERSI_PointM = 21;
  ERSI_PolyLineM = 23;
  ERSI_PolygonM = 25;
  ERSI_MultiPointM = 28;
  ERSI_MultiPatch = 31;

type
  TERSIMainFileHeader = packed record
    FileCode: LongInt; { Big endian = 9994 }
    Unused: array[0..4] of LongInt;
    FileLength: LongInt; { Big endian; Measured in 16-bit words }
    Version: LongInt; { = 1000 }
    ShapeType: LongInt;
    Xmin, Ymin, Xmax, Ymax: Double;
    Zmin, Zmax, Mmin, Mmax: Double; { Unused, with value 0.0, if not Measured or Z type }
  end;

  TERSIMainFileRecordHeader = record
    RecordNumber: LongInt; { Big endian;  Record numbers begin at 1 }
    ContentLength: LongInt; { Big endian; Measured in 16-bit words }
    ShapeType: LongInt;
  end;

  TERSINull = class
  private
    FHeader: TERSIMainFileRecordHeader;
  public
    constructor Create(AHeader: TERSIMainFileRecordHeader);

    property Header: TERSIMainFileRecordHeader read FHeader;
  end;

  TERSIPoint = class(TERSINull)
  private
    FPoint: TDoublePoint;
  public
    constructor Create(AHeader: TERSIMainFileRecordHeader; Stream: TStream);
    property Point: TDoublePoint read FPoint;
  end;

  TDoublePointM = record
    X, Y, M: Double;
  end;
  TDoublePointMArray = array of TDoublePointM;

  TERSIPointM = class(TERSINull)
  private
    FPointM: TDoublePointM;
  public
    constructor Create(AHeader: TERSIMainFileRecordHeader; Stream: TStream);
    property PointM: TDoublePointM read FPointM;
  end;

  TDoublePointZ = record
    X, Y, Z, M: Double;
  end;
  TDoublePointZArray = array of TDoublePointZ;

  TERSIPointZ = class(TERSINull)
  private
    FPointZ: TDoublePointZ;
  public
    constructor Create(AHeader: TERSIMainFileRecordHeader; Stream: TStream);
    property PointZ: TDoublePointZ read FPointZ;
  end;

  TBox = record
    Xmin, Ymin, Xmax, Ymax: Double;
  end;

  TERSIMultiPoint = class(TERSINull)
  private
    FBox: TBox;
    FNumPoints: LongInt;
    FPoints: TDoublePointArray;
  public
    constructor Create(AHeader: TERSIMainFileRecordHeader; Stream: TStream);

    property Box: TBox read FBox;
    property NumPoints: LongInt read FNumPoints;
    property Points: TDoublePointArray read FPoints;
  end;

  TERSIMultiPointM = class(TERSIMultiPoint)
  private
    FMMin: Double;
    FMMax: Double;
    FPointsM: TDoubleArray;
  public
    constructor Create(AHeader: TERSIMainFileRecordHeader; Stream: TStream);

    property MMin: Double read FMMin;
    property MMax: Double read FMMax;
    property PointsM: TDoubleArray read FPointsM;
  end;

  TERSIMultiPointZ = class(TERSIMultiPoint)
  private
    FZMin: Double;
    FZMax: Double;
    FPointsZ: TDoubleArray;
    FMMin: Double;
    FMMax: Double;
    FPointsM: TDoubleArray;
  public
    constructor Create(AHeader: TERSIMainFileRecordHeader; Stream: TStream);

    property ZMin: Double read FZMin;
    property ZMax: Double read FZMax;
    property PointsZ: TDoubleArray read FPointsZ;
    property MMin: Double read FMMin;
    property MMax: Double read FMMax;
    property PointsM: TDoubleArray read FPointsM;
  end;

  TLongIntArray = array of LongInt;

  TERSIPolyLine = class(TERSINull)
  private
    FBox: TBox;
    FNumParts: LongInt;
    FNumPoints: LongInt;
    FPartFirsPointIndex: TLongIntArray;
    FPoints: TDoublePointArray;
  public
    constructor Create(AHeader: TERSIMainFileRecordHeader; Stream: TStream);

    function PartLastPointIndex(PartNumber: LongInt): LongInt;
    function PartCount(PartNumber: LongInt): LongInt;

    property Box: TBox read FBox;
    property NumParts: LongInt read FNumParts; { Number of Parts }
    property NumPoints: LongInt read FNumPoints; { Total Number of Points }
    property PartFirsPointIndex: TLongIntArray read FPartFirsPointIndex;
    property Points: TDoublePointArray read FPoints; { Points for All Parts }
  end;

  TERSIPolygon = class(TERSIPolyLine);

  TERSIPolyLineM = class(TERSIPolyLine)
  private
    FMMin: Double;
    FMMax: Double;
    FPointsM: TDoubleArray;
  public
    constructor Create(AHeader: TERSIMainFileRecordHeader; Stream: TStream);

    property MMin: Double read FMMin;
    property MMax: Double read FMMax;
    property PointsM: TDoubleArray read FPointsM;
  end;

  TERSIPolygonM = class(TERSIPolyLineM);

  TERSIPolyLineZ = class(TERSIPolyLine)
  private
    FZMin: Double;
    FZMax: Double;
    FPointsZ: TDoubleArray;
    FMMin: Double;
    FMMax: Double;
    FPointsM: TDoubleArray;
  public
    constructor Create(AHeader: TERSIMainFileRecordHeader; Stream: TStream);

    property ZMin: Double read FZMin;
    property MZax: Double read FZMax;
    property PointsZ: TDoubleArray read FPointsZ;
    property MMin: Double read FMMin;
    property MMax: Double read FMMax;
    property PointsM: TDoubleArray read FPointsM;
  end;

  TERSIPolygonZ = class(TERSIPolyLineZ);

  TMultiPath = class(TERSINull)
  private
    FBox: TBox;
    FNumParts: LongInt;
    FNumPoints: LongInt;
    FPartFirsPointIndex: TLongIntArray;
    FPartTypes: TLongIntArray;
    FPoints: TDoublePointArray;
    FZMin: Double;
    FZMax: Double;
    FPointsZ: TDoubleArray;
    FMMin: Double;
    FMMax: Double;
    FPointsM: TDoubleArray;
  public
    constructor Create(AHeader: TERSIMainFileRecordHeader; Stream: TStream);

    function PartLastPointIndex(PartNumber: LongInt): LongInt;
    function PartCount(PartNumber: LongInt): LongInt;

    property Box: TBox read FBox;
    property NumParts: LongInt read FNumParts; { Number of Parts }
    property NumPoints: LongInt read FNumPoints; { Total Number of Points }
    property PartFirsPointIndex: TLongIntArray read FPartFirsPointIndex;
    property PartTypes: TLongIntArray read FPartTypes;
    property Points: TDoublePointArray read FPoints; { Points for All Parts }
    property ZMin: Double read FZMin;
    property MZax: Double read FZMax;
    property PointsZ: TDoubleArray read FPointsZ;
    property MMin: Double read FMMin;
    property MMax: Double read FMMax;
    property PointsM: TDoubleArray read FPointsM;
  end;

  TERSIShapeFile = class(TfrxAbstractMapFile)
  private
    function GetRecordCount: Integer;
    function GetPolyPartsCount(iRecord: Integer): Integer;
    function GetPolyPointsCount(iRecord, iPart: Integer): Integer;
    function GetMultiPointCount(iRecord: Integer): Integer;
    function GetPolyPoint(iRecord, iPart, iPoint: Integer): TDoublePoint;
    function GetPoint(iRecord: Integer): TDoublePoint;
    function GetMultiPoint(iRecord, iPoint: Integer): TDoublePoint;
    function GetLegend(FieldName: AnsiString; iRecord: Integer): AnsiString;
    function GetERSIShapeType(iRecord: Integer): LongInt;
    function GetLegendByColumn(iColumn, iRecord: Integer): AnsiString;
    function GetLegendToString(iColumn, iRecord: Integer): string;
  protected
    FMainFileHeader: TERSIMainFileHeader;
    XRange, YRange: Double;
    FList: TObjectList;
    FDBFFile: TERSIDBFFile;

    procedure SwapEndianness(var Value: LongInt);
    procedure ReadFromStream(Stream: TStream);
  public
    constructor Create(const FileName: string);
    destructor Destroy; override;

    procedure GetColumnList(List: TStrings);
    procedure GetPartPoints(var PartPoints: TDoublePointArray; iRecord, iPart: Integer);

    property RecordCount: Integer read GetRecordCount;
    property PolyPartsCount[iRecord: Integer]: Integer read GetPolyPartsCount;
    property PolyPointsCount[iRecord, iPart: Integer]: Integer read GetPolyPointsCount;
    property MultiPointCount[iRecord: Integer]: Integer read GetMultiPointCount;
    property Legend[FieldName: AnsiString; iRecord: Integer]: AnsiString read GetLegend;
    property LegendByColumn[iColumn, iRecord: Integer]: AnsiString read GetLegendByColumn;
    property LegendToString[iColumn, iRecord: Integer]: string read GetLegendToString;

    property ERSIShapeType[iRecord: Integer]: LongInt read GetERSIShapeType;
    property PolyPoint[iRecord, iPart, iPoint: Integer]: TDoublePoint read GetPolyPoint;
    property Point[iRecord: Integer]: TDoublePoint read GetPoint;
    property MultiPoint[iRecord, iPoint: Integer]: TDoublePoint read GetMultiPoint;
  end;

implementation
end.

