
{******************************************}
{                                          }
{             FastReport VCL               }
{                  Gauge                   }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxGauge;

{$I frx.inc}

interface

uses
  Classes, Graphics, Types, Controls;

type
  TFloatPoint = record
    X, Y: Extended;
  end;

  TRay = record
    FP: TFloatPoint;
    Angle, Tangent: Extended; // Radians
  end;

  TfrxBaseGauge = class;

  TGaugeCoordinateConverter = class(TPersistent)
  private
  protected
    FRect: TRect;
    FWidth: Integer;
    FHeight: Integer;
    FIndent: TFloatPoint;
    FBaseGauge: TfrxBaseGauge;
    FRange: Extended;

    FCenter: TFloatPoint;
    FRadius: Extended;
    FAngle: Extended; // Radians

    function Part(Minimum, Value, Maximum: Extended): Extended; // 0..1
  public
    constructor Create(ABaseGauge: TfrxBaseGauge);
    procedure AssignTo(Dest: TPersistent); override;
    procedure Init(Rect: TRect; Indent: TFloatPoint);
    function ScreenRay(Value: Extended): TRay;
    function Value(Point: TPoint): Extended;
    procedure OutsideInside(fp1, fp2: TFloatPoint;
      out Outside, Inside: TFloatPoint);
  end;

  TGaugeScaleTicks = class(TPersistent)
  private
    FLength: Integer;
    FWidth: Integer;
    FColor: TColor;
    procedure SetColor(const Value: TColor);
    procedure SetLength(const Value: Integer);
    procedure SetWidth(const Value: Integer);
  protected
    FBaseGauge: TfrxBaseGauge;
  public
    constructor Create(ABaseGauge: TfrxBaseGauge);
    procedure AssignTo(Dest: TPersistent); override;
    procedure Draw(Canvas: TCanvas; Ray: TRay);
  published
    property Length: Integer read FLength write SetLength;
    property Width: Integer read FWidth write SetWidth;
    property Color: TColor read FColor write SetColor;
  end;

  TGaugeMargin = class(TPersistent)
  private
    FLeft: Integer;
    FTop: Integer;
    FRight: Integer;
    FBottom: Integer;
    procedure SetBottom(const Value: Integer);
    procedure SetLeft(const Value: Integer);
    procedure SetRight(const Value: Integer);
    procedure SetTop(const Value: Integer);
  protected
    FBaseGauge: TfrxBaseGauge;
  public
    constructor Create(ABaseGauge: TfrxBaseGauge);
    procedure AssignTo(Dest: TPersistent); override;
    procedure Init(ALeft, ATop, ARight, ABottom: Integer);
  published
    property Left: Integer read FLeft write SetLeft;
    property Top: Integer read FTop write SetTop;
    property Right: Integer read FRight write SetRight;
    property Bottom: Integer read FBottom write SetBottom;
  end;

  TGaugePointerKind = (pkSegment, pkTriangle, pkDiamond, pkPentagon, pkBand);

  TGaugePointer = class(TPersistent)
  private
    FColor: TColor;
    FBorderWidth: Integer;
    FBorderColor: TColor;
    FWidth: Integer;
    FHeight: Integer;
    procedure SetBorderColor(const Value: TColor);
    procedure SetBorderWidth(const Value: Integer);
    procedure SetColor(const Value: TColor);
    procedure SetWidth(const Value: Integer);
    procedure SetHeight(const Value: Integer);
  protected
    FBaseGauge: TfrxBaseGauge;
    procedure Paint(Canvas: TCanvas; Value: Extended); virtual; abstract;
    procedure PaintPair(Canvas: TCanvas; StartValue, EndValue: Extended);
      virtual; abstract;
    function fp(x, y: Extended): TFloatPoint;
    procedure PaintFloatPolygon(Canvas: TCanvas; Value: Extended;
      const FPs: array of TFloatPoint);
  public
    constructor Create(ABaseGauge: TfrxBaseGauge);
    procedure AssignTo(Dest: TPersistent); override;
    procedure SetupCanvas(Canvas: TCanvas);
    function IsPublishedWidth: boolean; virtual;
    function IsPublishedHeight: boolean; virtual;
    function IsPublishedColor: boolean; virtual;

    property Width: Integer read FWidth write SetWidth;
    property Height: Integer read FHeight write SetHeight;
    property Color: TColor read FColor write SetColor;
  published
    property BorderWidth: Integer read FBorderWidth write SetBorderWidth;
    property BorderColor: TColor read FBorderColor write SetBorderColor;
  end;

  TGaugeScale = class(TPersistent)
  private
    FFont: TFont;
    FTicks: TGaugeScaleTicks;
    FValueFormat: String;
    FVisible: Boolean;
    FVisibleDigits: Boolean;
    FBilateral: Boolean;
    procedure SetBilateral(const Value: Boolean);
    procedure SetValueFormat(const Value: String);
    procedure SetVisible(const Value: Boolean);
    procedure SetVisibleDigits(const Value: Boolean);
    procedure SetTicks(const Value: TGaugeScaleTicks);
    procedure SetFont(const Value: TFont);
  protected
    FTextSize: TSize;
    FOneDigit: Boolean;
    FOneValue: Extended;
    FBaseGauge: TfrxBaseGauge;

    procedure FontChanged(Sender: TObject);
    procedure DrawDigits(Canvas: TCanvas; Ray: TRay; Value: Extended);
    function Correction: Integer;
  public
    constructor Create(ABaseGauge: TfrxBaseGauge);
    destructor Destroy; override;
    procedure AssignTo(Dest: TPersistent); override;
    procedure SetupCanvas(Canvas: TCanvas);
    procedure Draw(Canvas: TCanvas; ValueStep: Extended);
    function Indent: TFloatPoint;
    procedure OneDigitAt(Value: Extended);
    procedure MultiDigits;
  published
    property Font: TFont read FFont write SetFont;
    property Ticks: TGaugeScaleTicks read FTicks write SetTicks;
    property ValueFormat: String read FValueFormat write SetValueFormat;
    property Visible: Boolean read FVisible write SetVisible;
    property VisibleDigits: Boolean read FVisibleDigits write SetVisibleDigits;
    property Bilateral: Boolean read FBilateral write SetBilateral;
  end;

  TGaugeKind = (gkHorizontal, gkVertical, gkCircle);
  TUpdateEvent = procedure of object;

  TfrxBaseGauge = class(TPersistent)
  private
    FOnUpdateOI: TUpdateEvent;
    FOnRepaint: TUpdateEvent;
    FMinorScale: TGaugeScale;
    FMajorScale: TGaugeScale;
    FPointer: TGaugePointer;
    FMajorStep: Extended;
    FMinorStep: Extended;
    FMinimum: Extended;
    FMaximum: Extended;
    FMargin: TGaugeMargin;
    FPointerKind: TGaugePointerKind;
    FKind: TGaugeKind;
    FAngle: Integer;
    procedure SetMaximum(const Value: Extended);
    procedure SetMinimum(const Value: Extended);
    procedure SetMajorStep(const Value: Extended);
    procedure SetMinorStep(const Value: Extended);
    procedure SetPointerKind(const Value: TGaugePointerKind);
    procedure SetMajorScale(const Value: TGaugeScale);
    procedure SetMargin(const Value: TGaugeMargin);
    procedure SetMinorScale(const Value: TGaugeScale);
    procedure SetPointer(const Value: TGaugePointer);
    procedure SetKind(const Value: TGaugeKind);
    procedure SetAngle(const Value: Integer);
  protected
    FScaleX, FScaleY, FScale: Extended;
    FLeftPressed: Boolean;
    FCC: TGaugeCoordinateConverter;
    FIsPrinting: Boolean;

    function CreatePointer(Old: TGaugePointer = nil; DestroyOld: Boolean = False): TGaugePointer;
    procedure UpdateOI;
    procedure Repaint;
    procedure OneDigit(P: Tpoint); virtual; abstract;
    procedure MultiDigits; virtual;
    procedure CorrectByMaximum; virtual; abstract;
    procedure CorrectByMinimum; virtual; abstract;
  public
    constructor Create;
    destructor Destroy; override;
    procedure AssignTo(Dest: TPersistent); override;
    procedure Draw(Canvas: TCanvas; ClientRect: TRect);
    procedure DrawPointer(Canvas: TCanvas); virtual; abstract;
    procedure SetXYScales(AScaleX, AScaleY: Extended; IsPrinting: Boolean = False);
    function Radians: Extended;

    // Return True if the owner has to be updated
    function DoMouseDown(X, Y: Integer; Button: TMouseButton;
      Shift: TShiftState): Boolean;
    function DoMouseMove(X, Y: Integer; Shift: TShiftState): Boolean;
    function DoMouseUp(X, Y: Integer; Button: TMouseButton;
      Shift: TShiftState): Boolean;
    function DoMouseWheel(Shift: TShiftState; WheelDelta: Integer;
      MousePos: TPoint): Boolean; virtual; abstract;

    property ScaleX: Extended read FScaleX;
    property ScaleY: Extended read FScaleY;
    property Scale: Extended read FScale;
    property Minimum: Extended read FMinimum write SetMinimum;
    property Maximum: Extended read FMaximum write SetMaximum;
  published
    property OnUpdateOI: TUpdateEvent read FOnUpdateOI write FOnUpdateOI;
    property OnRepaint: TUpdateEvent read FOnRepaint write FOnRepaint;
    property Kind: TGaugeKind read FKind write SetKind;
    property Angle: Integer read FAngle write SetAngle; // Degrees - for people
    property MinorScale: TGaugeScale read FMinorScale write SetMinorScale;
    property MajorScale: TGaugeScale read FMajorScale write SetMajorScale;
    property Pointer: TGaugePointer read FPointer write SetPointer;
    property PointerKind: TGaugePointerKind read FPointerKind write SetPointerKind;
    property MajorStep: Extended read FMajorStep write SetMajorStep;
    property MinorStep: Extended read FMinorStep write SetMinorStep;
    property Margin: TGaugeMargin read FMargin write SetMargin;
  end;

  TfrxGauge = class(TfrxBaseGauge)
  private
    FCurrentValue: Extended;
    procedure SetCurrentValue(const Value: Extended);
  protected
    procedure OneDigit(P: Tpoint); override;
    procedure CorrectByMaximum; override;
    procedure CorrectByMinimum; override;
  public
    constructor Create;
    procedure AssignTo(Dest: TPersistent); override;
    procedure DrawPointer(Canvas: TCanvas); override;
    function DoMouseWheel(Shift: TShiftState; WheelDelta: Integer;
      MousePos: TPoint): Boolean; override;
  published
    property Minimum;
    property Maximum;
    property CurrentValue: Extended read FCurrentValue write SetCurrentValue;
  end;

  TSelectedPointer = (spNo, spStart, spEnd, spBoth);
  TfrxIntervalGauge = class(TfrxBaseGauge)
  private
    FStartValue: Extended;
    FEndValue: Extended;
    procedure SetEndValue(const Value: Extended);
    procedure SetStartValue(const Value: Extended);
  protected
    FSelectedPointer: TSelectedPointer;
    procedure OneDigit(P: Tpoint); override;
    procedure MultiDigits; override;
    procedure CorrectByMaximum; override;
    procedure CorrectByMinimum; override;
  public
    constructor Create;
    procedure AssignTo(Dest: TPersistent); override;
    procedure DrawPointer(Canvas: TCanvas); override;

    function DoMouseWheel(Shift: TShiftState; WheelDelta: Integer;
      MousePos: TPoint): Boolean; override;
  published
    property Maximum;
    property Minimum;
    property EndValue: Extended read FEndValue write SetEndValue;
    property StartValue: Extended read FStartValue write SetStartValue;
  end;

    TSegmentPointer = class(TGaugePointer)
  protected
    procedure Paint(Canvas: TCanvas; Value: Extended); override;
    procedure PaintPair(Canvas: TCanvas; StartValue, EndValue: Extended); override;
  public
    function IsPublishedHeight: boolean; override;
  published
    property Height;
  end;

  TBasePolygonPointer = class(TGaugePointer)
  public
    function IsPublishedWidth: boolean; override;
    function IsPublishedHeight: boolean; override;
    function IsPublishedColor: boolean; override;
  published
    property Height;
    property Width;
    property Color;
  end;

  TTrianglePointer = class(TBasePolygonPointer)
  protected
    procedure Paint(Canvas: TCanvas; Value: Extended); override;
    procedure PaintPair(Canvas: TCanvas; StartValue, EndValue: Extended); override;
  end;

  TDiamondPointer = class(TBasePolygonPointer)
  protected
    procedure Paint(Canvas: TCanvas; Value: Extended); override;
    procedure PaintPair(Canvas: TCanvas; StartValue, EndValue: Extended); override;
  end;

  TPentagonPointer = class(TBasePolygonPointer)
  protected
    procedure Paint(Canvas: TCanvas; Value: Extended); override;
    procedure PaintPair(Canvas: TCanvas; StartValue, EndValue: Extended); override;
  end;

  TBandPointer = class(TGaugePointer)
  protected
    procedure Paint(Canvas: TCanvas; Value: Extended); override;
    procedure PaintPair(Canvas: TCanvas; StartValue, EndValue: Extended); override;
    procedure PaintRectBand(Canvas: TCanvas; Start, Current: TRay);
    procedure PaintCircleBand(Canvas: TCanvas; Start, Current: TRay);
  public
    function IsPublishedWidth: boolean; override;
    function IsPublishedColor: boolean; override;
  published
    property Width;
    property Color;
  end;

implementation
end.
