
{******************************************}
{                                          }
{             FastReport VCL               }
{           GeoJSON Format File            }
{                                          }
{         Copyright (c) 1998-2021          }
{           by Fast Reports Inc.           }
{                                          }
{******************************************}

unit frxGeoJSONFileFormat;

{$I frx.inc}

interface

uses
  Contnrs, Classes, Types,
  frxJSON, frxMapHelpers, frxAnaliticGeometry;

type
  TIntegerMatrix = array of TIntegerDynArray;
  TDoublePointMatrix3D = array of TDoublePointMatrix;

// A position is an array of numbers.  There MUST be two or more elements.
// The first two elements are longitude and latitude ... precisely in that order and using decimal numbers.
// Altitude ... MAY be included as an optional third element.
  TGeoJSONPosition = TDoublePoint;

// A GeoJSON object has a member with the name "type". The value of the member
//   MUST be one of the GeoJSON types (TGeoJSONType).
  TGeoJSONType = (
    gjtPoint, gjtMultiPoint,
    gjtLineString, gjtMultiLineString,
    gjtPolygon, gjtMultiPolygon,
    gjtGeometryCollection,
    gjtFeature, gjtFeatureCollection,
    gjtTopology // TopoJSON
  );

// A GeoJSON object MAY have a "bbox" member, the value of which MUST
//   be a bounding box array (see Section 5).
  TGeoJSONBBox = record
    Defined: Boolean;
    SouthWest, NorthEast: TGeoJSONPosition;
  end;

  TGeoJSONTopology = class;

  TGeoJSONObject = class(TAbstractTaggedElement)
  private
    FGeoType: TGeoJSONType;
    FBBox: TGeoJSONBBox;
    FParent: TGeoJSONObject;
    FProperties: TStringList;

    procedure ReadBBox(BBoxObject: TObject);
    function IsValidate(Condition: Boolean): Boolean;
  protected
    FIsValid: Boolean;
    FTopology: TGeoJSONTopology;

    class function ToDoublePoint(ja: TfrxJSONArray; i1, i2: Integer): TDoublePoint;
    class function ToDouble(ja: TfrxJSONArray; i: Integer): Double;
    class function MinCount: integer; virtual;

    function IsNeedDequantize: Boolean;
    procedure DeQuantizePoint(var Position: TGeoJSONPosition);
    procedure DeQuantizeLine(Line: TDoublePointArray);
    function IsCanSplitArcs: Boolean; virtual; // True

    procedure ReadProperties(PropertiesObject: TObject);

    // These seven functions frees ja
    function IsReadPosition(ja: TfrxJSONArray; out Position: TGeoJSONPosition): Boolean;
    function IsReadLine(ja: TfrxJSONArray; out Line: TDoublePointArray): Boolean;
    function IsReadMatrix(ja: TfrxJSONArray; out Matrix: TDoublePointMatrix): Boolean;
    function IsReadMatrix3D(ja: TfrxJSONArray; out Matrix3D: TDoublePointMatrix3D): Boolean;
    function IsReadTopoLine(ja: TfrxJSONArray; out Line: TDoublePointArray): Boolean;
    function IsReadTopoMatrix(ja: TfrxJSONArray; out Matrix: TDoublePointMatrix): Boolean;
    function IsReadTopoMatrix3D(ja: TfrxJSONArray; out Matrix3D: TDoublePointMatrix3D): Boolean;

    function GetTags: TStringList; override;

    property Parent: TGeoJSONObject read FParent;
  public
    constructor Create(JSON: TfrxJSON; AParent: TGeoJSONObject = nil);
    destructor Destroy; override;

    property GeoType: TGeoJSONType read FGeoType;
    property IsValid: Boolean read FIsValid;
    property BBox: TGeoJSONBBox read FBBox;
    property Properties: TStringList read FProperties;
    property Topology: TGeoJSONTopology read FTopology;
  end;

  TGeoJSONObjectArray = array of TGeoJSONObject;

// For type "Point", the "coordinates" member is a single position
  TGeoJSONPoint = class(TGeoJSONObject)
  private
    FPosition: TGeoJSONPosition;
  public
    constructor Create(JSON: TfrxJSON; AParent: TGeoJSONObject = nil);

    property Position: TGeoJSONPosition read FPosition;
  end;

// For type "MultiPoint", the "coordinates" member is an array of positions
  TGeoJSONMultiPoint = class(TGeoJSONObject)
  private
    FPositionArray: TDoublePointArray;

    function GetPointCount: integer;
    function GetPosition(nPoint: Integer): TGeoJSONPosition;
  protected
    procedure ReadPositionArray(JSON: TfrxJSON); virtual;
  public
    constructor Create(JSON: TfrxJSON; AParent: TGeoJSONObject = nil);

    property PointCount: integer read GetPointCount;
    property Position[nPoint: Integer]: TGeoJSONPosition read GetPosition;
  end;

// For type "LineString", the "coordinates" member is an array of two or more positions
  TGeoJSONLineString = class(TGeoJSONMultiPoint)
  protected
    class function MinCount: integer; override;
    function IsCanSplitArcs: Boolean; override; // False

    procedure ReadPositionArray(JSON: TfrxJSON); override;
  end;

// For type "MultiLineString", the "coordinates" member is an array of LineString coordinate arrays.
  TGeoJSONMultiLineString = class(TGeoJSONObject)
  private
    FPositionMatrix: TDoublePointMatrix;

    function GetPosition(nLine, nPoint: Integer): TGeoJSONPosition;
    function GetLineCount: Integer;
    function GetPointCount(nLine: Integer): Integer;
  protected
    class function MinCount: integer; override;
  public
    constructor Create(JSON: TfrxJSON; AParent: TGeoJSONObject = nil);

    property LineCount: Integer read GetLineCount;
    property PointCount[nLine: Integer]: Integer read GetPointCount;
    property Position[nLine, nPoint: Integer]: TGeoJSONPosition read GetPosition;
  end;

// A linear ring is a closed LineString with four or more positions.
// For type "Polygon", the "coordinates" member MUST be an array of linear ring coordinate arrays.
  TGeoJSONPolygon = class(TGeoJSONMultiLineString)
  protected
    class function MinCount: integer; override;
  end;

// For type "MultiPolygon", the "coordinates" member is an array of Polygon coordinate arrays.
  TGeoJSONMultiPolygon = class(TGeoJSONObject)
  private
    FPositionMatrix3D: TDoublePointMatrix3D;

    function GetPosition(nPoligon, nLine, nPoint: Integer): TGeoJSONPosition;
    function GetPoligonCount: Integer;
    function GetLineCount(nPoligon: Integer): Integer;
    function GetPointCount(nPoligon, nLine: Integer): Integer;
  protected
    class function MinCount: integer; override;
  public
    constructor Create(JSON: TfrxJSON; AParent: TGeoJSONObject = nil);

    property PoligonCount: Integer read GetPoligonCount;
    property LineCount[nPoligon: Integer]: Integer read GetLineCount;
    property PointCount[nPoligon, nLine: Integer]: Integer read GetPointCount;
    property Position[nPoligon, nLine, nPoint: Integer]: TGeoJSONPosition read GetPosition;
  end;

// A GeoJSON object with type "GeometryCollection" is a Geometry object.
// A GeometryCollection has a member with the name "geometries".  The value of "geometries"
//   is an array. Each element of this array is a GeoJSON Geometry object.
//  TGeoJSONGeometryType = gjtPoint .. gjtGeometryCollection;
  TGeoJSONGeometryCollection = class(TGeoJSONObject)
  private
    FGeometries: TGeoJSONObjectArray;

    function GetGeometry(Index: Integer): TGeoJSONObject;
    function GetGeometryCount: Integer;
  public
    constructor Create(JSON: TfrxJSON; AParent: TGeoJSONObject = nil);
    destructor Destroy; override;

    property GeometryCount: Integer read GetGeometryCount;
    property Geometry[Index: Integer]: TGeoJSONObject read GetGeometry;
  end;

// A Feature object has a "type" member with the value "Feature".
// A Feature object has a member with the name "geometry".  The value of the geometry member SHALL be either
//   a Geometry object as defined above or, in the case that the Feature is unlocated, a JSON null value.
// A Feature object has a member with the name "properties".  The value of the properties member
//   is an object (any JSON object or a JSON null value).
  TGeoJSONFeature = class(TGeoJSONObject)
  private
    FGeometry: TGeoJSONObject;
  protected
    function GetTags: TStringList; override;
  public
    constructor Create(JSON: TfrxJSON; AParent: TGeoJSONObject = nil);
    destructor Destroy; override;

    property Geometry: TGeoJSONObject read FGeometry;
  end;

// A FeatureCollection object has a member with the name "features".  The value of "features"
//   is a JSON array. Each element of the array is a Feature object as defined above.  It is
//   possible for this array to be empty.
  TGeoJSONFeatureCollection = class(TGeoJSONObject)
  private
    FFeatures: TGeoJSONObjectArray;

    function GetFeature(Index: Integer): TGeoJSONObject;
    function GetFeatureCount: Integer;
  public
    constructor Create(JSON: TfrxJSON; AParent: TGeoJSONObject = nil);
    destructor Destroy; override;

    property FeatureCount: Integer read GetFeatureCount;
    property Feature[Index: Integer]: TGeoJSONObject read GetFeature;
  end;

  TTransform = record
    Defined: Boolean;
    Scale: TGeoJSONPosition;
    Translate: TGeoJSONPosition;
  end;

// A topology is a TopoJSON object where the type members value is Topology.
//   A topology must have a member with the name objects whose value is another object.
//   The value of each member of this object is a geometry object.
//   A topology must have a member with the name arcs whose value is an array of arcs.
//   A topology may have a transform member, whose value must be a transform.
//   A topology may have a bbox member, whose value must be a bounding box array.
  TGeoJSONTopology = class(TGeoJSONObject)
  private
    FObjects: TGeoJSONObjectArray;
    FTransform: TTransform;
    FArcs: TDoublePointMatrix;

    function GetGeometry(Index: Integer): TGeoJSONObject;
    function GetObjectsCount: Integer;
    function GetArcsCount: Integer;
    function GetPointCount(nArc: Integer): Integer;
    function GetPosition(nArc, nPoint: Integer): TGeoJSONPosition;
  protected
    class function MinCount: integer; override;
    procedure DeQuantization;

    property ArcsCount: Integer read GetArcsCount;
  public
    constructor Create(JSON: TfrxJSON; AParent: TGeoJSONObject = nil);
    destructor Destroy; override;

    procedure ReadTransform(TransformObject: TObject);

    property ObjectsCount: Integer read GetObjectsCount;
    property Objects[Index: Integer]: TGeoJSONObject read GetGeometry;
    property PointCount[nArc: Integer]: Integer read GetPointCount;
    property Position[nArc, nPoint: Integer]: TGeoJSONPosition read GetPosition;
    property Transform: TTransform read FTransform;
    property Arcs: TDoublePointMatrix read FArcs;
  end;

  TGeoJSONFile = class(TfrxAbstractMapFile)
  protected
    FRoot: TGeoJSONObject;
    FIsValidBounds: Boolean;

    procedure Load(const FileName: string);
  public
    constructor Create(const FileName: string);
    destructor Destroy; override;

    property Root: TGeoJSONObject read FRoot;
    property IsValidBounds: Boolean read FIsValidBounds;
  end;

const
  GeoJSONGeometryType = [gjtPoint .. gjtGeometryCollection];
  GeoJSONNonGeometryType = [gjtFeature .. gjtTopology];

implementation
end.
