
{******************************************}
{                                          }
{             FastReport v7.0              }
{           HTML Add-In Object             }
{                                          }
{           Copyright (c) 2020             }
{            by Oleg Adibekov,             }
{            Fast Reports Inc.             }
{                                          }
{******************************************}

unit frxHTML;

interface

{$I frx.inc}

{$IFNDEF FPC}
  {$Define UseMetaFile }
{$ENDIF}

uses
  Classes, Graphics,
  frCore, frCoreClasses, frxClass, frxHTMLViewer,
  frxProtocolFactory, frBaseGraphicsTypes;

type

{$IFDEF DELPHI16}
  [ComponentPlatformsAttribute(pidWin32 or pidWin64)]
{$ENDIF}
  TfrxHTMLObject = class(TComponent) // fake component
  end;

  TfrxEmbeddedDataType = (edNone, edInternal, edExternal);

  TfrxHtmlView = class(TfrxStretcheable, IfrxDataLinkObject, IfrxCachedView, IfrxDataBindingComponent)
  private
    FAllowExpressions: Boolean;
    FExpressionDelimiters: String;
    FGapX: Extended;
    FGapY: Extended;
    FWysiwyg: Boolean;
    FFilePath: String;
    FDataLink: TfrxDataLink;
    FEmbeddedObjects: TfrxEmbeddedDataType;
    FData: WideString;
    FNewCacheType: TfrxCachedGraphicType;
    FIsIndexesRead: Boolean;
    FNeedReload: Boolean;
    FReloadLocked: Boolean;

    function IsExprDelimitersStored: Boolean;
    function UsePrinterCanvas: Boolean;

    procedure ReadData(Stream: TStream);
    procedure WriteData(Stream: TStream);

    procedure ReadCSSCacheData(Stream: TStream);
    procedure WriteCSSCacheData(Stream: TStream);

    procedure ReadImageCacheData(Stream: TStream);
    procedure WriteImageCacheData(Stream: TStream);

    procedure ReadIndexCacheData(Stream: TStream);
    procedure WriteIndexCacheData(Stream: TStream);

    function GetDefBackground: TColor;
    procedure SetDefBackground(const Value: TColor);
    function GetFontColor: TColor;
    procedure SetFontColor(const Value: TColor);
    function GetFontName: TFontName;
    procedure SetFontName(const Value: TFontName);
    function GetFontSize: Integer;
    procedure SetFontSize(const Value: Integer);
    function GetHotSpotColor: TColor;
    procedure SetHotSpotColor(const Value: TColor);
    function GetPreFontName: TFontName;
    procedure SetPreFontName(const Value: TFontName);
    function GetMarginHeight: Integer;
    procedure SetMarginHeight(const Value: Integer);
    function GetMarginWidth: Integer;
    procedure SetMarginWidth(const Value: Integer);

    procedure Reload;
    function GetFilePath: String;
    procedure SetFilePath(const Value: String);
    procedure ReadFilePath(Reader: TReader);
    procedure WriteFilePath(Writer: TWriter);

    procedure ClearObjectData;
    function LoadDataStream(Stream: TStream; const NewLink: String): Boolean;
    function GetLink(LoadMethod: TfrxDataLinkLoadMethod): String;
    function IsExpressionLink: Boolean;
    function GetDataLink: TfrxDataLink;
    procedure SetDataLink(const Value: TfrxDataLink);
    function IsDataLinkStored: Boolean;
    procedure SetCachedGraphic(const PictureCache: IfrxPictureCache);
    procedure ReleaseCachedGraphic;
    function IsUpdateRequired(NewCacheType: TfrxCachedGraphicType): Boolean;
    procedure DoDelayLoad;
    procedure GetCachedGraphic(ACacheType: TfrxCachedGraphicType; const PictureCache: IfrxPictureCache);
    procedure SetEmbeddedObjects(const Value: TfrxEmbeddedDataType);
    function IsEmbeddedObjects: Boolean;
    procedure DoOnImageCacheChanged(Sender: TObject);
  protected
    FHtmlViewer: TfrxHtmlViewer;
    FPartStart: Extended;
    FPartHeight: Extended;
    FTempPartStart: Extended;
    FOldText: WideString;
    FOldStart: Extended;
    FOldHeight: Extended;

    function CreateGraphic: TGraphic;
    function CalcPartHeight: Integer;

    procedure DefineProperties(Filer: TFiler); override;

    function IsLoadPicture(const ComplexName: string; out BLOB: WideString): Boolean;
    procedure LoadFromDataField;
    procedure GetExpressionDelimiters(out LeftDlm, RightDlm: WideString);
    procedure Loaded; override;
    function IsPreviewPages: Boolean;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;

    procedure BeforePrint; override;
    procedure GetData; override;
    procedure AfterPrint; override;
    function CalcHeight: Extended; override; // Calculates and returns the object's height according to the data placed in it.

    class function GetDescription: string; override;

    procedure InitPart; override;
    function DrawPart: Extended; override;
    function HasNextDataPart(aFreeSpace: Extended): Boolean; override;

    property HtmlViewer: TfrxHtmlViewer read FHtmlViewer;
  published
    property AllowExpressions: Boolean read FAllowExpressions write FAllowExpressions default True;
    property BrushStyle;
    property Color;
    property Cursor;
    property DataField;
    property DataSet;
    property DataSetName;
    property DataLink: TfrxDataLink read GetDataLink write SetDataLink stored IsDataLinkStored;
    property ExpressionDelimiters: String read FExpressionDelimiters write FExpressionDelimiters stored IsExprDelimitersStored;
    property FillType;
    property Fill;
    property Frame;
    /// <summary>
    ///   The left indent of the text, in pixels.
    /// </summary>
    property GapX: Extended read FGapX write FGapX;
    /// <summary>
    ///   The top indent of the text, in pixels.
    /// </summary>
    property GapY: Extended read FGapY write FGapY;
    property TagStr;
    property URL;
    /// <summary>
    ///   Determines if the object should use the printer canvas to format the
    ///   text. A printer should be installed and ready.
    /// </summary>
    property Wysiwyg: Boolean read FWysiwyg write FWysiwyg default True;
    /// <summary>
    ///   Determines if images and CSS files should be saved in the report.
    /// </summary>
    property EmbeddedObjects: TfrxEmbeddedDataType read FEmbeddedObjects write SetEmbeddedObjects default edNone;

    { HTMLViever properties }
    property DefBackground: TColor read GetDefBackground write SetDefBackground;
    property DefFontColor: TColor read GetFontColor write SetFontColor;
    property DefFontName: TFontName read GetFontName write SetFontName;
    property DefFontSize: Integer read GetFontSize write SetFontSize; // pt
    property DefHotSpotColor: TColor read GetHotSpotColor write SetHotSpotColor;
    property DefPreFontName: TFontName read GetPreFontName write SetPreFontName;
    property MarginHeight: Integer read GetMarginHeight write SetMarginHeight;
    property MarginWidth: Integer read GetMarginWidth write SetMarginWidth;
    property Style;
  end;

implementation
end.
