{
HtmlViewer Version 11.7
Copyright (c) 2010-2016 by Bernd Gabriel

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Note that the source modules HTMLGIF1.PAS and DITHERUNIT.PAS
are covered by separate copyright notices located in those modules.
}

{$I frxHTMLCons.inc}

unit frxHTMLBuffer;

interface

uses
{$ifdef LCL}
  Classes, SysUtils,
  LclIntf, LclType, frLazHTMLMisc,
{$else}
  Windows,
  Classes, SysUtils,
{$endif}
  Graphics, 
{$ifdef UNICODE}
  AnsiStrings,
{$endif}
  //
  frxHTMLGlobals;

type
  EConversionError = class(Exception);

  TBuffCodePage = LongInt;
  TBuffCharSet = SmallInt;  // slightly more than a TFontCharset.

  TBuffChar = ThtChar;      // the type of a wide char
  PBuffChar = PhtChar;
  TBuffString = ThtString;  // the type of a wide string
  TBuffStringList = ThtStringList;  // the type of a wide string

  TBuffPointer = record
    case Integer of
      0: (BytePtr: PByte;);
      1: (WordPtr: PWord;);
      2: (AnsiChr: PAnsiChar;);
      3: (WideChr: PWideChar;);
    end;

  TBuffConverter = class
  protected
    FPos, FEnd: TBuffPointer;
    FCodePage, FInitalCodePage: TBuffCodePage;
    function GetNext: Word;
    procedure Assign(Source: TBuffConverter); virtual;
  public
    constructor Create(CurrPtr, EndPtr: TBuffPointer; CodePage, InitalCodePage: TBuffCodePage); virtual;
    constructor CreateCopy(Source: TBuffConverter); overload;

    function NextChar: TBuffChar; virtual;
    function PeekChar: TBuffChar;
    function AsString: TBuffString; virtual;
    property CodePage: TBuffCodePage read FCodePage;
  end;
  TBuffConverterClass =  class of TBuffConverter;

  TBuffArray = array of Byte;
  TBuffArray4 = array [0..3] of Byte;

  // BG, 11.11.2011: Issue 93: LoadStrings failure
  TBufferState = set of (
    bsFixedCodePage // ignore writing to property CodePage.
                    // Helps to ignore <META http-equiv="Content-Type" ...>,
                    // if buffer is an already converted TBuffString.
  );

  // BG, 17.12.2010: helps converting any kind of stream to WideChars.
  //
  // Although loading files via ThtStringList.LoadFromFile(FileName) produces WideChars,
  // these WideChars may not be coded correctly, if the file is not a unicode file.
  // For those cases you can apply a CodePage to the TBuffString versions of Convert()/Create().
  // Or you can used the PByte versions with the InitialCodePage parameters and set InitialCodePage to CP_UTF16LE.
  //
  // Avoid passing such raw ThtStringList.Text to HtmlViewer.
  // HtmlViewer assumes correct unicode in ThtString.
  // You MUST convert it before passing it to HtmlViewer: HtmlViewer.LoadFromString(TBuffer.Convert(StringList.Text, CodePage));
  // Or you let HtmlViewer load it from file or a TFileStream.
  //
  TBuffer = class
  private
    FBuffer: TBuffArray;
    FStart: TBuffPointer;
    FEnd: TBuffPointer;
    FName: TBuffString;
    FConverter: TBuffConverter;
    FState: TBufferState;
    FBomLength: Cardinal;

    function GetCodePage: TBuffCodePage;  {$ifdef UseInline} inline; {$endif}
    function GetPosition: Integer; {$ifdef UseInline} inline; {$endif}
    procedure DetectCodePage;
    procedure Reset;
    procedure SetCodePage(Value: TBuffCodePage); {$ifdef UseInline} inline; {$endif}
    procedure SetCodePages(Value, Initial: TBuffCodePage);
    procedure SetPosition(const Value: Integer); {$ifdef UseInline} inline; {$endif}
    procedure SetStream(Stream: TStream); {$ifdef UseInline} inline; {$endif}
  public
    class function Convert(Text: PByte; ByteCount: Integer; CodePage, InitialCodePage: TBuffCodePage): TBuffString; overload; {$ifdef UseInline} inline; {$endif}
    class function Convert(Text: PByte; ByteCount: Integer; CodePage: TBuffCodePage): TBuffString; overload; {$ifdef UseInline} inline; {$endif}
    class function Convert(const Text: TBuffString; CodePage: TBuffCodePage): TBuffString; overload; {$ifdef UseInline} inline; {$endif}
    constructor Create(const Doc: TBuffer); overload;
    constructor Create(Stream: TStream; CodePage: TBuffCodePage; Name: TBuffString = ''); overload;
    constructor Create(Stream: TStream; Name: TBuffString = ''); overload;
    constructor Create(Text: PByte; ByteCount: Integer; CodePage, InitialCodePage: TBuffCodePage; const Name: TBuffString = ''); overload;
    constructor Create(Text: PByte; ByteCount: Integer; CodePage: TBuffCodePage; const Name: TBuffString = ''); overload;
    constructor Create(const Text: TBuffString; CodePage: TBuffCodePage; const Name: TBuffString = ''); overload;
    constructor Create(const Text: TBuffString; const Name: TBuffString = ''; CodePage: TBuffCodePage = CP_UTF16LE); overload;
    destructor Destroy; override;
    function AsString: TBuffString; {$ifdef UseInline} inline; {$endif}
    function GetString(FromIndex, UntilIndex: Integer): TBuffString;
    function NextChar: TBuffChar; {$ifdef UseInline} inline; {$endif}
    function PeekChar: TBuffChar; {$ifdef UseInline} inline; {$endif}
    function Size: Integer; {$ifdef UseInline} inline; {$endif}
    procedure Assign(Source: TBuffer); virtual;
    procedure AssignTo(Destin: TObject); virtual;
    property CodePage: TBuffCodePage read GetCodePage write SetCodePage;
    property Name: TBuffString read FName;
    property Position: Integer read GetPosition write SetPosition;
    property BomLength: Cardinal read FBomLength;
  end;

// BG, 13.10.2012: Code Page Management
//
// You can register your own code page to unicode converters or unregister your own or predefined converters.
// Registering an existing converter again updates it.
//
// As each code page may have any number of aliases there are 2 sets of un-/register methods.
// One to register a code page name to code page number mappings and
// one to register the converter. 

// Add a name of a code page. Code pages can have many names, but each name names one code page only:
procedure RegisterCodePageName(const CodePageName: TBuffString; CodePage: TBuffCodePage);
// Remove a name of a code page:
procedure UnregisterCodePageName(const CodePageName: TBuffString);
// Get code page by name:
function StrToCodePage(const CodePageName: TBuffString): TBuffCodePage;

// Add a code page to unicode converter. Only one converter per code page is allowed,
// but a converter may handle more than one code page:
procedure RegisterCodePageToUnicodeConverter(
  CodePage: TBuffCodePage; Converter: TBuffConverterClass;
  const ConverterName: TBuffString = ''; CharSet: TBuffCharSet = UNKNOWN_CHARSET);
// Remove a code page to unicode converter:
procedure UnregisterCodePageToUnicodeConverter(CodePage: TBuffCodePage);
// Get charset of code page:
function CodePageToCharSet(CodePage: TBuffCodePage): TBuffCharSet;

implementation
end.


