{
Version   11.9
Copyright (c) 1995-2008 by L. David Baldwin
Copyright (c) 2008-2018 by HtmlViewer Team

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Note that the source modules HTMLGIF1.PAS and DITHERUNIT.PAS
are covered by separate copyright notices located in those modules.
}
{
ANGUS - LoadDocument, don't add physical disk file path to URL, it corrupts referrer
}

{$I frxHTMLCons.inc}

unit frxHTMLViewer;

interface

uses
{$ifdef LCL}
  LclIntf, LclType, LMessages, types, FPimage, frLazHTMLMisc, frLazGraphicClasses,
{$else}
  Windows,
{$endif}
  Messages, Classes, Graphics, Contnrs, SysUtils,
  frxHTMLURLSubs,
  frxHTMLGlobals,
  frxHTMLBuffer,
  frxHTMLImages,
  frxHTMLUn2,
  frxHTMLRead,
  frxHTMLSubs,
  frxHTMLStyleTypes,
  frxHTMLStyleUn,
  frxHTMLCaches
{$IFDEF FPC}
  , LazHelper
{$ENDIF}
;

type
  EhtException = class(Exception);
  EhtLoadError = class(EhtException);
  EExcessiveSizeError = class(EhtException);

  THtmlViewerOption = (
    htOverLinksActive, htNoLinkUnderline, htPrintTableBackground,
    htPrintBackground, htPrintMonochromeBlack, htShowDummyCaret,
    htShowVScroll, htNoWheelMouse, htNoLinkHilite,
    htNoFocusRect, //MK20091107
    htAllowHotSpotDblClick
    );
  THtmlViewerOptions = set of THtmlViewerOption;

  THtmlFileType = ThtDocType;

  THtmlViewerStateBit = (
    vsBGFixed,
    vsDontDraw,
    vsCreating,
    vsProcessing,
    vsLocalImageCache,
    vsHotSpotAction,
    vsMouseScrolling,
    vsLeftButtonDown,
    vsMiddleScrollOn,
    vsHiliting);
  THtmlViewerState = set of THtmlViewerStateBit;

  //TODO -oBG, 16.08.2015: split TfrxHtmlViewer into TfrxHtBase and TfrxHtmlViewer.
  //
  // TfrxHtBase (or a derivate of it) will be the control that TFVBase will create
  // to embed HtmlViewers.
  //
  // The properties of TfrxHtViewerBase will no longer be members to avoid the copy
  // orgies and to simplify implementing more features that are unique to the
  // entire TfrxHtmlViewer or TFrameViewer/TFrameBrowser component. Instead they
  // are gotten by virtual methods.
  //
  // The implementation of these properties will go back to TfrxHtmlViewer and
  // TFrameViewer/TFrameBrowser from where they came.
  //
  // TfrxHtmlViewer implements getters getting the properties from itself.
  // The TfrxHtBase derivate, that TFVBase uses, implements the getters getting the
  // properties from the TFrameViewer/TFrameBrowser component.

  TfrxHtmlViewer = class(TfrxHtmlViewerBase)
  private
    FPaintColor: TColor;
    FPaintBitmap: TBitmap;
    Font: TFont;

    // stuff copied in CreateCopy
    FBase: ThtString;
    FBaseEx: ThtString;
    FBaseTarget: ThtString;
    FOptions: THtmlViewerOptions;

    // events (also copied in CreateCopy)
    FOnLinkDrawn: TLinkDrawnEvent;

    // status info
    FViewerState: THtmlViewerState;

    // document related stuff
    FSectionList: TfrxHtDocument;
    FCurrentFile: ThtString;
    FCurrentFileType: ThtmlFileType;
    FDocument: TBuffer;
    FTitle: ThtString;
    // document related status stuff
    FLinkAttributes: ThtStringList;
    FLinkStart: TPoint;
    FLinkText: UnicodeString;
    FNameList: ThtStringList;
    FScrollWidth: Integer;

    // set in LoadStream and read in doImage to get the image into the document.
    FImageStream: TStream;

    FCSSCache: ThtCSSCache;
    FOnImageCacheChanged: TNotifyEvent;

    function GetDocumentSource: ThtString;
    function GetIDControl(const ID: ThtString): TfrxHtIDObject;
    function GetIDDisplay(const ID: ThtString): ThtDisplayStyle;
    function GetLinkList: TfrxHtLinkList;
    function GetNameList: ThtStringList;
    function GetOurPalette: HPalette;
    function GetViewImages: Boolean;
    procedure BackgroundChange(Sender: TObject);
    procedure DoLogic;
    procedure MatchMediaQuery(Sender: TObject; const MediaQuery: ThtMediaQuery; var MediaMatchesQuery: Boolean);
    procedure Parsed(const Title, Base, BaseTarget: ThtString);
    procedure ParseXHtml;
    procedure ParseHtml;
    procedure ParseText;
    procedure SetBase(Value: ThtString);
    procedure SetIDDisplay(const ID: ThtString; Value: ThtDisplayStyle);
    procedure SetOptions(Value: THtmlViewerOptions);
    procedure SetOurPalette(Value: HPalette);
    procedure SetProcessing(Value: Boolean);
    procedure SetViewerStateBit(Index: THtmlViewerStateBit; Value: Boolean);
    procedure SetViewImages(Value: Boolean);
    function GetDocumentCodePage: Integer;
    procedure SetText(const Value: ThtString);
    procedure LoadFromUrl(const Url: ThtString; DocType: THtmlFileType);
    procedure DoOnImageCacheChanged(Sender: TObject);
//    procedure LoadResource(HInstance: TfrHandle; const ResourceName: ThtString; DocType: THtmlFileType);
  protected
    function IsProcessing: Boolean;
    procedure DoBackground1(ACanvas: TCanvas; ATop, AWidth, AHeight, FullHeight: Integer);
    procedure DoGetImage(Sender: TObject; const SRC: ThtString; var Stream: TStream); virtual;
    procedure HandleMeta(Sender: TObject; const HttpEq, Name, Content: ThtString); virtual;
    procedure HTMLPaint(ACanvas: TCanvas; const ARect: TRect);
    procedure InitLoad; virtual;
    procedure LoadDocument(Document: TBuffer; const DocName: ThtString; DocType: THtmlFileType);
    procedure LoadFile(const FileName: ThtString; ft: ThtmlFileType); virtual;
    procedure LoadString(const Source, Reference: ThtString; ft: ThtmlFileType);
    procedure LoadStream(const URL: ThtString; AStream: TStream; ft: ThtmlFileType);
    procedure SetActiveColor(const Value: TColor); override;
//    procedure SetCharset(const Value: TFontCharset); override;
    procedure SetDefBackground(const Value: TColor); override;
    procedure SetHotSpotColor(const Value: TColor); override;
    procedure SetPreFontName(const Value: TFontName); override;
    procedure SetVisitedColor(const Value: TColor); override;
    property ScrollWidth: Integer read FScrollWidth;
  public
    constructor Create; override;
    constructor CreateCopy(Source: TfrxHtViewerBase); override;
    destructor Destroy; override;
    function Find(const S: UnicodeString; MatchCase: Boolean): Boolean;
    function FindDisplayPos(SourcePos: Integer; Prev: Boolean): Integer;
    function FindEx(const S: UnicodeString; MatchCase, Reverse: Boolean): Boolean;
    function FindSourcePos(DisplayPos: Integer): Integer;
    function FullDisplaySize(FormatWidth: Integer): TSize;
    function PartDisplaySize(FormatWidth, FormatHeight: Integer): TSize;
    function GetCharAtPos(Pos: Integer; var Ch: WideChar; var Font: TFont): Boolean;
    function HtmlExpandFilename(const Filename: ThtString; const CurrentFilename: ThtString = ''): ThtString; override;
    function MakeBitmap(YTop, FormatWidth, Width, Height: Integer): TBitmap;
    function MakeMetaFile(YTop, FormatWidth, Width, Height: Integer): TMetaFile;
    function PositionTo(Dest: ThtString): Boolean;
    procedure Clear; virtual;
    procedure Draw(Canvas: TCanvas; YTop, FormatWidth, Width, Height: Integer);

    procedure htStreamRequest(var Url: ThtString; var Stream: TStream; out PathOfUrl: ThtString);
    procedure htStreamRequested(const Url: ThtString; Stream: TStream);

    procedure Load(const Url: ThtString); override;
    //
    procedure LoadFromDocument(Document: TBuffer; const Reference: ThtString; DocType: THtmlFileType = HtmlType);
    procedure LoadFromFile(const FileName: ThtString; DocType: THtmlFileType = HtmlType);
    procedure LoadFromStream(const AStream: TStream; const Reference: ThtString = ''; DocType: THtmlFileType = HtmlType);
    procedure LoadFromString(const S: ThtString; const Reference: ThtString = ''; DocType: THtmlFileType = HtmlType);
//    procedure LoadFromResource(HInstance: TfrHandle; const ResourceName: ThtString; DocType: THtmlFileType = HtmlType);
    //
    procedure InitCache;
    procedure Reload;
    procedure SetImageCache(ImageCache: TfrxHtImageCache);
    procedure ToggleIDExpand(const URL: ThtString; var Handled: Boolean);
    property Base: ThtString read FBase write SetBase;
    property BaseTarget: ThtString read FBaseTarget;
    property CurrentFile: ThtString read FCurrentFile write FCurrentFile;
    property DocumentCodePage: Integer read GetDocumentCodePage;
    property DocumentTitle: ThtString read FTitle write FTitle;
    property IDControl[const ID: ThtString]: TfrxHtIDObject read GetIDControl;
    property IDDisplay[const ID: ThtString]: ThtDisplayStyle read GetIDDisplay write SetIDDisplay;
    property LinkAttributes: ThtStringList read FLinkAttributes;
    property LinkList: TfrxHtLinkList read GetLinkList;
    property LinkStart: TPoint read FLinkStart;
    property LinkText: UnicodeString read FLinkText write FLinkText;
    property NameList: ThtStringList read GetNameList;
    property OnLinkDrawn: TLinkDrawnEvent read FOnLinkDrawn write FOnLinkDrawn;
    property Palette: HPalette read GetOurPalette write SetOurPalette;
    property SectionList: TfrxHtDocument read FSectionList;
    property ViewerState: THtmlViewerState read FViewerState;
    property CharSet default DEFAULT_CHARSET;
    property HtOptions: THtmlViewerOptions read FOptions write SetOptions default [htPrintTableBackground, htPrintMonochromeBlack];
    property Text: ThtString read GetDocumentSource write SetText;
    property ViewImages: Boolean read GetViewImages write SetViewImages default True;
    property CSSCache: ThtCSSCache read FCSSCache;
    property OnImageCacheChanged: TNotifyEvent read FOnImageCacheChanged write FOnImageCacheChanged;
  end;

function GetFileType(const S: ThtString): THtmlFileType;

var
  FileTypes: ThtStringList;

function GetFileMask : ThtString;

implementation
end.
