
{******************************************}
{                                          }
{             FastReport VCL               }
{                Map Shape                 }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxMapShape;

interface

{$I frx.inc}

uses
{$IFDEF FPC}
  LCLType, LMessages, LazHelper, LCLIntf, LConvEncoding, LazUTF8,
{$ELSE}
  Windows,
{$ENDIF}
  Classes, Contnrs, frCore, frxClass, Graphics, Types,
  frxMapHelpers, frxAnaliticGeometry,
  frxOSMFileFormat, frxGPXFileFormat, frxGeoJSONFileFormat, frxERSIShapeFileFormat;

type
  TShapeStyle = class(TPersistent)
  private
    FBorderColor: TColor;
    FBorderStyle: TPenStyle;
    FBorderWidth: Integer;
    FFillColor: TColor;
    FPointSize: Extended;
  public
    constructor Create;
    procedure TunePen(Pen: TPen);
  published
    property BorderColor: TColor read FBorderColor write FBorderColor;
    property BorderStyle: TPenStyle read FBorderStyle write FBorderStyle;
    property BorderWidth: Integer read FBorderWidth write FBorderWidth;
    property FillColor: TColor read FFillColor write FFillColor;
    property PointSize: Extended read FPointSize write FPointSize; // Diameter
  end;
(******************************************************************************)
  TShape = class(TPersistent)
  private
    FZoom: Extended;
    FCenterOffsetX: Extended;
    FCenterOffsetY: Extended;
    FOffsetX: Extended;
    FOffsetY: Extended;
    FValue: Extended;
    function GetShapeType: TShapeType;
    function GetLegend(const Name: string): string;
    function GetPartCount: integer;
    function GetShapeCenter: TfrxPoint;
    function GetShapeTags: TStringList;
  protected
    FShapeData: TShapeData;

    procedure Clear; virtual;
  public
    procedure Write(Writer: TWriter);
    procedure WriteData(Writer: TWriter);
    procedure Read(Reader: TReader);
    procedure ReadData(Reader: TReader);

    constructor CreateClear;
    destructor Destroy; override;
    function IsValueEmpty: Boolean;
    function IsMercatorSuitable: Boolean;

    property Value: Extended read FValue write FValue;
    property ShapeType: TShapeType read GetShapeType;
    property Legend[const Name: string]: string read GetLegend;
    property PartCount: integer read GetPartCount;
    property ShapeCenter: TfrxPoint read GetShapeCenter;
    property ShapeData: TShapeData read FShapeData;

    property OffsetX: Extended read FOffsetX write FOffsetX;
    property OffsetY: Extended read FOffsetY write FOffsetY;
    property Zoom: Extended read FZoom write FZoom;
  published
    property CenterOffsetX: Extended read FCenterOffsetX write FCenterOffsetX;
    property CenterOffsetY: Extended read FCenterOffsetY write FCenterOffsetY;
    property ShapeTags: TStringList read GetShapeTags;
  end;

  TAdjustableShape = class(TShape)
  published
    property OffsetX;
    property OffsetY;
    property Zoom;
  end;

(******************************************************************************)
  TShapeList = class(TObjectList)
  private
    FEmbeddedData: Boolean;
    FAdjustableShape: Boolean;

    function GetShape(Index: TfrListInteger): TShape;
    procedure SetShape(Index: TfrListInteger; const AShape: TShape);
  protected
    FXMin, FXMax, FYMin, FYMax: Extended;
    FConverter: TMapToCanvasCoordinateConverter;
    FValidMapRect: Boolean;

    procedure SetMapRect(XMin, XMax, YMin, YMax: Extended);
    function Data(iRecord, iPart, iPoint: Integer): TDoublePoint;
  public
    constructor Create(Converter: TMapToCanvasCoordinateConverter);
    function AddShapeData(const AShapeData: TShapeData): integer;
    procedure ReplaceShapeData(iRecord: Integer; const AShapeData: TShapeData);

    procedure ReadDFM(Stream: TStream);
    procedure WriteDFM(Stream: TStream);
    procedure Read(Reader: TReader);
    procedure Write(Writer: TWriter);

    function IsGetValues(var Values: TDoubleArray): boolean;
    procedure ClearValues;

    function IsCanvasPolyPoints(iRecord, iPart: Integer;  MapAccuracy, PixelAccuracy: Extended; var PolyPoints: TPointArray): boolean;
    function CanvasPoint(iRecord: Integer): TfrxPoint;
    function CanvasRect(iRecord: Integer): TDoubleRect;
    function CanvasPoly(iRecord: Integer): TDoublePointArray;
    function CanvasMatrix(iRecord: Integer): TDoublePointMatrix;
    function CanvasWidestPartBounds(iRecord: Integer): TfrxRect;
    procedure GetColumnList(List: TStrings);
    function IsValidMapRect(out MapRect: TfrxRect): boolean;
    procedure SetMapRectByData;
    function IsHasLegend(FieldName: string; Legend: string; out iRecord: Integer): boolean;
    function CanvasDistance(iRecord: Integer; P: TfrxPoint): Extended;
    function IsInside(iRecord: Integer; P: TfrxPoint): Boolean; // Canvas Point
    procedure SaveToTextFile(FileName: string);
    function IsMercatorSuitable: Boolean;

    property Items[Index: TfrListInteger]: TShape read GetShape write SetShape; default;
    property EmbeddedData: Boolean read FEmbeddedData write FEmbeddedData;
    property AdjustableShape: Boolean read FAdjustableShape write FAdjustableShape;
  end;
(******************************************************************************)
  TOSMFileList = class (TStringList)
  public
    constructor Create;
    destructor Destroy; override;
    procedure AddFile(FileName: string);
    procedure AddLink(Link: string; Stream: TStream);
    function FileByName(FileName: string): TOSMFile;
  end;

  TOSMShapeList = class (TShapeList)
  protected
    procedure LoadData(OSMFile: TOSMFile; LayerTags: TStrings);
  public
    class function IsSupportedFile(FileName: string): Boolean;
    constructor CreateFromFile(FileName: string; Converter: TMapToCanvasCoordinateConverter;
      FileList: TOSMFileList; LayerTags: TStrings);
  end;
(******************************************************************************)
  TGPXShapeList = class (TShapeList)
  protected
    procedure LoadData(GPXFile: TGPXFile);
  public
    class function IsSupportedFile(FileName: string): Boolean;
    constructor CreateFromFile(FileName: string; Converter: TMapToCanvasCoordinateConverter);
  end;
(******************************************************************************)
  TERSIShapeList = class (TShapeList)
  protected
    procedure LoadData(ERSIFile: TERSIShapeFile);
  public
    class function IsSupportedFile(FileName: string): Boolean;
    constructor CreateFromFile(FileName: string; Converter: TMapToCanvasCoordinateConverter);
  end;
(******************************************************************************)
  TGeoJSONShapeList = class (TShapeList)
  protected
    procedure LoadData(GeoJSONFile: TGeoJSONFile);
  public
    class function IsSupportedFile(FileName: string): Boolean;
    constructor CreateFromFile(FileName: string; Converter: TMapToCanvasCoordinateConverter);
  end;
(******************************************************************************)

implementation
end.
