
{******************************************}
{                                          }
{             FastReport VCL               }
{            Signature Engine              }
{                                          }
{         Copyright (c) 1998-2023          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxSignatureEngine;

interface

{$I frx.inc}

uses
  Types,
  frxHelpers, frxCertificateList;

type
  TfrxSignatureFormat = (sfCADES_BES, sfCADES_T, sfCADES_X_LONG_TYPE_1);
  TfrxSigningMethod = (simFastReport, simCryptoAPI, simCryptoPRO, simCryptoAPICryptoPRO);

const
  IsTimeStampFormat: array[TfrxSignatureFormat] of Boolean = (False, True, True);
  IsSigningCompatible: array[TfrxSignatureFormat, TfrxSigningMethod] of Boolean = (
    (True,  True,  True, True),  // sfCADES_BES
    (False, True,  True, True),  // sfCADES_T
    (False, False, True, True)); // sfCADES_X_LONG_TYPE_1
  IsDetachedCompatible: array[Boolean {Detached}, TfrxSigningMethod] of Boolean = (
    (False, True, True, True),  // Attached
    ( True, True, True, True)); // Detached
  IsGOSTCompatible: array[Boolean {IsGOST}, TfrxSigningMethod] of Boolean = (
    ( True, True, True, True),  // non GOST
    (False, True, True, True)); // GOST
  IsStoreCompatible: array[TfrxCertificateStoreKind, TfrxSigningMethod] of Boolean = (
    (False, True, True, True),  // csWindowsMy
    ( True, True, True, True)); // csPFXFile

type
  TfrxSignatureData = record
    Format: TfrxSignatureFormat;
    Signing: TfrxSigningMethod;
    TSAURL: string;
    Detached: Boolean;
    Chain: Boolean;
  end;

function frxSignatureData(
  Format: TfrxSignatureFormat;
  Signing: TfrxSigningMethod;
  TSAURL: string;
  Detached: Boolean;
  Chain: Boolean): TfrxSignatureData;
function frxSignatureDataPDF(
  Format: TfrxSignatureFormat;
  Signing: TfrxSigningMethod;
  TSAURL: string): TfrxSignatureData;

function frxSignatureDataDescription(SD: TfrxSignatureData): string;

type
  TfrxSignatureEngine = class
  private
    FStatus: TSignatureSatus;
    FDebugLog: TLogList;
    FSD: TfrxSignatureData;
    FCertificate: TfrxCertificate;
    function GetFormat: TfrxSignatureFormat;
    procedure SetSigning(const Value: TfrxSigningMethod);
  protected
    function GetCertificateCommonName: WideString; virtual; abstract;
    function GetCertificateDistinguishedName: WideString; virtual; abstract;

    function IsError(ErrorCondition: Boolean; ErrorStatus: TSignatureSatus;
      IsUseLastError: Boolean = False): Boolean;

    property SD: TfrxSignatureData read FSD;
    property Certificate: TfrxCertificate read FCertificate;
    property Signing: TfrxSigningMethod write SetSigning;
  public
    constructor Create(ASD: TfrxSignatureData; ACertificate: TfrxCertificate; ADebugLog: TLogList);
    destructor Destroy; override;

    function Sign(MessageData: TByteDynArray; out SignedMessage: TByteDynArray):
      TSignatureSatus; virtual; abstract;

    procedure SignatureHashMD5; virtual; abstract;
    procedure SignatureHashSHA1; virtual; abstract;
    procedure SignatureHashSHA256; virtual; abstract;

    class function SignatureTypeByFormat(Format: TfrxSignatureFormat): WORD;
    class function FormatBySignatureType(SignatureType: WORD): TfrxSignatureFormat;

    property Status: TSignatureSatus read FStatus write FStatus;
    property DebugLog: TLogList read FDebugLog;
    property CertificateCommonName: WideString read GetCertificateCommonName;
    property CertificateDistinguishedName: WideString read GetCertificateDistinguishedName;
    property Format: TfrxSignatureFormat read GetFormat;
  end;

implementation
end.
