{******************************************}
{                                          }
{             FastReport VCL               }
{              Table Object                }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

/// <summary>
///   This unit contains classes of Table report object.
/// </summary>
unit frxTableObject;

interface

{$I frx.inc}

uses
  SysUtils, Types, Classes, Graphics, Variants, Controls,
  frxClass;

const
  /// <summary>
  ///   Default row height for new table.
  /// </summary>
  DefaultRowHeight: Extended = 37.7953 * 0.5;
  /// <summary>
  ///   Default column width for new table.
  /// </summary>
  DefaultColumnWidth: Extended = 37.7953 * 2;
type
  TfrxTableObject = class;
  TfrxCustomTableRow = class;
  TfrxTableColumn = class;
  TfrxHackView = class(TfrxView);
  TfrxHackComponent = class(TfrxComponent);
  TfrxCustomTableBuilder = class;
  TfrxTablePagination = (tpAcrossThenDown, tpDownThenAcross, tpDownThenAcrossWrapped);

  /// <summary>
  ///   The class represents padding data of table container (cell).
  /// </summary>
  TfrxContainerPadding = class(TPersistent)
  private
    FLeftPading: Extended;
    FTopPading: Extended;
    FRightPading: Extended;
    FBottomPading: Extended;
  public
    procedure Assign(Source: TPersistent); override;
  published
    /// <summary>
    ///   Left padding value.
    /// </summary>
    property LeftPading: Extended read FLeftPading write FLeftPading;
    /// <summary>
    ///   Top padding value.
    /// </summary>
    property TopPading: Extended read FTopPading write FTopPading;
    /// <summary>
    ///   Right padding value.
    /// </summary>
    property RightPading: Extended read FRightPading write FRightPading;
    /// <summary>
    ///   Bottom padding value.
    /// </summary>
    property BottomPading: Extended read FBottomPading write FBottomPading;
  end;

{$IFDEF DELPHI16}
/// <summary>
///   The TfrxReportTableObject allows use of the Table component in a report.
///   TfrxReportTableObject is an empty component. It is used to add table unit
///   to uses clause of an application. <br />
/// </summary>
[ComponentPlatformsAttribute(pidWin32 or pidWin64)]
{$ENDIF}
  TfrxReportTableObject = class(TComponent)  // fake component
  end;

  TfrxTablePrintData = (pdNone, pdRow, pdColumn);
  TfrxTableBandType = (tbData, tbHeader, tbFooter, tbAggregateFooter, tbLocalAggregateFooter);

  TfrxRange = record
    RangeStart: Integer;
    RangeEnd: Integer
  end;

  TfrxTableCell = class;

  TfrxCellData = class
  private
    FIsOwnCell: Boolean;
    FReport: TfrxReport;
    FColSpan: Integer;
    FRowSpan: Integer;
    FIndex: Integer;
    FCell: TfrxTableCell;
    FOriginalCell: TfrxTableCell;
    FValue: Variant;
    FParent: TfrxComponent;
    FText: WideString;
    FNeedUpdate: Boolean;
    procedure SetCell(const Value: TfrxTableCell);
    function GetOriginalCell: TfrxTableCell;
    procedure SetText(const Value: WideString);
  public
    property Report: TfrxReport read FReport write FReport;
    property ColSpan: Integer read FColSpan write FColSpan;
    property RowSpan: Integer read FRowSpan write FRowSpan;
    property Index: Integer read FIndex write FIndex;
    property Cell: TfrxTableCell read FCell write SetCell;
    property OriginalCell: TfrxTableCell read GetOriginalCell write FOriginalCell;
    property Value: Variant read FValue write FValue;
    property Parent: TfrxComponent read FParent write FParent;
    property Text: WideString read FText write SetText;
    property NeedUpdate: Boolean read FNeedUpdate write FNeedUpdate;
    destructor Destroy; override;
  end;

//  TfrxCellData = record
//    ColSpan: Integer;
//    RowSpan: Integer;
//    Index: Integer;
//    Cell: TfrxTableCell;
//    Value: Variant;
//    Parent: TfrxComponent;
//    Text: WideString
//  end;
//
//  TfrxPCellData = ^TfrxCellData;

  /// <summary>
  ///   This class represent a table cell. The table call can be used as a
  ///   container for other component or as Text object. It inherits most of
  ///   the properties from "Text" object.
  /// </summary>
  TfrxTableCell = class(TfrxCustomMemoView, IfrxAggregateObject)
  private
    FColSpan: Integer;
    FRowSpan: Integer;
    FHidden: Boolean;
    FContainerPadding: TfrxContainerPadding;
    FSavedObjects: TList;
    FSplitObjects: TList;
    FSavedColSpan: Integer;
    FSavedRowSpan: Integer;
    procedure SetColSpan(Value: Integer);
    procedure SetRowSpan(Value: Integer);
    function CellWidth: Extended;
    function CellHeight: Extended;
    procedure SetContainerPadding(const Value: TfrxContainerPadding);
    function IsPaddingStored: Boolean;
    function GetParentContainer: TfrxComponent;
    function GetDataContainers: TList;
    function GetExpression: String;
    function GetExpressionDelimiters: String;
    function GetInstance: TfrxComponent;
    function GetDataRowContainer: TfrxComponent;
    function GetColSpan: Integer;
    function GetRowSpan: Integer;
  protected
    FIndex: Integer;
    FCellData: TfrxCellData;
    procedure SetParent(AParent: TfrxComponent); override;
    procedure SetWidth(Value: Extended); override;
    procedure SetHeight(Value: Extended); override;
    procedure SetVisible(Value: Boolean); override;
    function GetRestrictions: TfrxRestrictions; override;
    procedure SetIsSelected(const Value: Boolean); override;
    function IsTopStored: Boolean; override;
    function IsLeftStored: Boolean; override;
    function IsWidthStored: Boolean; override;
    function IsHeightStored: Boolean; override;
    procedure DrawSizeBox(aCanvas: TCanvas; aScale: Extended; bForceDraw: Boolean = False); override;
    procedure StartSplit(AReport: TfrxReport);
    procedure FinishSplit(AReport: TfrxReport);
    function SetCellData(CellData: TfrxCellData; bUpdate: Boolean = False): TfrxTableCell;
    function GetText: WideString; override;
    function GetReport: TfrxReport; override;
    function GetValue: Variant; override;
    procedure SetText(const Value: WideString); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure CreateCellTag(Ln, LinesCount, sTextCount, IDX, I: Integer); virtual;
    procedure AfterPrint; override;
    procedure BeforePrint; override;
    function CalcHeight: Extended; override;
    function CalcWidth: Extended; override;
    procedure ResetCellData;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;
    /// <summary>
    ///   Draws cell selection area.
    /// </summary>
    /// <param name="Color">
    ///   Selection color.
    /// </param>
    procedure DrawSelected(Color: TColor = clSkyBlue);
    function IsAcceptControl(aControl: TfrxComponent): Boolean; override;
    function IsOwnerDraw: Boolean; override;
    function Diff(AComponent: TfrxComponent): String; override;
    function DiffText(AComponent: TfrxComponent): String; override;
    procedure DoMouseMove(X, Y: Integer; Shift: TShiftState;
      var EventParams: TfrxInteractiveEventsParams); override;
    function DoMouseDown(X, Y: Integer; Button: TMouseButton;
      Shift: TShiftState; var EventParams: TfrxInteractiveEventsParams)
      : Boolean; override;
    procedure DoMouseUp(X, Y: Integer; Button: TMouseButton; Shift: TShiftState;
      var EventParams: TfrxInteractiveEventsParams); override;
    function DoDragOver(Source: TObject; X, Y: Integer; State: TDragState;
      var Accept: Boolean; var EventParams: TfrxInteractiveEventsParams)
      : Boolean; override;
    procedure GetData; override;
    function GetClientArea: TfrxRect; override;
    /// <summary>
    ///   Returns reference to a parent "Table" object.
    /// </summary>
    function ParentTable: TfrxTableObject;
    /// <summary>
    ///   Returns reference to a parent "Row" object.
    /// </summary>
    function ParentRow: TfrxCustomTableRow;
    function SplitObjects(AHeight: Extended; var IsSplit: Boolean): Extended;
    procedure CleanSplit;
    property Index: Integer read FIndex;
  published
    property AllowExpressions;
    property AllowHTMLTags;
    property BrushStyle;
    property CharSpacing;
    property Clipped;
    property Color;
    property DataField;
    property DataSet;
    property DataSetName;
    property DisplayFormat;
    property ExpressionDelimiters;
    property Font;
    property Frame;
    property FillType;
    property Fill;
    property FlowTo;
    property GapX;
    property GapY;
    property HAlign;
    property HideZeros;
    property Highlight;
    property LineSpacing;
    property Memo;
    property ParagraphGap;
    property ParentFont;
    /// <summary>
    ///   Composite padding object which stores values for padding inside Table
    ///   cell used for inner objects align.
    /// </summary>
    property ContainerPadding: TfrxContainerPadding read FContainerPadding write SetContainerPadding stored IsPaddingStored;
    property Rotation;
    property RTLReading;
    property Style;
    property SuppressRepeated;
    property Underlines;
    property UseDefaultCharset;
    property WordBreak;
    property WordWrap;
    property Wysiwyg;
    property VAlign;
    /// <summary>
    ///   Count of column spans for current cell.
    /// </summary>
    property ColSpan: Integer read GetColSpan write SetColSpan default 1;
    /// <summary>
    ///   Count of row spans for current cell.
    /// </summary>
    property RowSpan: Integer read GetRowSpan write SetRowSpan default 1;
  end;

  /// <summary>
  ///   The base class for Table rows and columns.
  /// </summary>
  TfrxTableRowColumnBase = class(TfrxComponent)
  private
    FAutoSize: Boolean;
    FPageBreak: Boolean;
    FBandType: TfrxTableBandType;
    procedure SetAutoSize(Value: Boolean);
  protected
    FIndex: Integer;
    FInternalObjects: TList;
    procedure SetParent(AParent: TfrxComponent); override;
    procedure SetWidth(Value: Extended); override;
    procedure SetHeight(Value: Extended); override;
    procedure SetIsSelected(const Value: Boolean); override;
    function IsTopStored: Boolean; override;
    function IsLeftStored: Boolean; override;
    property InternalObjects: TList read FInternalObjects;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    /// <summary>
    ///   Returns reference to a parent "Table" object.
    /// </summary>
    function ParentTable: TfrxTableObject;
    procedure CreateUniqueNames;
    procedure DoMouseMove(X, Y: Integer; Shift: TShiftState;
      var EventParams: TfrxInteractiveEventsParams); override;
    function DoMouseDown(X, Y: Integer; Button: TMouseButton;
      Shift: TShiftState; var EventParams: TfrxInteractiveEventsParams)
      : Boolean; override;
    procedure DoMouseUp(X, Y: Integer; Button: TMouseButton; Shift: TShiftState;
      var EventParams: TfrxInteractiveEventsParams); override;
    /// <summary>
    ///   Index of current row or column.
    /// </summary>
    property Index: Integer read FIndex;
    property BandType: TfrxTableBandType read FBandType;
  published
    /// <summary>
    ///   When this property is set, Table object automatically calculates
    ///   width of column or height of row based on content inside cells.
    /// </summary>
    property AutoSize: Boolean read FAutoSize write SetAutoSize default False;
  end;

  /// <summary>
  ///   Represents Table column.
  /// </summary>
  TfrxTableColumn = class(TfrxTableRowColumnBase)
  private
    FMinWidth: Extended;
    FMaxWidth: Extended;
  protected
    function IsHeightStored: Boolean; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DoMouseUp(X, Y: Integer; Button: TMouseButton; Shift: TShiftState;
      var EventParams: TfrxInteractiveEventsParams); override;
  published
    function IsContain(X, Y: Extended): Boolean; override;
    property Width;
    /// <summary>
    ///   Minimal width of current column. Column width can not be less than
    ///   this value when AutoSize is turned on.
    /// </summary>
    property MinWidth: Extended read FMinWidth write FMinWidth;
    /// <summary>
    ///   Maximum width of current column. Column width can not be grater than
    ///   this value when AutoSize is turned on.
    /// </summary>
    property MaxWidth: Extended read FMaxWidth write FMaxWidth;
  end;

  /// <summary>
  ///   Represents Table Row.
  /// </summary>
  TfrxCustomTableRow = class(TfrxTableRowColumnBase)
  private
    FMinHeight: Extended;
    FMaxHeight: Extended;
    FAllowSplit: Boolean;
    //FCellsData: array of TfrxCellData;
    //FCellsData: TList;
    function GetCellCount: Integer;

  protected
    FIsSplit: Boolean;
    function GetCell(Index: Integer): TfrxTableCell; virtual;
    procedure InitCells(Value: Integer); virtual;
    procedure CorrectCellsOnColumnChange(Index, Correct: Integer); virtual;
    procedure SetLeft(Value: Extended); override;
    function IsWidthStored: Boolean; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function CreateTableCell(Index: Integer; bSetUniqueName: Boolean = False): TObject; virtual;
    procedure AfterPrint;
    function IsContain(X, Y: Extended): Boolean; override;
    procedure DoMouseUp(X, Y: Integer; Button: TMouseButton; Shift: TShiftState;
      var EventParams: TfrxInteractiveEventsParams); override;
    function GetSplitPart(AHeight: Extended; var IsLastPart: Boolean): TfrxCustomTableRow; virtual;
    function IsSplitPart: Boolean;
    /// <summary>
    ///   Returns count of cell in current row.
    /// </summary>
    property CellCount: Integer read GetCellCount;
    /// <summary>
    ///   List of the cell in current row.
    /// </summary>
    /// <param name="Index">
    ///   Column index.
    /// </param>
    property Cells[Index: Integer]: TfrxTableCell read GetCell;
  published
    property AllowSplit: Boolean read FAllowSplit write FAllowSplit default False;
    /// <summary>
    ///   Minimal height of current row. Row height can not be less than this
    ///   value when AutoSize is turned on.
    /// </summary>
    property MinHeight: Extended read FMinHeight write FMinHeight;
    /// <summary>
    ///   Maximum height of current row. Row height can not be greater than
    ///   this value when AutoSize is turned on.
    /// </summary>
    property MaxHeight: Extended read FMaxHeight write FMaxHeight;
    property Height;
  end;

  /// <summary>
  ///   The TfrxTableObject component represents a Table object. The table
  ///   object used to group report data or object in one big tabular
  ///   structure. With table object it's more easy to change appearance of the
  ///   tabular based reports. Table may grow in size based on content inside
  ///   cells. The table cells can contain other objects. The table size
  ///   controls by rows and column count.
  /// </summary>
  TfrxTableObject = class(TfrxStretcheable)
  private
    FRowCount: Integer;
    FColumnCount: Integer;
    FLockCorrectSpans: Boolean;
    FLockObjectsUpdate: Boolean;
    FSpanList: PfrxRectArray;
    FDragDropActive: Boolean;
    FColumnSelection: Boolean;
    FRowSelection: Boolean;
    FSelectionStart: TRect;
    FSelectionEnd: TRect;
    FCellsSelection: TRect;
    FModified: Boolean;
    FNewColumnDim: Integer;
    FNewRowDim: Integer;
    FResizeActive: Boolean;
    FBoundsUpdating: Boolean;
    FSavedTable: TfrxTableObject;
    FTableActive: Boolean;

    FSelectorPoint: TPoint;
    FSelectionFill: TfrxCustomFill;
    FSelectedGridCol: Integer;
    FSelectedGridRow: Integer;
    FCopyAppearance: Boolean;

    FVertSplitter: Integer;
    FHorzSplitter: Integer;
    FBreakRowIndex: Integer;
    FTableBuilder: TfrxCustomTableBuilder;
    FIsManualBuild: Boolean;
    FOnManualBuild: TfrxNotifyEvent;

    FLastMousePos: TPoint;
    FBrakeTo: TfrxTableObject;
    FBrakeToRows: TList;
    FBrakeToRowHeight: Extended;
    FDefaultCellHeight: Extended;
    FDefaultCellWidth: Extended;

    function GetColumn(Index: Integer): TfrxTableColumn;
    function GetRow(Index: Integer): TfrxCustomTableRow;
    function GetCell(X, Y: Integer): TfrxTableCell;
    procedure SetColumnCount(Value: Integer);
    procedure SetRowCount(Value: Integer);
    procedure CorrectSpansOnColumnChange(ColumnIndex, Correct: Integer);
    procedure CorrectSpansOnRowChange(RowIndex, Correct: Integer);
    procedure ResetSpanList;
    procedure DrawTable(Highlighted: Boolean = False);
    procedure CalcWidthInternal;
    procedure CalcHeightInternal;
    procedure NormalizeSpans;
    procedure AssignCellAppearance(FromCell, ToCell: TfrxTableCell);
    procedure UpdateDesigner;
    procedure UpdateCellDimensions(cell: TfrxTableCell);
    function GetTableHeight: Extended;
    function GetTableWidth: Extended;
    procedure SetTableHeight(const Value: Extended);
    procedure SetTableWidth(const Value: Extended);
    function GetTableBuilder: TfrxCustomTableBuilder;
  protected
    FSelectedRowCol: TfrxTableRowColumnBase;
    FSelectedRowColCount: Integer;
    procedure NormalizeObjectsList;
    procedure FillSpanList(var SpanList: TfrxRectArray);
    function IsInsideSpan(SpanList: TfrxRectArray; p: TPoint): Boolean;
    function IsTableActive: Boolean;
    function CheckColumnSelector(X, Y: Extended): Boolean;
    function CheckRowSelector(X, Y: Extended): Boolean;
    function CheckMoveArrow(X, Y: Extended): Boolean;
    function CheckSizeArrow(X, Y: Extended): Boolean;
    procedure ObjectListNotify(Ptr: Pointer; Action: TListNotification); override;
    procedure SetWidth(Value: Extended); override;
    procedure SetHeight(Value: Extended); override;
    function GetRowColumnByClass(aClass: TClass; Index: Integer): TfrxTableRowColumnBase;
    function IsWidthStored: Boolean; override;
    function IsHeightStored: Boolean; override;
    function CreateTableColumn: TfrxTableColumn; virtual;
    function CreateUniqueTableColumn: TfrxTableColumn; virtual;
    function CreateUniqueTableRow: TfrxCustomTableRow; virtual;
    procedure Loaded; override;
    procedure DoMirror(MirrorModes: TfrxMirrorControlModes); override;
    procedure MoveRowsColumns(DestTable: TfrxTableObject; bClearSource: Boolean = False; RowStartIndex: Integer = 0; RowEndIndex: Integer = -1; ColStartIndex: Integer = 0; ColEndIndex: Integer = -1);
    procedure RestoreBreakTable;
    procedure RestoreSourceTable;
    procedure RestoreAfterBuilder;
    procedure AddBrakeToRow(ARow: TfrxCustomTableRow);
    procedure ClearBrakeToRows;
    function GetSplitRow(ARow: TfrxCustomTableRow; AHeight: Extended; var IsLastPart: Boolean): TfrxCustomTableRow; virtual;
  public
    FSerializeColumns: TfrxRange;
    FSerializeRows: TfrxRange;
    constructor Create(AOwner: TComponent); override;
    constructor DesignCreate(AOwner: TComponent; Flags: Word); override;
    class function CreateTableRow: TfrxCustomTableRow; virtual;
    destructor Destroy; override;
    procedure AlignChildren(IgnoreInvisible: Boolean = False; MirrorModes: TfrxMirrorControlModes = []); override;
    procedure Clear; override;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;
    function DrawPart: Extended; override;
    procedure InitPart; override;
    function HasNextDataPart(aFreeSpace: Extended): Boolean; override;
    procedure DoMouseEnter(aPreviousObject: TfrxComponent;
      var EventParams: TfrxInteractiveEventsParams); override;
    procedure DoMouseLeave(aPreviousObject: TfrxComponent;
      var EventParams: TfrxInteractiveEventsParams); override;
    function DoMouseDown(X, Y: Integer; Button: TMouseButton;
      Shift: TShiftState; var EventParams: TfrxInteractiveEventsParams)
      : Boolean; override;
    procedure DoMouseUp(X, Y: Integer; Button: TMouseButton; Shift: TShiftState;
      var EventParams: TfrxInteractiveEventsParams); override;
    procedure DoMouseMove(X, Y: Integer; Shift: TShiftState;
      var EventParams: TfrxInteractiveEventsParams); override;
    function IsContain(X, Y: Extended): Boolean; override;
    function IsAcceptAsChild(aParent: TfrxComponent): Boolean; override;
    function GetContainedComponent(X, Y: Extended; IsCanContain: TfrxComponent = nil): TfrxComponent; override;
//    function ContainerAdd(Obj: TfrxComponent): Boolean; override;
    procedure BeforePrint; override;
    procedure AfterPrint; override;
    function CalcHeight: Extended; override;
    { script helpers }
    /// <summary>
    ///   Helper method used to create new column and insert it by given index
    ///   into current table object.
    /// </summary>
    /// <param name="Index">
    ///   Column index.
    /// </param>
    function CreateNewColumn(Index: Integer): TfrxTableColumn;
    /// <summary>
    ///   Helper method used to create new row and insert it by given index
    ///   into current table object.
    /// </summary>
    /// <param name="Index">
    ///   Row index.
    /// </param>
    function CreateNewRow(Index: Integer): TfrxCustomTableRow;

    procedure GetData; override;
    class function GetDescription: String; override;
    function ExportInternal(Filter: TfrxCustomExportFilter): Boolean; override;
    procedure LockObjectsUpdate;
    procedure UnlockObjectsUpdate;
    function GetSaveToComponent: TfrxReportComponent; override;
    /// <summary>
    ///   Adds column by it reference to the end of the table.
    /// </summary>
    /// <param name="Value">
    ///   Reference to a column object.
    /// </param>
    procedure AddColumn(Value: TfrxTableColumn);
    /// <summary>
    ///   Inserts column by it reference to given position.
    /// </summary>
    /// <param name="Index">
    ///   Column index.
    /// </param>
    /// <param name="Value">
    ///   Reference to a column object. <br />
    /// </param>
    procedure InsertColumn(Index: Integer; Value: TfrxTableColumn);
    /// <summary>
    ///   Moves culumn of the table from one position to another.
    /// </summary>
    /// <param name="Index">
    ///   Index of column to move.
    /// </param>
    /// <param name="NewIndex">
    ///   New position of the column.
    /// </param>
    procedure MoveColumn(Index, NewIndex: Integer);
    /// <summary>
    ///   Deletes column from the table.
    /// </summary>
    /// <param name="Index">
    ///   Column index.
    /// </param>
    procedure DeleteColumn(Index: Integer);
    /// <summary>
    ///   Adds row by it reference to the end of the table.
    /// </summary>
    /// <param name="Value">
    ///   Reference to a row object. <br />
    /// </param>
    procedure AddRow(Value: TfrxCustomTableRow);
    /// <summary>
    ///   Inserts row by it reference to given position.
    /// </summary>
    /// <param name="Index">
    ///   Index of the row.
    /// </param>
    /// <param name="Value">
    ///   Reference to a row object.
    /// </param>
    procedure InsertRow(Index: Integer; Value: TfrxCustomTableRow);
    /// <summary>
    ///   Swaps positions of two rows in the table.
    /// </summary>
    /// <param name="Row1">
    ///   Index of first row.
    /// </param>
    /// <param name="Row2">
    ///   index of second row.
    /// </param>
    procedure SwapRows(Row1, Row2: Integer);
    /// <summary>
    ///   Deletes row from the table.
    /// </summary>
    /// <param name="Index">
    ///   Row index to delete.
    /// </param>
    procedure DeleteRow(Index: Integer);
    procedure UpdateBounds; override;
    function IsDynamicTable: Boolean;
    /// <summary>
    ///   Joins selected cells to one big by setting ColSpan and RowSpan
    ///   properties.
    /// </summary>
    /// <param name="TopX">
    ///   Top column position of selection.
    /// </param>
    /// <param name="TopY">
    ///   Top row position of selection. <br />
    /// </param>
    /// <param name="BottomX">
    ///   Bottom column position of selection. <br />
    /// </param>
    /// <param name="BottomY">
    ///   Bottom row position of selection.
    /// </param>
    procedure JoinSelection(TopX: Integer = MaxInt; TopY: Integer = MaxInt; BottomX: Integer = -1; BottomY: Integer = -1);
    /// <summary>
    ///   Splits selected cells by reset ColSpan and RowSpan properties.
    /// </summary>
    procedure SplitSelected;
    /// <summary>
    ///   Splits cell by reset ColSpan and RowSpan properties.
    /// </summary>
    /// <param name="aCell">
    ///   Reference to a cell object to split.
    /// </param>
    procedure SplitCell(aCell: TfrxTableCell);
    procedure SetSpanFillList(const SpanList: PfrxRectArray);
    /// <summary>
    ///   List of table columns.
    /// </summary>
    /// <param name="Index">
    ///   Column index.
    /// </param>
    property Columns[Index: Integer]: TfrxTableColumn read GetColumn;
    /// <summary>
    ///   List of table rows.
    /// </summary>
    /// <param name="Index">
    ///   Row index.
    /// </param>
    property Rows[Index: Integer]: TfrxCustomTableRow read GetRow;
    /// <summary>
    ///   Matrix of cell objects.
    /// </summary>
    /// <param name="X">
    ///   Column index.
    /// </param>
    /// <param name="Y">
    ///   Row index.
    /// </param>
    property Cells[X, Y: Integer]: TfrxTableCell read GetCell;
    /// <summary>
    ///   Default cell height used during creation of new row or column.
    /// </summary>
    property DefaultCellHeight: Extended read FDefaultCellHeight write FDefaultCellHeight;
    /// <summary>
    ///   Default cell width used during creation of new row or column.
    /// </summary>
    property DefaultCellWidth: Extended read FDefaultCellWidth write FDefaultCellWidth;
    /// <summary>
    ///   Height of the table object.
    /// </summary>
    property TableHeight: Extended read GetTableHeight write SetTableHeight;
    /// <summary>
    ///   Width of the table object
    /// </summary>
    property TableWidth: Extended read GetTableWidth write SetTableWidth;
    property TableBuilder: TfrxCustomTableBuilder read GetTableBuilder;
  published
    /// <summary>
    ///   Count of columns in the table object.
    /// </summary>
    property ColumnCount: Integer read FColumnCount write SetColumnCount stored False;
    /// <summary>
    ///   Count of rows in the table object.
    /// </summary>
    property RowCount: Integer read FRowCount write SetRowCount stored False;
    property OnManualBuild: TfrxNotifyEvent read FOnManualBuild write FOnManualBuild;
  end;

  TfrxStateTableObject = class(TfrxTableObject)
  private
    FReport: TfrxReport;
    FOriginalParent: TfrxComponent;
  protected
    function GetReport: TfrxReport; override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure AssignOriginals(Source: TfrxComponent); override;
  end;

  TfrxTableRow = class(TfrxCustomTableRow);

  TfrxTableObjectClass = class of TfrxTableObject;

  TfrxCustomTableBuilder = class(TPersistent)
  protected
    FResultTable: TfrxTableObject;
    FSourceTable: TfrxTableObject;
    FOutTable: TfrxTableObject;
    FIsFirstPart: Boolean;
    FLeft: Extended;
    FTop: Extended;
    FFixedColumns: TList;
    FFixedFooterColumns: TList;
    FFixedFooterRows: TList;
    FFixedRows: TList;
    FReport: TfrxReport;
    FDataPrinting: TfrxTablePrintData;
    FIsRowPriority: Boolean;
    FPageBreak: Boolean;
    FAutoSpans: Boolean;
    FColumnSpans: TList;
    FRowSpans: TList;
    FPaginationOrder: TfrxTablePagination;
    FPartHeight: Extended;
    FOriginalRowIndex: Integer;
    FOriginalColumnIndex: Integer;
    FPrintingRowIndex: Integer;
    FPrintingColumnIndex: Integer;
    FPaginateStart: TPoint;
    FPaginateEnd: TPoint;
    FFirstPartY: Extended;
    FIsLastPart: Boolean;
    FCurrentPrintPart: TfrxTablePagination;
    FIsNewPage: Boolean;
    FTempList: TList;
    FLastUpdateValue: Cardinal;
    FResultSpans: TfrxRectArray;
    FBreakeSpanCells: TList;
    FResetBreakeSpan: Boolean;
    FCurFixedColumns: TList;
    FSavedTerminated: Boolean;
    class function GetResultTableClass: TfrxTableObjectClass; virtual;
    procedure CopyCells(OriginalColumnIndex, OriginalRowIndex,
      ResultColumnIndex, ResultRowIndex: Integer); virtual; abstract;
    procedure FillUpdate;
    procedure RestoreSpanBreak;
  protected
    procedure AddFixedColumn(FixedColumn: TfrxTableColumn; DestTable: TfrxTableObject); virtual;
    procedure AddFooterColumn(ColumnFooter: TfrxTableColumn; DestTable: TfrxTableObject); virtual;
    procedure AddFooterRow(RowFooter: TfrxCustomTableRow; DestTable: TfrxTableObject); virtual;
    procedure AddFixedRow(FixedRow: TfrxCustomTableRow; DestTable: TfrxTableObject); virtual;
    procedure ClearResultObjects(ATable: TfrxTableObject); virtual;
    function GetCellData(X, Y: Integer): TfrxCellData; virtual;
  public
    constructor Create(SourceTable: TfrxTableObject); virtual;
    destructor Destroy; override;
    procedure PrintRow(Index: Integer; TableBandType: TfrxTableBandType = tbData);
    procedure PrintRows(StartIndex: Integer = 0);
    procedure PrintColumn(Index: Integer; TableBandType: TfrxTableBandType = tbData);
    procedure PrintColumns(StartIndex: Integer = 0);
    procedure PageBreak;
    procedure PrepareFirstPart(PrepareToTable: TfrxTableObject);
    function PaginateNextPartTo(ATable: TfrxTableObject; Report: TfrxReport; var UsedHeight: Extended): Boolean;
    procedure DoPaginateTo(ATable: TfrxTableObject; Report: TfrxReport);
    function ColumnCount: Integer;
    function RowCount: Integer;
    property PrintingRowIndex: Integer read FPrintingRowIndex;
    property PrintingColumnIndex: Integer read FPrintingColumnIndex;
    property CellData[X, Y: Integer]: TfrxCellData read GetCellData;
    property ResultTable: TfrxTableObject read FResultTable;
    property OutTable: TfrxTableObject read FOutTable;// hide from public
    property PaginationOrder: TfrxTablePagination read FPaginationOrder write FPaginationOrder;
    property PartHeight: Extended read FPartHeight;
  end;

  TfrxTableBuilderClass = class of TfrxCustomTableBuilder;

  TfrxTableBuilder = class(TfrxCustomTableBuilder)
  protected
    procedure CopyCells(OriginalColumnIndex, OriginalRowIndex,
      ResultColumnIndex, ResultRowIndex: Integer); override;
  end;

var
  frxTableBuilder: TfrxTableBuilderClass;

implementation
end.
