
{******************************************}
{                                          }
{             FastReport VCL               }
{          HTTP connection client          }
{                                          }
{         Copyright (c) 1998-2021          }
{            Fast Reports Inc.             }
{                                          }
{******************************************}

unit frxTransportHTTP;

{$I frx.inc}
interface

uses
  Windows, Messages, SysUtils, Classes, ScktComp, frxServerUtils, frxNetUtils, Math,
  frGzip, frHashUtils, WinSock, frxBaseTransportConnection, frxBaseSocketIOHandler;

const
  HTTP_POST = 'POST';
  HTTP_GET = 'GET';
  HTTP_DELETE = 'DELETE';
  HTTP_MKCOL = 'MKCOL';
  HTTP_PUT = 'PUT';
  HTTP_PROPFIND = 'PROPFIND';
  HTTP_VER1 = 'HTTP/1.0';
  HTTP_LINK_PREFIX: AnsiString = 'http://';
  HTTPS_LINK_PREFIX: AnsiString = 'https://';

type
  TfrxRequestType = (xrtPost, xrtGet, xrtDelete, xrtMkcol, xrtPut, xrtPropFind);
  TfrxHTTPContentType = (htcNone, hctDefaultHTML, htcDefaultXML, htcDefaultApp);

type
  TfrxHTTPRequest = class (TPersistent)
  private
    FURL: AnsiString;
    FContentType: AnsiString;
    FAuthorization: AnsiString;
    FReqType: TfrxRequestType;
    FSourceStream: TStream;
    FAcceptTypes: AnsiString;
    FDefAcceptTypes: TfrxHTTPContentType;
    FEncoding: AnsiString;
    FUserAgent: AnsiString;
    FPort: Integer;
    function GetText: AnsiString;
    procedure SetText(const Value: AnsiString);
  protected
    FRequest: TStringList;
    FCustomHeader: TStrings;
    procedure SureEmptyLineAtEnd;
  public
    constructor Create;
    destructor Destroy; override;
    procedure BuildRequest; virtual;
    procedure Redirect(const NewAddress: AnsiString);
    function IsValidAddress: Boolean;
    function Host(bTruncPort: Boolean = True): AnsiString;
    function GetPort: AnsiString;
    property Authorization: AnsiString read FAuthorization write FAuthorization;
    property ReqType: TfrxRequestType read FReqType write FReqType;
    property Port: Integer read FPort write FPort;
    property SourceStream: TStream read FSourceStream write FSourceStream;
    property AcceptTypes: AnsiString read FAcceptTypes write FAcceptTypes;
    property ContentType: AnsiString read FContentType write FContentType;
    property DefAcceptTypes: TfrxHTTPContentType read FDefAcceptTypes write FDefAcceptTypes;
    property CustomHeader: TStrings read FCustomHeader;
    property Encoding: AnsiString read FEncoding write FEncoding;
    property UserAgent: AnsiString read FUserAgent write FUserAgent;

    property URL: AnsiString read FURL write FURL;
    property Text: AnsiString read GetText write SetText;
  end;

  TfrxHTTPServerFields = class;

{$IFDEF DELPHI16}
[ComponentPlatformsAttribute(pidWin32 or pidWin64)]
{$ENDIF}
  TfrxTransportHTTP = class(TfrxBaseTransportConnection)
  private
    FActive: Boolean;
    FConnected: Boolean;
    FRequestString: AnsiString;
    FAnswer: TStrings;
    FBreaked: Boolean;
    FErrors: TStrings;
    FMIC: Boolean;
    FProxyHost: String;
    FProxyPort: Integer;
    FRetryCount: Integer;
    FRetryTimeOut: Integer;
    FServerFields: TfrxHTTPServerFields;
    FRequestStream: TMemoryStream;
    FRawDataStream: TMemoryStream;
    FTimeOut: Integer;
    FProxyLogin: String;
    FProxyPassword: String;
    FPort: Integer;
    FConnectDelay: Cardinal;
    FAnswerDelay: Cardinal;
    FIOHandler: TfrxCustomIOHandler;
    FHandleOnlySocketMessages: Boolean;
//    FSocketType: TfrxSocketType;
    procedure ParseAnswer(aToDataStream: TStream);
//    procedure DoConnect(Sender: TObject; Socket: TCustomWinSocket);
//    procedure DoLookup(Sender: TObject; Socket: TCustomWinSocket);
    procedure DoDisconnect(Sender: TObject; Socket: TCustomWinSocket);
    procedure DoError(Sender: TObject; Socket: TCustomWinSocket;
      ErrorEvent: TErrorEvent; var ErrorCode: Integer);

    procedure SetActive(const Value: Boolean);
    procedure SetServerFields(const Value: TfrxHTTPServerFields);
    procedure SetIOHandler(const Value: TfrxCustomIOHandler);
    function PutBase(const aURL: AnsiString; const aSource: TStream; const frxRequestType: TfrxRequestType): AnsiString;
    function DeleteBase(const aURL: AnsiString; const frxRequestType: TfrxRequestType): AnsiString;
  protected
    FHTTPRequest: TfrxHTTPRequest;
    FClientSocket: TClientSocket;
    StreamSize: Cardinal;
    procedure SetSocketDestination;
    function IsAnswerCodeIn(Answers: array of Integer): boolean;
    function GetProxyHost: String; override;
    function GetProxyLogin: String; override;
    function GetProxyPassword: String; override;
    function GetProxyPort: Integer; override;
    procedure SetProxyHost(const Value: String); override;
    procedure SetProxyLogin(const Value: String); override;
    procedure SetProxyPassword(const Value: String); override;
    procedure SetProxyPort(const Value: Integer); override;
    function GetSocketType: TfrxSocketType; override;
    procedure SetSocketType(const Value: TfrxSocketType); override;
    procedure DoMessages;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Connect; override;
    procedure StartListening;
    procedure Disconnect; override;
    procedure Open;
    procedure Close;
    procedure DoWrite(Sender: TObject; Socket: TCustomWinSocket);
    procedure DoRead(Sender: TObject; Socket: TCustomWinSocket);
    procedure Send(aSource: TStream); // sync method does not return untill finish
    procedure SetDefaultParametersWithToken(AToken: String); override;
    procedure Receive(aSource: TStream); // sync method does not return untill finish
    function Post(aURL: AnsiString; aSource: TStream = nil): AnsiString; overload; // sync method does not return untill finish
    function Get(aURL: AnsiString; aSource: TStream = nil): AnsiString; overload; // sync method does not return untill finish
    function PropFind(const AURL: AnsiString; const ASource: TStream = nil): AnsiString; overload; // sync method does not return untill finish
    function Put(const AURL: AnsiString; const ASource: TStream = nil): AnsiString; overload; // sync method does not return untill finish
    function Post(aURL: WideString; aSource: TStream = nil): WideString; overload; // sync method does not return untill finish
    function Get(aURL: WideString; aSource: TStream = nil): WideString; overload; // sync method does not return untill finish
    function PropFind(const AURL: WideString; const ASource: TStream = nil): WideString; overload; // sync method does not return untill finish
    function Put(const AURL: WideString; const ASource: TStream = nil): WideString; overload; // sync method does not return untill finish
    function Delete(const AURL: AnsiString): AnsiString; overload; // sync method does not return untill finish
    function Mkcol(const AURL: AnsiString): AnsiString; overload; // sync method does not return untill finish
    function Delete(const AURL: WideString): WideString; overload; // sync method does not return untill finish
    function Mkcol(const AURL: WideString): WideString; overload; // sync method does not return untill finish
    property Answer: TStrings read FAnswer write FAnswer;
    property Errors: TStrings read FErrors write Ferrors;
    property RequestStream: TMemoryStream read FRequestStream;
    property IOHandler: TfrxCustomIOHandler read FIOHandler write SetIOHandler;
    property HandleOnlySocketMessages: Boolean read FHandleOnlySocketMessages write FHandleOnlySocketMessages;
  published
    property Active: Boolean read FActive write SetActive;
    property RequestString: AnsiString read FRequestString write FRequestString;
    property HTTPRequest: TfrxHTTPRequest read FHTTPRequest;
    property ClientSocket: TClientSocket read FClientSocket;
    property MIC: Boolean read FMIC write FMIC;
    property Port: Integer read FPort write FPort;
    property ProxyHost;
    property ProxyPort;
    property ProxyLogin;
    property ProxyPassword;
    property RetryCount: Integer read FRetryCount write FRetryCount;
    property RetryTimeOut: Integer read FRetryTimeOut write FRetryTimeOut;
    property ServerFields: TfrxHTTPServerFields read FServerFields write SetServerFields;
    property TimeOut: Integer read FTimeOut write FTimeOut;
    property ConnectDelay: Cardinal read FConnectDelay;
    property AnswerDelay: Cardinal read FAnswerDelay;
    property OnWork;
    property OnWorkBegin;
    property OnWorkEnd;
  end;

  TfrxHTTPServerFields = class (TPersistent)
  private
    FAnswerCode: Integer;
    FContentEncoding: String;
    FContentMD5: String;
    FContentLength: Integer;
    FLocation: String;
    FSessionId: String;
    FCookie: String;
  public
    constructor Create;
    procedure Assign(Source: TPersistent); override;
    procedure ParseAnswer(st: AnsiString);
  published
    property AnswerCode: Integer read FAnswerCode write FAnswerCode;
    property ContentEncoding: String read FContentEncoding write FContentEncoding;
    property ContentMD5: String read FContentMD5 write FContentMD5;
    property ContentLength: Integer read FContentLength write FContentLength;
    property Location: String read FLocation write FLocation;
    property SessionId: String read FSessionId write FSessionId;
    property Cookie: String read FCookie write FCookie;
  end;

  procedure frxTestToken(const URL: String; var sToken: String; bUsePOST: Boolean = False; PrefixAut: String = 'Bearer ');

  var
    msc: TFPUExceptionMask;

implementation
end.
