{******************************************}
{                                          }
{             FastReport VCL               }
{            Various routines              }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

/// <summary>
///   The unit contains additional helper function used by other FastReports
///   units like type conversions, CRC calculation, image draw and others.
/// </summary>
unit frxUtils;

interface

{$I frx.inc}

uses
{$IFNDEF FPC}
  Windows, Messages,
{$ELSE}
  LCLType, LCLIntf, LazHelper, DateUtils,
  {$IFDEF LCLGTK2} Printers, {$ENDIF}
{$ENDIF}
  SysUtils, {$IFDEF DELPHI16} System.UITypes, {$ENDIF} Classes, Graphics,
  Controls, Forms, StdCtrls, Menus, ImgList, ActnList, ComCtrls,
  frXML, frCore, frGraphics, frCoreClasses, frGraphicUtils,
  frxClass, frxVariables;

type
  TfrxRectArea = class
  public
    X, Y, X1, Y1: Extended;
    constructor Create(c: TfrxComponent); overload;
    constructor Create(Left, Top, Right, Bottom: Extended); overload;
    function InterceptsX(a: TfrxRectArea): Boolean;
    function InterceptsY(a: TfrxRectArea): Boolean;
    function InterceptX(a: TfrxRectArea): TfrxRectArea;
    function InterceptY(a: TfrxRectArea): TfrxRectArea;
    function Max(x1, x2: Extended): Extended;
    function Min(x1, x2: Extended): Extended;
  end;

procedure frxParseDilimitedText(aList: TStrings; aText: String; aDelimiter: Char);
/// <summary>
///   Searches for a component on owner by name.
/// </summary>
/// <param name="Owner">
///   Owner component.
/// </param>
/// <param name="Name">
///   Name of the searched component.
/// </param>
function frxFindComponent(Owner: TComponent; const Name: String): TComponent;
/// <summary>
///   Fills list with all found component in application with class type
///   specified in ClassRef parameter. <br />
/// </summary>
/// <param name="Owner">
///   Components owner.
/// </param>
/// <param name="ClassRef">
///   Class reference to search for.
/// </param>
/// <param name="List">
///   Out list parameter.
/// </param>
/// <param name="Skip">
///   Component to skip.
/// </param>
procedure frxGetComponents(Owner: TComponent; ClassRef: TClass;
  List: TStrings; Skip: TComponent);
/// <summary>
///   Returns full name of the component includes full sequences of owners
///   delimired by '.' sign.
/// </summary>
/// <param name="Owner">
///   Owner component.
/// </param>
/// <param name="c">
///   Component reference.
/// </param>
function frxGetFullName(Owner: TComponent; c: TComponent): String;
function frxRemoveQuotes(const s: String): String;
/// <summary>
///   Converts stream bytes to hex string representation.
/// </summary>
/// <param name="Stream">
///   Reference to stream.
/// </param>
/// <returns>
///   Hex string.
/// </returns>
function frxStreamToString(Stream: TStream): String;
/// <summary>
///   Converts hex string representation to stream bytes and writes them to
///   stream.
/// </summary>
/// <param name="s">
///   String with hex representation of bytes.
/// </param>
/// <param name="Stream">
///   Destination stream.
/// </param>
procedure frxStringToStream(const S: string; AStream: TStream);
function frxStringToHex(const Buffer: WideString): String;
function frxHexToString(const Buffer: String): WideString;
function frxDecToBinBytes(d: Longint): string;
function frxStrToFloat(s: String): Extended;

function frxRect(ALeft, ATop, ARight, ABottom: Extended): TfrxRect;
function frxPoint(X, Y: Extended): TfrxPoint;
function frxGetBrackedVariable(const Str, OpenBracket, CloseBracket: AnsiString;
  var i, j: Integer): AnsiString;
function frxGetBrackedVariableW(const Str, OpenBracket, CloseBracket: WideString;
  var i, j: Integer): WideString;
procedure frxCommonErrorHandler(Report: TfrxReport; const Text: String);
procedure frxInfoMsg(const Text: String);
function frxConfirmMsg(const Text: String; Buttons: Integer): Integer;
function frxIsValidFloat(const Value: string): Boolean;
procedure frxAssignImages(Bitmap: TBitmap; dx, dy: Integer;
  ImgList1: TImageList; ImgList2: TImageList = nil);
procedure frxDrawIcon(ImgList: TCustomImageList; aBitmap: TBitmap; IconIndex: Integer; BgColor: TColor = clNone);
procedure frxDrawTransparent(Canvas: TCanvas; x, y: Integer; bmp: TBitmap);
procedure frxDrawCheckBox(Canvas: TCanvas; ARect: TRect; aCheckColor: TColor; aFrameColor: TColor; bIsChecked: Boolean);
procedure frxDrawArrow(Canvas: TCanvas; aRect: TRect; aColor: TColor; bVertical: Boolean = True; bMirrored: Boolean = False);
procedure frxDrawArrows(Canvas: TCanvas; aRect: TRect; aColor: TColor; bVertical: Boolean = True; bMirrored: Boolean = False; LineWidth: Integer = 2; LineGap: Integer = 2; Gap: Integer = 3; Count: Integer = 2);
procedure frxParsePageNumbers(const PageNumbers: String; List: TStrings;
  Total: Integer);
function ColorText(Color: TColor) : String;
function FFColorText(Color: TColor) : String;
function HTMLRGBColor(Color: TColor): string;
procedure frxWriteCollection(Collection: TCollection; Writer: TWriter; Owner: TfrxComponent); overload;
procedure frxReadCollection(Collection: TCollection; Reader: TReader; Owner: TfrxComponent); overload;
function frxWriteCollection(Collection: TCollection; const Name: String; Item: TfrXMLItem; Owner: TfrxComponent; aAncestor: TfrCollection; ACreateNewItem: Boolean = True): TfrXMLItem; overload;
procedure frxReadCollection(Collection: TCollection; Item: TfrXMLItem; Owner: TfrxComponent;  aAncestor: TfrCollection); overload;
procedure frxWriteProperties(Instance: TPersistent; Writer: TWriter; Owner: TfrxComponent);
procedure frxReadProperties(Instance: TPersistent; Reader: TReader; Owner: TfrxComponent);

function frxCreateTempFile(const TempDir: String): String;
{$IFNDEF DELPHI7}
function frFloat2Str(const Value: Extended; const Prec: Integer = 2): String;
{$ELSE}
function frFloat2Str(const Value: Extended; const Prec: Integer = 2; const Sep: Char = '.'): String;
{$ENDIF}
function frxReverseString(const AText: string): string;
function frxStreamCRC32(Stream: TStream): Cardinal;
function frxUnixPath2WinPath(const Path: string): string;
{$IFNDEF DELPHI6}
function DirectoryExists(const Name: string): Boolean;
{$ENDIF}
{$IFNDEF DELPHI7}
function PosEx(const SubStr, S: string; Offset: Cardinal = 1): Integer;
{$ENDIF}
procedure TransparentFillRect(aDC: HDC; Left, Top, Right, Bottom: Integer;
  Color: TColor);
function GetStr(const Id: string): string;
function frxFloatDiff(const Val1, Val2: Extended): Boolean;

procedure ChangeGlyphColor(Index: Integer; Color: TColor);
function DateTimeToUTC(DateTime: TDateTime): TDateTime;

type
  TByteDinArray = array of Byte;
  TWordDinArray = array of Word;
  TLongWordDinArray = array of LongWord;
  TIntegerDinArray = array of Integer;

const
  PageBGName = '_pagebackground';
function IsPageBG(Obj: TfrxComponent): Boolean;

implementation
end.
