
{******************************************}
{                                          }
{          FastReport VCL/FMX/LCL          }
{             Graphics Library             }
{                                          }
{         Copyright (c) 1998-2022          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

{$IFNDEF FMX}
unit frGraphicUtils;
{$ENDIF}

interface

{$I frVer.inc}

uses
  {$IFNDEF FPC}
  Windows, Messages, WideStrings,
  {$ELSE}
  LCLType, LCLIntf, LazUTF8, LCLProc, Printers, frLazGraphicClasses,
  {$ENDIF}
  {$IFDEF DELPHI16}System.UITypes, {$ENDIF}
  SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Types, 
  frCore, frCoreClasses, frUnicodeUtils, Variants, frVectorCanvas,
  frCanvasHelper, frTypes, frDialectHelper, frBaseGraphicsTypes;

type
 {$IFNDEF FPC}
  /// <summary>
  ///   This class used to override standard behaviour of metafile draw
  ///   function. By default metafile size includes bottom and right bounds, so
  ///   the real size of metafile is increased by 1. Default draw function in
  ///   VCL tries to compensate it and decreases right and bottom bounds by 1
  ///   in draw method. This doesn't work because PlayEnhMetaFile stretches
  ///   output. As a result output may lose 1 pixel in width and that's very
  ///   critical for outputs like barcodes lines.
  /// </summary>
  TfrxMetafile = class(TMetafile)
  protected
    procedure Draw(ACanvas: TCanvas; const Rect: TRect); override;
  end;
{$ELSE}
  TfrxMetafile = class(TMetafile);
{$ENDIF}
 (*$HPPEMIT '#undef NewLine'*)
  TIntArray = array[0..MaxInt div 4 - 1] of Integer;
  PIntArray = ^TIntArray;
  TSubStyle = (ssNormal, ssSubscript, ssSuperscript);

  TfrxHTMLTag = class(TObject)
  public
    Position: Integer;
    Size: Integer;
    AddY: Integer;
    Style: TFontStyles;
    Color: Integer;
    Default: Boolean;
    Small: Boolean;
    DontWRAP: Boolean;
    SubType: TSubStyle;
    procedure Assign(Tag: TfrxHTMLTag);
  end;

  TfrxHTMLTags = class(TObject)
  private
    FItems: TList;
    procedure Add(Tag: TfrxHTMLTag);
    function GetItems(Index: Integer): TfrxHTMLTag;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    function Count: Integer;
    property Items[Index: Integer]: TfrxHTMLTag read GetItems; default;
  end;

  TfrxBaseHandler = class(TObject)
  protected
    FSQLKeyWordsList: TStringList;
    FSynDialectStyle: TfrSynDialectStyle;
    FCommentAttr: TFont;
    FKeywordAttr: TFont;
    FNumberAttr: TFont;
    FStringAttr: TFont;
    FTextAttr: TFont;
    FItems: TList;
    FTempArray: PIntArray;
    FTempArraySize: Integer;
    FAddY: Integer;
    FColor: LongInt;
    FDefColor: LongInt;
    FDefOffsColor: LongInt;
    FDefSize: Integer;
    FDefStyle: TFontStyles;
    FPosition: Integer;
    FSize: Integer;
    FStyle: TFontStyles;
    FDontWRAP: Boolean;
    FSubStyle: TSubStyle;

    FInterpretationType: TfrxInterpretationsTypes;
    FHightlightingType: TfrxHightlightingsTypes;

    FLeftDlm: String;
    FRightDlm: String;

    function GetAllowTags: Boolean;
    procedure SetAllowTags(val: Boolean);
    procedure SetSynDialectStyle(Value: TfrSynDialectStyle);
    procedure SetCommentAttr(Value: TFont);
    procedure SetKeywordAttr(Value: TFont);
    procedure SetNumberAttr(Value: TFont);
    procedure SetStringAttr(Value: TFont);
    procedure SetTextAttr(Value: TFont);

    procedure NewLine;
    procedure Wrap(TagsCount: Integer; AddBreak: Boolean);
    function Add: TfrxHTMLTag;
    function AddTag: TfrxHTMLTag; virtual;
    function FillCharSpacingArray(var ar: PIntArray; const s: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF};
      Canvas: TCanvas; LineIndex, Add: Integer; Convert: Boolean; DefCharset: Boolean): Integer;
    function GetItems(Index: Integer): TfrxHTMLTags;
    function GetPrevTag: TfrxHTMLTag;
    procedure ReallocTempArray(ANewSize: Integer);
    procedure FreeTempArray;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure ExpandTags(var s: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF});
    procedure Clear;
    procedure SetDefaults(DefColor: TColor; DefSize: Integer;
      DefStyle: TFontStyles);
    function Count: Integer;

    property SynDialectStyle: TfrSynDialectStyle read FSynDialectStyle write SetSynDialectStyle;
    property CommentAttr: TFont read FCommentAttr write SetCommentAttr;
    property KeywordAttr: TFont read FKeywordAttr write SetKeywordAttr;
    property NumberAttr: TFont read FNumberAttr write SetNumberAttr;
    property StringAttr: TFont read FStringAttr write SetStringAttr;
    property TextAttr: TFont read FTextAttr write SetTextAttr;
    property Items[Index: Integer]: TfrxHTMLTags read GetItems; default;
    property Position: Integer read FPosition write FPosition;
    property AllowTags: Boolean read GetAllowTags write SetAllowTags;
    property InterpretationType: TfrxInterpretationsTypes read FInterpretationType write FInterpretationType;
    property HightlightingType: TfrxHightlightingsTypes read FHightlightingType write FHightlightingType;
    property LeftDlm: String read FLeftDlm write FLeftDlm;
    property RightDlm: String read FRightDlm write FRightDlm;
  end;

  TfrxHTMLTagsList = class(TfrxBaseHandler) //legacy
  public
    procedure ExpandHTMLTags(var s: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF});
  end;

  TfrxWrapTextLineEvent  = procedure(s: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF};
    Width, FirstLineWidth, CharSpacing: Integer) of object;
  TfrxTextAddEvent  = procedure (s: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF}; start, len: Integer) of object;
  TfrxTextAddMode = (tamOriginalToDraw, tamDrawToOriginal);

  TfrxDrawText = class(TObject)
  private
// internals
    FBMP: TBitmap;
//    FLocked: Boolean; commented by Samuray
    FCanvas: TCanvas;
    FDefPPI: Integer;
    FScrPPI: Integer;
    FTempArray: PIntArray;
    FTempArraySize: Integer;

    FHandler: TfrxBaseHandler;

// data passed by SetXXX calls
    FFontSize: Integer;
    FCharSpacing: Extended;
    FLineSpacing: Extended;
    FOptions: Integer;
    FOriginalRect: TRect;
    FParagraphGap: Extended;
    FPlainText: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF};
    FPrintScale: Extended;
    FRotation: Integer;
    FRTLReading: Boolean;
    FScaledRect: TRect;
    FScaleX: Extended;
    FScaleY: Extended;
    FText: TWideStrings;
    FWordBreak: Boolean;
    FWordWrap: Boolean;
    FWysiwyg: Boolean;
    FMonoFont: Boolean;
    FUseDefaultCharset: Boolean;
    FVC: TVectorCanvas;
    FUseOldLineHeight: Boolean;
    FForcedWordBreak: Boolean;
    procedure BaseTextAdd(s: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF}; start, len: Integer);
    function GetWrappedText: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF};
    function IsPrinter(C: TCanvas): Boolean;
    procedure frxExtTextOut(C: TCanvas; X, Y, add1, i, j: Integer; {$IFDEF FPC} taj: Integer; {$ENDIF}
      CosA, SinA: Extended; s: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF};
      AnsiStr: AnsiString);
    procedure DrawTextLine(C: TCanvas; const s: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF};
      X, Y, DX, LineIndex: Integer; Align: TfrxHAlign; var fh, oldfh: HFont);
    procedure DrawSelectedTextLine(C: TCanvas; const s: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF};
      X, Y, DX, LineIndex: Integer; Align: TfrxHAlign; var fh, oldfh: HFont);
    function GetFontHeightMetric(aCanvas: TCanvas): Integer;
    procedure ReallocTempArray(ANewSize: Integer);
    procedure FreeTempArray;
    function GetLineHeightCorrection: Extended;
    function GetLineHeight: Extended;
    procedure SetOptions(WordWrap, RTLReading, WordBreak, Clipped,
      Wysiwyg: Boolean; Rotation: Integer); overload;
  protected
    FLineY: array of Integer;
    DoTextAdd: TfrxTextAddEvent;
    DoWrapTextLine: TfrxWrapTextLineEvent;
    procedure BaseWrapTextLine(s: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF};
      Width, FirstLineWidth, CharSpacing: Integer);
  public
    constructor Create;
    destructor Destroy; override;

// Call these methods in the same order
    procedure SetFont(Font: TFont);
    procedure SetOptions(WordWrap, HTMLTags, RTLReading, WordBreak,
      Clipped, Wysiwyg: Boolean; Rotation: Integer); overload; //old
    procedure SetOptions(WordWrap: Boolean; vInterpretationType: TfrxInterpretationsTypes;
      vHightlightingType: TfrxHightlightingsTypes; RTLReading, WordBreak,
      Clipped, Wysiwyg: Boolean; Rotation: Integer;
      vCommentAttr, vKeywordAttr, vNumberAttr, vStringAttr, vTextAttr: TFont;
      vSynDialectStyle: TfrSynDialectStyle = nil); overload;
    procedure SetOptions(WordWrap, HTMLTags: Boolean; vInterpretationType: TfrxInterpretationsTypes;
      vHightlightingType: TfrxHightlightingsTypes; RTLReading, WordBreak,
      Clipped, Wysiwyg: Boolean; Rotation: Integer); overload;
    procedure SetGaps(ParagraphGap, CharSpacing, LineSpacing: Extended);
    procedure SetExpressionDelimiters(ExpressionDelimiters: String);
    procedure SetDimensions(ScaleX, ScaleY, PrintScale: Extended;
      OriginalRect, ScaledRect: TRect);
    // need for wrap correction
    procedure SetText(Text: TWideStrings; FirstParaBreak: Boolean = False; TopLineIndex: Integer = 0);
    procedure SetParaBreaks(FirstParaBreak, LastParaBreak: Boolean);
    function DeleteTags(const Txt: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF}): {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF};

// call these methods only after methods listed above
    procedure DrawText(C: TCanvas; HAlign: TfrxHAlign; VAlign: TfrxVAlign; UnderlinesTextMode: TfrxUnderlinesTextMode = ulmNone;
      SelectedLine: Integer = -1);
    function CalcHeight: Extended;
    function CalcWidth: Extended;
    function LineHeight: Extended;
    function TextHeight: Extended;
    function GetFontHeight(aCanvas: TCanvas): Integer;
    function GetLineY: TIntegerDynArray; // Last array element - top of non-existing line "after last"
// returns the text that don't fit in the bounds
    function GetInBoundsText: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF};
    function GetOutBoundsText(var ParaBreak: Boolean): {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF};
    function UnusedSpace: Extended;

// call these methods before and after doing something
    procedure Lock;
    procedure Unlock;
    procedure LockCanvas;
    procedure UnlockCanvas;

    property Canvas: TCanvas read FCanvas;
    property DefPPI: Integer read FDefPPI;
    property ScrPPI: Integer read FScrPPI;
    property WrappedText: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF} read GetWrappedText;
    property UseDefaultCharset: Boolean read FUseDefaultCharset write FUseDefaultCharset;
    property UseMonoFont: Boolean read FMonoFont write FMonoFont;
    property Text: TWideStrings read FText;
    property VC: TVectorCanvas read FVC write FVC;
    property UseOldLineHeight: Boolean read FUseOldLineHeight write FUseOldLineHeight;
    property ForcedWordBreak: Boolean read FForcedWordBreak;
  end;

  TfrxDrawTextSupEditMemo = class(TfrxDrawText)
  private
    FFindLine: TPoint;
    FCurLine: TPoint;
    FResultLine: TPoint;

    procedure OTDTextAdd(s: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF}; start, len: Integer);
    procedure OTDWrapTextLine(s: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF};
      Width, FirstLineWidth, CharSpacing: Integer);
    procedure DTOTextAdd(s: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF}; start, len: Integer);  
    procedure DTOWrapTextLine(s: {$IFDEF FPCUNICODE}String{$ELSE}WideString{$ENDIF};
      Width, FirstLineWidth, CharSpacing: Integer);
  public
    procedure DrawSimpleTextLine(Canvas: TCanvas; LineNum: Integer; Align: TfrxHAlign);
    function SetTextWithMode(Text: TWideStrings; mode: TfrxTextAddMode; p: TPoint; FirstParaBreak: Boolean = False; TopLineIndex: Integer = 0): TPoint;
  {$IFDEF EMWYSIWYG}
    function GetY(val: Integer): Integer;
  {$ELSE}
    function GetRightLineY(C: TCanvas): Integer;
  {$ENDIF}
  end;

  function frxCreateRotatedFont(Font: TFont; Rotation: Integer): HFont;

var
  frxDrawText: TfrxDrawText;

implementation
end.
