
{******************************************}
{                                          }
{          FastReport VCL/FMX/LCL          }
{             Graphics Library             }
{                                          }
{         Copyright (c) 1998-2022          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frPictureCache;

interface

{$I frVer.inc}

uses
  {$IFNDEF FPC}
  Windows, Messages,
  {$ELSE}
  Types,
  {$ENDIF}
  SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  frXML, frBaseGraphicsTypes, SyncObjs, Variants, frHashUtils;


type
  TfrxCacheItem = packed record
    Segment: Longint;
    Offset: Longint;
    Loading: Integer;
    Width: Integer;
    Height: Integer;
  end;

  TfrxThumbnailsUpdateState = (tusAdded, tusClear);
  PfrxCacheItem = ^TfrxCacheItem;

  TfrxThumbnailsUpdate = procedure (Sender: TObject; UpdateState: TfrxThumbnailsUpdateState) of object;

  TFindOrAddGraphic = function (g: TGraphic; ResultType: TfrPictureHashMapResultType; IntData: Integer = -1): Integer of object;
  TGetIsLastNew = function: Boolean of object;

  TfrxPictureHashMap = class
  private
    //For Real
    FfrxDataHashMap: TfrDataHashMap;
    FIsLastNew: Boolean;
    //For Fake
    FCount: Integer;

    FRealHash: Boolean;
    FFindOrAddGraphic: TFindOrAddGraphic;
    FGetIsLastNew: TGetIsLastNew;

    function RealFindOrAddGraphic(g: TGraphic; ResultType: TfrPictureHashMapResultType; IntData: Integer = -1): Integer;
    function RealGetIsLastNew: Boolean;
    function FakeFindOrAddGraphic(g: TGraphic; ResultType: TfrPictureHashMapResultType; IntData: Integer = -1): Integer;
    function FakeGetIsLastNew: Boolean;

    procedure SetRealHash(RealHash: Boolean);

  public
    constructor Create(RealHash: Boolean);
    destructor Destroy; override;

    property FindOrAddGraphic: TFindOrAddGraphic read FFindOrAddGraphic;
    property IsLastNew: TGetIsLastNew read FGetIsLastNew;
    property RealHash: Boolean read FRealHash write SetRealHash;
  end;

  TfrxCacheList = class(TObject)
  private
    function Get(Index: Integer): PfrxCacheItem;
  protected
    FItems: TList;
  protected
    procedure Clear;
  public
    constructor Create;
    destructor Destroy; override;
    function Add: PfrxCacheItem;
    function Count: Integer;
    function GetSize(const Index: Integer): TSize;
    property Items[Index: Integer]: PfrxCacheItem read Get; default;
  end;

  TfrxFileStream = class(TFileStream)
  private
    FSz: LongWord;
  public
    function Seek(Offset: Longint; Origin: Word): Longint; override;
{$IFDEF DELPHI6}
    function Seek(const Offset: Int64; Origin: TSeekOrigin): Int64; override;
{$ENDIF}
  end;

  TfrxMemoryStream = class(TMemoryStream)
  private
    FSz: LongWord;
  public
    function Seek(Offset: Longint; Origin: Word): Longint; override;
{$IFDEF DELPHI6}
    function Seek(const Offset: Int64; Origin: TSeekOrigin): Int64; override;
{$ENDIF}
  end;


  TfrxInternalPictureCache = class(TObject)
  private
    FItems: TfrxCacheList;
    FCacheStreamList: TList;
    FTempFile: TStringList;
    FTempDir: String;
    FUseFileCache: Boolean;
    FCrit: TCriticalSection;
    FOnlyInternalFormats: Boolean;
    function Add: PfrxCacheItem;
    procedure SetTempDir(const Value: String);
    procedure SetUseFileCache(const Value: Boolean);
    procedure AddSegment;
    function InternalGetPicture(APicture: TPicture; aGraphic: TGraphic; ImageIndex: Integer): TGraphic;
  protected
    function Count: Integer;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    function AddPicture(aGraphic: TGraphic): Integer;
    procedure GetPicture(aGraphic: TGraphic; ImageIndex: Integer); overload;
    procedure GetPicture(Picture: TPicture; ImageIndex: Integer); overload;
    function GetPicture(const ImageIndex: Integer): TGraphic; overload;
    function GetSize(const ImageIndex: Integer): TSize;
    procedure SetSize(const ImageIndex: Integer; Width, Height: Integer);
    function IsExist(const ImageIndex: Integer): Boolean;
    procedure SaveToXML(Item: TfrXMLItem);
    procedure LoadFromXML(Item: TfrXMLItem);
    property UseFileCache: Boolean read FUseFileCache write SetUseFileCache;
    property OnlyInternalFormats: Boolean read FOnlyInternalFormats write FOnlyInternalFormats;
    property TempDir: String read FTempDir write SetTempDir;
  end;

  TfrxPictureCache = class;

  TfrxThumbnailThread = class(TThread)
  private
    FCache: TfrxPictureCache;
    procedure DoUpdate;
  public
    constructor Create(ACache: TfrxPictureCache);
    destructor Destroy; override;
    procedure Execute; override;
  end;

  TfrxPictureCache = class(TInterfacedObject, IfrxPictureCache)
  private
    FPictureCache: TfrxInternalPictureCache;
    FThumbnailCache: TfrxInternalPictureCache;
    FCahcedBitmap: TStringList;
//    FThumbnailThread: TfrxThumbnailThread;
    FOnThumbnailUpdated: TfrxThumbnailsUpdate;
    FIsUpdating: Boolean;
    FPictureCacheOptions: TfrxPictureCacheOptions;
    FfrxPictureHashMap: TfrxPictureHashMap;
    function GetTempDir: String;
    function GetUseFileCache: Boolean;
    procedure SetTempDir(const Value: String);
    procedure SetUseFileCache(const Value: Boolean);
    function GetPicture(const ImageIndex: Integer; const CacheType: TfrxCachedGraphicType): TGraphic; overload;
    procedure BuildThumbnailFromCache;
    procedure StartThumbnailThread;
    procedure FreeThumbnailThread;
    function CalcThumbnailSize(const Width, Height: Integer; const CacheType: TfrxCachedGraphicType): TSize;
    procedure ThumbnailUpdate(State: TfrxThumbnailsUpdateState);
    procedure SetPictureCacheOptions(const Value: TfrxPictureCacheOptions);
    function AddPictureTo(aGraphic: TGraphic; InternalCache: TfrxInternalPictureCache): Integer;
    function IsAutoRefMode: Boolean;
    function GetSize(const CacheType: TfrxCachedGraphicType; const ImageIndex: Integer): TSize;
    function GetCount: Integer;
    procedure UpdatePictureCacheOptions;
    function GetOnlyInternalFormats: Boolean;
    procedure SetOnlyInternalFormats(const Value: Boolean);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    procedure BuildFinished;
    function AddPicture(aGraphic: TGraphic): Integer;
    procedure GetGraphic(aGraphic: TGraphic; ImageIndex: Integer); overload;
    procedure GetPicture(APicture: TPicture; ImageIndex: Integer); overload;
    function GetGraphic(ImageIndex: Integer): TGraphic; overload;
    function GetCachedBitmap(CacheType: TfrxCachedGraphicType; ImageIndex: Integer): IfrxCachedGraphic;
    function GetPictureTo(APicture: TPicture; aGraphic: TGraphic; ImageIndex: Integer): TGraphic; overload;
    procedure SaveToXML(Item: TfrXMLItem);
    procedure LoadFromXML(Item: TfrXMLItem);
    property UseFileCache: Boolean read GetUseFileCache write SetUseFileCache;
    property TempDir: String read GetTempDir write SetTempDir;
    property PictureCacheOptions: TfrxPictureCacheOptions read FPictureCacheOptions write SetPictureCacheOptions;
    property OnThumbnailUpdated: TfrxThumbnailsUpdate read FOnThumbnailUpdated write FOnThumbnailUpdated;
    property OnlyInternalFormats: Boolean read GetOnlyInternalFormats write SetOnlyInternalFormats;
    property Count: Integer read GetCount;
  end;

implementation
end.
