
{******************************************}
{                                          }
{          FastReport VCL/FMX/LCL          }
{             Graphics Library             }
{                                          }
{         Copyright (c) 1998-2022          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frPictureGraphics;

{$I frVer.inc}

interface

uses
  SysUtils, Classes,
{$IFNDEF Linux}
  Windows,
{$ENDIF}
  Graphics, Variants, frCore, frBaseGraphicsTypes,
{$IFDEF FPC}
  Types, LCLType, LCLIntf, LCLProc, frLazGraphicClasses,
{$ENDIF}
{$IFDEF DELPHI16}
  System.Types,
{$ENDIF}
  SyncObjs;

const
  frxBitmapFileFormat = 0;

type
  TfrxGraphicFormatCap = (gcSaveTo, gcLoadFrom, gcGetCanvas, gcDraw, gcConvert, gcConvertToBitmap, gcInternalAdapters);
  TfrxGraphicStringType = (stClassName, stName, stExtension, stMime);
  TfrxGraphicFormatCaps = set of TfrxGraphicFormatCap;
  TfrxGraphicCanvasHelper = class;
  TfrxGraphicCanvasHelperClass = class of TfrxGraphicCanvasHelper;

  TfrxCanvasClass = class of TCanvas;

  TfrxGraphicProps = packed record
    HasAlpha: Boolean;
    Transparent: Boolean;
    TransparentColor: TColor;
    Quality: Integer;
    PixelFormat: TPixelFormat;
  end;

  TfrxDrawGraphicExt = packed record
    DrawProps: TfrxGraphicDrawProps;
    TransparentColor: TColor;
    Quality: TfrxGraphicQuality;
  end;

  EAbstractGraphicFormatException = class(Exception);

  TfrxCustomGraphicFormat = class
  protected
    class function GetCanvasHelperClass: TfrxGraphicCanvasHelperClass; virtual;
    class procedure DoDrawExt(const GProps: TfrxDrawGraphicExt; Canvas: TCanvas; AGraphic: TGraphic; const Area: TRect; ScaleX: Double; ScaleY: Double); virtual;
  public
    class function ConvertFrom(Graphic: TGraphic; DestPixelFormat: TPixelFormat; DestQuality: Integer = 100): TGraphic; virtual;
    class function ConvertToBitmap(Graphic: TGraphic; DestPixelFormat: TPixelFormat): TGraphic; virtual;
    class function ConvertToMaskedBitmap(Graphic: TGraphic; DestPixelFormat: TPixelFormat; TransparentMask: TColor): TGraphic; virtual;
    class function CreateNew(Width: Integer; Height: Integer; PixelFormat: TPixelFormat; Transparent: Boolean; Quality: Integer = 100): TGraphic; virtual;
    class function CreateFromStream(const Stream: TStream): TGraphic; virtual;
    class function CreateCanvasHelper(Graphic: TGraphic): TfrxGraphicCanvasHelper; virtual;
    class procedure ScaleGraphic(Graphic: TGraphic; NewWidth, NewHeight: Integer); virtual;
    class function ScaleGraphicToNew(Graphic: TGraphic; NewWidth, NewHeight: Integer): TGraphic; virtual;
    class procedure Draw(Canvas: TCanvas; AGraphic: TGraphic; Area: TRect; Quality: TfrxGraphicQuality); virtual;
    class procedure DrawTransparent(Canvas: TCanvas; AGraphic: TGraphic; Area: TRect; MaskColor: TColor; Quality: TfrxGraphicQuality); virtual;
    class procedure DrawExt(const GProps: TfrxDrawGraphicExt; Canvas: TCanvas; AGraphic: TGraphic; const Area: TRect; ScaleX: Double; ScaleY: Double); virtual;
    class function GetGraphicClass: TGraphicClass; virtual;
    class function GetGraphicMime: String; virtual;
    class function GetGraphicName: String; virtual;
    class function GetGraphicExt: String; virtual;
    class function GetGraphicConst: Integer; virtual;
    class function GetGraphicProps(Graphic: TGraphic): TfrxGraphicProps; virtual;
    class function GetFormatCapabilities: TfrxGraphicFormatCaps; virtual;
    class function GetAlphaBitmap(Graphic: TGraphic): TBitmap; virtual;
    class function GetMaskBitmap(Graphic: TGraphic; TransparentMask: TColor): TBitmap; virtual;
    class function HasAlphaChanel(Graphic: TGraphic): Boolean; virtual;
    class function HasMaskColor(Graphic: TGraphic): Boolean; virtual;
    class function IsTransparent(Graphic: TGraphic): Boolean; virtual;
    class function IsTranslucent: Boolean; virtual;
    class function IsSupportedFormat(const Stream: TStream): Boolean; virtual;
    class function IsVector: Boolean; virtual;
    class function LoadFromStream(const aPictire: TPicture; Stream: TStream): Boolean; virtual;
    class procedure SetTransparent(Graphic: TGraphic; Transparent: Boolean); virtual;
    class procedure SetTransparentColor(Graphic: TGraphic; Color: TColor); virtual;
    class function GetTransparentColor(Graphic: TGraphic): TColor; virtual;

    class function IsTextData(Graphic: TGraphic): Boolean; virtual;
    class function GetTextData(Graphic: TGraphic): string; virtual;
    class procedure SetTextData(Graphic: TGraphic; st: string); virtual;
    class function CreateFromText(st: string): TGraphic; virtual;
  end;

  TfrxFullAbilitiesGraphicFormat = class(TfrxCustomGraphicFormat)
  public
    class function GetFormatCapabilities: TfrxGraphicFormatCaps; override;
  end;

  TfrxCustomVectorGraphicFormat = class(TfrxFullAbilitiesGraphicFormat)
  protected
    class function ConvertToBitmapSize(Graphic: TGraphic; DestPixelFormat: TPixelFormat; Width, Height: Integer): TGraphic;
    class function ConvertToBitmapScale(Graphic: TGraphic; DestPixelFormat: TPixelFormat; Scale: Extended): TGraphic;
    class function GetAlphaBitmapSize(Graphic: TGraphic; Width, Height: Integer): TBitmap;
  public
    class function ConvertToBitmap(Graphic: TGraphic; DestPixelFormat: TPixelFormat): TGraphic; override;
    class function HasAlphaChanel(Graphic: TGraphic): Boolean; override;
    class function GetAlphaBitmap(Graphic: TGraphic): TBitmap; override;
    class function IsVector: Boolean; override;
  end;

  TfrxCustomGraphicFormatClass = class of TfrxCustomGraphicFormat;

  TfrxGraphicCanvasHelper = class
  private
    FFormatClass: TfrxCustomGraphicFormatClass;
  protected
    FGraphic: TGraphic;
    function GetCanvas: TCanvas; virtual;
    function GetCanvasGraphic: TGraphic; virtual;
  public
    constructor Create(Graphic: TGraphic; FormatClass: TfrxCustomGraphicFormatClass); virtual;
    destructor Destroy; override;
    procedure ReleaseCanvas; virtual;
    procedure SwitchFullOpacity; virtual; // Empty
    property Canvas: TCanvas read GetCanvas;
    property CanvasGraphic: TGraphic read GetCanvasGraphic;
    property GraphicHelper: TfrxCustomGraphicFormatClass read FFormatClass;
  end;

  TfrxBitmapCanvasHelper = class(TfrxGraphicCanvasHelper)
  protected
    function GetCanvas: TCanvas; override;
  end;

  TfrxBitmapGraphicFormat = class(TfrxFullAbilitiesGraphicFormat)
  protected
    class function GetCanvasHelperClass: TfrxGraphicCanvasHelperClass; override;
    class procedure DoDrawExt(const GProps: TfrxDrawGraphicExt; Canvas: TCanvas; AGraphic: TGraphic; const Area: TRect; ScaleX: Double; ScaleY: Double); override;
  public
    class function ConvertFrom(Graphic: TGraphic; DestPixelFormat: TPixelFormat; DestQuality: Integer = 100): TGraphic; override;
    class function ConvertFromVector(Graphic: TGraphic; DestPixelFormat: TPixelFormat; Scale: Extended = 1.0): TGraphic;
    class function CreateNew(Width: Integer; Height: Integer; PixelFormat: TPixelFormat; Transparent: Boolean; Quality: Integer = 100): TGraphic; override;
    class procedure Draw(Canvas: TCanvas; AGraphic: TGraphic; Area: TRect; Quality: TfrxGraphicQuality); override;
    class function HasAlphaChanel(Graphic: TGraphic): Boolean; override;
    class function GetGraphicClass: TGraphicClass; override;
    class function GetGraphicMime: String; override;
    class function GetGraphicName: String; override;
    class function GetGraphicExt: String; override;
    class function GetGraphicConst: Integer; override;
    class function GetGraphicProps(Graphic: TGraphic): TfrxGraphicProps; override;
    class function IsSupportedFormat(const Stream: TStream): Boolean; override;
    class procedure SetTransparent(Graphic: TGraphic; Transparent: Boolean); override;
    class procedure SetTransparentColor(Graphic: TGraphic; Color: TColor); override;
    class function GetTransparentColor(Graphic: TGraphic): TColor; override;
    class function GetAlphaBitmap(Graphic: TGraphic): TBitmap; override;
  end;

  { Default null graphic , used for draw unregistred classes }
  TfrxDefaultGraphicFormat = class(TfrxCustomGraphicFormat)
  public
    class function GetGraphicClass: TGraphicClass; override;
    class function GetGraphicMime: String; override;
    class function GetGraphicName: String; override;
    class function GetGraphicExt: String; override;
    class function GetGraphicConst: Integer; override;
    class function GetFormatCapabilities: TfrxGraphicFormatCaps; override;
  end;

  TfrxGraphicFormats = class(TList)
  private
    FGFormatsCS: TCriticalSection;
    function GetItem(Index: TfrListInteger): TfrxCustomGraphicFormatClass;
    procedure PutItem(Index: TfrListInteger;
      const Value: TfrxCustomGraphicFormatClass);
    function IsCapsSupported(ReqCaps, SourceCaps: TfrxGraphicFormatCaps): Boolean;
  public
    constructor Create;
    destructor Destroy; override;
    function BuildFormatsString(stType: TfrxGraphicStringType; const Separator: String = ' '; RequiredCaps: TfrxGraphicFormatCaps = []): String;
    function BuildFormatsStrings(stType: TfrxGraphicStringType; RequiredCaps: TfrxGraphicFormatCaps = []): TStringList;
    function Convert(SourceGraphic: TGraphic; const DestFormat: String; DestPixelFormat: TPixelFormat; DestQuality: Integer = 100): TGraphic;
    function ConvertToBitmap(SourceGraphic: TGraphic; DestPixelFormat: TPixelFormat): TGraphic;
    procedure Draw(Canvas: TCanvas; AGraphic: TGraphic; Area: TRect; Quality: TfrxGraphicQuality);
    procedure DrawExt(const GProps: TfrxDrawGraphicExt; Canvas: TCanvas; AGraphic: TGraphic; const Area: TRect; ScaleX: Double; ScaleY: Double);
    function FindByGraphic(GraphicClass: TGraphicClass; RequiredCaps: TfrxGraphicFormatCaps = []): TfrxCustomGraphicFormatClass;
    function FindByExt(const Ext: String; RequiredCaps: TfrxGraphicFormatCaps = []): TfrxCustomGraphicFormatClass;
    function FindByFormat(Stream: TStream; RequiredCaps: TfrxGraphicFormatCaps = []): TfrxCustomGraphicFormatClass;
    function FindByConst(const Index: Integer; RequiredCaps: TfrxGraphicFormatCaps = []): TfrxCustomGraphicFormatClass;
    function FindByName(const Name: String; RequiredCaps: TfrxGraphicFormatCaps = []): TfrxCustomGraphicFormatClass;
    function GetDefault: TfrxCustomGraphicFormatClass;
    function HasAlphaChanel(Graphic: TGraphic): Boolean;
    function IsTransparent(Graphic: TGraphic): Boolean;
    function LoadFromStream(const APictire: TPicture; AStream: TStream): Boolean; overload;
    function LoadFromStream(const APictire: TPicture; AStream: TStream; ARequiredCaps: TfrxGraphicFormatCaps): Boolean; overload;
    function LoadFromStream(AStream: TStream): TGraphic; overload;
    function LoadFromStream(AStream: TStream; ARequiredCaps: TfrxGraphicFormatCaps): TGraphic; overload;
    procedure ScaleGraphic(Graphic: TGraphic; NewWidth, NewHeight: Integer);
    function ScaleGraphicToNew(Graphic: TGraphic; NewWidth, NewHeight: Integer): TGraphic;
    procedure RegisterFormat(AClass: TfrxCustomGraphicFormatClass);
    procedure UnregisterFormat(AClass: TfrxCustomGraphicFormatClass);
    property Items[Index: TfrListInteger]: TfrxCustomGraphicFormatClass read GetItem write PutItem; default;
  end;

  function GetGraphicFormats: TfrxGraphicFormats;

var
  frxDefaultMetaCanvasClass: TfrxCanvasClass = nil;

implementation
end.
