
{******************************************}
{                                          }
{             FastReport VCL               }
{            SVG Base Elements             }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frSVGBase;

interface

{$I frVer.inc}

uses
  Windows, Classes, Graphics, Types,
  frCoreClasses,
  frSVGHelpers, frSVGCanvas, frSVGColor,
  frSVGElement, frSVGComponents, frCSSStyle, frSVGParse;

type
  TSVGBaseObj = class(TSVGElementObj)
  private
    FCanvasClipList: TfrOwnObjList;

    function GetStrokeData: TSVGStrokeData;
    procedure SetFiller(DrawType: TSVGDrawType; SVGPaint: TSVGPaint; Opacity: Single; const Graphics: TSVGCanvas);

  protected
    FPath: TSVGCanvasPath;

    procedure ConstructPath; override;
    procedure CalcClipPathList;
    function CalcSelfBounds: TSingleBounds; override;

    procedure DoPaintToGraphics(Graphics: TSVGCanvas); virtual;
    procedure PaintToPath(Path: TSVGCanvasPath); virtual;
  public
    destructor Destroy; override;
    procedure PaintToGraphics(Graphics: TSVGCanvas); virtual;
  end;

  Tel_svg = class(TSVGBaseObj)
  private
    procedure Paint(const Graphics: TSVGCanvas);
  protected
    FRootMatrix: TSVGTransform;

    function New(Parent: TSVGElementObj): TSVGElementObj; override;
    function CalcSelfBounds: TSingleBounds; override;
    function CalcXYTranslation(X, Y: Single): TSVGTransform; virtual;
    procedure CalcRootMatrix;
    function CalcCenteredShift(Offset: TSinglePoint): TSinglePoint; virtual;
    function CalcExternalTranslateMatrix: TSVGTransform; virtual;
    function CalcExternalScaleMatrix: TSVGTransform; virtual;
    function CalcAngleMatrix: TSVGTransform; virtual;
    function GetViewBox: TSingleBounds; virtual;
    procedure PaintTo(DC: HDC); virtual;
  public
    constructor Create; override;

    function GetHeight: Single; override;
    function GetWidth: Single; override;
    function GetTransform: TSVGTransform; override;

    function GetViewBoxWidth: Single;
    function GetViewBoxHeight: Single;
    function GetViewBoxDiagonal: Single;

    property RootMatrix: TSVGTransform read FRootMatrix;
  end;

  TSVGRootObj = class(Tel_svg)
  private
    FSource: string;
    FFileName: string;
    FStyleList: TfrxCSSList;
    FExternalOpacity: Single;
    FExternalAngle: Single;
    FExternalFontSize: Single;
    FExternalFontName: string;
    FExternalBounds: TSingleBounds;
    FExternalScale: TSinglePoint;
    FExternalCentered: Boolean;

    procedure SetExternalBounds(const Value: TSingleBounds);
    procedure SetExternalAngle(const Value: Single);
    procedure SetExternalFontName(const Value: string);
    procedure SetExternalFontSize(const Value: Single);
    procedure SetExternalOpacity(const Value: Single);
    procedure SetExternalScale(const Value: TSinglePoint);
    procedure SetExternalCentered(const Value: Boolean);
  protected
    FSVGCanvasClass: TSVGCanvasClass;
    FIsPrepared: Boolean;
    FIsFullOpacity: Boolean;

    procedure AssignTo(Dest: TPersistent); override;
    procedure AssignStylesApplyingSequenceTo(Obj: TSVGElementObj); override; // Empty
    function New(Parent: TSVGElementObj): TSVGElementObj; override;
    function GetCanvasClass: TSVGCanvasClass; override;
    function CalcCenteredShift(Offset: TSinglePoint): TSinglePoint; override;
    function CalcExternalTranslateMatrix: TSVGTransform; override;
    function CalcExternalScaleMatrix: TSVGTransform; override;
    function CalcAngleMatrix: TSVGTransform; override;
    function GetIsFullOpacity: Boolean; override;
    procedure SetIsFullOpacity(const Value: Boolean); override;
    function IsSizeless: Boolean;
    procedure Prepare;
  public
    constructor Create; override; // SVGCanvasClass = TSVGGDIPCanvas;
    constructor Create(SVGCanvasClass: TSVGCanvasClass); overload;
    destructor Destroy; override;
    procedure Clear; override;
    function IsRoot: Boolean; override;

    procedure ReadStyle(const Node: TfrxSVGXMLItem);
    procedure ClarifyOwnShorthandProperties; override;

    procedure LoadFromText(const Text: string);
    procedure LoadFromFile(const FileName: string);
    procedure LoadFromStream(Stream: TStream); overload;
    procedure SaveToFile(const FileName: string);
    procedure SaveToStream(Stream: TStream);

    procedure PaintTo(DC: HDC); override;

    function GetHeight: Single; override;
    function GetWidth: Single; override;

    function CalcFillOpacity: Single; override;
    function CalcStrokeOpacity: Single; override;

    function GetOuterHeight: Single;
    function GetOuterWidth: Single;
    procedure SureSizeAndViewBox;
    procedure atSetLengthSW(a: TSVGAttribute; LengthSW: TSVGLengthSW);

    function GetHostWidth: Single; override;
    function GetHostHeight: Single; override;
    function GetHostDiagonal: Single; override;

    property StyleList: TfrxCSSList read FStyleList;

    property ExternalOpacity: Single read FExternalOpacity write SetExternalOpacity;
    property ExternalAngle: Single write SetExternalAngle;  // Degree
    property ExternalFontSize: Single read FExternalFontSize write SetExternalFontSize;
    property ExternalFontName: string read FExternalFontName write SetExternalFontName;
    property ExternalBounds: TSingleBounds read FExternalBounds write SetExternalBounds;
    property ExternalScale: TSinglePoint write SetExternalScale;
    property ExternalCentered: Boolean write SetExternalCentered;

    property Source: string read FSource;
  end;

  Tel_g = class(TSVGBaseObj)
  protected
    function New(Parent: TSVGElementObj): TSVGElementObj; override;
  public
    constructor Create; override;
  end;

  Tel_a = class(TSVGBaseObj)
  protected
    function New(Parent: TSVGElementObj): TSVGElementObj; override;
  public
    constructor Create; override;
  end;

  Tel_switch = class(TSVGBaseObj)
  protected
    function New(Parent: TSVGElementObj): TSVGElementObj; override;
    function IsAnyChildHaveSystemLanguage: boolean;
    procedure SwitchChildren; override;
    procedure GetLanguage(out Language, PartLanguage: string);
  public
    constructor Create; override;
  end;

  Tel_use = class(TSVGBaseObj)
  protected
    FHrefObj: TSVGElementObj;

    function New(Parent: TSVGElementObj): TSVGElementObj; override;
    function IsHrefObjExists: Boolean;
  public
    constructor Create; override;
    procedure Construct;
    procedure PaintToGraphics(Graphics: TSVGCanvas); override;
    procedure PaintToPath(Path: TSVGCanvasPath); override;
  end;

  Tel_rect = class(TSVGBaseObj)
  protected
    function New(Parent: TSVGElementObj): TSVGElementObj; override;
    procedure ConstructPath; override;
  public
    constructor Create; override;
  end;

  Tel_line = class(TSVGBaseObj)
  protected
    function New(Parent: TSVGElementObj): TSVGElementObj; override;
    procedure ConstructPath; override;
  public
    constructor Create; override;
  end;

  Tel_polyline = class(TSVGBaseObj)
  private
    FPoints: TSinglePointDynArray;
  protected
    function New(Parent: TSVGElementObj): TSVGElementObj; override;
    procedure ConstructPath; override;
    function IsHasPoints: Boolean;
  public
    constructor Create; override;
  end;

  Tel_polygon = class(Tel_polyline)
  private
  protected
    function New(Parent: TSVGElementObj): TSVGElementObj; override;
    procedure ConstructPath; override;
  public
    constructor Create; override;
  end;

  Tel_circle = class(TSVGBaseObj)
  protected
    function New(Parent: TSVGElementObj): TSVGElementObj; override;
    procedure ConstructPath; override;
  public
    constructor Create; override;
  end;

  Tel_ellipse = class(TSVGBaseObj)
  protected
    function New(Parent: TSVGElementObj): TSVGElementObj; override;
    procedure ConstructPath; override;
  public
    constructor Create; override;
  end;

  Tel_image = class(TSVGBaseObj)
  private
    FImage: TSVGCanvasImage;
    FStream: TMemoryStream;
  protected
    function New(Parent: TSVGElementObj): TSVGElementObj; override;
    function IsFoundImage(S: string): boolean;
    function CalcSelfBounds: TSingleBounds; override;
    procedure ConstructPath; override;
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure PaintToGraphics(Graphics: TSVGCanvas); override;
  end;

  Tel_clipPath = class(TSVGBaseObj)
    private
      FInnerClipPath: TSVGCanvasPath;
      FTransformedPath: TSVGCanvasPath;
      FLastMatrix: TSVGTransform;
    protected
      function New(Parent: TSVGElementObj): TSVGElementObj; override;
      procedure ConstructClipPath;
    public
      constructor Create; override;
      destructor Destroy; override;
      function GetInnerClipPath(BoundingBox: TSingleBounds; UserMatrix: TSVGTransform): TSVGCanvasPath;
    end;

  TSVGFontMetric = record
    Descent, Ascent, InternalLeading, Capital, ExternalLeading: double;
  end;

  TSVGCustomText = class(TSVGBaseObj)
  private
    FText: string;
    FTextWidth: Single;
    FDecorations: TTextDecorations;
    FTextOrigin: TTextOrigin;
    FFontHeight: Single;

    function GetCompleteWidth: Single;
    procedure SetSize;
    function IsCalcSVGFontMetric(var Font: TFont; out FM: TSVGFontMetric): Boolean; // Frees Font

    function IsInTextPath: Boolean;
  protected
    procedure AssignTo(Dest: TPersistent); override;
    procedure ConstructPath; override;
    procedure DoPaintToGraphics(Graphics: TSVGCanvas); override;
    function GetFontData: TSVGFontData;
    procedure ClearDecorations;
    procedure ReadTextOrigin;
    procedure DeleteCRLF;
  public
    class function CreatePlainText(AParent: TSVGElementObj; ChildNodeText: string;
      ChildNodeIndex, ChildNodeCount: Integer): TSVGCustomText;
    destructor Destroy; override;
  end;

  Tel_text = class(TSVGCustomText)
  protected
    function New(Parent: TSVGElementObj): TSVGElementObj; override;
  public
    constructor Create; override;
  end;

  Tel_tspan = class(TSVGCustomText)
  protected
    function New(Parent: TSVGElementObj): TSVGElementObj; override;
  public
    constructor Create; override;
  end;

  Tel_textPath = class(TSVGCustomText)
  protected
    function New(Parent: TSVGElementObj): TSVGElementObj; override;
    procedure ConstructPath; override;
  public
    constructor Create; override;
  end;

implementation
end.
