
{******************************************}
{                                          }
{             FastReport VCL               }
{               SVG Canvas                 }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frSVGCanvas;

interface

{$I frVer.inc}

uses
  Windows,
  Graphics,
  Classes,
  Types,
  frSVGHelpers,
  frSVGColor,
  frSVGComponents,
  frSVGParse;

type
  TSVGDashData = record
    Offset: Single;
    Arr: TSingleDynArray;
  end;

  TSVGStrokeData = record
    Width: Single;
    Miterlimit: Single;
    LineCap: TSVGSpecificWord;
    LineJoin: TSVGSpecificWord;
    Dash: TSVGDashData;
  end;

  TSVGFontData = record
    Names: string; // May be like "Georgia, 'Times New Roman', Times, serif"
    Size: Single;
    Style: TSVGSpecificWord;
    Weight: Integer;
    Decoration: TSVGSpecificWordSet;
  end;

  TSVGGradientStopData = record
    SVGColor: TSVGColor;
    Offset: Single;
  end;
  TSVGGradientArray = array of TSVGGradientStopData;

  TSVGDrawType = (dtFill, dtStroke);
  TSVGFillerType = (ftSolidColor, ftLinearGradient, ftRadialGradient, ftPattern);

  TSVGGradientData = record
    gradientUnits: TSVGSpecificWord;
    spreadMethod: TSVGSpecificWord;
    UserBounds: TSingleBounds;
    gradientTransform: TSVGTransform;
    GradientArray: TSVGGradientArray;
    case FillerType: TSVGFillerType of
      ftLinearGradient: (x1, x2, y1, y2: Single);
      ftRadialGradient: (cx, cy, r, fx, fy, fr: Single);
  end;
//      cx, cy, r  defines the x|y-axis coordinate, radius of the end circle for a radial gradient.
//      fx, fy, fr defines the x|y-axis coordinate, radius of the focal point for a radial gradient

  TSVGCanvasClass = class of TSVGCanvas;
  TSVGCanvas = class;

  TSVGPatternData = record
    Bitmap: TBitmap;
    x, y, Width, Height, Factor: Single;
    Offset: TSinglePoint;
    Matrix: TSVGTransform;
  end;

  TSVGCanvasImage = class
  private
  public
    constructor Create(SA: TStreamAdapter); virtual; abstract;
    function GetWidth: Cardinal; virtual; abstract;
    function GetHeight: Cardinal; virtual; abstract;
  end;

  TSVGCanvasPath = class
  private
    FMatrix: TSVGTransform;
    FFillRule: TSVGSpecificWord;
  public
    constructor Create; virtual;
    function Clone: TSVGCanvasPath; virtual; abstract;
    procedure Transform(const Matrix: TSVGTransform); virtual; abstract;
    function Bounds: TSingleBounds; virtual; abstract;

    procedure AddArc(x, y, width, height, startAngle, sweepAngle: Single); virtual; abstract;

    procedure AddBezier(x1, y1, x2, y2, x3, y3, x4, y4: Single); overload; virtual; abstract;
    procedure AddBezier(p1, p2, p3, p4: TSinglePoint); overload;

    procedure AddCircle(cx, cy, r: Single); virtual; abstract;

    procedure AddEllipse(cx, cy, rx, ry: Single); virtual; abstract;

    procedure AddLine(x1, y1, x2, y2: Single); overload; virtual; abstract;
    procedure AddLine(p1, p2: TSinglePoint); overload;

    procedure AddPath(addingPath: TSVGCanvasPath; connect: Bool); virtual; abstract;

    procedure AddRectangle(x, y, width, height, rx, ry: Single); overload; virtual; abstract;
    procedure AddRectangle(Bounds: TSingleBounds; rx, ry: Single); overload;

    procedure CloseFigure; virtual; abstract;

    procedure StartFigure; virtual; abstract;

    function Log: string; virtual; abstract;

    property Matrix: TSVGTransform read FMatrix write FMatrix;
    property FillRule: TSVGSpecificWord read FFillRule write FFillRule;
  end;

  TSVGCanvasPathText = class
  protected
    FAdditionalMatrix: TSVGTransform;
  public
    function AddPathText(const Path: TSVGCanvasPath;
      const Text: string; const Indent: Single; FontData: TSVGFontData;
      const DistanceFactor: Single = 1; const KerningFactor: Single = 1): Single; virtual; abstract;

    property AdditionalMatrix: TSVGTransform read FAdditionalMatrix write FAdditionalMatrix;
  end;

  TTextDecoration = (tdUnderline, tdLineThrow, tdOwerline);
  TTextDecorations = array[TTextDecoration] of TSVGCanvasPath;

  TTextOrigin = record
    X, Y, DX, DY: TSingleDynArray;
  end;

  TSVGCanvas = class
  private
  protected
    FFillerType: array [TSVGDrawType] of TSVGFillerType;
    FSolidColorData: array [TSVGDrawType] of TSVGColor;
    FGradientData: array [TSVGDrawType] of TSVGGradientData;
    FPatternData: array [TSVGDrawType] of TSVGPatternData;

    procedure ChangeSolidColorFiller(DrawType: TSVGDrawType); virtual; abstract;
    procedure ChangeLinearGradientFiller(DrawType: TSVGDrawType); virtual; abstract;
    procedure ChangeRadialGradientFiller(DrawType: TSVGDrawType); virtual; abstract;
    procedure ChangePatternFiller(DrawType: TSVGDrawType); virtual; abstract;

    procedure StrokePath(Path: TSVGCanvasPath); virtual; abstract;
    procedure FillPath(Path: TSVGCanvasPath); virtual; abstract;
  public
    constructor Create(hdc: HDC); virtual; abstract;
    destructor Destroy; override;

    procedure SetTransform(const Matrix: TSVGTransform); virtual; abstract;
    procedure GetTransform(out Matrix: TSVGTransform); virtual; abstract;
    procedure ResetTransform; virtual; abstract;

    procedure SetSolidColor(DrawType: TSVGDrawType; SVGColor: TSVGColor);
    procedure SetGradient(DrawType: TSVGDrawType; GradientData: TSVGGradientData);
    procedure SetPattern(DrawType: TSVGDrawType; PatternData: TSVGPatternData);

    procedure SetClip(Path: TSVGCanvasPath); virtual; abstract;
    procedure IntersectClip(Path: TSVGCanvasPath); virtual; abstract;
    procedure ResetClip; virtual; abstract;

    procedure SetStroke(StrokeData: TSVGStrokeData); virtual; abstract;

    procedure PaintPath(Path: TSVGCanvasPath; FillRule: TSVGSpecificWord);

    function MeasureString(st: string; FontData: TSVGFontData;
      out FontHeight: Single): TSingleSize; virtual; abstract;

    procedure DrawImage(FImage: TSVGCanvasImage; SVGRect: TSingleBounds; FillOpacity: Single); virtual; abstract;

    class procedure AddStringToPath(st: string; Path: TSVGCanvasPath; Decorations: TTextDecorations;
      TextOrigin: TTextOrigin; FontData: TSVGFontData); virtual; // Empty: abstract; dosn't supported in D2010

    class function CreateImage(SA: TStreamAdapter): TSVGCanvasImage; virtual; // Empty: abstract; dosn't supported in D2010
    class function CreatePath: TSVGCanvasPath; virtual; // Empty: abstract; dosn't supported in D2010
    class function CreatePathText(Path: TSVGCanvasPath): TSVGCanvasPathText; virtual; // Empty: abstract; dosn't supported in D2010

    class function GetPathLength(const Path: TSVGCanvasPath): Single; virtual; // Empty: abstract; dosn't supported in D2010
    class function GetFont(FontData: TSVGFontData): TFont; virtual; // Empty: abstract; dosn't supported in D2010
  end;

function IsSameSVGColor(F1, F2: TSVGColor): Boolean;
function IsSameGradient(G1, G2: TSVGGradientData): Boolean;
function ToSVGGradientStopData(SVGColor: TSVGColor; Offset: Single): TSVGGradientStopData;
function IsSamePattern(P1, P2: TSVGPatternData): Boolean;

implementation
end.
