
{******************************************}
{                                          }
{             FastReport VCL               }
{               SVG Helpers                }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frSVGHelpers;

interface

{$I frVer.inc}

uses
  {$IFNDEF Linux}
  Windows,
  {$ENDIF}
  Classes, Types, Graphics, frXML;

type
  TfrxSVGXMLReader = class(TfrValuedXMLReader)
  protected
    FLastText: Boolean; // '<' already read
    constructor Create(Stream: TStream);
    procedure ProcessSecondLeftBrocket; override;
    function CreateItem: TfrXMLItem; override;
    procedure ReadValuedItem(out {$IFDEF DELPHI12}NameS, ValueS{$ELSE}Name, Value{$ENDIF}, Text: String); override;
  end;

  TfrxSVGXMLItem = class;

  TfrxSVGXMLDocument = class(TfrValuedXMLDocument)
  protected
    procedure ReadXMLHeader; override;
    function CreateReader: TfrValuedXMLReader; override;
    procedure LoadFromValuedStream;
  public
    constructor Create(Ignored: Pointer);
    procedure LoadFromXML(const XML: String);
  end;

  TfrxSVGIXMLChildNodes = class
  private
    function GetItems(Index: Integer): TfrxSVGXMLItem;
    function GetCount: Integer;
  protected
    FItems: TList;
  public
    constructor Create(AItems: TList);
    property Items[Index: Integer]: TfrxSVGXMLItem read GetItems; default;
    property Count: Integer read GetCount;
  end;

  TfrxSVGNameValue = record
    NodeName, NodeValue: string;
  end;

  TfrxSVGAttributeNodes = class
  private
    function GetCount: Integer;
    function GetNameValue(i: Integer): TfrxSVGNameValue;
    function GetValueByName(NodeName: string): string;
  protected
    SL: TStringList;
  public
    constructor Create(Text: string);
    destructor Destroy; override;
    function IndexOfName(const s: string): Integer;
    function IsHasAttribute(const s: string): Boolean;
    function FindNode(s: string): TfrxSVGXMLItem;

    property Count: Integer read GetCount;
    property NameValue[i: Integer]: TfrxSVGNameValue read GetNameValue; default;
    property ValueByName[NodeName: string]: string read GetValueByName;
  end;

  TfrxSVGXMLItem = class(TfrXMLItem)
  private
    FChildNodes: TfrxSVGIXMLChildNodes;
    FAttributeNodes: TfrxSVGAttributeNodes;

    function GetChildNodes: TfrxSVGIXMLChildNodes;
    function GetNodeValue: string;
    function GetNodeName: string;
    function GetAttributes(s: string): string;
    function GetHRef: string;
  protected
  public
    destructor Destroy; override;
    procedure AddItem(Item: TfrXMLItem); override;
    function HasAttribute(s: string): Boolean;
    procedure SetData(const AName, AText, AValue: String); override;

    property ChildNodes: TfrxSVGIXMLChildNodes read GetChildNodes;
    property NodeValue: string read GetNodeValue;
    property NodeName: string read GetNodeName;
    property AttributeNodes: TfrxSVGAttributeNodes read FAttributeNodes;
    property Attributes[s: string]: string read GetAttributes;
    property HRef: string read GetHRef;
  end;

function SplitCSSFont(st: string): TStringList;
function SuitableFont(FontList: string): string;

function IsStrInSet(Str: string; StrSet: array of string): Boolean;

{$IFNDEF DELPHI12}
type
  TSysCharset = set of Char; 
function CharInSet(Ch: Char; CharSet: TSysCharset): Boolean;
{$ENDIF}

type
  TSinglePoint = record
    X, Y: Single;
  end;

  TSinglePointDynArray = array of TSinglePoint;

function spCreate(X, Y: Single): TSinglePoint;
function spMultiply(SP: TSinglePoint; Factor: Single): TSinglePoint;
function spDivide(SP: TSinglePoint; Factor: Single): TSinglePoint;

function DynArrayCopy(DynArray: TSinglePointDynArray): TSinglePointDynArray; overload;
function DynArrayCopy(DynArray: TSingleDynArray): TSingleDynArray; overload;

type
  TSingleSize = record
    Width, Height: Single;
  end;

  TSingleBounds = record
    case Byte of
      0: (X, Y, Width, Height: Single);
      1: (TopLeft: TSinglePoint; Size: TSingleSize);
  end;

  TSingleRect = record
    case Byte of
      0: (Left, Top, Right, Bottom: Single);
      1: (TopLeft, BottomRight: TSinglePoint);
  end;

const
  EmptySingleBounds: TSingleBounds = (X: 0.0; Y: 0.0; Width: 0.0; Height:0.0);

function ToSinglePoint(const X, Y: Single): TSinglePoint; overload;
function ToSinglePoint(const SX, SY: string): TSinglePoint; overload;
function SinglePointSum(const p1, p2: TSinglePoint): TSinglePoint;
function SinglePointDiff(const p1, p2: TSinglePoint): TSinglePoint;
function SinglePointCenter(const p1, p2: TSinglePoint): TSinglePoint;
function ToSingleSize(const Width, Height: Single): TSingleSize;
function ToSingleBounds(const X, Y, Width, Height: Single): TSingleBounds; overload;
function ToSingleBounds(const Rect: TSingleRect): TSingleBounds; overload;
function ToSingleBounds(const Rect: TRect): TSingleBounds; overload;
function SingleBoundsUnion(const B1, B2: TSingleBounds): TSingleBounds;
function ToSingleRect(const Bounds: TSingleBounds): TSingleRect; overload;
function ToSingleRect(const Left, Top, Right, Bottom: Single): TSingleRect; overload;
function ToSingleRect(const A: TSinglePointDynArray): TSingleRect; overload;
function SingleRectUnion(const R1, R2: TSingleRect): TSingleRect;

function IsSameSingle(const A, B: Single; Epsilon: Single = 0): Boolean; // for D7
function IsSameBounds(const B1, B2: TSingleBounds): boolean;
function IsSameSinglePoint(const p1, p2: TSinglePoint): boolean;

type
//[0, 0] / a / m11 / eM11     [1, 0] / b / m12 / eM12
//[0, 1] /  / m21 / eM21     [1, 1] / d / m22 / eM22
//[0, 2] / e / m31 / eDx      [1, 2] / f / m32 / eDy
  TSVGTransform = record
    a, b, c, d, e, f: Single;  // m11, m12, m21, m22, m31, m32
  end;

const
  EmptyTransform: TSVGTransform = (a: 0.0; b: 0.0; c: 0.0; d: 0.0; e: 0.0; f: 0.0);

function tmIdentity: TSVGTransform;
function tmMultiply(const M1, M2: TSVGTransform): TSVGTransform;

function tmMatrix(const a, b, c, d, e, f: Single): TSVGTransform; overload;
function tmMatrix(const Data: TSingleDynArray): TSVGTransform; overload;

function tmRotation(const Alpha: Single): TSVGTransform; overload;
function tmRotation(const Alpha, X, Y: Single): TSVGTransform; overload;
function tmRotation(const Data: TSingleDynArray): TSVGTransform; overload;

function tmScaling(const Data: TSingleDynArray): TSVGTransform; overload;
function tmScaling(const ScaleX, ScaleY: Single): TSVGTransform; overload;
function tmScaling(const Scale: Single): TSVGTransform; overload;
function tmScalingX(const ScaleX: Single): TSVGTransform;
function tmScalingY(const ScaleY: Single): TSVGTransform;

function tmSkewing(const Data: TSingleDynArray): TSVGTransform; overload;
function tmSkewing(const SkewX, SkewY: Single): TSVGTransform; overload;
function tmSkewing(const Skew: Single): TSVGTransform; overload;
function tmSkewingX(const SkewX: Single): TSVGTransform;
function tmSkewingY(const SkewY: Single): TSVGTransform;

function tmTranslation(const Data: TSingleDynArray): TSVGTransform; overload;
function tmTranslation(const X, Y: Single): TSVGTransform; overload;
function tmTranslation(const X: Single): TSVGTransform; overload;
function tmTranslationX(const X: Single): TSVGTransform;
function tmTranslationY(const Y: Single): TSVGTransform;

function tmTransform(const P: TSinglePoint; const M: TSVGTransform): TSinglePoint;
function IsSameTransform(const M1, M2: TSVGTransform): Boolean;
function IsEmptyTransform(const M: TSVGTransform): Boolean;
function tmToStr(const M: TSVGTransform): string;

function TryStrToSingle(const S: string; out Value: Single): Boolean;
function StrToSingle(const S: string): Single;

function Dequote(const Value: string): string;

function IsBitmapBottomUp(Bitmap: TBitmap): Boolean;
implementation
end.
