{******************************************}
{                                          }
{             FastReport VCL               }
{             Report classes               }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

/// <summary>
///   Main FastReport unit. This unit contains base classes of report objects
///   and report class itself. To use FastReport components this unit should be
///   included in uses clause.
/// </summary>
unit frxClass;

interface

{$I frx.inc}
{$IFNDEF DELPHI16}
(*$HPPEMIT '#pragma link "Gdiplus.lib"'*)
{$ENDIF}

uses
{$IFNDEF FPC}
  Windows, Messages, WideStrings,
{$ELSE}
  LResources, LMessages, LCLType, LCLIntf, LazarusPackageIntf,
  LCLProc, FileUtil, LazHelper, frLazGraphicClasses, StdCtrls,
{$ENDIF}
  SysUtils, Types, Classes, Graphics, Controls, Forms, Dialogs,
  IniFiles, ExtCtrls, Printers, Math, frPictureCache,
  {$IFDEF DELPHI16}System.UITypes,{$ENDIF}
  {$IFNDEF NO_CRITICAL_SECTION} SyncObjs, {$ENDIF}
  frCore, frCoreClasses, frTypes, frXML,
  frBaseGraphicsTypes, frxVariables, frxProgress,
  fs_iinterpreter, frUnicodeUtils, Variants,
  frVectorCanvas, frxBaseForm, frPictureGraphics, frxProtocolFactory, frGraphicUtils,
  frDPIAwareUtils, frxDPIAwareBaseControls;

const
  /// <summary>
  ///   Used to convert Pixels to millimeters.
  /// </summary>
  fr01cm: Extended = 3.77953;
  /// <summary>
  ///   Used to convert Pixels to centimeters.
  /// </summary>
  fr1cm: Extended = 37.7953;
  /// <summary>
  ///   Used to convert Pixels to 0.X inches.
  /// </summary>
  fr01in: Extended = 9.6;
  /// <summary>
  ///   Used to convert Pixels to inches.
  /// </summary>
  fr1in: Integer = 96;
  /// <summary>
  ///   Used to convert Pixels to char width(dot-matrix report).
  /// </summary>
  fr1CharX: Extended = 9.6;
  /// <summary>
  ///   Used to convert Pixels to char height(dot-matrix report).
  /// </summary>
  fr1CharY: Integer = 17;
  /// <summary>
  ///   Default transparent color value.
  /// </summary>
  clTransparent: TColor = clNone;
  /// <summary>
  ///   Cursor represents hand sign.
  /// </summary>
  crHand: Integer = 150;
  /// <summary>
  ///   Cursor represents zoom sign.
  /// </summary>
  crZoom: Integer = 151;
  /// <summary>
  ///   Cursor represents format sign.
  /// </summary>
  crFormat: Integer = 152;
  DEF_REG_CONNECTIONS: String = '\Software\Fast Reports\Connections';
  /// <summary>
  ///   Internal message. Sends to window controls during creation to
  ///   synchronize creation of window control with main GDI thread.
  /// </summary>
  WM_CREATEHANDLE = WM_USER + 1;
  /// <summary>
  ///   Internal message. Sends to window controls during destroy to
  ///   synchronize destroying of window control with main GDI thread.
  /// </summary>
  WM_DESTROYHANDLE = WM_USER + 2;
  /// <summary>
  ///   Internal message. Used for synchronization between report component and
  ///   Engine ShiftTree item. Sends when component destroying.
  /// </summary>
  FRX_OWNER_DESTROY_MESSAGE = $D001;
  /// <summary>
  ///   Represents default pixel per inch value.
  /// </summary>
  frx_DefaultPPI = FR_DefaultPPI;

type
  {$IFDEF NONWINFPC}

  { TfrxStringList }

  TfrxStringList = class(TStringList)
  protected
    function DoCompareText(const s1,s2 : string) : PtrInt; override;
  end;

  {$ENDIF}

  TfrxReport = class;
  TfrxPage = class;
  TfrxReportPage = class;
  TfrxDialogPage = class;
  TfrxCustomEngine = class;
  TfrxCustomDesigner = class;
  TfrxCustomPreview = class;
  TfrxCustomPreviewPages = class;
  TfrxComponent = class;
  TfrxReportComponent = class;
  TfrxView = class;
  TfrxCustomStyleItem = class;
  TfrxCustomExportFilter = class;
  TfrxCustomCompressor = class;
  TfrxCustomDatabase = class;
  TfrxFrame = class;
  TfrxDataSet = class;
  TfrxHyperlink = class;
  TfrxInPlaceEditor = class;
  TfrxCustomIOTransport = class;
  TfrxComponentEditorsList = class;
  TfrxComponentEditorsManager = class;
  TfrxPostProcessor = class;
  TfrxObjectProcessing = class;
  TfrxMacrosItem = class;
  TfrxSelectedObjectsList = class;
  TfrxDelayedCommandManager = class;

  TfrxNotifyEvent = type String;
  TfrxCloseQueryEvent = type String;
  TfrxKeyEvent = type String;
  TfrxKeyPressEvent = type String;
  TfrxMouseEvent = type String;
  TfrxMouseEnterViewEvent = type String;
  TfrxMouseLeaveViewEvent = type String;
  TfrxMouseMoveEvent = type String;
  TfrxPreviewClickEvent = type String;
  TfrxRunDialogsEvent = type String;
  TfrxContentChangedEvent = type String;

  TfrxStyleName = type String;

  /// <summary>
  ///   Duplicate name exception class.
  /// </summary>
  EDuplicateName = class(Exception);
  /// <summary>
  ///   Export termination exception class.
  /// </summary>
  EExportTerminated = class(TObject);
  EReportScriptError = class(Exception);

  SYSINT = Integer;

  // TODO
  // remove unnecesary later
  // csContainer is legacy and used only with current crossTab for back compatibility
  // csObjectsContainer new flag which used in default Copy/Paste buffer for containers like Table
  // csContained means that object of this class only can be used inside other View object
  // csAcceptsFrxComponents used when view can accept other components as childs

  TfrxComponentState = set of (csFrxDesigning, csFrxWriting, csFrxSerializeToDict, csFrxSerializeToPreviewPages, csFrxModifyObject, csFrxExporting, csFrxInteractiveForms);

  /// <summary>
  ///   Styles on report component.
  /// </summary>
  TfrxComponentStyle = set of (
    /// <summary>
    ///   Deprecated. used only for cross tabs object. Use csObjectsContainer
    ///   instead.
    /// </summary>
    csContainer,
    /// <summary>
    ///   Report component visible in the report preview(prepared report).
    /// </summary>
    csPreviewVisible,
    /// <summary>
    ///   Use default RTTI serialization when object saves to prepared
    ///   report(Default serialization to prepared report).
    /// </summary>
    csDefaultDiff,
    /// <summary>
    ///   Determinates how to serialize internal non published composite
    ///   objects. When flag is set composite object saves as new XML node.
    /// </summary>
    csHandlesNestedProperties,
    /// <summary>
    ///   Report component can be child of another parent report container.
    /// </summary>
    csContained,
    /// <summary>
    ///   Report component may contain another report component, i.e. can be as
    ///   parent container for others.
    /// </summary>
    csAcceptsFrxComponents,
    /// <summary>
    ///   Report component is a object container. Flag used in default
    ///   Copy/Paste buffer for containers like Table.
    /// </summary>
    csObjectsContainer);
  /// <summary>
  ///   Stretch mode of the text objects.
  /// </summary>
  TfrxStretchMode = (
    /// <summary>
    ///   Stretch the object so all text would fit.
    /// </summary>
    smDontStretch,
    /// <summary>
    ///   Don't stretch the object.
    /// </summary>
    smActualHeight,
    /// <summary>
    ///   Stretch the object to the max height of the parent band.
    /// </summary>
    smMaxHeight,
    /// <summary>
    ///   Stretch the object to the max height of each part of the split band.
    /// </summary>
    smPartMaxHeight);
  /// <summary>
  ///   Shift mode of the object. Behaviour depends on what type of
  ///   TfrxShiftEngine selected in parent container.
  /// </summary>
  TfrxShiftMode = (
    /// <summary>
    ///   Don't shift the object.
    /// </summary>
    smDontShift,
    /// <summary>
    ///   Shift the object if the above objects are stretching(for seLinear
    ///   shifts even if there is no overlying objects). <br />
    /// </summary>
    smAlways,
    /// <summary>
    ///   Shift the object if it is overlapped by the above objects(for
    ///   seLinear consider vertical space between objects).
    /// </summary>
    smWhenOverlapped);
  /// <summary>
  ///   Determine how container shifts its child object.
  /// </summary>
  TfrxShiftEngine = (
    /// <summary>
    ///   Linear shift mechanism used.
    /// </summary>
    seLinear,
    /// <summary>
    ///   Tree shift mechanism used. Default.
    /// </summary>
    seTree,
    /// <summary>
    ///   Shift mechanism disabled. May increase speed.
    /// </summary>
    seDontShift);
  /// <summary>
  ///   Mode of the duplex printing.
  /// </summary>
  TfrxDuplexMode = (dmNone, dmVertical, dmHorizontal, dmSimplex);

  /// <summary>
  ///   The alignment of the object relative to its parent.
  /// </summary>
  TfrxAlign = (
    /// <summary>
    ///   No align.
    /// </summary>
    baNone,
    /// <summary>
    ///   Aligned to left.
    /// </summary>
    baLeft,
    /// <summary>
    ///   Aligned to right. <br />
    /// </summary>
    baRight,
    /// <summary>
    ///   Aligned to center. <br />
    /// </summary>
    baCenter,
    /// <summary>
    ///   Aligned to width of parent container. <br />
    /// </summary>
    baWidth,
    /// <summary>
    ///   Aligned to bottom. <br />
    /// </summary>
    baBottom,
    /// <summary>
    ///   Aligned to parent sizes. <br />
    /// </summary>
    baClient,
    /// <summary>
    ///   Objects ignored(invisible) in align mechanism.
    /// </summary>
    baHidden);

  /// <summary>
  ///   Style of the frame line.
  /// </summary>
  TfrxFrameStyle = (
    /// <summary>
    ///   Solid line.
    /// </summary>
    fsSolid,
    /// <summary>
    ///   Dash style.
    /// </summary>
    fsDash,
    /// <summary>
    ///   Dot style.
    /// </summary>
    fsDot,
    /// <summary>
    ///   Dash dot style.
    /// </summary>
    fsDashDot,
    /// <summary>
    ///   Dash dot dot style.
    /// </summary>
    fsDashDotDot,
    /// <summary>
    ///   Double line style.
    /// </summary>
    fsDouble,
    /// <summary>
    ///   Alternative dots style.
    /// </summary>
    fsAltDot,
    /// <summary>
    ///   Squares instead of dots.
    /// </summary>
    fsSquare);

  /// <summary>
  ///   Type of the frame line.
  /// </summary>
  TfrxFrameType = (
    /// <summary>
    ///   Left frame. <br />
    /// </summary>
    ftLeft,
    /// <summary>
    ///   Right frame. <br />
    /// </summary>
    ftRight,
    /// <summary>
    ///   Top frame. <br />
    /// </summary>
    ftTop,
    /// <summary>
    ///   Bottom frame.
    /// </summary>
    ftBottom);
  /// <summary>
  ///   Type of the frame line.
  /// </summary>
  TfrxFrameTypes = set of TfrxFrameType;

  /// <summary>
  ///   Visibility of object in the result report.
  /// </summary>
  TfrxVisibilityType  = (
    /// <summary>
    ///   Objects visible for preview.
    /// </summary>
    vsPreview,
    /// <summary>
    ///   Objects visible for exports.
    /// </summary>
    vsExport,
    /// <summary>
    ///   Objects visible for printing.
    /// </summary>
    vsPrint);
  /// <summary>
  ///   Visibility of object in the result report.
  /// </summary>
  TfrxVisibilityTypes = set of TfrxVisibilityType;

  /// <summary>
  ///   Reserved for future use.
  /// </summary>
  TfrxPrintOnType  = (ptFirstPage, ptLastPage, ptOddPages, ptEvenPages, ptRepeatedBand);
  TfrxPrintOnTypes = set of TfrxPrintOnType;

  /// <summary>
  ///   Kind of the formatting.
  /// </summary>
  TfrxFormatKind = (
    /// <summary>
    ///   Text format.
    /// </summary>
    fkText,
    /// <summary>
    ///   <para>
    ///     Numeric format.
    ///   </para>
    ///   <para>
    ///     %g - a number with the minimal signs number after decimal point; <br />
    ///     %2.2f - a number with the fixed number of signs after decimal
    ///     point; <br />%2.2n - a number with bits delimiter; <br />%2.2m -
    ///     a monetary format, accepted in the Windows OS, depending on the
    ///     regional settings in the control panel.
    ///   </para>
    /// </summary>
    fkNumeric,
    /// <summary>
    ///   <para>
    ///     Date time format.
    ///   </para>
    ///   <para>
    ///     dd.mm.yyyy - date of the 23.12.2003 type; <br />dd mmm yyyy -
    ///     dateof the 23 Nov. 2003 type; <br />dd mmmm yyyy - date of the 23
    ///     November 2003 type; <br />hh:mm - time of the 23:12 type; <br />
    ///     hh:mm:ss - time of the 23:12:00 type; <br />dd mmmm yyyy, hh:mm
    ///     -time and date of the 23 November 2003, 23:12 type
    ///   </para>
    /// </summary>
    fkDateTime,
    /// <summary>
    ///   Boolean format.
    /// </summary>
    fkBoolean);

  TfrxSilentMode = (simMessageBoxes, simSilent, simReThrow);
  /// <summary>
  ///   Flags, used to restrict some object operations in the designer.
  /// </summary>
  TfrxRestriction = (rfDontModify, rfDontSize, rfDontMove, rfDontDelete, rfDontEdit, rfDontEditInPreview, rfDontCopy);
  /// <summary>
  ///   Set of flags, which restrict some object operations in the designer.
  /// </summary>
  TfrxRestrictions = set of TfrxRestriction;
    /// <summary>
  ///   Flags, used to allow some edit object operations in the preview and export.
  /// </summary>
  TfrxEditableRight = (ferAllowInPreview, ferAllowInExport);
   /// <summary>
  ///   Set of flags, which used to allow some edit object operations in the preview and export.
  /// </summary>
  TfrxEditableRights = set of TfrxEditableRight;

  /// <summary>
  ///   Kind of shape.
  /// </summary>
  TfrxShapeKind = (
    /// <summary>
    ///   Rectangle.
    /// </summary>
    skRectangle,
    /// <summary>
    ///   Rectangle with round corners.
    /// </summary>
    skRoundRectangle,
    /// <summary>
    ///   Ellipse.
    /// </summary>
    skEllipse,
    /// <summary>
    ///   Triangle.
    /// </summary>
    skTriangle,
    /// <summary>
    ///   Diamond.
    /// </summary>
    skDiamond,
    /// <summary>
    ///   Diagonal line from bottom left to top right.
    /// </summary>
    skDiagonal1,
    /// <summary>
    ///   Diagonal line from top left to bottom right. <br />
    /// </summary>
    skDiagonal2);

  /// <summary>
  ///   Buttons that can be displayed in the preview window.
  /// </summary>
  TfrxPreviewButton = (
    /// <summary>
    ///   Print button.
    /// </summary>
    pbPrint,
    /// <summary>
    ///   Load.
    /// </summary>
    pbLoad,
    /// <summary>
    ///   Save button.
    /// </summary>
    pbSave,
    /// <summary>
    ///   Export button.
    /// </summary>
    pbExport,
    /// <summary>
    ///   Zoom buttons.
    /// </summary>
    pbZoom,
    /// <summary>
    ///   Find.
    /// </summary>
    pbFind,
    /// <summary>
    ///   Outline button.
    /// </summary>
    pbOutline,
    /// <summary>
    ///   Page setup button.
    /// </summary>
    pbPageSetup,
    /// <summary>
    ///   Deprecated. Not used.
    /// </summary>
    pbTools,
    /// <summary>
    ///   Edit button. Calls report designer for preview page.
    /// </summary>
    pbEdit,
    /// <summary>
    ///   Page navigator.
    /// </summary>
    pbNavigator,
    /// <summary>
    ///   Quick export button.
    /// </summary>
    pbExportQuick,
    /// <summary>
    ///   Disables close button.
    /// </summary>
    pbNoClose,
    /// <summary>
    ///   Disables full screen mode.
    /// </summary>
    pbNoFullScreen,
    /// <summary>
    ///   Disables e-mail export button.
    /// </summary>
    pbNoEmail,
    /// <summary>
    ///   Copy in context menu.
    /// </summary>
    pbCopy,
    /// <summary>
    ///   Paste in context menu.
    /// </summary>
    pbPaste,
    /// <summary>
    ///   Enables selection in the report preview (Shift + left mouse button).
    /// </summary>
    pbSelection,
    /// <summary>
    ///   Allows InPlace editor in the report preview. Enables "Highlight
    ///   editable text" button.
    /// </summary>
    pbInplaceEdit);
  /// <summary>
  ///   Set of buttons that can be displayed in the preview window.
  /// </summary>
  TfrxPreviewButtons = set of TfrxPreviewButton;
  /// <summary>
  ///   Zoom mode of the preview window.
  /// </summary>
  TfrxZoomMode = (
    /// <summary>
    ///   Default mode.
    /// </summary>
    zmDefault,
    /// <summary>
    ///   Fit whole page.
    /// </summary>
    zmWholePage,
    /// <summary>
    ///   Fit page by its width.
    /// </summary>
    zmPageWidth,
    /// <summary>
    ///   Fit several pages.
    /// </summary>
    zmManyPages);
  /// <summary>
  ///   Pages to print.
  /// </summary>
  TfrxPrintPages = (
    /// <summary>
    ///   All pages.
    /// </summary>
    ppAll,
    /// <summary>
    ///   Only odd pages.
    /// </summary>
    ppOdd,
    /// <summary>
    ///   Only even pages.
    /// </summary>
    ppEven);
  /// <summary>
  ///   Sets behaviour of TfrxPreviewPages.AddPage method. <br />
  /// </summary>
  TfrxAddPageAction = (
    /// <summary>
    ///   Write over current page if it isn't last page.
    /// </summary>
    apWriteOver,
    /// <summary>
    ///   Always add new page.
    /// </summary>
    apAdd);
  /// <summary>
  ///   The navigation start point.
  /// </summary>
  TfrxRangeBegin = (
    /// <summary>
    ///   From the current record.
    /// </summary>
    rbFirst,
    /// <summary>
    ///   From the beginning of the data.
    /// </summary>
    rbCurrent);
  /// <summary>
  ///   The endpoint of navigation.
  /// </summary>
  TfrxRangeEnd = (
    /// <summary>
    ///   Till the end of the data
    /// </summary>
    reLast,
    /// <summary>
    ///   Till the current record
    /// </summary>
    reCurrent,
    /// <summary>
    ///   By the number of records set in the "RangeEndCount" property.
    /// </summary>
    reCount);
  /// <summary>
  ///   Field type.
  /// </summary>
  TfrxFieldType = (
    /// <summary>
    ///   Numeric.
    /// </summary>
    fftNumeric,
    /// <summary>
    ///   Text.
    /// </summary>
    fftString,
    /// <summary>
    ///   Boolean.
    /// </summary>
    fftBoolean,
    /// <summary>
    ///   Data time.
    /// </summary>
    fftDateTime);
  /// <summary>
  ///   The kind of action in the OnProgressXXX event: run a report, export,
  ///   print.
  /// </summary>
  TfrxProgressType = (
    /// <summary>
    ///   Report preparing state.
    /// </summary>
    ptRunning,
    /// <summary>
    ///   Export state.
    /// </summary>
    ptExporting,
    /// <summary>
    ///   Printing state.
    /// </summary>
    ptPrinting);
  /// <summary>
  ///   The print mode.
  /// </summary>
  TfrxPrintMode = (
    /// <summary>
    ///   Default mode.
    /// </summary>
    pmDefault,
    /// <summary>
    ///   Split large page on several smaller pages.
    /// </summary>
    pmSplit,
    /// <summary>
    ///   Join several small pages on a large page.
    /// </summary>
    pmJoin,
    /// <summary>
    ///   Print a page on a specified paper size (use scaling).
    /// </summary>
    pmScale);
  /// <summary>
  ///   The inherit mode type.
  /// </summary>
  TfrxInheriteMode = (
    /// <summary>
    ///   Shows inherit dialog.
    /// </summary>
    imDefault,
    /// <summary>
    ///   Delete duplicates.
    /// </summary>
    imDelete,
    /// <summary>
    ///   Rename duplicates
    /// </summary>
    imRename);
  /// <summary>
  ///   Controls serialization state (Not used yet).
  /// </summary>
  TfrxSerializeState = (
    /// <summary>
    ///   Object serializes to a report template.
    /// </summary>
    ssTemplate,
    /// <summary>
    ///   Object serializes to a prepared report.
    /// </summary>
    ssPreviewPages);
  /// <summary>
  ///   Type of hyperlinks supported by report objects.
  /// </summary>
  TfrxHyperlinkKind = (
    /// <summary>
    ///   hyper link to HTTP.
    /// </summary>
    hkURL,
    /// <summary>
    ///   Anchor to position inside prepared report.
    /// </summary>
    hkAnchor,
    /// <summary>
    ///   Link to a page number in prepared report.
    /// </summary>
    hkPageNumber,
    /// <summary>
    ///   Link to detail report.
    /// </summary>
    hkDetailReport,
    /// <summary>
    ///   Reference to detail page.
    /// </summary>
    hkDetailPage,
    /// <summary>
    ///   Reserved.
    /// </summary>
    hkCustom,
    /// <summary>
    ///   Disabled.
    /// </summary>
    hkNone);

  /// <summary>
  ///   Fill types of report objects.
  /// </summary>
  TfrxFillType = (
    /// <summary>
    ///   Brush fill.
    /// </summary>
    ftBrush,
    /// <summary>
    ///   Gradient fill.
    /// </summary>
    ftGradient,
    /// <summary>
    ///   Glass fill.
    /// </summary>
    ftGlass);
  /// <summary>
  ///   Gradient fill styles.
  /// </summary>
  TfrxGradientStyle = (
    /// <summary>
    ///   Horizontal gradient.
    /// </summary>
    gsHorizontal,
    /// <summary>
    ///   Vertical gradient.
    /// </summary>
    gsVertical,
    /// <summary>
    ///   Elliptic gradient.
    /// </summary>
    gsElliptic,
    /// <summary>
    ///   Rectangle gradient.
    /// </summary>
    gsRectangle,
    /// <summary>
    ///   Vertical centered gradient.
    /// </summary>
    gsVertCenter,
    /// <summary>
    ///   Horizontal centered gradient.
    /// </summary>
    gsHorizCenter);
  { preview pages events }
  /// <summary>
  ///   Type of mouse interactive events supported by report objects.
  /// </summary>
  TfrxMouseIntEvents = (
    /// <summary>
    ///   Click on object.
    /// </summary>
    meClick,
    /// <summary>
    ///   Double click on object.
    /// </summary>
    meDbClick,
    /// <summary>
    ///   Mouse moved over object.
    /// </summary>
    meMouseMove,
    /// <summary>
    ///   Mouse up.
    /// </summary>
    meMouseUp,
    /// <summary>
    ///   Mouse down.
    /// </summary>
    meMouseDown,
    /// <summary>
    ///   Drag something over on object.
    /// </summary>
    meDragOver,
    /// <summary>
    ///   Drop something on object.
    /// </summary>
    meDragDrop,
    /// <summary>
    ///   Mouse wheel event.
    /// </summary>
    meMouseWheel);
  /// <summary>
  ///   Type of the report designer tools.
  /// </summary>
  TfrxDesignTool = (
    /// <summary>
    ///   Selection.
    /// </summary>
    dtSelect,
    /// <summary>
    ///   Move hand.
    /// </summary>
    dtHand,
    /// <summary>
    ///   Zoom.
    /// </summary>
    dtZoom,
    /// <summary>
    ///   InPlace text editor.
    /// </summary>
    dtText,
    /// <summary>
    ///   Format copy tool.
    /// </summary>
    dtFormat,
    /// <summary>
    ///   Editor.
    /// </summary>
    dtEditor);
  /// <summary>
  ///   Access type of IO filters.
  /// </summary>
  TfrxFilterAccess = (
    /// <summary>
    ///   Read only.
    /// </summary>
    faRead,
    /// <summary>
    ///   Write.
    /// </summary>
    faWrite);

  /// <summary>
  ///   Flags which is controlling duplicated behaviour.
  /// </summary>
  TfrxDuplicateMerge = (
    /// <summary>
    ///   Show all duplicates. Default.
    /// </summary>
    dmShow,
    /// <summary>
    ///   Hide objects with duplicated values.
    /// </summary>
    dmHide,
    /// <summary>
    ///   Clear all duplicated values.
    /// </summary>
    dmClear,
    /// <summary>
    ///   Merge all duplicated objects to one.
    /// </summary>
    dmMerge);

  /// <summary>
  ///   Controls when to process expression calculation(for some TfrxView
  ///   objects).
  /// </summary>
  TfrxProcessAtMode = (
    /// <summary>
    ///   By Default. When engine processes object.
    /// </summary>
    paDefault,
    /// <summary>
    ///   Processes when report finished. <br />
    /// </summary>
    paReportFinished,
    /// <summary>
    ///   Processes when report page finished. <br />
    /// </summary>
    paReportPageFinished,
    /// <summary>
    ///   Processes when page finished. <br />
    /// </summary>
    paPageFinished,
    /// <summary>
    ///   Processes when column finished.
    /// </summary>
    paColumnFinished,
    /// <summary>
    ///   Processes when data finished. <br />
    /// </summary>
    paDataFinished,
    /// <summary>
    ///   Processes when group finished. <br />
    /// </summary>
    paGroupFinished,
    /// <summary>
    ///   Processes by user from script code.
    /// </summary>
    paCustom);
  /// <summary>
  ///   Grid type in the report designer.
  /// </summary>
  TfrxGridType = (gt1pt, gt1cm, gt1in, gtDialog, gtChar, gtNone);
  /// <summary>
  ///   Copy paste format for inPlace editors.
  /// </summary>
  TfrxCopyPasteType = (
    /// <summary>
    ///   Default.
    /// </summary>
    cptDefault,
    /// <summary>
    ///   Text.
    /// </summary>
    cptText,
    /// <summary>
    ///   Image.
    /// </summary>
    cptImage,
    /// <summary>
    ///   Native control format.
    /// </summary>
    cptNative);

  /// <summary>
  ///   Object anchors type.
  /// </summary>
  TfrxAnchorsKind = (
    /// <summary>
    ///   Left.
    /// </summary>
    fraLeft,
    /// <summary>
    ///   Top.
    /// </summary>
    fraTop,
    /// <summary>
    ///   Right.
    /// </summary>
    fraRight,
    /// <summary>
    ///   Bottom.
    /// </summary>
    fraBottom);
  /// <summary>
  ///   Set of object anchors.
  /// </summary>
  TfrxAnchors = set of TfrxAnchorsKind;

  /// <summary>
  ///   Mirror mode. Determinate how engine align controls and content base on
  ///   original layout.
  /// </summary>
  TfrxMirrorControlMode = (
    /// <summary>
    ///   Mirror bands positions(page columns) from right to left.
    /// </summary>
    mcmRTLBands,
    /// <summary>
    ///   Mirror objects positions from right to left.
    /// </summary>
    mcmRTLObjects,
    /// <summary>
    ///   Mirror appearance from right to left.
    /// </summary>
    mcmRTLAppearance,
    /// <summary>
    ///   Mirror content of objects from right to left.
    /// </summary>
    mcmRTLContent,
    /// <summary>
    ///   Sets RTLReading flag of "Text" object.
    /// </summary>
    mcmRTLSpecial,
    /// <summary>
    ///   Mirror bands positions from Top to Bottom.
    /// </summary>
    mcmBTTBands,
    /// <summary>
    ///   Mirror objects positions from Top to Bottom.
    /// </summary>
    mcmBTTObjects,
    /// <summary>
    ///   Mirror appearance from Top to Bottom.
    /// </summary>
    mcmBTTAppearance,
    /// <summary>
    ///   Mirror content of objects from Top to Bottom.
    /// </summary>
    mcmBTTContent,
    /// <summary>
    ///   Reserved.
    /// </summary>
    mcmBTTSpecial,
    /// <summary>
    ///   Apply mirror mode setting only for objects which have AllowMirrorMode
    ///   property set to true.
    /// </summary>
    mcmOnlyAllowed);
  /// <summary>
  ///   Set of mirror modes.
  /// </summary>
  TfrxMirrorControlModes = set of TfrxMirrorControlMode;

  /// <summary>
  ///   A set of flags, which inhibit specific operations in the designer.
  /// </summary>
  TfrxDesignerRestriction =
    (
    /// <summary>
    ///   Forbids insertion of objects.
    /// </summary>
    drDontInsertObject,
    /// <summary>
    ///   Forbids deletion of pages.
    /// </summary>
    drDontDeletePage,
    /// <summary>
    ///   Forbids creation of new pages.
    /// </summary>
    drDontCreatePage,
    /// <summary>
    ///   Forbids modifying page's properties.
    /// </summary>
    drDontChangePageOptions,
     /// <summary>
     ///   Forbids creation of a new report.
     /// </summary>
     drDontCreateReport,
    /// <summary>
    ///   Forbids report's loading.
    /// </summary>
    drDontLoadReport,
    /// <summary>
    ///   Forbids report's saving.
    /// </summary>
    drDontSaveReport,
     /// <summary>
     ///   Forbids report's preview.
     /// </summary>
     drDontPreviewReport,
    /// <summary>
    ///   Forbids editing of variables.
    /// </summary>
    drDontEditVariables,
    /// <summary>
    ///   Forbids modifying the report's properties.
    /// </summary>
    drDontChangeReportOptions,
     /// <summary>
     ///   Forbids Report|Data... menu.
     /// </summary>
     drDontEditReportData,
    /// <summary>
    ///   Forbids display of recent file list.
    /// </summary>
    drDontShowRecentFiles,
    /// <summary>
    ///   Forbids changes in report script.
    /// </summary>
    drDontEditReportScript,
    /// <summary>
    ///   Forbids changes in internal DataSets(DataPage).
    /// </summary>
    drDontEditInternalDatasets);

  TfrxShowDialogOption =
    (
    /// <summary>
    ///   Export.ShowExportSettings
    /// </summary>
    doShowExportSettings,
    /// <summary>
    ///   Export.ShowSaveDialog
    /// </summary>
    doShowSaveDialog);

  TfrxShowDialogOptions = set of TfrxShowDialogOption;

  //TODO: make enum rect, it should reduce size
  /// <summary>
  ///   Used to pass parameters from preview workspace to preview pages. Stores
  ///   addition information event.
  /// </summary>
  TfrxPreviewIntEventParams = packed record
    /// <summary>
    ///   Interactive event type.
    /// </summary>
    MouseEventType: TfrxMouseIntEvents;
    /// <summary>
    ///   Reference to Sender (Preview control).
    /// </summary>
    Sender: TObject;
    /// <summary>
    ///   Current cursor. PreviewPages, Editors, or report objects may change
    ///   it.
    /// </summary>
    Cursor: TCursor;
    /// <summary>
    ///   Selected tool. Preview supports only dtHand and dtEditor.
    /// </summary>
    EditMode: TfrxDesignTool;
    // DragDrop
    /// <summary>
    ///   State of Drag and Drop event if it was generated.
    /// </summary>
    State: TDragState;
    /// <summary>
    ///   Return parameter for drag and drop event. If True, drag and drop
    ///   accepts.
    /// </summary>
    Accept: Boolean;
    // Mouse Wheel
    /// <summary>
    ///   Mouse wheel delta for mouse wheel events.
    /// </summary>
    WheelDelta: Integer;
    /// <summary>
    ///   Mouse position for mouse wheel events.
    /// </summary>
    MousePos: TPoint;
    /// <summary>
    ///   Return value. If True, event processed by preview pages, otherwise
    ///   default preview behaviour will execute.
    /// </summary>
    RetResult: Boolean;
  end;

  /// <summary>
  ///   Rect with float values.
  /// </summary>
  TfrxRect = TfrRectF;

  /// <summary>
  ///   float point.
  /// </summary>
  TfrxPoint = TfrPointF;

  /// <summary>
  ///   Used with FRX_OWNER_DESTROY_MESSAGE.
  /// </summary>
  TfrxDispatchMessage = record
    MsgID: Word;
    Sender: TObject;
  end;

  /// <summary>
  ///   Class uses in DragDrop events. It helps to determinate when object
  ///   event was called from another object editor. Check
  ///   TfrxInPlaceBandEditor.DoCustomDragDrop.
  /// </summary>
  TfrxEventObject = class(TObject)
  public
    Sender: TObject;
    Index: NativeInt;
  end;

  /// <summary>
  ///   Type of interactive event sender.
  /// </summary>
  TfrxInteractiveEventSender = (
    /// <summary>
    ///   The report designer.
    /// </summary>
    esDesigner,
    /// <summary>
    ///   The report preview.
    /// </summary>
    esPreview);
  /// <summary>
  ///   Used in InPlace editors. Event fires when the report designer or the
  ///   report preview close InPlace editor.
  /// </summary>
  TfrxOnFinishInPlaceEdit = procedure(Sender: TObject; Refresh, Modified: Boolean) of object;

  { remove unnecessary fields later }
  TfrxInteractiveEventsParams = packed record
    Sender: TObject;
    Refresh, Modified: Boolean;
    FireParentEvent: Boolean;
    EventSender: TfrxInteractiveEventSender;
    OnFinish: TfrxOnFinishInPlaceEdit;
    EditMode: TfrxDesignTool;
    PopupVisible: Boolean;
    EditRestricted: Boolean;
    { lets the object decide what editor to chose, not designer }
    EditorsList: TfrxComponentEditorsList;
    SelectionList: TfrxSelectedObjectsList;
    OffsetX: Extended;
    OffsetY: Extended;
    Scale: Extended;
    DevicePPI: Integer;
    GridAlign: Boolean;
    GridType: TfrxGridType;
    GridX: Extended;
    GridY: Extended;
  end;

  /// <summary>
  ///   This event fires when engine loads report XML file and file contains
  ///   XML item &lt;FrxCustomData&gt;(i.e. store user data inside XML file).
  ///   All child nodes in this item ignores by <br />serializer. XMLItem
  ///   contains all the custom user data. <br />
  /// </summary>
  TfrxGetCustomDataEvent = procedure(XMLItem: TfrXMLItem) of object;
  /// <summary>
  ///   This event fires when serializer saves report template to XML(i.e.
  ///   store user data inside XML file). If event is set, serializer creates
  ///   &lt;FrxCustomData&gt; node and pass it as a parameter. Programmer may
  ///   use this node to create child nodes with custom data.
  /// </summary>
  TfrxSaveCustomDataEvent = procedure(XMLItem: TfrXMLItem) of object;
  /// <summary>
  ///   This event is fired when performing some continues operation (print,
  ///   export, run report). The event handler may display a dialog window with
  ///   progress bar. Progress parameter is the page number which is currently
  ///   being processed.
  /// </summary>
  TfrxProgressEvent = procedure(Sender: TfrxReport;
    ProgressType: TfrxProgressType; Progress: Integer) of object;
  /// <summary>
  ///   This event is fired each time before printing an report's object.
  /// </summary>
  TfrxBeforePrintEvent = procedure(Sender: TfrxReportComponent) of object;
  /// <summary>
  ///   This event is fired when an unknown variable (VarName) is met. An event
  ///   handler must return the value of this variable in the Value parameter.
  /// </summary>
  TfrxGetValueEvent = procedure(const VarName: String; var Value: Variant) of object;
  /// <summary>
  ///   This event is fired when an unknown variable (VarName) is met. An event
  ///   handler must return the value of this variable in the Value parameter.
  ///   New event has a Sender parameter and pass a report objects which calls
  ///   an event.
  /// </summary>
  TfrxNewGetValueEvent = procedure(Sender: TObject; const VarName: String; var Value: Variant) of object;
  /// <summary>
  ///   This event is fired when an unknown variable (VarName) is met and
  ///   IsBlobField event returns True(BLOB type). AssignTo parameter contains
  ///   an object in which data should be <br />loaded (TWideStrings for "Text"
  ///   object). <br />
  /// </summary>
  TfrxGetBlobValueEvent = procedure(Sender: TObject; const FieldName: String; AssignTo: TObject) of object;
  /// <summary>
  ///   An event fires when engine checks type of an assigned field. The result
  ///   tells report engine how to interpret the field with name 'FieldName'.
  ///   Returns True if it's a BLOB field. In this case OnGetBlobValue event
  ///   will fire.
  /// </summary>
  TfrxIsBlobFieldEvent = function(Sender: TObject; const FieldName: String): Boolean of object;
  /// <summary>
  ///   This event is fired when calling a function (MethodName), added with
  ///   the help of the TfrxReport.AddFunction method. The event handler should
  ///   return a value of the function.
  /// </summary>
  TfrxUserFunctionEvent = function(const MethodName: String;
    var Params: Variant): Variant of object;
  /// <summary>
  ///   If the handler of this event is allocated, then the FastReport engine
  ///   is blocked and you thus would have to construct a report manually.
  /// </summary>
  TfrxManualBuildEvent = procedure(Page: TfrxPage) of object;
  /// <summary>
  ///   Manual build of the report objects wich support it
  /// </summary>
  TfrxObjectManualBuildEvent = function(ReportObject: TfrxReportComponent): Boolean of object;
  /// <summary>
  ///   Occurs when report is previewed in the preview window and user click on
  ///   object.
  /// </summary>
  TfrxClickObjectEvent = procedure(Sender: TfrxView;
    Button: TMouseButton; Shift: TShiftState; var Modified: Boolean) of object;
  /// <summary>
  ///   Occurs when report is previewed in the preview window and user moves
  ///   the mouse over the object.
  /// </summary>
  TfrxMouseOverObjectEvent = procedure(Sender: TfrxView) of object;
  /// <summary>
  ///   Occurs when report is previewed in the preview window and mouse pointer
  ///   enters object rect area.
  /// </summary>
  TfrxMouseEnterEvent = procedure(Sender: TfrxView) of object;
  /// <summary>
  ///   Occurs when report is previewed in the preview window and mouse pointer
  ///   leaves object rect area.
  /// </summary>
  TfrxMouseLeaveEvent = procedure(Sender: TfrxView) of object;
  /// <summary>
  ///   This event's handler must return the Eof = True parameter, if the end
  ///   of the data set is reached.
  /// </summary>
  TfrxCheckEOFEvent = procedure(Sender: TObject; var Eof: Boolean) of object;
  /// <summary>
  ///   Occurs when report dialog page is previewed.
  /// </summary>
  TfrxRunDialogEvent = procedure(Page: TfrxDialogPage) of object;
  /// <summary>
  ///   Occurs when user edit connection in the report wizard.
  /// </summary>
  TfrxEditConnectionEvent = function(const ConnString: String): String of object;
  /// <summary>
  ///   Occurs when user set ConnectionNeme property. Internal use.
  /// </summary>
  TfrxSetConnectionEvent = procedure(const ConnString: String) of object;
  /// <summary>
  ///   Occurs when the report engine establish connection to database for
  ///   internal database components.
  /// </summary>
  TfrxBeforeConnectEvent = procedure(Sender: TfrxCustomDatabase; var Connected: Boolean) of object;
  /// <summary>
  ///   Occurs when the report engine close connection to database for internal
  ///   database components.
  /// </summary>
  TfrxAfterDisconnectEvent = procedure(Sender: TfrxCustomDatabase) of object;
  /// <summary>
  ///   Occurs when a report is an inherited report. The event handler must
  ///   load the report template (given by TemplateName parameter) into Report
  ///   object.
  /// </summary>
  TfrxPrintPageEvent = procedure(Page: TfrxReportPage; CopyNo: Integer) of object;
  /// <summary>
  ///   Occurs when a report is an inherited report. The event handler must
  ///   load the report template (given by TemplateName parameter) into Report
  ///   object.
  /// </summary>
  TfrxLoadTemplateEvent = procedure(Report: TfrxReport; const TemplateName: String) of object;
  /// <summary>
  ///   Occurs when a report object loads detail report. The event handler must
  ///   load the report template (given by TemplateName parameter) into Report
  ///   object with using parameters of AHyperlink. <br />
  /// </summary>
  TfrxLoadDetailTemplateEvent = function(Report :TfrxReport; const TemplateName :String; const AHyperlink :TfrxHyperlink): Boolean of object;
  /// <summary>
  ///   The event responsible for the sheet name generation template.
  /// </summary>
  TfrxGenerateSheetName = function(PageName: String; PageNumber: Integer): String of object;

{ Root classes }

  /// <summary>
  ///   List notification event.
  /// </summary>
  TfrxObjectsNotifyEvent = procedure(Ptr: Pointer; Action: TListNotification) of object;

  /// <summary>
  ///   Standard list with Notify event. Used as helper to detect added/deleted
  ///   objects.
  /// </summary>
  TfrxObjectsNotifyList = class(TList)
  private
    FOnNotifyList: TfrxObjectsNotifyEvent;
{$IFDEF DELPHI20}
    FDirection: TDirection;
{$ENDIF}
  protected
    procedure Notify(Ptr: Pointer; Action: TListNotification); override;
    function Get(Index: TfrListInteger): Pointer; virtual;
    procedure Put(Index: TfrListInteger; const Value: Pointer); virtual;
    function GetCount: TfrListInteger; virtual;
  public
    property Count: TfrListInteger read GetCount;
{$IFDEF DELPHI20}
    property SearchDirection: TDirection read FDirection write FDirection;
{$ENDIF}
    /// <summary>
    ///   List notification event.
    /// </summary>
    property OnNotifyList: TfrxObjectsNotifyEvent read FOnNotifyList write FOnNotifyList;
    property Items[Index: TfrListInteger]: Pointer read Get write Put; default;
  end;

  TfrxObjectsNotifyListClass = class of TfrxObjectsNotifyList;

  TfrxDataLinkProcessing = (dpString, dpExpression);
  TfrxDataLinkLoadType = (dltOnGetData, dltOnPreview);
  TfrxDataLinkLoading = set of TfrxDataLinkLoadType;

  { Datalink helper class for fast implementation }
  TfrxDataLink = class(TPersistent)
  private
    FLink: String;
    FTempLink: String;
    FProcessingType: TfrxDataLinkProcessing;
    FLoadingType: TfrxDataLinkLoading;
    procedure SetLink(const Value: String);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    function Diff(const PropName: String; ADataLink: TfrxDataLink): String;
    class procedure SaveState(ADataLink: TfrxDataLink);
    class procedure RestoreState(ADataLink: TfrxDataLink);
    class function GetLoadType(LoadMethod: TfrxDataLinkLoadMethod): TfrxDataLinkLoadType;
    class function GetLink(ADatalink: TfrxDataLink; LoadMethod: TfrxDataLinkLoadMethod): String;
    class function IsExpressionLink(ADataLink: TfrxDataLink): Boolean;
    class function IsDataLinkStored(ADataLink: TfrxDataLink; ComponentState: TfrxComponentState): Boolean;
  published
    property Link: String read FLink write SetLink;
    property ProcessingType: TfrxDataLinkProcessing read FProcessingType write FProcessingType default dpString;
    property LoadingType: TfrxDataLinkLoading read FLoadingType write FLoadingType default [dltOnGetData];
  end;

  /// <summary>
  ///   TfrxComponent is the basic class for all FastReport components. Objects
  ///   of this type have such parameters as "coordinates", "size", "font",
  ///   "visibility" and lists of subordinate objects. The class also contains
  ///   methods, which allow saving/restoring object's state to/from the
  ///   stream.
  /// </summary>
  TfrxComponent = class(TfrCustomComponent)
  private
    FFont: TFont;
    FObjects: TfrxObjectsNotifyList;
    FAllObjects: TList;
    FSortedObjects: TStringList;
    FLeft: Extended;
    FTop: Extended;
    FWidth: Extended;
    FHeight: Extended;
    FParentFont: Boolean;
    FGroupIndex: Integer;
    FIsLoading: Boolean;
    FIsPrinting: Boolean;
    FMouseInView: Boolean;
    //FIsSelected: Boolean;
    FRestrictions: TfrxRestrictions;
    FEditable: TfrxEditableRights;
    FVisible: Boolean;
    FDescription: String;
    FComponentStyle: TfrxComponentStyle;
    FfrComponentState: TfrxComponentState;
    FAncestorOnlyStream: Boolean;
    FIndexTag: frInteger;
    FAnchors: TfrxAnchors;
    FAnchorsUpdating: Boolean;
    FAllowMirrorMode: Boolean;
    function GetAbsTop: Extended;
    function GetPage: TfrxPage;
    function GetTopParent: TfrxComponent;
    function IsFontStored: Boolean;
    function GetAllObjects: TList;
    function GetSortedObjects: TStringList;
    function GetAbsLeft: Extended;
    function GetIsLoading: Boolean;
    function GetIsAncestor: Boolean;
    function GetIsSelected: Boolean;
    function GetIsDesigning: Boolean;
    procedure SetIsDesigning(const Value: Boolean);
    function GetIsWriting: Boolean;
    procedure SetIsWriting(const Value: Boolean);
  protected
    FSelectList: TList;
    FParent: TfrxComponent;
    FAliasName: String;
    FBaseName: String;
    FAncestor: Boolean;
    FOriginalComponent: TfrxComponent;
    FOriginalRect: TfrxRect;
    FOriginalBand: TfrxComponent;
    FComponentEditors: TfrxComponentEditorsManager;
    FSerializedItem: TfrXMLItem;
    FHighlighted: Boolean;
    class function GetObjectListClass: TfrxObjectsNotifyListClass; virtual;
    function IsTopStored: Boolean; virtual;
    function IsLeftStored: Boolean; virtual;
    function IsWidthStored: Boolean; virtual;
    function IsHeightStored: Boolean; virtual;
    function IsIndexTagStored: Boolean;
    procedure SetAnchors(const Value: TfrxAnchors); virtual;
    procedure SetIsSelected(const Value: Boolean); virtual;
    procedure SetParent(AParent: TfrxComponent); virtual;
    procedure SetLeft(Value: Extended); virtual;
    procedure SetTop(Value: Extended); virtual;
    procedure SetWidth(Value: Extended); virtual;
    procedure SetHeight(Value: Extended); virtual;
    procedure SetName(const AName: TComponentName); override;
    procedure SetFont(Value: TFont); virtual;
    procedure SetParentFont(const Value: Boolean); virtual;
    procedure SetVisible(Value: Boolean); virtual;
    procedure FontChanged(Sender: TObject); virtual;
    function DiffFont(f1, f2: TFont; const Add: String): String;
    function InternalDiff(AComponent: TfrxComponent): String;
    function GetContainerObjects: TList; virtual;
    function GetRestrictions: TfrxRestrictions; virtual;
    function GetReport: TfrxReport; virtual;
    procedure InintComponentInPlaceEditor(var EventParams: TfrxInteractiveEventsParams);
    { override to detect what was added or deleted from Objects list }
    procedure ObjectListNotify(Ptr: Pointer; Action: TListNotification); virtual;
    procedure LockAnchorsUpdate;
    procedure UnlockAnchorsUpdate;
    procedure ThreadSynchronize(Method: TThreadMethod);

    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
    function GetChildOwner: TComponent; override;

        { interactive object behaviour }
    procedure DoMouseMove(X, Y: Integer; Shift: TShiftState;
      var EventParams: TfrxInteractiveEventsParams); virtual;
    procedure DoMouseUp(X, Y: Integer; Button: TMouseButton;
      Shift: TShiftState;
      var EventParams: TfrxInteractiveEventsParams); virtual;
    function DoMouseDown(X, Y: Integer; Button: TMouseButton;
      Shift: TShiftState; var EventParams: TfrxInteractiveEventsParams)
      : Boolean; virtual;
    procedure DoMouseEnter(aPreviousObject: TfrxComponent;
      var EventParams: TfrxInteractiveEventsParams); virtual;
    procedure DoMouseLeave(aNextObject: TfrxComponent;
      var EventParams: TfrxInteractiveEventsParams); virtual;
    procedure DoMouseClick(Double: Boolean;
      var EventParams: TfrxInteractiveEventsParams); virtual;
    function DoMouseWheel(Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint;
      var EventParams: TfrxInteractiveEventsParams): Boolean; virtual;
    function DoDragOver(Source: TObject; X, Y: Integer; State: TDragState;
      var Accept: Boolean; var EventParams: TfrxInteractiveEventsParams)
      : Boolean; virtual;
    function DoDragDrop(Source: TObject; X, Y: Integer;
      var EventParams: TfrxInteractiveEventsParams): Boolean; virtual;
    procedure MirrorContent(MirrorModes: TfrxMirrorControlModes); virtual;
    procedure DoMirror(MirrorModes: TfrxMirrorControlModes); virtual;
    procedure UpdateAnchors(DeltaX, Deltay: Extended); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    /// <summary>
    ///   Constructor uses by the report designer when create new report
    ///   object.
    /// </summary>
    /// <param name="AOwner">
    ///   Component owner.
    /// </param>
    /// <param name="Flags">
    ///   Addition creation parameter depends on component class.
    /// </param>
    constructor DesignCreate(AOwner: TComponent; Flags: Word); virtual;
    destructor Destroy; override;
    /// <summary>
    ///   Get description of the report component. By default returns class
    ///   name.
    /// </summary>
    class function GetDescription: String; virtual;
    /// <summary>
    ///   Align all child objects relative to the container based on mirror
    ///   control modes and child objects Align's.
    /// </summary>
    /// <param name="IgnoreInvisible">
    ///   Ignore objects which's Visible property set to false.
    /// </param>
    /// <param name="MirrorModes">
    ///   Mirror modes of top container.
    /// </param>
    procedure AlignChildren(IgnoreInvisible: Boolean = False; MirrorModes: TfrxMirrorControlModes = []); virtual;
    /// <summary>
    ///   Copies the state of Source object (ignores all child objects).
    /// </summary>
    procedure Assign(Source: TPersistent); override;
    procedure AssignToDict(Source: TfrxComponent);
    /// <summary>
    ///   Copies the state of Source object with all child objects.
    /// </summary>
    procedure AssignAll(Source: TfrxComponent; Streaming: Boolean = False);
    /// <summary>
    ///   Internal use only. Copies the state of Source object with all child
    ///   objects. Also assign all original components (FOriginalComponent)
    ///   used for prepared report object dictionary.
    /// </summary>
    procedure AssignAllWithOriginals(Source: TfrxComponent; Streaming: Boolean = False);
    procedure AssignOriginals(Source: TfrxComponent); virtual;
    /// <summary>
    ///   Deprecated. Used only for cross tab object for saving nested objects
    ///   to source object dictionary.
    /// </summary>
    procedure AddSourceObjects; virtual;
    /// <summary>
    ///   Report engine call this method before start report preparation. Used
    ///   to prepare objects before build.
    /// </summary>
    procedure BeforeStartReport; virtual;
    /// <summary>
    ///   Creates a unique name for an object placed into the report.
    /// </summary>
    procedure Clear; virtual;
    /// <summary>
    ///   Creates a unique name for an object placed into the report.
    /// </summary>
    procedure CreateUniqueName(DefaultReport: TfrxReport = nil);
    /// <summary>
    ///   Loads object contents and all its child objects from the stream.
    /// </summary>
    procedure LoadFromStream(Stream: TStream); virtual;
    /// <summary>
    ///   Saves an object to the stream. The SaveChildren parameter
    ///   defines,whether state of all child objects should also be saved.
    ///   TheSaveDefaultValues parameter defines, whether to save properties
    ///   thathave default values.
    /// </summary>
    procedure SaveToStream(Stream: TStream; SaveChildren: Boolean = True;
      SaveDefaultValues: Boolean = False; Streaming: Boolean = False); virtual;
    /// <summary>
    ///   Sets coordinates and size of an object.
    /// </summary>
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Extended);
    /// <summary>
    ///   The report engine notifies all objects after show band. Used only for
    ///   Chart object with dtBandData type.
    /// </summary>
    procedure OnNotify(Sender: TObject); virtual;
    /// <summary>
    ///   Object receives message when users paste report object in the report
    ///   designer. Old method.
    /// </summary>
    procedure OnPaste; virtual;
    /// <summary>
    ///   Generates serialized object XML string based on comparison of current
    ///   object with AComponent (Diff between two objects). Used to generate
    ///   object dictionary when csDefaultDiff flag is set. <br />
    /// </summary>
    function AllDiff(AComponent: TfrxComponent): String;
    /// <summary>
    ///   Generates serialized object XML string based on comparison of current
    ///   object with AComponent (Diff between two objects). Light version of
    ///   AllDiff function used to generate object dictionary when
    ///   csDefaultDiff flag is NOT set. <br />
    /// </summary>
    function Diff(AComponent: TfrxComponent): String; virtual;
    /// <summary>
    ///   Searches for an object with specified name among the child objects.
    /// </summary>
    function FindObject(const AName: String): TfrxComponent; virtual;
    /// <summary>
    ///   The serializer calls this method when write nested(composite) non
    ///   published object. Used with csHandlesNestedProperties flag to write
    ///   XML representation of an object instead of binary stream(PropData).
    /// </summary>
    procedure WriteNestedProperties(Item: TfrXMLItem; AAncestor: TPersistent = nil); virtual;
    /// <summary>
    ///   The serializer calls this method when read nested(composite) non
    ///   published object. Used with csHandlesNestedProperties flag to read
    ///   XML representation of an object instead of binary stream(PropData).
    /// </summary>
    function ReadNestedProperty(Item: TfrXMLItem): Boolean; virtual;
    /// <summary>
    ///   Detects if point is inside an object coordinates.
    /// </summary>
    function IsContain(X, Y: Extended): Boolean; overload; virtual;
    /// <summary>
    ///   Returns True if part of object bounds is inside source rectangle.
    /// </summary>
    /// <param name="aRect">
    ///   Source rectangle.
    /// </param>
    function IsContain(X, Y, AScale: Extended): Boolean; overload;
    function IsInRect(const aRect: TfrxRect): Boolean; virtual;
    /// <summary>
    ///   Return client area of object.
    /// </summary>
    function GetClientArea: TfrxRect; virtual;
    /// <summary>
    ///   Returns object used to draw and calculate text bounds. For objects
    ///   with report components it returns object owned by report. otherwise
    ///   global frxDrawText.
    /// </summary>
    function GetDrawTextObject: Pointer;
    /// <summary>
    ///   Method returns topmost report component for prepared report or report
    ///   template (preview pages are not owned by report object and 'Report'
    ///   property returns nil for them). <br />
    /// </summary>
    function GetGlobalReport: TfrxReport;
    /// <summary>
    ///   Returns top-most object which contain point at passed coordinates
    /// </summary>
    /// <param name="X">
    ///   X coordinate.
    /// </param>
    /// <param name="Y">
    ///   Y coordinate.
    /// </param>
    /// <param name="IsCanContain">
    ///   Child Component. If IsCanContain = nil - returns top-most object,
    ///   otherwise IsCanContain &lt;&gt; nil - return top-most object which
    ///   can accept IsCanContain component as child.
    /// </param>
    /// <returns>
    ///   Found component or nil.
    /// </returns>
    function GetContainedComponent(X, Y: Extended; IsCanContain: TfrxComponent = nil): TfrxComponent; virtual;
    /// <summary>
    ///   Returns objects inside rectangle area with class type inherit from
    ///   specified class type.
    /// </summary>
    /// <param name="aRect">
    ///   Rectangle area.
    /// </param>
    /// <param name="InheriteFrom">
    ///   Class of objects which should be selected by function.
    /// </param>
    /// <param name="aComponents">
    ///   Result List of components that suits function parameters.
    /// </param>
    /// <param name="bSelectContainers">
    ///   Include containers into selection result.
    /// </param>
    procedure GetContainedComponents(const aRect: TfrxRect; InheriteFrom: TClass; aComponents: TList; bSelectContainers: Boolean = False); virtual;
    /// <summary>
    ///   Export engine calls this method before process object inside export
    ///   filter. If this method returns true, object processed by itself and
    ///   default method will not call.
    /// </summary>
    function ExportInternal(Filter: TfrxCustomExportFilter): Boolean; virtual;

    { control behaviour for detection childs which it can holds }
    /// <summary>
    ///   Returns true when component can accept component in aControl as
    ///   child.
    /// </summary>
    /// <param name="aControl">
    ///   Component that supposedly should become a child.
    /// </param>
    function IsAcceptControl(aControl: TfrxComponent): Boolean; virtual;
    /// <summary>
    ///   Returns true when component in aParent can be accepted as parent for
    ///   current. <br />
    /// </summary>
    /// <param name="aParent">
    ///   Component that supposedly should become a parent for current.
    /// </param>
    function IsAcceptAsChild(aParent: TfrxComponent): Boolean; virtual;
    { general function just quick check of flag }
    function IsAcceptControls: Boolean; virtual;
    { events trigger }
    /// <summary>
    ///   Generates mouse move event for report object.
    /// </summary>
    /// <param name="X">
    ///   X coordinate.
    /// </param>
    /// <param name="Y">
    ///   Y coordinate. <br />
    /// </param>
    /// <param name="Shift">
    ///   Shift keys state.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with additional parameters.
    /// </param>
    procedure MouseMove(X, Y: Integer; Shift: TShiftState;
      var EventParams: TfrxInteractiveEventsParams);
    /// <summary>
    ///   Generates mouse up event for report object.
    /// </summary>
    /// <param name="X">
    ///   X coordinate.
    /// </param>
    /// <param name="Y">
    ///   Y coordinate.
    /// </param>
    /// <param name="Button">
    ///   Mouse button.
    /// </param>
    /// <param name="Shift">
    ///   Shift keys state.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with additional parameters.
    /// </param>
    procedure MouseUp(X, Y: Integer; Button: TMouseButton;
      Shift: TShiftState;
      var EventParams: TfrxInteractiveEventsParams);
    /// <summary>
    ///   Generates mouse down event for report object.
    /// </summary>
    /// <param name="X">
    ///   X coordinate.
    /// </param>
    /// <param name="Y">
    ///   Y coordinate.
    /// </param>
    /// <param name="Button">
    ///   Mouse button.
    /// </param>
    /// <param name="Shift">
    ///   Shift keys state.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with additional parameters.
    /// </param>
    function MouseDown(X, Y: Integer; Button: TMouseButton;
      Shift: TShiftState; var EventParams: TfrxInteractiveEventsParams)
      : Boolean;
    /// <summary>
    ///   Generates mouse enter event for report object.
    /// </summary>
    /// <param name="aPreviousObject">
    ///   Reference to previous object which was in focus(Can be nil).
    /// </param>
    /// <param name="EventParams">
    ///   Structure with additional parameters.
    /// </param>
    procedure MouseEnter(aPreviousObject: TfrxComponent;
      var EventParams: TfrxInteractiveEventsParams);
    /// <summary>
    ///   Generates mouse leave event for report object.
    /// </summary>
    /// <param name="X">
    ///   X coordinate.
    /// </param>
    /// <param name="Y">
    ///   Y coordinate.
    /// </param>
    /// <param name="aNextObject">
    ///   Reference to an object which should receive focus(Can be nil).
    /// </param>
    /// <param name="EventParams">
    ///   Structure with additional parameters.
    /// </param>
    procedure MouseLeave(X, Y: Integer; aNextObject: TfrxComponent;
      var EventParams: TfrxInteractiveEventsParams);
    /// <summary>
    ///   Generates mouse click event for report object.
    /// </summary>
    /// <param name="Double">
    ///   If True, it's a double click event.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with additional parameters.
    /// </param>
    procedure MouseClick(Double: Boolean;
      var EventParams: TfrxInteractiveEventsParams);
    /// <summary>
    ///   Generates mouse wheel event for report object.
    /// </summary>
    /// <param name="Shift">
    ///   Shift keys state.
    /// </param>
    /// <param name="WheelDelta">
    ///   Delta of mouse wheel.
    /// </param>
    /// <param name="MousePos">
    ///   Position of mouse cursor.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with additional parameters.
    /// </param>
    function MouseWheel(Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint;
      var EventParams: TfrxInteractiveEventsParams): Boolean;
    /// <summary>
    ///   Generates drag over event for report object when drag and drop
    ///   initialized.
    /// </summary>
    /// <param name="Source">
    ///   The object being dragged.
    /// </param>
    /// <param name="X">
    ///   X coordinate.
    /// </param>
    /// <param name="Y">
    ///   Y coordinate. <br />
    /// </param>
    /// <param name="State">
    ///   Indicates how the dragged object is moving in relation to the
    ///   control.
    /// </param>
    /// <param name="Accept">
    ///   Returns true when object can accept dragged object.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with additional parameters.
    /// </param>
    function DragOver(Source: TObject; X, Y: Integer; State: TDragState;
      var Accept: Boolean; var EventParams: TfrxInteractiveEventsParams)
      : Boolean;
    /// <summary>
    ///   Generates drag drop event for report object when drag and drop
    ///   initialized.
    /// </summary>
    /// <param name="Source">
    ///   The object being dropped.
    /// </param>
    /// <param name="X">
    ///   X coordinate
    /// </param>
    /// <param name="Y">
    ///   Y coordinate
    /// </param>
    /// <param name="EventParams">
    ///   Structure with additional parameters.
    /// </param>
    function DragDrop(Source: TObject; X, Y: Integer;
      var EventParams: TfrxInteractiveEventsParams): Boolean;

//    procedure FireEvent(Sender: TObject; X, Y: Integer; Button: TMouseButton; Shift: TShiftState; Scale, OffsetX, OffsetY: Extended; var aEventParams: TfrxPreviewIntEventParams);
    /// <summary>
    ///   Used to validate object real bounds.
    /// </summary>
    procedure UpdateBounds; virtual;
    /// <summary>
    ///   Deprecated. Used only for cross tab component. Adds internal child
    ///   objects for cross cells.
    /// </summary>
    /// <param name="Obj">
    ///   Report object.
    /// </param>
    function ContainerAdd(Obj: TfrxComponent): Boolean; virtual;
    /// <summary>
    ///   Looking for data set with filename specified. The result of search
    ///   depends on flag state Tfrxreport Engine options .Use Grobal data set
    ///   list. If the flag is installed in True ,then search is realized in
    ///   Grobal data set list , otherwuise in TfrxReport list . Enabled data
    ///   sets.
    /// </summary>
    /// <param name="DataSet">
    ///   Dataset.
    /// </param>
    /// <param name="DSName">
    ///   Dataset name.
    /// </param>
    function FindDataSet(DataSet: TfrxDataSet; const DSName: String): TfrxDataSet;
    /// <summary>
    ///   Used with TfrxMirrorControlModes when mcmOnlyAllowed set on parent
    ///   container. It marks report object follows TfrxMirrorControlModes
    ///   rules.
    /// </summary>
    property AllowMirrorMode: Boolean read FAllowMirrorMode write FAllowMirrorMode default False;
    /// <summary>
    ///   Anchors relative to parent container (Default left, top).
    /// </summary>
    property Anchors: TfrxAnchors read FAnchors write SetAnchors default [fraLeft, fraTop];
    /// <summary>
    ///   Used when save object to XML. If this property set to true only
    ///   values of ancestor object will be serialized.
    /// </summary>
    property AncestorOnlyStream: Boolean read FAncestorOnlyStream write FAncestorOnlyStream;
    /// <summary>
    ///   The list of child objects.
    /// </summary>
    property Objects: TfrxObjectsNotifyList read FObjects;
    /// <summary>
    ///   The list of all subordinate objects.
    /// </summary>
    property AllObjects: TList read GetAllObjects;
    /// <summary>
    ///   Deprecated. Used to support old cross tab behaviour.
    /// </summary>
    property ContainerObjects: TList read GetContainerObjects;
    /// <summary>
    ///   Link to the parent object.
    /// </summary>
    property Parent: TfrxComponent read FParent write SetParent;
    /// <summary>
    ///   Link to the report page, which the object belongs to.
    /// </summary>
    property Page: TfrxPage read GetPage;
    /// <summary>
    ///   Link to the report, which the object belongs to.
    /// </summary>
    property Report: TfrxReport read GetReport;
    /// <summary>
    ///   Returns topmost parent of the object.
    /// </summary>
    property TopParent: TfrxComponent read GetTopParent;
    /// <summary>
    ///   Returns True when object has an ancestor.
    /// </summary>
    property IsAncestor: Boolean read GetIsAncestor;
    /// <summary>
    ///   The object in designing mode.
    /// </summary>
    property IsDesigning: Boolean read GetIsDesigning write SetIsDesigning;
    /// <summary>
    ///   Object is loading.
    /// </summary>
    property IsLoading: Boolean read GetIsLoading write FIsLoading;
    /// <summary>
    ///   Object is printing.
    /// </summary>
    property IsPrinting: Boolean read FIsPrinting write FIsPrinting;
    /// <summary>
    ///   Object is in serialization state.
    /// </summary>
    property IsWriting: Boolean read GetIsWriting write SetIsWriting;
    /// <summary>
    ///   Object selected in the report designer or report preview.
    /// </summary>
    property IsSelected: Boolean read GetIsSelected write SetIsSelected;
    /// <summary>
    ///   Base name of the report object. Used for generation of unique names.
    /// </summary>
    property BaseName: String read FBaseName;
    property GroupIndex: Integer read FGroupIndex write FGroupIndex default 0;
    /// <summary>
    ///   Report object component styles.
    /// </summary>
    property frComponentStyle: TfrxComponentStyle read FComponentStyle write FComponentStyle;
    property frComponentState: TfrxComponentState read FfrComponentState write FfrComponentState;
    /// <summary>
    ///   Object's X coordinate (relatively to a parent, in pixels).
    /// </summary>
    property Left: Extended read FLeft write SetLeft stored IsLeftStored;
    /// <summary>
    ///   Object's Y coordinate (relatively to a parent, in pixels).
    /// </summary>
    property Top: Extended read FTop write SetTop stored IsTopStored;
    /// <summary>
    ///   Object's width, in pixels.
    /// </summary>
    property Width: Extended read FWidth write SetWidth stored IsWidthStored;
    /// <summary>
    ///   Object's height, in pixels.
    /// </summary>
    property Height: Extended read FHeight write SetHeight stored IsHeightStored;
    /// <summary>
    ///   Object's X absolute coordinate, in pixels.
    /// </summary>
    property AbsLeft: Extended read GetAbsLeft;
    /// <summary>
    ///   Object's Y absolute coordinate, in pixels.
    /// </summary>
    property AbsTop: Extended read GetAbsTop;

    /// <summary>
    ///   Object's description.
    /// </summary>
    property Description: String read FDescription write FDescription;
    /// <summary>
    ///   Allows edit objects in the preview and export(some export filters).
    /// </summary>
    property Editable: TfrxEditableRights read FEditable write FEditable default [];
    /// <summary>
    ///   If "True", then uses the parent object font settings.
    /// </summary>
    property ParentFont: Boolean read FParentFont write SetParentFont default True;
    /// <summary>
    ///   Set of flags, which restrict some object operations.
    /// </summary>
    property Restrictions: TfrxRestrictions read GetRestrictions write FRestrictions default [];
    /// <summary>
    ///   Global visibility of the object in the result report.
    /// </summary>
    property Visible: Boolean read FVisible write SetVisible default True;
    /// <summary>
    ///   Object's font.
    /// </summary>
    property Font: TFont read FFont write SetFont stored IsFontStored;
    /// <summary>
    ///   Not used.
    /// </summary>
    property IndexTag: frInteger read FIndexTag write FIndexTag stored IsIndexTagStored;
  end;

  IfrxAggregateObject = interface
  ['{63ABB456-F4C9-49DA-ADA2-E3ABC7125B15}']
    function GetParentContainer: TfrxComponent;
    function GetDataRowContainer: TfrxComponent;
    function GetDataContainers: TList;
    function GetExpression: String;
    function GetExpressionDelimiters: String;
    function GetInstance: TfrxComponent;
  end;

  /// <summary>
  ///   Objects of this type can be placed into a report. The class contains
  ///   the "Draw" method for object's painting, as well as
  ///   "BeforePrint/GetData/AfterPrint" methods, which are called as soon as a
  ///   report runs.
  /// </summary>
  TfrxReportComponent = class(TfrxComponent)
  private
    FOnAfterData: TfrxNotifyEvent;
    FOnAfterPrint: TfrxNotifyEvent;
    FOnBeforePrint: TfrxNotifyEvent;
    FOnPreviewClick: TfrxPreviewClickEvent;
    FOnPreviewDblClick: TfrxPreviewClickEvent;
    FOnContentChanged: TfrxContentChangedEvent;
    FPrintOn: TfrxPrintOnTypes;
  protected
    /// <summary>
    ///   Draws size boxes around report objects in the report designer.
    /// </summary>
    procedure DrawSizeBox(aCanvas: TCanvas; aScale: Extended; bForceDraw: Boolean = False); virtual;
  public
    { TODO : hide this Stuff !}
    /// <summary>
    ///   Used in new shift mechanism, stores shift tree or list. Internal use
    ///   only.
    /// </summary>
    FShiftObject: TObject;
    /// <summary>
    ///   Used for TfrxSubReport.PrintOnParent. Internal use only.
    /// </summary>
    FOriginalObjectsCount: Integer;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    /// <summary>
    ///   Used from the report designer and preview calls Draw/DrawHighlight. <br />
    /// </summary>
    /// <param name="Canvas">
    ///   Canvas to draw object.
    /// </param>
    /// <param name="ScaleX">
    ///   X Scale factor.
    /// </param>
    /// <param name="ScaleY">
    ///   Y Scale factor. <br />
    /// </param>
    /// <param name="OffsetX">
    ///   X Offset.
    /// </param>
    /// <param name="OffsetY">
    ///   Y Offset.
    /// </param>
    /// <param name="Highlighted">
    ///   Force DrawHighlight call.
    /// </param>
    /// <param name="hlColor">
    ///   Highlight color.
    /// </param>
    procedure InteractiveDraw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended; Highlighted: Boolean = False; hlColor: TColor = clNone);
    { back compat }
    /// <summary>
    ///   The Draw method is called when painting an object. <br />
    /// </summary>
    /// <param name="Canvas">
    ///   Canvas
    /// </param>
    /// <param name="ScaleX">
    ///   Zoom by X-axis.
    /// </param>
    /// <param name="ScaleY">
    ///   Zoom by Y-axis.
    /// </param>
    /// <param name="OffsetX">
    ///   X offset relatively to the edges of the canvas.
    /// </param>
    /// <param name="OffsetY">
    ///   Y offset relatively to the edges of the canvas. <br />
    /// </param>
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended);
      virtual; abstract;
    /// <summary>
    ///   Reserved. Not implemented.
    /// </summary>
    procedure DrawChilds(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended; Highlighted: Boolean; IsDesigning: Boolean);
      virtual;
    /// <summary>
    ///   Reserved. Not implemented.
    /// </summary>
    procedure DrawWithChilds(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended; Highlighted: Boolean; IsDesigning: Boolean);
    /// <summary>
    ///   Draws selection highlight.
    /// </summary>
    /// <param name="Canvas">
    ///   Canvas to draw.
    /// </param>
    /// <param name="ScaleX">
    ///   X scale factor.
    /// </param>
    /// <param name="ScaleY">
    ///   Y scale factor. <br />
    /// </param>
    /// <param name="OffsetX">
    ///   X offset.
    /// </param>
    /// <param name="OffsetY">
    ///   X offset.
    /// </param>
    /// <param name="hlColor">
    ///   Highlight color.
    /// </param>
    procedure DrawHighlight(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended; hlColor: TColor = clNone); virtual;
    /// <summary>
    ///   The report engine calls this method before prepare object and place
    ///   it to preview pages.
    /// </summary>
    procedure BeforePrint; virtual;
    /// <summary>
    ///   The report engine calls this method to get Data inside object or
    ///   calculate expressions.
    /// </summary>
    procedure GetData; virtual;
    /// <summary>
    ///   The report engine calls this method after object was prepared and
    ///   placed to preview pages.
    /// </summary>
    procedure AfterPrint; virtual;
    /// <summary>
    ///   Returns text of the report object. Used in export filters.
    /// </summary>
    function GetComponentText: String; virtual;
    /// <summary>
    ///   Deprecated. Returns real size of object.
    /// </summary>
    function GetRealBounds: TfrxRect; virtual;
    /// <summary>
    ///   Returns component with part of the data of the current (by default
    ///   returns Self). Used in split mechanism when one object breaks to
    ///   several parts. The component should be prepared by DrawPart method.
    ///   Engine adds this object to preview pages.
    /// </summary>
    function GetSaveToComponent: TfrxReportComponent; virtual;
    /// <summary>
    ///   Returns True when parent container should draw child.
    /// </summary>
    function IsOwnerDraw: Boolean; virtual;
    /// <summary>
    ///   The name of the script routine that will be called after the object
    ///   gets data.
    /// </summary>
    property OnAfterData: TfrxNotifyEvent read FOnAfterData write FOnAfterData;
    /// <summary>
    ///   The name of the script routine that will be called after the object
    ///   is handled.
    /// </summary>
    property OnAfterPrint: TfrxNotifyEvent read FOnAfterPrint write FOnAfterPrint;
    /// <summary>
    ///   The name of the script routine that will be called before the object
    ///   is handled.
    /// </summary>
    property OnBeforePrint: TfrxNotifyEvent read FOnBeforePrint write FOnBeforePrint;
    /// <summary>
    ///   The name of the script routine that will be called if user clicks on
    ///   the object in the preview window.
    /// </summary>
    property OnPreviewClick: TfrxPreviewClickEvent read FOnPreviewClick write FOnPreviewClick;
    /// <summary>
    ///   The name of the script routine that will be called if user double
    ///   clicks on the object in the preview window.
    /// </summary>
    property OnPreviewDblClick: TfrxPreviewClickEvent read FOnPreviewDblClick write FOnPreviewDblClick;
    /// <summary>
    ///   The name of the script routine that will be called if content of
    ///   object was changes(for example text in "Text" object).
    /// </summary>
    property OnContentChanged: TfrxContentChangedEvent read FOnContentChanged write FOnContentChanged;
    // todo
    /// <summary>
    ///   Reserved for future use.
    /// </summary>
    property PrintOn: TfrxPrintOnTypes read FPrintOn write FPrintOn;
  published
    property Description;
    property IndexTag;
  end;

  /// <summary>
  ///   The TfrxDialogComponent class is the basic one for writing non-visual
  ///   components, which can be placed on a dialog form in a report.
  /// </summary>
  TfrxDialogComponent = class(TfrxReportComponent)
  const frPropLeft = 'pLeft';
        frPropTop = 'pTop';
  private
    FComponent: TComponent;
    procedure ReadLeft(Reader: TReader);
    procedure ReadTop(Reader: TReader);
    procedure WriteLeft(Writer: TWriter);
    procedure WriteTop(Writer: TWriter);
    procedure WriteDesignLeft(Writer: TWriter);
    procedure WriteDesignTop(Writer: TWriter);
    procedure UpdateDesignInfo;
  protected
    FImageIndex: Integer;
    procedure DefineProperties(Filer: TFiler); override;
    procedure SetLeft(Value: Extended); override;
    procedure SetTop(Value: Extended); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;
    /// <summary>
    ///   Link to the component itself.
    /// </summary>
    property Component: TComponent read FComponent write FComponent;
  end;

  { TfrxDialogControl }

  /// <summary>
  ///   The "TfrxDialogControl" class is the basic one for writing common
  ///   control, which can be placed on a dialogue form in a report. The class
  ///   contains a large number of general properties and events shared with
  ///   most of the common controls.
  /// </summary>
  TfrxDialogControl = class(TfrxReportComponent)
  private
    FControl: TControl;
    FOnClick: TfrxNotifyEvent;
    FOnDblClick: TfrxNotifyEvent;
    FOnEnter: TfrxNotifyEvent;
    FOnExit: TfrxNotifyEvent;
    FOnKeyDown: TfrxKeyEvent;
    FOnKeyPress: TfrxKeyPressEvent;
    FOnKeyUp: TfrxKeyEvent;
    FOnMouseDown: TfrxMouseEvent;
    FOnMouseMove: TfrxMouseMoveEvent;
    FOnMouseUp: TfrxMouseEvent;
    FOnActivate: TNotifyEvent;
    function GetColor: TColor;
    function GetEnabled: Boolean;
    procedure DoOnClick(Sender: TObject);
    procedure DoOnDblClick(Sender: TObject);
    procedure DoOnEnter(Sender: TObject);
    procedure DoOnExit(Sender: TObject);
    procedure DoOnKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure DoOnKeyPress(Sender: TObject; var Key: Char);
    procedure DoOnKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure DoOnMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure DoOnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure DoOnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure DoOnMouseWheel(Sender: TObject; Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure SetColor(const Value: TColor);
    procedure SetEnabled(const Value: Boolean);
    function GetCaption: String;
    procedure SetCaption(const Value: String);
    function GetHint: String;
    procedure SetHint(const Value: String);
    function GetShowHint: Boolean;
    procedure SetShowHint(const Value: Boolean);
    function GetTabStop: Boolean;
    procedure SetTabStop(const Value: Boolean);
    procedure DoDestroyControl;
    procedure DoSetParent;
    procedure DoInitControl;
  protected
    FCurrentPPI: Integer;
    procedure CorrectControlCoordinates; virtual;
    function GetCoordinateOffset: TPoint; virtual;
    function GetScaledControlHeight: Integer;
    function GetScaledControlLeft: Integer;
    function GetScaledControlTop: Integer;
    function GetScaledControlWidth: Integer;
    procedure SetAnchors(const Value: TfrxAnchors); override;
    procedure SetLeft(Value: Extended); override;
    procedure SetTop(Value: Extended); override;
    procedure SetWidth(Value: Extended); override;
    procedure SetHeight(Value: Extended); override;
    procedure SetParentFont(const Value: Boolean); override;
    procedure SetVisible(Value: Boolean); override;
    procedure SetParent(AParent: TfrxComponent); override;
    procedure FontChanged(Sender: TObject); override;
    procedure InitControl(AControl: TControl);
    procedure SetName(const AName: TComponentName); override;
    procedure UpdateControlPPI(aNewPPI: Integer); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;
    function IsOwnerDraw: Boolean; override;
    function IsAcceptControls: Boolean; override;
    function IsAcceptAsChild(aParent: TfrxComponent): Boolean; override;
    /// <summary>
    ///   Caption of the control.
    /// </summary>
    property Caption: String read GetCaption write SetCaption;
    /// <summary>
    ///   Color of the control.
    /// </summary>
    property Color: TColor read GetColor write SetColor;
    /// <summary>
    ///   Link to the control itself.
    /// </summary>
    property Control: TControl read FControl write FControl;
    /// <summary>
    ///   Determines if the user can tab to a control.
    /// </summary>
    property TabStop: Boolean read GetTabStop write SetTabStop default True;
    /// <summary>
    ///   The name of the script routine that will be called if user clicks the
    ///   control.
    /// </summary>
    property OnClick: TfrxNotifyEvent read FOnClick write FOnClick;
    /// <summary>
    ///   The name of the script routine that will be called if user double
    ///   clicks the control.
    /// </summary>
    property OnDblClick: TfrxNotifyEvent read FOnDblClick write FOnDblClick;
    /// <summary>
    ///   The name of the script routine that will be called if control gets
    ///   the focus.
    /// </summary>
    property OnEnter: TfrxNotifyEvent read FOnEnter write FOnEnter;
    /// <summary>
    ///   The name of the script routine that will be called if user pressed a
    ///   key.
    /// </summary>
    property OnExit: TfrxNotifyEvent read FOnExit write FOnExit;
    /// <summary>
    ///   The name of the script routine that will be called if user pressed a
    ///   key.
    /// </summary>
    property OnKeyDown: TfrxKeyEvent read FOnKeyDown write FOnKeyDown;
    /// <summary>
    ///   The name of the script routine that will be called if user pressed a
    ///   key.
    /// </summary>
    property OnKeyPress: TfrxKeyPressEvent read FOnKeyPress write FOnKeyPress;
    /// <summary>
    ///   The name of the script routine that will be called if user released a
    ///   key.
    /// </summary>
    property OnKeyUp: TfrxKeyEvent read FOnKeyUp write FOnKeyUp;
    /// <summary>
    ///   The name of the script routine that will be called if user pressed a
    ///   mouse button.
    /// </summary>
    property OnMouseDown: TfrxMouseEvent read FOnMouseDown write FOnMouseDown;
    /// <summary>
    ///   The name of the script routine that will be called if user moved a
    ///   mouse over control.
    /// </summary>
    property OnMouseMove: TfrxMouseMoveEvent read FOnMouseMove write FOnMouseMove;
    /// <summary>
    ///   The name of the script routine that will be called if user released a
    ///   mouse button.
    /// </summary>
    property OnMouseUp: TfrxMouseEvent read FOnMouseUp write FOnMouseUp;
    /// <summary>
    ///   The name of the script routine that will be called if user activate
    ///   control.
    /// </summary>
    property OnActivate: TNotifyEvent read FOnActivate write FOnActivate;
  published
    property Anchors;
    property Left;
    property Top;
    property Width;
    property Height;
    property Font;
    property GroupIndex;
    property ParentFont;
    /// <summary>
    ///   Determines if the control is enabled.
    /// </summary>
    property Enabled: Boolean read GetEnabled write SetEnabled default True;
    /// <summary>
    ///   Contains the text string that can appear when the user moves the
    ///   mouse over the control.
    /// </summary>
    property Hint: String read GetHint write SetHint;
    /// <summary>
    ///   Show/hide hint.
    /// </summary>
    property ShowHint: Boolean read GetShowHint write SetShowHint;
    property Visible;
  end;

  /// <summary>
  ///   The TfrxDataSet class is used for navigation of a dataset. It is base
  ///   class for TfrxDBDataSet, TfrxUserDataSet components.
  /// </summary>
  TfrxDataSet = class(TfrxDialogComponent)
  private
    FCloseDataSource: Boolean;
    FEnabled: Boolean;
    FEof: Boolean;
    FOpenDataSource: Boolean;
    FRangeBegin: TfrxRangeBegin;
    FRangeEnd: TfrxRangeEnd;
    FRangeEndCount: Integer;
    FReportRef: TfrxReport;
    FUserName: String;
    FOnCheckEOF: TfrxCheckEOFEvent;
    FOnFirst: TNotifyEvent;
    FOnNext: TNotifyEvent;
    FOnPrior: TNotifyEvent;
    FOnOpen: TNotifyEvent;
    FOnClose: TNotifyEvent;
  protected
    FInitialized: Boolean;
    FRecNo: Integer;
    // TODO const String
    function GetDisplayText(Index: String): WideString; virtual;
    function GetDisplayWidth(Index: String): Integer; virtual;
    function GetFieldType(Index: String): TfrxFieldType; virtual;
    function GetValue(Index: String): Variant; virtual;
    procedure GetSourceFieldList(List: TStrings); virtual;
    procedure SetName(const NewName: TComponentName); override;
    procedure SetUserName(const Value: String); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    { Navigation methods }
    /// <summary>
    ///   Initializes DataSet before open.
    /// </summary>
    procedure Initialize; virtual;
    /// <summary>
    ///   Finalizes DataSet after close.
    /// </summary>
    procedure Finalize; virtual;
    /// <summary>
    ///   Opens the dataset.
    /// </summary>
    procedure Open; virtual;
    /// <summary>
    ///   Closes the dataset.
    /// </summary>
    procedure Close; virtual;
    /// <summary>
    ///   Sets the cursor to the first record.
    /// </summary>
    procedure First; virtual;
    /// <summary>
    ///   Moves the cursor to the next record.
    /// </summary>
    procedure Next; virtual;
    /// <summary>
    ///   Moves the cursor to the previous record. This method is used only
    ///   when printing groups.
    /// </summary>
    procedure Prior; virtual;
    /// <summary>
    ///   Function returns True if end of the dataset is reached.
    /// </summary>
    function Eof: Boolean; virtual;

    { Data access }
    /// <summary>
    ///   Returns the number of fields defined in the dataset.
    /// </summary>
    function FieldsCount: Integer; virtual;
    /// <summary>
    ///   Returns True if dataset has field with specified name.
    /// </summary>
    function HasField(const fName: String): Boolean;
    /// <summary>
    ///   Returns True if dataset has BLOB field with specified name.
    /// </summary>
    function IsBlobField(const fName: String): Boolean; virtual;
    /// <summary>
    ///   Returns True if dataset has wide memo BLOB field with specified name.
    /// </summary>
    function IsWideMemoBlobField(const fName: String): Boolean; virtual;
    /// <summary>
    ///   Returns True if dataset has memo BLOB field with specified name.
    /// </summary>
    function IsMemoBlobField(const fName: String): Boolean; virtual;
    /// <summary>
    ///   Returns True if DataSet has master RecordSet.
    /// </summary>
    function IsHasMaster: Boolean; virtual;
    /// <summary>
    ///   Returns records count if DataSet supports it.
    /// </summary>
    function RecordCount: Integer; virtual;
    /// <summary>
    ///   Assigns stream in Obj to BLOB field with fName name.
    /// </summary>
    /// <param name="fName">
    ///   Name of the BLOB field.
    /// </param>
    /// <param name="Obj">
    ///   Object serialized to blob (supports TStream successors). <br />
    /// </param>
    procedure AssignBlobTo(const fName: String; Obj: TObject); virtual;
    /// <summary>
    ///   Returns list of dataset fields.
    /// </summary>
    procedure GetFieldList(List: TStrings); virtual;
    procedure GetFieldDefsList(List: TStrings); virtual;
    procedure UpdateFieldDefs(bForceUpdate: Boolean = False; AFromList: TStrings = nil); virtual;
    procedure UpdateFieldTypes; virtual;
    procedure ResetFieldDefs; virtual;
    function CreateFieldBindComponent(const Index: String; AOwner: TfrxComponent): TfrxComponent; virtual;
    /// <summary>
    ///   Returns the field value as a string.
    /// </summary>
    property DisplayText[Index: String]: WideString read GetDisplayText;
    /// <summary>
    ///   Returns the size of a specified field.
    /// </summary>
    property DisplayWidth[Index: String]: Integer read GetDisplayWidth;
    /// <summary>
    ///   Returns the field type
    /// </summary>
    property FieldType[Index: String]: TfrxFieldType read GetFieldType;
    /// <summary>
    ///   Returns the field value as a variant.
    /// </summary>
    property Value[Index: String]: Variant read GetValue;

    /// <summary>
    ///   Close a data set after report construction is completed.
    /// </summary>
    property CloseDataSource: Boolean read FCloseDataSource write FCloseDataSource;
    { OpenDataSource is kept for backward compatibility only }
    /// <summary>
    ///   Open a data set before report construction is started.
    /// </summary>
    property OpenDataSource: Boolean read FOpenDataSource write FOpenDataSource default True;
    /// <summary>
    ///   A current record number. The first record's number is "0".
    /// </summary>
    property RecNo: Integer read FRecNo;
    /// <summary>
    ///   Reference for the report owner if any.
    /// </summary>
    property ReportRef: TfrxReport read FReportRef write FReportRef;
    /// <summary>
    ///   An event occurs when closing a data set.
    /// </summary>
    property OnClose: TNotifyEvent read FOnClose write FOnClose;
    /// <summary>
    ///   An event occurs when opening a data set.
    /// </summary>
    property OnOpen: TNotifyEvent read FOnOpen write FOnOpen;
  published
    /// <summary>
    ///   Defines, if the given component is available from the designer.
    /// </summary>
    property Enabled: Boolean read FEnabled write FEnabled default True;
    /// <summary>
    ///   The navigation start point.
    /// </summary>
    property RangeBegin: TfrxRangeBegin read FRangeBegin write FRangeBegin default rbFirst;
    /// <summary>
    ///   The endpoint of navigation.
    /// </summary>
    property RangeEnd: TfrxRangeEnd read FRangeEnd write FRangeEnd default reLast;
    /// <summary>
    ///   A number of records in the data set, if the "RangeEnd" property =
    ///   reCount.
    /// </summary>
    property RangeEndCount: Integer read FRangeEndCount write FRangeEndCount default 0;
    /// <summary>
    ///   A symbolic name, under which the dataset will be displayed in the
    ///   designer.
    /// </summary>
    property UserName: String read FUserName write SetUserName;
    /// <summary>
    ///   This event's handler must return the Eof = True parameter, if the end
    ///   of the data set is reached.
    /// </summary>
    property OnCheckEOF: TfrxCheckEOFEvent read FOnCheckEOF write FOnCheckEOF;
    /// <summary>
    ///   This event's handler must move the cursor to the beginning of the
    ///   data set.
    /// </summary>
    property OnFirst: TNotifyEvent read FOnFirst write FOnFirst;
    /// <summary>
    ///   This event's handler must move the cursor to the next record.
    /// </summary>
    property OnNext: TNotifyEvent read FOnNext write FOnNext;
    /// <summary>
    ///   This event's handler must move the cursor to the previous record.
    /// </summary>
    property OnPrior: TNotifyEvent read FOnPrior write FOnPrior;
  end;

  TfrxComponentClass = class of TfrxComponent;
  TfrxExtPropList = class;

  IfrxDataBindingComponent = interface
  ['{079181D1-ED18-4083-8EC3-0B874518F493}']
    procedure BindDataField(ADataSet: TfrxDataSet; const AFieldName: String);
  end;

  TfrxStringClassName = type String;

  IInspectedProperties = interface
  ['{BDBE0326-DA24-4FB8-ADA0-2F740AB6C199}']
    function GetProperies: TfrxExtPropList;
  end;

  IStringClassNameContainer = interface
  ['{97FFA69C-C491-458E-98A7-EFDCFFE05465}']
    function GetSupportedClasses: TList;
  end;

  /// <summary>
  ///   Represents field definitions used by the report designer
  /// </summary>

  { TfrxFieldDef }

  TfrxFieldDef = class(TfrCollectionItem, IStringClassNameContainer)
  private
    FFieldName: String;
    FFieldAlias: String;
    FSize: Integer;
    FBindComponent: TfrxStringClassName;
    FFieldType: TfrxFieldType;
    FActive: Boolean;
    procedure SetBindComponent(const Value: TfrxStringClassName);
    procedure SetFieldAlias(const Value: String);
    procedure SetFieldName(const Value: String);
//    FBindableFieldProps: TfrxBindableFieldProps;
  protected
    function GetInheritedName: String; override;
    { IInterface }
    {$REGION 'IUnknown'}
    {$IFDEF FPC}
    function QueryInterface({$IFDEF FPC_HAS_CONSTREF}constref{$ELSE}const{$ENDIF} iid : tguid;out obj) : longint;{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};
    {$ELSE}
    function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    {$ENDIF}
    function _AddRef: Integer; {$IFDEF FPC}{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};{$ELSE} stdcall;{$ENDIF}
    function _Release: Integer; {$IFDEF FPC}{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};{$ELSE} stdcall;{$ENDIF}
    {$ENDREGION}
    function GetSupportedClasses: TList;
  public
    constructor Create(ACollection: TCollection); override;
    procedure Assign(Source: TPersistent); override;
    function CreateUniqueFieldName: String;
    function CreateBindComponent(AOwner: TfrxComponent): TfrxComponent;
    function GetInternalName: String;
    function IsManualDefine: Boolean;
  published
    property Active: Boolean read FActive write FActive default True;
    property FieldName: String read FFieldName write SetFieldName;
    property FieldAlias: String read FFieldAlias write SetFieldAlias;
    property BindComponent: TfrxStringClassName read FBindComponent write SetBindComponent;
    property FieldType: TfrxFieldType read FFieldType write FFieldType default fftNumeric;
    property Size: Integer read FSize write FSize default 10;
    //property BindableProperties: TfrxBindableFieldProps read FBindableFieldProps write FBindableFieldProps;
  end;

  TfrxFieldDefs = class(TfrCollection)
  private
    FLegacyAliases: TStringList;
    FAliasesSorted: TStringList;
    FFieldsSorted: TStringList;
    FLegacyAliasesUpdating: Boolean;
    FIsLegacyAliasesFilled: Boolean;
    function GetItem(Index: Integer): TfrxFieldDef;
    procedure DoAliasesChange(Sender: TObject);
    procedure DoBeforeAliasesChange(Sender: TObject);
    procedure SetLegacyAliases(const Value: TStringList); deprecated 'Use TfrxFieldDefs.Items.FieldAlias';
    function ConvertAlias(const AAliasName: String): String;
    function GetAlias(const AFieldName: String): String;
    function GetValueSeparator: String;
    procedure InitializaLegacyAliases;
    function GetLegacyAliases: TStringList; deprecated 'Use TfrxFieldDefs.Items.FieldAlias';
    procedure UpdateNameOrAlias(const AOldFieldName, ANewFieldName, AOldAlias, ANewAlias: String; AFieldDef: TfrxFieldDef);
    function GetUpdated: Boolean;
    function GetIsManualDefined: Boolean;
  protected
    procedure Notify(Item: TCollectionItem; Action: TCollectionNotification); override;
    procedure UpdateFromAliasesList(AList: TStrings);
    procedure UpdateFromDefines;
  public
    constructor Create;
    destructor Destroy; override;
    procedure AssignAll(Source: TfrxFieldDefs);
    function AddFieldDef: TfrxFieldDef;
    procedure EndUpdate; override;
    function FindFieldDef(const AFieldName: String): TfrxFieldDef;
    function FindAliasDef(const AAliasName: String): TfrxFieldDef;
    procedure GetFieldList(AList: TStrings);
    property LegacyAliases: TStringList read GetLegacyAliases write SetLegacyAliases;
    property Items[Index: Integer]: TfrxFieldDef read GetItem; default;
    property IsManualDefined: Boolean read GetIsManualDefined;
    property Updated: Boolean read GetUpdated;
  end;

{$IFDEF DELPHI16}
/// <summary>
///   The TfrxUserDataSet component allows constructing reports, which are not
///   connected to DB data, but do receive data from other sources (for
///   example, array, file, etc.). A programmer should only provide navigation
///   in such source (define OnFirst, OnNext, OnPrior, OnCheckEOF events). Data
///   receiving is performed not via this component, but using other ways (for
///   example, via the "TfrxReport.OnGetValue" event). In most cases you can
///   only set the navigation range by setting RangeBegin, RangeEnd,
///   RangeEndCount properties. The current position can be obtained via RecNo
///   property.
/// </summary>
[ComponentPlatformsAttribute(pidWin32 or pidWin64)]
{$ENDIF}
  TfrxUserDataSet = class(TfrxDataset)
  private
    FFields: TStrings;
    FOnGetValue: TfrxGetValueEvent;
    FOnNewGetValue: TfrxNewGetValueEvent;
    FOnGetBlobValue: TfrxGetBlobValueEvent;
    FOnIsBlobField : TfrxIsBlobFieldEvent;
    procedure SetFields(const Value: TStrings);
  protected
    function GetDisplayText(Index: String): WideString; override;
    function GetValue(Index: String): Variant; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function FieldsCount: Integer; override;
    procedure GetFieldList(List: TStrings); override;
    function IsBlobField(const FieldName: String): Boolean; override;
    procedure AssignBlobTo(const FieldName: String; Obj: TObject); override;
  published
    /// <summary>
    ///   List of the fields. These names used in the report designer "Data
    ///   tree".
    /// </summary>
    property Fields: TStrings read FFields write SetFields;
    /// <summary>
    ///   This event is fired when an unknown variable (VarName) is met. An
    ///   eventhandler must return the value of this variable in the Value
    ///   parameter.
    /// </summary>
    property OnGetValue: TfrxGetValueEvent read FOnGetValue write FOnGetValue;
    /// <summary>
    ///   This event is fired when an unknown variable (VarName) is met. An
    ///   event handler must return the value of this variable in the Value
    ///   parameter.New event has a Sender parameter and pass a report objects
    ///   which calls an event.
    /// </summary>
    property OnNewGetValue: TfrxNewGetValueEvent read FOnNewGetValue write FOnNewGetValue;
    /// <summary>
    ///   This event is fired when an unknown variable (VarName) is met and
    ///   IsBlobField event returns True(BLOB type). AssignTo parameter
    ///   contains an object in which data should be loaded (TWideStrings for
    ///   "Text"object).
    /// </summary>
    property OnGetBlobValue: TfrxGetBlobValueEvent read FOnGetBlobValue write FOnGetBlobValue;
    /// <summary>
    ///   An event fires when engine checks type of an assigned field. The
    ///   result tells report engine how to interpret the field with name
    ///   'FieldName'.Returns True if it's a BLOB field. In this case
    ///   OnGetBlobValue event will fire.
    /// </summary>
    property OnIsBlobField: TfrxIsBlobFieldEvent read FOnIsBlobField write FOnIsBlobField;
  end;

  /// <summary>
  ///   The TfrxDBDataSet component is designed for connecting to the DB
  ///   components, which are based on TDataSet, such as "TTable" and "TQuery".
  ///   Such functions as navigation in the data source and referring to fields
  ///   are performed automatically, which means that a programmer does not
  ///   have to worry about it.
  /// </summary>
  TfrxCustomDBDataSet = class(TfrxDataSet)
  private
    FFields: TStringList;
    FFieldDefs: TfrxFieldDefs;
    FFieldDefTypesUpdated: Boolean;
    //TODO
    //FFields: TfrxFieldsStringList;
    //procedure SetFieldAliases(const Value: TfrxFieldsStringList);
    procedure SetFieldAliases(const Value: TStrings); deprecated 'Use TfrxCustomDBDataSet.FieldDefs.Items.FieldAlias';
    procedure WriteFieldDefs(Writer: TWriter);
    procedure ReadFieldDefs(Reader: TReader);
    function GetFieldAliases: TStrings; deprecated 'Use TfrxCustomDBDataSet.FieldDefs.Items.FieldAlias';
  protected
    function DoUpdateDefs(AFieldDefs: TfrxFieldDefs; ADeleteUnbound: Boolean; AFromList: TStrings = nil): Boolean;
    procedure LoadFieldDef(AFieldDef: TfrxFieldDef; var UpdateFieldDefs: Boolean); virtual;
    procedure DefineProperties(Filer: TFiler); override;
    function GetDisplayWidth(Index: String): Integer; override;
    function GetFieldType(Index: String): TfrxFieldType; override;
    function IsFieldDefsUpdated: Boolean; virtual;
    procedure UpdateAutoFieldDefs;
    property FieldDefs: TfrxFieldDefs read FFieldDefs;
    property Fields: TStringList read FFields;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function ConvertAlias(const fName: String): String;
    function GetAlias(const fName: String): String;
    function FieldsCount: Integer; override;
    procedure UpdateFieldDefs(bForceUpdate: Boolean = False; AFromList: TStrings = nil); override;
    procedure UpdateFieldTypes; override;
    procedure ResetFieldDefs; override;
    procedure WriteNestedProperties(Item: TfrXmlItem; AAncestor: TPersistent = nil); override;
    function ReadNestedProperty(Item: TfrXmlItem): Boolean; override;
    function CreateFieldBindComponent(const Index: String; AOwner: TfrxComponent): TfrxComponent; override;
  published
    property CloseDataSource;
    //property FieldAliases: TfrxFieldsStringList read FFields write SetFieldAliases;
    /// <summary>
    ///   List of field aliases.
    /// </summary>
    /// <example>
    ///   FieldName1=FieldAlias1 <br />FieldName2=FieldAlias2
    /// </example>
    property FieldAliases: TStrings read GetFieldAliases write SetFieldAliases stored False; // deprecated
    property OpenDataSource;
    property OnClose;
    property OnOpen;
  end;

  TfrxDBComponents = class(TfrCustomComponent)
  public
    function GetDescription: String; virtual;
  end;

  /// <summary>
  ///   The base class for DB engine components such as TfrxXXXDatabase. Used
  ///   internally in reports.
  /// </summary>
  TfrxCustomDatabase = class(TfrxDialogComponent)
  protected
    procedure BeforeConnect(var Value: Boolean);
    procedure AfterDisconnect;
    procedure SetConnected(Value: Boolean); virtual;
    procedure SetDatabaseName(const Value: String); virtual;
    procedure SetLoginPrompt(Value: Boolean); virtual;
    procedure SetParams(Value: TStrings); virtual;
    function GetConnected: Boolean; virtual;
    function GetDatabaseName: String; virtual;
    function GetLoginPrompt: Boolean; virtual;
    function GetParams: TStrings; virtual;
  public
    /// <summary>
    ///   Converts connection settings to string.
    /// </summary>
    function ToString: WideString{$IFDEF DELPHI12}; reintroduce{$ENDIF}; virtual;
    procedure FromString(const Connection: WideString); virtual;
    /// <summary>
    ///   Sets the login and password required for database connection.
    /// </summary>
    procedure SetLogin(const Login, Password: String); virtual;
    /// <summary>
    ///   Determines whether the connection is active.
    /// </summary>
    property Connected: Boolean read GetConnected write SetConnected default False;
    /// <summary>
    ///   The database name or connection string.
    /// </summary>
    property DatabaseName: String read GetDatabaseName write SetDatabaseName;
    /// <summary>
    ///   Determines whether to show the login dialog when connecting to a
    ///   database. Default value is True.
    /// </summary>
    property LoginPrompt: Boolean read GetLoginPrompt write SetLoginPrompt default True;
    /// <summary>
    ///   Parameters of the connection.
    /// </summary>
    property Params: TStrings read GetParams write SetParams;
  published
    property Left;
    property Top;
  end;

{ Report Objects }

  /// <summary>
  ///   Encapsulate data for fill gaps. Band objects use it as a gap parameter
  ///   for fill object.
  /// </summary>
  TfrxFillGaps = class(TPersistent)
  private
    FTop: Integer;
    FLeft: Integer;
    FBottom: Integer;
    FRight: Integer;
  published
    /// <summary>
    ///   Top gap.
    /// </summary>
    property Top: Integer read FTop write FTop;
    /// <summary>
    ///   Left gap.
    /// </summary>
    property Left: Integer read FLeft write FLeft;
    /// <summary>
    ///   Bottom gap.
    /// </summary>
    property Bottom: Integer read FBottom write FBottom;
    /// <summary>
    ///   Right gap.
    /// </summary>
    property Right: Integer read FRight write FRight;
  end;

  /// <summary>
  ///   Base abstract class for representation of object fills interface. Every
  ///   object fill should inherit from this class.
  /// </summary>
  TfrxCustomFill = class(TPersistent)
  public
//    procedure Assign(Source: TfrxCustomFill); virtual; abstract;
    /// <summary>
    ///   Draw fill on a canvas.
    /// </summary>
    /// <param name="ACanvas">
    ///   Source canvas.
    /// </param>
    /// <param name="X">
    ///   Start X coordinate of draw area.
    /// </param>
    /// <param name="Y">
    ///   Start Y coordinate of draw area.
    /// </param>
    /// <param name="X1">
    ///   End X coordinate of draw area.
    /// </param>
    /// <param name="Y1">
    ///   End Y coordinate of draw area.
    /// </param>
    /// <param name="ScaleX">
    ///   X scale factor.
    /// </param>
    /// <param name="ScaleY">
    ///   Y scale factor.
    /// </param>
    procedure Draw(ACanvas: TCanvas; X, Y, X1, Y1: Integer; ScaleX, ScaleY: Extended); virtual; abstract;
    /// <summary>
    ///   Returns XML string with serialized comparison of stored data between
    ///   two objects( Self and AFill).
    /// </summary>
    function Diff(AFill: TfrxCustomFill; const Add: String): String; virtual; abstract;
  end;

  /// <summary>
  ///   Represents standard brush fill class for all report objects.
  /// </summary>
  TfrxBrushFill = class(TfrxCustomFill)
  private
    FBackColor: TColor;
    FForeColor: TColor;
    FStyle: TBrushStyle;
  public
    constructor Create;
    procedure Assign(Source: TPersistent); override;
    procedure Draw(ACanvas: TCanvas; X, Y, X1, Y1: Integer; ScaleX, ScaleY: Extended); override;
    function Diff(AFill: TfrxCustomFill; const Add: String): String; override;
  published
    /// <summary>
    ///   Background color.
    /// </summary>
    property BackColor: TColor read FBackColor write FBackColor default clNone;
    /// <summary>
    ///   Fore color for hatch brush.
    /// </summary>
    property ForeColor: TColor read FForeColor write FForeColor default clBlack;
    /// <summary>
    ///   Brush styles.
    /// </summary>
    property Style: TBrushStyle read FStyle write FStyle default bsSolid;
  end;

  /// <summary>
  ///   Represents standard gradient fill class for all report objects.
  /// </summary>
  TfrxGradientFill = class(TfrxCustomFill)
  private
    FStartColor: TColor;
    FEndColor: TColor;
    FGradientStyle: TfrxGradientStyle;
  public
    constructor Create;
    procedure Assign(Source: TPersistent); override;
    procedure Draw(ACanvas: TCanvas; X, Y, X1, Y1: Integer; ScaleX, ScaleY: Extended); override;
    function Diff(AFill: TfrxCustomFill; const Add: String): String; override;
    function GetColor: TColor;
  published
    /// <summary>
    ///   Begin color of the gradient.
    /// </summary>
    property StartColor: TColor read FStartColor write FStartColor default clWhite;
    /// <summary>
    ///   End color of the gradient.
    /// </summary>
    property EndColor: TColor read FEndColor write FEndColor default clBlack;
    /// <summary>
    ///   Style of the gradient fill.
    /// </summary>
    property GradientStyle: TfrxGradientStyle read FGradientStyle write FGradientStyle;
  end;


  /// <summary>
  ///   Orientation type of glass fill.
  /// </summary>
  TfrxGlassFillOrientation = (
    /// <summary>
    ///   Vertical.
    /// </summary>
    foVertical,
    /// <summary>
    ///   Horizontal.
    /// </summary>
    foHorizontal,
    /// <summary>
    ///   Vertical mirrored.
    /// </summary>
    foVerticalMirror,
    /// <summary>
    ///   Horizontal mirrored.
    /// </summary>
    foHorizontalMirror);

  TfrxGlassFill = class(TfrxCustomFill)
  private
    FColor: TColor;
    FBlend: Extended;
    FHatch: Boolean;
    FOrient: TfrxGlassFillOrientation;
  public
    constructor Create;
    procedure Assign(Source: TPersistent); override;
    procedure Draw(ACanvas: TCanvas; X, Y, X1, Y1: Integer; ScaleX, ScaleY: Extended); override;
    function Diff(AFill: TfrxCustomFill; const Add: String): String; override;
    function GetColor: TColor;
    function BlendColor: TColor;
    function HatchColor: TColor;
  published
    /// <summary>
    ///   Base color.
    /// </summary>
    property Color: TColor read FColor write FColor default clWhite;
    /// <summary>
    ///   Blend value used in calculation of blend color. Value lays between
    ///   0.0 and 1.0. Highest value means more contrast.
    /// </summary>
    property Blend: Extended read FBlend write FBlend;
    /// <summary>
    ///   Draws hatch lines.
    /// </summary>
    property Hatch: Boolean read FHatch write FHatch default False;
    /// <summary>
    ///   Orientation type of glass fill.
    /// </summary>
    property Orientation: TfrxGlassFillOrientation read FOrient write FOrient default foHorizontal;
  end;


  /// <summary>
  ///   This class encapsulates data for hyperlinks. Some report objects may
  ///   have hyperlink property which can be used to: change location inside a
  ///   report, open HTTP link, open and build detailed report or detailed
  ///   report page.
  /// </summary>
  TfrxHyperlink = class(TPersistent)
  private
    FKind: TfrxHyperlinkKind;
    FDetailReport: String;
    FDetailPage: String;
    FExpression: String;
    FReportVariable: String;
    FValue: String;
    FValuesSeparator: String;
    FTabCaption: String;
    function IsValuesSeparatorStored: Boolean;
  public
    constructor Create;
    procedure Assign(Source: TPersistent); override;
    /// <summary>
    ///   Returns XML string with serialized comparison of stored data between
    ///   two objects( Self and ALink).
    /// </summary>
    /// <param name="ALink">
    ///   Base object used to compare properties.
    /// </param>
    /// <param name="Add">
    ///   Additional XML text.
    /// </param>
    function Diff(ALink: TfrxHyperlink; const Add: String): String;
    /// <summary>
    ///   Assigns values of variables declared in ReportVariable property to
    ///   aReport object.
    /// </summary>
    /// <param name="aReport">
    ///   Report object.
    /// </param>
    procedure AssignVariables(aReport: TfrxReport);
    /// <summary>
    ///   Returns True if data for detail reports is not assigned.
    /// </summary>
    function DetailIsEmpty: Boolean;
    /// <summary>
    ///   Returns True if data is not filled and properties have default
    ///   values.
    /// </summary>
    function IsEmpty: Boolean;
  published
    /// <summary>
    ///   Type of hyperlinks supported by report objects.
    /// </summary>
    property Kind: TfrxHyperlinkKind read FKind write FKind default hkURL;
    /// <summary>
    ///   Symbolic link to a detail report.
    /// </summary>
    property DetailReport: String read FDetailReport write FDetailReport;
    /// <summary>
    ///   Symbolic link to a detail page.
    /// </summary>
    property DetailPage: String read FDetailPage write FDetailPage;
    /// <summary>
    ///   Expression for variable declared in ReportVariable property. The
    ///   report Engine calculates it in GetData method. May contain several
    ///   expressions delimited with ValuesSeparator sign(Default ';') like
    ///   &lt;Expr1&gt;;&lt;Expr2&gt;;&lt;Expr3&gt;.
    /// </summary>
    property Expression: String read FExpression write FExpression;
    /// <summary>
    ///   Variable names in destination report which will be used to broadcast
    ///   parameters of one report to another. May contain several variables
    ///   delimited with ValuesSeparator sign(Default ';') like Var1;Var2;Var3.
    ///   <br />
    /// </summary>
    property ReportVariable: String read FReportVariable write FReportVariable;
    /// <summary>
    ///   Caption of the new report tab in the report preview.
    /// </summary>
    property TabCaption: String read FTabCaption write FTabCaption;
    /// <summary>
    ///   Default values for variables declared in ReportVariable property. May
    ///   contain several variables delimited with ValuesSeparator sign(Default
    ///   ';') like Val1;Val2;Val3. <br />
    /// </summary>
    property Value: String read FValue write FValue;
    /// <summary>
    ///   Separator sign. It used in expressions, variables and values as a
    ///   delimiter. Default ';'.
    /// </summary>
    property ValuesSeparator: String read FValuesSeparator write FValuesSeparator stored IsValuesSeparatorStored;
  end;

  /// <summary>
  ///   The TfrxFrameLine class represents the parameters of one line of the
  ///   frame.
  /// </summary>
  TfrxFrameLine = class(TPersistent)
  private
    FFrame: TfrxFrame;
    FColor: TColor;
    FStyle: TfrxFrameStyle;
    FWidth: Extended;
    function IsColorStored: Boolean;
    function IsStyleStored: Boolean;
    function IsWidthStored: Boolean;
  public
    constructor Create(AFrame: TfrxFrame);
    procedure Assign(Source: TPersistent); override;

    /// <summary>
    ///   Returns XML string with serialized comparison of stored data between
    ///   two objects( Self and ALine). <br />
    /// </summary>
    function Diff(ALine: TfrxFrameLine; const LineName: String;
      ColorChanged, StyleChanged, WidthChanged: Boolean): String;
  published
    /// <summary>
    ///   Color of the line.
    /// </summary>
    property Color: TColor read FColor write FColor stored IsColorStored;
    /// <summary>
    ///   Style of the line.
    /// </summary>
    property Style: TfrxFrameStyle read FStyle write FStyle stored IsStyleStored;
    /// <summary>
    ///   Width of the line.
    /// </summary>
    property Width: Extended read FWidth write FWidth stored IsWidthStored;
  end;


  /// <summary>
  ///   The TfrxFrame class represents a frame. Almost all Frx visual objects
  ///   have a frame.
  /// </summary>
  TfrxFrame = class(TPersistent)
  private
    FLeftLine: TfrxFrameLine;
    FTopLine: TfrxFrameLine;
    FRightLine: TfrxFrameLine;
    FBottomLine: TfrxFrameLine;
    FColor: TColor;
    FDropShadow: Boolean;
    FShadowWidth: Extended;
    FShadowColor: TColor;
    FStyle: TfrxFrameStyle;
    FTyp: TfrxFrameTypes;
    FWidth: Extended;
    function IsShadowWidthStored: Boolean;
    function IsTypStored: Boolean;
    function IsWidthStored: Boolean;
    procedure SetBottomLine(const Value: TfrxFrameLine);
    procedure SetLeftLine(const Value: TfrxFrameLine);
    procedure SetRightLine(const Value: TfrxFrameLine);
    procedure SetTopLine(const Value: TfrxFrameLine);
    procedure SetColor(const Value: TColor);
    procedure SetStyle(const Value: TfrxFrameStyle);
    procedure SetWidth(const Value: Extended);
  public
    /// <summary>
    ///   Constructor.
    /// </summary>
    constructor Create;
    destructor Destroy; override;
    /// <summary>
    ///   Assigns properties from other frame object.
    /// </summary>
    procedure Assign(Source: TPersistent); override;
    /// <summary>
    ///   Returns XML string with serialized comparison of stored data between
    ///   two objects( Self and AFrame). <br />
    /// </summary>
    function Diff(AFrame: TfrxFrame): String;
    /// <summary>
    ///   Draws frame on a Canvas.
    /// </summary>
    /// <param name="Canvas">
    ///   Source canvas.
    /// </param>
    /// <param name="FX">
    ///   X start coordinate.
    /// </param>
    /// <param name="FY">
    ///   Y start coordinate.
    /// </param>
    /// <param name="FX1">
    ///   X end coordinate.
    /// </param>
    /// <param name="FY1">
    ///   Y end coordinate. <br />
    /// </param>
    /// <param name="ScaleX">
    ///   X scale factor.
    /// </param>
    /// <param name="ScaleY">
    ///   Y scale factor.
    /// </param>
    procedure Draw(Canvas: TCanvas; FX, FY, FX1, FY1: Integer; ScaleX, ScaleY: Extended);
  published
    /// <summary>
    ///   Color of frame.
    /// </summary>
    property Color: TColor read FColor write SetColor default clBlack;
    /// <summary>
    ///   Determines whether to draw a shadow.
    /// </summary>
    property DropShadow: Boolean read FDropShadow write FDropShadow default False;
    /// <summary>
    ///   Color of shadow.
    /// </summary>
    property ShadowColor: TColor read FShadowColor write FShadowColor default clBlack;
    /// <summary>
    ///   Width of shadow.
    /// </summary>
    property ShadowWidth: Extended read FShadowWidth write FShadowWidth stored IsShadowWidthStored;
    /// <summary>
    ///   Style of frame.
    /// </summary>
    property Style: TfrxFrameStyle read FStyle write SetStyle default fsSolid;
    /// <summary>
    ///   Set of frame lines.
    /// </summary>
    property Typ: TfrxFrameTypes read FTyp write FTyp stored IsTypStored;
    /// <summary>
    ///   Width of frame.
    /// </summary>
    property Width: Extended read FWidth write SetWidth stored IsWidthStored;
    /// <summary>
    ///   Left line of the frame.
    /// </summary>
    property LeftLine: TfrxFrameLine read FLeftLine write SetLeftLine;
    /// <summary>
    ///   Top line of the frame.
    /// </summary>
    property TopLine: TfrxFrameLine read FTopLine write SetTopLine;
    /// <summary>
    ///   Right line of the frame.
    /// </summary>
    property RightLine: TfrxFrameLine read FRightLine write SetRightLine;
    /// <summary>
    ///   Bottom line of the frame.
    /// </summary>
    property BottomLine: TfrxFrameLine read FBottomLine write SetBottomLine;
  end;



//  TfrxWatermark = class(TPersistent)
//  private
//    FVisibility: TfrxVisibilityTypes;
//    FTextFont: TFont;
//    FImage:
//    FImageTransparentColor
//    FShowImageOnTop
//    ShowTextOnTop
//    Text
//    TextFill
//    TextRotation
//
//  Visible
//Font
//Image
//ImageTransparentColor
//ShowImageOnTop
//ShowTextOnTop
//Text
//TextFill
//TextRotation
//  end;

  /// <summary>
  ///   The TfrxView class is the basic one for most components, which can be
  ///   placed on a report design page. An object of this type has parameters
  ///   such as "Frame" and "Color" and also can be connected to the data
  ///   source. Practically all FastReport standard objects are inherited from
  ///   this class.
  /// </summary>
  TfrxView = class(TfrxReportComponent)
  private
    FAlign: TfrxAlign;
    FCursor: TCursor;
    FDataField: String;
    FDataSet: TfrxDataSet;
    FDataSetName: String;
    FFrame: TfrxFrame;
    FShiftMode: TfrxShiftMode;
    FTagStr: String;
    FTempTag: String;
    FTempHyperlink: String;
    FHint: String;
    FShowHint: Boolean;
    FURL: String;
    FPlainText: Boolean;
    FHyperlink: TfrxHyperlink;
    FFill: TfrxCustomFill;
    FVisibility: TfrxVisibilityTypes;
    FDrawAsMask: Boolean;
    FAllowVectorExport: Boolean;
    {  interactive script events  }
    FOnMouseEnter: TfrxMouseEnterViewEvent;
    FOnMouseLeave: TfrxMouseLeaveViewEvent;
    FOnMouseDown: TfrxMouseEvent;
    FOnMouseMove: TfrxMouseMoveEvent;
    FOnMouseUp: TfrxMouseEvent;
    procedure SetFrame(const Value: TfrxFrame);
    procedure SetDataSet(const Value: TfrxDataSet);
    procedure SetDataSetName(const Value: String);
    function GetDataSetName: String;
    procedure SetFillType(const Value: TfrxFillType);
    function GetFillType: TfrxFillType;
    procedure SetBrushStyle(const Value: TBrushStyle);
    function GetBrushStyle: TBrushStyle;
    procedure SetColor(const Value: TColor);
    function GetColor: TColor;
    procedure SetHyperLink(const Value: TfrxHyperlink);
    procedure SetFill(const Value: TfrxCustomFill);
    procedure SetURL(const Value: String);
    function GetPrintable: Boolean;
    procedure SetPrintable(Value: Boolean);
    procedure SetProcessing(const Value: TfrxObjectProcessing);
  protected
    FX: Integer;
    FY: Integer;
    FX1: Integer;
    FY1: Integer;
    FDX: Integer;
    FDY: Integer;
    FFrameWidth: Integer;
    FScaleX: Extended;
    FScaleY: Extended;
    FOffsetX: Extended;
    FOffsetY: Extended;
    FCanvas: TCanvas;
    FProcessing: TfrxObjectProcessing;
    FObjAsMetafile: Boolean;
    FDrawFillOnMetaFile: Boolean;
    FVC: TVectorCanvas;
    FStyle: TfrxStyleName;
    procedure BeginDraw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); virtual;
    procedure DrawBackground; virtual;
    procedure DrawFrame; virtual;
    procedure DrawFrameEdges;
    procedure DrawLine(x, y, x1, y1, w: Integer);
    procedure ExpandVariables(var Expr: String; ARaiseException: Boolean = True);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure GetScreenScale(var aScaleX, aScaleY: Extended);
	  procedure SetStyle(const Value: TfrxStyleName); virtual;
    procedure BindDataField(ADataSet: TfrxDataSet; const AFieldName: String);
    { interactive object behaviour }
    function dm: Integer; virtual; // = 0
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    /// <summary>
    ///   Applies the given style to the object.
    /// </summary>
    /// <param name="Style">
    ///   Style.
    /// </param>
    procedure ApplyStyle(Style: TfrxCustomStyleItem); overload; virtual;
    procedure AssignToStyle(Style: TfrxCustomStyleItem); virtual;
    function Diff(AComponent: TfrxComponent): String; override;
    /// <summary>
    ///   Highlighting a component.
    /// </summary>
    procedure Highlighting(Canvas: TCanvas; AColor: TColor);
    /// <summary>
    ///   Returns True when report object has assigned DataSet and DataField.
    /// </summary>
    function IsDataField: Boolean;
    /// <summary>
    ///   Through this method objects tell the export engine that it allows
    ///   export via meta file output. When function returns False object will
    ///   be exported as picture or via native processing(if supported).
    /// </summary>
    function IsEMFExportable: Boolean; virtual;
    function IsAcceptAsChild(aParent: TfrxComponent): Boolean; override;
    /// <summary>
    ///   Returns vector graphic image derived from TGraphic (Currently
    ///   supports meta file) with object output on it.
    /// </summary>
    /// <param name="DrawFill">
    ///   If True , object draws fill on result vector image.
    /// </param>
    function GetVectorGraphic(DrawFill: Boolean = False): TGraphic; overload; virtual;
    function GetVectorGraphic(DrawFill: Boolean; AScaleX, AScaleY: Double): TGraphic; overload; virtual;
    /// <summary>
    ///   Returns TVectorCanvas used for export engine. <br />
    /// </summary>
    function GetVectorCanvas: TVectorCanvas; virtual;
    procedure DrawHighlight(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended; hlColor: TColor = clNone); override;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;
    /// <summary>
    ///   Draw object with clip bounds on. Any output behind bounds cuts off.
    ///   By default just calls standard Draw method. Used for some objects.
    /// </summary>
    procedure DrawClipped(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); virtual;
    procedure BeforePrint; override;
    procedure GetData; override;
    /// <summary>
    ///   Deprecated. Used to get scale factor of object content for export
    ///   filters.
    /// </summary>
    procedure GetScaleFactor(var ScaleX: Extended; var ScaleY: Extended); virtual;
    procedure MirrorContent(MirrorModes: TfrxMirrorControlModes); override;

    procedure AfterPrint; override;
    /// <summary>
    ///   Calculates size of dropped shadows.
    /// </summary>
    function ShadowSize: Extended; // In order to not test DropShadow every time

    { post processing interfaces }
    /// <summary>
    ///   This method used by the report engine when adds object to preview
    ///   pages when object has post-processing enabled. It should save
    ///   expression from current object in PostProcessor instance and save
    ///   returned index in current object.
    /// </summary>
    /// <param name="aReport">
    ///   Reference to the report object.
    /// </param>
    /// <param name="PostProcessor">
    ///   Reference to TfrxPostProcessor instance of post processing manager.
    /// </param>
    procedure SaveContentToDictionary(aReport: TfrxReport; PostProcessor: TfrxPostProcessor); virtual;
    /// <summary>
    ///   Method loads data from post-processor item (aItem) to current object
    ///   based on index stored before during SaveContentToDictionary function
    ///   call.
    /// </summary>
    /// <param name="aReport">
    ///   Reference to the report object.
    /// </param>
    /// <param name="aItem">
    ///   TfrxMacrosItem which stores processed data for current object.
    /// </param>
    function LoadContentFromDictionary(aReport: TfrxReport; aItem: TfrxMacrosItem): Boolean; virtual;
    /// <summary>
    ///   The report engine call this method when event specified in
    ///   Processing.ProcessAt occurs. It retrieves saved expression from <br />
    ///   PostProcessor by index saved during SaveContentToDictionary call and
    ///   calculates expression(calls to GetData). The result saves back to
    ///   PostProcessor for future use in LoadContentFromDictionary. <br />
    /// </summary>
    procedure ProcessDictionary(aItem: TfrxMacrosItem; aReport: TfrxReport; PostProcessor: TfrxPostProcessor); virtual;
    /// <summary>
    ///   Return True when post processing for current object is enabled
    ///   (Default ProcessAt &lt;&gt; paDefault). <br />
    /// </summary>
    function IsPostProcessAllowed: Boolean; virtual;
    { end post process interfaces}

    /// <summary>
    ///   Object fill type.
    /// </summary>
    property FillType: TfrxFillType read GetFillType write SetFillType default ftBrush;
    /// <summary>
    ///   Reference to fill instance based on FillType.
    /// </summary>
    property Fill: TfrxCustomFill read FFill write SetFill;
    /// <summary>
    ///   Object's filling style (deprecated check Fill property). <br />
    /// </summary>
    property BrushStyle: TBrushStyle read GetBrushStyle write SetBrushStyle stored False;
    /// <summary>
    ///   Object's filling color (deprecated check Fill property). <br />
    /// </summary>
    property Color: TColor read GetColor write SetColor stored False;
    /// <summary>
    ///   Data field's name, which the object is connected to.
    /// </summary>
    property DataField: String read FDataField write FDataField;
    /// <summary>
    ///   Data source, which the object is connected to.
    /// </summary>
    property DataSet: TfrxDataSet read FDataSet write SetDataSet;
    /// <summary>
    ///   Name of data source, which the object is connected to. This property
    ///   duplicates the DataSet property.
    /// </summary>
    property DataSetName: String read GetDataSetName write SetDataSetName;
    /// <summary>
    ///   Object's frame.
    /// </summary>
    property Frame: TfrxFrame read FFrame write SetFrame;
    /// <summary>
    ///   Used in export matrix to get a plain text of object.
    /// </summary>
    property PlainText: Boolean read FPlainText write FPlainText;
    /// <summary>
    ///   Object's cursor (in the preview mode).
    /// </summary>
    property Cursor: TCursor read FCursor write FCursor default crDefault;
    /// <summary>
    ///   The field for storing different information. This property can
    ///   contain an expression in square brakets which will be calculated when
    ///   report runs.
    /// </summary>
    property TagStr: String read FTagStr write FTagStr;
    /// <summary>
    ///   Object's URL (Deprecated. Check HyperLink property). Determines what
    ///   action will occur when a user clicks this <br />object in the preview
    ///   window. The property may contain string like: <br />- '@number' -
    ///   jumps to specified page number;/n <br />- '#name' - jumps to
    ///   specified anchor;/n <br />- 'mailto:', 'htpp:', 'ftp:' - execute
    ///   shell command./n/n <br />The property may contain an expression in
    ///   square brakets which will be calculated when report runs.
    /// </summary>
    property URL: String read FURL write SetURL stored False;
    /// <summary>
    ///   Internal use. The export engine uses it to generate transparency mask
    ///   of object.
    /// </summary>
    property DrawAsMask: Boolean read FDrawAsMask write FDrawAsMask;
    /// <summary>
    ///   Defines how the report engine should process all expressions and
    ///   GetData calls for current object. It controls post processing of
    ///   expressions inside objects.
    /// </summary>
    property Processing: TfrxObjectProcessing read FProcessing write SetProcessing;
        /// <summary>
    ///   Name of the object's style. Styles are stored in the
    ///   TfrxReport.Styles.
    /// </summary>
    property Style: TfrxStyleName read FStyle write SetStyle;
  published
    property Anchors;
    property AllowMirrorMode;
    /// <summary>
    ///   Object's aligning relatively to its parent.
    /// </summary>
    property Align: TfrxAlign read FAlign write FAlign default baNone;
    /// <summary>
    ///   This object can be exported through vector graphic processor(meta
    ///   file) if object and export filter supports it.
    /// </summary>
    property AllowVectorExport: Boolean read FAllowVectorExport write FAllowVectorExport;
    /// <summary>
    ///   Defines whether the given object should be printed out (Deprecated
    ///   check Visibility property).
    /// </summary>
    property Printable: Boolean read GetPrintable write SetPrintable stored False default True;
    /// <summary>
    ///   The mode of object's shifting in cases when a stretchable object is
    ///   placed over the given one
    /// </summary>
    property ShiftMode: TfrxShiftMode read FShiftMode write FShiftMode default smAlways;
    property Left;
    property Top;
    property Width;
    property Height;
    property GroupIndex;
    property Editable;
    property Restrictions;
    property Visible;
    property OnAfterData;
    property OnAfterPrint;
    property OnBeforePrint;
    property OnContentChanged;
    /// <summary>
    ///   The name of the script routine that will be called when mouse cursor
    ///   in the report preview enters object area.
    /// </summary>
    property OnMouseEnter: TfrxMouseEnterViewEvent read FOnMouseEnter write FOnMouseEnter;
    /// <summary>
    ///   The name of the script routine that will be called when mouse cursor
    ///   in the report preview leaves object area.
    /// </summary>
    property OnMouseLeave: TfrxMouseLeaveViewEvent read FOnMouseLeave write FOnMouseLeave;
    /// <summary>
    ///   The name of the script routine that will be called when user push
    ///   mouse button in the report preview on current object.
    /// </summary>
    property OnMouseDown: TfrxMouseEvent read FOnMouseDown write FOnMouseDown;
    /// <summary>
    ///   The name of the script routine that will be called when user move
    ///   mouse pointer in the report preview over current object.
    /// </summary>
    property OnMouseMove: TfrxMouseMoveEvent read FOnMouseMove write FOnMouseMove;
    /// <summary>
    ///   The name of the script routine that will be called when user release
    ///   mouse button in the report preview on current object.
    /// </summary>
    property OnMouseUp: TfrxMouseEvent read FOnMouseUp write FOnMouseUp;
    property OnPreviewClick;
    property OnPreviewDblClick;
    /// <summary>
    ///   The report object hyperlink property which can be used to: change
    ///   location inside a report, open HTTP link, open and build detailed
    ///   report or detailed report page.
    /// </summary>
    property Hyperlink: TfrxHyperlink read FHyperlink write SetHyperLink;
    /// <summary>
    ///   The hint, shows when hold mouse pointer at object.
    /// </summary>
    property Hint: String read FHint write FHint;
    /// <summary>
    ///   Show/hide hint.
    /// </summary>
    property ShowHint: Boolean read FShowHint write FShowHint default False;
    /// <summary>
    ///   Visibility of object in the result report.
    /// </summary>
    property Visibility: TfrxVisibilityTypes read FVisibility write FVisibility default [vsPreview, vsExport, vsPrint];
  end;

  TfrxStyledView = class(TfrxView);

  /// <summary>
  ///   The type of method used by AutoScale mechanism.
  /// </summary>
  TfrxContentAutoScale = (
    /// <summary>
    ///   AutoScale is Disabled.
    /// </summary>
    casNone,
    /// <summary>
    ///   AutoScale in static mode. The content of the object should fit its constant bounds.
    /// </summary>
    casStatic,
    /// <summary>
    ///   AutoScale in dynamic mode. The content of object decreases with object height
    ///   until objects on a parent container fit a report page.
    /// </summary>
    casDynamic);

  /// <summary>
  ///   The type of direction that can be used by AutoScale mechanism.
  /// </summary>
  TfrxContentAutoScaleType = (
    /// <summary>
    ///   Allows to scale down an object content.
    /// </summary>
    cstShrink,
    /// <summary>
    ///   Allows to scale up an object content.
    /// </summary>
    cstGrow);
  TfrxContentAutoScaleTypes = set of TfrxContentAutoScaleType;

  /// <summary>
  ///  The TfrxContentScaleConstraints class is a data class used to store constraints settings for
  ///  content auto-scale mechanism. The report objects uses settings from this calss
  ///  when report engine calls to auto-scale mechanism interfaces.
  /// </summary>
  TfrxContentScaleConstraints = class(TPersistent)
  private const
    DefaultMaxStepValue = 32;
    DefaultMinStepValue = 4;
    DefaultMaxStepValueLimit = 2000;
  private
    FMinStepValue: Integer;
    FMaxStepValue: Integer;
    FMinIterationValue: Integer;
    FMaxIterationValue: Integer;
    procedure SetMaxStepValue(const Value: Integer);
    procedure SetMinStepValue(const Value: Integer);
    procedure SetMinIterationValue(const Value: Integer);
    procedure SetMaxIterationValue(const Value: Integer);
    function CheckMinValue(const AValue: Integer; const AMaxValue: Integer): Boolean;
    function CheckMaxValue(const AValue: Integer; const AMinValue: Integer): Boolean;
  public
    constructor Create;
    procedure Assign(Source: TPersistent); override;
    /// <summary>
    ///   The function checks if parameter value in rage of iterator constraint values.
    /// </summary>
    function IsIterationInRange(const AIteration: Integer): Boolean;
    /// <summary>
    ///   The function checks if parameter value in rage of step constraint values.
    /// </summary>
    function IsCurentStepInRange(const AStep: Integer): Boolean;
    function GetMaxStepValueLimit: Integer;
  published
    /// <summary>
    ///   The maximum range of iterator value. Zero value means it has not upper limit.
    /// </summary>
    property MaxIterationValue: Integer read FMaxIterationValue write SetMaxIterationValue;
    /// <summary>
    ///   The minimum range of iterator value. Zero value means it has not lower limit.
    /// </summary>
    property MinIterationValue: Integer read FMinIterationValue write SetMinIterationValue;
    /// <summary>
    ///   The maximum range of step value. Zero value means it has not upper limit.
    /// </summary>
    property MaxStepValue: Integer read FMaxStepValue write SetMaxStepValue default DefaultMaxStepValue;
    /// <summary>
    ///   The minimum range of step value. Zero value means it has not lower limit.
    /// </summary>
    property MinStepValue: Integer read FMinStepValue write SetMinStepValue default DefaultMinStepValue;
  end;

  /// <summary>
  ///  The TfrxContentScaleOptions class is a data class used to store content scale settings for
  ///  content auto-scale mechanism. The report objects uses settings from this calss
  ///  when report engine calls to auto-scale mechanism interfaces. Used to scale down fonts in a text
  ///  object.
  /// </summary>
  TfrxContentScaleOptions = class(TPersistent)
  private const
    DefaultStepValue = 2;
  private
    FAutoScale: TfrxContentAutoScale;
    FConstraints: TfrxContentScaleConstraints;
    FStepValue: Integer;
    FScaleType: TfrxContentAutoScaleTypes;
    procedure SetAutoScale(const Value: TfrxContentAutoScale);
    procedure SetConstraints(const Value: TfrxContentScaleConstraints);
    procedure SetStepValue(const Value: Integer);
    procedure SetScaleType(const Value: TfrxContentAutoScaleTypes);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
  published
    /// <summary>
    ///   The mode of Auto-scale for the current object. Determines how the object scales content inside.
    /// </summary>
    property AutoScale: TfrxContentAutoScale read FAutoScale write SetAutoScale default casNone;
    /// <summary>
    ///   The constraints settings for Auto-scale mode.
    /// </summary>
    property Constraints: TfrxContentScaleConstraints read FConstraints write SetConstraints;
    /// <summary>
    ///   The step value for dynamic auto-scale mode. Is used as increment for each scale-down iteration.
    /// </summary>
    property StepValue: Integer read FStepValue write SetStepValue default DefaultStepValue;
    /// <summary>
    ///   The type of scale operation, grow or shrink content.
    /// </summary>
    property ScaleType: TfrxContentAutoScaleTypes read FScaleType write SetScaleType default [cstShrink, cstGrow];

  end;

  /// <summary>
  ///   The TfrxStretcheable class is the basic one for writing components,
  ///   which modify their height depending on the data placed in them. Objects
  ///   of the given class can be not only stretched, but they can also be
  ///   "broken" into pieces in cases when an object does not find room on the
  ///   page. At the same time, the object is displayed piecemeal until all its
  ///   data is displayed.
  /// </summary>
  TfrxStretcheable = class(TfrxStyledView)
  private
    FStretchMode: TfrxStretchMode;
    FCanShrink: Boolean;
    FContentScaleOptions: TfrxContentScaleOptions;
    function GetContentScaleOptions: TfrxContentScaleOptions;
    procedure SetContentScaleOptions(const Value: TfrxContentScaleOptions);
  protected
    function IsContentScaleOptionsStored: Boolean;
    /// <summary>
    ///  The Settings of content scale behaviour. Used by the report engine to scale down or up content of the object.
    ///  Typically by change a font size for a Text object
    /// </summary>
    property ContentScaleOptions: TfrxContentScaleOptions read GetContentScaleOptions write SetContentScaleOptions stored IsContentScaleOptionsStored;
  public
    /// <summary>
    ///   Internal use only. Stores object Height during stretch operation.
    /// </summary>
    FSaveHeight: Extended;
    /// <summary>
    ///   Internal use only. Stores original top during shift operation.
    /// </summary>
    FSavedTop: Extended;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    /// <summary>
    ///   Calculates and returns the object's height according to the data
    ///   placed in it.
    /// </summary>
    function CalcHeight: Extended; virtual;
    /// <summary>
    ///   Objects part drawing function , if the object is splittable.
    /// </summary>
    function DrawPart: Extended; virtual;
    /// <summary>
    ///   The method returns True when object supports content AutoScale behaviour.
    /// </summary>
    function IsCanScaleContent: TfrxContentAutoScale; virtual;
    /// <summary>
    ///   The method used by the report engine to change content scale for a single iteration.
    /// </summary>
    function ScaleContent(const AIteration: Integer; const ADelta: Extended; const AScalableObjects: IfrReadonlyList): Extended; virtual;
    /// <summary>
    ///   The method used by the report engine to change content scale for static scale. It decreases scale until one of the following conditions are meet:
    ///     - Object content fits the bounds;
    ///     - Scale factor reaches minimum limit;
    /// </summary>
    procedure ScaleContentToFitBounds; virtual;
    function RestoreContentScale: Extended; virtual;
    /// <summary>
    ///   Initialize part of Splittable object.
    /// </summary>
    procedure InitPart; virtual;
    /// <summary>
    ///   Returns True if object has a part to draw on a new page.
    /// </summary>
    /// <param name="aFreeSpace">
    ///   Free space on current page.
    /// </param>
    function HasNextDataPart(aFreeSpace: Extended): Boolean; virtual;
    function GetLastShiftAmount(aFreeSpace: Extended): Extended; virtual;
  published
    /// <summary>
    ///   True if object can shrink height base on its content.
    /// </summary>
    property CanShrink: Boolean read FCanShrink write FCanShrink default False;
    /// <summary>
    ///   Object's stretch mode.
    /// </summary>
    property StretchMode: TfrxStretchMode read FStretchMode write FStretchMode
      default smDontStretch;
  end;

  /// <summary>
  ///   Reserved.
  /// </summary>
  TfrxInteractiveHighlightEvent = (ihNone, ihClick, ihEnter, ieLeave);

  /// <summary>
  ///   The TfrxHighlight class holds the conditional highlight settings. The
  ///   conditional highlight is used in the TfrxMemoView object and allows
  ///   changing the font and color settings depending on some condition.
  /// </summary>
  TfrxHighlight = class(TfrCollectionItem)
  private
    FActive: Boolean;
    FApplyFont: Boolean;
    FApplyFill: Boolean;
    FApplyFrame: Boolean;
    FCondition: String;
    FFont: TFont;
    FFill: TfrxCustomFill;
    FFrame: TfrxFrame;
    FVisible: Boolean;
    FInteractiveType: TfrxInteractiveHighlightEvent;
    procedure SetFont(const Value: TFont);
    procedure SetFill(const Value: TfrxCustomFill);
    procedure SetFillType(const Value: TfrxFillType);
    function GetFillType: TfrxFillType;
    procedure SetColor(const Value: TColor);
    function GetColor: TColor;
    procedure SetFrame(const Value: TfrxFrame);
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    /// <summary>
    ///   Copies all properties from other object.
    /// </summary>
    procedure Assign(Source: TPersistent); override;
    /// <summary>
    ///   True if instance should generate unique name before serialization.
    ///   Used in inherit mechanism for unnamed collections.
    /// </summary>
    function IsUniqueNameStored: Boolean; override;
  published
    // backward compatibility
    /// <summary>
    ///   Highlight's font. Determines whether the highlight is active. When active,
    ///   the TfrxMemoView object uses highlight settings.
    /// </summary>
    property Active: Boolean read FActive write FActive stored False default False;
    /// <summary>
    ///   Deprecated. Apply font styles to the object. <br />
    /// </summary>
    property ApplyFont: Boolean read FApplyFont write FApplyFont default True;
    /// <summary>
    ///   Deprecated. Apply Fill to the object. <br />
    /// </summary>
    property ApplyFill: Boolean read FApplyFill write FApplyFill default True;
    /// <summary>
    ///   Deprecated. Apply frame to the object. <br />
    /// </summary>
    property ApplyFrame: Boolean read FApplyFrame write FApplyFrame default False;
    /// <summary>
    ///   Highlight's font.
    /// </summary>
    property Font: TFont read FFont write SetFont;
    // backward compatibility
    /// <summary>
    ///   Highlight's color.
    /// </summary>
    property Color: TColor read GetColor write SetColor stored False default clNone;
    /// <summary>
    ///   Highlight's condition. Contains any valid expression.
    /// </summary>
    property Condition: String read FCondition write FCondition;
    /// <summary>
    ///   Reserved.
    /// </summary>
    property InteractiveType: TfrxInteractiveHighlightEvent read FInteractiveType write FInteractiveType stored False;
 	  /// <summary>
 	  ///   Highlight's Fill type.
 	  /// </summary>
 	  property FillType: TfrxFillType read GetFillType write SetFillType;
    /// <summary>
    ///   Highlight's Fill object.
    /// </summary>
    property Fill: TfrxCustomFill read FFill write SetFill;
    /// <summary>
    ///   Highlight's frame object.
    /// </summary>
    property Frame: TfrxFrame read FFrame write SetFrame;
    /// <summary>
    ///   Visibility of Highlight.
    /// </summary>
    property Visible: Boolean read FVisible write FVisible default True;
  end;

  /// <summary>
  ///   Collection of TfrxHighlight. Each object can hold several highlights
  ///   which serves different conditions.
  /// </summary>
  TfrxHighlightCollection = class(TfrCollection)
  private
    function GetItem(Index: Integer): TfrxHighlight;
  public
    constructor Create;
    /// <summary>
    ///   TfrxHighlight items.
    /// </summary>
    property Items[Index: Integer]: TfrxHighlight read GetItem; default;
  end;

  /// <summary>
  ///   The constructor. Do not call it directly because an instance of this
  ///   class is stored in the TfrxCustomMemoView.DisplayFormat.
  /// </summary>
  TfrxFormat = class(TCollectionItem)
  private
    FDecimalSeparator: String;
    FThousandSeparator: String;
    FFormatStr: String;
    FKind: TfrxFormatKind;
  public
    procedure Assign(Source: TPersistent); override;
  published
    /// <summary>
    ///   Decimal separator symbol. This symbol is used when the Kind property
    ///   is set to fkNumeric. If this property is empty, default separator is
    ///   used.
    /// </summary>
    property DecimalSeparator: String read FDecimalSeparator write FDecimalSeparator;
    /// <summary>
    ///   Thousand separator symbol. It is used when numbers formatting (Kind
    ///   property =fk Numeric).
    /// </summary>
    property ThousandSeparator: String read FThousandSeparator write FThousandSeparator;
    /// <summary>
    ///   Formatting string in every case is an argument for the function, with
    ///   the help of which formatting is accomplished. Thus, for numerical
    ///   formatting, such function would be the Delphi's Format function, for
    ///   date/time it is the FormatDateTime function. One can get the possible
    ///   values from the Delphi help system. Below are several values used in
    ///   FastReport: <br />for the numerical formatting: <br />%g - a number
    ///   with the minimal signs number after decimal point; <br />%2.2f - a
    ///   number with the fixed number of signs after decimal point; <br />
    ///   %2.2n - a number with bits delimiter; <br />%2.2m - a monetary
    ///   format, accepted in the Windows OS, depending on the regional
    ///   settings in the control panel. <br />for the date/time format: <br />
    ///   dd.mm.yyyy - date of the 23.12.2003 type; <br />dd mmm yyyy - date of
    ///   the 23 Nov. 2003 type; <br />dd mmmm yyyy - date of the 23 November
    ///   2003 type; <br />hh:mm - time of the 23:12 type; <br />hh:mm:ss -
    ///   time of the 23:12:00 type; <br />dd mmmm yyyy, hh:mm - time and date
    ///   of the 23 November 2003, 23:12 type. <br />As for formatting of the
    ///   boolean type, the formatting string is presented as two values
    ///   separated by comma. The first value corresponds to "False," the
    ///   second one corresponds to "True".
    /// </summary>
    property FormatStr: String read FFormatStr write FFormatStr;
    /// <summary>
    ///   Kind of formatting.
    /// </summary>
    property Kind: TfrxFormatKind read FKind write FKind default fkText;
  end;

  /// <summary>
  ///   Collection of TfrxFormat. An object can hold several <br />formats
  ///   which serves different expressions. <br />
  /// </summary>
  TfrxFormatCollection = class(TCollection)
  private
    function GetItem(Index: Integer): TfrxFormat;
  public
    constructor Create;
    /// <summary>
    ///   Format items array.
    /// </summary>
    property Items[Index: Integer]: TfrxFormat read GetItem; default;
  end;

  // internal Type move to TfrxCustomMemoView
  TfrxCustomMemoViewBehaviourFlag = (mbfAllowExpressions, mbfAllowHTMLTags,
    mbfAutoWidth, mbfClipped, mbfFirstParaBreak, mbfLastParaBreak,
    mbfRTLReading, mbfSuppressRepeated, mbfWordBreak, mbfWordWrap, mbfWysiwyg,
    mbfUseDefaultCharset, mbfHighlightActivated, mbfMacroLoaded,
    mbfClearEmptyLines, mbfHideZeros);
  TfrxCustomMemoViewBehaviourFlags = set of TfrxCustomMemoViewBehaviourFlag;
  /// <summary>
  ///   The TfrxCustomMemoView class is the base class for the "Text", "System
  ///   text" and "Dot-matrix text" objects. This object represents a
  ///   rectangular area with the multiline text inside.
  /// </summary>
  TfrxCustomMemoView = class(TfrxStretcheable, IfrxDataLinkObject)
  private
    FCharSpacing: Extended;
    FFormats: TfrxFormatCollection;
    FExpressionDelimiters: String;
    FFlowTo: TfrxCustomMemoView;
    FGapX: Extended;
    FGapY: Extended;
    FHAlign: TfrxHAlign;
    FHighlights: TfrxHighlightCollection;
    FLineSpacing: Extended;
    FMemo: TWideStrings;
    FParagraphGap: Extended;
    FRotation: Integer;
    FDuplicates: TfrxDuplicateMerge;
    FDataLink: TfrxDataLink;
    FTempMemo: WideString;
    FUnderlinesTextMode: TfrxUnderlinesTextMode;
    FVAlign: TfrxVAlign;
    FValue: Variant;
    FTempFill: TfrxCustomFill;
    FTempFont: TFont;
    FTempFrame: TfrxFrame;
    FTempVisible: Boolean;
    FScaledRect: TRect;
    FStoredFontSize: Integer;
    FMemoBehaviourFlags: TfrxCustomMemoViewBehaviourFlags;
    procedure SetMemo(const Value: TWideStrings);
    procedure SetRotation(Value: Integer);
    procedure SetAnsiText(const Value: AnsiString);
    function AdjustCalcHeight: Extended;
    function AdjustCalcWidth: Extended;
    function GetAnsiText: AnsiString;
    function IsExprDelimitersStored: Boolean;
    function IsLineSpacingStored: Boolean;
    function IsGapXStored: Boolean;
    function IsGapYStored: Boolean;
    function IsHighlightStored: Boolean;
    function IsDisplayFormatStored: Boolean;
    function IsParagraphGapStored: Boolean;
    function GetHighlight: TfrxHighlight;
    function GetDisplayFormat: TfrxFormat;
    procedure SetHighlight(const Value: TfrxHighlight);
    procedure SetDisplayFormat(const Value: TfrxFormat);
    function IsCharSpacingStored: Boolean;
    procedure WriteFormats(Writer: TWriter);
    procedure WriteHighlights(Writer: TWriter);
    procedure ReadFormats(Reader: TReader);
    procedure ReadHighlights(Reader: TReader);
    procedure SavePreHighlightState;
    procedure RestorePreHighlightState;
    function GetUnderlines: Boolean;
    procedure SetUnderlines(const Value: Boolean);
    function GetDataLink: TfrxDataLink;
    function IsDataLinkStored: Boolean;
    procedure SetDataLink(const Value: TfrxDataLink);
    function GetFlagValue(const Index: Integer): Boolean;
    procedure SetFlagValue(const Index: Integer; const Value: Boolean);
    procedure CalculateAutoWidth(ADrawText: TObject);
    function CalcSize(ACalcWidth: Boolean; ACalcHeight: Boolean): TfrxPoint;
    function GetInterpretationType: TfrxInterpretationsTypes;
    function GetHightlightingType: TfrxHightlightingsTypes;
  protected
    FPartMemo: WideString;
    FLastValue: Variant;
    FTotalPages: Integer;
    FCopyNo: Integer;
    FTextRect: TRect;
    FPrintScale: Extended;
    FMacroIndex: Integer;
    FMacroLine: Integer;
    procedure Loaded; override;
    function CalcAndFormat(const Expr: WideString; Format: TfrxFormat): WideString;
    function CalcTextRect(OffsetX, OffsetY, ScaleX, ScaleY: Extended): TRect;

    procedure BeginDraw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;
    procedure SetDrawParams(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended);
    procedure DrawText;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DefineProperties(Filer: TFiler); override;
    function GetText: WideString; virtual;
    function GetValue: Variant; virtual;
    procedure SetText(const Value: WideString); virtual;
    procedure SetPartMemoText(const Value: WideString); virtual;

    procedure ClearObjectData;
    function LoadDataStream(Stream: TStream; const NewLink: String): Boolean;
    function GetLink(LoadMethod: TfrxDataLinkLoadMethod): String;
    function IsExpressionLink: Boolean;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    class function GetDescription: String; override;
    function Diff(AComponent: TfrxComponent): String; override;
    function DiffText(AComponent: TfrxComponent): String; virtual;
    function HasOutboundContent: Boolean;
    function CalcHeight: Extended; override;
    /// <summary>
    ///   Calculates and returns the object's width, in pixels.
    /// </summary>
    function CalcWidth: Extended; virtual;
    function DrawPart: Extended; override;
    procedure WriteNestedProperties(Item: TfrXMLItem; AAncestor: TPersistent = nil); override;
    function ReadNestedProperty(Item: TfrXMLItem): Boolean; override;
    /// <summary>
    ///   Apply active highlight for object in prepared report.
    /// </summary>
    procedure ApplyPreviewHighlight;
    function ScaleContent(const AIteration: Integer; const ADelta: Extended; const AScalableObjects: IfrReadonlyList): Extended; override;
    procedure ScaleContentToFitBounds; override;
    function RestoreContentScale: Extended; override;
    procedure SaveContentToDictionary(aReport: TfrxReport; PostProcessor: TfrxPostProcessor); override;
    function LoadContentFromDictionary(aReport: TfrxReport; aItem: TfrxMacrosItem): Boolean; override;
    procedure ProcessDictionary(aItem: TfrxMacrosItem; aReport: TfrxReport; PostProcessor: TfrxPostProcessor); override;
    function IsPostProcessAllowed: Boolean; override;

    { interactive object behaviour }
    procedure DoMouseEnter(aPreviousObject: TfrxComponent; var EventParams: TfrxInteractiveEventsParams); override;
    procedure DoMouseLeave(aNextObject: TfrxComponent; var EventParams: TfrxInteractiveEventsParams); override;

    procedure MirrorContent(MirrorModes: TfrxMirrorControlModes); override;
    function GetComponentText: String; override;
    /// <summary>
    ///   Formats the value passed in the "Value" parameter and returns the
    ///   result as a string. This method is used format settings stored in the
    ///   DisplayFormat property. If AFormat parameter is set, it is used for
    ///   formatting instead of DisplayFormat property.
    /// </summary>
    function FormatData(const Value: Variant; AFormat: TfrxFormat = nil): WideString;
    // when aParaText is not nil function fills stringlist with paragraph indents in Objects property
    // 1 - begin of paragraph, 2 - end of paragraph, 3 -  both begin and end at one line, 0 - continue of paragraph
    /// <summary>
    ///   Breaks the text so all strings would fit in the object's width, and
    ///   returns the resulting text.
    /// </summary>
    /// <param name="WrapWords">
    ///   Wrap words if possible.
    /// </param>
    /// <param name="aParaText">
    ///   If not nil , holds result of function in TWideStrings with paragraph
    ///   marks.
    /// </param>
    function WrapText(WrapWords: Boolean; aParaText: TWideStrings = nil): WideString;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;
    procedure BeforePrint; override;
    procedure GetData; override;
    procedure AfterPrint; override;
    procedure InitPart; override;
    /// <summary>
    ///   Apply Highlight to object.
    /// </summary>
    /// <param name="AHighlight">
    ///   Highlight.
    /// </param>
    procedure ApplyHighlight(AHighlight: TfrxHighlight);
    /// <summary>
    ///   Extract macros from the object text and replaces them with values.
    /// </summary>
    procedure ExtractMacros(Dictionary: TfrxPostProcessor = nil);
    /// <summary>
    ///   Deprecated. Reset suppressed values.
    /// </summary>
    procedure ResetSuppress;
    /// <summary>
    ///   Normalize angle to 360 degrees.
    /// </summary>
    function ReducedAngle: Integer; // 0..359
    procedure UpdateBounds; override;
    /// <summary>
    ///   Text of the object.
    /// </summary>
    property Text: WideString read GetText write SetText;
    /// <summary>
    ///   Text of the object in ANSI format based on current Locale settings.
    /// </summary>
    property AnsiText: AnsiString read GetAnsiText write SetAnsiText;
    /// <summary>
    ///   Stores value of Format operation(for single expression or DataSet
    ///   only) with respect of its type.
    /// </summary>
    property Value: Variant read GetValue write FValue;
    // analogue of Memo property
    /// <summary>
    ///   Text lines of the object.
    /// </summary>
    property Lines: TWideStrings read FMemo write SetMemo;

    /// <summary>
    ///   Determines if the object may contain expressions inside the text.
    ///   Expressions are enclosed in the square brackets (by default). The
    ///   ExpressionDelimiters defines which symbols to use for detecting the
    ///   expression. Default value is True.
    /// </summary>
    property AllowExpressions: Boolean index Integer(mbfAllowExpressions) read GetFlagValue write SetFlagValue default True;
    /// <summary>
    ///   Determines if the object may contain HTML-tags inside the text. <br />
    ///   The following tags are supported: <br />&lt;b&gt; - bold; <br />
    ///   &lt;i&gt; - italic; <br />&lt;u&gt; - underline; <br />&lt;sub&gt; -
    ///   bottom index; <br />&lt;sup&gt; - top index; <br />&lt;font color&gt;
    ///   - font color. <br />Default value is False.
    /// </summary>
    property AllowHTMLTags: Boolean index Integer(mbfAllowHTMLTags) read GetFlagValue write SetFlagValue default False;
    /// <summary>
    ///   Determines if the text object should handle its width automatically.
    /// </summary>
    property AutoWidth: Boolean index Integer(mbfAutoWidth) read GetFlagValue write SetFlagValue default False;
    /// <summary>
    ///   Character spacing, in pixels. Default value is 0.
    /// </summary>
    property CharSpacing: Extended read FCharSpacing write FCharSpacing stored IsCharSpacingStored;
    /// <summary>
    ///   Deletes empty lines which only have line break at the end after object gets data.
    /// </summary>
    property ClearEmptyLines: Boolean index Integer(mbfClearEmptyLines) read GetFlagValue write SetFlagValue default False;
    /// <summary>
    ///   Determines whether to clip the text on object's bounds.
    /// </summary>
    property Clipped: Boolean index Integer(mbfClipped) read GetFlagValue write SetFlagValue default True;
    /// <summary>
    ///   Deprecated(Check Formats property). Format settings. Used for
    ///   formatting expression's values <br />or DB fields which object is
    ///   linked to. <br />
    /// </summary>
    property DisplayFormat: TfrxFormat read GetDisplayFormat write SetDisplayFormat stored IsDisplayFormatStored;
    property DataLink: TfrxDataLink read GetDataLink write SetDataLink stored IsDataLinkStored;
    /// <summary>
    ///   Set of symbols, designating the expression. Default value is '[,]'.
    ///   The comma divides opening and closing symbols. There is one
    ///   limitation however: the opening and closing symbols cannot be
    ///   similar, so '%,%' will not work. One can set several symbols, for
    ///   example '&lt;%,%&gt;'.
    /// </summary>
    property ExpressionDelimiters: String read FExpressionDelimiters
      write FExpressionDelimiters stored IsExprDelimitersStored;
    /// <summary>
    ///   Link to the "Text" object that will be used for displaying the
    ///   textthat not fit in this object. The linked object should have the
    ///   samefont settings as the main object. You also should place the
    ///   linkedobject after the main object.
    /// </summary>
    property FlowTo: TfrxCustomMemoView read FFlowTo write FFlowTo;
    /// <summary>
    ///   Gap between text and left and right edges of object, in pixels.
    ///   Default value is 2.
    /// </summary>
    property GapX: Extended read FGapX write FGapX stored IsGapXStored;
    /// <summary>
    ///   Gap between text and top and bottom edges of object, in pixels.
    ///   Default value is 1.
    /// </summary>
    property GapY: Extended read FGapY write FGapY stored IsGapYStored;
    /// <summary>
    ///   The horizontal text alignment.
    /// </summary>
    property HAlign: TfrxHAlign read FHAlign write FHAlign default haLeft;
    /// <summary>
    ///   Determines whether to show the zero values of expressions or DB
    ///   fields. Default value is True.
    /// </summary>
    property HideZeros: Boolean index Integer(mbfHideZeros) read GetFlagValue write SetFlagValue default False;
    /// <summary>
    ///   Deprecated(see Highlights property). Conditional highlight settings.
    ///   The hightlight allows <br />changing the appearance of the object
    ///   depending on some condition. <br />
    /// </summary>
    property Highlight: TfrxHighlight read GetHighlight write SetHighlight stored IsHighlightStored;
    /// <summary>
    ///   Spacing between text lines, in pixels. Default value is 2.
    /// </summary>
    property LineSpacing: Extended read FLineSpacing write FLineSpacing stored IsLineSpacingStored;
    /// <summary>
    ///   Text lines of the object.
    /// </summary>
    property Memo: TWideStrings read FMemo write SetMemo;
    /// <summary>
    ///   Paragraph gap, in pixels. Default value is 0.
    /// </summary>
    property ParagraphGap: Extended read FParagraphGap write FParagraphGap stored IsParagraphGapStored;
    /// <summary>
    ///   Determines the text rotation, in degrees. Valid values are 0..360.
    /// </summary>
    property Rotation: Integer read FRotation write SetRotation default 0;
    /// <summary>
    ///   Used to print text right-to-left for far-east languages.
    /// </summary>
    property RTLReading: Boolean index Integer(mbfRTLReading) read GetFlagValue write SetFlagValue default False;
    property Style;
    /// <summary>
    ///   Deprecated(See Duplicates property). Determines if it hides repeated
    ///   values. Default value is False.
    /// </summary>
    property SuppressRepeated: Boolean index Integer(mbfSuppressRepeated) read GetFlagValue write SetFlagValue default False;
    /// <summary>
    ///   Determines how to process repeated values. Default value is dmShow.
    /// </summary>
    property Duplicates: TfrxDuplicateMerge read FDuplicates write FDuplicates default dmShow;
    /// <summary>
    ///   Deprecated(See UnderlinesTextMode property). Determines whether to
    ///   show underlines under each text string. Default value is False.
    /// </summary>
    property Underlines: Boolean read GetUnderlines write SetUnderlines default False;
    /// <summary>
    ///   Determines how to show underlines in text object. Default value is
    ///   ulmNone.
    /// </summary>
    property UnderlinesTextMode: TfrxUnderlinesTextMode read FUnderlinesTextMode write FUnderlinesTextMode default ulmNone;
    /// <summary>
    ///   Determines whether to break russian words. Default value is False.
    /// </summary>
    property WordBreak: Boolean index Integer(mbfWordBreak) read GetFlagValue write SetFlagValue default False;
    /// <summary>
    ///   Determines whether the text object inserts soft carriage returns so
    ///   text wraps at the right margin. Default value is True.
    /// </summary>
    property WordWrap: Boolean index Integer(mbfWordWrap) read GetFlagValue write SetFlagValue default True;
    /// <summary>
    ///   Determines whether to use printer canvas to arrange the text more
    ///   accurately. If set to False, you cannot use the "width align" mode.
    /// </summary>
    property Wysiwyg: Boolean index Integer(mbfWysiwyg) read GetFlagValue write SetFlagValue default True;
    /// <summary>
    ///   The vertical text alignment.
    /// </summary>
    property VAlign: TfrxVAlign read FVAlign write FVAlign default vaTop;
    /// <summary>
    ///   Use Unicode during export even if Font Charset is selected.
    /// </summary>
    property UseDefaultCharset: Boolean index Integer(mbfUseDefaultCharset) read GetFlagValue write SetFlagValue default False;
    /// <summary>
    ///   Collection of Highlights. The object can hold several highlights
    ///   which serves different conditions.
    /// </summary>
    property Highlights: TfrxHighlightCollection read FHighlights;
    /// <summary>
    ///   Collection of Formats. The object can hold several formats for each
    ///   expression inside an object.
    /// </summary>
    property Formats: TfrxFormatCollection read FFormats;
  published
    property ContentScaleOptions;
    /// <summary>
    ///   Internal use. Determines first paragraph break if object where split.
    ///   <br />
    /// </summary>
    property FirstParaBreak: Boolean index Integer(mbfFirstParaBreak) read GetFlagValue write SetFlagValue default False;
    /// <summary>
    ///   Internal use. Determines last paragraph break if object where split.
    /// </summary>
    property LastParaBreak: Boolean index Integer(mbfLastParaBreak) read GetFlagValue write SetFlagValue default False;
    property Cursor;
    property TagStr;
    property URL;
  end;

  /// <summary>
  ///   The TfrxMemoView class represents the "Text" object that can show
  ///   multiline text in a rectangular area.
  /// </summary>
  TfrxMemoView = class(TfrxCustomMemoView, IfrxDataBindingComponent)
  published
    property AutoWidth;
    property AllowExpressions;
    property AllowHTMLTags;
    property BrushStyle;
    property CharSpacing;
    property ClearEmptyLines;
    property Clipped;
    property Color;
    property DataField;
    property DataSet;
    property DataSetName;
    property DisplayFormat;
    property DataLink;
    property ExpressionDelimiters;
    property FlowTo;
    property Font;
    property Frame;
    property FillType;
    property Fill;
    property GapX;
    property GapY;
    property HAlign;
    property HideZeros;
    property Highlight;
    property LineSpacing;
    property Memo;
    property ParagraphGap;
    property ParentFont;
    property Processing;
    property Rotation;
    property RTLReading;
    property Style;
    property SuppressRepeated;
    property Duplicates;
    property Underlines;
    property UnderlinesTextMode;
    property UseDefaultCharset;
    property WordBreak;
    property WordWrap;
    property Wysiwyg;
    property VAlign;
  end;

  /// <summary>
  ///   The TfrxSysMemoView represents the "System text" object. This object
  ///   can show system variables, aggregate values or text line.
  /// </summary>
  TfrxSysMemoView = class(TfrxCustomMemoView)
  public
    class function GetDescription: String; override;
  published
    property AutoWidth;
    property CharSpacing;
    property Color;
    property DisplayFormat;
    property Font;
    property Frame;
    property FillType;
    property Fill;
    property GapX;
    property GapY;
    property HAlign;
    property HideZeros;
    property Highlight;
    property Memo;
    property ParentFont;
    property Processing;
    property Rotation;
    property RTLReading;
    property Style;
    property SuppressRepeated;
    property Duplicates;
    property VAlign;
    property WordWrap;
  end;

  /// <summary>
  ///   The TfrxCustomLineView class is the base class for "Line" and
  ///   "Dot-matrix line" objects.
  /// </summary>
  TfrxCustomLineView = class(TfrxStretcheable)
  private
    FColor: TColor;
    FDiagonal: Boolean;
    FArrowEnd: Boolean;
    FArrowLength: Integer;
    FArrowSolid: Boolean;
    FArrowStart: Boolean;
    FArrowWidth: Integer;
    procedure DrawArrow(x1, y1, x2, y2: Extended);
    procedure DrawDiagonalLine;
  public
    constructor Create(AOwner: TComponent); override;
    constructor DesignCreate(AOwner: TComponent; Flags: Word); override;
    function Diff(AComponent: TfrxComponent): String; override;
    function GetVectorGraphic(DrawFill: Boolean = False): TGraphic; override;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;
    function IsContain(X, Y: Extended): Boolean; override;
    function IsInRect(const aRect: TfrxRect): Boolean; override;
    /// <summary>
    ///   Determines whether to draw an arrow in the end of line.
    /// </summary>
    property ArrowEnd: Boolean read FArrowEnd write FArrowEnd default False;
    /// <summary>
    ///   The length of arrow.
    /// </summary>
    property ArrowLength: Integer read FArrowLength write FArrowLength default 20;
    /// <summary>
    ///   Determines whether to draw a solid arrow.
    /// </summary>
    property ArrowSolid: Boolean read FArrowSolid write FArrowSolid default False;
    /// <summary>
    ///   Determines if an arrow is used at the start and end of the line.
    /// </summary>
    property ArrowStart: Boolean read FArrowStart write FArrowStart default False;
    /// <summary>
    ///   The width of arrow.
    /// </summary>
    property ArrowWidth: Integer read FArrowWidth write FArrowWidth default 5;
    /// <summary>
    ///   Determines if the line is diagonal.
    /// </summary>
    property Diagonal: Boolean read FDiagonal write FDiagonal default False;
  published
    /// <summary>
    ///   Color of the shape.
    /// </summary>
    property Color: TColor read FColor write FColor default clNone;
    property TagStr;
  end;

  /// <summary>
  ///   The TfrxLineView class represents the "Line" object. It can show a
  ///   horizontal, vertical or diagonal line.
  /// </summary>
  TfrxLineView = class(TfrxCustomLineView)
  public
    class function GetDescription: String; override;
  published
    property ArrowEnd;
    property ArrowLength;
    property ArrowSolid;
    property ArrowStart;
    property ArrowWidth;
    property Frame;
    property Diagonal;
  end;

  /// <summary>
  ///   The TfrxPictureView class represents the "Picture" object. It can show
  ///   a picture in the PNG, BMP, JPEG, ICO, WMF, EMF formats.
  /// </summary>
  TfrxPictureView = class(TfrxStyledView, IfrxCachedView, IfrxDataLinkObject, IfrxDataBindingComponent)
  private
    FGraphicProps: TfrxGraphicDrawProps;
    FFileLink: String;
    FDataLink: TfrxDataLink;
    FImageIndex: Integer;
    FIsImageIndexStored: Boolean;
    FIsPictureStored: Boolean;
    FPicture: TPicture;
    FPictureChanged: Boolean;
    FTransparentColor: TColor;
    FCached: IfrxCachedGraphic;
    FGFormat: TfrxCustomGraphicFormatClass;
    procedure SetPicture(const Value: TPicture);
    procedure PictureChanged(Sender: TObject);
    function GetPicture: TPicture;
    procedure UpdateGraphicHelper;
    function GetValue(const Index: Integer): Boolean;
    procedure SetValue(const Index: Integer; const Value: Boolean);
    
    procedure ClearObjectData;
    function LoadDataStream(Stream: TStream; const NewLink: String): Boolean;
    function GetLink(LoadMethod: TfrxDataLinkLoadMethod): String;
    function IsExpressionLink: Boolean;
        
    function GetDataLink: TfrxDataLink;
    procedure SetDataLink(const Value: TfrxDataLink);
    function IsDataLinkStored: Boolean;
  protected
    procedure SetCachedGraphic(const PictureCache: IfrxPictureCache);
    procedure GetCachedGraphic(ACacheType: TfrxCachedGraphicType; const PictureCache: IfrxPictureCache);
    procedure ReleaseCachedGraphic;
    function IsUpdateRequired(NewCacheType: TfrxCachedGraphicType): Boolean;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    class function GetDescription: String; override;
    function Diff(AComponent: TfrxComponent):String; override;
    procedure AfterPrint; override;
    procedure BeforePrint; override;
    /// <summary>
    ///   Load picture from stream.
    /// </summary>
    function LoadPictureFromStream(s: TStream; ResetStreamPos: Boolean = True): HResult; virtual;
    function IsEMFExportable: Boolean; override;
    function IsEmpty: Boolean;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;

    procedure GetData; override;
    function GetGraphic: TGraphic;
    function GraphicWidth: Integer;
    function GraphicHeight: Integer;
    function GraphicTransparency: Boolean;
    function GraphicIsVector: Boolean;
    function GraphicHasAlpha: Boolean;
    function GraphicHasMaskColor: Boolean;
    function GraphicTransparentColor: TColor;
    function GraphicIsTranslucent: Boolean;

    /// <summary>
    ///   Internal use. Determines if images stored in a picture cache.
    /// </summary>
    property IsImageIndexStored: Boolean read FIsImageIndexStored write FIsImageIndexStored;
    /// <summary>
    ///   Internal use. Determines if images stored in a picture object.
    /// </summary>
    property IsPictureStored: Boolean read FIsPictureStored write FIsPictureStored;
  published
    property Cursor;
    /// <summary>
    ///   Determines whether the picture should handle its size automatically.
    ///   Default value is False.
    /// </summary>
    property AutoSize: Boolean index Integer(fgdAutoSize) read GetValue write SetValue default False;
    /// <summary>
    ///   Determines whether the image should be centered inside the object.
    ///   Default value is False.
    /// </summary>
    property Center: Boolean index Integer(fgdCenter) read GetValue write SetValue default False;
    property DataField;
    property DataSet;
    property DataSetName;
    property DataLink: TfrxDataLink read GetDataLink write SetDataLink stored IsDataLinkStored;
    property Frame;
    property FillType;
    property Fill;
    /// <summary>
    ///   File name or expression which contains a name of the picture file.
    /// </summary>
    property FileLink: String read FFileLink write FFileLink;
    /// <summary>
    ///   Internal use only. Index of image in a picture cache if image stored
    ///   in a picture cache.
    /// </summary>
    property ImageIndex: Integer read FImageIndex write FImageIndex stored FIsImageIndexStored;
    /// <summary>
    ///   Keep the original aspect ratio of the image when stretching the
    ///   object. The Stretched property should also be set.
    /// </summary>
    property KeepAspectRatio: Boolean index Integer(fgdKeepAspectRatio) read GetValue write SetValue default True;
    /// <summary>
    ///   Reference to internal TPicture object that stores the image.
    /// </summary>
    property Picture: TPicture read GetPicture write SetPicture stored FIsPictureStored;
    /// <summary>
    ///   Stretches the picture to fit the object bounds. Default value is
    ///   True.
    /// </summary>
    /// <summary>
    ///   Stretches the picture to fit the object bounds. Default value is
    ///   True.
    /// </summary>
    property Stretched: Boolean index Integer(fgdStretch) read GetValue write SetValue default True;
    property Style;
    property TagStr;
    property URL;
    /// <summary>
    ///   Activates drawing with better smooth.
    /// </summary>
    property HightQuality: Boolean index Integer(fgdHiQuality) read GetValue write SetValue;
    /// <summary>
    ///   Determines if picture has transparent parts or Alpha mask.
    /// </summary>
    property Transparent: Boolean index Integer(fgdTransparent) read GetValue write SetValue;
    /// <summary>
    ///   Color of transparent parts.
    /// </summary>
    property TransparentColor: TColor read FTransparentColor write FTransparentColor;
  end;

  TfrxCustomSignatureGraphicType = (gtNo, gtImported, gtName);
  /// <summary>
  ///   Represents data for a custom digital signature image.
  /// </summary>
  TfrxCustomSignatureData = class(TPersistent)
  private
    FGraphicType: TfrxCustomSignatureGraphicType;
    FNamePicture: TPicture;
    FLogoPicture: TPicture;
    FFRGraphic: TGraphic;

    procedure SetNamePicture(const Value: TPicture);
    procedure SetLogoPicture(const Value: TPicture);
    function GetLogoPicture: TPicture;
  protected
    FShowName: Boolean;
    FShowLocation: Boolean;
    FShowDate: Boolean;
    FShowReason: Boolean;
    FShowDistinguishedName: Boolean;
    FShowProgramVersion: Boolean;
    FShowLogo: Boolean;
    FShowLabels: Boolean;
    FRTL: Boolean;
  public
    constructor Create;
    destructor Destroy; override;
  published
    property GraphicType: TfrxCustomSignatureGraphicType read FGraphicType write FGraphicType;
    property NamePicture: TPicture read FNamePicture write SetNamePicture;
    property ShowName: Boolean read FShowName write FShowName;
    property ShowLocation: Boolean read FShowLocation write FShowLocation;
    property ShowDate: Boolean read FShowDate write FShowDate;
    property ShowReason: Boolean read FShowReason write FShowReason;
    property ShowDistinguishedName: Boolean read FShowDistinguishedName write FShowDistinguishedName;
    property ShowProgramVersion: Boolean read FShowProgramVersion write FShowProgramVersion;
    property ShowLogo: Boolean read FShowLogo write FShowLogo;
    property ShowLabels: Boolean read FShowLabels write FShowLabels;
    property LogoPicture: TPicture read GetLogoPicture write SetLogoPicture;
    property RTL: Boolean read FRTL write FRTL;
  end;

  TPDFSignatureKind = (skInvisible, skVisible, skEmpty);

  TfrxDigitalSignatureView = class (TfrxPictureView)
  private
    FKind: TPDFSignatureKind;
    FDescription: WideString;
    FCustomizable: Boolean;
    FCustomData: TfrxCustomSignatureData;
    FCustomPicture: TPicture;
  protected
    FCommonName: WideString;
    FDate: WideString;
    FLocation: WideString;
    FReason: WideString;
    FDistinguishedName: WideString;
    FProgramVersion: WideString;
  protected
    FSavedPicture: TPicture;

    procedure CreateCustomPicture;
    function CreateBitmap32: TBitmap;
    function CreateBitmap24: TBitmap;
    function IsGraphicVisible: Boolean;
    function IsTextVisible: Boolean;
    procedure DrawLogo(Canvas: TCanvas);
    procedure DrawNamePicture(Canvas: TCanvas; Rect: TRect);
    procedure DrawNameText(Canvas: TCanvas; Rect: TRect);
    procedure DrawText(Canvas: TCanvas; Rect: TRect);
    procedure OutText(Canvas: TCanvas; Rect: TRect; WS: TWideStrings);
    procedure DrawWithAcpectRatio(Canvas: TCanvas; Rect: TRect; Graphic: TGraphic);
    procedure DrawTextParams(Rect: TRect);
    procedure DrawTextFontSizeAndText(dFontSize: Integer; WS: TWideStrings);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;
    procedure CustomFirst;
    procedure CustomLast;

    property CommonName: WideString write FCommonName;
    property Date: WideString write FDate;
    property Location: WideString read FLocation write FLocation;
    property Reason: WideString read FReason write FReason;
    property DistinguishedName: WideString write FDistinguishedName;
    property ProgramVersion: WideString write FProgramVersion;
  published
    property Font;
    property Kind: TPDFSignatureKind read FKind write FKind;
    /// <summary>
    ///   Designed to identify the signature in the export dialog if there are
    ///   multiple signatures.
    /// </summary>
    property Description: WideString read FDescription write FDescription;
    /// <summary>
    ///   Allows you to customize the digital signature image by adding a name,
    ///   signing time, location, reason etc.
    /// </summary>
    property Customizable: Boolean read FCustomizable write FCustomizable default False;
    property CustomData: TfrxCustomSignatureData read FCustomData write FCustomData;
  end;

  /// <summary>
  ///   The TfrxShapeView class represents the "Shape" object. It can show one
  ///   of the following shapes: rectangle, ellipse, rounded rectangle,
  ///   triangle, diamond.
  /// </summary>
  TfrxShapeView = class(TfrxView)
  private
    FCurve: Integer;
    FShape: TfrxShapeKind;
    FRotation: Integer;
    procedure SetRotation(vRot: Integer);
  protected
    function dm: Integer; override; // = 1
  public
    constructor Create(AOwner: TComponent); override;
    constructor DesignCreate(AOwner: TComponent; Flags: Word); override;
    function Diff(AComponent: TfrxComponent): String; override;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;

    class function GetDescription: String; override;
  published
    property Rotation: Integer read FRotation write SetRotation default 0;
    property Color;
    property Cursor;
    /// <summary>
    ///   Curvature of the round rectangle edges.
    /// </summary>
    property Curve: Integer read FCurve write FCurve default 0;
    property FillType;
    property Fill;
    property Frame;
    /// <summary>
    ///   Kind of shape.
    /// </summary>
    property Shape: TfrxShapeKind read FShape write FShape default skRectangle;
    property TagStr;
    property URL;
  end;

  /// <summary>
  ///   The TfrxSubreport class represents the "Subreport" object. Object is
  ///   linked to a report page that stores the objects of subreport. When
  ///   running a report, as soon as the "Subreport" object is encountered, the
  ///   report, allocated to the subreport design page, will be displayed.
  ///   After that, basic report creation will continue.
  /// </summary>
  TfrxSubreport = class(TfrxView)
  private
    FPage: TfrxReportPage;
    FPrintOnParent: Boolean;
    procedure SetPage(const Value: TfrxReportPage);
  protected
    procedure SetParent(AParent: TfrxComponent); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;
    class function GetDescription: String; override;
    procedure ActivatePage;
    procedure UnActivatePage;
  published
    /// <summary>
    ///   Reference to a report page that stores the objects of subreport.
    /// </summary>
    property Page: TfrxReportPage read FPage write SetPage;
    /// <summary>
    ///   Determines whether the subreport contents should be printed on a
    ///   parent band (which contains the "subreport" object). Default value is
    ///   False. In this case the subreport's objects are printed as a <br />
    ///   separate bands and do not affect the parent band.
    /// </summary>
    property PrintOnParent: Boolean read FPrintOnParent write FPrintOnParent
      default False;
  end;


{ Bands }
  TfrxChild = class;

  /// <summary>
  ///   The TfrxBand is the base class for all bands. Band allows creation of a
  ///   report area, which has a specific behaviour according to type.
  /// </summary>
  TfrxBand = class(TfrxReportComponent)
  private
    FAllowSplit: Boolean;
    FChild: TfrxChild;
    FKeepChild: Boolean;
    FOnAfterCalcHeight: TfrxNotifyEvent;
    FOutlineText: String;
    FOverflow: Boolean;
    FStartNewPage: Boolean;
    FStretched: Boolean;
    FPrintChildIfInvisible: Boolean;
    FVertical: Boolean;
    FFill: TfrxCustomFill;
    FFillType: TfrxFillType;
    FFillMemo: TfrxMemoView;
    FFillgap: TfrxFillGaps;
    FFrame: TfrxFrame;
    FBandDesignHeader: Extended;
    FShiftEngine: TfrxShiftEngine;
    function GetBandName: String;
    procedure SetChild(Value: TfrxChild);
    procedure SetVertical(const Value: Boolean);
    procedure SetFill(const Value: TfrxCustomFill);
    procedure SetFillType(const Value: TfrxFillType);
    procedure SetFrame(const Value: TfrxFrame);
    procedure SetGaps(const Value: TfrxFillGaps);
  protected
    function GetBandTitleColor: TColor; virtual;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetLeft(Value: Extended); override;
    procedure SetTop(Value: Extended); override;
    procedure SetHeight(Value: Extended); override;
  public
    /// <summary>
    ///   List of sub bands.
    /// </summary>
    FSubBands: TList;
    FHeader, FFooter, FGroup: TfrxBand;  { h./f./g. bands   }
    /// <summary>
    ///   Used for &lt;Line&gt;.
    /// </summary>
    FLineN: Integer;
    /// <summary>
    ///   Used for &lt;Line#&gt;.
    /// </summary>
    FLineThrough: Integer;
    /// <summary>
    ///   Whether the band should show vertical bands.
    /// </summary>
    FHasVBands: Boolean;
    /// <summary>
    ///   Height of the band after stretch.
    /// </summary>
    FStretchedHeight: Extended;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    /// <summary>
    ///   Returns band index from frxBands array.
    /// </summary>
    function BandNumber: Integer;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;
    /// <summary>
    ///   Creates background object which represent band fill and frame.
    /// </summary>
    procedure CreateFillMemo();
    /// <summary>
    ///   Frees background object which represent band fill and frame.
    /// </summary>
    procedure DisposeFillMemo();
    class function GetDescription: String; override;
    function IsContain(X, Y: Extended): Boolean; override;
    function GetContainedComponent(X, Y: Extended; IsCanContain: TfrxComponent = nil): TfrxComponent; override;
    /// <summary>
    ///   Determines whether the band contents may be splitted if band does not
    ///   find a room in the page. Only "Text", "Line" and "RichText" objects
    ///   may be split. Default value is False.
    /// </summary>
    property AllowSplit: Boolean read FAllowSplit write FAllowSplit default False;
    /// <summary>
    ///   Returns name of the band based on its ClassName (It's not similar to
    ///   Name property).
    /// </summary>
    property BandName: String read GetBandName;
    /// <summary>
    ///   Returns band header height if band caption enabled in the report
    ///   designer.
    /// </summary>
    property BandDesignHeader: Extended read FBandDesignHeader write FBandDesignHeader;
    /// <summary>
    ///   Link to a child band that will be shown after this band.
    /// </summary>
    property Child: TfrxChild read FChild write SetChild;
    /// <summary>
    ///   Determines whether to show a child band on the same page as a parent
    ///   band. If child band does not find a room on the page, it will be
    ///   carried to the next page along with its parent band. Default value is
    ///   False.
    /// </summary>
    property KeepChild: Boolean read FKeepChild write FKeepChild default False;
    /// <summary>
    ///   "Report outline" object represents the "Report outline" control
    ///   element in a preview window. This element displays a treelike
    ///   structure of a finished report. When clicking on any tree node, there
    ///   is a jump to the page connected to this node. <br />All bands have
    ///   the "OutlineText" property, into which a line-expression can be put,
    ///   and this in turn helps to automatically create a tree. The expression
    ///   will be calculated when forming a report, and its value will be added
    ///   to the tree when printing the band. Thus, elements' hierarchy in the
    ///   tree is similar to the bands' hierarchy in a report. That means, that
    ///   in the tree there will be main and subordinate elements,
    ///   corresponding to main and subordinate bands in a report (a report
    ///   with two levels of data or with groups can exemplify the point).
    /// </summary>
    property OutlineText: String read FOutlineText write FOutlineText;
    /// <summary>
    ///   Returns True if band was reprinted on a new page (only for "Group
    ///   header" or "Header" bands with ReprintOnNewPage property set to
    ///   True).
    /// </summary>
    property Overflow: Boolean read FOverflow write FOverflow;
    /// <summary>
    ///   Determines whether to show child band if this band is invisible.
    ///   Default value is False.
    /// </summary>
    property PrintChildIfInvisible: Boolean read FPrintChildIfInvisible
      write FPrintChildIfInvisible default False;
    /// <summary>
    ///   Determines whether to form a new output page before printing this
    ///   band. Default value is False.
    /// </summary>
    property StartNewPage: Boolean read FStartNewPage write FStartNewPage default False;
    /// <summary>
    ///   Determines whether the band can stretch depending on size of its
    ///   objects. To use this property, band should contain stretchable
    ///   objects such as "Text", "Picture", "RichText". Default value is
    ///   False.
    /// </summary>
    property Stretched: Boolean read FStretched write FStretched default False;
    { make IContainer Interface and move some base properties set/get to it }
    /// <summary>
    ///   Determine how band shifts its child object.
    /// </summary>
    property ShiftEngine: TfrxShiftEngine read FShiftEngine write FShiftEngine default seTree;
  published
    property AllowMirrorMode;
    /// <summary>
    ///   Band fill type.
    /// </summary>
    property FillType: TfrxFillType read FFillType write SetFillType;
    /// <summary>
    ///   Reference to a band fill object.
    /// </summary>
    property Fill: TfrxCustomFill read FFill write SetFill;
    /// <summary>
    ///   Reference to an object winch stores fill gaps settings.
    /// </summary>
    property FillGap: TfrxFillGaps read FFillGap write SetGaps;
    /// <summary>
    ///   Reference to band Frame.
    /// </summary>
    property Frame: TfrxFrame read FFrame write SetFrame;
    property Font;
    property Height;
    property Left;
    property ParentFont;
    property Restrictions;
    property Top;
    /// <summary>
    ///   Determines if the band is a vertical band.
    /// </summary>
    property Vertical: Boolean read FVertical write SetVertical default False;
    property Visible;
    property Width;
    /// <summary>
    ///   Name of the script procedure that will be called after calculating
    ///   the band's height.
    /// </summary>
    property OnAfterCalcHeight: TfrxNotifyEvent read FOnAfterCalcHeight
      write FOnAfterCalcHeight;
    property OnAfterPrint;
    property OnBeforePrint;
  end;

  TfrxBandClass = class of TfrxBand;

  /// <summary>
  ///   This flags used in a child band to complete data band upto specified
  ///   number of rows (fill the space with an empty rows). If data band has
  ///   less number of rows, the empty row will be printed.
  /// </summary>
  TfrxToNRowsMode = (
    /// <summary>
    ///   It prints rows until reaches N records count.
    /// </summary>
    rmCount,
    /// <summary>
    ///   It adds N rows to current Data band.
    /// </summary>
    rmAddToCount,
    /// <summary>
    ///   It fills page with "Child" band till the end of the page reached.
    /// </summary>
    rmTillPageEnds);

  /// <summary>
  ///   The TfrxDataBand class is the base class for all data-bands. Data-band
  ///   is a band connected to a dataset. Data-band is displayed as many times
  ///   as there are records in the dataset.
  /// </summary>
  TfrxDataBand = class(TfrxBand)
  private
    FColumnGap: Extended;
    FColumnWidth: Extended;
    FColumns: Integer;
    FCurColumn: Integer;
    FDataSet: TfrxDataSet;
    FDataSetName: String;
    FFilter: String;
    FFooterAfterEach: Boolean;
    FKeepFooter: Boolean;
    FKeepHeader: Boolean;
    FKeepTogether: Boolean;
    FPrintIfDetailEmpty: Boolean;
    FRowCount: Integer;
    FOnMasterDetail: TfrxNotifyEvent;
    FVirtualDataSet: TfrxUserDataSet;
    procedure SetCurColumn(Value: Integer);
    procedure SetRowCount(const Value: Integer);
    procedure SetDataSet(const Value: TfrxDataSet);
    procedure SetDataSetName(const Value: String);
    function GetDataSetName: String;
    function GetVirtualDataSet: TfrxUserDataSet;
  protected
    function GetBandTitleColor: TColor; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    FMaxY: Extended;                             { used for columns }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;
    class function GetDescription: String; override;
    /// <summary>
    ///   Current column of the band.
    /// </summary>
    property CurColumn: Integer read FCurColumn write SetCurColumn;
    property VirtualDataSet: TfrxUserDataSet read GetVirtualDataSet;
  published
    property AllowSplit;
    property Child;
    /// <summary>
    ///   The "Columns" property, available in all data-bands, allows to set
    ///   number of columns for a particular band and not for the whole page.
    ///   Thus, displaying of data will be not "from the top to the bottom,
    ///   then from the left to the right," but "from left to right, then from
    ///   top to bottom". You also should set ColumnWidth and ColumnGap
    ///   properties.
    /// </summary>
    property Columns: Integer read FColumns write FColumns default 0;
    /// <summary>
    ///   Width of the band's column, in pixels.
    /// </summary>
    property ColumnWidth: Extended read FColumnWidth write FColumnWidth;
    /// <summary>
    ///   Gap between band's columns, in pixels.
    /// </summary>
    property ColumnGap: Extended read FColumnGap write FColumnGap;
    /// <summary>
    ///   Dataset for a band.
    /// </summary>
    property DataSet: TfrxDataSet read FDataSet write SetDataSet;
    /// <summary>
    ///   Dataset name for a band. This property duplicates the DataSet
    ///   property.
    /// </summary>
    property DataSetName: String read GetDataSetName write SetDataSetName;
    /// <summary>
    ///   Determines whether to show band footer after each record. Default
    ///   value is False. By default the footer is displayed after printing all
    ///   dataset records.
    /// </summary>
    property FooterAfterEach: Boolean read FFooterAfterEach write FFooterAfterEach default False;
    /// <summary>
    ///   An expression used to manage visibility of a band record. if result
    ///   of this expression is True, band will be shown.
    /// </summary>
    property Filter: String read FFilter write FFilter;
    /// <summary>
    ///   Determines whether to show band footer on the same page as the band
    ///   itself. If footer does not find room on the page, it will be carried
    ///   to the next page along with last record of its parent band. Default
    ///   value is False.
    /// </summary>
    property KeepChild;
    /// <summary>
    ///   Determines whether to show band footer on the same page as the band
    ///   itself. If footer does not find room on the page, it will be carried
    ///   to the next page along with last record of its parent band. Default
    ///   value is False.
    /// </summary>
    property KeepFooter: Boolean read FKeepFooter write FKeepFooter default False;
    /// <summary>
    ///   Determines whether to show band header on the same page as the
    ///   banditself. If header does not find room on the page, it will be
    ///   carriedto the next page. Default value is False.
    /// </summary>
    property KeepHeader: Boolean read FKeepHeader write FKeepHeader default False;
    /// <summary>
    ///   Determines whether to print the band together with all its subbands
    ///   on the same page. Default value is False.
    /// </summary>
    property KeepTogether: Boolean read FKeepTogether write FKeepTogether default False;
    property OutlineText;
    property PrintChildIfInvisible;
    /// <summary>
    ///   Determines whether to print the band if all its subbands are empty.
    ///   Default value is False.
    /// </summary>
    property PrintIfDetailEmpty: Boolean read FPrintIfDetailEmpty
      write FPrintIfDetailEmpty default False;
    /// <summary>
    ///   Allows printing the band N times. This property is used if the band
    ///   is NOT connected to the dataset.
    /// </summary>
    property RowCount: Integer read FRowCount write SetRowCount;
    property StartNewPage;
    property Stretched;
    property ShiftEngine;
    /// <summary>
    ///   Name of the script procedure that will be called before the band will
    ///   check if any subbands are empty.
    /// </summary>
    property OnMasterDetail: TfrxNotifyEvent read FOnMasterDetail write FOnMasterDetail;
  end;

  /// <summary>
  ///   The TfrxHeader class represents a "Header" band. This kind of band is a
  ///   header for all data-bands.
  /// </summary>
  TfrxHeader = class(TfrxBand)
  private
    FReprintOnNewPage: Boolean;
  published
    property AllowSplit;
    property Child;
    property KeepChild;
    property PrintChildIfInvisible;
    /// <summary>
    ///   Determines if it is necessary to reprint a header on each new page.
    /// </summary>
    property ReprintOnNewPage: Boolean read FReprintOnNewPage write FReprintOnNewPage default False;
    property StartNewPage;
    property Stretched;
    property ShiftEngine;
  end;

  /// <summary>
  ///   The TfrxFooter class represents a "Footer" band. This kind of band is a
  ///   footer for all data-bands.
  /// </summary>
  TfrxFooter = class(TfrxBand)
  private
  public
  published
    property AllowSplit;
    property Child;
    property KeepChild;
    property PrintChildIfInvisible;
    property Stretched;
    property ShiftEngine;
  end;

  /// <summary>
  ///   The TfrxMasterData represents a "Master data" band.
  /// </summary>
  TfrxMasterData = class(TfrxDataBand)
  private
  public
  published
  end;

  /// <summary>
  ///   The TfrxDetailData represents a "Detail data" band.
  /// </summary>
  TfrxDetailData = class(TfrxDataBand)
  private
  public
  published
  end;

  /// <summary>
  ///   The TfrxSubDetailData represents a "Subdetail data" band.
  /// </summary>
  TfrxSubdetailData = class(TfrxDataBand)
  private
  public
  published
  end;

  /// <summary>
  ///   The TfrxDataBand4 represents a "4th data level" band.
  /// </summary>
  TfrxDataBand4 = class(TfrxDataBand)
  private
  public
  published
  end;

  /// <summary>
  ///   The TfrxDataBand5 represents a "5th data level" band.
  /// </summary>
  TfrxDataBand5 = class(TfrxDataBand)
  private
  public
  published
  end;

  /// <summary>
  ///   The TfrxDataBand6 represents a "6th data level" band.
  /// </summary>
  TfrxDataBand6 = class(TfrxDataBand)
  private
  public
  published
  end;

  /// <summary>
  ///   The TfrxPageHeader class represents a "Page header" band. This kind of
  ///   band is displayed at the top of each page.
  /// </summary>
  TfrxPageHeader = class(TfrxBand)
  private
    FPrintOnFirstPage: Boolean;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property Child;
    property PrintChildIfInvisible;
    /// <summary>
    ///   Determines whether to print a band on the first page.
    /// </summary>
    property PrintOnFirstPage: Boolean read FPrintOnFirstPage write FPrintOnFirstPage default True;
    property Stretched;
    property ShiftEngine;
  end;

  /// <summary>
  ///   The TfrxPageFooter class represents a "Page footer" band. This kind of
  ///   band is displayed at the bottom of each page.
  /// </summary>
  TfrxPageFooter = class(TfrxBand)
  private
    FPrintOnFirstPage: Boolean;
    FPrintOnLastPage: Boolean;
    FPrintOnSinglePage: Boolean;
  public
    constructor Create(AOwner: TComponent); override;
  published
    /// <summary>
    ///   Determines whether to print a band on the first page.
    /// </summary>
    property PrintOnFirstPage: Boolean read FPrintOnFirstPage write FPrintOnFirstPage default True;
    /// <summary>
    ///   Determines whether to print a band on the last page.
    /// </summary>
    property PrintOnLastPage: Boolean read FPrintOnLastPage write FPrintOnLastPage default True;
    /// <summary>
    ///   Determines whether to print a band on the page when it last and first
    ///   page (i.e. single page).
    /// </summary>
    property PrintOnSinglePage: Boolean read FPrintOnSinglePage write FPrintOnSinglePage default False;
  end;

  /// <summary>
  ///   The TfrxColumnHeader class represents a "Column header" band. This kind
  ///   of band is displayed at the top of each column.
  /// </summary>
  TfrxColumnHeader = class(TfrxBand)
  private
  public
  published
    property Child;
    property Stretched;
    property ShiftEngine;
  end;

  /// <summary>
  ///   The TfrxColumnFooter class represents a "Column footer" band. This kind
  ///   of band is displayed at the bottom of each column.
  /// </summary>
  TfrxColumnFooter = class(TfrxBand)
  private
  public
  published
  end;

  /// <summary>
  ///   The TfrxGroupHeader class represents a "Group header" band. This kind
  ///   of band is displayed at the top of each group.
  /// </summary>
  TfrxGroupHeader = class(TfrxBand)
  private
    FCondition: String;
    FDrillName: String;               { used instead Tag property in drill down }
    FDrillDown: Boolean;
    FExpandDrillDown: Boolean;
    FShowFooterIfDrillDown: Boolean;
    FShowChildIfDrillDown: Boolean;
    FKeepTogether: Boolean;
    FReprintOnNewPage: Boolean;
    FResetPageNumbers: Boolean;
  public
    /// <summary>
    ///   Used to store previous value of group condition. Internal only.
    /// </summary>
    FLastValue: Variant;
    function Diff(AComponent: TfrxComponent): String; override;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;
  published
    property AllowSplit;
    property Child;
    /// <summary>
    ///   A special condition (DB field value or an expression); the band is
    ///   displayed as soon as the condition's value is changed.
    /// </summary>
    property Condition: String read FCondition write FCondition;
    /// <summary>
    ///   Determines whether a group is drill-down group. It means you can
    ///   click on the header in the preview window and group will expand or
    ///   collapse.
    /// </summary>
    property DrillDown: Boolean read FDrillDown write FDrillDown default False;
    /// <summary>
    ///   Determines whether to expand all groups at first run of report. See
    ///   also DrillDown property.
    /// </summary>
    property ExpandDrillDown: Boolean read FExpandDrillDown write FExpandDrillDown default False;
    property KeepChild;
    /// <summary>
    ///   Determines whether to print a group on the same page. If the whole
    ///   group does not find room on the page, it is transferred to a new
    ///   page. Default value is False.
    /// </summary>
    property KeepTogether: Boolean read FKeepTogether write FKeepTogether default False;
    /// <summary>
    ///   Determines whether is necessary to reprint a header on each new page.
    /// </summary>
    property ReprintOnNewPage: Boolean read FReprintOnNewPage write FReprintOnNewPage default False;
    property OutlineText;
    property PrintChildIfInvisible;
    /// <summary>
    ///   Determines whether to reset the page numbers on when group starts.
    ///   You should also set the StartNewPage property.
    /// </summary>
    property ResetPageNumbers: Boolean read FResetPageNumbers write FResetPageNumbers default False;
    /// <summary>
    ///   Determines whether to show the footer of a collapsed group. See also
    ///   DrillDown property.
    /// </summary>
    property ShowFooterIfDrillDown: Boolean read FShowFooterIfDrillDown
      write FShowFooterIfDrillDown default False;
    /// <summary>
    ///   Determines whether to show the child band of a collapsed group. See
    ///   also DrillDown property.
    /// </summary>
    property ShowChildIfDrillDown: Boolean read FShowChildIfDrillDown
      write FShowChildIfDrillDown default False;
    property StartNewPage;
    property Stretched;
    property ShiftEngine;
    /// <summary>
    ///   The name of a drill down group.
    /// </summary>
    property DrillName: String read FDrillName write FDrillName;
  end;

  /// <summary>
  ///   The TfrxGroupFooter class represents a "Group footer" band. This kind
  ///   of band is displayed at the bottom of each group.
  /// </summary>
  TfrxGroupFooter = class(TfrxBand)
  private
    FHideIfSingleDataRecord: Boolean;
  public
  published
    property AllowSplit;
    property Child;
    /// <summary>
    ///   Determines whether to hide a footer if a group has only one data
    ///   record.
    /// </summary>
    property HideIfSingleDataRecord: Boolean read FHideIfSingleDataRecord
      write FHideIfSingleDataRecord default False;
    property KeepChild;
    property PrintChildIfInvisible;
    property Stretched;
    property ShiftEngine;
  end;

  /// <summary>
  ///   The TfrxReportTitle class represents a "Report title" band. This kind
  ///   of band is displayed at the top of the report.
  /// </summary>
  TfrxReportTitle = class(TfrxBand)
  private
  public
  published
    property AllowSplit;
    property Child;
    property KeepChild;
    property PrintChildIfInvisible;
    property StartNewPage;
    property Stretched;
    property ShiftEngine;
  end;

  /// <summary>
  ///   The TfrxReportSummary class represents a "Report summary" band. This
  ///   kind of band is displayed at the end of the report.
  /// </summary>
  TfrxReportSummary = class(TfrxBand)
  private
    FPrintAtBottom: Boolean;
  public
  published
    property AllowSplit;
    property Child;
    property KeepChild;
    property PrintChildIfInvisible;
    /// <summary>
    ///   Determines if summary band should print at bottom of the page.
    /// </summary>
    property PrintAtBottom: Boolean read FPrintAtBottom write FPrintAtBottom default False;
    property StartNewPage;
    property Stretched;
    property ShiftEngine;
  end;

  /// <summary>
  ///   The TfrxChild class represents a "Child" band. This kind of band is
  ///   linked to the parent band via its "Child" property and displayed after
  ///   it.
  /// </summary>
  TfrxChild = class(TfrxBand)
  private
    FToNRows: Integer;
    FToNRowsMode: TfrxToNRowsMode;
  public
  published
    property AllowSplit;
    property Child;
    property KeepChild;
    property PrintChildIfInvisible;
    property StartNewPage;
    property Stretched;
    property ShiftEngine;
    /// <summary>
    ///   Row count used for ToNRowsMode. The value depends on TfrxToNRowsMode
    ///   flag.
    /// </summary>
    property ToNRows: Integer read FToNRows write FToNRows;
    /// <summary>
    ///   This flags used in a child band to complete data band upto specified
    ///   number of rows (fill the space with an empty rows). If data band has
    ///   less number of rows, the empty row will be printed.
    /// </summary>
    property ToNRowsMode: TfrxToNRowsMode read FToNRowsMode write FToNRowsMode;
  end;

  /// <summary>
  ///   The TfrxOverlay class represents an "Overlay" band. This kind of band
  ///   is printed on each page as a bottom layer.
  /// </summary>
  TfrxOverlay = class(TfrxBand)
  private
    FPrintOnTop: Boolean;
  public
  published
    /// <summary>
    ///   Determines whether to print this band on top of other bands.
    /// </summary>
    property PrintOnTop: Boolean read FPrintOnTop write FPrintOnTop default False;
  end;

  TfrxNullBand = class(TfrxBand);


{ Pages }

  /// <summary>
  ///   The TfrxPage class is the base class for report design pages such as
  ///   TfrxReportPage, TfrxDialogPage, TfrxDMPPage.
  /// </summary>
  TfrxPage = class(TfrxComponent)
  private
    FHGuides: TStrings;
    FVGuides: TStrings;
    procedure SetHGuides(const Value: TStrings);
    procedure SetVGuides(const Value: TStrings);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); virtual;
    /// <summary>
    ///   Clears all guide lines. Guide lines are used only in the designer.
    /// </summary>
    procedure ClearGuides;
    function IsSupportsGuidlines: Boolean; virtual;
    destructor Destroy; override;
  published
    property Font;
    property Visible;
    /// <summary>
    ///   Horizontal guide lines. Used only in the designer.
    /// </summary>
    property HGuides: TStrings read FHGuides write SetHGuides;
    /// <summary>
    ///   Vertical guide lines. Used only in the designer.
    /// </summary>
    property VGuides: TStrings read FVGuides write SetVGuides;
  end;

  TfrxSubreportsList = class(TObject)
  private
    FList: TList;
    FActiveSubreport: Integer;
    function GetActiveSubreport: TfrxSubreport;
    procedure SetActiveSubreport(ASubreport: TfrxSubreport);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Add(ASubreport: TfrxSubreport);
    procedure Clear;
    procedure Remove(ASubreport: TfrxSubreport);
    function GetLastSubreport: TfrxSubreport;
    property ActiveSubreport: TfrxSubreport read GetActiveSubreport write SetActiveSubreport;
  end;

  /// <summary>
  ///   The TfrxReportPage class represents a report design page that may
  ///   contain bands and other objects such as "Text", "Picture", etc.
  /// </summary>
  TfrxReportPage = class(TfrxPage)
  private
    FBackPicture: TfrxPictureView;
    FBin: Integer;
    FBinOtherPages: Integer;
    FBottomMargin: Extended;
    FColumns: Integer;
    FColumnWidth: Extended;
    FColumnPositions: TStrings;
    FDataSet: TfrxDataSet;
    FDuplex: TfrxDuplexMode;
    FEndlessHeight: Boolean;
    FEndlessWidth: Boolean;
    FLargeDesignHeight: Boolean;
    FLeftMargin: Extended;
    FMirrorMargins: Boolean;
    FOrientation: TPrinterOrientation;
    FOutlineText: String;
    FPrintIfEmpty: Boolean;
    FPrintOnPreviousPage: Boolean;
    FResetPageNumbers: Boolean;
    FRightMargin: Extended;
    FSubReports: TfrxSubreportsList;
    FTitleBeforeHeader: Boolean;
    FTopMargin: Extended;
    FOnAfterPrint: TfrxNotifyEvent;
    FOnBeforePrint: TfrxNotifyEvent;
    FOnManualBuild: TfrxNotifyEvent;
    FDataSetName: String;
    FBackPictureVisible: Boolean;
    FBackPicturePrintable: Boolean;
    FBackPictureStretched: Boolean;
    FPageCount: Integer;
    FShowTitleOnPreviousPage: Boolean;
    FReport: TfrxReport;
    FMirrorMode: TfrxMirrorControlModes;
    procedure SetPageCount(const Value: Integer);
    procedure SetColumns(const Value: Integer);
    procedure SetOrientation(Value: TPrinterOrientation);

    procedure SetColumnPositions(const Value: TStrings);
    procedure SetFrame(const Value: TfrxFrame);
    function GetFrame: TfrxFrame;
    function GetColor: TColor;
    procedure SetColor(const Value: TColor);
    function GetBackPicture: TPicture;
    procedure SetBackPicture(const Value: TPicture);
    procedure SetDataSet(const Value: TfrxDataSet);
    procedure SetDataSetName(const Value: String);
    function GetDataSetName: String;
    function GetSubReport: TfrxSubreport;
  protected
    FPaperHeight: Extended;
    FPaperSize: Integer;
    FPaperWidth: Extended;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetPaperHeight(const Value: Extended); virtual;
    procedure SetPaperWidth(const Value: Extended); virtual;
    procedure SetPaperSize(const Value: Integer); virtual;
    procedure AddSubreport(ASubreport: TfrxSubreport);
    procedure RemoveSubreport(ASubreport: TfrxSubreport);
    procedure UpdateDimensions;
  public
    /// <summary>
    ///   List of master bands.
    /// </summary>
    FSubBands: TList;
    /// <summary>
    ///   List of vertical master bands.
    /// </summary>
    FVSubBands: TList;
    constructor Create(AOwner: TComponent); overload; override;
    { C++ Builder doesn't support constructor names , use parameters}
    /// <summary>
    ///   Constructor used in preview pages/
    /// </summary>
    /// <param name="AOwner">
    ///   Owner component.
    /// </param>
    /// <param name="AReport">
    ///   Parent report for this page.
    /// </param>
    constructor CreateInPreview(AOwner: TComponent; AReport: TfrxReport); overload;
    destructor Destroy; override;
    class function GetDescription: String; override;
    function GetActiveSubreport: TfrxSubreport;
    /// <summary>
    ///   Search for band of the specified class on the page.
    /// </summary>
    /// <param name="Band">
    ///   Class of the band to search for.
    /// </param>
    function FindBand(Band: TfrxBandClass): TfrxBand;
    /// <summary>
    ///   Returns True, if the report design page is linked to the "Subreport"
    ///   object.
    /// </summary>
    function IsSubReport: Boolean;
    function IsSupportsGuidlines: Boolean; override;
    procedure AlignChildren(IgnoreInvisible: Boolean = False; MirrorModes: TfrxMirrorControlModes = []); override;
    procedure Draw(Canvas: TCanvas; ScaleX, ScaleY, OffsetX, OffsetY: Extended); override;
    procedure SetMarginOffset(PageIndex: Integer);
    procedure ClearMarginOffset;
    /// <summary>
    ///   Sets default page size and margins.
    /// </summary>
    procedure SetDefaults; virtual;
    /// <summary>
    ///   Sets the paper size and dimensions.
    /// </summary>
    procedure SetSizeAndDimensions(ASize: Integer; AWidth, AHeight: Extended);
    /// <summary>
    ///   Reference to a Subreport object.
    /// </summary>
    property SubReport: TfrxSubreport read GetSubReport;
    /// <summary>
    ///   parent report for current page.
    /// </summary>
    property ParentReport: TfrxReport read FReport;
  published
    { paper }
    /// <summary>
    ///   The paper orientation.
    /// </summary>
    property Orientation: TPrinterOrientation read FOrientation
      write SetOrientation default poPortrait;
    /// <summary>
    ///   The paper width, in millimeters.
    /// </summary>
    property PaperWidth: Extended read FPaperWidth write SetPaperWidth;
    /// <summary>
    ///   The paper height, in millimeters.
    /// </summary>
    property PaperHeight: Extended read FPaperHeight write SetPaperHeight;
    /// <summary>
    ///   The paper size. One of the values DMPAPER_xx, defined in the
    ///   Windows.pas file (for example, DMPAPER_A4). When changing this
    ///   property, other properties (PaperWidth, PaperHeight) changed too. If
    ///   you set the DMPAPER_USER paper size, you should set PaperWidth,
    ///   PaperHeight also.
    /// </summary>
    property PaperSize: Integer read FPaperSize write SetPaperSize;
    { margins }
    /// <summary>
    ///   Left margin, in millimeters.
    /// </summary>
    property LeftMargin: Extended read FLeftMargin write FLeftMargin;
    /// <summary>
    ///   Right margin, in millimeters.
    /// </summary>
    property RightMargin: Extended read FRightMargin write FRightMargin;
    /// <summary>
    ///   Top margin, in millimeters.
    /// </summary>
    property TopMargin: Extended read FTopMargin write FTopMargin;
    /// <summary>
    ///   Bottom margin, in millimeters.
    /// </summary>
    property BottomMargin: Extended read FBottomMargin write FBottomMargin;
    /// <summary>
    ///   Determines whether is necessary to switch left and right margins when
    ///   printing even pages. Default value is False.
    /// </summary>
    property MirrorMargins: Boolean read FMirrorMargins write FMirrorMargins
      default False;
    { columns }
    /// <summary>
    ///   Number of columns on a page. If this value is greather than 1, the
    ///   printing will be performed in several columns. You also should set
    ///   the column width in the ColumnWidth property.
    /// </summary>
    property Columns: Integer read FColumns write SetColumns default 0;
    /// <summary>
    ///   Width of a column, in millimeters.
    /// </summary>
    property ColumnWidth: Extended read FColumnWidth write FColumnWidth;
    /// <summary>
    ///   Column starting positions, in millimeters.
    /// </summary>
    property ColumnPositions: TStrings read FColumnPositions write SetColumnPositions;
    { bins }
    /// <summary>
    ///   Paper tray for the first report page. One of the DMBIN_XXX values,
    ///   defined in the Windows.pas file.
    /// </summary>
    property Bin: Integer read FBin write FBin default DMBIN_AUTO;
    /// <summary>
    ///   Paper tray for report pages, except first page. One of the DMBIN_XXX
    ///   values, defined in the Windows.pas file.
    /// </summary>
    property BinOtherPages: Integer read FBinOtherPages write FBinOtherPages
      default DMBIN_AUTO;
    { other }
    /// <summary>
    ///   The page's background picture. It is printed on each page.
    /// </summary>
    property BackPicture: TPicture read GetBackPicture write SetBackPicture;
    /// <summary>
    ///   If the property is equal to True , then the background page layout
    ///   will be displayed while preview.
    /// </summary>
    property BackPictureVisible: Boolean read FBackPictureVisible write FBackPictureVisible default True;
    /// <summary>
    ///   If the property is equal to True, the background page layout will be
    ///   printed.
    /// </summary>
    property BackPicturePrintable: Boolean read FBackPicturePrintable write FBackPicturePrintable default True;
    /// <summary>
    ///   Stretches background picture to page size.
    /// </summary>
    property BackPictureStretched: Boolean read FBackPictureStretched write FBackPictureStretched default True;
    /// <summary>
    ///   The count of printed pages is similar to RowCout of the TfrxDataBand.
    /// </summary>
    property PageCount: Integer read FPageCount write SetPageCount default 1;
    /// <summary>
    ///   Color of the page.
    /// </summary>
    property Color: TColor read GetColor write SetColor default clNone;
    /// <summary>
    ///   Dataset for the report page. If this property is not empty, the page
    ///   will be printed as many times as there are records in the dataset.
    /// </summary>
    property DataSet: TfrxDataSet read FDataSet write SetDataSet;
    /// <summary>
    ///   Name of the dataset for the report page. This property duplicates the
    ///   DataSet property.
    /// </summary>
    property DataSetName: String read GetDataSetName write SetDataSetName;
    /// <summary>
    ///   Duplex mode.
    /// </summary>
    property Duplex: TfrxDuplexMode read FDuplex write FDuplex default dmNone;
    /// <summary>
    ///   Frame of the page.
    /// </summary>
    property Frame: TfrxFrame read GetFrame write SetFrame;
    /// <summary>
    ///   If this property is True, the page height will grow depending on
    ///   number of data records on it.
    /// </summary>
    property EndlessHeight: Boolean read FEndlessHeight write FEndlessHeight default False;
    /// <summary>
    ///   If this property is True, the page width will grow depending on
    ///   number of data records on it.
    /// </summary>
    property EndlessWidth: Boolean read FEndlessWidth write FEndlessWidth default False;
    /// <summary>
    ///   If this property is True, the page will have large height in the
    ///   designer. It is useful if you have many bands in the page.
    /// </summary>
    property LargeDesignHeight: Boolean read FLargeDesignHeight
      write FLargeDesignHeight default False;
    /// <summary>
    ///   Set of Mirror modes for current page. Determinate how engine align
    ///   controls and content base on original layout.
    /// </summary>
    property MirrorMode: TfrxMirrorControlModes read FMirrorMode write FMirrorMode;
    /// <summary>
    ///   "Report outline" object represents the "Report outline" control
    ///   element in a preview window. This element displays a treelike
    ///   structure of a finished report. When clicking on any tree node, there
    ///   is a jump to the page connected to this node.
    /// </summary>
    property OutlineText: String read FOutlineText write FOutlineText;
    /// <summary>
    ///   Determines whether to print a page if all its databands are connected
    ///   to empty datasets. Default value is True.
    /// </summary>
    property PrintIfEmpty: Boolean read FPrintIfEmpty write FPrintIfEmpty default True;
    /// <summary>
    ///   Allows to print pages, beginning from blank space of the previous
    ///   page. This option is used when a report template consists of several
    ///   pages or when printing batch (composite) reports.
    /// </summary>
    property PrintOnPreviousPage: Boolean read FPrintOnPreviousPage
      write FPrintOnPreviousPage default False;
    /// <summary>
    ///   Determines whether to print the "Report title" when
    ///   PrintOnPreviousPage set to true and previous page already has "Report
    ///   title" band.
    /// </summary>
    property ShowTitleOnPreviousPage: Boolean read FShowTitleOnPreviousPage
      write FShowTitleOnPreviousPage default True;
    /// <summary>
    ///   If this property is True, the report engine reset logical page number
    ///   when start printing current report template page.
    /// </summary>
    property ResetPageNumbers: Boolean read FResetPageNumbers
      write FResetPageNumbers default False;
    /// <summary>
    ///   Determines whether to print the "Report title" band before "Page
    ///   header" band. Default value is True.
    /// </summary>
    property TitleBeforeHeader: Boolean read FTitleBeforeHeader
      write FTitleBeforeHeader default True;
    /// <summary>
    ///   Name of the script procedure that will be called after printing the
    ///   page.
    /// </summary>
    property OnAfterPrint: TfrxNotifyEvent read FOnAfterPrint write FOnAfterPrint;
    /// <summary>
    ///   Name of the script procedure that will be called before printing the
    ///   page.
    /// </summary>
    property OnBeforePrint: TfrxNotifyEvent read FOnBeforePrint write FOnBeforePrint;
    /// <summary>
    ///   Name of the script procedure that represents an event handler. If the
    ///   handler of this event is allocated, then the FastReport engine is
    ///   blocked and you thus would have to construct a report manually.
    /// </summary>
    property OnManualBuild: TfrxNotifyEvent read FOnManualBuild write FOnManualBuild;
  end;

  { TfrxDialogPage }

  /// <summary>
  ///   The TfrxDialogPage class represents a dialogue form that may contain
  ///   non-visual and visual controls such as "TfrxBDETable", "Button", etc.
  /// </summary>
  TfrxDialogPage = class(TfrxPage)
  private
    FBorderStyle: TFormBorderStyle;
    FCaption: String;
    FColor: TColor;
    FForm: TForm;
    FOnActivate: TfrxNotifyEvent;
    FOnClick: TfrxNotifyEvent;
    FOnDeactivate: TfrxNotifyEvent;
    FOnHide: TfrxNotifyEvent;
    FOnKeyDown: TfrxKeyEvent;
    FOnKeyPress: TfrxKeyPressEvent;
    FOnKeyUp: TfrxKeyEvent;
    FOnResize: TfrxNotifyEvent;
    FOnShow: TfrxNotifyEvent;
    FOnCloseQuery: TfrxCloseQueryEvent;
    FPosition: TPosition;
    FWindowState: TWindowState;
    FClientWidth: Extended;
    FClientHeight: Extended;
    FCurrentPPI: Integer;
    procedure DoInitialize;
    procedure DoOnActivate(Sender: TObject);
    procedure DoOnClick(Sender: TObject);
    procedure DoOnCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure DoOnDeactivate(Sender: TObject);
    procedure DoOnHide(Sender: TObject);
    procedure DoOnKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure DoOnKeyPress(Sender: TObject; var Key: Char);
    procedure DoOnKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure DoOnShow(Sender: TObject);
    procedure DoOnResize(Sender: TObject);
    procedure DoModify(Sender: TObject);
    procedure DoOnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure DoShowModal;
    procedure SetBorderStyle(const Value: TFormBorderStyle);
    procedure SetCaption(const Value: String);
    procedure SetColor(const Value: TColor);
    function GetModalResult: TModalResult;
    procedure SetModalResult(const Value: TModalResult);
    function GetDoubleBuffered: Boolean;
    procedure SetDoubleBuffered(const Value: Boolean);
    function IsClientSizeStored: Boolean;
  protected
    procedure SetLeft(Value: Extended); override;
    procedure SetTop(Value: Extended); override;
    procedure SetWidth(Value: Extended); override;
    procedure SetHeight(Value: Extended); override;
    procedure SetClientWidth(Value: Extended);
    procedure SetClientHeight(Value: Extended);
    procedure SetScaled(Value: Boolean);
    function GetScaled: Boolean;
    function GetClientWidth: Extended;
    function GetClientHeight: Extended;
    function GetPPIScale: Single;
    procedure FontChanged(Sender: TObject); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    class function GetDescription: String; override;
    /// <summary>
    ///   Initialize dialog page.
    /// </summary>
    procedure Initialize;
    function IsContain(X, Y: Extended): Boolean; override;
    /// <summary>
    ///   Show dialog page as Modal form.
    /// </summary>
    function ShowModal: TModalResult;
    /// <summary>
    ///   Dialog page receives this message when user changes PPI of the screen
    ///   or move dialog form to a screen with different PPI.
    /// </summary>
    procedure UpdateDialogPPI(aNewPPI: Integer); virtual;
    /// <summary>
    ///   Used to validate coordinates of some child objects. FreePacal only.
    /// </summary>
    procedure UpdateControlsCoords;
    /// <summary>
    ///   Link to TForm object.
    /// </summary>
    property DialogForm: TForm read FForm;
    /// <summary>
    ///   Modal result of the form.
    /// </summary>
    property ModalResult: TModalResult read GetModalResult write SetModalResult;
    function IsSupportsGuidlines: Boolean; override;
  published
    /// <summary>
    ///   The style of the window.
    /// </summary>
    property BorderStyle: TFormBorderStyle read FBorderStyle write SetBorderStyle default bsSizeable;
    /// <summary>
    ///   The caption of the window.
    /// </summary>
    property Caption: String read FCaption write SetCaption;
    /// <summary>
    ///   The color of the window.
    /// </summary>
    property Color: TColor read FColor write SetColor default clBtnFace;
    /// <summary>
    ///   Enable double buffering for dialog form.
    /// </summary>
    property DoubleBuffered: Boolean read GetDoubleBuffered write SetDoubleBuffered;
    property Height;
    /// <summary>
    ///   Height of client area.
    /// </summary>
    property ClientHeight: Extended read GetClientHeight write SetClientHeight stored IsClientSizeStored;
    property Left;
    /// <summary>
    ///   The initial position of the window.
    /// </summary>
    property Position: TPosition read FPosition write FPosition default poScreenCenter;
    property Top;
    property Width;
    property Scaled: Boolean read GetScaled write SetScaled;
    /// <summary>
    ///   Width of client area.
    /// </summary>
    property ClientWidth: Extended read GetClientWidth write SetClientWidth stored IsClientSizeStored;
    /// <summary>
    ///   Initial state of the window.
    /// </summary>
    property WindowState: TWindowState read FWindowState write FWindowState default wsNormal;
    /// <summary>
    ///   Name of the script procedure that will be called on dialogue form
    ///   activation.
    /// </summary>
    property OnActivate: TfrxNotifyEvent read FOnActivate write FOnActivate;
    /// <summary>
    ///   Name of the script procedure that will be called on click on the
    ///   form.
    /// </summary>
    property OnClick: TfrxNotifyEvent read FOnClick write FOnClick;
    /// <summary>
    ///   Name of the script procedure that will be called on closing the form.
    ///   The event handler should return True to allow closing.
    /// </summary>
    property OnCloseQuery: TfrxCloseQueryEvent read FOnCloseQuery write FOnCloseQuery;
    /// <summary>
    ///   Name of the script procedure that will be called on deactivating the
    ///   form.
    /// </summary>
    property OnDeactivate: TfrxNotifyEvent read FOnDeactivate write FOnDeactivate;
    /// <summary>
    ///   Name of the script procedure that will be called on hiding the form.
    /// </summary>
    property OnHide: TfrxNotifyEvent read FOnHide write FOnHide;
    /// <summary>
    ///   Name of the script procedure that will be called on pressing a key.
    /// </summary>
    property OnKeyDown: TfrxKeyEvent read FOnKeyDown write FOnKeyDown;
    /// <summary>
    ///   Name of the script procedure that will be called on pressing a key.
    /// </summary>
    property OnKeyPress: TfrxKeyPressEvent read FOnKeyPress write FOnKeyPress;
    /// <summary>
    ///   Name of the script procedure that will be called on releasing a key.
    /// </summary>
    property OnKeyUp: TfrxKeyEvent read FOnKeyUp write FOnKeyUp;
    /// <summary>
    ///   Name of the script procedure that will be called on showing the form.
    /// </summary>
    property OnShow: TfrxNotifyEvent read FOnShow write FOnShow;
    /// <summary>
    ///   Name of the script procedure that will be called on resizing the
    ///   form.
    /// </summary>
    property OnResize: TfrxNotifyEvent read FOnResize write FOnResize;
  end;

  /// <summary>
  ///   The TfrxDataPage represents a datamodule.
  /// </summary>
  TfrxDataPage = class(TfrxPage)
  private
  protected
  public
    constructor Create(AOwner: TComponent); override;
    class function GetDescription: String; override;
  published
    property Height;
    property Left;
    property Top;
    property Width;
  end;


{ Report }

  /// <summary>
  ///   The TfrxEngineOptions class represents a set of properties related to
  ///   the FastReport engine. Instance of this class is stored in the
  ///   TfrxReport.EngineOptions.
  /// </summary>
  TfrxEngineOptions = class(TPersistent)
  private
    const MaxIterations = 10;
  private
    FConvertNulls: Boolean;
    FDestroyForms: Boolean;
    FDoublePass: Boolean;
    FMaxMemSize: Integer;
    FPrintIfEmpty: Boolean;
    FReportThread: TThread;
    FEnableThreadSafe: Boolean;
    FSilentMode: TfrxSilentMode;
    FTempDir: String;
    FUseFileCache: Boolean;
    FUseGlobalDataSetList: Boolean;
    FIgnoreDevByZero: Boolean;
    FIgnoreExprError: Boolean;
    FZeroPrecisionValue: Extended;
    FContentScaleMaxIterations: Integer;

    procedure SetSilentMode(Mode: Boolean);
    function GetSilentMode: Boolean;
    procedure SetContentScaleMaxIterations(const Value: Integer);
  public
    constructor Create;
    /// <summary>
    ///   Copies the contents from another object.
    /// </summary>
    procedure Assign(Source: TPersistent); override;
    /// <summary>
    ///   Copies properties responsible for multi-thread behaviour from another
    ///   object.
    /// </summary>
    procedure AssignThreadProps(Source: TPersistent);
    /// <summary>
    ///   Clears all properties and fills them by default values.
    /// </summary>
    procedure Clear;
    /// <summary>
    ///   Deprecated. Not used.
    /// </summary>
    property ReportThread: TThread read FReportThread write FReportThread;
    /// <summary>
    ///   Determines whether to save the report to the temporary stream before
    ///   run a report and restore it after report is finished. Default is
    ///   True.
    /// </summary>
    property DestroyForms: Boolean read FDestroyForms write FDestroyForms;
    /// <summary>
    ///   Switch TfrxReport component in multi-thread mode. It disables unsafe
    ///   routines like ProcessMessages cycle. <br />
    /// </summary>
    property EnableThreadSafe: Boolean read FEnableThreadSafe write FEnableThreadSafe;
    /// <summary>
    ///   The property determines if its necessary to use the global list of
    ///   DataSet or to use the list of EnabledDataSet collection of TfrxReport
    ///   component. Default-True.
    /// </summary>
    property UseGlobalDataSetList: Boolean read FUseGlobalDataSetList write FUseGlobalDataSetList;
    /// <summary>
    ///   This property contains precision value for zero comparison operation
    ///   in Text objects(used with HideZeros property). The value in Text
    ///   object below precision value interpret as zero. Default is 1E-17.
    /// </summary>
    property ZeroPrecisionValue: Extended read FZeroPrecisionValue write FZeroPrecisionValue;
  published
    /// <summary>
    ///   Converts the "Null" value of the DB field into "0," "False," or empty
    ///   string, depending on the field type. Default value is True.
    /// </summary>
    property ConvertNulls: Boolean read FConvertNulls write FConvertNulls default True;
    /// <summary>
    ///   Makes a report a two-pass one. Default value is False.
    /// </summary>
    property DoublePass: Boolean read FDoublePass write FDoublePass default False;
    /// <summary>
    ///   The maximum size of memory in Mbytes, allocated to the report pages'
    ///   cache. It becomes useful in cases when the "UseFileCache" property is
    ///   equal to "True." If a report begins to occupy more memory during
    ///   construction, caching of the constructed report pages into a
    ///   temporary file is performed. This property is inexact and allows only
    ///   approximate determination of the memory limit. Default value is 10.
    /// </summary>
    property MaxMemSize: Integer read FMaxMemSize write FMaxMemSize default 10;
    /// <summary>
    ///   Defines, whether it is necessary to print a blank report (one which
    ///   contains no data lines). Default value is True.
    /// </summary>
    property PrintIfEmpty: Boolean read FPrintIfEmpty write FPrintIfEmpty default True;
    /// <summary>
    ///   The maximum value of iterations that can be used to scale down objects content
    ///   when dynamic auto scale mode active for one of the report object.
    /// </summary>
    property ContentScaleMaxIterations: Integer read FContentScaleMaxIterations write SetContentScaleMaxIterations default MaxIterations;
    /// <summary>
    ///   Deprecated(See NewSilentMode). "Silent" mode. When errors occurred
    ///   during loading or <br />running report, no dialog windows will be
    ///   shown. All errors will be <br />contained in the TfrxReport.Errors
    ///   property. This mode is useful for <br />server applications. Default
    ///   value is False. <br />
    /// </summary>
    property SilentMode: Boolean read GetSilentMode write SetSilentMode default False;
    /// <summary>
    ///   Set behaviour of exceptions handling during report execution.
    /// </summary>
    property NewSilentMode: TfrxSilentMode read FSilentMode write FSilentMode default simMessageBoxes;
    /// <summary>
    ///   Specifies a path to the directory for storing temporary files.
    /// </summary>
    property TempDir: String read FTempDir write FTempDir;
    /// <summary>
    ///   Defines, whether it is necessary to use report pages caching into a
    ///   file (see the "MaxMemSize" property). Default value is False.
    /// </summary>
    property UseFileCache: Boolean read FUseFileCache write FUseFileCache default False;
    /// <summary>
    ///   Ignore all "Division by zero" exceptions inside expression
    ///   calculation.
    /// </summary>
    property IgnoreDevByZero: Boolean read FIgnoreDevByZero write FIgnoreDevByZero default False;
    /// <summary>
    ///   Ignore all expression exceptions inside expression calculation.
    /// </summary>
    property IgnoreExprError: Boolean read FIgnoreExprError write FIgnoreExprError default False;
  end;

  /// <summary>
  ///   The TfrxPrintOptions class represents a set of properties, which relate
  ///   to the report printing. Instance of this class is stored in the
  ///   TfrxReport.PrintOptions.
  /// </summary>
  TfrxPrintOptions = class(TPersistent)
  private
    FCopies: Integer;
    FCollate: Boolean;
    FPageNumbers: String;
    FPagesOnSheet: Integer;
    FPrinter: String;
    FPrintMode: TfrxPrintMode;
    FPrintOnSheet: Integer;
    FPrintPages: TfrxPrintPages;
    FReverse: Boolean;
    FShowDialog: Boolean;
    FSwapPageSize: Boolean;
    FPrnOutFileName: String;
    FDuplex: TfrxDuplexMode;
    FSplicingLine: Integer;
  public
    constructor Create;
    /// <summary>
    ///   Copies the contents from another object.
    /// </summary>
    procedure Assign(Source: TPersistent); override;
    /// <summary>
    ///   Clears all properties and fills them by default values.
    /// </summary>
    procedure Clear;
    /// <summary>
    ///   File name of printed output when prints to a file.
    /// </summary>
    property PrnOutFileName: String read FPrnOutFileName write FPrnOutFileName;
    /// <summary>
    ///   Global duplex settings. Set only after prepare report.
    /// </summary>
    property Duplex: TfrxDuplexMode read FDuplex write FDuplex;// set only after prepare report, need to store global duplex
    /// <summary>
    ///   Size of slicing line used for some Print modes.
    /// </summary>
    property SplicingLine: Integer read FSplicingLine write FSplicingLine default 3;
  published
    /// <summary>
    ///   The number of the printable copies by default. Default value is 1.
    /// </summary>
    property Copies: Integer read FCopies write FCopies default 1;
    /// <summary>
    ///   Whether to collate the copies. Default value is True.
    /// </summary>
    property Collate: Boolean read FCollate write FCollate default True;
    /// <summary>
    ///   Page numbers, which are to be printed. For example, "1,3,5-12,17-".
    /// </summary>
    property PageNumbers: String read FPageNumbers write FPageNumbers;
    /// <summary>
    ///   Printer name.
    /// </summary>
    property Printer: String read FPrinter write FPrinter;
    /// <summary>
    ///   The print mode: default; split large page on several smaller pages;
    ///   join several small pages on a large page; print a page on a specified
    ///   paper size (use scaling).
    /// </summary>
    property PrintMode: TfrxPrintMode read FPrintMode write FPrintMode default pmDefault;
    /// <summary>
    ///   The paper size to print a report on. Used if PrintMode is not
    ///   pmDefault. One of DMBIN_XXX values defined in the Windows.pas.
    /// </summary>
    property PrintOnSheet: Integer read FPrintOnSheet write FPrintOnSheet;
    /// <summary>
    ///   Defines the pages to be printed. The following values are available: <br />
    ///   ppAll - all; <br />ppOdd - odd; <br />ppEven - even. <br />
    /// </summary>
    property PrintPages: TfrxPrintPages read FPrintPages write FPrintPages default ppAll;
    /// <summary>
    ///   Determines whether to print pages in reverse order.
    /// </summary>
    property Reverse: Boolean read FReverse write FReverse default False;
    /// <summary>
    ///   Whether to display a print dialogue. Default value is True.
    /// </summary>
    property ShowDialog: Boolean read FShowDialog write FShowDialog default True;
    /// <summary>
    ///   Not used.
    /// </summary>
    property SwapPageSize: Boolean read FSwapPageSize write FSwapPageSize stored False;// remove it
  end;

  /// <summary>
  ///   The TfrxPreviewOptions class represents a set of properties, relating
  ///   to the report preview. Instance of this class is stored in the
  ///   TfrxReport.PreviewOptions.
  /// </summary>
  TfrxPreviewOptions = class(TPersistent)
  private
    FAllowEdit: Boolean;
    FAllowPreviewEdit: Boolean;
    FButtons: TfrxPreviewButtons;
    FDoubleBuffered: Boolean;
    FMaximized: Boolean;
    FMDIChild: Boolean;
    FModal: Boolean;
    FOutlineExpand: Boolean;
    FOutlineVisible: Boolean;
    FOutlineWidth: Integer;
    FPagesInCache: Integer;
    FShowCaptions: Boolean;
    FThumbnailVisible: Boolean;
    FZoom: Extended;
    FZoomMode: TfrxZoomMode;
    FPictureCacheInFile: Boolean;
    FRTLPreview: Boolean;
  public
    constructor Create;
    /// <summary>
    ///   Copies the contents from another object.
    /// </summary>
    procedure Assign(Source: TPersistent); override;
    /// <summary>
    ///   Clears all properties and fills them by default values.
    /// </summary>
    procedure Clear;
    /// <summary>
    ///   Enables RTL page output in preview.
    /// </summary>
    property RTLPreview: Boolean read FRTLPreview write FRTLPreview;
  published
    /// <summary>
    ///   Enables or disables a finished report editing. Default value is True.
    /// </summary>
    property AllowEdit: Boolean read FAllowEdit write FAllowEdit default True;
    /// <summary>
    ///   Enables InPlace editors in the report preview. Allows user to edit
    ///   prepared report text and other properties without the report
    ///   designer.
    /// </summary>
    property AllowPreviewEdit: Boolean read FAllowPreviewEdit write FAllowPreviewEdit default True;
    /// <summary>
    ///   A set of buttons, which will be available in the preview window. You
    ///   can combine any of these values. Default value is: all buttons.
    /// </summary>
    property Buttons: TfrxPreviewButtons read FButtons write FButtons;
    /// <summary>
    ///   A double-buffer mode for the preview window. If enabled (by default),
    ///   the preview window will not flicker during repainting, but the
    ///   process speed will be reduced.
    /// </summary>
    property DoubleBuffered: Boolean read FDoubleBuffered write FDoubleBuffered default True;
    /// <summary>
    ///   Defines if the preview window is maximized. Default value is True.
    /// </summary>
    property Maximized: Boolean read FMaximized write FMaximized default True;
    /// <summary>
    ///   Defines if the preview window is MDIChild (for MDI interface
    ///   organizing). Default value is False.
    /// </summary>
    property MDIChild: Boolean read FMDIChild write FMDIChild default False;
    /// <summary>
    ///   Defines if the preview window is modal. Default value is True.
    /// </summary>
    property Modal: Boolean read FModal write FModal default True;
    /// <summary>
    ///   Determines whether to expand all outline items.
    /// </summary>
    property OutlineExpand: Boolean read FOutlineExpand write FOutlineExpand default True;
    /// <summary>
    ///   Defines if the panel with the report outline is visible. Default
    ///   value is False.
    /// </summary>
    property OutlineVisible: Boolean read FOutlineVisible write FOutlineVisible default False;
    /// <summary>
    ///   Defines width of the panel with the report outline, in pixels.
    ///   Default value is 120.
    /// </summary>
    property OutlineWidth: Integer read FOutlineWidth write FOutlineWidth default 120;
    /// <summary>
    ///   Determines how many prepared pages are located in memory. Other
    ///   stored as serialized XML and loads by request.
    /// </summary>
    property PagesInCache: Integer read FPagesInCache write FPagesInCache default 50;
    /// <summary>
    ///   Defines if the panel with the report thumbnail is visible. Default
    ///   value is False.
    /// </summary>
    property ThumbnailVisible: Boolean read FThumbnailVisible write FThumbnailVisible default False;
    /// <summary>
    ///   Defines if button captions are displayed. When enabling thisproperty,
    ///   you should limit the number of the displayed buttons in theButtons
    ///   property, since all the buttons will not find room on thescreen.
    /// </summary>
    property ShowCaptions: Boolean read FShowCaptions write FShowCaptions default False;
    /// <summary>
    ///   Default scale of the report in the report preview.
    /// </summary>
    property Zoom: Extended read FZoom write FZoom;
    /// <summary>
    ///   Default zooming mode.
    /// </summary>
    property ZoomMode: TfrxZoomMode read FZoomMode write FZoomMode default zmDefault;
    /// <summary>
    ///   Activates the mode of caching all pictures in the file, saves memory
    ///   while displaying the reports with a large quantity of pictures.
    /// </summary>
    property PictureCacheInFile: Boolean read FPictureCacheInFile write FPictureCacheInFile default False;
  end;

  /// <summary>
  ///   The TfrxReportOptions class represents a set of properties relating to
  ///   the report. Instance of this class is stored in the
  ///   TfrxReport.ReportOptions.
  /// </summary>
  TfrxReportOptions = class(TPersistent)
  private
    FAuthor: String;
    FCompressed: Boolean;
    FConnectionName: String;
    FCreateDate: TDateTime;
    FDescription: TStrings;
    FInitString: String;
    FName: String;
    FLastChange: TDateTime;
    FPassword: String;
    FPicture: TPicture;
    FReport: TfrxReport;
    FVersionBuild: String;
    FVersionMajor: String;
    FVersionMinor: String;
    FVersionRelease: String;
    FPrevPassword: String;
    FHiddenPassword: String;
    FInfo: Boolean;
    procedure SetDescription(const Value: TStrings);
    procedure SetPicture(const Value: TPicture);
    procedure SetConnectionName(const Value: String);
  public
    constructor Create(AOwner: TComponent);
    destructor Destroy; override;
    /// <summary>
    ///   Copies the contents from another object.
    /// </summary>
    procedure Assign(Source: TPersistent); override;
    /// <summary>
    ///   Clears all properties and fills them by default values.
    /// </summary>
    procedure Clear;
    /// <summary>
    ///   If HiddenPassword is not set, shows password dialog and check
    ///   password inputted by user. Otherwise uses HiddenPassword as a user
    ///   defined password. <br />
    /// </summary>
    function CheckPassword: Boolean;
    /// <summary>
    ///   Stores previous used password.
    /// </summary>
    property PrevPassword: String write FPrevPassword;
    property Info: Boolean read FInfo write FInfo;
    /// <summary>
    ///   Allows to set password which will be used for report in LoadFrom
    ///   method without showing dialog to user.
    /// </summary>
    property HiddenPassword: String read FHiddenPassword write FHiddenPassword;
  published
    /// <summary>
    ///   Report author.
    /// </summary>
    property Author: String read FAuthor write FAuthor;
    /// <summary>
    ///   Determines whether to compress a report file. The gzip compression is
    ///   used. You must put the TfrxGZipCompressor component onto your form.
    /// </summary>
    property Compressed: Boolean read FCompressed write FCompressed default False;
    /// <summary>
    ///   Not used.
    /// </summary>
    property ConnectionName: String read FConnectionName write SetConnectionName;
    /// <summary>
    ///   Report creation date.
    /// </summary>
    property CreateDate: TDateTime read FCreateDate write FCreateDate;
    /// <summary>
    ///   Report description.
    /// </summary>
    property Description: TStrings read FDescription write SetDescription;
    /// <summary>
    ///   Dot-matrix printer init string. Used only in dot-matrix reports.
    /// </summary>
    property InitString: String read FInitString write FInitString;
    /// <summary>
    ///   Report name.
    /// </summary>
    property Name: String read FName write FName;
    /// <summary>
    ///   The date the report was last modified.
    /// </summary>
    property LastChange: TDateTime read FLastChange write FLastChange;
    /// <summary>
    ///   Report password. If this property is not blank, a password is
    ///   required when opening a report.
    /// </summary>
    property Password: String read FPassword write FPassword;
    /// <summary>
    ///   Report picture.
    /// </summary>
    property Picture: TPicture read FPicture write SetPicture;
    /// <summary>
    ///   Properties, which detect the number of a version.
    /// </summary>
    property VersionBuild: String read FVersionBuild write FVersionBuild;
    /// <summary>
    ///   Properties, which detect the number of a version.
    /// </summary>
    property VersionMajor: String read FVersionMajor write FVersionMajor;
    /// <summary>
    ///   Properties, which detect the number of a version.
    /// </summary>
    property VersionMinor: String read FVersionMinor write FVersionMinor;
    /// <summary>
    ///   Properties, which detect the number of a version.
    /// </summary>
    property VersionRelease: String read FVersionRelease write FVersionRelease;
  end;


  /// <summary>
  ///   The TfrxReportOptions class represents an expression cache. The
  ///   instance of this class used in the report object to store most used
  ///   expressions and link them with script items. It significantly increases
  ///   expression calculation (like aggregate functions) in a huge reports
  ///   because syntax parsing of expressions called only once.
  /// </summary>
  TfrxExpressionCache = class(TObject)
  private
    FExpressions: TStringList;
    FMainScript: TfsScript;
    FScript: TfsScript;
    FScriptLanguage: String;
    procedure SetCaseSensitive(const Value: Boolean);
    function GetCaseSensitive: Boolean;
  public
    /// <summary>
    ///   Constructor winch receives default script object.
    /// </summary>
    /// <param name="AScript">
    ///   Default script object.
    /// </param>
    constructor Create(AScript: TfsScript);
    destructor Destroy; override;
    /// <summary>
    ///   Clears expression cache.
    /// </summary>
    procedure Clear;
    /// <summary>
    ///   Calculates an expression.
    /// </summary>
    /// <param name="Expression">
    ///   Expression to calculate.
    /// </param>
    /// <param name="ErrorMsg">
    ///   Out. Contains error message if error occurred.
    /// </param>
    /// <param name="AScript">
    ///   Script object used to calculate an expression.
    /// </param>
    function Calc(const Expression: String; var ErrorMsg: String;
      AScript: TfsScript): Variant;
    /// <summary>
    ///   If True , then the expression will be catching subject to the letter
    ///   list . It is necessary to use the same text with different letter
    ///   list in expression.
    /// </summary>
    property CaseSensitive: Boolean read GetCaseSensitive write SetCaseSensitive;
  end;

  /// <summary>
  ///   The TfrxDataSetItem class represents a dataset in the report dataset's
  ///   list. Dataset's list is contained in the TfrxReportDataSets class.
  /// </summary>
  TfrxDataSetItem = class(TCollectionItem)
  private
    FDataSet: TfrxDataSet;
    FDataSetName: String;
    procedure SetDataSet(const Value: TfrxDataSet);
    procedure SetDataSetName(const Value: String);
    function GetDataSetName: String;
  published
    /// <summary>
    ///   Reference to the dataset.
    /// </summary>
    property DataSet: TfrxDataSet read FDataSet write SetDataSet;
    /// <summary>
    ///   Name of the dataset. This property duplicates the DataSet property.
    /// </summary>
    property DataSetName: String read GetDataSetName write SetDataSetName;
  end;

  /// <summary>
  ///   The TfrxReportDataSets class represents the report dataset's list.
  ///   Instance of this class is stored in the TfrxReport.Datasets property.
  /// </summary>
  TfrxReportDataSets = class(TCollection)
  private
    FReport: TfrxReport;
    function GetItem(Index: Integer): TfrxDataSetItem;
  public
    /// <summary>
    ///   Creates the class instance. Do not use this method directly, because
    ///   instance of this class is stored in the TfrxReport.Datasets property.
    /// </summary>
    constructor Create(AReport: TfrxReport);
    /// <summary>
    ///   Initializes dataset list.
    /// </summary>
    procedure Initialize;
    /// <summary>
    ///   Finalize all datasets in the list.
    /// </summary>
    procedure Finalize;
    /// <summary>
    ///   Adds an element to the list.
    /// </summary>
    procedure Add(ds: TfrxDataSet); overload;
    procedure Add(AName: String); overload;
    /// <summary>
    ///   Finds an element in the list.
    /// </summary>
    function Find(ds: TfrxDataSet): TfrxDataSetItem; overload;
    /// <summary>
    ///   Finds an element in the list.
    /// </summary>
    function Find(const Name: String): TfrxDataSetItem; overload;
    /// <summary>
    ///   Delete an element in the list.
    /// </summary>
    procedure Delete(const Name: String); overload;
    /// <summary>
    ///   List of elements.
    /// </summary>
    property Items[Index: Integer]: TfrxDataSetItem read GetItem; default;
  end;

  TfrxStyleSheets = class;

  TfrxNamedCollectionItem = class(TfrCollectionItem)
  private
    FName: String;
    procedure SetName(const Value: String);
  protected
    function GetInheritedName: String; override;
    procedure NameChange(const ANewName: string; const AOldName: string); virtual;
  public
    /// <summary>
    ///   Creates an unique name of style.
    /// </summary>
    procedure CreateUniqueName;
    property Name: String read FName write SetName;
  end;

  TfrxNamedCollection = class(TfrCollection)
  private
    function GetItem(Index: Integer): TfrxNamedCollectionItem;
  protected
    function InternalFind(const Name: String): TfrxNamedCollectionItem;
    property Items[Index: Integer]: TfrxNamedCollectionItem read GetItem; default;
  end;

  /// <summary>
  ///   The TfrxCustomStyleItem represents a base style. Style is an element, which
  ///   possesses a name and properties, and determines design attributes, i.e.
  ///   color, font and frame. The style determines the way a report object
  ///   should be designed. The objects such as TfrxMemoView have the Style
  ///   property, which is a property intended to set the style name. When
  ///   applying a value to this property, the style design attributes are
  ///   copied to the object.
  /// </summary>
  TfrxCustomStyleItem = class(TfrxNamedCollectionItem)
  private
    FFont: TFont;
    FFrame: TfrxFrame;
    FApplyFont: Boolean;
    FApplyFill: Boolean;
    FApplyFrame: Boolean;
    FFill: TfrxCustomFill;
    procedure SetFont(const Value: TFont);
    procedure SetFrame(const Value: TfrxFrame);
    function GetFillType: TfrxFillType;
    procedure SetFill(const Value: TfrxCustomFill);
    procedure SetFillType(const Value: TfrxFillType);
    function GetColor: TColor;
    procedure SetColor(const Value: TColor);
    function GetFill: TfrxCustomFill;
    function GetFont: TFont;
    function GetFrame: TfrxFrame;
  protected
    function GetFontRef: TFont; virtual;
    function GetFillRef: TfrxCustomFill; virtual;
    procedure SetFillRef(const Value: TfrxCustomFill); virtual;
    function GetFrameRef: TfrxFrame; virtual;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
  published
    /// <summary>
    ///   Style name.
    /// </summary>
    property Name;
    /// <summary>
    ///   Deprecated. Background color.
    /// </summary>
    property Color: TColor read GetColor write SetColor stored False;
    /// <summary>
    ///   Font settings.
    /// </summary>
    property Font: TFont read GetFont write SetFont;
    /// <summary>
    ///   Frame settings.
    /// </summary>
    property Frame: TfrxFrame read GetFrame write SetFrame;
    /// <summary>
    ///   Apply font from the style to an object.
    /// </summary>
    property ApplyFont: Boolean read FApplyFont write FApplyFont default True;
    /// <summary>
    ///   Apply fill from the style to an object.
    /// </summary>
    property ApplyFill: Boolean read FApplyFill write FApplyFill default True;
    /// <summary>
    ///   Apply frame from the style to an object.
    /// </summary>
    property ApplyFrame: Boolean read FApplyFrame write FApplyFrame default True;
 	  /// <summary>
 	  ///   Fill type.
 	  /// </summary>
 	  property FillType: TfrxFillType read GetFillType write SetFillType default ftBrush;
    /// <summary>
    ///   Reference to Fill object.
    /// </summary>
    property Fill: TfrxCustomFill read GetFill write SetFill;
  end;

  /// <summary>
  ///   The TfrxStyleItem represents a style. Style is an element, which
  ///   possesses a name and properties, and determines design attributes, i.e.
  ///   color, font and frame. The style determines the way a report object
  ///   should be designed. The objects such as TfrxMemoView have the Style
  ///   property, which is a property intended to set the style name. When
  ///   applying a value to this property, the style design attributes are
  ///   copied to the object.
  /// </summary>
  TfrxStyleItem = class(TfrxCustomStyleItem)
  private
    FStyleSheet: TfrxStyleSheets;
    procedure SetStyleSheet(const Value: TfrxStyleSheets);
    function IsStyleSheetActive: Boolean;
    function GetActiveStyleSheet: String;
    function IsStyleSheetStored: Boolean;
  protected
    function GetFontRef: TFont; override;
    function GetFillRef: TfrxCustomFill; override;
    procedure SetFillRef(const Value: TfrxCustomFill); override;
    function GetFrameRef: TfrxFrame; override;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure WriteProperties(Writer: TObject; Ancestor: TfrCollection; Owner: TComponent); override;
    procedure ReadProperties(PropName: String; Reader: TObject; Ancestor: TfrCollection); override;
    /// <summary>
    ///   Updates style sheets of an element from a list.
    /// </summary>
    procedure UpdateSheets(AList: TStrings);
  published
    /// <summary>
    ///   Reference to StyleSheet collection object.
    /// </summary>
    property StyleSheet: TfrxStyleSheets read FStyleSheet write SetStyleSheet stored IsStyleSheetStored;
  end;

  TfrxStyleBookMode = (fsbDesigner, fsbPreview);
  TfrxStyleBookModes = set of TfrxStyleBookMode;

  /// <summary>
  ///   The TfrxStyles class represents a set of styles. It comprises methods
  ///   for performing such set operations as reading, saving, adding,
  ///   deleting, as well as searching for a style. The set of styles file has
  ///   FS3 extension by default.
  /// </summary>
  TfrxStyles = class(TfrxNamedCollection)
  private
    FName: String;
    FReport: TfrxReport;
    FStyleSheetNames: TStringList;
    FActiveStyleSheet: String;
    FDefaultStyleLocked: Boolean;
    FStyleBookMode: TfrxStyleBookModes;
    function GetItem(Index: Integer): TfrxStyleItem;
    procedure UpdateSheets(AList: TStrings);
    procedure SetActiveStyleSheet(const Value: String);
    function IsStyleSheetActive: Boolean;
    procedure InternalSynchSheetName(const ANewName: string; const AOldName: string);
    procedure SetStyleBookMode(const Value: TfrxStyleBookModes);
    function GetActiveStyleSheet: String;
  protected
    property UpdateCount;
    procedure LockDefaultStyle;
    procedure UnlockDefaultStyle;
  public
    /// <summary>
    ///   Creates the styles set. One can specify "nil" instead of "AReport,"
    ///   however in this case a user would be unable to use the "Apply"
    ///   method.
    /// </summary>
    constructor Create(AReport: TfrxReport);
    destructor Destroy; override;
    /// <summary>
    ///   Adds a new style.
    /// </summary>
    function Add: TfrxStyleItem;
     /// <summary>
    ///   Adds a new style sheet with generated name.
    /// </summary>
    function AddStyleSheet: String;
    procedure Assign(Source: TPersistent); override;
    /// <summary>
    ///   Assigns while style collection with style sheets to a preview pages styles.
    /// </summary>
    procedure AssignAsPreview(Source: TfrxStyles);
    procedure ClearAll;
    procedure EndUpdate; override;
    /// <summary>
    ///   Returns the style with the given name.
    /// </summary>
    function Find(const AName: String): TfrxStyleItem;
    /// <summary>
    ///   Applies a set to a report.
    /// </summary>
    procedure Apply;
    /// <summary>
    ///   Deletes  StyleSheet with given name from a collection
    /// </summary>
    procedure DeleteStyleSheet(const AName: String);
    /// <summary>
    ///   Returns the list of the styles names.
    /// </summary>
    procedure GetList(List: TStrings);
    /// <summary>
    ///   Returns the list of the style sheets names.
    /// </summary>
    procedure GetStyleSheetList(List: TStrings);
    /// <summary>
    ///   Reads a set from a file.
    /// </summary>
    procedure LoadFromFile(const FileName: String);
    /// <summary>
    ///   Reads a set from a stream.
    /// </summary>
    procedure LoadFromStream(Stream: TStream);
    /// <summary>
    ///   Reads a set from a XML item.
    /// </summary>
    procedure LoadFromXMLItem(AItem: TfrXMLItem; AOldXMLFormat: Boolean = True); overload;
    /// <summary>
    ///   Reads a set from a XML item.
    /// </summary>
    procedure LoadFromXMLItem(AItem: TfrXMLItem; AAncestor: TfrxStyles; AOldXMLFormat: Boolean = True); overload;
    /// <summary>
    ///   Saves a set to a file.
    /// </summary>
    procedure SaveToFile(const FileName: String);
    /// <summary>
    ///   Saves a set to a stream.
    /// </summary>
    procedure SaveToStream(Stream: TStream);
    /// <summary>
    ///   Saves a set to a XML item.
    /// </summary>
    procedure SaveToXMLItem(AItem: TfrXMLItem); overload;
    /// <summary>
    ///   Saves a set to a XML item.
    /// </summary>
    procedure SaveToXMLItem(AItem: TfrXMLItem; AAncestor: TfrxStyles); overload;
    /// <summary>
    ///   Renames StyleSheet with a name to a new name.
    /// </summary>
    function RenameStyleSheet(const AName: String; const ANewName: String): Boolean;
    /// <summary>
    ///   Assigns appearance from an objects with assigned styles to style elements
    /// </summary>
    procedure UpdateFromObjects(AList: TList);
    /// <summary>
    ///   Active style sheet in current styles collection
    /// </summary>
    property ActiveStyleSheet: String read GetActiveStyleSheet write SetActiveStyleSheet;
    /// <summary>
    ///   List of styles in the set.
    /// </summary>
    property Items[Index: Integer]: TfrxStyleItem read GetItem; default;
    /// <summary>
    ///   A set's name.
    /// </summary>
    property Name: String read FName write FName;
  published
    property StyleBookMode: TfrxStyleBookModes read FStyleBookMode write SetStyleBookMode default [];
  end;

  TfrxStyleSheetItem = class(TfrxCustomStyleItem)
  protected
    procedure NameChange(const ANewName: string; const AOldName: string); override;
  end;

  TfrxStyleSheets = class(TfrxNamedCollection)
  private
    FName: String;
    FStyles: TfrxStyles;
    FParentStyle: TfrxStyleItem;
    FCachedItem: TfrxStyleSheetItem;
    function GetItem(Index: Integer): TfrxStyleSheetItem;
  public
    /// <summary>
    ///   Creates the styles set. One can specify "nil" instead of "AReport,"
    ///   however in this case a user would be unable to use the "Apply"
    ///   method.
    /// </summary>
    constructor Create(AStyles: TfrxStyles; AParentStyle: TfrxStyleItem);
    /// <summary>
    ///   Adds a new style.
    /// </summary>
    function Add: TfrxStyleSheetItem;
    procedure AddSheet(const AName: String);
    procedure DeleteByName(const AName: String);
    function GetSheetItem(const AName: String): TfrxStyleSheetItem;
    /// <summary>
    ///   Returns the style with the given name.
    /// </summary>
    function Find(const Name: String): TfrxStyleSheetItem;
    procedure UpdateSheets(AList: TStrings);
    /// <summary>
    ///   List of styles in the set.
    /// </summary>
    property Items[Index: Integer]: TfrxStyleSheetItem read GetItem; default;
    /// <summary>
    ///   A set's name.
    /// </summary>
    property Name: String read FName write FName;
  end;

  /// <summary>
  ///   The TfrxStyleSheet class represents a styles' library. It has methods
  ///   for the library reading/saving, as well as adding, deleting, and style
  ///   sets' searching.
  /// </summary>
  TfrxStyleScheme = class(TObject)
  private
    FItems: TList;
    function GetItems(Index: Integer): TfrxStyles;
  public
    /// <summary>
    ///   Constructs a library.
    /// </summary>
    constructor Create;
    destructor Destroy; override;
    /// <summary>
    ///   Clears a library.
    /// </summary>
    procedure Clear;
    /// <summary>
    ///   Deletes a set with certain number.
    /// </summary>
    procedure Delete(Index: Integer);
    /// <summary>
    ///   Returns the list of the names of styles' sets.
    /// </summary>
    procedure GetList(List: TStrings);
    /// <summary>
    ///   Loads a library from a file.
    /// </summary>
    procedure LoadFromFile(const FileName: String);
    /// <summary>
    ///   Loads a library from a stream.
    /// </summary>
    procedure LoadFromStream(Stream: TStream);
    /// <summary>
    ///   Saves a library to a file.
    /// </summary>
    procedure SaveToFile(const FileName: String);
    /// <summary>
    ///   Saves a library to a stream.
    /// </summary>
    procedure SaveToStream(Stream: TStream);
    /// <summary>
    ///   Adds a new set of styles to the library.
    /// </summary>
    function Add: TfrxStyles;
    /// <summary>
    ///   Returns a number of styles' sets in the library.
    /// </summary>
    function Count: Integer;
    /// <summary>
    ///   Returns a set with the set name.
    /// </summary>
    function Find(const Name: String): TfrxStyles;
    /// <summary>
    ///   Returns a set number with the given name.
    /// </summary>
    function IndexOf(const Name: String): Integer;
    /// <summary>
    ///   The list of styles' sets.
    /// </summary>
    property Items[Index: Integer]: TfrxStyles read GetItems; default;
  end;

  /// <summary>
  ///   The TfrxPreviewPagesList class represents list of
  ///   TfrxCustomPreviewPages. Each TfrxReport instance can have several
  ///   TfrxCustomPreviewPages instances. This class used to manage all preview
  ///   pages in the report component.
  /// </summary>
  TfrxPreviewPagesList = class
  private
    FReport: TfrxReport;
    { TODO: Make Named list, can use it latet in server components }
    FList: TList;
    FCurrentIndex: Integer;
    function GetCurrent: TfrxCustomPreviewPages;
    function GetMain: TfrxCustomPreviewPages;
    procedure SetCurrent(const Value: TfrxCustomPreviewPages);
    procedure SetCurrentIndex(Index: Integer);
    procedure SetMain(const Value: TfrxCustomPreviewPages);
    procedure CurPagesChanged;
  public
    /// <summary>
    ///   Constructor.
    /// </summary>
    /// <param name="aReport">
    ///   The report component.
    /// </param>
    constructor Create(aReport: TfrxReport);
    destructor Destroy; override;
    /// <summary>
    ///   Clears and free all preview pages instances instead main.
    /// </summary>
    procedure Clear;
    /// <summary>
    ///   Returns count of preview pages in collection.
    /// </summary>
    function Count: Integer;
    /// <summary>
    ///   Adds new instance of PreviewPages.
    /// </summary>
    function Add: TfrxCustomPreviewPages;
    /// <summary>
    ///   Adds new instance of PreviewPages and set it active(current).
    /// </summary>
    function AddAsCurrent: TfrxCustomPreviewPages;
    /// <summary>
    ///   Delete instance of TfrxCustomPreviewPages from list.
    /// </summary>
    /// <param name="aPages">
    ///   Reference to instance that should be deleted.
    /// </param>
    procedure Delete(aPages: TfrxCustomPreviewPages); overload;
    /// <summary>
    ///   Delete instance of TfrxCustomPreviewPages from list by index.
    /// </summary>
    /// <param name="Index">
    ///   Index in list to delete.
    /// </param>
    procedure Delete(Index: Integer); overload;
    /// <summary>
    ///   Stores current active preview pages. TfrxReport uses this property
    ///   for any operations with preview pages.
    /// </summary>
    property Current: TfrxCustomPreviewPages read GetCurrent write SetCurrent;
    /// <summary>
    ///   Stores main preview pages. This pages could not be deleted by clear
    ///   method, it frees at destruction.
    /// </summary>
    property Main: TfrxCustomPreviewPages read GetMain write SetMain;
  end;

{$IFDEF DELPHI16}
/// <summary>
///   The TfrxReport component is the main one. One TfrxReport component
///   contains one report. In design-time, double-clicking the component calls
///   the report designer. The component has all necessary properties and
///   methods for report loading and saving, design and viewing.
/// </summary>
[ComponentPlatformsAttribute(pidWin32 or pidWin64)]
{$ENDIF}
  TfrxReport = class(TfrxComponent)
  private
    FCurObject: String;
    FDataSet: TfrxDataSet;
    FDataSetName: String;
    FDataSets: TfrxReportDatasets;
    FDesigner: TfrxCustomDesigner;
    FDotMatrixReport: Boolean;
    FDrawText: Pointer;
    FDrillState: TStrings;
    FEnabledDataSets: TfrxReportDataSets;
    FEngine: TfrxCustomEngine;
    FEngineOptions: TfrxEngineOptions;
    FErrors: TStrings;
    FExpressionCache: TfrxExpressionCache;
    FFileName: String;
    FIniFile: String;
    FLoadStream: TStream;
    FLocalValue: TfsCustomVariable;
    FSelfValue: TfsCustomVariable;
    FModified: Boolean;
    FOldStyleProgress: Boolean;
    FParentForm: TForm;
    FParentReport: String;
    FParentReportObject: TfrxReport;
    FPreviewPagesList: TfrxPreviewPagesList;
    FPreview: TfrxCustomPreview;
    FPreviewForm: TForm;
    FPreviewOptions: TfrxPreviewOptions;
    FPrintOptions: TfrxPrintOptions;
    FProgress: TfrxProgress;
    FReloading: Boolean;
    FReportOptions: TfrxReportOptions;
    FScript: TfsScript;
    FSaveParentScript: TfsScript;
    FScriptLanguage: String;
    FScriptText: TStrings;
    FFakeScriptText: TStrings; {fake object}
    FShowProgress: Boolean;
    FStoreInDFM: Boolean;
    FStyles: TfrxStyles;
    FSysVariables: TStrings;
    FTerminated: Boolean;
    FTimer: TTimer;
    FVariables: TfrxVariables;
    FVersion: String;
    FXMLSerializer: TObject;
    FStreamLoaded: Boolean;
    FIsScriptObjectsAdded: Boolean;
    FCommandManager: TfrxDelayedCommandManager;
    FPreparing: Boolean;

    FOnAfterPrint: TfrxBeforePrintEvent;
    FOnAfterPrintReport: TNotifyEvent;
    FOnBeforeConnect: TfrxBeforeConnectEvent;
    FOnAfterDisconnect: TfrxAfterDisconnectEvent;
    FOnBeforePrint: TfrxBeforePrintEvent;
    FOnBeginDoc: TNotifyEvent;
    FOnClickObject: TfrxClickObjectEvent;
    FOnDblClickObject: TfrxClickObjectEvent;
    FOnEditConnection: TfrxEditConnectionEvent;
    FOnEndDoc: TNotifyEvent;
    FOnGetValue: TfrxGetValueEvent;
    FOnNewGetValue: TfrxNewGetValueEvent;
    FOnLoadTemplate: TfrxLoadTemplateEvent;
    FOnLoadDetailTemplate: TfrxLoadDetailTemplateEvent;
    FOnManualBuild: TfrxManualBuildEvent;
    FOnObjectManualBuild: TfrxObjectManualBuildEvent;
    FOnMouseOverObject: TfrxMouseOverObjectEvent;
    FOnPreview: TNotifyEvent;
    FOnPrintPage: TfrxPrintPageEvent;
    FOnPrintReport: TNotifyEvent;
    FOnProgressStart: TfrxProgressEvent;
    FOnProgress: TfrxProgressEvent;
    FOnProgressStop: TfrxProgressEvent;
    FOnRunDialogs: TfrxRunDialogsEvent;
    FOnSetConnection: TfrxSetConnectionEvent;
    FOnStartReport: TfrxNotifyEvent;
    FOnStopReport: TfrxNotifyEvent;
    FOnUserFunction: TfrxUserFunctionEvent;
    FOnClosePreview: TNotifyEvent;
    FOnReportPrint: TfrxNotifyEvent;
    FOnAfterScriptCompile: TNotifyEvent;
    FOnMouseEnter: TfrxMouseEnterEvent;
    FOnMouseLeave: TfrxMouseLeaveEvent;

    FOnGetCustomDataEvent: TfrxGetCustomDataEvent;
    FOnSaveCustomDataEvent: TfrxSaveCustomDataEvent;

    function CallMethod(Instance: TObject; ClassType: TClass;
      const MethodName: String; var Params: Variant): Variant;
    function DoGetValue(const Expr: String; var Value: Variant): Boolean;
    function GetScriptValue(Instance: TObject; ClassType: TClass;
      const MethodName: String; var Params: Variant): Variant;
    function SetScriptValue(Instance: TObject; ClassType: TClass;
      const MethodName: String; var Params: Variant): Variant;
    function DoUserFunction(Instance: TObject; ClassType: TClass;
      const MethodName: String; var Params: Variant): Variant;
    function GetDataSetName: String;
    function GetLocalValue: Variant;
    function GetSelfValue: TfrxView;
    function GetPages(Index: Integer): TfrxPage;
    function GetPagesCount: Integer;
    function GetCaseSensitive: Boolean;
    function GetScriptText: TStrings;
    procedure AncestorNotFound(Reader: TReader; const ComponentName: string;
      ComponentClass: TPersistentClass; var Component: TComponent);
    procedure DoClear;
    procedure DoLoadFromStream;
    procedure OnTimer(Sender: TObject);
    procedure ReadDatasets(Reader: TReader);
    procedure ReadStyle(Reader: TReader);
    procedure ReadStyleBookMode(Reader: TReader);
    procedure ReadActiveStyleSheet(Reader: TReader);
    procedure ReadVariables(Reader: TReader);
    procedure SetScriptVar(const vName: String; Obj: TObject);
    procedure SetDataSet(const Value: TfrxDataSet);
    procedure SetDataSetName(const Value: String);
    procedure SetEngineOptions(const Value: TfrxEngineOptions);
    procedure SetSelfValue(const Value: TfrxView);
    procedure SetLocalValue(const Value: Variant);
    procedure SetParentReport(const Value: String);
    procedure SetPreviewOptions(const Value: TfrxPreviewOptions);
    procedure SetPrintOptions(const Value: TfrxPrintOptions);
    procedure SetReportOptions(const Value: TfrxReportOptions);
    procedure SetScriptText(const Value: TStrings);
    procedure SetStyles(const Value: TfrxStyles);
    procedure SetTerminated(const Value: Boolean);
    procedure SetCaseSensitive(const Value: Boolean);
    procedure WriteDatasets(Writer: TWriter);
    procedure WriteStyle(Writer: TWriter);
    procedure WriteStyleBookMode(Writer: TWriter);
    procedure WriteActiveStyleSheet(Writer: TWriter);
    procedure WriteVariables(Writer: TWriter);
    procedure SetPreview(const Value: TfrxCustomPreview);
    procedure SetVersion(const Value: String);
    procedure SetPreviewPages(const Value: TfrxCustomPreviewPages);
    {$IFDEF FPC}
    function GetLazIniFile:string;
    {$ENDIF}
    procedure UpdateGlobalScritVars;
    procedure SetMainPreviewPages(const Value: TfrxCustomPreviewPages);
    procedure FixUpParentReport;
    function GetMainPreviewPages: TfrxCustomPreviewPages;
    function GetPreviewPages: TfrxCustomPreviewPages;
    procedure RestoreScriptObjectsRef;
    function GetPictureCacheOptions: TfrxPictureCacheOptions;
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DefineProperties(Filer: TFiler); override;
    procedure DoGetAncestor(const Name: String; var Ancestor: TPersistent);
    procedure PreviewPagesChanged;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Clear; override;
    class function GetDescription: String; override;
    procedure WriteNestedProperties(AItem: TfrXMLItem; AAncestor: TPersistent = nil); override;
    function ReadNestedProperty(AItem: TfrXMLItem): Boolean; override;

    { internal methods }
    /// <summary>
    ///   Calculates the "Expr" expression and returns the result.
    /// </summary>
    function Calc(const Expr: String; AScript: TfsScript = nil): Variant;
    /// <summary>
    ///   Calls designer for set report editing (deletes bands, blocks inlay
    ///   code e.t.c).
    /// </summary>
    function DesignPreviewPage: Boolean;
    /// <summary>
    ///   Returns an alias for the given data set. Dataset should be included
    ///   in the TfrxReport.Datasets.
    /// </summary>
    function GetAlias(DataSet: TfrxDataSet): String;
    /// <summary>
    ///   Returns the data set on its alias name. Dataset should be included in
    ///   the TfrxReport.Datasets.
    /// </summary>
    function GetDataset(const Alias: String): TfrxDataset;
    /// <summary>
    ///   Returns reference to current DrawText object. For thread safe saferty
    ///   each report instance has own DrawText instance.
    /// </summary>
    function GetReportDrawText: Pointer;
    /// <summary>
    ///   Returns the ini-file instance that stores all the designer settings.
    ///   It may be either ini-file or registry, depending on IniFile property
    ///   value.
    /// </summary>
    function GetIniFile: TCustomIniFile;
    /// <summary>
    ///   This function returns path to application executable.
    /// </summary>
    function GetApplicationFolder: String;
    /// <summary>
    ///   Searches for an object by name in the current report template.
    /// </summary>
    /// <param name="AName">
    ///   Object name.
    /// </param>
    function FindObject(const AName: String): TfrxComponent; override;
    /// <summary>
    ///   Prepares script engine for work. It compiles script code, adds report
    ///   objects and user functions to script engine.
    /// </summary>
    /// <param name="ActiveReport">
    ///   Reference to report component(by default nil).
    /// </param>
    function PrepareScript(ActiveReport: TfrxReport = nil): Boolean;
    /// <summary>
    ///   The current loaded template inherits from the specified . The first
    ///   parameter is the name of the parent template, the second one is an
    ///   inheritance mode: derive inheritance dialogue, delete duplicates ,
    ///   rename duplicates .
    /// </summary>
    function InheritFromTemplate(const templName: String; InheriteMode: TfrxInheriteMode = imDefault): Boolean;
    /// <summary>
    ///   Serves IDE report designer.
    /// </summary>
    procedure DesignReport(IDesigner: {$IFDEF FPC}TObject{$ELSE}IUnknown{$ENDIF}; Editor: TObject); overload;
    /// <summary>
    ///   Script Notify event handler.
    /// </summary>
    procedure DoNotifyEvent(Obj: TObject; const EventName: String;
      RunAlways: Boolean = False);
    /// <summary>
    ///   Script parameterized event handler.
    /// </summary>
    procedure DoParamEvent(const EventName: String; var Params: Variant;
      RunAlways: Boolean = False);
    /// <summary>
    ///   Serves AfterPrint event.
    /// </summary>
    procedure DoAfterPrint(c: TfrxReportComponent);
    /// <summary>
    ///   Serves BeforePrint event.
    /// </summary>
    procedure DoBeforePrint(c: TfrxReportComponent);
    /// <summary>
    ///   Serves PreviewClick event.
    /// </summary>
    procedure DoPreviewClick(v: TfrxView; Button: TMouseButton;
      Shift: TShiftState; var Modified: Boolean; var EventParams: TfrxInteractiveEventsParams; DblClick: Boolean = False);
    /// <summary>
    ///   Serves MouseEnter script event.
    /// </summary>
    procedure DoMouseEnter(Sender: TfrxView; aPreviousObject: TfrxComponent; var EventParams: TfrxInteractiveEventsParams); reintroduce;
    /// <summary>
    ///   Serves MouseLeave script event.
    /// </summary>
    procedure DoMouseLeave(Sender: TfrxView; X, Y: Integer; aNextObject: TfrxComponent; var EventParams: TfrxInteractiveEventsParams); reintroduce;
    /// <summary>
    ///   Serves several mouse script events (UP/DOWN).
    /// </summary>
    function DoBaseMouseEvents(AMouseEventType: TfrxMouseIntEvents; ASender: TfrxView; X, Y: Integer; Button: TMouseButton; Shift: TShiftState; var EventParams: TfrxInteractiveEventsParams): Boolean;
    /// <summary>
    ///   Serves MouseMove script event.
    /// </summary>
    procedure DoBaseMouseMove(Sender: TfrxView; X, Y: Integer; Shift: TShiftState; var EventParams: TfrxInteractiveEventsParams);
    /// <summary>
    ///   Retrieves dataset and field name from an expression like
    ///   DSName."field name".
    /// </summary>
    /// <param name="ComplexName">
    ///   Name of dataset and field.
    /// </param>
    /// <param name="Dataset">
    ///   Our dataset reference.
    /// </param>
    /// <param name="Field">
    ///   Our field reference.
    /// </param>
    procedure GetDatasetAndField(const ComplexName: String;
      var Dataset: TfrxDataset; var Field: String);
    /// <summary>
    ///   Returns the list of the data sets available in the report in the
    ///   "List" parameter. The second parameter specifies whether it is
    ///   necessary to return only the data sets, connected with DB data.
    /// </summary>
    procedure GetDataSetList(List: TStrings; OnlyDB: Boolean = False);
    /// <summary>
    ///   Fill list with active datasets.
    /// </summary>
    /// <param name="List">
    ///   Out list.
    /// </param>
    procedure GetActiveDataSetList(List: TStrings);
    /// <summary>
    ///   The report engine calls it when start some work over
    ///   report(preparation, print, export).
    /// </summary>
    procedure InternalOnProgressStart(ProgressType: TfrxProgressType); virtual;
    /// <summary>
    ///   The report engine calls it when some work over report is in
    ///   progress(preparation, print, export).
    /// </summary>
    procedure InternalOnProgress(ProgressType: TfrxProgressType; Progress: Integer); virtual;
    /// <summary>
    ///   The report engine calls it when some work over report is in
    ///   finished(preparation, print, export).
    /// </summary>
    procedure InternalOnProgressStop(ProgressType: TfrxProgressType); virtual;
    /// <summary>
    ///   Returns True when report is preparing or script routine is
    ///   executing(interactive event).
    /// </summary>
    function IsReportActionRunning: Boolean;
    /// <summary>
    ///   Selects a printer specified in PrintOptions.Printer property.
    /// </summary>
    procedure SelectPrinter;
    /// <summary>
    ///   Sets a message to progress bar window if it's active or for status
    ///   bar text in the report preview.
    /// </summary>
    /// <param name="Value">
    ///   Message text.
    /// </param>
    /// <param name="Ishint">
    ///   Set as hint.
    /// </param>
    /// <param name="bHandleMessage">
    ///   Use HandleMessage for main thread GUI synchronization.
    /// </param>
    procedure SetProgressMessage(const Value: String; Ishint: Boolean = False; bHandleMessage: Boolean = True);
    /// <summary>
    ///   This method checks for DataPage in current report and if it's not
    ///   exist, creates one.
    /// </summary>
    procedure CheckDataPage;
    { public methods }
    /// <summary>
    ///   Loads a report from a file with given name. If the second parameter
    ///   is equal to "True," and the file was not found, an exception will be
    ///   generated. It returns "True," if the file is loaded successfully.
    /// </summary>
    /// <param name="FileName">
    ///   Name of the file.
    /// </param>
    /// <param name="ExceptionIfNotFound">
    ///   If true, function generates an exception if file does not exist.
    /// </param>
    function LoadFromFile(const FileName: String;
      ExceptionIfNotFound: Boolean = False): Boolean;
    /// <summary>
    ///   Loads a report from the stream.
    /// </summary>
    /// <param name="Stream">
    ///   Source stream.
    /// </param>
    procedure LoadFromStream(Stream: TStream); override;
    /// <summary>
    ///   Loads a report from using IO transport class to get stream of a
    ///   report.
    /// </summary>
    /// <param name="Filter">
    ///   IO filter transport instance used to get a stream data.
    /// </param>
    /// <param name="FileName">
    ///   Name used if filter doesn't return any name.
    /// </param>
    function LoadFromFilter(Filter: TfrxCustomIOTransport; const FileName: String): Boolean;
    /// <summary>
    ///   Loads or saves report by using IO transport.
    /// </summary>
    /// <param name="Filter">
    ///   IO transport used to get stream for save and load operations.
    /// </param>
    /// <param name="FileName">
    ///   File name used to identify stream if IO filter doesn't have name.
    /// </param>
    /// <param name="fAccess">
    ///   Access type of file operation, read or write.
    /// </param>
    function ProcessIOTransport(Filter: TfrxCustomIOTransport; const FileName: String; fAccess: TfrxFilterAccess): Boolean;
    /// <summary>
    ///   Saves a report from using IO transport class to get stream of a
    ///   report.
    /// </summary>
    /// <param name="Filter">
    ///   IO filter transport instance used to get a stream data.
    /// </param>
    /// <param name="FileName">
    ///   File name used to identify stream if IO filter doesn't have name.
    /// </param>
    function SaveToFilter(Filter: TfrxCustomIOTransport; const FileName: String): Boolean;
    /// <summary>
    ///   Saves a report to a file with given name.
    /// </summary>
    /// <param name="FileName">
    ///   Name of a file.
    /// </param>
    procedure SaveToFile(const FileName: String);
    /// <summary>
    ///   Saves a report to the stream. SaveChildren, SaveDefaultValues,
    ///   UseGetAncestor parameters are not used.
    /// </summary>
    /// <param name="Stream">
    ///   Source stream.
    /// </param>
    /// <param name="SaveChildren">
    ///   Not used.
    /// </param>
    /// <param name="SaveDefaultValues">
    ///   Not used. <br />
    /// </param>
    procedure SaveToStream(Stream: TStream; SaveChildren: Boolean = True;
      SaveDefaultValues: Boolean = False; UseGetAncestor: Boolean = False); override;
    /// <summary>
    ///   Calls the report designer. The designer component must be included
    ///   into your project (to perform this, you can either place the
    ///   "TfrxDesigner" component on the form , or include the "frxDesgn" unit
    ///   into the "Uses" list). The Modal parameter determines whether the
    ///   designer should be modal. The MDIChild parameter allows to make a
    ///   designer window a MDI child window.
    /// </summary>
    /// <param name="Modal">
    ///   Show as modal form.
    /// </param>
    /// <param name="MDIChild">
    ///   Shows as MDI child form.
    /// </param>
    procedure DesignReport(Modal: Boolean = True; MDIChild: Boolean = False); overload; {$IFDEF UNIX}cdecl;{$ELSE}stdcall;{$ENDIF}
    /// <summary>
    ///   Starts a report without a preview window. The parameter assignment is
    ///   the same as in the "ShowReport" method. If a report is constructed
    ///   successfully, this function returns "True."
    /// </summary>
    /// <param name="ClearLastReport">
    ///   If True, clears previous prepared report result.
    /// </param>
    /// <remarks>
    ///   When the report component builds a report and receives another
    ///   PrepareReport or Export command it puts it in deferred commands list
    ///   and run a command only after current action is complete. Used in
    ///   single-threaded application for synchronization with ProcessMessages
    ///   actions.
    /// </remarks>
    function PrepareReport(ClearLastReport: Boolean = True): Boolean;
  	/// <summary>
  	///   Prepares specified report template page in current report.
  	/// </summary>
  	/// <param name="APage">
  	///   A report template page.
  	/// </param>
  	function PreparePage(APage: TfrxPage): Boolean;
    /// <summary>
    ///   Runs all deferred report commands (prepare report, export, print).
    ///   Used in single thread application. When the report component builds a
    ///   report and receives another PrepareReport or Export command it puts
    ///   it in deferred commands list and run a command only after current
    ///   action is complete.
    /// </summary>
    procedure RunDelayedCommands;
    /// <summary>
    ///   This method rebuilds current active preview pages.
    /// </summary>
    procedure RefreshActivePreviewedReport;
    /// <summary>
    ///   Displays the report, which was previously built via the
    ///   "PrepareReport" call.
    /// </summary>
    /// <remarks>
    ///   When the report component builds a report and receives another
    ///   PrepareReport or Export command it puts it in deferred commands list
    ///   and run a command only after current action is complete. Used in
    ///   single-threaded application for synchronization with ProcessMessages
    ///   actions.
    /// </remarks>
    procedure ShowPreparedReport; {$IFDEF UNIX}cdecl;{$ELSE}stdcall;{$ENDIF}
    /// <summary>
    ///   Starts a report and displays a result in the preview window. If the
    ///   "ClearLastReport" parameter is equal to "False," then a report will
    ///   be added to the previously constructed one, otherwise the previously
    ///   constructed report is cleared (by default).
    /// </summary>
    /// <param name="ClearLastReport">
    ///   Clear previous result.
    /// </param>
    /// <remarks>
    ///   When the report component builds a report and receives another
    ///   PrepareReport or Export command it puts it in deferred commands list
    ///   and run a command only after current action is complete. Used in
    ///   single-threaded application for synchronization with ProcessMessages
    ///   actions.
    /// </remarks>
    procedure ShowReport(ClearLastReport: Boolean = True); {$IFDEF UNIX}cdecl;{$ELSE}stdcall;{$ENDIF}
    /// <summary>
    ///   Adds a user function to the list of the functions available in the
    ///   report. See the FastReport developer's manual about this method.
    /// </summary>
    /// <param name="FuncName">
    ///   Function name with parameters ('function MyFunc(s: String):String'). <br />
    /// </param>
    /// <param name="Category">
    ///   Function category.
    /// </param>
    /// <param name="Description">
    ///   Description of function behaviour.
    /// </param>
    procedure AddFunction(const FuncName: String; const Category: String = '';
      const Description: String = '');
    /// <summary>
    ///   Shows designer in specified panel.
    /// </summary>
    /// <param name="Panel">
    ///   Parent panel.
    /// </param>
    procedure DesignReportInPanel(Panel: TWinControl);
    /// <summary>
    ///   Prints a report using printer settings from a PrintOptions property.
    /// </summary>
    /// <remarks>
    ///   When the report component builds a report and receives another
    ///   PrepareReport or Export command it puts it in deferred commands list
    ///   and run a command only after current action is complete. Used in
    ///   single-threaded application for synchronization with ProcessMessages
    ///   actions.
    /// </remarks>
    function Print: Boolean; {$IFDEF UNIX}cdecl;{$ELSE}stdcall;{$ENDIF}
    /// <summary>
    ///   Exports a report using the specified export filter.
    /// </summary>
    /// <param name="Filter">
    ///   Export filter instance.
    /// </param>
    /// <remarks>
    ///   When the report component builds a report and receives another
    ///   PrepareReport or Export command it puts it in deferred commands list
    ///   and run a command only after current action is complete. Used in
    ///   single-threaded application for synchronization with ProcessMessages
    ///   actions.
    /// </remarks>
    function Export(Filter: TfrxCustomExportFilter): Boolean;

    { internals }
    /// <summary>
    ///   Reference to a current processed object by report engine. The engine
    ///   sets it before process report object.
    /// </summary>
    property CurObject: String read FCurObject write FCurObject;
    /// <summary>
    ///   The list with drill down states names. Groups names in this list are
    ///   expanded. To collapse or expand groups remove or add group name to
    ///   the list.
    /// </summary>
    property DrillState: TStrings read FDrillState;
    /// <summary>
    ///   Current value used for &lt;Value&gt; variable inside expressions.
    /// </summary>
    property LocalValue: Variant read GetLocalValue write SetLocalValue;
    /// <summary>
    ///   Current value used for &lt;Self&gt; variable inside expressions.
    /// </summary>
    property SelfValue: TfrxView read GetSelfValue write SetSelfValue;
    /// <summary>
    ///   Current preview from if preview for is open.
    /// </summary>
    property PreviewForm: TForm read FPreviewForm;
    /// <summary>
    ///   Reference to active XML serializer instance.
    /// </summary>
    property XMLSerializer: TObject read FXMLSerializer;
    /// <summary>
    ///   True when report is reloading(When component loads report).
    /// </summary>
    property Reloading: Boolean read FReloading write FReloading;

    { public }
    /// <summary>
    ///   The list of data sets available for the report.
    /// </summary>
    property DataSets: TfrxReportDataSets read FDataSets;
    /// <summary>
    ///   A link to the report designer (actual when the designer is enabled).
    /// </summary>
    property Designer: TfrxCustomDesigner read FDesigner write FDesigner;
    /// <summary>
    ///   The list of enabled data sets available for the report. When
    ///   EngineOptions.UseGlobalDataSetList is True, th report component uses
    ///   this list instead of Global.
    /// </summary>
    property EnabledDataSets: TfrxReportDataSets read FEnabledDataSets;
    /// <summary>
    ///   A link to the report engine. It is useful in cases when it is
    ///   necessary to operate a process of the report construction from a
    ///   code.
    /// </summary>
    property Engine: TfrxCustomEngine read FEngine;
    /// <summary>
    ///   The list of errors, occurring during one or another operation.
    /// </summary>
    property Errors: TStrings read FErrors;
    /// <summary>
    ///   Defines a report file name; it is displayed in the designer. The
    ///   LoadFromFile method fills this property.
    /// </summary>
    property FileName: String read FFileName write FFileName;
    /// <summary>
    ///   This property indicates that report was modified by the report
    ///   designer.
    /// </summary>
    property Modified: Boolean read FModified write FModified;
    /// <summary>
    ///   Reference to main preview pages instance of finished report belongs
    ///   to the report. <br />
    /// </summary>
    property MainPreviewPages: TfrxCustomPreviewPages read GetMainPreviewPages write SetMainPreviewPages;
    /// <summary>
    ///   A link to current pages of a finished report. It is used in all
    ///   operations, which are connected with a finished report (printing,
    ///   saving, export, etc.).
    /// </summary>
    property PreviewPages: TfrxCustomPreviewPages read GetPreviewPages write SetPreviewPages;
    /// <summary>
    ///   List of preview pages generated by current instance of the report.
    ///   This class used to manage all previewpages in the report component.
    /// </summary>
    property PreviewPagesList: TfrxPreviewPagesList read FPreviewPagesList;
    /// <summary>
    ///   The list of report pages. The pages of the "dialogue form" type are
    ///   also included in this list.
    /// </summary>
    property Pages[Index: Integer]: TfrxPage read GetPages;
    /// <summary>
    ///   Number of pages in a report template.
    /// </summary>
    property PagesCount: Integer read GetPagesCount;
    property PictureCacheOptions: TfrxPictureCacheOptions read GetPictureCacheOptions;
    /// <summary>
    ///   The "TfsScript" component, linked to the report. By using this link,
    ///   you can add your variables, classes, functions, which can be used in
    ///   a report script in the future. See more in the FastScript developer's
    ///   manual.
    /// </summary>
    property Script: TfsScript read FScript;
    /// <summary>
    ///   Report style. See more about operating with styles in the FastReport
    ///   programmer's manual.
    /// </summary>
    property Styles: TfrxStyles read FStyles write SetStyles;
    /// <summary>
    ///   Indicates that report was terminated during preparation. Set it to
    ///   true manually will cause exit from the report preparation state. The
    ///   termination is not instant.
    /// </summary>
    property Terminated: Boolean read FTerminated write SetTerminated;
    /// <summary>
    ///   The list of report variables. See more about operating with variables
    ///   in the FastReport programmer's manual.
    /// </summary>
    property Variables: TfrxVariables read FVariables;
    /// <summary>
    ///   If True , then the expression will be catching subject to the letter
    ///   list . It is necessary to use the same text with different letter
    ///   list in expression.
    /// </summary>
    property CaseSensitiveExpressions: Boolean read GetCaseSensitive write SetCaseSensitive;

    /// <summary>
    ///   Occurs when user edit connection in the report wizard.
    /// </summary>
    property OnEditConnection: TfrxEditConnectionEvent read FOnEditConnection write FOnEditConnection;
    /// <summary>
    ///   Occurs when user set ConnectionNeme property. Internal use.
    /// </summary>
    property OnSetConnection: TfrxSetConnectionEvent read FOnSetConnection write FOnSetConnection;
  published
    property Version: String read FVersion write SetVersion;
    /// <summary>
    ///   The name of a base report. Used if you inherit the report from a
    ///   base.
    /// </summary>
    property ParentReport: String read FParentReport write SetParentReport;
    /// <summary>
    ///   Dataset for a report. If this property is set, the report will be
    ///   printed as many times as many records in the dataset.
    /// </summary>
    property DataSet: TfrxDataSet read FDataSet write SetDataSet;
    /// <summary>
    ///   The same as DataSet property but contains symbolic link to a DataSet.
    /// </summary>
    property DataSetName: String read GetDataSetName write SetDataSetName;
    /// <summary>
    ///   Defines whether the report is dot-matrix. When setting this property
    ///   to True, the report may contain dot-matrix pages (TfrxDMPPage) and
    ///   objects. Don't set this property directly. Use the "File|New..." menu
    ///   item to create dot-matrix reports.
    /// </summary>
    property DotMatrixReport: Boolean read FDotMatrixReport write FDotMatrixReport;
    /// <summary>
    ///   A set of properties related to the FastReport engine.
    /// </summary>
    property EngineOptions: TfrxEngineOptions read FEngineOptions write SetEngineOptions;
    /// <summary>
    ///   A file name or a name of the registry key for storing the FastReport
    ///   environment settings.
    /// </summary>
    property IniFile: String read {$IFNDEF FPC}FIniFile{$ELSE}GetLazIniFile{$ENDIF} write FIniFile;
    /// <summary>
    ///   Shows legacy progress message window. Do not use it.
    /// </summary>
    property OldStyleProgress: Boolean read FOldStyleProgress write FOldStyleProgress default False;
    /// <summary>
    ///   A link to the "TfrxPreview" component, in which the finished report
    ///   should be displayed. If this property is blank, a report is displayed
    ///   in the standard preview window.
    /// </summary>
    property Preview: TfrxCustomPreview read FPreview write SetPreview;
    /// <summary>
    ///   A set of properties, relating to the report preview.
    /// </summary>
    property PreviewOptions: TfrxPreviewOptions read FPreviewOptions write SetPreviewOptions;
    /// <summary>
    ///   A set of properties, which relate to the report printing.
    /// </summary>
    property PrintOptions: TfrxPrintOptions read FPrintOptions write SetPrintOptions;
    /// <summary>
    ///   Defines a set of properties relating to the report.
    /// </summary>
    property ReportOptions: TfrxReportOptions read FReportOptions write SetReportOptions;
    /// <summary>
    ///   Script language used in a report. The following values are valid:
    ///   PascalScript, C++Script, BasicScript, JScript.
    /// </summary>
    property ScriptLanguage: String read FScriptLanguage write FScriptLanguage;
    /// <summary>
    ///   Script text.
    /// </summary>
    property ScriptText: TStrings read GetScriptText write SetScriptText;
    /// <summary>
    ///   Set to True when need progress message to be shown.
    /// </summary>
    property ShowProgress: Boolean read FShowProgress write FShowProgress default True;
    /// <summary>
    ///   Set to False when do not need to save report to DFM.
    /// </summary>
    property StoreInDFM: Boolean read FStoreInDFM write FStoreInDFM default True;

    /// <summary>
    ///   When starting a report. It occurs after handling each object.
    /// </summary>
    property OnAfterPrint: TfrxBeforePrintEvent read FOnAfterPrint write FOnAfterPrint;
    /// <summary>
    ///   Occurs when the report engine establish connection to database
    ///   forinternal database components.
    /// </summary>
    property OnBeforeConnect: TfrxBeforeConnectEvent read FOnBeforeConnect write FOnBeforeConnect;
    /// <summary>
    ///   Occurs when the report engine close connection to database for
    ///   internaldatabase components.
    /// </summary>
    property OnAfterDisconnect: TfrxAfterDisconnectEvent read FOnAfterDisconnect write FOnAfterDisconnect;
    /// <summary>
    ///   When starting a report. It occurs before handling each object.
    /// </summary>
    property OnBeforePrint: TfrxBeforePrintEvent read FOnBeforePrint write FOnBeforePrint;
    /// <summary>
    ///   Occurs when running a report.
    /// </summary>
    property OnBeginDoc: TNotifyEvent read FOnBeginDoc write FOnBeginDoc;
    /// <summary>
    ///   When previewing a report in the preview window. Occurs when clicking
    ///   the object.
    /// </summary>
    property OnClickObject: TfrxClickObjectEvent read FOnClickObject write FOnClickObject;
    /// <summary>
    ///   When previewing a report in the preview window. Occurs when double
    ///   clicking the object.
    /// </summary>
    property OnDblClickObject: TfrxClickObjectEvent read FOnDblClickObject write FOnDblClickObject;
    /// <summary>
    ///   Occurs after finishing a report.
    /// </summary>
    property OnEndDoc: TNotifyEvent read FOnEndDoc write FOnEndDoc;
    /// <summary>
    ///   When starting a report. Occurs when an unknown variable is met. An
    ///   event handler must return the value of this variable.
    /// </summary>
    property OnGetValue: TfrxGetValueEvent read FOnGetValue write FOnGetValue;
    /// <summary>
    ///   When starting a report. Occurs when an unknown variable is met. An
    ///   event handler must return the value of this variable.
    /// </summary>
    property OnNewGetValue: TfrxNewGetValueEvent read FOnNewGetValue write FOnNewGetValue;
    /// <summary>
    ///   When starting a report. If the handler of this event is allocated,
    ///   then the FastReport engine is blocked and you thus would have to
    ///   construct a report manually.
    /// </summary>
    property OnManualBuild: TfrxManualBuildEvent read FOnManualBuild write FOnManualBuild;
    property OnObjectManualBuild: TfrxObjectManualBuildEvent read FOnObjectManualBuild write FOnObjectManualBuild;
    /// <summary>
    ///   When report is previewed in the preview window. Occurs when mouse
    ///   cursor passes over the object.
    /// </summary>
    property OnMouseOverObject: TfrxMouseOverObjectEvent read FOnMouseOverObject
      write FOnMouseOverObject;
    /// <summary>
    ///   Occurs when report is previewed in the preview window and mouse
    ///   pointerenters object rect area.
    /// </summary>
    property OnMouseEnter: TfrxMouseEnterEvent read FOnMouseEnter write FOnMouseEnter;
    /// <summary>
    ///   Occurs when report is previewed in the preview window and mouse
    ///   pointerleaves object rect area.
    /// </summary>
    property OnMouseLeave: TfrxMouseLeaveEvent read FOnMouseLeave write FOnMouseLeave;
    /// <summary>
    ///   Occurs before displaying a preview window.
    /// </summary>
    property OnPreview: TNotifyEvent read FOnPreview write FOnPreview;
    /// <summary>
    ///   Occurs when printing a report page. The CopyNo parameter determines
    ///   which copy is printing now (if user choose to print several copies).
    /// </summary>
    property OnPrintPage: TfrxPrintPageEvent read FOnPrintPage write FOnPrintPage;
    /// <summary>
    ///   Occurs before printing a report.
    /// </summary>
    property OnPrintReport: TNotifyEvent read FOnPrintReport write FOnPrintReport;
    /// <summary>
    ///   Occurs after printing a report.
    /// </summary>
    property OnAfterPrintReport: TNotifyEvent read FOnAfterPrintReport write FOnAfterPrintReport;
    /// <summary>
    ///   This event is fired at start of some continuous operation (print,
    ///   export, run report). The event handler may display a dialog window
    ///   with progress bar. Progress parameter is a page number which is
    ///   currently processed.
    /// </summary>
    property OnProgressStart: TfrxProgressEvent read FOnProgressStart write FOnProgressStart;
    /// <summary>
    ///   This event is fired when performing some continuous operations
    ///   (print, export, run report). The event handler may display a dialog
    ///   window with progress bar. Progress parameter is a page number which
    ///   is currently being processed.
    /// </summary>
    property OnProgress: TfrxProgressEvent read FOnProgress write FOnProgress;
    /// <summary>
    ///   This event is fired at stop of some continuous operation (print,
    ///   export, run report). The event handler may display a dialog window
    ///   with progress bar. Progress parameter is a page number which is
    ///   currently processed.
    /// </summary>
    property OnProgressStop: TfrxProgressEvent read FOnProgressStop write FOnProgressStop;
    /// <summary>
    ///   Name of the script procedure which will be called when report dialog
    ///   page is previewed. <br />
    /// </summary>
    property OnRunDialogs: TfrxRunDialogsEvent read FOnRunDialogs write FOnRunDialogs;
    /// <summary>
    ///   Name of the script procedure which will be called when report starts.
    /// </summary>
    property OnStartReport: TfrxNotifyEvent read FOnStartReport write FOnStartReport;
    /// <summary>
    ///   Name of the script procedure which will be called when report
    ///   finishes.
    /// </summary>
    property OnStopReport: TfrxNotifyEvent read FOnStopReport write FOnStopReport;
    /// <summary>
    ///   Occurs when calling a function, added with the help of the
    ///   "AddFunction" method.
    /// </summary>
    property OnUserFunction: TfrxUserFunctionEvent read FOnUserFunction write FOnUserFunction;
    /// <summary>
    ///   Occurs when a report is an inherited report. The event handler must
    ///   load the report template (given by TemplateName parameter) into
    ///   Report object.
    /// </summary>
    property OnLoadTemplate: TfrxLoadTemplateEvent read FOnLoadTemplate write FOnLoadTemplate;
    /// <summary>
    ///   Occurs when a report object loads detail report. The event handler
    ///   must load the report template (given by TemplateName parameter) into
    ///   Report object with using parameters of AHyperlink.
    /// </summary>
    property OnLoadDetailTemplate: TfrxLoadDetailTemplateEvent read FOnLoadDetailTemplate write FOnLoadDetailTemplate;
    /// <summary>
    ///   Occurs when preview form closed.
    /// </summary>
    property OnClosePreview: TNotifyEvent read FOnClosePreview write FOnClosePreview;
    /// <summary>
    ///   Occurs when papered report being printed.
    /// </summary>
    property OnReportPrint: TfrxNotifyEvent read FOnReportPrint write FOnReportPrint;
    /// <summary>
    ///   Occurs after script compilation and before the report engine stars
    ///   report preparation.
    /// </summary>
    property OnAfterScriptCompile: TNotifyEvent read FOnAfterScriptCompile write FOnAfterScriptCompile;
    /// <summary>
    ///   This event fires when engine loads report XML file and file contains
    ///   XML item &lt;FrxCustomData&gt;(i.e. store user data inside XML
    ///   file).All child nodes in this item ignores by <br />serializer.
    ///   XMLItem contains all the custom user data.
    /// </summary>
    property OnGetCustomData: TfrxGetCustomDataEvent read FOnGetCustomDataEvent write FOnGetCustomDataEvent;
    /// <summary>
    ///   This event fires when serializer saves report template to
    ///   XML(i.e.store user data inside XML file). If event is set, serializer
    ///   creates &lt;FrxCustomData&gt; node and pass it as a parameter.
    ///   Programmer may use this node to create child nodes with custom data.
    /// </summary>
    property OnSaveCustomData: TfrxSaveCustomDataEvent read FOnSaveCustomDataEvent write FOnSaveCustomDataEvent;
  end;

  /// <summary>
  ///   This interface declares deferred command abstraction.
  /// </summary>
  IfrxDelayedCommand = interface ['{3A04A39B-B516-42A5-A597-53F99F0EDB7B}']
    /// <summary>
    ///   Runs command.
    /// </summary>
    procedure Run;
    /// <summary>
    ///   Get symbolic name of current command instance.
    /// </summary>
    function GetName: String;
    /// <summary>
    ///   Get command instance.
    /// </summary>
    function GetInstance: TObject;
  end;

  /// <summary>
  ///   The base abstract class which implements IfrxDelayedCommand interface.
  ///   It used as a base for all report commands. All commands stored in the
  ///   command list in the report component.
  /// </summary>
  TfrxBaseDelayedCommand = class(TPersistent, IfrxDelayedCommand)
  private
    FRefCount: Integer;
  protected
    function _AddRef: Integer; {$IFDEF FPC}{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};{$ELSE} stdcall;{$ENDIF}
    function _Release: Integer; {$IFDEF FPC}{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};{$ELSE} stdcall;{$ENDIF}
  public
    function QueryInterface({$IFDEF FPC_HAS_CONSTREF}constref{$ELSE}const{$ENDIF} IID: TGUID; out Obj): HResult; virtual; {$IFDEF FPC}{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};{$ELSE} stdcall;{$ENDIF}
  public
    procedure AfterConstruction; override;
    procedure Run; virtual; abstract;
    function GetName: String;
    function GetInstance: TObject;
  end;

  /// <summary>
  ///   This class is a base for all prepare report commands.
  /// </summary>
  TfrxBasePrepareReportCommand = class(TfrxBaseDelayedCommand)
  private
    FReport: TfrxReport;
    FClearLast: Boolean;
    FEngineOptions: TfrxEngineOptions;
  public
    /// <summary>
    ///   Command constructor.
    /// </summary>
    /// <param name="AReport">
    ///   Reference to a report component.
    /// </param>
    /// <param name="ClearLast">
    ///   Clears previous report. Parameter of PrepareReport method.
    /// </param>
    constructor Create(AReport: TfrxReport; ClearLast: Boolean);
    destructor Destroy; override;
    procedure Run; override;
  end;

  /// <summary>
  ///   This class implements realization of deferred ShowReport command.
  /// </summary>
  TfrxShowReportCommand = class(TfrxBasePrepareReportCommand)
  public
    procedure Run; override;
  end;

  /// <summary>
  ///   This class implements realization of deferred PrepareReport command.
  /// </summary>
  TfrxPrepareReportCommand = class(TfrxBasePrepareReportCommand)
  public
    procedure Run; override;
  end;

  /// <summary>
  ///   This class implements realization of deferred RefreshReport command.
  /// </summary>
  TfrxRefreshReportCommand = class(TfrxBasePrepareReportCommand)
  public
    procedure Run; override;
  end;

  /// <summary>
  ///   This class implements realization of deferred LoadFrom command.
  /// </summary>
  TfrxLoadFromCommand = class(TfrxBaseDelayedCommand)
  private
    FReport: TfrxReport;
    FTransport: TfrxCustomIOTransport;
    FFileName: String;
  public
    /// <summary>
    ///   LoadFrom command constructor.
    /// </summary>
    /// <param name="AReport">
    ///   Reference to a report component.
    /// </param>
    /// <param name="Transport">
    ///   Reference to a IO transport used for file operations(command creates
    ///   copy of it in constructor).
    /// </param>
    /// <param name="FileName">
    ///   File name.
    /// </param>
    constructor Create(AReport: TfrxReport; Transport: TfrxCustomIOTransport; const FileName: String);
    destructor Destroy; override;
    procedure Run; override;
  end;

  /// <summary>
  ///   This class implements realization of deferred LoadFromFile command.
  /// </summary>
  TfrxLoadFromFileCommand = class(TfrxBaseDelayedCommand)
  private
    FReport: TfrxReport;
    FFileName: String;
    FExceptionIfNotFound: Boolean;
  public
    /// <summary>
    ///   LoadFromFile command constructor.
    /// </summary>
    /// <param name="AReport">
    ///   Reference to a report component.
    /// </param>
    /// <param name="FileName">
    ///   File name.
    /// </param>
    /// <param name="ExceptionIfNotFound">
    ///   Generates exception if file now found. LoadFromFile parameter.
    /// </param>
    constructor Create(AReport: TfrxReport; const FileName: String; ExceptionIfNotFound: Boolean);
    procedure Run; override;
  end;

  /// <summary>
  ///   This class implements realization of deferred LoadFromStream command.
  /// </summary>
  TfrxLoadFromStreamCommand = class(TfrxBaseDelayedCommand)
  private
    FReport: TfrxReport;
    FStream: TMemoryStream;
  public
    /// <summary>
    ///   LoadFromStream command constructor.
    /// </summary>
    /// <param name="AReport">
    ///   Reference to a report component.
    /// </param>
    /// <param name="Stream">
    ///   Source stream(Command constructor creates a copy of in in memory).
    /// </param>
    constructor Create(AReport: TfrxReport; Stream: TStream);
    destructor Destroy; override;
    procedure Run; override;
  end;

  /// <summary>
  ///   The TfrxDelayedCommandManager class is used as storage and manager for
  ///   commands which implements <br />IfrxDelayedCommand interface.
  /// </summary>
  TfrxDelayedCommandManager = class
  private
    FCommands: TStringList;
    FIsRunning: Boolean;
  public
    /// <summary>
    ///   Constructor.
    /// </summary>
    constructor Create;
    destructor Destroy; override;
    /// <summary>
    ///   Clears all deferred commands in the list.
    /// </summary>
    procedure Clear;
    /// <summary>
    ///   Adds new command to the deferred commands list.
    /// </summary>
    /// <param name="Cmd">
    ///   Object which implements <br />IfrxDelayedCommand interface.
    /// </param>
    procedure AddCommand(Cmd: IfrxDelayedCommand);
    /// <summary>
    ///   Run all commands in the list one by one.
    /// </summary>
    procedure RunAll;
    /// <summary>
    ///   The current state of manager. It's True if manager runs some
    ///   commands.
    /// </summary>
    property IsRunning: Boolean read FIsRunning;
  end;

  { Incapsulate second list for Object ispector }
  /// <summary>
  ///   This class implements list of selected report objects in the report
  ///   designer and in the report preview. For the report designer it
  ///   encapsulates second list for the object inspector.
  /// </summary>
  TfrxSelectedObjectsList = class(TList)
  private
    FInspSelectedObjects: TList;
    FUpdated: Boolean;
  protected
    procedure Notify(Ptr: Pointer; Action: TListNotification); override;
  public
    /// <summary>
    ///   List constructor.
    /// </summary>
    constructor Create;
    destructor Destroy; override;
    /// <summary>
    ///   Clears only object inspector list.
    /// </summary>
    procedure ClearInspectorList;
    /// <summary>
    ///   Reference to the object inspector list. It used by the object
    ///   inspector. In the report designer selected objects and object
    ///   inspector selection may vary.
    /// </summary>
    property InspSelectedObjects: TList read FInspSelectedObjects;
    /// <summary>
    ///   State of the lists. It's updating when it's True.
    /// </summary>
    property Updated: Boolean read FUpdated write FUpdated;
  end;

  /// <summary>
  ///   The TfrxCustomDesigner class is the base class for the report designer.
  /// </summary>
  TfrxCustomDesigner = class(TfrxBaseForm)
  private
    FReport: TfrxReport;
    FIsPreviewDesigner: Boolean;
    FMemoFontName: String;
    FMemoFontSize: Integer;
    FUseObjectFont: Boolean;
    FParentForm: TForm;
  protected
    FModified: Boolean;
    FInspectorLock: Boolean;
    FObjects: TList;
    FPage: TfrxPage;
    FSelectedObjects: TfrxSelectedObjectsList;
    procedure SetModified(const Value: Boolean); virtual;
    procedure SetPage(const Value: TfrxPage); virtual;
    function GetCode: TStrings; virtual; abstract;
    function GetDefaultPreferences: TObject; virtual;
  public
    /// <summary>
    ///   The Report designer constructor.
    /// </summary>
    constructor CreateDesigner(AOwner: TComponent; AReport: TfrxReport;
      APreviewDesigner: Boolean = False);
    destructor Destroy; override;
    procedure DeleteObject(vComp: TfrxComponent); virtual;
    procedure AddPictureToCache(PicView: TfrxPictureView); virtual;
    /// <summary>
    ///   Used to check restriction of the report designer. Returns True when
    ///   restriction flag is set in the report designer.
    /// </summary>
    /// <param name="Op">
    ///   Restriction flag.
    /// </param>
    function CheckOp(Op: TfrxDesignerRestriction): Boolean; virtual;
    procedure ClearLastActiveObject; virtual; abstract;
    /// <summary>
    ///   Method shows an "Insert expression" dialogue. It should return an
    ///   expression string or an empty string if dialogue was canceled.
    /// </summary>
    function InsertExpression(const Expr: String): String; virtual; abstract;
    /// <summary>
    ///   Method shows an "Insert expression" dialogue. It should return an
    ///   expression string in out OutExpr parameter. Method returns True when
    ///   expression was changed.
    /// </summary>
    function IsChangedExpression(const InExpr: String; out OutExpr: String): Boolean; virtual; abstract;
    function GetPreferencesStorage(aDefault: Boolean): TObject; override;
    /// <summary>
    ///   Locks any drawings in the designer window. Use this method when
    ///   making any changes in the currently edited report.
    /// </summary>
    procedure Lock; virtual; abstract;
    /// <summary>
    ///   Reloads all report's pages. Use this method if you add or delete a
    ///   report page. The Index parameter defines which page should be made
    ///   active.
    /// </summary>
    procedure ReloadPages(Index: Integer); virtual; abstract;
    /// <summary>
    ///   Reloads all report contents.
    /// </summary>
    procedure ReloadReport; virtual; abstract;
    /// <summary>
    ///   Reloads all report objects in the report designer.
    /// </summary>
    /// <param name="ResetSelection">
    ///   Reset objects selection in the report designer.
    /// </param>
    procedure ReloadObjects(ResetSelection: Boolean = True); virtual; abstract;
    /// <summary>
    ///   Updates the "Data Tree" tool window.
    /// </summary>
    procedure UpdateDataTree; virtual; abstract;
    /// <summary>
    ///   Updates the report designer workspace.
    /// </summary>
    procedure UpdatePage; virtual; abstract;
    /// <summary>
    ///   Updates object inspector.
    /// </summary>
    procedure UpdateInspector; virtual; abstract;
    /// <summary>
    ///   Locks object inspector updating.
    /// </summary>
    procedure DisableInspectorUpdate; virtual; abstract;
    /// <summary>
    ///   Releases object inspector lock.
    /// </summary>
    procedure EnableInspectorUpdate; virtual; abstract;
    /// <summary>
    ///   This method processes clipboard copy behaviour.
    /// </summary>
    procedure InternalCopy; virtual; abstract;
    /// <summary>
    ///   This method processes clipboard paste behaviour.
    /// </summary>
    procedure InternalPaste; virtual; abstract;
    /// <summary>
    ///   The methods checks if clipboard contains data available to paste in
    ///   selected object. It serves copy/paste methods of InPlace editors.
    /// </summary>
    function InternalIsPasteAvailable: Boolean; virtual; abstract;
    /// <summary>
    ///   Calls DataSets selection dialog.
    /// </summary>
    procedure ReportDataEdit; virtual; abstract;
    /// <summary>
    ///   Calls report variables editor dialog.
    /// </summary>
    procedure ReportEditVariable; virtual; abstract;
    /// <summary>
    ///   Checks if passed object is designer DataTree window. Used for drag
    ///   and drop behaviour.
    /// </summary>
    /// <param name="aObject">
    ///   Reference to an object.
    /// </param>
    function IsDataTree(aObject: TObject): Boolean; virtual; abstract;
    procedure SetSelection(AComponent: TfrxComponent; const PropName: String; SelStart, SelEnd: TPoint); virtual; abstract;
    /// <summary>
    ///   Returns a dataset expression from DataTree selection.
    /// </summary>
    function GetDataSelectedAsExp: String; virtual; abstract;
    /// <summary>
    ///   Returns True if the designer was launched from the preview window.
    /// </summary>
    property IsPreviewDesigner: Boolean read FIsPreviewDesigner;
    /// <summary>
    ///   Determines whether the report was modified.
    /// </summary>
    property Modified: Boolean read FModified write SetModified;
    /// <summary>
    ///   List of objects on the current report page, including the page
    ///   itself.
    /// </summary>
    property Objects: TList read FObjects;
    /// <summary>
    ///   Reference to a report.
    /// </summary>
    property Report: TfrxReport read FReport;
    /// <summary>
    ///   List of currently selected object.
    /// </summary>
    property SelectedObjects: TfrxSelectedObjectsList read FSelectedObjects;
    /// <summary>
    ///   Reference to currently edited report page. May be nil if current page
    ///   is "code editor".
    /// </summary>
    property Page: TfrxPage read FPage write SetPage;
    /// <summary>
    ///   The script of the report.
    /// </summary>
    property Code: TStrings read GetCode;
    /// <summary>
    ///   If True , Memo editor will use font for editor from the report object
    ///   settings.
    /// </summary>
    property UseObjectFont: Boolean read FUseObjectFont write FUseObjectFont;
    /// <summary>
    ///   Default name of the font for memo text editor.
    /// </summary>
    property MemoFontName: String read FMemoFontName write FMemoFontName;
    /// <summary>
    ///   Default size of the font for memo text editor.
    /// </summary>
    property MemoFontSize: Integer read FMemoFontSize write FMemoFontSize;
    /// <summary>
    ///   Parent form for the report designer form.
    /// </summary>
    property ParentForm: TForm read FParentForm write FParentForm;
  end;

  TfrxDesignerClass = class of TfrxCustomDesigner;
  /// <summary>
  ///   Notification of Tree node item. Used to notify about changes of
  ///   item(Add/Remove).
  /// </summary>
  TfrxNodeChanging = procedure(const Name: String; bUseRemoveList: Boolean = False; aObject: TObject = nil) of object;

  /// <summary>
  ///   This class represents tree node which used to emulate tree structure of
  ///   directories and files in IO transports.
  /// </summary>
  TfrxNode = class(TObject)
  private
    FParent: TfrxNode;
    FNodesList: TStringList;
    FData: Variant;
    FObjectData: TObject;
    FName: String;
    FOriginalName: String;
    FOwnObject: Boolean;
    FOnAddItem: TfrxNodeChanging;
    FOnRemoveItem: TfrxNodeChanging;
    FFilterAccess: TfrxFilterAccess;
    FFilesCount: Integer;
    procedure SetName(const Value: String);
    procedure SetObjectData(const Value: TObject);
    function GetNode(Index: Integer): TfrxNode;
    procedure UpdateFilesCount(Count: Integer);
    procedure SetOriginalName(const Value: String);
  public
    /// <summary>
    ///   Default constructor.
    /// </summary>
    constructor Create;
    destructor Destroy; override;
    /// <summary>
    ///   Clears current node and destroys all child nodes with data(closes and
    ///   free all opened streams).
    /// </summary>
    procedure Clear;
    /// <summary>
    ///   Returns Root tree node.
    /// </summary>
    function GetRoot: TfrxNode;
    /// <summary>
    ///   Adds new child node to current one.
    /// </summary>
    /// <param name="Name">
    ///   Name of the child node.
    /// </param>
    /// <param name="aObject">
    ///   Object reference assigned to node(usually stream).
    /// </param>
    function Add(const Name: String; aObject: TObject = nil): TfrxNode;
    /// <summary>
    ///   Removes node.
    /// </summary>
    /// <param name="aNode">
    ///   Reference to a node.
    /// </param>
    procedure RemoveNode(aNode: TfrxNode);
    /// <summary>
    ///   Searches for node by name.
    /// </summary>
    /// <param name="Name">
    ///   Name of a node to search.
    /// </param>
    /// <param name="SearchInChilds">
    ///   True if need to search in all child nodes.
    /// </param>
    function Find(const Name: String; SearchInChilds: Boolean = False): TfrxNode; overload;
    /// <summary>
    ///   Searches for node by assigned object.
    /// </summary>
    /// <param name="aObject">
    ///   Reference to an object.
    /// </param>
    /// <param name="SearchInChilds">
    ///   True if need to search in all child nodes.
    /// </param>
    function Find(aObject: TObject; SearchInChilds: Boolean = False): TfrxNode; overload;
    /// <summary>
    ///   Returns count of nodes for current level.
    /// </summary>
    function Count: Integer;
    /// <summary>
    ///   Reference to a parent node.
    /// </summary>
    property Parent: TfrxNode read FParent;
    /// <summary>
    ///   Node data.
    /// </summary>
    property Data: Variant read FData write FData;
    /// <summary>
    ///   Node object. Usually stream.
    /// </summary>
    property ObjectData: TObject read FObjectData write SetObjectData;
    /// <summary>
    ///   Name of the current node.
    /// </summary>
    property Name: String read FName write SetName;
    /// <summary>
    ///   List of all child nodes for current level.
    /// </summary>
    property Items[Index: Integer]: TfrxNode read GetNode;
    /// <summary>
    ///   Original name of a node. internal use.
    /// </summary>
    property OriginalName: String read FOriginalName write SetOriginalName;
    /// <summary>
    ///   This event fires when child item added to current.
    /// </summary>
    property OnAddItem: TfrxNodeChanging read FOnAddItem write FOnAddItem;
    /// <summary>
    ///   This event fires when child item removed from current.
    /// </summary>
    property OnRemoveItem: TfrxNodeChanging read FOnRemoveItem write FOnRemoveItem;
    /// <summary>
    ///   Access type of IO filter.
    /// </summary>
    property FilterAccess: TfrxFilterAccess read FFilterAccess write FFilterAccess;
    /// <summary>
    ///   Count of files(ObjectData assigned) in current item.
    /// </summary>
    property FilesCount: Integer read FFilesCount;
  end;

  /// <summary>
  ///   These flags used in IO transports to define visibility of IO transport
  ///   filter. By default each transport filter component registers in global
  ///   list and these flags helps to sort items of the global list.
  /// </summary>
  TfrxFilterVisibleState = (
    /// <summary>
    ///   Filter available from the report designer in "File" -&gt; "Save as"
    ///   and "Open" submenu.
    /// </summary>
    fvDesigner,
    /// <summary>
    ///   Filter available from the report preview in "File" -&gt; "Save as"
    ///   and "Open" submenu. <br />
    /// </summary>
    fvPreview,
    /// <summary>
    ///   Filter available from an export filter dialog.
    /// </summary>
    fvExport,
    /// <summary>
    ///   Filter available from the report designer as a filter option (drop
    ///   down list) in standard Save and Open dialogs.
    /// </summary>
    fvDesignerFileFilter,
    /// <summary>
    ///   Filter available from the report preview as a filter option (drop
    ///   down list) in standard Save and Open dialogs.
    /// </summary>
    fvPreviewFileFilter,
    /// <summary>
    ///   Filter available from an export as a filter option (drop down list)
    ///   in standard Save and Open dialogs.
    /// </summary>
    fvExportFileFilter,
    /// <summary>
    ///   Reserved.
    /// </summary>
    fvOther);
  /// <summary>
  ///   Set of flags used in IO transports to define visibility of IO transport
  ///   filter.
  /// </summary>
  TfrxFilterVisibility = set of TfrxFilterVisibleState;

  /// <summary>
  ///   Supported stream type for current IO transport filter.
  /// </summary>
  TfrxIOTransportStream = (
    /// <summary>
    ///   IO filter supports Read operations.
    /// </summary>
    tsIORead,
    /// <summary>
    ///   IO filter supports Write operations.
    /// </summary>
    tsIOWrite);
  /// <summary>
  ///   Set of flags for supported stream types by IO transport filter.
  /// </summary>
  TfrxIOTransportStreams = set of TfrxIOTransportStream;


  /// <summary>
  ///   The TfrxCustomIOTransport class is the base class for all IO
  ///   (input/output filters). It represents Base adapter interface for file
  ///   or stream IO operations. Transport filters can be used in the report
  ///   designer (save/open operations), report preview (save/open operations)
  ///   and export filters. It possible to create one class of transport filter
  ///   which encapsulates work with some IO protocol(for example save/load
  ///   from cloud) and use this class where it's necessary(Designer, Preview,
  ///   Exports) by setting up Visibility property.
  /// </summary>
  TfrxCustomIOTransport = class(TComponent)
  private
    FReport: TfrxReport;
    FShowDialog: Boolean;
    FOverwritePrompt: Boolean;
    FDefaultPath: String;
    FBasePath: String;
    FFileName: String;
    FDefaultExt: String;
    FExtDescription: String;
    FFilterString: String;
    FFilterAccess: TfrxFilterAccess;
    FTempFilter: TfrxCustomIOTransport;

    FCurNode: TfrxNode;
    function GetTempFilter: TfrxCustomIOTransport;
    procedure SetTempFilter(const Value: TfrxCustomIOTransport);
    procedure SetInternalFilter(const Value: TfrxCustomIOTransport);
    function GetInternalFilter: TfrxCustomIOTransport;
    function GetCurrentContainer: String;
    procedure SetCurrentContainer(const Value: String);
  protected
    FNoRegister: Boolean;
    FVisibility: TfrxFilterVisibility;
    FCreatedFrom: TfrxFilterVisibleState;
    FFreeStream: Boolean;
    FClosedLoadLock: Boolean;
    FInternalFilter: TfrxCustomIOTransport;
    FDirTree: TfrxNode;
    FSupportedStreams: TfrxIOTransportStreams;
    FOriginalCopy: TfrxCustomIOTransport;
    function GetFilterString: String; virtual;
    function GetVisibility: TfrxFilterVisibility; virtual;
    procedure SetVisibility(const Value: TfrxFilterVisibility); virtual;
    function GetNodeFromPath(aPath: String; bCreateNodes: Boolean = False; bLastNodeAsFile: Boolean = False): TfrxNode;
    { override for container functional if needed }
    function AddContainer(Item: TfrxNode): Boolean; virtual;
    function RemoveContainer(Item: TfrxNode): Boolean; virtual;
    function DoCreateStream(var aFullFilePath: String; aFileName: String): TStream; virtual;
    property InternalFilter: TfrxCustomIOTransport read GetInternalFilter write SetInternalFilter;
  public
    constructor Create(AOwner: TComponent); override;
    /// <summary>
    ///   This constructor creates component without registration in global
    ///   list so the filter will be hidden for preview or report designer.
    /// </summary>
    constructor CreateNoRegister;
    /// <summary>
    ///   Copy constructor.
    /// </summary>
    /// <param name="CreatedFrom">
    ///   Instance to copy from.
    /// </param>
    /// <returns>
    ///   New copy of filter.
    /// </returns>
    function CreateFilterClone(CreatedFrom: TfrxFilterVisibleState): TfrxCustomIOTransport;
    destructor Destroy; override;
    // basic file structure methods
    /// <summary>
    ///   The method closes all opened streams.
    /// </summary>
    procedure CloseAllStreams;
    /// <summary>
    ///   It copies names of steams in current transport to TStrings passed as
    ///   a parameter.
    /// </summary>
    /// <param name="aStrings">
    ///   List to copy into.
    /// </param>
    /// <param name="OpenedOnly">
    ///   Copy only names of open streams.
    /// </param>
    procedure CopyStreamsNames(aStrings: TStrings; OpenedOnly: Boolean = False);
    /// <summary>
    ///   Loads and opens steams from the list.
    /// </summary>
    /// <param name="aStrings">
    ///   List with steam names.
    /// </param>
    procedure LoadStreamsList(aStrings: TStrings);
    /// <summary>
    ///   Reloads all closed streams and open them again.
    /// </summary>
    procedure LoadClosedStreams;
    { containers / Directories }
    /// <summary>
    ///   Creates a structure node represents a container or directory.
    /// </summary>
    /// <param name="cName">
    ///   Name of a container.
    /// </param>
    /// <param name="bSetCurrent">
    ///   Set this container as current.
    /// </param>
    procedure CreateContainer(const cName: String; bSetCurrent: Boolean = False);
    /// <summary>
    ///   Deletes a structure node represents a container or directory.
    /// </summary>
    procedure DeleteContainer(const cName: String);
    /// <summary>
    ///   Creates a structure node represents a temporary container or
    ///   directory.
    /// </summary>
    procedure CreateTempContainer; virtual;
    /// <summary>
    ///   Destroy a structure node represents a temporary container or
    ///   directory.
    /// </summary>
    procedure FreeTempContainer; virtual;
    { file streams }
    /// <summary>
    ///   Returns stream by name of a node or a file.
    /// </summary>
    /// <param name="aFileName">
    ///   File or node name.
    /// </param>
    function GetStream(aFileName: String = ''): TStream; virtual;
    /// <summary>
    ///   Closes an opened stream by reference or name.
    /// </summary>
    /// <param name="aStream">
    ///   Reference to opened stream.
    /// </param>
    /// <param name="aFileName">
    ///   Stream or file name.
    /// </param>
    procedure FreeStream(aStream: TStream; const aFileName: String = ''); virtual;
    /// <summary>
    ///   This method call before loading from stream is started. Can be used
    ///   to make a preparation of stream data before loading.
    /// </summary>
    function DoFilterProcessStream(aStream: TStream; ProcesssingObject: TObject): Boolean; virtual;
    /// <summary>
    ///   Returns true when physically all stream belongs to one container(zip
    ///   archive).
    /// </summary>
    function AllInOneContainer: Boolean; virtual;
    // filter (un)init
    /// <summary>
    ///   This method can be called to open filter. It uses to make preparation
    ///   like set connection or show user dialogs. It returns true when
    ///   initialization was processed.
    /// </summary>
    function OpenFilter: Boolean; virtual;
    /// <summary>
    ///   This method can be called to close a filter. It uses to make
    ///   finalization like close connection and streams.
    /// </summary>
    procedure CloseFilter; virtual;
    /// <summary>
    ///   Initialization of transport filter based on export filter
    ///   settings(like file name and file extension).
    /// </summary>
    /// <param name="ExportFilter">
    ///   Instance of an export filter.
    /// </param>
    procedure InitFromExport(ExportFilter: TfrxCustomExportFilter);
    /// <summary>
    ///   This method assigns properties of other transport filter to current.
    /// </summary>
    procedure AssignFilter(Source: TfrxCustomIOTransport); virtual;
    /// <summary>
    ///   This method assigns shared properties(like tokens) of other transport
    ///   filter to current.
    /// </summary>
    procedure AssignSharedProperties(Source: TfrxCustomIOTransport); virtual;
    procedure AssignOriginalSharedProperties;
    /// <summary>
    ///   Returns nearest file node in current container.
    /// </summary>
    function GetFileNode: TfrxNode;
    class function GetDescription: String; virtual;
    /// <summary>
    ///   Root node of the filter.
    /// </summary>
    property RootNode: TfrxNode read FDirTree;
    /// <summary>
    ///   Name of the current container.
    /// </summary>
    property CurrentContainer: String read GetCurrentContainer write SetCurrentContainer;
    /// <summary>
    ///   This flag shows where the current instance of the filter was created
    ///   from.
    /// </summary>
    property CreatedFrom: TfrxFilterVisibleState read FCreatedFrom write FCreatedFrom;
    /// <summary>
    ///   Returns reference to transport filter instance for temporary
    ///   operations.
    /// </summary>
    property TempFilter: TfrxCustomIOTransport read GetTempFilter write SetTempFilter;
    /// <summary>
    ///   Access type of IO filter.
    /// </summary>
    property FilterAccess: TfrxFilterAccess read FFilterAccess write FFilterAccess;
    /// <summary>
    ///   Description of filter instance.
    /// </summary>
    property FilterString: String read GetFilterString write FFilterString;
    /// <summary>
    ///   Reference to a report component.
    /// </summary>
    property Report: TfrxReport read FReport write FReport;
    /// <summary>
    ///   Whether to display a print dialogue. Default value is True.
    /// </summary>
    property ShowDialog: Boolean read FShowDialog write FShowDialog;
    property OverwritePrompt: Boolean read FOverwritePrompt write FOverwritePrompt;
    /// <summary>
    ///   Default path used in transport filter.
    /// </summary>
    property DefaultPath: String read FDefaultPath write FDefaultPath;
    /// <summary>
    ///   File name for current filter.
    /// </summary>
    property FileName: String read FFileName write FFileName;
    /// <summary>
    ///   Base path used in transport filter.
    /// </summary>
    property BasePath: String read FBasePath write FBasePath;
    /// <summary>
    ///   Default extension used in transport filter.
    /// </summary>
    property DefaultExt: String read FDefaultExt write FDefaultExt;
    /// <summary>
    ///   Description for default extension used in transport filter.
    /// </summary>
    property ExtDescription: String read FExtDescription write FExtDescription;
    /// <summary>
    ///   These flags used in IO transports to define visibility of IO
    ///   transport filter. By default each transport filter component
    ///   registers in global list and these flags helps to sort items of the
    ///   global list.
    /// </summary>
    property Visibility: TfrxFilterVisibility read GetVisibility write SetVisibility;
    /// <summary>
    ///   Supported stream type for current IO transport filter.
    /// </summary>
    property SupportedStreams: TfrxIOTransportStreams read FSupportedStreams write FSupportedStreams;
  end;

  /// <summary>
  ///   The class is the base IO transport filter class for all file
  ///   operations. IO filters which work directly with files use this class as
  ///   ancestor.
  /// </summary>
  TfrxIOTransportFile = class(TfrxCustomIOTransport)
  private
    FTempFolderCreated: Boolean;
  protected
    procedure SetVisibility(const Value: TfrxFilterVisibility); override;
    function AddContainer(Item: TfrxNode): Boolean; override;
    procedure DeleteFiles;
  public
    constructor Create(AOwner: TComponent); override;
    function OpenFilter: Boolean; override;
    function DoCreateStream(var aFullFilePath: String; aFileName: String): TStream; override;
    procedure CloseFilter; override; // Empty
    class function GetDescription: String; override;
    procedure CreateTempContainer; override;
    procedure FreeTempContainer; override;
  end;

  /// <summary>
  ///   The class represents compressed file transport filter.
  /// </summary>
  TfrxSaveToCompressedFilter = class(TfrxIOTransportFile)
  private
    FIsFR3File: Boolean;
    procedure SetIsFR3File(const Value: Boolean);
  protected
    function GetFilterString: String; override;
  public
    constructor Create(AOwner: TComponent); override;
    function OpenFilter: Boolean; override;
    class function GetDescription: String; override;
    /// <summary>
    ///   True when it's an report template file.
    /// </summary>
    property IsFR3File: Boolean read FIsFR3File write SetIsFR3File;
  end;

  /// <summary>
  ///   The TfrxCustomExportFilter is the base class for all export filters.
  /// </summary>
  TfrxCustomExportFilter = class(TComponent)
  private
    FCurPage: Boolean;
    FExportNotPrintable: Boolean;
    FName: String;
    FNoRegister: Boolean;
    FPageNumbers: String;
    FReport: TfrxReport;
    FShowDialogOptions: TfrxShowDialogOptions;
    FStream: TStream;
    FUseFileCache: Boolean;
    FDefaultPath: String;
    FSlaveExport: Boolean;
    FShowProgress: Boolean;
    FDefaultExt: String;
    FFilterDesc: String;
    FSuppressPageHeadersFooters: Boolean;
    FTitle: String;
    FOverwritePrompt: Boolean;
    FFIles: TStrings;
    FOnBeforeExport: TNotifyEvent;
    FOnBeginExport: TNotifyEvent;
    FOnGenerateSheetName: TfrxGenerateSheetName;
    FCreationTime: TDateTime;
    FDataOnly: Boolean;
    FIOTransport: TfrxCustomIOTransport;
    FOpenAfterExport: Boolean;
    FCalculatePictureHash: Boolean;
    function GetDefaultIOTransport: TfrxCustomIOTransport;
    procedure SetDefaultIOTransport(const Value: TfrxCustomIOTransport);
    procedure SetFileName(const Value: String);
    procedure SetShowDialog(b: Boolean);
    function GetShowDialog: Boolean;
  protected
    FDefaultIOTransport: TfrxCustomIOTransport;
    FTerminated: Boolean;
    FfrxPictureHashMap: TfrxPictureHashMap;
    procedure Finish; virtual;
    procedure AfterFinish; virtual;
    procedure BeforeStart; virtual;
    function Start: Boolean; virtual;
    function EnableCalculateHash: Boolean; virtual;

    /// <summary>
    ///   The event responsible for the sheet name generation template.
    /// </summary>
    property OnGenerateSheetName: TfrxGenerateSheetName read FOnGenerateSheetName write FOnGenerateSheetName;
  public
    constructor Create(AOwner: TComponent); override;
    /// <summary>
    ///   This constructor creates component without registration in global
    ///   list so the filter will be hidden for preview. Thread safe
    ///   constructor.
    /// </summary>
    constructor CreateNoRegister;
    destructor Destroy; override;
    /// <summary>
    ///   Method is called after finishing the export no matter if it was canceled or compleated.
    ///   It's used to clear intarnal resorces.
    /// </summary>
    procedure ClearResources; virtual;
    /// <summary>
    ///   Description of the export filter. This value is displayed in the
    ///   preview window.
    /// </summary>
    class function GetDescription: String; virtual;
    /// <summary>
    ///   Method should show the export settings dialog and return mrOk value
    ///   if dialog was accepted. This method is not called if ShowDialog
    ///   property is False.
    /// </summary>
    function ShowModal: TModalResult; virtual;
    /// <summary>
    ///   Method is called before export starts. To continue export, this
    ///   method should return True.
    /// </summary>
    function DoStart: Boolean;
    /// <summary>
    ///   Method created default instance of transport filter for current class
    ///   of export filter.
    /// </summary>
    function CreateDefaultIOTransport: TfrxCustomIOTransport; virtual;
    /// <summary>
    ///   Method is called when exporting the object. Bands are exported too.
    /// </summary>
    procedure ExportObject(Obj: TfrxComponent); virtual; abstract;
    /// <summary>
    ///   Method is called after finishing the export.
    /// </summary>
    procedure DoFinish;
    /// <summary>
    ///   Method is called after finishing a page. Page is the reference to a
    ///   page, Index is the page index in the prepared report.
    /// </summary>
    procedure FinishPage(Page: TfrxReportPage; Index: Integer); virtual;
    /// <summary>
    ///   Method is called before exporting a page. Page is the reference to a
    ///   page, Index is the page index in the prepared report.
    /// </summary>
    procedure StartPage(Page: TfrxReportPage; Index: Integer); virtual;
    /// <summary>
    ///   Begin clip operation.
    /// </summary>
    /// <param name="Obj">
    ///   Rect of this object use as a clip rect.
    /// </param>
    procedure BeginClip(Obj: TfrxView); virtual;
    /// <exception cref="">
    ///   Stop clip operation.
    /// </exception>
    procedure EndClip; virtual;
    /// <summary>
    ///   Returns True when filter export can process internal objects in other
    ///   objects.
    /// </summary>
    function IsProcessInternal: Boolean; virtual;
    property CalculatePictureHash: Boolean read FCalculatePictureHash write FCalculatePictureHash default True;
    /// <summary>
    ///   Default transport filter used for file operations in current export
    ///   filter.
    /// </summary>
    property DefaultIOTransport: TfrxCustomIOTransport read GetDefaultIOTransport write SetDefaultIOTransport;
    /// <summary>
    ///   If this property is True, only current page will be exported(in case
    ///   the export was called from the preview window).
    /// </summary>
    property CurPage: Boolean read FCurPage write FCurPage;
    /// <summary>
    ///   List of page numbers to export, for example '1-3,5,12'. If this
    ///   property is empty (by default), all pages will be exported.
    /// </summary>
    property PageNumbers: String read FPageNumbers write FPageNumbers;
    /// <summary>
    ///   Reference to a report.
    /// </summary>
    property Report: TfrxReport read FReport write FReport;
    /// <summary>
    ///   Reference to a stream used to save export filter result.
    /// </summary>
    property Stream: TStream read FStream write FStream;
    /// <summary>
    ///   Internal use.
    /// </summary>
    property SlaveExport: Boolean read FSlaveExport write FSlaveExport;
    /// <summary>
    ///   Default extension of export filter result.
    /// </summary>
    property DefaultExt: String read FDefaultExt write FDefaultExt;
    /// <summary>
    ///   Default description of export filter.
    /// </summary>
    property FilterDesc: String read FFilterDesc write FFilterDesc;
    /// <summary>
    ///   Suppress page Headers and page Footers during export.
    /// </summary>
    property SuppressPageHeadersFooters: Boolean read FSuppressPageHeadersFooters
      write FSuppressPageHeadersFooters;
    /// <summary>
    ///   Title of export filter.
    /// </summary>
    property ExportTitle: String read FTitle write FTitle;
    /// <summary>
    ///   List of files generated by export filter.
    /// </summary>
    property Files: TStrings read FFiles write FFiles;
    /// <summary>
    ///   If True, document will be opened by default viewer.
    /// </summary>
    property OpenAfterExport: Boolean read FOpenAfterExport write FOpenAfterExport;
    /// <summary>
    ///   If True, filter requests cancellation.
    /// </summary>
    property Terminated: Boolean read FTerminated write FTerminated;
    /// <summary>
    ///   Support for ShowDialog
    /// </summary>
    property ShowDialogOptions: TfrxShowDialogOptions read FShowDialogOptions write FShowDialogOptions default [doShowExportSettings, doShowSaveDialog];
  published
    property IOTransport: TfrxCustomIOTransport read FIOTransport write FIOTransport;
    /// <summary>
    ///   Determines if export settings dialogue is shown. Default value is
    ///   True.
    /// </summary>
    property ShowDialog: Boolean read GetShowDialog write SetShowDialog default True;
    /// <summary>
    ///   Default file name for export result.
    /// </summary>
    property FileName: String read FName write SetFileName;
    /// <summary>
    ///   If True, the report engine exports all objects with Printable flag
    ///   set to False.
    /// </summary>
    property ExportNotPrintable: Boolean read FExportNotPrintable write FExportNotPrintable default False;
    /// <summary>
    ///   Use a file cache for temporary operations.
    /// </summary>
    property UseFileCache: Boolean read FUseFileCache write FUseFileCache;
    /// <summary>
    ///   Default path of export result.
    /// </summary>
    property DefaultPath: String read FDefaultPath write FDefaultPath;
    /// <summary>
    ///   Shows dialog with export progress.
    /// </summary>
    property ShowProgress: Boolean read FShowProgress write FShowProgress;
    property OverwritePrompt: Boolean read FOverwritePrompt write FOverwritePrompt;
    /// <summary>
    ///   Creation time of the document.
    /// </summary>
    property CreationTime: TDateTime read FCreationTime write FCreationTime;
    /// <summary>
    ///   If True, the export engine exports only data bands objects.
    /// </summary>
    property DataOnly: Boolean read FDataOnly write FDataOnly;
    /// <summary>
    ///   This event fires before export starts.
    /// </summary>
    property OnBeforeExport: TNotifyEvent read FOnBeforeExport write FOnBeforeExport;
    /// <summary>
    ///   This event fires when export filter is being exported.
    /// </summary>
    property OnBeginExport: TNotifyEvent read FOnBeginExport write FOnBeginExport;
  end;

  /// <summary>
  ///   The TfrxCustomWizard class is the basic class for any wizards. Wizards
  ///   are called from the designer's menu "File|New...". Wizard should be
  ///   registered via frxWizards.Register method, defined in the frxDsgnIntf
  ///   unit.
  /// </summary>
  TfrxCustomWizard = class(TComponent)
  private
    FDesigner: TfrxCustomDesigner;
    FReport: TfrxReport;
  public
    constructor Create(AOwner: TComponent); override;
    /// <summary>
    ///   Returns the wizard description.
    /// </summary>
    class function GetDescription: String; virtual;
    /// <summary>
    ///   Method is called when running the wizard; it must return "True," if
    ///   the wizard finished working successfully and brought any changes to
    ///   the report.
    /// </summary>
    function Execute: Boolean; virtual; abstract;
    /// <summary>
    ///   Reference to the report designer.
    /// </summary>
    property Designer: TfrxCustomDesigner read FDesigner;
    /// <summary>
    ///   Reference to the report.
    /// </summary>
    property Report: TfrxReport read FReport;
  end;

  TfrxWizardClass = class of TfrxCustomWizard;

  TfrxCustomAggregateList = class(TObject)
  public
    procedure Clear; virtual; abstract;
    procedure ClearValues; virtual; abstract;
    procedure AddItems(Page: TfrxReportPage); virtual; abstract;
    procedure AddValue(DataRow: TfrxComponent; VColumn: Integer = 0); virtual; abstract;
    function AddCalculatedValue(DataRow: TfrxComponent; const Value: Variant; VColumn: Integer = 0): Boolean; virtual; abstract;
    function AddAggregatedItem(AObject: IfrxAggregateObject; Container: TfrxComponent; AList: TList = nil): Boolean; virtual; abstract;
    procedure DeleteValue(DataRow: TfrxComponent); virtual; abstract;
    procedure SaveValue(DataRow: TfrxComponent); virtual; abstract;
    procedure RestoreValue(DataRow: TfrxComponent); virtual; abstract;
    procedure EndKeep; virtual; abstract;
    procedure Reset(ParentContainer: TfrxComponent); virtual; abstract;
    procedure StartKeep; virtual; abstract;
    function GetValue(ParentContainer: TfrxComponent; const ComplexName: String;
      VColumn: Integer = 0): Variant; overload; virtual; abstract;
    function GetValue(ParentContainer: TfrxComponent; VColumn: Integer;
      const Name, Expression: String; DataRow: TfrxComponent; Flags: Integer): Variant; overload; virtual; abstract;
  end;

  /// <summary>
  ///   The TfrxCustomEngine class is the base class for report engine.
  ///   Instance of this class is stored in the TfrxReport.Engine property.
  /// </summary>
  TfrxCustomEngine = class(TPersistent)
  private
    FCurColumn: Integer;
    FCurVColumn: Integer;
    FCurLine: Integer;
    FCurLineThrough: Integer;
    FFinalPass: Boolean;
    FNotifyList: TList;
    FPageHeight: Extended;
    FPageWidth: Extended;
    FPreviewPages: TfrxCustomPreviewPages;
    FReport: TfrxReport;
    FRunning: Boolean;
    FStartDate: TDateTime;
    FStartTime: TDateTime;
    FTotalPages: Integer;
    FOnRunDialog: TfrxRunDialogEvent;
    FSecondScriptCall: Boolean;
    FCurTableRow: Integer;
    FCurTableColumn: Integer;
    function GetDoublePass: Boolean;
    procedure SetPreviewPages(const Value: TfrxCustomPreviewPages);
  protected
    FCurX: Extended;
    FCurY: Extended;
    function GetAggregates: TfrxCustomAggregateList; virtual; abstract;
    function GetCurColumnWidth: Extended; virtual; abstract;
    function GetCurColumnPos: Extended; virtual; abstract;
  protected
    function GetPageHeight: Extended; virtual;
  public
    constructor Create(AReport: TfrxReport); virtual;
    destructor Destroy; override;
    /// <summary>
    ///   Called when page printing is finished.
    /// </summary>
    procedure EndPage; virtual; abstract;
    /// <summary>
    ///   Breaks any keep operation currently active.
    /// </summary>
    procedure BreakAllKeep; virtual;
    function IsMultiColumnPage: Boolean; virtual; abstract;
    /// <summary>
    ///   Creates a new column in a multicolumn report. After the last column,
    ///   page break is automatically inserted.
    /// </summary>
    procedure NewColumn; virtual; abstract;
    /// <summary>
    ///   Creates a new page (page break).
    /// </summary>
    procedure NewPage; virtual; abstract;
    /// <summary>
    ///   Builds a shift tree of child objects for current container.
    /// </summary>
    /// <param name="Container">
    ///   Objects container.
    /// </param>
    procedure PrepareShiftTree(Container: TfrxReportComponent); virtual; abstract;
    /// <summary>
    ///   Displays a band with a specified name. After displaying a band, the
    ///   "CurY" position is automatically shifted.
    /// </summary>
    function ShowBand(Band: TfrxBand): TfrxBand; overload; virtual; abstract;
    /// <summary>
    ///   Displays a band with a specified name. After displaying a band, the
    ///   "CurY" position is automatically shifted.
    /// </summary>
    procedure ShowBand(Band: TfrxBandClass); overload; virtual; abstract;
    /// <summary>
    ///   Displays a band with a specified name. After displaying a band, the
    ///   "CurY" position is automatically shifted.
    /// </summary>
    procedure ShowBandByName(const BandName: String);
    /// <summary>
    ///   Performs stretch operation on passed container object.
    /// </summary>
    /// <param name="Band">
    ///   Container object. <br />
    /// </param>
    procedure Stretch(Band: TfrxReportComponent; SkipGetData: Boolean = False); virtual; abstract;
    /// <summary>
    ///   Returns back state of the container after stretch operation.
    /// </summary>
    /// <param name="Band">
    ///   Container object.
    /// </param>
    procedure UnStretch(Band: TfrxReportComponent); virtual; abstract;
    procedure InitializeSplit(Container: TfrxReportComponent; SavedObjects: TList; SplitObject: TList); virtual; abstract;
    function SplitObjects(Container: TfrxReportComponent; SavedObjects: TList; SplitObject: TList; AHeight: Extended): Boolean; virtual; abstract;
    procedure FinalizeSplit(Container: TfrxReportComponent; SavedObjects: TList; SplitObject: TList; SplitHeight: Extended); virtual; abstract;
    /// <summary>
    ///   Stops report preparation.
    /// </summary>
    procedure StopReport;
    function HeaderHeight(AddReprintOnNewPage: Boolean = False): Extended; virtual; abstract;
    /// <summary>
    ///   Returns height of the footers.
    /// </summary>
    function FooterHeight: Extended; virtual; abstract;
    /// <summary>
    ///   Returns height value of white space left on a page, in pixels.
    /// </summary>
    function FreeSpace: Extended; virtual; abstract;
    /// <summary>
    ///   Returns value of aggregate function for an expression.
    /// </summary>
    /// <param name="Name">
    ///   Method name (aggregate name).
    /// </param>
    /// <param name="Expression">
    ///   Expression.
    /// </param>
    /// <param name="Band">
    ///   Band reference.
    /// </param>
    /// <param name="Flags">
    ///   Flag.
    /// </param>
    function GetAggregateValue(const Name, Expression: String;
      DataRow: TfrxComponent;  Flags: Integer): Variant; virtual; abstract;
    function GetCurrentPage: TfrxPage; virtual; abstract;
    /// <summary>
    ///   Processed GetData for passed object. Used with paCustom flag.
    /// </summary>
    procedure ProcessObject(ReportObject: TfrxView); virtual; abstract;
    /// <summary>
    ///   This method runs the report engine.
    /// </summary>
    /// <param name="ARunDialogs">
    ///   If True, Processes report dialog pages.
    /// </param>
    /// <param name="AClearLast">
    ///   Clears last result from prepared pages.
    /// </param>
    /// <param name="APage">
    ///   Reference to a page(if nil, prepares all).
    /// </param>
    function Run(ARunDialogs: Boolean;
      AClearLast: Boolean = False; APage: TfrxPage = nil): Boolean; virtual; abstract;
    property Aggregates: TfrxCustomAggregateList read GetAggregates;
    property CurColumnPos: Extended read GetCurColumnPos;
    property CurColumnWidth: Extended read GetCurColumnWidth;
    /// <summary>
    ///   Current line number.
    /// </summary>
    property CurLine: Integer read FCurLine write FCurLine;
    /// <summary>
    ///   Current absolute line number (&lt;Line#&gt;).
    /// </summary>
    property CurLineThrough: Integer read FCurLineThrough write FCurLineThrough;
    /// <summary>
    ///   Current row of the table.
    /// </summary>
    property CurTableRow: Integer read FCurTableRow write FCurTableRow;
    /// <summary>
    ///   Current column of the table.
    /// </summary>
    property CurTableColumn: Integer read FCurTableColumn write FCurTableColumn;
    /// <summary>
    ///   List with objects that receives a notifications from report engine.
    /// </summary>
    property NotifyList: TList read FNotifyList;
    /// <summary>
    ///   Reference to prepared report pages.
    /// </summary>
    property PreviewPages: TfrxCustomPreviewPages read FPreviewPages write SetPreviewPages;
    /// <summary>
    ///   Reference to a report component.
    /// </summary>
    property Report: TfrxReport read FReport;
    /// <summary>
    ///   True when dialog pages are running.
    /// </summary>
    property Running: Boolean read FRunning write FRunning;
    /// <summary>
    ///   Occurs when report dialog page is previewed.
    /// </summary>
    property OnRunDialog: TfrxRunDialogEvent read FOnRunDialog write FOnRunDialog;
  published
    /// <summary>
    ///   The number of the current column in a multi-columned report. A value
    ///   can be assigned to this property.
    /// </summary>
    property CurColumn: Integer read FCurColumn write FCurColumn;
    /// <summary>
    ///   The number of the current column in a cross-tab report.
    /// </summary>
    property CurVColumn: Integer read FCurVColumn write FCurVColumn;
    /// <summary>
    ///   The current shift of the coordinates on the X-axis. A value can be
    ///   assigned to this property.
    /// </summary>
    property CurX: Extended read FCurX write FCurX;
    /// <summary>
    ///   The current shift of the coordinates on the Y-axis. A value can be
    ///   assigned to this property.
    /// </summary>
    property CurY: Extended read FCurY write FCurY;
    /// <summary>
    ///   Equal to "True," if the report is a two-pass one. Analogous to
    ///   Report.EngineOptions.DoublePass.
    /// </summary>
    property DoublePass: Boolean read GetDoublePass;
    /// <summary>
    ///   Equal to "True," if the last pass of the two-pass report is
    ///   performed.
    /// </summary>
    property FinalPass: Boolean read FFinalPass write FFinalPass;
    /// <summary>
    ///   Printable region's height, in pixels.
    /// </summary>
    property PageHeight: Extended read GetPageHeight write FPageHeight;
    /// <summary>
    ///   Printable region's width, in pixels.
    /// </summary>
    property PageWidth: Extended read FPageWidth write FPageWidth;
    /// <summary>
    ///   Time of report running. A counterpart of the &lt;Date&gt; system
    ///   variable.
    /// </summary>
    property StartDate: TDateTime read FStartDate write FStartDate;
    /// <summary>
    ///   Time of report running. A counterpart of the &lt;Time&gt; system
    ///   variable.
    /// </summary>
    property StartTime: TDateTime read FStartTime write FStartTime;
    /// <summary>
    ///   A number of output pages in a report. A counterpart of the
    ///   &lt;TotalPages&gt; system variable. The report should be a two-pass
    ///   one, so that this variable can be used.
    /// </summary>
    property TotalPages: Integer read FTotalPages write FTotalPages;
    /// <summary>
    ///   Flag for the determination of repeat call of event (in some cases the
    ///   event can be called repeatedly during grouping), if True then the
    ///   script has already been operated.
    /// </summary>
    property SecondScriptCall: Boolean read FSecondScriptCall write FSecondScriptCall;
  end;

  /// <summary>
  ///   This object represents the "Report tree" control element in a preview
  ///   window. Instance of this class is stored in the
  ///   TfrxReport.PreviewPages.Outline property.
  /// </summary>
  TfrxCustomOutline = class(TPersistent)
  private
    FCurItem: TfrXMLItem;
    FPreviewPages: TfrxCustomPreviewPages;
  protected
    function GetCount: Integer; virtual; abstract;
  public
    constructor Create(APreviewPages: TfrxCustomPreviewPages); virtual;
    /// <summary>
    ///   Adds an element with the "Text" name to the current tree position.
    ///   The current report's page and the Top position on the page are
    ///   associated with the element.
    /// </summary>
    procedure AddItem(const Text: String; Top: Integer); virtual; abstract;
    procedure DeleteItems(From: TfrXMLItem); virtual; abstract;
    procedure LevelDown(Index: Integer); virtual; abstract;
    /// <summary>
    ///   Shifts the current position in the tree to the root level.
    /// </summary>
    procedure LevelRoot; virtual; abstract;
    /// <summary>
    ///   Shifts the current position in the tree on one level up.
    /// </summary>
    procedure LevelUp; virtual; abstract;
    procedure GetItem(Index: Integer; var Text: String;
      var Page, Top: Integer); virtual; abstract;
    procedure ShiftItems(From: TfrXMLItem; NewTop: Integer); virtual; abstract;
    function Engine: TfrxCustomEngine;
    function GetCurPosition: TfrXMLItem; virtual; abstract;
    property Count: Integer read GetCount;
    property CurItem: TfrXMLItem read FCurItem write FCurItem;
    property PreviewPages: TfrxCustomPreviewPages read FPreviewPages;
  end;

  /// <summary>
  ///   This class encapsulate data for object processing engine.
  /// </summary>
  TfrxObjectProcessing = class(TPersistent)
  private
    FProcessAt: TfrxProcessAtMode;
    FGroupLevel: Integer;
  public
    procedure Assign(Source: TPersistent); override;
  published
    /// <summary>
    ///   This property controls when to process expression calculation(for
    ///   some TfrxView objects).
    /// </summary>
    property ProcessAt: TfrxProcessAtMode read FProcessAt write FProcessAt default paDefault;
    /// <summary>
    ///   It represent level of nested group which will be used to calculate
    ///   object data. This property used with paGroupFinished flag. Zero means
    ///   all groups call data calculation for current object. The value above
    ///   zero represents nested group level number.
    /// </summary>
    property GroupLevel: Integer read FGroupLevel write FGroupLevel default 0;
  end;

  /// <summary>
  ///   Internal class used by data processing engine. Stores values for post
  ///   processor.
  /// </summary>
  TfrxMacrosItem = class(TObject)
  private
    FItems: TWideStrings;
    FSupressedValue: Variant;
    FLastIndex: Integer;
    FComponent: TfrxComponent;
    FBaseComponent: TfrxComponent;
    FBandName: String;
    FNeedReset: Boolean;
    FProcessAt: TfrxProcessAtMode;
    FGroupLevel: Integer;
    FDataLevel: Integer;
    FParent: TfrxPostProcessor;
    function GetItem(Index: Integer): WideString;
    procedure SetItem(Index: Integer; const Value: WideString);
  public
    constructor Create;
    destructor Destroy; override;
    function AddObject(const S: WideString; AObject: TObject; bSupressed: Boolean = false): Integer;
    function Count: Integer;
    property Item[Index: Integer]: WideString read GetItem write SetItem;
  end;

  /// <summary>
  ///   This class used as post-process data dictionary. It controls behaviour
  ///   of post-processed objects and stores data and expressions of these
  ///   objects. Used by the report engine for delay expression calculation in
  ///   selected objects.
  /// </summary>
  TfrxPostProcessor = class(TPersistent)
  private
    FMacroses: TStrings;
    FProcessList: TList;
    FBands: TStrings;
    FGroupLevel: Integer;
    FDataLevel: Integer;
    FProcessing: Boolean;
  public
    /// <summary>
    ///   Default constructor.
    /// </summary>
    constructor Create;
    destructor Destroy; override;
    /// <summary>
    ///   Clears and destroys all items in dictionary list.
    /// </summary>
    procedure Clear;
    /// <summary>
    ///   Adds new View component to the list for delayed calculation. Places
    ///   an item to the list for future calculation.
    /// </summary>
    /// <param name="BandName">
    ///   Name of the parent band.
    /// </param>
    /// <param name="Name">
    ///   Item name(usually component name).
    /// </param>
    /// <param name="Content">
    ///   Content to be stored and calculated later(usually an expression).
    /// </param>
    /// <param name="ProcessMode">
    ///   Controls when to process expression calculation.
    /// </param>
    /// <param name="aComponent">
    ///   Reference to a view component.
    /// </param>
    /// <param name="bSupressed">
    ///   process suppressed values.
    /// </param>
    /// <param name="bEmpty">
    ///   Clear suppressed values.
    /// </param>
    function Add(const BandName: String; const Name: String; const Content: WideString; ProcessMode: TfrxProcessAtMode; aComponent: TfrxComponent; bSupressed: Boolean = false; bEmpty: Boolean = false): Integer;
    /// <summary>
    ///   Enter new group level.
    /// </summary>
    procedure EnterGroup;
    /// <summary>
    ///   Leaves current group level.
    /// </summary>
    procedure LeaveGroup;
    /// <summary>
    ///   Enters new data level.
    /// </summary>
    procedure EnterData;
    /// <summary>
    ///   Leaves new data level.
    /// </summary>
    procedure LeaveData;
    /// <summary>
    ///   Not used
    /// </summary>
    function GetValue(const MacroIndex: Integer; const MacroLine: Integer): WideString;
    /// <summary>
    ///   Returns list of items by name of object.
    /// </summary>
    /// <param name="MacroName">
    ///   Object name.
    /// </param>
    function GetMacroList(const MacroName: String): TWideStrings;
    /// <summary>
    ///   Loads value to component from dictionary.
    /// </summary>
    /// <param name="aComponent">
    ///   Reference to view component.
    /// </param>
    function LoadValue(aComponent: TfrxComponent): Boolean;
    /// <summary>
    ///   Save dictionary to XML node.
    /// </summary>
    procedure SaveToXML(Item: TfrXMLItem);
    /// <summary>
    ///   Load dictionary from XML node.
    /// </summary>
    procedure LoadFromXML(Item: TfrXMLItem);
    /// <summary>
    ///   Resets suppressed values in the list.
    /// </summary>
    procedure ResetSuppressed;
    /// <summary>
    ///   Resets duplicates values in the list.
    /// </summary>
    procedure ResetDuplicates(const BandName: String = '');
    /// <summary>
    ///   Calculates an expression for items in the list. It called by the
    ///   report engine when one of the events occurs. Removes objects from the
    ///   processing list when appropriate event occurs.
    /// </summary>
    /// <param name="AReport">
    ///   Reference to a report component.
    /// </param>
    /// <param name="ABand">
    ///   Name of the band container being processed.
    /// </param>
    /// <param name="ProcessMode">
    ///   Mode that should be processed.
    /// </param>
    procedure ProcessExpressions(AReport: TfrxReport; ABand: TfrxBand; ProcessMode: TfrxProcessAtMode);
    /// <summary>
    ///   It calculates an expressions for selected view component.
    /// </summary>
    /// <param name="AReport">
    ///   Reference to a report component.
    /// </param>
    /// <param name="aComponent">
    ///   Reference to a view component.
    /// </param>
    procedure ProcessObject(AReport: TfrxReport; aComponent: TfrxComponent);
    /// <summary>
    ///   List of items in the dictionary.
    /// </summary>
    property Macroses: TStrings read FMacroses;
  end;

  /// <summary>
  ///   The TfrxCustomPreviewPages class is the base class for pages of a
  ///   prepared report. Instance of this class is stored in the
  ///   TfrxReport.PreviewPages property.
  /// </summary>
  TfrxCustomPreviewPages = class(TObject)
  private
    FAddPageAction: TfrxAddPageAction; { used in the cross-tab renderer }
    FCurPage: Integer;
    FCurPreviewPage: Integer;
    FFirstPage: Integer;               {  used in the composite reports }
    FOutline: TfrxCustomOutline;
    FReport: TfrxReport;
    FEngine: TfrxCustomEngine;
    FPictureCacheOptions: TfrxPictureCacheOptions;
    FStyles: TfrxStyles;
    function GetStyles: TfrxStyles;
    procedure SetStyles(const Value: TfrxStyles);
  protected
    FCopyNo: Integer;
    FPrintScale: Extended;
    FPostProcessor: TfrxPostProcessor;
    procedure ClearPageCache; virtual; abstract;
    procedure ClearStyles;
    function FindStyle(const AStyleName: string): TfrxStyleItem;
    function GetCount: Integer; virtual; abstract;
    function GetPage(Index: Integer): TfrxReportPage; virtual; abstract;
    function GetPageSize(Index: Integer): TPoint; virtual; abstract;
    function GetPictureCacheOptions: TfrxPictureCacheOptions; virtual;
  public
    /// <summary>
    ///   Default constructor.
    /// </summary>
    /// <param name="AReport">
    ///   Reference to an owner report component.
    /// </param>
    constructor Create(AReport: TfrxReport); virtual;
    destructor Destroy; override;
    /// <summary>
    ///   Clears a prepared report.
    /// </summary>
    procedure Clear; virtual; abstract;
    /// <summary>
    ///   Initializes preview pages.
    /// </summary>
    procedure Initialize; virtual; abstract;

    /// <summary>
    ///   This method adds an object to current preview page.
    /// </summary>
    /// <param name="Obj">
    ///   Reference to report object.
    /// </param>
    procedure AddObject(Obj: TfrxComponent); virtual; abstract;
    /// <summary>
    ///   This method adds new preview page or increases number of current
    ///   preview page by 1 depends on AddPageAction flag.
    /// </summary>
    /// <param name="Page">
    ///   Reference to report page.
    /// </param>
    procedure AddPage(Page: TfrxReportPage); virtual; abstract;
    /// <summary>
    ///   This method adds report template page to preview page objects
    ///   dictionary.
    /// </summary>
    /// <param name="Page">
    ///   Reference to report template page.
    /// </param>
    procedure AddSourcePage(Page: TfrxReportPage); virtual; abstract;
    /// <summary>
    ///   This method adds report template object to preview page objects
    ///   dictionary.
    /// </summary>
    /// <param name="Obj">
    ///   Report object.
    /// </param>
    procedure AddToSourcePage(Obj: TfrxComponent); virtual; abstract;
    /// <summary>
    ///   The report engine call this method when start first pass of double
    ///   pass report.
    /// </summary>
    procedure BeginPass; virtual; abstract;
    /// <summary>
    ///   Clears results of the first pass.
    /// </summary>
    procedure ClearFirstPassPages; virtual; abstract;
    /// <summary>
    ///   Cuts objects from current serialized XML page by given XML node
    ///   position to temporary cut-buffer. <br />
    /// </summary>
    /// <param name="APosition">
    ///   Index of XML node in serialized preview page XML node.
    /// </param>
    procedure CutObjects(APosition: Integer); virtual; abstract;
    /// <summary>
    ///   Deletes serialized XML objects at given XML node position from
    ///   current serialized preview page.
    /// </summary>
    /// <param name="APosition">
    ///   XML node position.
    /// </param>
    procedure DeleteObjects(APosition: Integer); virtual; abstract;
    /// <summary>
    ///   Deletes all preview pages anchors starts at given position.
    /// </summary>
    /// <param name="From">
    ///   Start item position.
    /// </param>
    procedure DeleteAnchors(From: Integer); virtual; abstract;
    /// <summary>
    ///   Called by the report engine when report preparation finished.
    /// </summary>
    procedure Finish; virtual;
    /// <summary>
    ///   Fires mouse leave behaviour for preview pages objects.
    /// </summary>
    procedure FireMouseLeave; virtual;
    /// <summary>
    ///   Updates logical page number variable.
    /// </summary>
    procedure IncLogicalPageNumber; virtual; abstract;
    function IsStylesActive: Boolean;
    /// <summary>
    ///   Resets logical page number variable.
    /// </summary>
    procedure ResetLogicalPageNumber; virtual; abstract;
    /// <summary>
    ///   Pastes cut objects from temporary buffer to current serialized XML
    ///   page at given X, Y position on the preview page.
    /// </summary>
    /// <param name="X">
    ///   X coordinate (column position).
    /// </param>
    /// <param name="Y">
    ///   Y coordinate (band position).
    /// </param>
    procedure PasteObjects(X, Y: Extended); virtual; abstract;
    /// <summary>
    ///   Shifts all preview pages anchors starts at given position by new top
    ///   position.
    /// </summary>
    /// <param name="From">
    ///   Start item position.
    /// </param>
    /// <param name="NewTop">
    ///   New top position.
    /// </param>
    procedure ShiftAnchors(From, NewTop: Integer); virtual; abstract;

    function GetPicture(CacheType: TfrxCachedGraphicType; ImageIndex: Integer): IfrxCachedGraphic; virtual; abstract;
    /// <summary>
    ///   Adds picture object to preview pages picture cache.
    /// </summary>
    procedure AddPicture(Picture: TfrxPictureView); virtual; abstract;
    function AddGraphic(Graphic: TGraphic): Integer; virtual; abstract;
    /// <summary>
    ///   Checks if band exist on current preview page.
    /// </summary>
    /// <param name="Band">
    ///   Reference to Band in a report template.
    /// </param>
    function BandExists(Band: TfrxBand): Boolean; virtual; abstract;
    /// <summary>
    ///   Returns current XML node position in serialized preview page.
    /// </summary>
    function GetCurPosition: Integer; virtual; abstract;
    /// <summary>
    ///   Returns current anchor item position.
    /// </summary>
    function GetAnchorCurPosition: Integer; virtual; abstract;
    /// <summary>
    ///   Returns last Y position on preview page which is not occupied by band
    ///   objects.
    /// </summary>
    /// <param name="ColumnPosition">
    ///   Column X position.
    /// </param>
    function GetLastY(ColumnPosition: Extended = 0): Extended; virtual; abstract;
    /// <summary>
    ///   Returns logical page number for current page.
    /// </summary>
    function GetLogicalPageNo: Integer; virtual; abstract;
    /// <summary>
    ///   Returns logical total pages count.
    /// </summary>
    function GetLogicalTotalPages: Integer; virtual; abstract;
    /// <summary>
    ///   Inserts an empty preview page at given position.
    /// </summary>
    /// <param name="Index">
    ///   Page index.
    /// </param>
    procedure AddEmptyPage(Index: Integer); virtual; abstract;
    /// <summary>
    ///   Deletes preview page at given position.
    /// </summary>
    /// <param name="Index">
    ///   Page index.
    /// </param>
    procedure DeletePage(Index: Integer); virtual; abstract;
    /// <summary>
    ///   This method writes changes of prepared page with all objects to
    ///   serialized XML. Use this method to save all changes made on prepared
    ///   page.
    /// </summary>
    /// <param name="Index">
    ///   Page index in prepared report.
    /// </param>
    /// <param name="Page">
    ///   Reference to modified page object.
    /// </param>
    procedure ModifyPage(Index: Integer; Page: TfrxReportPage); virtual; abstract;
    /// <summary>
    ///   This method writes changes of report object on prepared report page
    ///   to XML node with using objects data dictionary. Use this method to
    ///   save changes made on object of prepared report. This method uses data
    ///   dictionary and works much faster than ModifyPage method.
    /// </summary>
    /// <param name="Component">
    ///   Reference to modified prepared report object.
    /// </param>
    procedure ModifyObject(Component: TfrxComponent); virtual; abstract;
    /// <summary>
    ///   Draws prepared page on Canvas.
    /// </summary>
    /// <param name="Index">
    ///   Index of page to draw.
    /// </param>
    /// <param name="Canvas">
    ///   Destination Canvas.
    /// </param>
    /// <param name="ScaleX">
    ///   Scale X.
    /// </param>
    /// <param name="ScaleY">
    ///   Scale Y.
    /// </param>
    /// <param name="OffsetX">
    ///   X offset.
    /// </param>
    /// <param name="OffsetY">
    ///   Y offset.
    /// </param>
    /// <param name="bHighlightEditable">
    ///   Draws with highlight of editable objects.
    /// </param>
    procedure DrawPage(Index: Integer; Canvas: TCanvas; ScaleX, ScaleY,
      OffsetX, OffsetY: Extended; bHighlightEditable: Boolean = False); virtual; abstract;
    /// <summary>
    ///   This method sends an interactive message to preview page or/and
    ///   objects on that page.
    /// </summary>
    /// <param name="Index">
    ///   Page index of the page.
    /// </param>
    /// <param name="X">
    ///   X coordinate relative to parent preview workspace view(mouse
    ///   coordinates).
    /// </param>
    /// <param name="Y">
    ///   Y coordinate relative to parent preview workspace view(mouse
    ///   coordinates).
    /// </param>
    /// <param name="Button">
    ///   Mouse button.
    /// </param>
    /// <param name="Shift">
    ///   Shift button.
    /// </param>
    /// <param name="Scale">
    ///   Current scale of preview window.
    /// </param>
    /// <param name="OffsetX">
    ///   Horizontal offset (start of draw area based on horizontal scroll).
    /// </param>
    /// <param name="OffsetY">
    ///   Vertical offset (start of draw area based on vertical scroll). <br />
    /// </param>
    /// <param name="EventParams">
    ///   Structure used for additional event parameters (like event type).
    ///   Also stores return value.
    /// </param>
    procedure ObjectOver(Index: Integer; X, Y: Integer; Button: TMouseButton;
      Shift: TShiftState; Scale, OffsetX, OffsetY: Extended;
      var EventParams: TfrxPreviewIntEventParams); virtual; abstract;
    /// <summary>
    ///   Creates a composite prepared report through concatenation of preview
    ///   pages in to current preview pages.
    /// </summary>
    /// <param name="Report">
    ///   Source report.
    /// </param>
    procedure AddFrom(Report: TfrxReport); virtual; abstract;
    procedure AddFromPreviewPages(vPreviewPages: TfrxCustomPreviewPages); virtual; abstract;

    /// <summary>
    ///   Loads a prepared report from a stream. The AllowPartialLoading
    ///   parameter allows to load large files partially for saving memory.
    /// </summary>
    procedure LoadFromStream(Stream: TStream;
      AllowPartialLoading: Boolean = False); virtual; abstract;
    /// <summary>
    ///   Saves a prepared report to a stream.
    /// </summary>
    procedure SaveToStream(Stream: TStream); virtual; abstract;
    /// <summary>
    ///   Saves a prepared report by using IO transport class to get stream for
    ///   a prepared pages.
    /// </summary>
    /// <param name="Filter">
    ///   IO transport filter instance.
    /// </param>
    /// <param name="FileName">
    ///   Name of the prepared file.
    /// </param>
    function SaveToFilter(Filter: TfrxCustomIOTransport; const FileName: String): Boolean; virtual; abstract;
    /// <summary>
    ///   Loads a prepared report from a file.
    /// </summary>
    function LoadFromFile(const FileName: String;
      ExceptionIfNotFound: Boolean = False): Boolean; virtual; abstract;
    /// <summary>
    ///   Saves a prepared report to a file.
    /// </summary>
    procedure SaveToFile(const FileName: String); virtual; abstract;
    procedure SetActiveStyleSheet(const AName: String);
    /// <summary>
    ///   Prints a report.
    /// </summary>
    function Print: Boolean; virtual; abstract;
    /// <summary>
    ///   Exports a report.
    /// </summary>
    function Export(Filter: TfrxCustomExportFilter): Boolean; virtual; abstract;

    /// <summary>
    ///   This flag used by AddPage method. apWriteOver - writes over current
    ///   page if it isn't last page (in case it's last page a new page will be
    ///   added). apWriteOver - always adds new page.
    /// </summary>
    property AddPageAction: TfrxAddPageAction read FAddPageAction write FAddPageAction;
    /// <summary>
    ///   Number of pages in the prepared report.
    /// </summary>
    property Count: Integer read GetCount;
    /// <summary>
    ///   Current page number.
    /// </summary>
    property CurPage: Integer read FCurPage write FCurPage;
    /// <summary>
    ///   Current previewed page number.
    /// </summary>
    property CurPreviewPage: Integer read FCurPreviewPage write FCurPreviewPage;
    /// <summary>
    ///   Reference to the report engine object.
    /// </summary>
    property Engine: TfrxCustomEngine read FEngine write FEngine;
    /// <summary>
    ///   First page number in composite report.
    /// </summary>
    property FirstPage: Integer read FFirstPage write FFirstPage;
    /// <summary>
    ///   Link to the "Report outline" class.
    /// </summary>
    property Outline: TfrxCustomOutline read FOutline;
    /// <summary>
    ///   List of pages in the prepared report.
    /// </summary>
    property Page[Index: Integer]: TfrxReportPage read GetPage;
    /// <summary>
    ///   List of page sizes in the prepared report.
    /// </summary>
    property PageSize[Index: Integer]: TPoint read GetPageSize;
    /// <summary>
    ///   Number of printed copy.
    /// </summary>
    property PrintCopyNo: Integer read FCopyNo;
    /// <summary>
    ///   Scale for print output.
    /// </summary>
    property PrintScale: Extended read FPrintScale;
    /// <summary>
    ///   Reference to a report object.
    /// </summary>
    property Report: TfrxReport read FReport;
    property Styles: TfrxStyles read GetStyles write SetStyles;
    /// <summary>
    ///   Post-process data dictionary object.
    /// </summary>
    property PostProcessor: TfrxPostProcessor read FPostProcessor;
    property PictureCacheOptions: TfrxPictureCacheOptions read GetPictureCacheOptions;
  end;

  /// <summary>
  ///   The TfrxCustomPreview class is the base class for TfrxPreview class.
  /// </summary>
  TfrxCustomPreview = class(TfrxDPIAwareCustomControl)
  private
    FUseReportHints: Boolean;
  protected
    FPreviewForm: TForm;
    FInPlaceditorsList: TfrxComponentEditorsList;
    FSelectionList: TfrxSelectedObjectsList;
    function GetReport: TfrxReport; virtual;
    procedure DoFinishInPlace(Sender: TObject; Refresh, Modified: Boolean); virtual;
    function GetPreviewPages: TfrxCustomPreviewPages; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    /// <summary>
    ///   This method initialize TfrxInteractiveEventsParams structure with
    ///   default values.
    /// </summary>
    procedure SetDefaultEventParams(var EventParams: TfrxInteractiveEventsParams); virtual;
    /// <summary>
    ///   Initializes report preview with giving report and preview pages.
    /// </summary>
    /// <param name="aReport">
    ///   Reference to a report component.
    /// </param>
    /// <param name="aPrevPages">
    ///   Reference to preview pages object.
    /// </param>
    function Init(aReport: TfrxReport; aPrevPages: TfrxCustomPreviewPages): Boolean; virtual; abstract;
    /// <summary>
    ///   Uninitializes report preview for passed report component. <br />
    /// </summary>
    /// <param name="aReport">
    ///   Reference to a report component.
    /// </param>
    procedure UnInit(aReport: TfrxReport); virtual; abstract;
    /// <summary>
    ///   Locks report preview for any update and draw operations.
    /// </summary>
    function Lock: Boolean; virtual; abstract;
    /// <summary>
    ///   Unlock report preview.
    /// </summary>
    /// <param name="DoUpdate">
    ///   Refresh after unlock.
    /// </param>
    procedure Unlock(DoUpdate: Boolean = True); virtual; abstract;
    /// <summary>
    ///   Rebuilds a report for current active preview pages.
    /// </summary>
    procedure RefreshReport; virtual; abstract;
    /// <summary>
    ///   Copies content for selected objects to clipboard. This method uses
    ///   InPlace editors to process copy behaviour of the selected objects.
    /// </summary>
    procedure InternalCopy; virtual;
    /// <summary>
    ///   Pastes content to selected objects from clipboard. This method uses
    ///   InPlace editors to process paste behaviour of the selected objects.
    /// </summary>
    procedure InternalPaste; virtual;
    /// <summary>
    ///   This method checks if paste is available for selected objects from
    ///   clipboard. It uses InPlace editors to process copy-paste behaviour of
    ///   the selected objects.
    /// </summary>
    function InternalIsPasteAvailable: Boolean; virtual;
    /// <summary>
    ///   The report engine calls this method when starts some action at
    ///   current report(preparation, export, print).
    /// </summary>
    /// <param name="Sender">
    ///   Sender object.
    /// </param>
    /// <param name="ProgressType">
    ///   Type of progress.
    /// </param>
    /// <param name="Progress">
    ///   Progress value.
    /// </param>
    procedure InternalOnProgressStart(Sender: TfrxReport;
      ProgressType: TfrxProgressType; Progress: Integer); virtual; abstract;
    /// <summary>
    ///   The report engine calls this method to update action progress at
    ///   current report(preparation, export, print).
    /// </summary>
    /// <param name="Sender">
    ///   Sender object.
    /// </param>
    /// <param name="ProgressType">
    ///   Type of progress
    /// </param>
    /// <param name="Progress">
    ///   Progress value.
    /// </param>
    procedure InternalOnProgress(Sender: TfrxReport;
      ProgressType: TfrxProgressType; Progress: Integer); virtual; abstract;
    /// <summary>
    ///   The report engine calls this method when stops some action at current
    ///   report(preparation, export, print).
    /// </summary>
    /// <param name="Sender">
    ///   Sender object.
    /// </param>
    /// <param name="ProgressType">
    ///   Type of progress
    /// </param>
    /// <param name="Progress">
    ///   Progress value.
    /// </param>
    procedure InternalOnProgressStop(Sender: TfrxReport;
      ProgressType: TfrxProgressType; Progress: Integer); virtual; abstract;
    /// <summary>
    ///   Creates a new tab in the preview window linked with passed report
    ///   object.
    /// </summary>
    /// <param name="AReport">
    ///   Reference to a report component.
    /// </param>
    /// <param name="TabName">
    ///   Name of the Tab.
    /// </param>
    /// <param name="TabCaption">
    ///   Caption of the Tab.
    /// </param>
    /// <param name="FreeObjects">
    ///   If True, report object which is belong to the Tab object will be
    ///   freed.
    /// </param>
    /// <param name="aDetailPage">
    ///   Name of Detail page.
    /// </param>
    procedure AddPreviewTab(AReport: TfrxReport; const TabName: String; const TabCaption: String = ''; FreeObjects: Boolean = True; aDetailPage: String = ''); virtual; abstract;
    /// <summary>
    ///   Creates or switch to a new tab in the preview window linked with
    ///   passed report object.
    /// </summary>
    /// <param name="AReport">
    ///   Reference to a report component.
    /// </param>
    /// <param name="TabName">
    ///   Name of the Tab.
    /// </param>
    /// <param name="TabCaption">
    ///   Caption of the Tab.
    /// </param>
    /// <param name="FreeObjects">
    ///   Creates a new tab in the preview window linked with passed report
    ///   object.
    /// </param>
    /// <param name="aDetailPage">
    ///   Name of Detail page.
    /// </param>
    procedure AddPreviewTabOrSwitch(AReport: TfrxReport; const TabName: String; const TabCaption: String = ''; FreeObjects: Boolean = True; aDetailPage: String = ''); virtual; abstract;
    /// <summary>
    ///   Closes and removes Tab from report preview.
    /// </summary>
    /// <param name="TabIndex">
    ///   Index of the Tab.
    /// </param>
    procedure RemoveTab(TabIndex: Integer); virtual; abstract;
    /// <summary>
    ///   Checks if preview has Tab with giving name.
    /// </summary>
    /// <param name="TabName">
    ///   Name of the Tab.
    /// </param>
    function HasTab(const TabName: String): Boolean; overload; virtual; abstract;
    /// <summary>
    ///   Checks if preview has Tab with giving report object.
    /// </summary>
    /// <param name="aReport">
    ///   Reference to a report component.
    /// </param>
    function HasTab(const aReport: TfrxReport): Boolean; overload; virtual; abstract;
    /// <summary>
    ///   Returns True when report preview has visible tabs.
    /// </summary>
    function HasVisibleTabs: Boolean; virtual; abstract;
    /// <summary>
    ///   Called when active preview pages was changed.
    /// </summary>
    procedure PreviewPagesChanged; virtual; abstract;
    /// <summary>
    ///   Switches to Tab by giving name.
    /// </summary>
    /// <param name="TabName">
    ///   Name of the Tab.
    /// </param>
    procedure SwitchToTab(const TabName: String); overload; virtual; abstract;
    /// <summary>
    ///   Switches to Tab by giving report.
    /// </summary>
    /// <param name="aReport">
    ///   Reference to a report component.
    /// </param>
    procedure SwitchToTab(const aReport: TfrxReport); overload; virtual; abstract;

    /// <summary>
    ///   Reference to an active preview pages object.
    /// </summary>
    property PreviewPages: TfrxCustomPreviewPages read GetPreviewPages;// write FPreviewPages;
    /// <summary>
    ///   Reference to an active report object.
    /// </summary>
    property Report: TfrxReport read GetReport;// write FReport;
    /// <summary>
    ///   If True, shows hints of the report objects. Default True.
    /// </summary>
    property UseReportHints: Boolean read FUseReportHints write FUseReportHints;
  end;

  TfrxCompressorClass = class of TfrxCustomCompressor;

{$IFDEF DELPHI16}
/// <summary>
///   The TfrxCustomCompressor class is the base class for compressors that
///   allows compression/decompression of the FR3 and FP3 files.
/// </summary>
[ComponentPlatformsAttribute(pidWin32 or pidWin64)]
{$ENDIF}
  TfrxCustomCompressor = class(TComponent)
  private
    FIsFR3File: Boolean;
    FOldCompressor: TfrxCompressorClass;
    FReport: TfrxReport;
    FStream: TStream;
    FTempFile: String;
    FFilter: TfrxSaveToCompressedFilter;
    procedure SetIsFR3File(const Value: Boolean);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    /// <summary>
    ///   Method will decompress the data from a Source parameter and save
    ///   resulting data to a stream defined in the Stream property.
    /// </summary>
    function Decompress(Source: TStream): Boolean; virtual; abstract;
    /// <summary>
    ///   Method will compress the data from the Stream property and save
    ///   resulting data to a Dest stream.
    /// </summary>
    procedure Compress(Dest: TStream); virtual; abstract;
    procedure CreateStream;
    /// <summary>
    ///   Determines if there is compressed data is the FR3 file. FR3 file can
    ///   be compressed using maximum compression.
    /// </summary>
    property IsFR3File: Boolean read FIsFR3File write SetIsFR3File;
    /// <summary>
    ///   Reference to a report.
    /// </summary>
    property Report: TfrxReport read FReport write FReport;
    /// <summary>
    ///   Stream used by compressing/decompressing operations.
    /// </summary>
    property Stream: TStream read FStream write FStream;
  end;

  TfrxCrypterClass = class of TfrxCustomCrypter;

{$IFDEF DELPHI16}
[ComponentPlatformsAttribute(pidWin32 or pidWin64)]
{$ENDIF}
  TfrxCustomCrypter = class(TComponent)
  private
    FStream: TStream;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Decrypt(Source: TStream; const Key: AnsiString): Boolean; virtual; abstract;
    procedure Crypt(Dest: TStream; const Key: AnsiString); virtual; abstract;
    procedure CreateStream;
    property Stream: TStream read FStream write FStream;
  end;

  TfrxLoadEvent = function(Sender: TfrxReport; Stream: TStream): Boolean of object;
  TfrxGetScriptValueEvent = function(var Params: Variant): Variant of object;

  TfrxFR2Events = class(TObject)
  private
    FOnGetValue: TfrxGetValueEvent;
    FOnPrepareScript: TNotifyEvent;
    FOnLoad: TfrxLoadEvent;
    FOnGetScriptValue: TfrxGetScriptValueEvent;
    FFilter: String;
  public
    property OnGetValue: TfrxGetValueEvent read FOnGetValue write FOnGetValue;
    property OnPrepareScript: TNotifyEvent read FOnPrepareScript write FOnPrepareScript;
    property OnLoad: TfrxLoadEvent read FOnLoad write FOnLoad;
    property OnGetScriptValue: TfrxGetScriptValueEvent read FOnGetScriptValue write FOnGetScriptValue;
    property Filter: String read FFilter write FFilter;
  end;

  /// <summary>
  ///   This class represents global list of datasets. Every ancestor of
  ///   TfrxDataSet registers in this list during construction and removes when
  ///   destroy.
  /// </summary>
  TfrxGlobalDataSetList = class(TList)
{$IFNDEF NO_CRITICAL_SECTION}
    FCriticalSection: TCriticalSection;
{$ENDIF}
  public
    /// <summary>
    ///   Default constructor.
    /// </summary>
    constructor Create;
    destructor Destroy; override;
    /// <summary>
    ///   Enters in critical section.
    /// </summary>
    procedure Lock;
    /// <summary>
    ///   Leaves critical section.
    /// </summary>
    procedure Unlock;
  end;

  { InPlace editors and custom draw for designer and preview }
  { remove unused properties }
  /// <summary>
  ///   The base class for InPlace editors. With InPlace editors programmers
  ///   may extend functionality of the report designer and the report preview.
  ///   This class handles mouse events, drag and drop, clipboard and custom
  ///   draw. It can be used to create custom in place editor for the report
  ///   designer, override copy/paste behaviour for selected object and draws
  ///   custom editors.
  /// </summary>
  TfrxInPlaceEditor = class(TObject)
  protected
    FComponent: TfrxComponent;
    FParentComponent: TComponent;
    FOwner: TWinControl;
    FOnFinishInPlace: TfrxOnFinishInPlaceEdit;
    FClassRef: TfrxComponentClass;
    FOffsetX: Extended;
    FOffsetY: Extended;
    FScale: Extended;
    FLocked: Boolean;
    FDevicePPI: Integer;
    { used to call other editors }
    FEditors: TfrxComponentEditorsList;
    FComponents: TfrxSelectedObjectsList;
    FClipboardObject: TPersistent;
    procedure SetComponent(const Value: TfrxComponent);
  public
    /// <summary>
    ///   Default constructor.
    /// </summary>
    /// <param name="aClassRef">
    ///   Class reference of report component used in the editor.
    /// </param>
    /// <param name="aOwner">
    ///   Widow control which owns editor(designer or preview).
    /// </param>
    constructor Create(aClassRef: TfrxComponentClass; aOwner: TWinControl); virtual;
    destructor Destroy; override;
    /// <summary>
    ///   Returns active work area of the editor.
    /// </summary>
    function GetActiveRect: TRect; virtual;
    /// <summary>
    ///   Not used.
    /// </summary>
    function HasInPlaceEditor: Boolean; virtual;
    /// <summary>
    ///   Not used.
    /// </summary>
    function HasCustomEditor: Boolean; virtual;
    /// <summary>
    ///   Draws object editor over object canvas.
    /// </summary>
    /// <param name="aCanvas">
    ///   Destination Canvas.
    /// </param>
    /// <param name="aRect">
    ///   Draw area.
    /// </param>
    procedure DrawCustomEditor(aCanvas: TCanvas; aRect: TRect); virtual;
    /// <summary>
    ///   Sets parameters used by the report preview or the report designer.
    /// </summary>
    /// <param name="aOffsetX">
    ///   X offset of work area.
    /// </param>
    /// <param name="aOffsetY">
    ///   Y offset of work area. <br />
    /// </param>
    /// <param name="aScale">
    ///   Scale of work area. <br />
    /// </param>
    procedure SetOffset(aOffsetX, aOffsetY, aScale: Extended);
    /// <summary>
    ///   Not used.
    /// </summary>
    function FillData: Boolean; virtual;
    procedure EditInPlace(aParent: TComponent; aRect: TRect); virtual;
    /// <summary>
    ///   This method should close in place editor and apply changes to edited
    ///   object.
    /// </summary>
    function EditInPlaceDone: Boolean; virtual;
    /// <summary>
    ///   Calls OnFinishInPlace event handler.
    /// </summary>
    procedure DoFinishInPlace(Sender: TObject; Refresh, Modified: Boolean); virtual;
    /// <summary>
    ///   The report engine calls it when DragOver operation is active over the
    ///   object with class that InPlace editor serves.
    /// </summary>
    /// <param name="Source">
    ///   Source object.
    /// </param>
    /// <param name="X">
    ///   Current mouse x coordinate.
    /// </param>
    /// <param name="Y">
    ///   Current mouse y coordinate. <br />
    /// </param>
    /// <param name="State">
    ///   State of drag and drop operation.
    /// </param>
    /// <param name="Accept">
    ///   Should return true when editor accepts drag and drop for current
    ///   object.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with parameters of interactive events.
    /// </param>
    function DoCustomDragOver(Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean; var EventParams: TfrxInteractiveEventsParams): Boolean; virtual;
    /// <summary>
    ///   The report engine calls it when DragOver operation is active over the
    ///   object with class that InPlace editor serves.
    /// </summary>
    /// <param name="Source">
    ///   Source object.
    /// </param>
    /// <param name="X">
    ///   Current mouse x coordinate.
    /// </param>
    /// <param name="Y">
    ///   Current mouse y coordinate.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with parameters of interactive events.
    /// </param>
    function DoCustomDragDrop(Source: TObject; X, Y: Integer; var EventParams: TfrxInteractiveEventsParams): Boolean; virtual;

    /// <summary>
    ///   The report engine calls it when DragOver operation is active over the
    ///   object with class that InPlace editor serves.
    /// </summary>
    /// <param name="X">
    ///   Current mouse x coordinate.
    /// </param>
    /// <param name="Y">
    ///   Current mouse y coordinate.
    /// </param>
    /// <param name="Shift">
    ///   Shift buttons states.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with parameters of interactive events.
    /// </param>
    function DoMouseMove(X, Y: Integer; Shift: TShiftState; var EventParams: TfrxInteractiveEventsParams): Boolean; virtual;
    /// <summary>
    ///   The report engine calls it when DragOver operation is active over the
    ///   object with class InPlace editor serves.
    /// </summary>
    /// <param name="X">
    ///   Current mouse x coordinate.
    /// </param>
    /// <param name="Y">
    ///   Current mouse y coordinate.
    /// </param>
    /// <param name="Button">
    ///   Mouse button.
    /// </param>
    /// <param name="Shift">
    ///   Shift buttons states.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with parameters of interactive events.
    /// </param>
    function DoMouseUp(X, Y: Integer; Button: TMouseButton;
      Shift: TShiftState; var EventParams: TfrxInteractiveEventsParams): Boolean; virtual;
    /// <summary>
    ///   The report engine calls it when DragOver operation is active over the
    ///   object with class that InPlace editor serves.
    /// </summary>
    /// <param name="X">
    ///   Current mouse x coordinate.
    /// </param>
    /// <param name="Y">
    ///   Current mouse y coordinate.
    /// </param>
    /// <param name="Button">
    ///   Mouse button.
    /// </param>
    /// <param name="Shift">
    ///   Shift buttons states.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with parameters of interactive events.
    /// </param>
    function DoMouseDown(X, Y: Integer; Button: TMouseButton;
      Shift: TShiftState; var EventParams: TfrxInteractiveEventsParams): Boolean; virtual;
    /// <summary>
    ///   The report engine calls it when DragOver operation is active over the
    ///   object with class that InPlace editor serves.
    /// </summary>
    /// <param name="Shift">
    ///   Shift buttons states.
    /// </param>
    /// <param name="WheelDelta">
    ///   Mouse wheel delta.
    /// </param>
    /// <param name="MousePos">
    ///   Cursor position when event were fired.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with parameters of interactive events.
    /// </param>
    function DoMouseWheel(Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var EventParams: TfrxInteractiveEventsParams): Boolean; virtual;

    /// <summary>
    ///   Initialize UI of in place editor when mouse cursor enters area which
    ///   inPlace editor serves.
    /// </summary>
    /// <param name="EventParams">
    ///   Structure with parameters of interactive events.
    /// </param>
    procedure InitializeUI(var EventParams: TfrxInteractiveEventsParams); virtual;
    /// <summary>
    ///   Finalize UI of in place editor when mouse cursor leaves area which
    ///   inPlace editor serves.
    /// </summary>
    /// <param name="EventParams">
    ///   Structure with parameters of interactive events.
    /// </param>
    procedure FinalizeUI(var EventParams: TfrxInteractiveEventsParams); virtual;

    /// <summary>
    ///   This method serves copy behaviour for the group of objects. The
    ///   object selection passed in EventParams.SelectionList parameter.
    /// </summary>
    /// <param name="CopyFrom">
    ///   Reference to an object used to copy from.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with parameters of interactive events.
    /// </param>
    /// <param name="Buffer">
    ///   Temporary buffer for copy.
    /// </param>
    /// <param name="CopyAs">
    ///   This flag determines kind of object data.
    /// </param>
    procedure CopyGoupContent(CopyFrom: TfrxComponent; var EventParams: TfrxInteractiveEventsParams; Buffer: TStream; CopyAs: TfrxCopyPasteType = cptDefault); virtual;
    /// <param name="PasteTo">
    ///   Reference to an object used to paste data.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with parameters of interactive events.
    /// </param>
    /// <param name="Buffer">
    ///   Temporary buffer for copy.
    /// </param>
    /// <param name="PasteAs">
    ///   This flag determines kind of object data.
    /// </param>
    procedure PasteGoupContent(PasteTo: TfrxComponent; var EventParams: TfrxInteractiveEventsParams; Buffer: TStream; PasteAs: TfrxCopyPasteType = cptDefault); virtual;

    /// <param name="CopyFrom">
    ///   Reference to an object used to copy from.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with parameters of interactive events.
    /// </param>
    /// <param name="Buffer">
    ///   Temporary buffer for copy.
    /// </param>
    /// <param name="CopyAs">
    ///   This flag determines kind of object data.
    /// </param>
    procedure CopyContent(CopyFrom: TfrxComponent; var EventParams: TfrxInteractiveEventsParams; Buffer: TStream; CopyAs: TfrxCopyPasteType = cptDefault); virtual;
    /// <param name="PasteTo">
    ///   Reference to an object used to paste data.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with parameters of interactive events.
    /// </param>
    /// <param name="Buffer">
    ///   Temporary buffer for copy.
    /// </param>
    /// <param name="PasteAs">
    ///   This flag determines kind of object data.
    /// </param>
    procedure PasteContent(PasteTo: TfrxComponent; var EventParams: TfrxInteractiveEventsParams; Buffer: TStream; PasteAs: TfrxCopyPasteType = cptDefault); virtual;
    /// <param name="PasteTo">
    ///   Reference to an object used to paste data.
    /// </param>
    /// <param name="EventParams">
    ///   Structure with parameters of interactive events.
    /// </param>
    /// <param name="PasteAs">
    ///   This flag determines kind of object data.
    /// </param>
    function IsPasteAvailable(PasteTo: TfrxComponent; var EventParams: TfrxInteractiveEventsParams; PasteAs: TfrxCopyPasteType = cptDefault): Boolean; virtual;
    /// <summary>
    ///   Returns default content type supported by InPlcae editor.
    /// </summary>
    function DefaultContentType: TfrxCopyPasteType; virtual;

    /// <summary>
    ///   Reference to edited component.
    /// </summary>
    property Component: TfrxComponent read FComponent write SetComponent;
    /// <summary>
    ///   Reference to a parent component.
    /// </summary>
    property ParentComponent: TComponent read FParentComponent write FParentComponent;
    /// <summary>
    ///   An event used in InPlace editors. Event fires when the report
    ///   designer or the report preview close InPlace editor.
    /// </summary>
    property OnFinishInPlace: TfrxOnFinishInPlaceEdit read FOnFinishInPlace write FOnFinishInPlace;
    /// <summary>
    ///   Editor is locked.
    /// </summary>
    property Locked: Boolean read FLocked;
    /// <summary>
    ///   Child editors for current editor class.
    /// </summary>
    property Editors: TfrxComponentEditorsList read FEditors write FEditors;
    /// <summary>
    ///   Reference to a clipboard object.
    /// </summary>
    property ClipboardObject: TPersistent read FClipboardObject write FClipboardObject;
  end;

  /// <summary>
  ///   Visibility of InPlace editors.
  /// </summary>
  TfrxComponentEditorVisibilityState = (
    /// <summary>
    ///   Visible only in the report preview.
    /// </summary>
    evPreview,
    /// <summary>
    ///   Visible only in the report designer.
    /// </summary>
    evDesigner);
  TfrxComponentEditorVisibility = set of TfrxComponentEditorVisibilityState;

  TfrxInPlaceEditorClass = class of TfrxInPlaceEditor;

  TfrxCustomIOTransportClass = class of TfrxCustomIOTransport;

  /// <summary>
  ///   This class represents registration item of InPlace editor in editors
  ///   list.
  /// </summary>
  TfrxComponentEditorsRegItem = class(TObject)
  private
    FEditorsClasses: TList;
    FEditorsVisibility: TList;
    FComponentClass: TfrxComponentClass;
    function GetEditorClass(Index: Integer): TfrxInPlaceEditorClass;
    procedure SetEditorClass(Index: Integer;
      const Value: TfrxInPlaceEditorClass);
    function GetComponentClassName: String;
    function GetEditorVisibility(Index: Integer): TfrxComponentEditorVisibility;
    procedure SetEditorVisibility(Index: Integer;
      const Value: TfrxComponentEditorVisibility);
  public
    constructor Create;
    destructor Destroy; override;
    function Count: Integer;
    procedure RemoveEditor(AClass: TfrxInPlaceEditorClass);
    property EditorClass[Index: Integer]: TfrxInPlaceEditorClass read GetEditorClass write SetEditorClass;
    property EditorVisibility[Index: Integer]: TfrxComponentEditorVisibility read GetEditorVisibility write SetEditorVisibility;
    property ComponentClassName: String read GetComponentClassName;
  end;

  TfrxRectArray = array of TRect;
  PfrxRectArray = ^TfrxRectArray;

  /// <summary>
  ///   This class controls group of InPlace editors registered for specified
  ///   object type. Methods description are the same as for TfrxInPlaceEditor
  ///   class.
  /// </summary>
  TfrxComponentEditorsManager = class(TObject)
  private
    FOffsetX: Extended;
    FOffsetY: Extended;
    FScale: Extended;
    FEditorsGlasses: TList;
    FComponentClass: TfrxComponentClass;
    function GetComponent: TfrxComponent;
    procedure SetComponent(const Value: TfrxComponent);
  public
    function EditorsActiveRects(aComponent: TfrxComponent): TfrxRectArray;
    procedure DrawCustomEditor(aCanvas: TCanvas; aRect: TRect);
    function DoCustomDragOver(Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean; var EventParams: TfrxInteractiveEventsParams): Boolean;
    function DoCustomDragDrop(Source: TObject; X, Y: Integer; var EventParams: TfrxInteractiveEventsParams): Boolean;
    procedure SetOffset(aOffsetX, aOffsetY, aScale: Extended);
    function DoMouseMove(X, Y: Integer; Shift: TShiftState; var EventParams: TfrxInteractiveEventsParams): Boolean;
    function DoMouseUp(X, Y: Integer; Button: TMouseButton;
      Shift: TShiftState; var EventParams: TfrxInteractiveEventsParams): Boolean;
    function DoMouseDown(X, Y: Integer; Button: TMouseButton;
      Shift: TShiftState; var EventParams: TfrxInteractiveEventsParams): Boolean;
    function DoMouseWheel(Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var EventParams: TfrxInteractiveEventsParams): Boolean;

    procedure InitializeUI(var EventParams: TfrxInteractiveEventsParams);
    procedure FinalizeUI(var EventParams: TfrxInteractiveEventsParams);

    procedure CopyContent(CopyFrom: TfrxComponent; var EventParams: TfrxInteractiveEventsParams; Buffer: TStream; CopyAs: TfrxCopyPasteType = cptDefault);
    procedure PasteContent(PasteTo: TfrxComponent; var EventParams: TfrxInteractiveEventsParams; Buffer: TStream; PasteAs: TfrxCopyPasteType = cptDefault);
    function IsPasteAvailable(PasteTo: TfrxComponent; var EventParams: TfrxInteractiveEventsParams; PasteAs: TfrxCopyPasteType = cptDefault): Boolean;
    function IsContain(ObjectRect: TRect; X, Y: Extended): Boolean;
    procedure SetClipboardObject(const aObject: TPersistent);
    constructor Create;
    destructor Destroy; override;
    property Component: TfrxComponent read GetComponent write SetComponent;
  end;

  /// <summary>
  ///   This class used to store component InPlace editors for current instance
  ///   of TfrxComponent.
  /// </summary>
  TfrxComponentEditorsList = class(TStringList)
  private
    FEditorsInstances: TList;
    function GetEditors(Index: Integer): TfrxComponentEditorsManager;
    procedure PutEditors(Index: Integer;
      const Value: TfrxComponentEditorsManager);
  public
    constructor Create; overload;
    destructor Destroy; override;
    procedure CreateInstanceFromItem(aItem: TfrxComponentEditorsRegItem; aOwner: TWinControl; aVisibility: TfrxComponentEditorVisibilityState);
    property Editors[Index: Integer]: TfrxComponentEditorsManager read GetEditors write PutEditors;
  end;

  /// <summary>
  ///   This class represents collection of registered component InPlace
  ///   editors for each type of components.
  /// </summary>
  TfrxComponentEditorsClasses = class(TObject)
  private
    FList: TStringList;
    function GetEditors(Index: Integer): TfrxComponentEditorsRegItem;
    procedure PutEditors(Index: Integer;
      const Value: TfrxComponentEditorsRegItem);
  public
    constructor Create;
    destructor Destroy; override;
    /// <summary>
    ///   Creates new instances of InPlace editors specified in the collection.
    /// </summary>
    /// <param name="aVisibility">
    ///   Visibility of an InPlcae editor.
    /// </param>
    /// <param name="aOwner">
    ///   Owner of editors controls.
    /// </param>
    function CreateEditorsInstances(aVisibility: TfrxComponentEditorVisibilityState; aOwner: TWinControl): TfrxComponentEditorsList;
    /// <summary>
    ///   Count of items.
    /// </summary>
    function Count: Integer;
    /// <summary>
    ///   Returns item index by it's name.
    /// </summary>
    /// <param name="s">
    ///   Item name.
    /// </param>
    function IndexOf(const s: String): Integer;
    /// <summary>
    ///   Registers new inPlcae editor class for specified component class.
    /// </summary>
    /// <param name="ComponentClass">
    ///   Class of a component.
    /// </param>
    /// <param name="ComponentEditors">
    ///   An open array of InPlace editor classes.
    /// </param>
    /// <param name="EditorsVisibility">
    ///   An open array of visibility for each InPlace editor specified in
    ///   ComponentEditors <br />parameter.
    /// </param>
    procedure Register(ComponentClass: TfrxComponentClass; ComponentEditors: array of TfrxInPlaceEditorClass; EditorsVisibility: array of TfrxComponentEditorVisibility);
    /// <summary>
    ///   Removes all registered editors for specified component class.
    /// </summary>
    /// <param name="ComponentClass">
    ///   Class of a component.
    /// </param>
    procedure UnRegister(ComponentClass: TfrxComponentClass); overload;
    procedure UnRegister(ComponentClass: TfrxComponentClass; EditroClass: TfrxInPlaceEditorClass); overload;
    /// <summary>
    ///   Removes all registered editors for specified editor class.
    /// </summary>
    /// <param name="EditroClass">
    ///   lass of an editor.
    /// </param>
    procedure UnRegisterEditor(EditroClass: TfrxInPlaceEditorClass);
    /// <summary>
    ///   Saves settings of editors visibility to Ini file.
    /// </summary>
    /// <param name="IniFile">
    ///   Reference to ini file. <br />
    /// </param>
    procedure SaveToIni(IniFile: TCustomIniFile);
    /// <summary>
    ///   Loads settings of editors visibility from Ini file.
    /// </summary>
    /// <param name="IniFile">
    ///   Reference to ini file.
    /// </param>
    procedure LoadFromIni(IniFile: TCustomIniFile);
    /// <summary>
    ///   List of registered items.
    /// </summary>
    property EditorsClasses[Index: Integer]: TfrxComponentEditorsRegItem read GetEditors write PutEditors;
  end;

  TfrxExtPropList = class(TObject)
    FPropInfos: TStringList;
  private
    function GetGetMethod(Index: Integer): Pointer;
    function GetName(Index: Integer): String;
    function GetSetMethod(Index: Integer): Pointer;
    function GetTypInfo(Index: Integer): Pointer;
    procedure SetGetMethod(Index: Integer; const Value: Pointer);
    procedure SetName(Index: Integer; const Value: String);
    procedure SetSetMethod(Index: Integer; const Value: Pointer);
    procedure SetTypInfo(Index: Integer; const Value: Pointer);
    procedure ClearPropInfo;
  public
    constructor Create;
    destructor Destroy; override;
    procedure AddProperty(aName: ShortString; pTypInf, StaticGetMethod, StaticSetMethod: Pointer);
    function Count: Integer;
    property TypInfo[Index: Integer]: Pointer read GetTypInfo write SetTypInfo;
    property Name[Index: Integer]: String read GetName write SetName;
    property GetMethod[Index: Integer]: Pointer read GetGetMethod write SetGetMethod;
    property SetMethod[Index: Integer]: Pointer read GetSetMethod write SetSetMethod;
  end;

  TfrxCustomObjectPreset = class(TPersistent)
  public
    constructor Create; virtual;
    function GetData(aReport: TfrxReport): String; virtual; abstract;
    procedure ApplySettings(aComponent: TfrxComponent); virtual; abstract;
    procedure SaveComponentState(aComponent: TfrxComponent); virtual; abstract;
    procedure RestoreComponentState(aComponent: TfrxComponent); virtual; abstract;
    procedure BeginDraw(Canvas: TCanvas; ScaleX, ScaleY: Extended; Area: TRect); virtual;
    procedure EndDraw(Canvas: TCanvas; ScaleX, ScaleY: Extended; Area: TRect); virtual;
  end;
  TfrxCustomObjectPresetClass = class of TfrxCustomObjectPreset;

/// <summary>
///   Function returns a global form that may be useful for some add-in
///   controls.
/// </summary>
function frxParentForm: TForm;
/// <summary>
///   This function searches for a DataSet on specified owner.
/// </summary>
function frxFindDataSet(DataSet: TfrxDataSet; const Name: String;
  Owner: TComponent): TfrxDataSet;
procedure frxGetDataSetList(List: TStrings);
function frxCreateFill(const Value: TfrxFillType): TfrxCustomFill;
function frxGetFillType(const Value: TfrxCustomFill): TfrxFillType;
function frxRegEditorsClasses: TfrxComponentEditorsClasses;
procedure frxUnregisterEditorsClass(ComponentClass: TfrxComponentClass; EditroClass: TfrxInPlaceEditorClass);
function frxGetInPlaceEditor(aList: TfrxComponentEditorsList; aComponent: TfrxComponent): TfrxComponentEditorsManager;

var
  frxDefaultIOTransportClass: TfrxCustomIOTransportClass = nil;
  frxDefaultIODialogTransportClass: TfrxCustomIOTransportClass = nil;
  frxDefaultTempFilterClass: TfrxCustomIOTransportClass = nil;
  frxDesignerClass: TfrxDesignerClass;
  frxDotMatrixExport: TfrxCustomExportFilter;
  frxCompressorClass: TfrxCompressorClass;
  frxCrypterClass: TfrxCrypterClass;
  frxCharset: Integer = DEFAULT_CHARSET;
  frxFR2Events: TfrxFR2Events;
{$IFNDEF NO_CRITICAL_SECTION}
  frxCS: TCriticalSection;
{$ENDIF}
  /// <summary>
  ///   The global variables list available to all reports.
  /// </summary>
  frxGlobalVariables: TfrxVariables;

const
  BND_COUNT = 18;
  frxBands: array[0..BND_COUNT - 1] of TfrxComponentClass =
    (TfrxReportTitle, TfrxReportSummary, TfrxPageHeader, TfrxPageFooter,
     TfrxHeader, TfrxFooter, TfrxMasterData, TfrxDetailData, TfrxSubdetailData,
     TfrxDataBand4, TfrxDataBand5, TfrxDataBand6, TfrxGroupHeader, TfrxGroupFooter,
     TfrxChild, TfrxColumnHeader, TfrxColumnFooter, TfrxOverlay);

{$IFDEF FPC}
//  procedure Register;
{$ENDIF}

implementation
end.
